/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * barrier.c: barrier specific protocol module for Midway.
 * 
 * [some stuff must be in lock.c -- the allocation & transfer stuff 
 * in particular... allocation 'cause of statics in lock.c (so poor excuse)
 * transfer 'cause I'm sharing the global "lock" (really synchro variable)
 * array, and might as well do it all in one place (better excuse).]
 *
 * 16 Jul 91, Matthew J. Zekauskas
 * 30 Jul 92 (mjz): finally put debugging printing under conditionals
 * 11 Aug 92 (was): added clock_t type.
 * 13 Aug 92 (was): changed parms to mid_addrmap_xxx
 * 28 Sep 92 (mjz): inc was fixes for no dirtybits (i.e. writing private)
 *  3 Oct 92 (mjz): move to new naming scheme
 * 15 Oct 92 (was): changed MAX_DATA_SIZE to mid_max_data_size
 * 19 Oct 92 (mjz): initial multipacket barrier request work
 * 26 Oct 92 (mjz): flush barriers
 * 13 Jan 93 (was): use macro to check if update queue is empty
 * 13 Mar 93 (was): added verification of user parms
 * 21 Jun 93 (mjz): switch to standard pack/unpack.  As a conseqence,
 *		we will no longer have coherence between times on all
 *		memory for a particular barrier operation (that is,
 *		times for a given data item should be the same, but 
 *		not all data items protected by a barrier will be the same).
 * 21 Jul 93 (was): changed parms to (and name of) mid_addrmap_delta_pack
 *  8 Jan 94 (was): added fastpath for acquire/release locks (FASTLOCKS)
 *                  locktable moved to a fixed/known address in memory
 * 23 Feb 94 (was): added support for page-based version of midway
 *  4 Mar 94 (was): added stats, changed parms to mid_extq_xxx
 *  4 May 94 (was): changed parms to CommReply
 */
 /* #include "midwayos.h" -- someday *//* get other pesky (std) include files */
#include <sys/types.h>		/* u_long & friends */
#include <mach.h>
#include <cthreads.h>
#include <errno.h>
#include <stdio.h>

#include "midway_types.h"       /* global type definitions */
#include "vm.h"			/* vm region definitions */
#include "lock.h"		/* perlock struct, lock request formats */
#include "comm.h"		/* M_xxx_OP */
#include "smexecd_types.h"	/* ensure get char255 */
#include "pbuf.h"		/* packet buffer stuff */
#include "updateq.h"		/* prototypes for calling update queue rtns */
#include "midway-dbg.h"		/* debugging levels */
#ifndef shared			/* allow us to be compiled w/o midway compiler */
#define shared
#endif
#include "midway.h"		/* prototypes */

/* where our thread id is kept.  Gosh... initialization????? *MJZ* */
#define OUR_TID ((u_long) cthread_data(cthread_self()))

#ifdef PAGE_BASED
   /* 
    * Page-based version cannot allow updates to memory to occur
    * while the application is executing. If the application were
    * to write to a clean page that we were in the middle of updating
    * (ie had mapped read-write), we would lose the write, leaving
    * the application with inconsistent memory.    WAS 8Mar94
    */
#  undef FLUSH_NOCOPY
#else
#  define FLUSH_NOCOPY	1	/* on flush, write to memory immediately */
#endif



/*
 * Global references
 */
extern int Comm_hnum;		/* number of hosts participating */
extern int Comm_index;		/* number of this host (for init) */
extern clock_t LocalClock;	/* to date lock structures...  */
extern u_long cpt_init();	/* for init too... BAD STRUCTURE */

/*
 * forward declarations
 */
static void mid_bar_sendpackets();
static void mid_bar_applychgs();


#if defined(STATS)

static struct {
  u_long	alloced;	/* # of barriers allocated */
  u_long	crossed;	/* # of barriers crossed */
  u_long	cross_data;  	/* total data bound to bars x crossing */
  u_long	flushed;	/* # of barriers flushed */
  u_long	flush_data;  	/* total data bound to bars x flushing */
  u_long        data_sent;	/* actual data sent */
  u_long        data_released;	/* actual data released (ie returned by master) */
} BarStats;
static mutex_t statLock;		/* just in case */

/* 
 * Define macro to conditionally maintain stat info
 */
# ifdef FAST_STATS
#  define INCR_STATS(x,by) (x) += (by)
# else
#  define INCR_STATS(x,by) \
    { mutex_lock(statLock);  (x) += (by);  mutex_unlock(statLock); }
# endif 

#else /* not STATS */

#define INCR_STATS(x,by)   /* nothing to do  */

#endif


/*
 ************************************************************************
 *		m i d _ b a r r i e r _ i n i t   			*
 *----------------------------------------------------------------------*
 * Initialize barrier code.						*
 ************************************************************************
 */
void
mid_barrier_init()
{
#ifdef STATS
  statLock = mutex_alloc ();
  MIDWAY_MUTEX_SET_NAME(statLock,"barrier.c:statLock");
  mid_barrier_reset_stats();
#endif
}


/*
 ************************************************************************
 *		m i d w a y _ b a r r i e r _ a l l o c			*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
midway_barrier_t
midway_barrier_alloc()
{
  pl_t *lock;

  /* allocate & initialize a barrier structure */
  lock = mid_alloc_synch_object (NULL, (L_BARRIER | L_BAR_NOREG));

  INCR_STATS(BarStats.alloced, 1);

  return((midway_barrier_t) lock);
} /* midway_barrier_alloc */

/*
 ************************************************************************
 *		  m i d _ r e l e a s e _ b a r r i e r			*
 *----------------------------------------------------------------------*
 * On the manager processor, perform all that needs to be done once all *
 * threads have hit the barrier.					*
 ************************************************************************
 */
typedef struct {u_long thread; u_long tid; } ph_t; 
static ph_t *participating_hosts = 0; /* HACK to prevent using up threads.  */
static int participating_hosts_l = 0;

static
mid_release_barrier( bar, hst )
     pl_t *bar;			/* the barrier info */
     u_long hst;		/* thread doing release: 0=master */
				/* x=> barrier hit thread from elsewhere */
{
  u_long	i;
  clock_t	now;		/* time now  */
  pbuf	pb;
  u_long pblen;			/* length of stuff in current packet buf */
  m_bar_req_t	*obr;
  s_addr_map_t	*amap;		/* amap for delta_pack */
  clock_t	maxtime;	/* max time for data to transmit */
  char          *amap_addr;	/* parm to delta_pack */
  long          amap_len;
  char          *data_addr;
  long          data_len;
  clock_t       data_clock;
  long		xferlen;
  u_long s_host, s_thread, s_tid; /* for popping waiter q */
  clock_t s_clock;
  clock_t s_incarn;
  pbufe_t *e;			/* for applying changes locally */
  char255 smsg;

 /*
  * it goes like this:
  * if busy > 0
  *	do deltamap of entries on this machine; add to output buffer
  * move output buffer list to old output buffer list
  * move bar_curtime to bar_lasttime
  * set new bar_curtime
  * send [old] output buffer to all machines on output waiter q
  * apply changes to local machine
  *	[probably need to apply *all* and in order, so that all machines
  *	 have consistent view...  could optimize by having machine tag
  *	 on updates... and skip those for your machine.  Hrrumph.]
  * reset busy, bar_subtotoal, clear output waiter q
  *  ==> we lock out new people that hit barrier until we are done
  * condition_broadcast to bar_wait.
  */

#ifdef DEBUG
  if (IS_DBG_BARR_EXTR) {
    sprintf(smsg, "mid_release_barrier: busy = %u", bar->busy);/**/
    StatusMsg(smsg);
  }
#endif /* DEBUG */

  ASSERT (bar->state & L_BARRIER);
  ASSERT (bar->held == MIDWAY_LOCKHELD_OTHER);

  /*
   * now must be set here, because it is used to set the barrier
   * "curtime" even if the barrier is not in use locally or this
   * is a flush barrier.  Duh.  28Apr93mjz
   */
  now = LocalClock;		/* save in case LocalClock changes... */

  if (bar->busy && !(bar->state & L_BAR_FLUSH)) {
    /* start dealing with data barrier protects... */
    /* the following should get only those things with 0 timestamps */
    xferlen = 0;
    amap = bar->amap;
    if (amap) {			/* then there REALLY is stuff to append */
      pb = mid_pbuf_get();
      ASSERT (pb != NULL);
      obr = (m_bar_req_t *) pb;
      obr->curtime = now;
      obr->num = 0;		/* ??? */
      obr->totnum = 0;		/* nothing in buf yet */
      pblen = sizeof(m_bar_req_t);
      maxtime = NULL_CLOCK_VALUE; amap_addr = NULL; amap_len = 0; 
      data_addr = NULL; data_len = 0; data_clock = NULL_CLOCK_VALUE;

#ifdef PAGE_BASED
      while(mid_paged_pack_data((u_long)bar,
				PAGE_PACK_DIFFS, 
				TRUE, /* mark data as clean */
				&amap, &amap_addr, &amap_len,
				now-1, 
				(char *)(obr + 1), &pblen,
				&xferlen)) {
#else
      while(mid_pack_data(&amap, &amap_addr, &amap_len,
			  now-1, &maxtime, 
			  (char *)(obr + 1), &pblen,
			  &data_addr, &data_len, &data_clock,
			  &xferlen)) {
#endif
	/* IF HERE, THEN THERE ARE CHANGES */
	INCR_STATS(BarStats.data_released, xferlen);
	obr->totnum = pblen;	/* save for output routine */
	mid_pbuf_enqueue(pb, &(bar->output));
	pb = mid_pbuf_get();
	ASSERT (pb != NULL);
	obr = (m_bar_req_t *) pb;
	obr->curtime = now;
	obr->num = 0;
	obr->totnum = 0;		/* nothing in buf yet */
	pblen = sizeof(m_bar_req_t);
      } /* while buffer filled up */
      INCR_STATS(BarStats.data_released, xferlen);

      if (pblen == sizeof(m_bar_req_t)) { /* no changes to "final" buffer */
	mid_pbuf_free(pb);
      } else {
	obr->totnum = pblen;	/* save for output routine */
	mid_pbuf_enqueue(pb, &(bar->output));
      }
    } /* if amap */

  } /* was busy -- collect stuff from this machine */

  bar->bar_lasttime = bar->bar_curtime;	/* done with this iteration */
  mid_pbuf_qmove(&(bar->output), &(bar->old_output));
  bar->bar_curtime = now;	/* a new time to aspire to */

  /*
   * HACK ALERT HACK ALERT HACK ALERT HACK ALERT HACK ALERT HACK ALERT
   * I used to just set the transaction state to done here.  However,
   * this is not what we want with the current system.  Why?  Because
   * say we have a few hosts (>4) participating on a barrier that backs
   * a ton o' data.  We release a few hosts, but they don't get all the
   * data for the reply.  They retransmit.  They are "DONE" so the
   * retransmission makes it back to this module, where they block on
   * the mutex that protects this barriers information.  This ties up
   * that worker thread.  Eventually, all workers are stuck, and there
   * is no one to get the ack back from a multi-packet reply.  The system
   * deadlocks.  There is much sadness.--15apr93mjz
   *
   * GAK.  The same thing happens for new requests.  We ought to be
   * able to deal with new requests, though...  so I am going to
   * try and unlock the barrier lock while we are sending stuff out.
   */
    if (participating_hosts) {
      bzero((char *)participating_hosts, participating_hosts_l);
    } else {
      participating_hosts_l = (Comm_hnum+1) * sizeof(ph_t);
      participating_hosts = (ph_t *) CALLOC(Comm_hnum+1, sizeof(ph_t));
    }
      

  /* send to all waiters *//*NOW NOTING ONLY 15apr93mjz*/
  while(!(mid_extq_pop(&(bar->bar_waiters), &s_host, &s_thread, &s_tid, &s_clock, &s_incarn))) {
#ifdef OLD_WAY
    /* got something (returns non-zero on empty) */
    mid_bar_sendpackets(bar, s_host, s_thread, s_tid);
    if (hst != s_host) {	/* really q'd */
      CommSetTranState(s_host,s_thread,s_tid,M_R_DONE); /* done this one */
      /* if hst non zero, then that thread will set MRDONE when finished. */
    }
#else
      participating_hosts[s_host].thread = s_thread;
      participating_hosts[s_host].tid = s_tid;
#endif
  }

  /* apply locally */
  e = MID_PBUF_QTOP(&(bar->old_output));
  while (e != 0) {
    obr = (m_bar_req_t *) Q2PBUF(e);
    /* need this if we're running in flush mode */
    obr->curtime = bar->bar_curtime;
    mid_bar_applychgs(bar, (char *) obr, obr->totnum);
    e = MID_PBUF_QNEXT(e);
  }

  bar->busy = 0;		/* done */
  bar->bar_subtotal = 0;	/* no one hit yet */
  mutex_unlock(&(bar->m));	/* try letting them go 15apr93mjz */

  /* SEND TO REMOTE HOSTS */
  for (i=2; i<=Comm_hnum; i++) {
    if (participating_hosts[i].thread)
      mid_bar_sendpackets(bar, i, participating_hosts[i].thread, 
			  participating_hosts[i].tid);
  }
    
  mutex_lock(&(bar->m));	/* reacquire */

  /*
   * HACK CONTINUES
   * now release the non-zero participating hosts
   */
  for (i=2/*ASSUMES MASTER IS 1*/; i<=Comm_hnum; i++)
    if ((i != hst) && (participating_hosts[i].thread))
      CommSetTranState(i, participating_hosts[i].thread,
		       participating_hosts[i].tid, M_R_DONE);
  
    
  condition_broadcast(&(bar->bar_wait)); /* let local people go */
} /* mid_release_barrier */


/*
 ************************************************************************
 *		m i d w a y _ c r o s s _ b a r r i e r			*
 *----------------------------------------------------------------------*
 * User level call -- you've hit the barrier.				*
 ************************************************************************
 */
void
mid_cross_barrier( bar, num )
     pl_t *bar;		/* barrier name */
     int num;			/* # procs you expect to hit barrier */
{
  u_long i;
  char255 smsg;

  /* assume already locked, and verified that we're in a barrier */
  ASSERT (bar->state & L_BARRIER);
  ASSERT (bar->held == MIDWAY_LOCKHELD_OTHER);

  /* see if on master, if so do things differently */
  if (Comm_index == bar->bestguess) { /* thread on manager */
    /* new iteration stuff.  This MUST be a new or current iteration,
       since we do not use the communication system so there can be
       no retransmissions or old packets. */
    if (bar->bar_subtotal == 0) { /* new iteration */
      bar->bar_total = num;	/* reset total */
    } /* new iteration */
    else {			/* already done some work on this iteration */
      if (num != bar->bar_total) { /* inconsistency */
	sprintf(smsg, "Barrier %u: num (%u) != existing total (%u) -- barfing.",
		bar->lock, num, bar->bar_total);
	StatusMsg(smsg);
	ASSERT(num == bar->bar_total);
      }
    } /* "old" iteration */
    bar->busy++;		/* note we hit barrier locally */
    bar->bar_subtotal++;	/* and bump total # of threads to hit */
    if (bar->bar_subtotal == bar->bar_total) { /* time to make the donuts */
      mid_release_barrier(bar,0/*master*/);
    }
    else if(bar->bar_subtotal < bar->bar_total) { /* not enough yet */
      condition_wait(&(bar->bar_wait), &(bar->m));
    }
    else { /* bar_subtotal > bar_total -- oops! */
      sprintf(smsg, "Barrier %u: bar_subtotal (%u) > bar_total (%u) -- barfing.",
	      bar->lock, bar->bar_subtotal, bar->bar_total);
      StatusMsg(smsg);
      botch("subtotal too high");
    }
  } /* thread on manager */

  else {			/* thread on remote machine */
    bar->busy++;		/* note another thread hit barrier */
    /* reset total number to hit barrier.  Note: we should probably just
       do this for the first thread to hit (busy == 1), and then
       complain if not the same for others...  For now, just punt.
       (used by midSendBarReq) 31jul91mjz */
    bar->bar_total = num;
    if (bar->busy == bar->bar_num) { /* got enough to send packet to master */
      bar->state |= L_ACQ;	/* note in the process of getting from rem */
      midSendBarReq(bar);	/* send off request */
      /* now memory has been updated, as has bar_curtime */
      /* (unlock lock? or just let replies that make it update bar_curtime..)*/
      /* gotta unlock, otherwise RegisterBarrierUse race condition will cause
	 deadlock, instead of being benign */
      bar->state &= L_XACQ;	/* turn it off */
      bar->busy = 0;		/* reset busy */
      condition_broadcast(&(bar->bar_wait)); /* let any other waiters go */
    }
    else if (bar->busy < bar->bar_num) { /* not enough -- just wait */
      condition_wait(&(bar->bar_wait), &(bar->m));
    }
    else { /* busy > bar->bar_num -- oops! */
      sprintf(smsg, "Barrier %u: busy (%u) > bar_num (%u) -- barfing.",
	      bar->lock, bar->busy, bar->bar_num);
      StatusMsg(smsg);
      StatusMsg("Use RegisterBarrierUse(<barrier>) in all threads!");
      ASSERT(1 == 0);
    }
  } /* thread on remote machine */

  mutex_unlock(&(bar->m));
  return;
} /* mid_cross_barrier [INTERNAL]*/


/*
 ************************************************************************
 *		m i d w a y _ c r o s s _ b a r r i e r			*
 *----------------------------------------------------------------------*
 * User level call -- you've hit the barrier.				*
 ************************************************************************
 */
void
midway_cross_barrier( bname, num )
     midway_barrier_t bname;	/* barrier name */
     int num;			/* # procs you expect to hit barrier */
{
  pl_t *bar;
  char255 smsg;
  int rc;

  /* convert barname to a pointer to the barrier structure */
  bar = (pl_t *) bname;


#ifdef DEBUG
  if (IS_DBG_BARR_OPS) {
    sprintf(smsg, "Cross_Barrier(%u, %u)", bname, num);
    StatusMsg(smsg);
  }
#endif /* DEBUG */

#ifdef DEBUG_USER
  if (((vm_offset_t) bname < vm_locktable_base) || 
      ((vm_offset_t) bname >= vm_locktable_end) ||
      ((vm_offset_t) bname != (vm_offset_t) bar->lock)) {
    sprintf(smsg, "midway_cross_barrier: Invalid barrier id %lu", bname);
    StatusMsg(smsg);
#if STOP_ON_ERROR
    botch("midway_cross_barrier");
#else
    return;
#endif
  }
  if ((bar->state & L_BARRIER) == 0) { /* not a barrier */
    sprintf(smsg, "midway_cross_barrier: %lu is not a barrier", bname);
    StatusMsg (smsg);
#if STOP_ON_ERROR
    botch("midway_cross_barrier");
#else
    return; 
#endif
 }
  if (num > Comm_hnum) {
    sprintf(smsg, "midway_cross_barrier: count %d > # of procs %d",
	    num, Comm_hnum);
    StatusMsg(smsg); 
#if STOP_ON_ERROR
    botch("midway_cross_barrier");
#else
   return;
#endif
  }
#endif

  ASSERT (bar->held == MIDWAY_LOCKHELD_OTHER);

  INCR_STATS(BarStats.crossed, 1);
  INCR_STATS(BarStats.cross_data,bar->amap_data_size);

  /*
   * Flush update queue, before acquiring a global lock 
   */
  if (UPDATEQ_NOT_EMPTY) {
    if (IS_DBG_PORT_QUEUE) StatusMsg("Barrier flushing update queue");
    rc = mid_updateq_flush (UPDATEQ_CALLER_BARRIER);
  }
#ifdef PAGE_BASED_RC
  /* Flush update queue for page-based code */
  paged_flush_updates(bar);
#endif

  mutex_lock(&(bar->m));	/* so it is safe to twiddle with DS */
  /*
   * what if multiple threads on one processor hit the barrier at once?
   * then brian's thought about taking advantage of barrier processing
   * is not necessarily valid. [Know values won't change until you hit
   * barrier.]  What about multiple people banging on the dirty bits?
   */

#ifdef DEBUG
  if (IS_DBG_BARR_OPS) {
    sprintf(smsg, "Cross_Barrier(%u, %u) [to master %u]", bname, num,
	    bar->bestguess);
    StatusMsg(smsg);
  }
/*  mid_addrmap_prt(bar->lock, bar->amap); */
#endif
  mid_cross_barrier(bar, num);

} /* midway_cross_barrier */


/*
 ************************************************************************
 *		m i d w a y _ f l u s h _ b a r r i e r			*
 *----------------------------------------------------------------------*
 * User level call -- you've hit the barrier, but don't want data back.	*
 ************************************************************************
 */
void
midway_flush_barrier( bname, num )
     midway_barrier_t bname;	/* barrier name */
     int num;			/* # procs you expect to hit barrier */
{
  pl_t *bar;
  int rc;
  char255 smsg;

  /* convert barname to a pointer to the barrier structure */
  bar = (pl_t *) bname;


#ifdef DEBUG
  if (IS_DBG_BARR_OPS) {
    sprintf(smsg, "Flush_Barrier(%u, %u)", bname, num);
    StatusMsg(smsg);
  }
#endif /* DEBUG */

#ifdef DEBUG_USER
  if (((vm_offset_t) bname < vm_locktable_base) || 
      ((vm_offset_t) bname >= vm_locktable_end) ||
      ((vm_offset_t) bname != (vm_offset_t) bar->lock)) {
    sprintf(smsg, "midway_flush_barrier: Invalid barrier id %lu", bname);
    StatusMsg(smsg);
#if STOP_ON_ERROR
    botch("midway_flush_barrier");
#else
    return;
#endif
  }
  if ((bar->state & L_BARRIER) == 0) { /* not a barrier */
    sprintf(smsg, "midway_flush_barrier: %lu is not a barrier", bname);
    StatusMsg (smsg);
#if STOP_ON_ERROR
    botch("midway_flush_barrier");
#else
    return;
#endif
  }
  if (num > Comm_hnum) {
    sprintf(smsg, "midway_flush_barrier: count %d > # of procs %d",
	    num, Comm_hnum);
    StatusMsg(smsg);
#if STOP_ON_ERROR
    botch("midway_flush_barrier");
#else
    return;
#endif
  }
#endif

  INCR_STATS(BarStats.flushed, 1);
  INCR_STATS(BarStats.flush_data,bar->amap_data_size);

  ASSERT (bar->held == MIDWAY_LOCKHELD_OTHER);

  /*
   * Flush update queue, before acquiring a global lock 
   */
  if (UPDATEQ_NOT_EMPTY) {
    if (IS_DBG_PORT_QUEUE) StatusMsg("Barrier flushing update queue");
    rc = mid_updateq_flush (UPDATEQ_CALLER_BARRIER);
  }
#ifdef PAGE_BASED_RC
  /* Flush update queue for page-based code */
  paged_flush_updates(bar);
#endif

  mutex_lock(&(bar->m));	/* so it is safe to twiddle with DS */
  /*
   * what if multiple threads on one processor hit the barrier at once?
   * then brian's thought about taking advantage of barrier processing
   * is not necessarily valid. [Know values won't change until you hit
   * barrier.]  What about multiple people banging on the dirty bits?
   */
#ifdef DEBUG
  if (IS_DBG_BARR_OPS) {
    sprintf(smsg, "FlushBarrier(%u, %u) [to master %u]", bname, num,
	    bar->bestguess);
    StatusMsg(smsg);
  }
/*  mid_addrmap_prt(bar->lock, bar->amap); */
#endif
  bar->state |= L_BAR_FLUSH;	/* note we will only flush */

  mid_cross_barrier(bar, num);
} /* midway_flush_barrier */

/*
 ************************************************************************
 *		     m i d S e n d B a r R e q				*
 *----------------------------------------------------------------------*
 * actually send off a barrier request to the barrier manager.  We scan *
 * for dirty "lines", package them all up, and send off.  Done here     *
 * instead of in "Barrier" to keep Barrier control flow looking clean.  *
 ************************************************************************
 */

#define MORE 0
#define LAST 1
midSendBarReq(bar)
     pl_t *bar;
{
  m_bar_req_t *br;
  pbuf pb;
  u_long	i,j,mask;
  int		number;
  clock_t	now;		/* time now  */
  u_long	packnum;	/* packet number */
  u_long	pblen;
  char255	smsg;
  u_long	tid = 0;	/* Transaction ID; set/used by CommXRequest */
  s_addr_map_t	*amap;		/* amap for delta_pack */
  clock_t	maxtime;	/* max time for data to transmit */
  char          *amap_addr;	/* parm to delta_pack */
  long          amap_len;
  char          *data_addr;
  long          data_len;
  clock_t       data_clock;
  long		xferlen;

  pb = mid_pbuf_get();		/* grab one */
  ASSERT (pb != NULL);
  packnum = 0;			/* first packet is 0 */

  /* now add lock reply data */
  br = (m_bar_req_t *) pb;
  br->curtime = bar->bar_curtime; /* this iteration's value */
  br->num = bar->bar_num;
  br->totnum = bar->bar_total;
  br->noreply = bar->state & L_BAR_FLUSH;

  pblen = sizeof(m_bar_req_t);

  /* start dealing with data barrier protects... */
  now = LocalClock;		/* save in case LocalClock changes... */
  amap = bar->amap;
  if (amap) {
    xferlen = 0;
    maxtime = NULL_CLOCK_VALUE; amap_addr = NULL; amap_len = 0; 
    data_addr = NULL; data_len = 0; data_clock = NULL_CLOCK_VALUE;
    
    /* note: now-1 -> now should just get 0 (dirty) entries */
#ifdef PAGE_BASED
    while(mid_paged_pack_data((u_long)bar,
			      PAGE_PACK_DIFFS,
			      TRUE, /* mark data as clean */
			      &amap, &amap_addr, &amap_len,
			      now-1, 
			      (char *)(br+1), &pblen,
			      &xferlen)) {
#else
    while(mid_pack_data(&amap, &amap_addr, &amap_len,
			now-1, &maxtime, 
			(char *)(br+1), &pblen,
			&data_addr, &data_len, &data_clock,
			&xferlen)) {
#endif
      /* returned 1, so filled current packet, and needs another */
      ASSERT (xferlen > 0);
      INCR_STATS(BarStats.data_sent, xferlen);

      CommXRequest(bar->lock, M_BARRIER_OP, bar->bestguess, pb, pblen,
		   packnum, MORE, &tid);

      /* CommXRequest holds onto packet, and will toss when done */
      packnum++;
      pb = mid_pbuf_get();	/* get a new one */
      ASSERT (pb != NULL);
      br = (m_bar_req_t *) pb;
      br->curtime = bar->bar_curtime;
      br->num = bar->bar_num;
      br->totnum = bar->bar_total;
      br->noreply = bar->state & L_BAR_FLUSH;
      pblen = sizeof(m_bar_req_t);
/*      mutex_lock(&(lock->m));*/
    } /* while delta pack */
    INCR_STATS(BarStats.data_sent, xferlen);
  }
	

#ifdef DEBUG
    /*								---DEBUG */
    if (IS_DBG_BARR_MSGS) {
      long *f = (long *) pb;
      sprintf(smsg, "%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx",
       *f, *(f+1), *(f+2), *(f+3), *(f+4), *(f+5), *(f+6),
       *(f+7), *(f+8), *(f+9), *(f+10),
       *(f+11), *(f+12));
      StatusMsg(smsg);
      sprintf(smsg,"Total outgoing message length: %u", pblen);
      StatusMsg(smsg);
    }
    /*								---DEBUG */
#endif /* DEBUG */

  /* unlock entry while waiting for barrier to take care of the case */
  mutex_unlock(&(bar->m));
  CommXRequest(bar->lock, M_BARRIER_OP, bar->bestguess, pb, pblen, 
	       packnum, LAST, &tid);
  mutex_lock(&(bar->m));
} /* midSendBarReq */

/*
 ************************************************************************
 *		    m i d B a r r i e r R e q u e s t			*
 *----------------------------------------------------------------------*
 * Handle barrier hit request from a thread; the processor should be	*
 * the master.  Queue the request unless all threads are in, in which   *
 * case we should service it....					*
 ************************************************************************
 */
comm_request_t
midBarrierRequest( lockn, host, thread, tid, retrans, msg, msglen)
     u_long lockn;		/* name of barrier this is all about */
     u_long host;		/* host requesting transfer */
     u_long thread;		/* thread@host that hit barrier */
     u_long tid;		/* transaction id for this request */
     int retrans;		/* non zero if this is a retransmission */
     char *msg;			/* request specific info */
     u_long msglen;		/* length of data in msg */
{
  char255 smsg;
  pl_t *bar;			/* pointer to barrier entry */
  int empty;			/* to check for empty q's */
  m_bar_req_t *br;
  pbuf pb;			/* a packet buffer pointer */
  addrseg_t *as;		/* address segment */
  u_long msgl = msglen;		/* internal message length */
  comm_request_t rc = M_R_DONE;

  /* convert barname to a pointer to the barrier structure */
  bar = (pl_t *) lockn;
  br = (m_bar_req_t *) msg;	/* message should start with a request */

#ifdef DEBUG
  if (IS_DBG_BARR_EXTR) {
    sprintf(smsg,"Barrier [%s] hit from %lu(%lu)@%lu for %u[%lx]@%lu (#=%u; tot=%u)",
	    ((retrans & M_FINAL)?"last":"more"),
	    host,thread,tid,lockn,bar,br->curtime, br->num, br->totnum);
    StatusMsg(smsg);
  }
#endif DEBUG

  ASSERT (bar->state & L_BARRIER);
  ASSERT (bar->lock == lockn);
  ASSERT (bar->held == MIDWAY_LOCKHELD_OTHER);


  /*
   * Acquire mutex guarding barrier structure
   */
  mutex_lock(&(bar->m));	/* ensure consistent access */

  /* check for first one to hit barrier on new iteration */
  if (br->curtime == bar->bar_curtime) { /* this is part of "new" iteration */
    if (bar->bar_subtotal == 0) { /* this is first one to hit barrier */
      bar->bar_total = br->totnum; /* update requested number */
    } else {			/* not first one to hit barrier */
      if (bar->bar_total != br->totnum) { /* sanity failure */
      }
    } /* not first to hit barrier */

    /* if slave says flush, ensure set */
    if (br->noreply && !(bar->state & L_BAR_FLUSH))
      bar->state |= L_BAR_FLUSH;

    if (retrans & M_FINAL) {		/* only for last one... */
      /* add self to output waiter q for release barrier */
      if (mid_extq_dup(&(bar->bar_waiters), host, thread, tid)) {
	/* already in Q, don't do it again */
	CommAck(host, thread, tid, lockn, 0, 0); /* ReplyPending */
	rc = M_R_QUEUED;
	goto done;
      }
      if (mid_extq_add(&(bar->bar_waiters), host, thread, tid, br->curtime, 0)) {
	/* no memory *mjz* */
      }
      bar->bar_subtotal += br->num; /* bump by #of threads that hit barrier */
    } /* retrans & M_FINAL */

    /*
     * 24Jun93(mjz): previous method: we copy the input data to the end
     * of the last queued pbuf, and continue queuing pbufs (at most 1)
     * until the input is exausted, potentially compacting pbufs if
     * not much data is in each packet.  We could not just use the input
     * packets, because the Ethernet comm routines presented a static
     * buffer, not a pbuf which could be queued.
     *
     * new method: if data in buffer, simply copy entire thing to a pbuf.
     * if flushing, don't queue but apply the changes.
     */
    /* COPY DATA TO END OF OUTPUT QUEUE.  GAK */
    /* goes against Brian's do as little as possible now philosophy */
    msgl -= sizeof(m_bar_req_t);
    as = (addrseg_t *) (br+1);	/* point after inital msg */
    while(msgl) {		/* stuff still in incoming message */
      switch(as->segtype) {

      case DATASEG:		/* only valid one for barriers */
#ifdef PAGE_BASED
      case DIFFSEG:
#endif

#ifdef FLUSH_NOCOPY
	  /*
	   * Weird semantics.  To try and do as much in parallel,
	   * (and avoiding a data copy), copy directly to memory
	   * if we are in "flush" mode.  If any thread declares
	   * flush mode, it will apply to all, and no results
	   * will be redistributed.
	   */
	  if (!(bar->state & L_BAR_FLUSH)) {
	    /* NORMAL (not flush) CASE */
#endif
#ifdef DEBUG
	    if (IS_DBG_BARR_EXTR) {
	      sprintf(smsg,"midBarrierReq: copying %u to output buf",
		      msglen); /**/
	      StatusMsg(smsg);
	    }
#endif DEBUG
	    pb = mid_pbuf_get();	/* get one */
	    ASSERT (pb != NULL);
	    bcopy(msg, pb, msglen); /* just copy the WHOLE thing */
	    /* one possible problem: br->curtime.  We don't know
	       what time to stick in the message yet.  The sendpackets
	       routine will Do The Right Thing, and insert it for us.
	       The old code just used to set it to zero here
	       */
	    ((m_bar_req_t *) pb) ->totnum = msglen; /* remember length */
	    mid_pbuf_enqueue(pb, &(bar->output)); /* place it on Q */
	    msgl = 0;		/* done -- copied whole thing */
#ifdef FLUSH_NOCOPY
	  } /* not flush, copy */
	  else {		/* FLUSH */
	    /* NOTE: originally, we just copyied the data to local
	       memory without updating dirty bits.  For now, use
	       common unpack routine */
#ifdef PAGE_BASED
	    mid_paged_unpack_data((u_long)bar,
				  (char **)&as, &msgl, bar->amap,
				  "midBarrierRequest",  mid_MUD_barrier);
#else
	    mid_unpack_data((char **)&as, &msgl, bar->amap,
			    "midBarrierRequest",  mid_MUD_barrier);
#endif
	  }
#endif
	break;

      default:
	sprintf(smsg,"Unknown segment received (%x) decoding barrier reply for %lu -- ignoring rest of message [msglen=%u]",
		as->segtype, bar->lock, msgl);
	StatusMsg(smsg);
	msgl = 0;		/* so quit loop */
	break;
      } /* switch on segtype */
    } /* while msgl */

    if (bar->bar_subtotal == 0) {
      /* this must be part of a multipacket request message that made it
	 before this processor (master) hit the barrier.  Bail out now,
	 otherwise we will succeed in the "subtotal == total" test below.
	 Yuck!
       */
      rc = M_R_INPRMULTI;	/* will be ignored, actually, if MP request */
      goto done;
    }
    if (bar->bar_subtotal == bar->bar_total) { /* time to make the donuts */
#ifdef DEBUG
      if (IS_DBG_BARR_EXTR) {
	StatusMsg("midBarrierRequest: releasing.");
      }
#endif /* DEBUG */
      if (!(retrans & M_FINAL)) {
	sprintf(smsg, "BarrierRequest(%lu): releasing, but not last packet! [h= %lu th=%lu tid=%lu",
		lockn, host, thread, tid);
	StatusMsg(smsg);
	botch("not lastpacket"); /* So I can catch it... */
      }

      mid_release_barrier(bar, host);
    }
    else if(bar->bar_subtotal < bar->bar_total) { /* not enough yet */
      rc = M_R_QUEUED;		/* we were queued. */
      goto done;			/* someone later will take care of this */
    }
    else { /* bar_subtotal > bar_total -- oops! */
      sprintf(smsg, "midBarrierRequest %u: bar_subtotal (%u) > bar_total (%u) -- barfing.",
	      bar->lock, bar->bar_subtotal, bar->bar_total);
      StatusMsg(smsg);
      botch("req subtotal too high");
    }
    
  } /* part of "new" iteration */

  /* case: part of last iteration -- retransmission or delayed packet */
  else if (br->curtime == bar->bar_lasttime) {
    /* RESEND */
#ifdef DEBUG
    if (IS_DBG_BARR_WNG) {
      StatusMsg("midBarrierRequest: retransmitting");
    }
#endif /* DEBUG */
    mid_bar_sendpackets(bar, host, thread, tid);
  }

  /* case: probably very old packet -- should not hit this one */
  else {
    StatusMsg("midBarrierRequest: packet not in current or last generation: dropping.");
    rc = M_R_DROPPED;
  }


done:
/*  fprintf(stderr,"%s\n-Current State:\n", smsg);fflush(stderr); */
/*  PrintLock(bar->lock); */ /* better way to do this? 30jul92mjz */

  /* release mutex before returning */
  mutex_unlock(&(bar->m));
  return(rc);
} /* midBarrierRequest */


/*
 ************************************************************************
 *		  m i d _ b a r _ s e n d p a c k e t s			*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
static void
mid_bar_sendpackets( bar, host, thread, tid )
     pl_t *bar;			/* barrier we're using */
     u_long host;		/* host to get packets */
     u_long thread;		/* thread on host */
     u_long tid;		/* transaction id for this request */
{
  pbufe_t	*e;		/* for running thru packets */
  pbufe_t	*enext;		/* ditto, need to tell if at end */
  u_long	pnum;		/* packet sequence number */
  m_bar_req_t 	*br;		/* to stamp last packet */
  m_bar_req_t 	request;	/* a real one! */

  e = MID_PBUF_QTOP(&(bar->old_output)); /* what if none? */
  pnum = 0;			/* start at 0 */
  
  if ((e == 0)||(bar->state & L_BAR_FLUSH)) {	/* no data to transmit */
    /* waste of time to do this every time? 31jul91mjz */
    request.curtime = bar->bar_curtime; /* reply with next barrier it val */
    request.num = 0;
    request.totnum = sizeof(request);
    CommReply(M_BARRIER_OP, host, thread, tid, bar->lock,
	      1 /* lastpacket*/, 0 /*pnum*/, NULL_ADDR_VECTOR/*nacks*/,
	      (char *)&request, sizeof(request), C_NOFREE);
    return;
  }

  while (e != 0) {		/* something left to send */
    enext = MID_PBUF_QNEXT(e);
    br = (m_bar_req_t *) Q2PBUF(e); /* get pbuf head, which is bar req */
    br->curtime = bar->bar_curtime;  /* redundant after 1st pass */
    /* but probably in noise, checking would probably cost more...*/

#ifdef DEBUG
    if (IS_DBG_BARR_MSGS)
    /*								---DEBUG */
    {
      long *f = (long *) br;
      char255 smsg;

      sprintf(smsg, "%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx",
       *f, *(f+1), *(f+2), *(f+3), *(f+4), *(f+5), *(f+6),
       *(f+7), *(f+8), *(f+9), *(f+10),
       *(f+11), *(f+12));
      StatusMsg(smsg);
      sprintf(smsg,"Total outgoing message length: %u", br->totnum);
      StatusMsg(smsg);
    }
    /*								---DEBUG */
#endif /* DEBUG */
    CommReply(M_BARRIER_OP, host, thread, tid, bar->lock,
	      (enext == 0) /* last */, pnum, NULL_ADDR_VECTOR /* nacks */, 
	      (char *) br, br->totnum, C_NOFREE);
    e = enext;
    pnum++;
  }

  return;
} /* mid_bar_sendpackets */

/*
 ************************************************************************
 *		    m i d _ b a r _ a p p l y c h g s			*
 *----------------------------------------------------------------------*
 * apply all changes in a given packet to the local memory		*
 * bar_lasttime, bar_curtime updated by our caller.  How?		*
 ************************************************************************
 */
static void
mid_bar_applychgs( bar, msg, msglen )
     pl_t	*bar;		/* the barrier */
     char	*msg;		/* the incoming message */
     u_long	msglen;		/* and its length */
{
  m_bar_req_t	*br;		/* the barrier request message */
  addrseg_t	*as;		/* container for addrseg's */
  char255	smsg;

  /* done under lock? */
#ifdef DEBUG
  if (IS_DBG_BARR_EXTR) {
    sprintf(smsg, "mid_bar_applychgs: incoming message length %u", msglen);
    StatusMsg(smsg);
  }
#endif /* DEBUG */
  br = (m_bar_req_t *) msg;
  msglen -= sizeof(m_bar_req_t);
  as = (addrseg_t *) (br+1);	/* point after barrier info */
  while (msglen) {
    switch(as->segtype) {

    case DATASEG:
	mid_unpack_data((char **)&as, &msglen, bar->amap, "mid_bar_applychgs",
			mid_MUD_barrier);
      break; /* case DATASEG */

#ifdef PAGE_BASED
    case DIFFSEG:
	mid_paged_unpack_data((u_long)bar,
			      (char **)&as, &msglen, bar->amap, "mid_bar_applychgs",
			      mid_MUD_barrier);
      break; /* case DIFFSEG */
#endif

    default:
      sprintf(smsg,"Bad segment received (%x) decoding barrier reply for %lu -- giving up [msglen=%u]",
	      as->segtype, bar->lock, msglen);
      StatusMsg(smsg);
      return;			/* better not be under lock */
      break;
    } /* switch */
  } /* while(msglen) */

} /* mid_bar_applychgs */

/*
 ************************************************************************
 *		    m i d B a r r i e r R e p l y			*
 *----------------------------------------------------------------------*
 * accept reply from net, and process it				*
 ************************************************************************
 */
void
midBarrierReply( lockn, srchost, dstthread, tid,  msg, msglen )
     u_long lockn;		/* name of barrier this is all about */
     u_long srchost;		/* should be manager */
     u_long dstthread;
     u_long tid;		/* transaction id for this request */
     char *msg;			/* request specific info */
     u_long msglen;		/* length of data in msg */
{
  pl_t *bar;			/* pointer to barrier entry */
  m_bar_req_t *br;
  clock_t teatime;
  char255 smsg;

  /* convert barname to a pointer to the barrier structure */
  bar = (pl_t *) lockn;
  br = (m_bar_req_t *) msg;	/* message should start with a request */

  ASSERT (bar->state & L_BARRIER);
  ASSERT (bar->lock == lockn);
  ASSERT (bar->held == MIDWAY_LOCKHELD_OTHER);

  mutex_lock(&(bar->m));
  if (bar->state & L_ACQ) {
    /*
     * only apply changes to memory if we are in the midst of getting
     * past the barrier.  When might we be here with L_ACQ off?  Well,
     * if two threads on this machine hit the barrier, with the first
     * hitting the barrier before the second gets a chance to register
     * (which means the mutations are disjoint), the second will also
     * send off a barrier request.  However, once one of them processes
     * all changes & releases the waiting thread(s), we cannot have
     * the second's update messages go through (because now we are
     * mutating again).  Hmmmm...  a good reason to not unlock while
     * applying changes!  think about unlocking more...31jul91mjz
     */
    /*** check this out 30jul91mjz */
    teatime = bar->bar_curtime;
    if (teatime != br->curtime) {
      bar->bar_lasttime = teatime;
      bar->bar_curtime = br->curtime;
    }
    
    mid_bar_applychgs(bar, msg, msglen); /* real work */
  } /* L_ACQ was on */
  mutex_unlock(&(bar->m));	/* restr. parallelism; do before apply?mjz*/

  return;
} /* midBarrierReply */

RegisterBarrierUse( bname )
     midway_barrier_t bname;		/* barrier name */
{
  pl_t *bar;

  /* convert barname to a pointer to the barrier structure */
  bar = (pl_t *) bname;

  ASSERT (bar->state & L_BARRIER);
  ASSERT ((vm_offset_t) bar->lock == (vm_offset_t) bname);
  ASSERT (bar->held == MIDWAY_LOCKHELD_OTHER);

  mutex_lock(&(bar->m));	/* so it is safe to twiddle with DS */

  if (bar->state & L_BAR_NOREG) { /* first time to register */
    bar->state &= L_XBAR_NOREG;	/* remove it -- we've registered */
  }
  else {
    bar->bar_num++;		/* bump # that must hit barrier */
  }

  mutex_unlock(&(bar->m));
  return;
} /* RegisterBarrierUse */

/*
 ************************************************************************
 *	    m i d _ b a r r i e r _ p r i n t _ s t a t s	        *
 *----------------------------------------------------------------------*
 * Print barrier statistics						*
 ************************************************************************
 */
mid_barrier_print_stats ()
{
#if !defined(STATS)
  StatusMsg ("Not compiled to collect Barrierr statistics");
#else
  int i, len, slen;
  u_long acqs, reqs;
  char255 smsg, tmp;
#define MAX_STR_LEN 80

  sprintf(smsg,
  "Bars alloc=%lu cross %lu (map %lu %f) flush %lu (map %lu %f) Data Sent %lu Rel %lu",
	  BarStats.alloced, BarStats.crossed, BarStats.cross_data, 
	  (BarStats.crossed == 0) ? 0.0 :
	     (double)(((double) BarStats.cross_data)
		          / ((double) BarStats.crossed)),
	  BarStats.flushed, BarStats.flush_data, 
	  (BarStats.flushed == 0) ? 0.0 :
	     (double)(((double) BarStats.flush_data)
		          / ((double) BarStats.flushed)),
	  BarStats.data_sent, BarStats.data_released);
  StatusMsg (smsg);
#endif
}
mid_barrier_reset_stats ()
{
#if defined(STATS)
    bzero (&BarStats, sizeof(BarStats));
#endif
}


