/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
#ifndef _BITWIDDLE_H
#define _BITWIDDLE_H
/*
 * bitwiddle.h BIT twiddle routines.
 *
 * bit array stuff.  this is used in sets of hosts and sets of packets.
 *
 * typedef BITARRAY unsigned int;
 *
 * WE ASSUME AN unsigned int IS EXACTLY 4 BYTES (OCTETS) LONG!
 *
 *  BITARRAY bitarray[(#things + 32) div 32]; (#things indexed at 0)
 *
 * mostly, these will be malloc'd; malloc guarantees this will start on
 * a "word" boundry.
 *
 * bits are numbered from 0 to #things, starting at the left most bit
 *   (most significant bit of the first word of the array)
 *
 * To find the proper word, use (item div 32) (#things index origin 0 again)
 *
 * To find the proper bit within the word, (item mod 32) give the spot;
 * it's either 0x80000000 >> (item mod 32) (provided shift is logical)
 * or 1 << (31-(item mod 32)).
 *
 * 22 Jan 91, Matthew J. Zekauskas
 * 17 Mar 91 (mjz): added midCompareBit; should switch /32's with shifts...
 * 16 Jul 92 (was): fixed bug in computing midBITARRAYsize
 * 20 Oct 92 (was): added midZeroBitArray
 * 10 Oct 93 (was): added midSetBitRange
 * 15 Nov 93 (was): made midCompareBit an extern routine instead of static
 */

#ifdef C_WORKED_CORRECTLY
/* sizeof not allowed in #if expression :-( */
#if (4 != (sizeof unsigned int))
  THE BIT TWIDDLING ROUTINES ARE GOING TO BREAK AND MUST BE MODIFIED
#endif
#endif C_WORKED_CORRECTLY


/*
 * Define base type for bit arrays
 */
typedef unsigned int BITARRAY;

/* define assumed size of this type for assertion checks */
#define BITARRAY_BITS 32
#define BITARRAY_CHARS 4



/* in this, and other, macros, the arguments had better be integer! */
#define midBITARRAYsize(x) (((x+31)/32)*sizeof(BITARRAY))

#ifndef NOT_AS_MACROS

#define midSetBit(ba,bit) ((ba)[(bit)/32] |= ((unsigned)0x80000000 >> ((bit)%32)))
#define midClearBit(ba,bit) ((ba)[(bit)/32] &= (0xffffffff ^ ((unsigned)0x80000000 >> ((bit)%32))))
#define midTestBit(ba,bit) ((ba)[(bit)/32] & ((unsigned)0x80000000 >> ((bit)%32)))
#define midZeroBitArray(ba,bitlen) bzero(ba, midBITARRAYsize(bitlen))

#define midSetBitRange(ba,startbit,endbit)					\
{ register BITARRAY *bap = ba;     /* cache ptr in a register */		\
  register BITARRAY *firstwordP = &bap[(startbit) / 32];			\
  register unsigned  firstbit = (startbit) % 32;				\
  register BITARRAY *lastwordP = &bap[(endbit) / 32];				\
  register unsigned  lastbit = (endbit) % 32;					\
  if (firstwordP < lastwordP) {							\
    *firstwordP++ |= ((unsigned)0xFFFFFFFF >> firstbit);			\
    while (firstwordP < lastwordP)						\
     *firstwordP++ = ((unsigned)0xFFFFFFFF);					\
    firstbit = 0;								\
  }										\
  *firstwordP |= ((((unsigned)0xFFFFFFFF) >> ((31 + firstbit) - lastbit))	\
		          << (31 - lastbit));					\
}

#define midOrBitArray(ba1,ba2,bitlen) \
{ register int __i,__j; \
    for (__i=0, __j=0; __j<=bitlen; __i++, __j+=32) { \
      ba1[__i] |= ba2[__i]; \
    } \
}

#else NOT_AS_MACROS
/* this here for documentation */
void
midSetBit( ba, bit )
     BITARRAY	ba[];
     u_long	bit;		/* bit to set */
{
  ba[bit/32] |= (unsigned)0x80000000 >> (bit%32);
} /* midSetBit */
void
midClearBit( ba, bit )
     BITARRAY	ba[];
     u_long	bit;
{
  ba[bit/32] &= 0xffffffff ^ ((unsigned)0x80000000 >> (bit%32));
  /* XOR first to ensure we have right mask for AND */
} /* midClearBit */
int
midTestBit( ba, bit )
     BITARRAY	ba[];
     u_long	bit;
{
  return((ba[bit/32] & ((unsigned)0x80000000 >> (bit%32))));
} /* midTestBit */

void midSetBitRange(ba,startbit,endbit)
     BITARRAY	ba[];
     u_long	startbit,endbit;
{ register u_long firstword = startbit / 32;
  register u_long firstbit = startbit % 32;
  register u_long lastword = endbit / 32;
  register u_long lastbit = endbit % 32;
  if (firstword < lastword) {
    ba[firstword++] |= ((unsigned)0xFFFFFFFF >> firstbit);
    while (firstword < lastword)
      ba[firstword++] = ((unsigned)0xFFFFFFFF);
    firstbit = 0;
  }
  ba[firstword] |= ((((unsigned)0xFFFFFFFF) >> ((31 + firstbit) - lastbit))
		          << (31 - lastbit));
}

#endif NOT_AS_MACROS


#ifdef DEFINE_BITWIDDLE_H
/*
 * midCompareBit: test ba1 against ba2 for equality out until max bit.
 *	we actually might check a bit more, since we only do word compares.
 *	This should be fine, but implies ALL WORDS MUST BE INITIALIZED TO 0!
 */
int
midCompareBit( ba1, ba2, max )
     BITARRAY	ba1[];
     BITARRAY	ba2[];
     unsigned long max;		/* maximum bit that could be set */
{
  int i;			/* index, of course */

  i = max/32 + 1;		/* max index into ba1 & ba2 */
  do {
    i--;			/* go down array */
    if (ba1[i] != ba2[i])
      return(0);		/* not equal */
  } while (i>0);		/* something left to do */
  return(1);			/* got here, must be equal! */
} /* midCompareBit */

#else

/*
 * Define prototype instead
 */
extern int
midCompareBit( BITARRAY	ba1[], BITARRAY	ba2[], unsigned long max );

#endif /* DEFINE_BITWIDDLE_H */

#endif /* _BITWIDDLE_H */
