/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * comm.h: Headers related to communication
 *
 * 16 Dec 90: Matthew J. Zekauskas
 *
 *  8 jul 91 (mjz): add retransmission bit to retransmitted packet
 * 31 Jul 91 (mjz): add C_FREE/C_NOFREE (free packet buffer indicator)
 * 17 Sep 91 (mjz): moved SCHED_OP, added ALLOCATE and ALLOC_INFO for shmalloc.
 * 11 Aug 92 (was): added clock_t type
 * 18 Aug 92 (was): added new message types M_PAGE_xxx
 * 26 Aug 92 (was): added flag in cpt_t to indicate multicast requests.
 *  4 Oct 92 (mjz): add incarnation number to guard against old stuff[UDP SUX]
 *  9 Oct 92 (was): added NET_xxxx macros
 * 12 Oct 92 (mjz): add some fields to cpt + new results for ackafter support
 * 15 Oct 92 (was): added ifndef _COMM_H_ to allow multiple inclusions
 *  4 May 94 (was): changed layout of comm msg and cpt_t
 */
#ifndef _COMM_H_
#define _COMM_H_ 1


#define C_FREE	1
#define C_NOFREE 0

#include <sys/types.h>          /* define u_long etc */
#include <mach.h>
#include <cthreads.h>
#include <netinet/in.h>		/* sockaddr_in, + ntohl()... */
#include <netdb.h>		/* for hostent */
#include <sys/uio.h>		/* for iovec */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>		/* for inet_ntoa() */

#include "midway_types.h"       /* for clock_t */
#include "bitwiddle.h"		/* BITARRAY */


/*
 * Define types for host/thread ids in the comm msg.
 * Note: Size depends on MAX_PROCESSORS.
 */
typedef u_char c_hostid_t;
typedef u_char c_threadid_t;

#define MAX_COMM_HOSTID  255	/* used for assertion checks on above types */
#define MAX_COMM_THREADID  255


/* the message, sent as a UDP datagram */
typedef struct CommMsg {
  u_char 	version;	/* version of this protocol */
  u_char	type;		/* type of this header */
  u_char	flags;		/* we all need flags */
  u_char        unused;
  u_long	incarnation;	/* incarnation # of this system */
  c_hostid_t	source;		/* source host id */
  c_threadid_t	id;		/* identifies the calling thread */
  c_hostid_t 	rhost;		/* reply-to host */
  c_threadid_t 	rid;		/* reply-to id / destination id */
  u_long	lock;		/* object of this message */
  clock_t	clock;		/* sender's clock value */
  u_long	tid;		/* transaction ID */
  /* initially LocalClock on send, and same value on reply... */
  u_long	pacnum;		/* seq # of this packet, multi-pac replies */
  addr_vector_t nacks;		/* bit array of acksRequired */
  /* type dependent data follows */
} COMMMSG, *COMMMSGP;

#define M_PROTOVER 3		/* V3: + incarnation */

/* flags */
#define M_LASTPACK	0x01 /* last packet of a series */
#define M_RETRANS	0x10 /* this is a retransmission */
/* following flag NOT sent over the wire... */
#define M_DOAGAIN	0x20 /* set by receiver if this is a do-over */
#define M_FORWARDED	0x40 /* set by receiver if packet was fwded */
#define M_FINAL		0x80 /* set if this is last packet to arrive */
/* how know processed last??? */

/* types */
#define M_REPLY_BIT      0x80	/* Use high-order bit for OR'ing  */

#define M_OP_BITS        0x7F	/* bits used for message ops  */

#define M_ACK_OP	  1
#define M_NACK_OP	  2
#define M_ACK		 (M_ACK_OP | M_REPLY_BIT)	/* always a reply */
#define M_NACK		 (M_NACK_OP | M_REPLY_BIT)  	/* always a reply */
#define M_INVALIDATE_OP   3
#define M_SCHED_OP        4        /* scheduling barrier (data is lock info) */
#define M_ALLOCATE	  5        /* global allocation of malloc'd store */
#define M_ALLOC_INFO      6        /* if region allocated, get linesize */
#define M_HOST_OP	  7        
#define M_BARRIER_OP	  8        /* hit,release */
#define M_LOCK_OP	  9        
#define M_LOCK_INV_OP	 10        /* only used in replies */
#define M_LOCK_NEWOWN	 11        /* new owner for lock (update hint) */
#define M_NOTIFY_OP	 12        /* asy.  Could be LOCK+INVAL; new reader */
#define M_PAGE_DATA      32        
#define M_PAGE_UPDATE    33        
#define M_PAGE_COPYSET   34        
#define M_SEPPUKU	127        /* terminate self with honor */
#define M_TYPEMAX	do not use obsolete M TYPEMAX define

/* host_reply specific... */
typedef struct midHostReply {
  u_long	index;		/* index of this host */
  u_long	ipaddr;		/* in network order */
  u_short	udpport;
  u_short	reserved;	/* keep word aligned */
} HREPLYMSG, *HREPLYMSGP;

/*
 * Request routines called by comm
 */
typedef u_long comm_request_t;

/* possible values (also "state" values in comm's trantab) */
#define M_R_STATUS	0x80000000 /* is not forwarded place */
#define M_R_DONE	0x80000001 /* transaction completed here */
#define M_R_DROPPED	0x80000002 /* request packet dropped */
#define M_R_QUEUED	0x80000004 /* request packet queued -- still in proc.*/
#define M_R_INPROCESS	M_R_QUEUED /* ditto */

#define M_R_WAITING	0x80000100 /* [INPROCESS] Replyer waiting for ACK */
#define M_R_ACKED	0x80000200 /* [INPROCESS] ACK arrived for replier */
#define M_R_INPRMULTI	0x80000400 /* [INPROCESS, multi packet request.]

/* RC, not state: rc => don't change state */
#define M_R_UNCHANGED   0x80000010
#define M_R_DOAGAIN	0	   /* allow next packet to be treated as new */
/* operation terminated because it could or would block */
#define M_R_COULDBLOCK	0x80000020

/* values with M_R_STATUS off mean the message was forwarded to that host */
/* no longer 16sep92mjz; here for history -- delete soon :-) */

/*
 * Defns for fast network communication
 */
typedef int (*comm_filter)(char *buf, u_long len, int opaque, unsigned int t);






/* cpt, htblent comm.c specific, but useful for testing too, so it's here */

typedef struct htblent {	/* host # to ip/udp/(thread?) */
  u_long	ip;		/* ip addr of host */
  u_short	uport;		/* udp port */
} htblent_t;

#define _PNUM 4
typedef struct packetlist {	/* outstanding packet list */
  BITARRAY	pacvec[_PNUM];	/* can rep 4*32 packets for about 1 meg data */
  struct packetlist *pacnext;	/* pointer to more of these */
} plist_t;

typedef struct {
  char * msg;			/* keep old msgs around for retransmission */
  u_long mlen;			/* (was pbuf, but .h depend on each other  */
} c_pbuf_q_t;			/* ..20oct92mjz) */


/* sigh, used when sending replies too.. need tid and pnum */
/* no, did could be rid... SIGH */
typedef struct cperthread {	/* client per thread data structure */
  struct mutex	m;		/* client waits on this for req to be done */
  struct condition c;		/* actually, it waits on this */
  int		dest;		/* place it's going (into htbl) */
  char		*msg;		/* type dependant message */
  u_long	mlen;		/* length of above message */
  int		retries;	/* see retries flag below */
  addr_vector_t acks_required;	/* acks required before we can be released */
  addr_vector_t	acks_received;	/* acks received so far */
  u_long	maxlinear;	/* max # received in sequence << 1 */
  u_long	lastpacket;	/* biggest packet seen so far */
  u_long	writers;	/* if > 0, some thread is writing to cache */
  int           multicast;      /* multicast request, multi-replies expected */
  u_long	ackpnum;	/* if getting lg reply, should ack & pnum */
  u_long	ackthread;	/* thread ack should go to (sigh) */
  plist_t	plist;		/* for keeping track of multi-packet msgs */
  c_pbuf_q_t   *oldp;		/* keep old messages around */
  int		pflow;		/* for packet flow times (if enabled) */

  COMMMSG	msgheader;	/* Initialized message header */
  struct iovec	iov[2];		/* for scatter/gather io */
  struct sockaddr_in sin;	/* for destination */
  struct msghdr msgh;

} cpt_t;

#define c__rhost  msgheader.rhost
#define c__rid    msgheader.rid  
#define c__did    msgheader.rid  
#define c__tid    msgheader.tid
#define c__lock   msgheader.lock
#define c__mflags msgheader.flags
#define c__optype msgheader.type
#define c__pnum   msgheader.pacnum
#define c__nacks  msgheader.nacks
#define acksRequired acks_required.bitArray
#define acksReceived acks_received.bitArray

/* for retries field */
#define M_SCANNABLE	0x00000001 /* entry available for scanning */
#define M_SCANCLEAR	0xfffffffe /* clear M_SCANNABLE bit */
#define M_RETRANSMIT	0x00000010 /* entry can be retransmitted */
#define M_RETRANSCLR	0xffffffef /* clear M_RETRANSMIT */
#define M_ALLPACKETS	0x00000100 /* we know we saw the last packet */
#define M_NEWTARGET	0x00001000 /* retargeted; don't set M_RETRANS */
#define M_NEWTARGCLR    0xffffefff
#define M_WAITING	0x00010000 /* replier waiting here */
#define M_WAITCLEAR	0xfffeffff
#define M_WAITFREE	0x00020000 /* free packet after waiting. */
#define M_WAITFREECLR	0xfffdffff


/*
 * servertab - the table kept in comm for each requestor -- so we can
 * supress duplicates, mostly.
 *
 * for now, keep only the packet number per host.
 *
 */

typedef struct servertab {
  clock_t	clock;		/* most recent host clock value */
  u_long	tid_hint;	/* most recent workertid entry */
} servertab_t, *servertab_pt;






/*
 * Define macros to access udp calls.
 * May be defined to invoke standard udp calls
 * or fast udp library calls.
 */
#ifdef OLD_FAST_UDP

/*
 * Define network macros to invoke fast library routines
 */
#define NET_SOCKET       e_socket
#define NET_SETSOCKOPT   e_setsockopt
#define NET_BIND         e_bind
#define NET_SENDTO       e_sendto 
#define NET_SENDMSG      e_sendmsg
#define NET_RECVFROM     e_recvfrom
#define NET_CLOSE        e_close

#else

/*
 * Define network macros to invoke standard Unix routines
 */
#define NET_SOCKET       socket
#define NET_SETSOCKOPT   setsockopt
#define NET_BIND         bind
#define NET_SENDTO       sendto 
#define NET_SENDMSG      sendmsg
#define NET_RECVFROM     recvfrom
#define NET_CLOSE        close

#endif /* FAST_UDP */


/*
 * Define prototypes for comm routines 
 */
extern int CommReply
    (
     u_long	type,		/* kind of message */
     u_long	host,		/* index into host table */
     u_long	thread,		/* thread@host reply is destined for */
     u_long	tid,		/* tid for this transaction */
     u_long	lock,		/* ??? (u_long) */
     int	last,		/* non zero if last packet */
     u_long	pnum,		/* this packet number (?) */
     addr_vector_t nacks,	/* bitarray of acksRequired */
     char	*msg,		/* type dependent message info */
     u_long	msglen,		/* size of message */
     int        freepack 	/* =C_FREE to free pbuf pointed to by msg */
     );



#endif /* _COMM_H_ */

