/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * exception.c: Code to handle exceptions, this one is for Mach 3.0 only.
 *
 * 25 Aug 94 (mjz): ressurected.
 * 29 Sep 94 (was): added to page based version of midway
 */
#include <mach.h>
#include <mach/kern_return.h>
#include <cthreads.h>
#include <stdio.h>
#include "midway_types.h"
#include "midway_mem.h"
#include "shmalloc.h"

int midway_exception_debug = 0;	/* should add code to realmain to set */
/* for exception handling */
#include <mach/message.h>
#include <mach/exception.h>


extern int exc_server();
static mach_port_t exception_port;	/* our new exception port */
static mach_port_t old_exception_port;  /* old, so we can chain ourselves in */

#if 0
/* for testing, but this is wrong, since it should BE a demux,
    and not have mach_msg_server parameters!
*/
my_exc_server(demux, max_size, rcv_name)
	boolean_t (*demux)();
    mach_msg_size_t max_size;
    mach_port_t rcv_name;
{
    fprintf(stderr,"my_exc_server: called\n");
    exc_server(demux, max_size, rcv_name);
}
#endif


/*
 ************************************************************************
 *		    e x c e p t i o n _ s e r v e r			*
 *----------------------------------------------------------------------*
 * This routine 						        *
 * catches exception RPC's from the kernel, and then calls exc_server	*
 * (from libmach) to decode the message.  The only real decode routine  *
 * is catch_exception_raise() [from the exception_raise kernel call]	*
 * This should also be a library routine!				*
 *									*
 ************************************************************************
 */
static void
exception_server( void ) 
{
    kern_return_t krc;

    /* if we fail, we can't catch the exception.  Hand it off. */
    krc = thread_set_exception_port(mach_thread_self(), old_exception_port);
    if (krc != KERN_SUCCESS)  {
	fprintf(stderr,"exception_server: thread_set_exception_port (1)failed: %s\n",
		mach_error_string(krc));
    }

    /* OK, let's go get 'em */
    mach_msg_server(exc_server, /* was my_  for testing*/
		    512 /* max msg size */,
		    exception_port
		    );
  }

/*
 * midway_setup_exception_handler(): set up the exeception server.
 */

void
midway_setup_exception_handler(void)
{
    kern_return_t krc;


    krc = task_get_exception_port(mach_task_self(), &old_exception_port);
    if (krc != KERN_SUCCESS) {
	mach_error("task_get_special_port [exception]", krc);
	exit(1);
    }

    if (midway_exception_debug)  {
	printf("about to port_allocate exception port\n");
    }
    
    krc =  mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE,
			      &exception_port);
    if (krc != KERN_SUCCESS) {
	mach_error("port_allocate of exc port", krc);
	exit(1);
    }

    /* task_set_exception_port requires send right */
    krc = mach_port_insert_right(mach_task_self(),
				 exception_port,
				 exception_port,
				 MACH_MSG_TYPE_MAKE_SEND);
    
    if (krc != KERN_SUCCESS)  {
	mach_error("mach_port_insert_right", krc);
	exit(1);
    }

    if (midway_exception_debug) {
	printf("about to fork exception server, port is %d\n", exception_port);
	fflush(stdout);
    }
    cthread_detach(cthread_fork((cthread_fn_t)exception_server, 0));

    if (midway_exception_debug) {
	printf("done setting exception port, old was %d\n",
	       old_exception_port);
	fflush(stdout);
    }
    
    krc = task_set_exception_port(mach_task_self(), exception_port);
    if (krc != KERN_SUCCESS) {
	mach_error("task_set_special_port [exception]", krc);
	exit(1);
    }


}

/*
 ************************************************************************
 *		c a t c h _ e x c e p t i o n _ r a i s e		*
 *----------------------------------------------------------------------*
 *  Called by the mig-generated exc_server() by an exception_raise()	*
 * kernel call.  Reasons in /usr/include/mach/exception.h.  If we can	*
 * handle the fault, do so and return KERN_SUCCESS.  A reply message	*
 * will be sent back to kernel, and the thread released.  If we can't   *
 * handle, return KERN_FAILURE... then the usual stuff happens, which   *
 * is to first try the thread exception port, and then let the unix     *
 * server handle it (by converting it to a signal and zapping the	*
 * original process, probably resulting in a neat-o message like	*
 * "bus error").  [I got this by reading the code --mjz.]		*
 * Oh, and /usr/mach/include/exc.h has the mig generated include file...*
 ************************************************************************
 */
#include<mach/thread_status.h>	/* for thread_get_state calls below */


kern_return_t
do_catch_exception_raise( exception_port, thread, task,
		      exception, code, subcode)
    mach_port_t exception_port;	/* port that got exception */
    mach_port_t thread;		/* thread that hiccuped */
    mach_port_t task;		/* the thread's task */
    int	   exception;		/* what happend <mach/exception.h>*/
    int    code;		/* more detail <mach/machine/exception.h>*/
    int    subcode;		/* yet more detail */
{
    kern_return_t krc;
    extern kern_return_t paged_memory_data_access_request();

    /* PROBLEM: handling only one at a time.  We could deadlock this and
       any other thread trying to generate an exception to kill us.
       Must have either other threads servicing (generate on fly?) or
       not have the loop block waiting for an answer.
     */

    if (midway_exception_debug) {
        printf("exception: type=%s:%d[0x%lx] code=%s:%d[0x%lx] subcode=%d[0x%lx]\n",
	       (exception == EXC_BAD_ACCESS) ? "BadAddr" : "Other",
	       exception,exception, 
	       (code == KERN_INVALID_ADDRESS) ? "InvAddr" : 
	       (code == KERN_PROTECTION_FAILURE) ? "Protect" : "Other",
	       code,code, subcode,subcode);
	fflush(stdout);
    }

    if (exception == EXC_BAD_ACCESS) { /* see exception.h for other types */
      /* 
       * code contains kern_return_t describing error [machine dependant?]
       * It is either hard coded to KERN_INVALID_ADDRESS or set to the
       * result of vm_fault() for EXC_BAD_ACCESS (as seen by perusing
       * a few versions of "trap.c" for various machines).
       *
       * for BAD_ACCESS, subcode contains invalid memory address
       */
      switch(code) {
      case KERN_INVALID_ADDRESS:

#ifdef MJZTESTING
	/* sample resolution - just allocate the page */

	krc = VM_ALLOCATE(mach_task_self(), &subcode, vm_page_size, 0);
	printf("back from vm_allocate\n");fflush(stdout);
	if (krc != KERN_SUCCESS) {
	    mach_error("vm_allocate of page", krc);
	    return(KERN_FAILURE);
	}
	return(KERN_SUCCESS);
#endif MJZTESTING

	/*
	** Code goes here for addresses that have no memory backing them
	**
	** If you need to tell a read from write access, you will have
	** to issue a thread_get_state() call (I believe --mjz aug 94).
	** (There is an example at the bottom of this file.)
	**
	** Remember to add return(KERN_SUCCESS) if you allocate the page.
	** (Returning KERN_FAILURE (the default) causes the Unix exception
	**  handler to be invoked.)
	*/

	/*
	** For page based code, the xpager handles all allocation requests.
	** WAS 30 Sep 94
	*/

	break; /* INVALID_ADDRESS */


      case KERN_PROTECTION_FAILURE:
	/*
	 * this is also set for an instruction bus error or data bus
	 * error on the mips;  the comment in trap.c says "GDB"
	 */

	/*
	** Code goes here for writes to read-only pages
	**
	** Remember to add return(KERN_SUCCESS) if you allocate the page.
	** (Returning KERN_FAILURE (the default) causes the Unix exception
	**  handler to be invoked.)
	*/

	krc = paged_memory_data_access_request ((vm_offset_t) 0, subcode,
						vm_page_size, VM_PROT_WRITE);
	if (krc == KERN_SUCCESS) return krc;

	break; /* PROTECTION_FAILURE */


      default: /* do nothing, will return KERN_FAILURE */;

      } /* switch(code) */
    } /*EXC_BAD_ACCESS*/

    return(KERN_FAILURE);	/* we can't handle this exception */
} /* do_catch_exception_raise */



/*
 * catch_exception_raise: cover for handling exceptions.  If we
 * cannot handle it, pass it on to the real handler.  [In 2.5,
 * I did the MiG loop myself, and passed the same _message_ to
 * my code and then the old port.  This does the same thing...]
 */
kern_return_t
catch_exception_raise( exception_port, thread, task,
		      exception, code, subcode)
    mach_port_t exception_port;	/* port that got exception */
    mach_port_t thread;		/* thread that hiccuped */
    mach_port_t task;		/* the thread's task */
    int	   exception;		/* what happend <mach/exception.h>*/
    int    code;		/* more detail <mach/machine/exception.h>*/
    int    subcode;		/* yet more detail */
{
    kern_return_t krc;
    
    krc = do_catch_exception_raise(exception_port,
				   thread,
				   task,
				   exception,
				   code,
				   subcode);
    if (krc == KERN_SUCCESS)  {
      if (midway_exception_debug)
	    fprintf(stderr,"exception handled properly");
	return krc;
    }
    

#ifdef STOP_ON_ERROR
    fret("exception: type=%s:%d[0x%lx] code=%s:%d[0x%lx] subcode=%d[0x%lx]\n",
	 (exception == EXC_BAD_ACCESS) ? "BadAddr" : "Other",
	 exception,exception, 
	 (code == KERN_INVALID_ADDRESS) ? "InvAddr" : 
	 (code == KERN_PROTECTION_FAILURE) ? "Protect" : "Other",
	 code,code, subcode,subcode);
    botch("exception caught");
#else
    krc = exception_raise(old_exception_port,
			  thread,
			  task,
			  exception,
			  code,
			  subcode);
#endif /* STOP_ON_ERROR */

    return krc;
} /* catch_exception_raise */

	



#if 0
    if ((exception == EXC_BAD_ACCESS) && (code == KERN_INVALID_ADDRESS)) { 
	/* 
	 * code contains kern_return_t describing error [machine dependant?]
	 * It is either hard coded to KERN_INVALID_ADDRESS or set to the
	 * result of vm_fault() for EXC_BAD_ACCESS (as seen by perusing
	 * a few versions of "trap.c" for various machines).  Other possible
	 * problems (than INVALID_ADDRESS) are things like the protection
	 * is wrong... we will barf on those.
	 *
	 * subcode contains invalid memory address
	 */

#ifdef MJZTESTING
	/* sample resolution - just allocate the page */

	krc = VM_ALLOCATE(mach_task_self(), &subcode, vm_page_size, 0);
	printf("back from vm_allocate\n");fflush(stdout);
	if (krc != KERN_SUCCESS) {
	    mach_error("vm_allocate of page", krc);
	    return(KERN_FAILURE);
	}
	return(KERN_SUCCESS);
#endif MJZTESTING
	/* old BAD_ACCESS + INVALID_ADDRESS body */

	/* This code was used for EC-only addressing exceptions. */
	/* I'm leaving it here because of the example get state calls */

	/* 13 jul 92: getting an exception for a static region.  Add stuff
	 * to try and figure out how this happens...
	 *  3 Aug 92: it was because a subprocess(task) had an error --
	 *	it wasn't our task! (figured this out in july.)  However,
	 *	if we have multiple messages coming in for a region, they
	 *	may all fault on update;  only the first should cause
	 *	an alloc_info request; we should silently ack the others.
	 */
	ri = region_from_addr( (int *) subcode);
	if (ri->line_size != 0) { /* already mapped, sigh */
	  char smsg[255];
	  thread_state_data_t tsd; /* generic */
	  int tsdcount = THREAD_STATE_MAX;
	  int j;
#ifdef mips
	  struct mips_exc_state mes;
	  int mescount = MIPS_EXC_STATE_COUNT;
#endif mips

	  /*
	   * see if this should be silently ignored
	   */
	  if (task == mach_task_self()) { /* there is hope */
	    /* since regions are large, verify that the page is
	       actually mapped. */
	    vm_address_t adr = (vm_address_t) subcode;
	    vm_size_t sz;
	    vm_prot_t maxprot;
	    vm_prot_t prot;
	    vm_inherit_t inh;
	    boolean_t shrd;
	    mach_port_t object_name;
	    vm_offset_t offset;

	    krc = vm_region(task, &adr, &sz, &maxprot, &prot,
			    &inh, &shrd, &object_name, &offset);
	    if ((krc == KERN_SUCCESS) && (adr <= subcode)) {
	      /* all is well.  return. */
	      return(KERN_SUCCESS);
	    }
	    if (krc == KERN_SUCCESS) {
	      sprintf(smsg,"vm_region for faulting address returns \
address 0x%x size %d prot 0x%x maxprot 0x%x inh 0x%x shared %d \
object 0x%x offset 0x%x",
		      adr, sz, prot, maxprot, inh, shrd, object_name,
		      offset);
	      StatusMsg(smsg);
	    } else {
	      StatusMsg("vm_region says no vm at or above faulting address");
	    }
	  } /* task == task_self(): silently ignore? */

	  sprintf(smsg,"catch_exception_raise: EXC_BAD_ACCESS and KERN_INVALID_ADDRESS for task %lx thread %lx addr %lx, but line_size not zero (%lu)",
		  task, thread, subcode, ri->line_size);
	  StatusMsg(smsg);
	  sprintf(smsg,"our task self port: %lx", mach_task_self());
	  StatusMsg(smsg);

	  /* note: this should be machine dependant; I will do things as
	     generically as possible first.  Flavor 1 is defined on all
	     architectures I have seen (mips,i386,vax); and is the basic
	     thread state */
	  krc = thread_get_state(thread, 1, tsd, &tsdcount);
	  if (krc != KERN_SUCCESS) {
	    sprintf(smsg,"thread_get_state (flavor 1) fails, krc: %ld",
		    krc);
	    StatusMsg(smsg);
	    task_suspend(mach_task_self()); /* that should do it */
	  }
#ifdef mips
	  sprintf(smsg,"thread_get_state (flavor 1) says pc is 0x%lx",
		  ((struct mips_thread_state*)tsd)->pc);
	  StatusMsg(smsg);
	  krc = thread_get_state(thread, MIPS_EXC_STATE, &mes, &mescount);
	  if (krc != KERN_SUCCESS) {
	    sprintf(smsg,"thread_get_state (flavor MIPS_EXC_STATE) fails, krc: %ld",
		    krc);
	    StatusMsg(smsg);
	    task_suspend(mach_task_self()); /* that should do it */
	  }
	  sprintf(smsg,"mips_exc_state: cause 0x%x addr 0x%x coproc_state 0x%x",
		  mes.cause, mes.address, mes.coproc_state);
	  StatusMsg(smsg);
#endif mips
	  StatusMsg("catch_exception_raise: SPINNING");
	  for (j=0; j<0x7fffffff; j++)
	  /* What the hell are we spinning on? */
	    cthread_yield();   /* XX */
	} /* already mapped, sigh */

	
	krc = get_shmalloc_info(subcode);
	if (krc != KERN_SUCCESS) {
	  char smsg[255];
	  int i,j;
	  for (i=0; i<0x7fffffff; i++) {
	    sprintf(smsg, "catch_exception_raise: get_shmalloc_info(0x%lx) does not return KERN_SUCCESS -- SPINNING", subcode);
	    StatusMsg(smsg);
	    for (j=0; j<0x0fffffff; j++)
	      cthread_yield();
	  }
	}
	return(krc);
#endif /* 0 */
