/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/* 
 * Mach Operating System
 * Copyright (c) 1992 Carnegie-Mellon University
 * All rights reserved.  The CMU software License Agreement specifies
 * the terms and conditions for use and redistribution.
 */
/*
 * File:	uxkern/gprof_support.c
 * Author:	Chris Maeda
 * Date:	March 1992
 *
 * Defines routines for in-server profiling a la gprof.
 *
 * HISTORY
 * $Log: gprof_support.c,v $
 * Revision 1.11  1994/06/17  02:33:46  wsawdon
 * Changed gprof_support to call internal midway vm_allocate.
 *
 * Revision 1.10  1994/06/07  04:24:44  wsawdon
 * Changes for page based version of midway.
 * Also changed midway's message header format.
 *
 * Revision 1.9  1994/02/09  04:50:48  mattz
 * Fix three bugs:
 * 	- not transferring all amaps depending on packet breaks
 * 	- hanging by looping calling task_disable_pc_sampling (if profiling)
 * 	  (and only sometimes...)
 * 	- data race such that the gmon.out file could be written twice
 * 	  (usually on the slave)
 * We can now disable fastlocks in the library.  (Thanks, Wayne!)
 *
 * Revision 1.8  1994/02/01  18:33:05  mattz
 * Merge for version 0.8: MK83 and new in-kernel ATM code (with signalling).
 *
 * Major changes (see messages on MK82 branch for details):
 * 	- atm_iface revamped
 * 	- new socket_iface; now uses Cmaeda's -lsockets
 * 	- atm & fast (or slow) udp co-exist.  specify -MIDNETWORK_ATM
 * 		for ATM service
 * 	- (new in merge): removed silly extra message for invalidating
 * 		yourself
 * 	- new interfaces for MK82+.  Still don't support full sampling
 * 		(sampling of VM events) (have not modified crt0.c)
 * 		Now uses ode make tools (although not odemake itself).
 *
 * Revision 1.7  1993/12/30  03:05:55  wsawdon
 * New fast dirtybit code.
 * New shmalloc code.
 *
 * Revision 1.6.2.1  1993/10/09  23:26:05  mattz
 * Update for MK83/UX42 + sockets lib (new fast udp support).
 *         - FAST_UDP: the names changed.  Had to avoid clashes between
 *                 our usurping open/close/read/write & Chris' usurption.(!)
 *         - some header files have changed location
 *         - gprof support totally different.  For now, we are still using
 *           our crt0, not Brian's, so we don't have VM event gprof support.
 * [2nd try, hopefully this one in branch MK82, not mainline.]
 *
 * Revision 1.6  1993/03/14  03:34:37  wsawdon
 * Use modern midway_uxpass_local/remote instead of obsolete
 * uxpass_go_local/remote. Also changed updateq.h to flush non-ec data
 * on demand rather than on every lock acquire/release.
 *
 * Revision 1.5  1993/03/14  03:02:26  wsawdon
 * call uxpass_go_remote after dumping gmon file.
 *
 * Revision 1.4  1993/02/07  20:37:05  wsawdon
 * Allocate dirty bits from own region. Dirty bit data is initialized
 * as it is faulted in.
 *
 * Revision 1.3  1992/12/02  20:46:46  mattz
 * ATM:
 * 	Add interrupt support.  Loop during request for reply even with ints.
 * Everywhere:
 * 	Add MIDWAY_MUTEX_NAMES, and MIDWAY_MUTEX_SET_NAME -- add names to
 * 	mutexes so that I can gather collision data with my cthreads library,
 * 	and have some kind of meaningful printout.  Add mapped timer and
 * 	mapped free running counter support files.
 *
 * Revision 1.2  1992/10/20  03:39:04  wsawdon
 * Increased region size to 2MB
 * Fixed bug in realmain that prevented slave from being profiled.
 * Also changed to prefix gmon.out with hostName-Id# to allow more
 * than 1 process at a time to be profiled.
 * Changed lock initialization in RemoteReleaseReply to prevent slave
 * from copying uninitialized lock data.
 *
 * Revision 1.1  1992/10/08  19:54:39  wsawdon
 * Added support for gprof.
 * Enable profiling by -MIDPROFILE command line parameter.
 * Also, fixed various errors found by gcc (ANSI) compiler.
 *
 */


#include <sys/param.h>
#include <sys/types.h>
#include <mach.h>
#include <cthreads.h>
#include "timer.h"
#define KERNEL 1
#include "gprof.h"		/* chris modified */
#include <sys/time.h>		/* so doesn't include /usr/cs/include/time.h */

#include <fcntl.h>		/* open, creat, close, unlink, etc. */
#include <midway-dbg.h>		/* MIDWAY_MUTEX_SET_NAME */

#define panic(s) {printf(s); exit(10);}

#ifndef NULL
#  define NULL 0
#endif


/*
 * GPROF for UX.
 *
 * _mcount protects its data structures with a mutex.  Therefore
 * we have to be very careful to not call any functions defined
 * outside of this file since they can recursively call _mcount.
 * Note that the BSD trick of setting a variable to turn off
 * _mcount in recursive calls doesn't work since we're multithreaded.
 *
 * The only functions defined outside of this file are various
 * kernel stubs (eg vm_allocate) from libmach and the timeout
 * function which is used to put the pc sampling function on the
 * callout queue.  However, timeout is never called from _mcount
 * so we're ok.
 */

struct mutex gprof_mutex;	/* protects all profiling tables */
int gprof_do_profiling = 0;	/* do profiling when non-zero */
int gprof_init_complete = 0;	/* =1 when one-time initialization is complete */
unsigned long gprof_bounds = 0;	/* counter of bounds check errors */
unsigned long gprof_text_low, gprof_text_high;

int gprof_sample_type = SAMPLED_PC_PERIODIC;

/*
 * Call graph data structures.
 */
callarc_t **gprof_calls = (callarc_t **) NULL; /* hash table of callarcs */
callarc_t *gprof_callarc_freelist = (callarc_t *) NULL;
callarc_t *gprof_callarc_pagelist = (callarc_t *) NULL;

/*
 * PC histogram data structures.
 */ 
struct mutex gprof_sample_mutex;
CHUNK *gprof_pchist;		/* histogram of pc values */
unsigned long gprof_pchist_size;

int gprof_tick_sample;		/* ticks per sample */
int gprof_sample_interval = 50; /* how long in mS to sleep before next sample*/
static mach_port_t gprof_task;
sampled_pc_seqno_t gprof_seqno, gprof_oseqno;
unsigned long gprof_seqnum_gaps = 0;
sampled_pc_t gprof_pc_samples[512];

/*
 * Memory allocator for callarcs.
 * Can't use zones since they call us recursively.
 * Assumes caller has locked gprof_mutex.
 */
void
gprof_callarc_get(n)
	int n;			/* this many callarcs */
{
    kern_return_t kr;
    vm_address_t page;
    vm_size_t size;
    int num_structs, i;
    register callarc_t **arcp, *arcs;

    size = n * sizeof(callarc_t);
    kr = VM_ALLOCATE(mach_task_self(), &page, size, TRUE);
    if (kr != KERN_SUCCESS) {
	panic("_mcount: allocating callarcs");
    }

    /*
     * First callarc of each page is a pointer to the next region
     * of callarcs.  ca_count holds the size.
     */
    arcs = (callarc_t *)page;
    arcs->ca_count = size;
    arcs->ca_next = gprof_callarc_pagelist;
    gprof_callarc_pagelist = arcs;
    arcs++;			/* point to next callarc */

    /*
     * Build a freelist of callarcs in the new pages.
     */
    num_structs = (size - sizeof(callarc_t)) / sizeof(callarc_t);
    arcp = &gprof_callarc_freelist;	/* freelist should be NULL */
    for (i = 0; i < num_structs; i++) {
	*arcp = arcs;
	arcp = &arcs->ca_next;
	arcs++;
    }
    *arcp = (callarc_t *) NULL; /* null terminate list */
}

/*
 * Allocate memory so we can do sampling.
 */
void
monstartup()
{
    kern_return_t kr;
    unsigned int objtype;
    vm_address_t callarcs;
    extern int eprol();		/* first function in text seg */
    extern int etext;		/* comes right after text seg */

    gprof_task = mach_task_self();

    gprof_text_low = (unsigned long) eprol;
    gprof_text_high = (unsigned long) &etext;
    gprof_pchist_size = ((gprof_text_high - gprof_text_low) / CG_RESOL) + 1;
    gprof_text_high = gprof_text_low + (gprof_pchist_size - 1) * CG_RESOL;
    
    /*
     * Allocate pc histogram.
     */
    kr = VM_ALLOCATE(gprof_task,
		     (vm_address_t *) &gprof_pchist,
		     gprof_pchist_size * sizeof(CHUNK),
		     TRUE);	/* anywhere */
    if (kr != KERN_SUCCESS)
	panic("monstartup can't allocate pchist");

    /*
     * Allocate callarc table.
     */
    kr = VM_ALLOCATE(gprof_task,
		     (vm_address_t *) &gprof_calls,
		     gprof_pchist_size * sizeof(callarc_t *),
		     TRUE);	/* anywhere */
    if (kr != KERN_SUCCESS)
	panic("monstartup can't allocate callarc table");

    /*
     * Allocate callarcs.
     */
    gprof_callarc_get(65336);
	
    mutex_init(&gprof_mutex);
    MIDWAY_MUTEX_SET_NAME(&gprof_mutex, "gprof_mutex");
    mutex_init(&gprof_sample_mutex);
    MIDWAY_MUTEX_SET_NAME(&gprof_sample_mutex, "gprof_sample_mutex");
}

/*
 * mcountaux -- called by _mcount
 */
#define MCOUNT_ACTIVE_BIT	 0x00000001
#define MCOUNT_ACTIVE_BIT_CLEAR (~MCOUNT_ACTIVE_BIT)
void
mcountaux(from, to)
	unsigned long from, to;
{
    register callarc_t *carc;
    struct timeval now;
    unsigned long now_ms;
    int samplecnt, old_prof_value;
    register int i;
    cthread_t me;

    /*
     * filter out anything down the call chain from mcount.  This way we
     * can use system services which might be profiled, and not shoot
     * ourselves in the foot.
     *
     * Use field "return" in struct cthread -- it is zeroed on fork, and
     * not assigned until return, so we should be safe.  This is quite
     * the hack, but will work with both my cthreads library and the
     * default one.
     */
#ifndef notyet
    me = cthread_self();	/* uses assy routine -- so is safe. */
    if ((me == 0) || (((int)(me->result)) & MCOUNT_ACTIVE_BIT)) {
      /* bail */
      return;
    }
    /* set "down the pipe from mcount" flag */
    me->result = (any_t) (((int)(me->result)) | MCOUNT_ACTIVE_BIT);
#endif

    mutex_lock(&gprof_mutex);

    /* maybe someone changed it while we were blocked */
    if (gprof_do_profiling == 0)
	goto mcount_done;

#ifndef ORIGINAL
    /* Handle calls from midway dirtybit template code.   WAS 4Dec93 */
    if (from > gprof_text_high) {
      from = gprof_text_high;
      gprof_bounds++;
    }
#endif

    /*
     * Find the callarc for the given values of FROM and TO.
     */
    {
	/* Bounds Check */
	register unsigned long idx;

	/* note: UX does it (from - gprof_text_low)>>CG_SHIFT - 7jul93mjz */
	idx = (from >> CG_SHIFT) - (gprof_text_low >> CG_SHIFT);
	if (idx >= gprof_pchist_size) {
	    gprof_bounds++;
	    goto mcount_done;
	}
	carc = gprof_calls[idx];
    }
    /* THIS IS DIFFERENT IN UX.  ASK WAYNE.  7jul93mjz */
    while (carc != (callarc_t *) NULL) {
	if (carc->ca_to == to) {
	    (carc->ca_count)++;
	    break;
	}
	else
	    carc = carc->ca_next;
    }
    if (carc != (callarc_t *) NULL)
	goto mcount_done;

    /*
     * If we didn't find a callarc, grab one.
     */
    if (gprof_callarc_freelist == (callarc_t *)NULL) {
	gprof_callarc_get(32768);
	if (gprof_callarc_freelist == (callarc_t *)NULL)
	    panic("_mcount: out of callarcs");
    }
	
    /* get a callarc from freelist */
    carc = gprof_callarc_freelist;
    gprof_callarc_freelist = carc->ca_next;
    
    /* initialize the new callarc */
    carc->ca_from = from;
    carc->ca_to = to;
    carc->ca_count = 1;
    carc->ca_next = gprof_calls[(from>>CG_SHIFT) - (gprof_text_low>>CG_SHIFT)];
    /* we already know this is in bounds: */
    gprof_calls[(from>>CG_SHIFT) - (gprof_text_low>>CG_SHIFT)] = carc;

 mcount_done:
    mutex_unlock(&gprof_mutex);
#ifndef notyet
    /* should be single threaded in this cthread, so don't have to lock */
    me->result = (any_t)(((int)(me->result) & MCOUNT_ACTIVE_BIT_CLEAR));
#endif
    return;
}

/*
 * Takes pc samples.  Called from pc_sample_thread.
 */

int gprof_total_samples;
/*ARGSUSED*/
void
gprof_take_samples()
{
    kern_return_t kr;
    int samplecnt, i;
    vm_offset_t pc;

    mutex_lock(&gprof_sample_mutex);

    if (gprof_do_profiling == 0) {
	mutex_unlock(&gprof_sample_mutex);
	return;
    }

    samplecnt = sizeof(gprof_pc_samples) / sizeof(sampled_pc_t);
    kr = task_get_sampled_pcs(gprof_task, &gprof_seqno,
			      gprof_pc_samples, &samplecnt);
    if (kr != KERN_SUCCESS) {
	mach_error("sampling pcs", kr);
	mutex_unlock(&gprof_sample_mutex);
	return;
    }

    /* store samples for our task in the pc histogram */
    for (i = 0; i < samplecnt; i++) {
            gprof_total_samples++;
	    pc = gprof_pc_samples[i].pc;

	    if ((pc >= gprof_text_low) && (pc <= gprof_text_high))
		(gprof_pchist[(pc-gprof_text_low)>>CG_SHIFT])++;
#ifndef ORIGINAL
	    /* handle samples in midway dirtybit template code. WAS 4Dec93 */
	    else if (pc > gprof_text_high)
	      gprof_pchist[gprof_pchist_size-1]++;
	    else
	      gprof_pchist[0]++;
#endif
	}

    /* see if we missed any samples */
    if ((gprof_oseqno + samplecnt) != gprof_seqno)
	printf("gprof_take_samples: gap from %d to %d\n",
	       gprof_oseqno, gprof_seqno - samplecnt);

    gprof_oseqno = gprof_seqno;

    mutex_unlock(&gprof_sample_mutex);
}

kern_return_t
gprof_start_profiling()		/* assumes caller has locked gprof_mutex */
{
	register int i, samplecnt;
	register callarc_t *carc, *carc_next;
	kern_return_t kr;

	/*
	 * (a) zero histogram
	 * (b) (re)initialize call graph
	 * (c) (re)start pc sampling
	 */
	
	bzero(gprof_pchist, gprof_pchist_size * sizeof(CHUNK));

	for (i = 0; i < gprof_pchist_size; i++) {
		carc = gprof_calls[i];
		while (carc != (callarc_t *) NULL) {
			carc_next = carc->ca_next;
			carc->ca_next = gprof_callarc_freelist;
			gprof_callarc_freelist = carc;
			carc = carc_next;
		}
		gprof_calls[i] = (callarc_t *) NULL;
	}

	if (!gprof_do_profiling) {
	  gprof_seqno = gprof_oseqno = 0;
	  kr = task_enable_pc_sampling(gprof_task, &gprof_tick_sample,
				       gprof_sample_type);
	  if (kr != KERN_SUCCESS)
	    return kr;
	
	  gprof_do_profiling++;
	}

	return KERN_SUCCESS;
}

kern_return_t
gprof_stop_profiling()		/* assumes caller has locked gprof_mutex */
{
    int samplecnt;
    kern_return_t kr;
	
    if (gprof_do_profiling == 0)
	return KERN_SUCCESS;
	
    gprof_do_profiling = 0;

#if 0 /*old way */
    do {
        samplecnt = 0;
	kr = task_disable_pc_sampling(gprof_task, &samplecnt);
	if (kr != KERN_SUCCESS)
	    return kr;
    } while (samplecnt);
#else
    /*
     * As of mk84, samplecnt is defined as OUT in MiG.  This means
     * it is blindly overwritten by the MiG code, and never given
     * to the routine in the first place.  In addition, in MK84,
     * the variable is never set.  Sooo... if the MiG reply buffer
     * ends up being non-zero, this will always return non-zero (at
     * least until a different reply buffer is allocated) leaving
     * us with very long, and sometimes infinite, loop.  Bah Bah Bah.
     * - 8feb94, mjz
     */
    kr = task_disable_pc_sampling(gprof_task, &samplecnt);
    if (kr != KERN_SUCCESS)
      return kr;
#endif /*~0*/

    return KERN_SUCCESS;
}

/*
 * Server monitoring messages from bsd_1.defs
 */

kern_return_t
bsd_mon_switch(onoff)
	int		*onoff;
{
	int old_value;
	kern_return_t kr;

	mutex_lock(&gprof_sample_mutex);
	mutex_lock(&gprof_mutex);

	old_value = gprof_do_profiling;

	if (*onoff)
		kr = gprof_start_profiling();
	else
		kr = gprof_stop_profiling();

	*onoff = old_value;
	mutex_unlock(&gprof_mutex);
	mutex_unlock(&gprof_sample_mutex);
	
	return kr;
}

kern_return_t
bsd_mon_dump(mon_data, mon_data_cnt)
	char		**mon_data;
	int		*mon_data_cnt;
{
	vm_address_t	mon_buffer;
	vm_size_t	mon_size;
	int		callarc_cnt, i;
	callarc_t	*ca;
	kern_return_t	kr;
	struct gprof_header *gh;
	CHUNK		*hist;

	mutex_lock(&gprof_sample_mutex);
	mutex_lock(&gprof_mutex);

	/* Count callarcs. */
	callarc_cnt = 0;
	for (i = 0; i < gprof_pchist_size; i++)
		for (ca = gprof_calls[i];
		     ca != (callarc_t *) NULL;
		     ca = ca->ca_next)
			callarc_cnt++;

	/* How big is the buffer? */
	mon_size = sizeof(struct gprof_header)
	                + gprof_pchist_size * sizeof(CHUNK)
			+ callarc_cnt * sizeof(struct gprof_call);
  
	kr = VM_ALLOCATE(mach_task_self(), &mon_buffer, mon_size, TRUE);
	if (kr != KERN_SUCCESS) {
		mutex_unlock(&gprof_mutex);
		return kr;
	}

	gh = (struct gprof_header *) mon_buffer;
	gh->low = gprof_text_low;
	gh->high = gprof_text_high;
	gh->nbytes = sizeof(struct gprof_header)
		+ gprof_pchist_size * sizeof(CHUNK);

	hist = (CHUNK *) (gh + 1);
	bcopy((char *)gprof_pchist, (char *)hist,
	      gprof_pchist_size * sizeof(CHUNK));

 	hist += gprof_pchist_size;
	{
		register struct gprof_call *gc;

		gc = (struct gprof_call *) hist;

		for (i = 0; i < gprof_pchist_size; i++)
			for (ca = gprof_calls[i];
			     ca != (callarc_t *) NULL;
			     ca = ca->ca_next) {
				/*
				 * We have to use bcopy since gc might
				 * not be word-aligned.
				 */
				bcopy(ca, gc, sizeof(struct gprof_call));
				gc++;
			}
	}

	*mon_data = (char *) mon_buffer;
	*mon_data_cnt = mon_size;
	
	mutex_unlock(&gprof_mutex);
	mutex_unlock(&gprof_sample_mutex);
	return KERN_SUCCESS;
}

int
pc_sample_thread(any_t arg)
{
        /* FOLLOWING DEFINED IN comm.c */
	extern int midway_gprof_priority; /* set by comm, if necessary */
	if (midway_gprof_priority) {
	  midway_set_good_priority(midway_gprof_priority);
	}
        timer_init();
	while(1) {
		timer_pause_ms( gprof_sample_interval );
		gprof_take_samples();
	}
	/*NOTREACHED*/
	return 0;
}


kern_return_t
gprof_mon_start ()
{
  kern_return_t krc;
  int onoff;

  if (!gprof_init_complete) {

    /* Spawn thread to perform sampling every timer tick */
    cthread_detach ( cthread_fork ( pc_sample_thread, (any_t) 10 ));
  
    /* initialize gprof */
    monstartup();
    
    gprof_init_complete = 1;
  }

  /* turn on sampling */
  onoff = 1;
  krc = bsd_mon_switch(&onoff);	

  return (krc);
}

kern_return_t
gprof_mon_stop (char *fileName)
{
    char *mon_data;
    int  mon_data_cnt;
    int off, size, result;
    int fd;
    int onoff;
    kern_return_t krc;

#define WRITE_SIZE_LIMIT 4096

    if (!gprof_init_complete)
      return (KERN_FAILURE);

    /* turn off sampling */
    onoff = 0;
    krc = bsd_mon_switch(&onoff);
    if (krc != KERN_SUCCESS) return (krc);

    /* obtain pc sampling data */
    krc = bsd_mon_dump(&mon_data, &mon_data_cnt);
    if (krc != KERN_SUCCESS) return (krc);
    
    printf("Writing %d bytes to file \"%s\"\n", mon_data_cnt, fileName);
    printf("gprof_total_samples: %d\n", gprof_total_samples);

    /* force writes to go to local host */
    onoff = midway_uxpass_local ();

    /* write file */
    fd = creat(fileName, 0666);
    if (fd < 0) return (KERN_FAILURE);

    for (off = 0; off < mon_data_cnt; off += WRITE_SIZE_LIMIT) {
      size = mon_data_cnt - off; 
      if (size > WRITE_SIZE_LIMIT) size = WRITE_SIZE_LIMIT;
      result = write (fd, &mon_data[off], size);
      if (result != size) {
	printf("Write of file \"%s\" failed, rc=%d\n", fileName, result);
	break;
      }
    }
    close(fd);

    /* restore remote writes */
    if (onoff)
      onoff = midway_uxpass_remote ();

    return (KERN_SUCCESS);
}
