/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * linex.c: Routines for line exceptions when dirting cache lines.
 * 
 *
 * 15 Aug 92 (was): Initial code 
 * 25 Oct 92 (was): Fixed bugs
 * 13 May 93 (was): Added code to set dense dirty bits
 * 19 Sep 93 (was): Changed parms for calls from new dirty bit code
 *                  and added routine mid_area_exception
 *
 */
#include <sys/types.h>		/* u_long & friends */
#include <mach.h>
#include <cthreads.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>

#include "lock.h"		/* perlock struct, lock request formats */
#include "comm.h"		/* M_xxx_OP */
#include "smexecd_types.h"	/* ensure get char255 */
#include "pbuf.h"		/* packet buffer stuff */
#include "shmalloc.h"		/* for FIRST/LAST_STATIC_SYM */
#include "updateq.h"		/* prototypes for calling update queue rtns */
#include "midway-dbg.h"		/* filter debug messages */

#define __LINEX_INTERNAL__ 
#include "linex.h"              /* get internal definitons */

extern clock_t *vm_dense_dirty_base; /* base address for dense dirty bits */


/*
 ************************************************************************
 * 
 * mid_line_exception
 *
 * An update to shared memory encountered a cache-line marked
 * with the exception flag to indicate out-of-line handling.
 * This routine handles all line exceptions.
 *
 ************************************************************************
 */
void 
mid_line_exception
  (
   vm_address_t  dataOffset,	/* offset in shared region being written */
   u_long        dataLen,	/* length being written */
   clock_t      *dbitP,		/* address of dbit to set */
   region_info   riP		/* ptr to region info for dataAddr */
   )
{
  register clock_t dbitVal;
  u_long  enqueueLen;
  vm_address_t enqueueAddr;
  vm_address_t lineAddr;
  u_long  lineOffset;
  u_long  lineSizeInBytes;
  u_long  dataInThisLine;

  dbitVal = *dbitP;
  
  if (IS_DBG_PORT_LINEX) {
    char255 smsg;
    sprintf(smsg, "))) line exception: Offset=0x%x len=%d dbit=0x%x *dbit=0x%x (%s) riP=0x%x base=0x%x",
	    dataOffset, dataLen, dbitP, dbitVal,
	    (dbitVal == NULL_CLOCK_VALUE) ? "NULL" : 
	    (dbitVal & LINE_EXCEPTION_FLAG) ? "EC" : 
	    (dbitVal & LINE_ENQUEUE_FLAG) ? "ENQUEUE" : 
	    (dbitVal & LINE_UNALIGNED_FLAG) ? "UNALIGNED" : 
	    (dbitVal & LINE_CONVERSION_FLAG) ? "CONVERSION" : 
	    "UNKNOWN",
	    riP, riP->base);
    StatusMsg(smsg);
  }

  enqueueLen = 0;
  enqueueAddr = 0;
  lineAddr = (dataOffset >> riP->line_size) << riP->line_size;
  lineOffset = dataOffset - lineAddr;
  lineSizeInBytes = RI_LINE_SIZE_TO_BYTES(riP->line_size);
  dataInThisLine = lineSizeInBytes - lineOffset;
  ASSERT ((lineOffset >= 0) && (lineOffset < lineSizeInBytes));

  while (TRUE) {

    /* Check for enqueued changes */
    if ((enqueueLen > 0) && 
	((dbitVal & (LINE_EXCEPTION_FLAG|LINE_ENQUEUE_FLAG)) 
	 != (LINE_EXCEPTION_FLAG|LINE_ENQUEUE_FLAG))) {
      mid_updateq_add (enqueueAddr, enqueueLen);
      enqueueLen = 0;
    }

    /* Check flags in dbit field */
    if (dbitVal != NULL_CLOCK_VALUE) {
      if ((dbitVal & LINE_EXCEPTION_FLAG) == 0) {
	
	/* EC fast path. Just set info to zero & continue */
	*dbitP = NULL_CLOCK_VALUE;
	
      } else if ((dbitVal & LINE_ENQUEUE_FLAG) != 0) {
	
	/* record changes */
	if (enqueueLen == 0) enqueueAddr = dataOffset + riP->base;
	enqueueLen += lineSizeInBytes;
	
	/* reset flag to prevent duplicates */
	*dbitP = NULL_CLOCK_VALUE;
	
      } else if ((dbitVal & LINE_UNALIGNED_FLAG) != 0) {
	
	/* Handle dense dirty bits */
	{ register u_long ddbOffset, ddbLen;
	  register clock_t *ddbP, *ddbEndP;
	  register u_long line_exception_value = LINE_EXCEPTION_FLAG;
	  register u_long line_enqueue_value = LINE_ENQUEUE_FLAG;

	  ddbOffset = (dbitVal & LINE_DATA_VALUE) + lineOffset;
	  ddbP = &vm_dense_dirty_base[ ddbOffset ];
	  ddbLen = (dataLen <= dataInThisLine) ? dataLen : dataInThisLine;
	
	  if (IS_DBG_PORT_LINEX) {
	    char255 smsg;
	    sprintf(smsg, "))) dense dirty bits at offset=%d addr=0x%x len=%d 1st val=0x%x\n",
		    ddbOffset, ddbP, ddbLen, *ddbP);
	    StatusMsg(smsg);
	  }
	
	  for (ddbEndP = ddbP + ddbLen; ddbP < ddbEndP; ddbP++) {
	    dbitVal = *ddbP;
	    if (dbitVal != NULL_CLOCK_VALUE) {
	      if ((dbitVal & line_exception_value) == 0) {
		*ddbP = NULL_CLOCK_VALUE;
	      } else if ((dbitVal & line_enqueue_value) != 0) {
		if (enqueueLen == 0) 
		  enqueueAddr = riP->base + dataOffset;
		else if ((enqueueAddr + enqueueLen) != (riP->base + dataOffset)) {
		  mid_updateq_add (enqueueAddr, enqueueLen);
		  enqueueLen = 0;
		  enqueueAddr = riP->base + dataOffset;
		}
		enqueueLen += 1;   /* Byte */
		*ddbP = NULL_CLOCK_VALUE;
	      } else {
		fprintf(stderr, 
			"Invalid dense dirty bit info at 0x%x, addr=0x%x len=%d, *dbitP=0x%x\n",
			dbitP, (riP->base+dataOffset), dataLen, dbitVal);
		fflush (stderr);
		ASSERT (FALSE);
	      }
	    }
	  } /* for (ddbEndP = ddbP + ddbLen; ... */
	}
      } else {
	fprintf(stderr, 
		"Invalid dirty bit info at 0x%x, addr=0x%x len=%d, *dbitP=0x%x\n",
		dbitP, (riP->base+dataOffset), dataLen, dbitVal);
	fflush (stderr);
	ASSERT (FALSE);
      } 
    }

    if (dataInThisLine >= dataLen) break;
    
    /* advance to next cache line */
    dbitP++;
    dbitVal = *dbitP;
    dataOffset += dataInThisLine;
    dataLen  -= dataInThisLine;
    dataInThisLine = lineSizeInBytes;
    
  } /* while (dataLen > 0) */
      
  /* Check for enqueued changes */
  if (enqueueLen > 0) {
    mid_updateq_add (enqueueAddr, enqueueLen);
  }

  return;
} /* mid_line_exception */

