/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * lock.h : structures used by the locking protocol
 *
 * HISTORY
 * 19 Jun 91 (mjz): update for startup protocol change and move to indirect
 * 		lock numbers instead of physical addresses.
 * 22 Jun 91 (mjz): update for address maps in locks
 *  1 Jul 91 (mjz): keep transaction id for remote requests
 * 14-25 Jul 91 (mjz): update for barriers
 * 31 jul 91 (mjz): add lock/barrier switch in reply message
 * 20 Mar 92 (mjz):  xfer bar_curtime on Create request
 * 19 May 92 (mjz): resurrect nreadq; reincarnate datac as xferc; L_DATAIN now
 * 			defunct
 *  1 Jul 92 (mjz): add L_(X)ACQINV
 *  9 Jul 92 (mjz): add new "iwrite" field.  We need to know which thread
 * 		is waiting to satisfy a request so that we can see if a r/o req
 * 		corresponds to an inv ack we are waiting for...  Also need
 * 		lock "incarnation" -- this will be dataclock except if data
 * 		is not updated.  Must be passed along in lock message too :-(.
 * 13 Jul 92 (was): added types and fields to lock structure to support
 *                  porting code to midway
 * 11 Aug 92 (was): added clock_t type
 * 26 Oct 92 (mjz): allow for "flush" barriers
 *  8 Oct 92 (was): added hint segs to update lock location hints
 *  8 Feb 93 (was): added amap_tailP to optimize addrmap_add calls.
 * 20 Jul 93 (was): added prototypes for pack/unpack routines.
 *  8 Jan 94 (was): added fastpath for acquire/release locks (FASTLOCKS)
 *                  locktable moved to a fixed/known address in memory
 *  6 Mar 94 (was): added page-based EC support
 * 23 Aug 94 (mjz): add extended addrmaps for fast dirtybit scanning. 
 */
#include "midway_types.h"       /* global type definitions */
#include "queue.h"		/* threadq's and extq's */
#include "bitwiddle.h"		/* for BITARRAY */
#include "pbuf.h"		/* for pbuf_q */


/*
 * address map of data guarded by synchronization variables
 */
typedef struct s_addr_map {
  char	*addr;			/* start of region */
  char  *end;			/* address of byte after end of region */
  u_long len;			/* length := end - addr*/
#if !defined(PAGE_BASED)
  clock_t *dbit;			/* 1st dbit address */
  clock_t *dbit_end;		/* last dbit address */
  int   line_size;		/* cache line size for the dbits */
  int   line_offset;		/* if needed ? */
  int   max_size;		/* max size in bytes of dirty data + headers */
  /* 0 => always go slow path */
#endif
  struct s_addr_map *next;	/* queued together */
} s_addr_map_t;

/*
 * define structure to hold program counter
 */
typedef struct prog_cntr {
  char *fileP;                  /* ptr to file name */
  int   lineNo;                 /* line number within file */  
} prog_cntr_t;

typedef u_long uniq_id;         /* type for unique process ids */


#ifdef PAGE_BASED

/*
 * Define structure for saving diffs.
 * Each element describes the diffs for a given incarnation.
 * The number of incarnations is limited, but there is
 * no limit on the size of diffs for a single incarnation.
 */
typedef struct {
  clock_t       diff_incarn;	/* incarnation number  */
  u_long        diff_status;	/* valid, unknown */
  u_long	diff_size;	/* size of diffs in this list */
  pbufq_t	diff_queue;	/* queue of diffs */
} diff_list_t;

#define DIFF_STAT_UNKNOWN  0	/* Diff not yet generated */
#define DIFF_STAT_VALID    1	/* Diff is valid */

#define MAX_DIFF_INCARN	1 /*4*/	/* Max number of incarns saved */



#endif


/* 
 * per lock structure -- the record that records all lock information 
 *
 * Space for the per lock structures is allocated in the locktable.
 * The locktable is mapped in to virtual memory starting after the 
 * application's data segment and growing towards shared memory.
 * Since the application sees the lock's address as the lock's name,
 * the address must be the same on all machines.   WAS 8Jan94
 *
 * NOTE: The first two fields MUST match those defined as the external 
 * representation of locks (typedef midway_lock_t in midway.h). 
 * If the external representation were changed to be the same size
 * as the internal, then the application could allocate space for
 * the structure without calling midway_lock_alloc(). (Appl would
 * need to call midway_lock_init(lock) instead).   WAS 8Jan94
 */

typedef struct perlock {	/* per lock data structure */
  struct mutex  m;		/* to protect this structure, local Qs */
  int           held;           /* if != 0 then lock is held by FASTLOCK acquire */
  				/* ABOVE FIELDS NOT ALLOWED TO MOVE!  WAS 8JAN94 */
  u_long        lock;		/* lock name (address of lock struct) */
  u_long	state;		/* state bits for lock */
  long   	busy;		/* #threads using this lock locally */
  u_long	rdrs;		/* #sites reading (including this one) */
  u_long	nreadq;		/* number of locally queued readers */
  u_long	nwriteq;	/* number of locally queued writers */
  u_long	iwrite;		/* local writer requesting lock externally */
  u_long	eread;		/* queued external writer (thrd id; need 1) */
  extq_t	erdrs;		/* need list of reader info */
  u_long	ewrite;		/* likewise, for writers */
  extq_t	ewrtrs;		/*  */
  BITARRAY	*aerdrs;	/* Active External ReaDeRS */
  struct condition ireadc;	/* to let internal readers go */
  struct condition ereadc;	/* to let external readers go */
  struct condition iwritec;	/* internal writers condition */ /* per-t?*/
  struct condition ewritec;	/* external writers condition */
  struct condition xferc;	/* xfer{site,tid} inconsistent */
  struct condition iupc;	/* Internal UPgrade; w wait for r req to complete */
  u_long	xfersite;	/* site we are transferring to */
  u_long	xfertid;	/* Trans. ID last time we transferred */
  u_long	bestguess;	/* owner */
  u_long	nextbest;	/* old owner  */
  clock_t	bgtime;		/* best guess time (Uses Incarnation NOT dataclock)*/
  clock_t	dataclock;	/* data time (newest data we've got) */
  u_long	incarnation;	/* need to disambiguate invalidates &r/o req */
  clock_t       notime;         /* NewOwner time (Incarnation NOT dataclock) */
  clock_t	tcreate;	/* create time */
  long		ROxfers;	/* For page_based - is RO xfer in progress? */
  s_addr_map_t	*amap;		/* address map for data guarded by SV */
  s_addr_map_t	*amap_tailP;	/* tail of address map list */
  s_addr_map_t	*amap_prevP;	/* previous address map added to list */
  u_long	amap_data_size; /* sum of sizes of addr maps */
  pbufq_t	output;		/* building new packets for barrier (mgr) */
  pbufq_t	old_output;	/* previous iteration results */

  /*
   * Define fields used only by midway porting code
   * If space is a problem, these can be eliminated
   * by judicious use of #ifdef MIDWAY_PORT commands.  WAS 1Jul92
   */
  uniq_id       xholder;        /* Unique id of thread holding write lock */
  prog_cntr_t   xalloc;         /* pc when lock was allocated */
  prog_cntr_t   xacquire;       /* pc when lock was acquired */
  struct change_log *clogP;     /* ptr to change log for this lock/barrier */

#ifdef PAGE_BASED
  /*
   * Define queue for saved diff buffers.
   */
  diff_list_t   diff[MAX_DIFF_INCARN];
#endif

} pl_t;

/*
 * barriers use some fields differently
 */
  /* --map of what machines participated (used by manager) */
#define bar_map aerdrs
  /* --number of threads that have hit barrier in this iteration [master] */
#define bar_subtotal nwriteq
  /* --number of machines supposed to participate [mast/slave]*/
#define bar_total ewrite
  /* --number of machines that have checked in [mast/slave]*/
#define bar_num eread
  /* --where slaves can wait for barrier to be released (>1 thread/proc) */
#define bar_wait iwritec
  /* --current iteration identification (time) */
#define bar_curtime dataclock
  /* --previous iteration identification (time) [mgr] */
#define bar_lasttime bgtime
  /* --need to keep trans id, thread id of hosts... done here */
#define bar_waiters ewrtrs

/*
 * state bits
 */
#define L_OWN		0x80000000 /* I own this lock */
#define L_ACQ		0x40000000 /* I am acquiring this lock */
#define L_ACQRW		0x20000000 /* I am acquiring this lock exclusively */
#define L_XFER		0x10000000 /* I am xferring ownership to someone else */
#define L_DATAIN	0x08000000 /* Data update (transfer in) in progress */
#define L_ACQINV	0x04000000 /* Invalidate arvd for pending RO acquire */
#define L_BARRIER	0x00000001 /* this is a barrier, not a lock! */
#define L_BAR_NOREG	0x00000002 /* no one has registered to use the bar */
#define L_BAR_FLUSH	0x00000004 /* do not redistribute data; master keeps */

#define L_XOWN		0x7fffffff /* turn off L_OWN */
#define L_XACQ		0xbfffffff
#define L_XACQRW	0xdfffffff
#define L_XXFER		0xefffffff
#define L_XDATAIN	0xf7ffffff
#define L_XACQINV	0xfbffffff
#define L_XBAR_NOREG	0xfffffffe

/* for BITARRAYs of hosts, like .aerdrs */
/* The high order bit is host 0, which does not participate in the locking
 * protocol -- so we steal it to indicate the bitarray his info in it */
#define BAH_USED	0x80000000



/*
 * Define number of lock structures to allocate/map each time.
 * Must be a page multiple.
 */
#define LOCKTABLE_ALLOC_SIZE (16 * vm_page_size)

/*
 * constants used for mid_unpack_data (which pulls apart ADDRSEGs)
 */
typedef enum {
  mid_MUD_lock,			/* a lock reply */
  mid_MUD_barrier,		/* a barrier reply */
  mid_MUD_static		/* static data during thread create */
  } MUD_t;

/* lock specific message contents */

typedef struct addrpair {
  char		*addr;		/* address of this block */
  u_long	len;		/* length of this block */
  u_char	data[4];	/* data starts here (could be any length) */
				/* ...padded w 0's to be multiple of 4 */
} addrpair_t;

#define ADDRPAIR_SIZE (sizeof(addrpair_t) - 4)

typedef struct addrseg {
  u_long	segtype;	/* to differentiate segments */
  clock_t	clock;		/* clock value for these addresses */
  u_long	num;		/* number of addrpairs that follow */
  addrpair_t	ap;		/* the address pair */
  /* more addrpair's could follow for this clock value */
} addrseg_t;

#define ADDRSEG_SIZE (sizeof(addrseg_t) - sizeof(addrpair_t))

#define ADDRSEG_ADDRPAIR_SIZE (sizeof(addrseg_t) - 4)

typedef struct queuer {
  u_long	host;		/* host of requestor */
  u_long	thread;		/* thread of requestor */
  clock_t	clock;		/* requestor's time */
  u_long	tid;		/* transaction id */
  u_long	mode;		/* request mode */
} queuer_t;

typedef struct qseg {
  u_long	segtype;	/* to differentiate segments */
  u_long	num;		/* # of segments */
  queuer_t	first;		/* first (and perhaps only) q'er */
  /* num-1 queuers follow */
} qseg_t;

typedef struct hint {
  u_long        lock;		/* location hint for this lock */
  clock_t       incarn;		/* lock's incarnation value hint was good */
} hint_t;

typedef struct hseg {
  u_long	segtype;	/* to differentiate segments */
  u_long	num;		/* # of segments */
  hint_t	first;		/* first (and perhaps only) hint */
} hseg_t;

#define DATASEG		1	/* this is an addrseg */
#define QSEG		2	/* this is a queue segment */
#define HINTSEG         4       /* updates to lock location hints */
/* #ifdef PAGE_BASED -- currently need since no such #ifdef around code in
   vm_paged.c -- 24aug94 mjz */
#  define DIFFSEG       8	/* this is a diffseg */
/* #endif */


/* lock request would have packet #'s (inserted by comm stuff), object (also
   in comm part of msg) all that's left is mode and last known object time */
typedef struct m_lock_req {
  u_long	incarnation;	/* need for r/o requests */
  clock_t	dataClock;	/* most recent update we've had */
  u_long	mode;		/* mode of this req !0 == excl */
} m_lock_req_t;

/* these used for RemoteRelease RemoteWait -- used to transfer global
 * lock info at thread creation... */
typedef struct m_lock_rem_amap { /* addrmap segment */
  char *	addr;		/* better be 4 bytes... *//*machdep*/
  u_long	len;		/* length */
} m_lock_rem_amap_t;
typedef struct m_lock_remote_seg {
  u_long	name;		/* name of lock */
  u_long	type;		/* lock/barrier */
  u_long	bestguess;	/* best guess */
  clock_t	tcreate;	/* create time */
  clock_t	bar_curtime;	/* xfer for barriers only */
  u_long	anum;		/* # m_lock_rem_amap's following */
#ifdef PAGE_BASED
  u_long	anum_data;	/* page-based debugging: total for amaps */
#endif
} m_lock_remote_seg_t;
typedef struct m_lock_remote {
  void		(*func)();	/* function to be performed */
  u_long	arg;		/* argument for function */
  u_long	nsegs;		/* number of m_lock_remote_segments after */
} m_lock_remote_t;


/*
 * Barriers
 */
typedef struct m_bar_req {
  u_long	curtime;	/* time of most recent iteration */
  u_long	num;		/* number of threads that modified data */
  u_long	totnum;		/* total number of waiters wanted */
  u_long	noreply;	/* nonzero => "flush" barrier. don't reply */
} m_bar_req_t;


/*
 * Prototypes
 */
int
mid_pack_data
  (
     s_addr_map_t **amapPP,	/* the map in/out -- in case more needed */
     char         **amap_addrPP,/* in/out -- ptr to uncopied portion of map */
     long          *amap_lenP,  /* in/out -- uncopied length of map */
     clock_t	old_clock,	/* other side is accurate as of this time */
     clock_t	*max_clockP,	/* maximum clock time found so far */
     char	*pbufP,		/* pointer into packetbuffer */
     long	*pbufLenP,	/* length of stuff already in buffer */
     char	**data_addrPP,	/* data address working on */
     long	*data_lenP,	/* length that needs to be copied */
     clock_t	*data_clockP,	/* time that needs to be copied */
     long	*xfer_lenP	/* for stats, actual amount xferred */
   );

void
mid_unpack_data 
  (
     char **pptr,
     u_long *lptr,
     s_addr_map_t *amap_listP,	/* Debug only: addrmaps for lock/barrier */
     char *cname,		/* caller for debug messages */
     MUD_t ctype 		/* another way to interpret the caller */
   );

pl_t *
mid_alloc_synch_object
  (
     u_long     lname,		/* =NULL for a new lock, =ADDR for existing lock */
     u_long	state		/* Initial state for object */
  );

#ifdef PAGE_BASED

#define PAGE_PACK_DIFFS   1
#define PAGE_PACK_DATA    0

int
mid_paged_pack_data (
		     u_long		synch_obj,	/* used for debugging only */
		     int		pack_diffs, 	/* =1 to pack diffs, =0 to pack all data  */
		     int		mark_clean, 	/* =1 to mark data as clean */
		     s_addr_map_t	**amapPP,	/* the map in/out -- in case more needed */
		     char         	**amap_addrPP,	/* in/out -- ptr to uncopied portion of map */
		     long          	*amap_lenP,  	/* in/out -- uncopied length of map */
		     clock_t		 clock_val,     /* in -- lock's incarnation number */
		     char		*pbufP,		/* pointer into packetbuffer */
		     long		*pbufLenP,	/* length of stuff already in buffer */
		     long		*xfer_lenP 	/* for stats, actual amount xferred */
		    );
void
mid_paged_unpack_data (
		       u_long		synch_obj,	/* used for debugging only */
		       char **pptr,
		       u_long *lptr,
		       s_addr_map_t *amap_listP,	/* DEBUG only: addr maps for this lock/barrier */
		       char *cname,		/* caller for debug messages */
		       MUD_t ctype 		/* another way to interpret the caller */
		       );

#endif
