/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * lockaddr.c: routines for diddling with the lock address maps
 *
 * 22 jun 91, mjz
 * 24 jun 91 (mjz): add running thru irty bits to find addrs to xfer
 * 22 Aug 91 (mjz): merge w/libmidway, fixup midway_mem.h
 * 18 Dec 91 (mjz): add mid_deltamap_free()
 *  9 Mar 92 (mjz): Fix _delta; we were not handling addrmaps that
 *			crossed cache lines (but did not fill up the
 *			last cache line) correctly.  In particular,
 *			we would not transfer the last chunk.
 *			Add LAST_STATIC_SYM instead of &end .
 * 16 Jul 92 (mjz): fix assert in mid_addrmap_delta to not check
 * 	for a megabyte boundry, but be in the *same* megabyte.
 *	see also midway.c.
 * 13 Aug 92 (was): changed mid_addrmap_xxx code to take a list ptr
 *                  as input rather than a lock structure, so that the
 *                  code can be reused for other lists.
 *  3 Sep 92 (mjz): fix bug introduced by Wayne...
 * 16 Sep 92 (was): slight optimizations to addrmap_delta & deltamap_free
 *  1 Dec 92 (mjz): deltamap-> packdelta: pack to pbuf at same time
 *  8 Feb 93 (was): Optimized mid_addrmap_add
 * 21 Feb 93 (was): ifdef'd out code to sort addrmaps when added.
 * 22 Mar 93 (was): fixed bug in computing data transferred in mid_addrmap_delta_pack
 *  6 Jul 93 (was): Added code to handle dense dirty bits
 *  1 Dec 93 (mjz): remove ASSERT(FALSE) in unpack code.
 *  5 Jan 94 (was): Get region_info for remote regions the first time we unpack into it.
 *                  also ifdef'ed out some debug code in mid_unpack_data
 * 16 Feb 94 (was): optimized inner loops of mid_pack_data/mid_unpack_data
 * 15 Jun 94 (was): re-enabled addrmap sorting
 * 23 Aug 94 (mjz): re-disable addrmap sorting for checked-in version
 *                  extended addrmap support; some cleanup.
 *  4 Sep 94 (was): Added addrmap_purge call for midway_rebind_synch_object.
 * 30 sep 94 (mjz): switch optimization in inner pack loop; don't check
 * 	for =0 and =clock, but just for =0.  This should never happen.
 *	if it does, we will loop around again, and create a new addrseg.
 */
#include <sys/types.h>		/* u_long & friends */
#include <mach.h>		/* for cthreads */
#include <cthreads.h>
#include <stdio.h>		/* NULL */
#include "lock.h"
#include "smexecd_types.h"	/* char255 */
#include "midway-dbg.h"		/* IS_DBG_LOCK_DLTA */
#include <flow_times.h>	/* deltaloop */
/* get rid of "shared" declaration in brian's midway.h -- means type of
 * shmalloc is wrong, but I don't use that here... (don't want to have
 * to use hacked gcc to compile this if I don't wanna */
#define OMITSHARED 1
#include "midway_mem.h"
#include "shmalloc.h" 


/*#define SORT_ADDRMAPS	1	*//* sort lock addr maps */

#define VERIFY_XAMAP_DBITS 1	/* verify new dbit mappings on xfer */

#if BENCHMARK
  /* for now, no non-ec if benchmarked. 22jun94mjz&wsawdon */
# define ADDR_ENABLE_NON_EC 0
#else
# define ADDR_ENABLE_NON_EC 1
#endif

int mid_addrmap_fast_cutoff = 64;	/* # dirtybits we will tolerate */
typedef enum { AMAP_SINGLE, AMAP_MULTIPLE } amap_num_t;

/*
 * Define external vars used for dense dirty bits
 */
extern clock_t *vm_dense_dirty_base;

/* 32 bit machine only warning */
#define FRC_MAX ((unsigned int) 0xffffffff)

#ifdef STATS

/*
 * Define table for collecting statistics
 */
struct {
  u_long mapadds;		/* # of calls to addrmap_add */
  u_long mapadd_empty;		/* # of adds to an empty list */
  u_long mapadd_tail;		/* # of adds to tail of list */
  u_long mapadd_head;		/* # of adds to head of list */
  u_long mapadd_next;		/* # of elements traversed to find location */
  u_long mapadd_after;		/* # of elements added between two existing */
  u_long mapadd_current_start;	/* # of elements added to start of existing */
  u_long mapadd_current_end;	/* # of elements added to end of existing */
  u_long mapadd_current;	/* # of elements completely enclosed by an existing */
  u_long mapadd_merged;		/* # of elements merged & deleted from list */
  u_long mapadd_afterprev;	/* # of elements immediately after prev added */
  u_long mapadd_beforeprev;	/* # of elements immediately before prev added */
  u_long map_purges;		/* # of calls to addrmap_purge */

  u_long pack_calls;		/* # of calls to mid_pack_data */
  u_long pack_dbit_lookups;	/* # of dirtybit info lookups */
  u_long pack_dbit_reads;	/* # of dirtybits read */
  u_long pack_dbit_writes;	/* # of dirtybits found null and written a new clock value */
  u_long pack_dbit_wbytes;	/* # of bytes for dbit_writes (*linesize) */
  u_long pack_ddbit_lookups;	/* # of dense dirtybit info lookups */
  u_long pack_ddbit_reads;	/* ~# of dense dirtybits read (may be > than actual #) */

  u_long unpack_calls;		/* # of calls to mid_unpack_data */
  u_long unpack_dbit_lookups;	/* # of dirtybit info lookups */
  u_long unpack_dbit_writes;	/* # of dirtybits written */
  u_long unpack_dbit_errors;	/* # of dirtybits found null or clock > new clock value */
  u_long unpack_ddbit_lookups;	/* # of dense dirtybit info lookups */
  u_long unpack_ddbit_writes;	/* ~# of dense dirtybits written (may be > than actual #) */
  u_long unpack_data_bytes;	/* # of bytes of data unpacked */

} midLockAddrStats;
static mutex_t statAddrLock;


/* 
 * Define macro to conditionally maintain stat info
 */
#  ifdef FAST_STATS
#    define INCR_STATS(x,by) (x) += (by)
#  else
#    define INCR_STATS(x,by) \
      { mutex_lock(statAddrLock);  (x) += (by);  mutex_unlock(statAddrLock); }
#  endif /* FAST_STATS */

#else /* not STATS */
#  define INCR_STATS(x,by)   /* nothing to do  */
#endif  /* STATS */


/*
 ************************************************************************
 *		    m i d _ a d d r m a p _ i n i t			*
 *----------------------------------------------------------------------*
 * Perform one-time initialization of addrmap code.                     *
 ************************************************************************
 */
void
mid_addrmap_init ()
{
#ifdef STATS
  statAddrLock = mutex_alloc ();
  MIDWAY_MUTEX_SET_NAME(statAddrLock,"lockaddr.c:statAddrLock");
  bzero (&midLockAddrStats, sizeof(midLockAddrStats));
#endif 

  return;
}

#if !defined(PAGE_BASED)
/*
 ************************************************************************
 *		m i d _ a d d r m a p _ s e t _ d b i t			*
 *----------------------------------------------------------------------*
 * Set up dbit information in the addrmap.  We would rather pre-compute *
 * this information than do it on every pack.  In addition, decide if   *
 * the "fast path" can be taken -- there can be only one addrmap and    *
 * few enough dbit entries such that even if all memory backed by this  *
 * addrmap is dirty, the information will fit in one pbuf.		*
 ************************************************************************
 */
mid_addrmap_set_dbit( s_addr_map_t *amap, amap_num_t num)
{
  region_info	riP;
  int           line_size, line_offset, offset, dbit_len;
  clock_t      *ri_line_info;
  int           ri_line_size, ri_line_size_in_bytes;
  vm_offset_t   ri_base;
  clock_t	*dbitP;
  char255	smsg;
  int		amap_len = amap->len;

  /* locate region descriptor */
  riP = (region_info) REGION_FROM_ADDR(amap->addr);
  if ((riP == NULL) || (riP->line_info == NULL)) {
    /* XXX - need to expand this addrmap now, else we can't pre-compute! */
    /* if this has no amap info, then it hasn't been allocated.
       perhaps we should just not do it here, and take the slow
       path later.  ...but the slow path no longer computes? */
    sprintf(smsg,"mid_addrmap_set_dbit: WARNING: dbits not set up for 0x%x+%d, entry will be ignored.",
	    amap->addr, amap_len);
    StatusMsg(smsg);
    StatusMsg("...perhaps this memory has not been shmalloc'd?");
    amap->dbit = 0;
    amap->dbit_end = 0;
    amap->line_size = 0;
    amap->line_offset = 0;
    amap->max_size = 0;
    return;
  }
  ri_base = riP->base;
  ri_line_info = riP->line_info;
  amap->line_size = ri_line_size = riP->line_size;
  ri_line_size_in_bytes = RI_LINE_SIZE_TO_BYTES(ri_line_size);

  /* locate dbit for 1st byte of amap_addr */
  offset = (long)(amap->addr) - ri_base;
  line_offset = offset >> ri_line_size;
  amap->dbit = dbitP = &ri_line_info[ line_offset ];
  amap->line_offset = line_offset = offset - (line_offset << ri_line_size);
  ASSERT ((line_offset >= 0) && (line_offset < ri_line_size_in_bytes));
  
  if (IS_DBG_LOCK_MLPD) {
    sprintf(smsg, "dbit: addr=0x%x+%ld dbitP 0x%x (%d) line_off=%d size=%d",
	    amap->addr, amap_len, dbitP, *dbitP, line_offset, ri_line_size_in_bytes);
    StatusMsg(smsg);
  }

  if (LINE_UNALIGNED(*dbitP)) {
    amap->max_size = 0;		/* never go fast path */
    return;
  }

  /* regular dirty bits */
  dbit_len = 0;
  if (amap_len > (ri_line_size_in_bytes - line_offset)) { /* len > data in 1st line */
    
    /* compute number of dirty bits covered by this amap.
     * subtract data in the first cache line, then round up the remaining length
     * to a cache line multiple. ie
     * (((data_len - (data_in_first_line)) + (line_size_in_bytes-1)) >> ri_line_size) + 1
     * which simplifies since data_in_first_line = line_size_in_bytes - line_offset
     */
    dbit_len = (amap_len + line_offset - 1) >> ri_line_size;
    amap->dbit_end = dbitP + 1 + dbit_len;
  } else
    amap->dbit_end = dbitP + 1;

  if ((num == AMAP_SINGLE) && (dbit_len < mid_addrmap_fast_cutoff)) {
    /* see if the data might fit in a packet */
    /* Worst case: each dbit is transferrable, and has a different clock
       value.  This will cause data guarded by each dbit to have an
       addrseg and an addrpair prepended. */
    int bytes = (dbit_len + 1) * (ADDRSEG_ADDRPAIR_SIZE);

    /* add in cost of data.  Could do by adding ri_line_size_in_bytes to
       above, but might be off by line_size_in_bytes-1, which might
       be significant. */
    bytes += amap->len;

    /* add in size of lock request, and see if it will fit. */
    if ((bytes + sizeof(m_lock_req_t)) < mid_max_data_size) {
      amap->max_size = bytes;
    } else
      amap->max_size = 0;	/* sorry, can't guarantee. take slow path */
  } else
    amap->max_size = 0;		/* sorry, can't guarantee. take slow path */

  ASSERT (!(LINE_UNALIGNED(*dbitP)));
} /* mid_addrmap_set_dbit */
#endif

/*
 ************************************************************************
 *		    m i d _ a d d r m a p _ a d d			*
 *----------------------------------------------------------------------*
 * Add an address range to the address map, coalescing as necessary.	*
 * ...probably just called from midway_bind_syncronization_object/lock.c*
 ************************************************************************
 */
void
mid_addrmap_add(listPP, tailPP, prevPP, daddr, len)
     s_addr_map_t **listPP;     /* ptr to head of list ptr */
     s_addr_map_t **tailPP;     /* ptr to tail of list ptr */
     s_addr_map_t **prevPP;     /* ptr to previous addrmap added to list */
     char *daddr;		/* start of address */
     u_long len;		/* length */
{
  register s_addr_map_t *c;		/* pointer to current block */
  register s_addr_map_t *n;		/* pointer to next block */
  register s_addr_map_t *t;		/* pointer to tail of addrmap list */
  register s_addr_map_t *p;		/* used to create new block for insertion */
  register char *s = (char *) daddr;	/* start address (pretend char for add/sub) */
  register char *e = s+len;		/* end of new block */

/* done under mutex of lock entry, so all of this is safe... */
/*
 * Block could be:
 *	wholly before current chain
 *	wholly between current values in chain
 *	wholly after current chain
 *	part of one or more blocks in current chain
 *		part of one but not another, abut two exactly
 *		(causing them to coalesce) or subsuming multiple blocks
 *		[tho I wouldn't expect this...]
 *
 *
 * Each block has a start address, end address (which is set to the first
 * byte after the block) and a length (so e = s + l)
 *
 * start:
 *    if e < s-cur insert at front, STOP
 *    else
 *	while (s > e-cur AND next-exists AND e >= s-next)
 *	  {not contained in this block, and perhaps part of next
 *	   (which means we don't need to didle our next pointer)
 *	  advance to next 
 *	{here, either contained in current or fits between or part of next}
 *	if (s > e-cur AND (!next-exists OR e < s-next))
 *	  {after this block and before next block, if any (fits between)}
 *	  insert new block after this block; STOP
 *	{here, either contained in current or part of next}
 *	if (s > e-cur)
 *	  {still not part of this block, so must be part of next... advance,
 *	   using next block to start coalescing process... so same as
 *	   being part of "current block"}
 *	  advance to next;
 *	{here definitely part of current block; but may start before...}
 *	if (s < s-cur)
 *	  {must start before now current block and (at least) run into it}
 *	  s-cur := s;
 *	{ok, now look at end}
 *	if (e <= e-cur)					--don't need
 *	  {block now subsumed by current block}		--subsumed by
 *	  STOP;						--while & following
 *	while (e > e-cur)  {extends past current block}
 *	  if (!next-exists OR e < s-next)
 *	    { not part of next block}
 *	    e-cur := e;
 *        else {next-exists AND e >= s-next ---subsumes part/all of next}
 *	    e-cur := e-next;
 *	    toss next (replace it with next's next)
 *      {e <= e-cur; block is now subsumed by current block}
 *	STOP;
 * ...here "STOP" will always fix the length in the current block too...
 */

  /* collect stats on number of calls */
  INCR_STATS (midLockAddrStats.mapadds, 1);

  /*
   * Check if this element immediately follows or
   * immediately proceeds the previous one added.
   */
  p = *prevPP;
  if (p != NULL) {
      if (s == p->end) {
	  p->end = e;
	  p->len += len;
	  INCR_STATS (midLockAddrStats.mapadd_afterprev, 1);
#if !(defined(PAGE_BASED))
	  mid_addrmap_set_dbit(p, (p->max_size?AMAP_SINGLE:AMAP_MULTIPLE));
#endif
	  return;
      }
      if (e == p->addr) {
	  p->addr = s;
	  p->len += len;
	  INCR_STATS (midLockAddrStats.mapadd_beforeprev, 1);
#if !(defined(PAGE_BASED))
	  mid_addrmap_set_dbit(p, (p->max_size?AMAP_SINGLE:AMAP_MULTIPLE));
#endif
	  return;
      }
  }


  /* 
   * Check for insert at tail of list.
   * Or for empty list.
   */
  t = *tailPP;

#ifdef SORT_ADDRMAPS

  if ((t == (s_addr_map_t *) NULL) ||
      (s > t->end)) {
#endif 

    p = (s_addr_map_t *) MALLOC(sizeof(s_addr_map_t));
    ASSERT(p != 0);
    p->addr = s;
    p->end = e;
    p->len = len;
    p->next = NULL;		/* insert after all other entries */
    if (t == (s_addr_map_t *) NULL)
      *listPP = p;
    else {
      t->next = p;
#if !(defined(PAGE_BASED))
      t->max_size = 0;		/* slow path, > 1 addrmap. */
#endif
    }
#if !defined(PAGE_BASED)
    mid_addrmap_set_dbit(p, (t?AMAP_MULTIPLE:AMAP_SINGLE));
#endif
    *tailPP = p;
    *prevPP = p;
#ifdef STATS
    if (t == NULL) {
	INCR_STATS (midLockAddrStats.mapadd_empty, 1);
    } else {
	INCR_STATS (midLockAddrStats.mapadd_tail, 1);
    }
#endif
    return;
#ifdef SORT_ADDRMAPS
  }

  c = *listPP;       /* get ptr to first element in list */
  ASSERT (c != NULL);

  if (e < c->addr) {  	   /* INSERT BEFORE current (first) entry */
    p = (s_addr_map_t *) MALLOC(sizeof(s_addr_map_t));
    ASSERT(p != 0);
    p->addr = s;
    p->end = e;
    p->len = len;
    p->next = c;		/* insert before any other entry */
    *listPP = p;
    *prevPP = p;
    INCR_STATS (midLockAddrStats.mapadd_head, 1);
#if !(defined(PAGE_BASED))
    c->max_size = 0;		/* no fast path, > 1 addrmap */
    mid_addrmap_set_dbit(p, AMAP_MULTIPLE);
#endif
    return;
  }

  /* 
   * We need to scan the list to locate
   * where to insert this element.
   * Start scan at beginning or at the previous element.
   */
  ASSERT (p != NULL);		/* if p == NULL, then t MUST == NULL above */
  if (s > p->end)
      c = p;

#if STATS 
  { register u_long next_count = 0;
#endif

    /* if here, know we don't insert at front at least... find right spot */
    while (s > c->end) {
      n = c->next;
      if (n == NULL) break;
      if (e < n->addr) break;
      ASSERT (c->end <= n->addr); /* Can be == due to prevP heuristic above */
      c = n;
#if STATS
      next_count++;
#endif      
    }

#if STATS 
    INCR_STATS (midLockAddrStats.mapadd_next, next_count);
  }
#endif

  /* contained in current OR fits between OR part of next */
  if ((s > c->end) && ((!c->next) || (e < c->next->addr))) {
    /* after this block and before next block, if any */
    p = (s_addr_map_t *) MALLOC(sizeof(s_addr_map_t));
    ASSERT(p != 0);
    p->addr = s;
    p->end = e;
    p->len = len;
    p->next = c->next;		/* insert after.. */
    c->next = p;		/* ..this block */
    *prevPP = p;
    INCR_STATS (midLockAddrStats.mapadd_after, 1);
#if !defined(PAGE_BASED)
    mid_addrmap_set_dbit(p, AMAP_MULTIPLE);
    /* ensure c will not take fast path */
    c->max_size = 0;
#endif
    return;
  }

  /* if here, either contained in current or part of next */
  if (s > c->end) {		/* must be part of next */
    c = c->next;		/* advance to next */
  }

  /* now must be contained in current */

  /* fix starting addr as necessary */
  if (s < c->addr) {		/* starts before this block */
    c->addr = s;		/* so fix starting address */
    INCR_STATS (midLockAddrStats.mapadd_current_start, 1);
  }

  /* fix ending address as necessary */
  while (e > c->end) {		/* extends past current block */
    if (!(c->next) || (e < c->next->addr)) {
      /* not part of next block */
      c->end = e;		/* fixup end */
      INCR_STATS (midLockAddrStats.mapadd_current_end, 1);
    }
    else {		       /* subsumes all/part of next; squash together */
      /* (sticking the new stuff in the current entry makes it "run into"
	 the next block; hence coalesce) */
      p = c->next;
      if (t == p) {
	t = c;
	*tailPP = c;
      }
      c->end = p->end;		/* swallow next */
      c->next = p->next;	/* delete it from chain */
      FREE(p);			/* and free it too */
      INCR_STATS (midLockAddrStats.mapadd_merged, 1);
    }
  }

  /* now know new totally subsumed by current block */
  c->len = c->end - c->addr;	/* fix in case diddled start/end pointers */
  *prevPP = c;
  INCR_STATS (midLockAddrStats.mapadd_current, 1);
#if !defined(PAGE_BASED)
  mid_addrmap_set_dbit(c, ((*tailPP == c) && (*listPP == c))?AMAP_SINGLE:
		       AMAP_MULTIPLE);
#endif
  return;			/* done! */

#endif /* SORT_ADDRMAPS */

} /* mid_addrmap_add */


/*
 ************************************************************************
 *		    m i d _ a d d r m a p _ p u r g e			*
 *----------------------------------------------------------------------*
 * Free all addrmaps in the given list.                             	*
 ************************************************************************
 */
void
mid_addrmap_purge(listPP, tailPP, prevPP)
     s_addr_map_t **listPP;     /* ptr to head of list ptr */
     s_addr_map_t **tailPP;     /* ptr to tail of list ptr */
     s_addr_map_t **prevPP;     /* ptr to previous addrmap added to list */
{
  register s_addr_map_t *c;		/* pointer to current block */
  register s_addr_map_t *n;		/* pointer to next block */

/* done under mutex of lock entry, so all of this is safe... */

  /* collect stats on number of calls */
  INCR_STATS (midLockAddrStats.map_purges, 1);

  /* 
   * Loop freeing all elements in the list
   */
  c = *listPP;
  while (c != NULL) {
      n = c->next;
      FREE (c);
      c = n;
  }
  
  /* reset list pointers before returning */
  *listPP = NULL;
  *tailPP = NULL;
  *prevPP = NULL;

  return;			/* done! */

} /* mid_addrmap_purge */


/*
 ************************************************************************
 *		    m i d _ a d d r m a p _ p r t			*
 *----------------------------------------------------------------------*
 * Send current addrmap entry to the status server...			*
 ************************************************************************
 */
mid_addrmap_prt(objId, p)
     u_long         objId;      /* Id of object being printed */
     s_addr_map_t  *p;          /* ptr to head of addrmap list */
{
  char255 smsg;
  char temp[100];
  int left;			/* amount of room left in smsg */

  sprintf(smsg,"Object %u amap: ", objId);

  if (p == (s_addr_map_t *) 0) {
    strcat(smsg,"[null addrmap]");
  }
  else {
    sprintf(temp,"%lx+%lu=%lx",p->addr,p->len,p->end);
    strcat(smsg,temp);		/* know will fit */
    left = sizeof(smsg) - strlen(smsg);
    p = p->next;
    while (p != (s_addr_map_t *) 0) {
      sprintf(temp,", %lx+%lu=%lx",p->addr,p->len,p->end);
      if (strlen(temp) > left) {
	StatusMsg(smsg);	/* clear out what we've got */
	sprintf(smsg,"+Lock %u amap: %s",objId,(temp+2));
	/* temp+2 overcomes comma and space */
      } else {
	strcat(smsg,temp);	/* add the new info */
      }
      left = sizeof(smsg) - strlen(smsg);
      p = p->next;
    }
  }
  StatusMsg(smsg);
  return;
} /* mid_addrmap_prt */


/*
 ************************************************************************
 *			m i d _ p a c k _ d a t a			*
 *----------------------------------------------------------------------*
 * Given an addrmap and a time, and a partially filled pbuf, fill pbuf  *
 * with DATASEGs containing data changed since the given time.		*
 *									*
 * Since we might run out of packet before running out of data, we must *
 * be passed in arguments to be able to tell where in the addrmap we    *
 * are.  We will do this under lock; we want only one of the		*
 * potentially many callers to update the dirty bits at a time.		*
 * If this is bad, we can probably keep some kind of pointer into the   *
 * addrmap whereby you don't need the lock if you are looking before    *
 * that point; or only lock if we find a value needs to be changed.	*
 ************************************************************************
 */
#if STATS
#if STATS_DBIT
unsigned int midway_dbit_pack_time = 0;
#endif
#endif

/* return 1 if ran out of buffer before addrmap, 0 if done */
int
mid_pack_data (
     s_addr_map_t **amapPP,	/* the map in/out -- in case more needed */
     char         **amap_addrPP,/* in/out -- ptr to uncopied portion of map */
     long          *amap_lenP,  /* in/out -- uncopied length of map */
     clock_t	old_clock,	/* other side is accurate as of this time */
     clock_t	*max_clockP,	/* maximum clock time found so far */
     char	*pbufP,		/* pointer into packetbuffer */
     long	*pbufLenP,	/* length of stuff already in buffer */
     char	**data_addrPP,	/* data address working on */
     long	*data_lenP,	/* length that needs to be copied */
     clock_t	*data_clockP,	/* time that needs to be copied */
     long	*xfer_lenP	/* for stats, actual amount xferred */
)
{
  int           rc;
  s_addr_map_t *amapP;
  char         *amap_addr;
  long          amap_len;
  addrseg_t    *asegP;
  addrpair_t   *apairP;
  long          pblen, pbleft;
  long          copy_len;
  int           aseg_num;
  char         *data_addr;
  long          data_len;
  register clock_t       data_clock;
  clock_t       max_clock;
  clock_t       aseg_clock;
  clock_t       current_clock;
  int           dbit_len;
  int           line_size, line_offset, offset;
  clock_t      *ri_line_info;
  int           ri_line_size, ri_line_size_in_bytes;
  vm_offset_t   ri_base;
  int           ri_number, amap_region_number;
  region_info   riP;
  register clock_t       dbit_val;
  register clock_t      *dbitP, *dbit_startP, *dbit_endP;
  clock_t      *first_dbitP;
  long          xfer_len;
  char255       smsg;
  extern clock_t LocalClock;
#if STATS
#if STATS_DBIT
  unsigned int	startfrc = 0, endfrc = 0;
  extern unsigned int *midway_kern_frc;
#endif /*STATS_DBIT*/
#endif /*STATS*/


  /* cache args */
  amapP = *amapPP;
  amap_addr = *amap_addrPP;
  amap_len = *amap_lenP;
  max_clock = *max_clockP;
  pblen = *pbufLenP;
  data_addr = *data_addrPP;
  data_len = *data_lenP;
  data_clock = *data_clockP;
  
  if (IS_DBG_LOCK_DLTA) {
    sprintf(smsg,"mid_pack_data(0x%lx,0x%x,%lu,%lu,0x%x,%lu,0x%x,%lu,%lu)", 
	    amapP, amap_addr, amap_len, old_clock, 
	    pbufP, pblen, data_addr, data_len, data_clock);
    StatusMsg(smsg);
  }
  INCR_STATS(midLockAddrStats.pack_calls, 1);

#ifdef STATS 
  xfer_len = 0;
#endif 


  rc = 0;			/* 0 = no more work to do */
  asegP = NULL;
  apairP = (addrpair_t *) pbufP;
    
  if (amapP == NULL)
    goto done;

  pbleft = (mid_max_data_size - pblen) & ~3;
  aseg_num = 0;
  aseg_clock = MAX_CLOCK_VALUE;
  ri_number = -1;
#if !defined(PAGE_BASED)
  /* x-amap setup */
  if (amap_len != 0) {
    ASSERT(amap_addr >= amapP->addr);
    ASSERT(amap_len <= amapP->len);
    line_size = ri_line_size = amapP->line_size;
    line_offset = amapP->line_offset;

    offset = amap_addr - amapP->addr; /* bytes, both are char* */
    offset += data_len;		/* gotta skip over that which you will copy */
    offset = offset >> line_size; /* # dirty words */
    dbitP = amapP->dbit + offset; /* clock_t *, so should inc correctly */
    dbit_endP = amapP->dbit_end;
    ri_line_size_in_bytes = RI_LINE_SIZE_TO_BYTES(ri_line_size);
    first_dbitP = dbitP;

#if (BENCHMARK == 0) && (VERIFY_XAMAP_DBITS != 0)
    riP = (region_info) REGION_FROM_ADDR(amap_addr);
    ASSERT(line_size == riP->line_size);
    offset = (long) amap_addr - riP->base;
    offset += data_len;
    offset = offset >> riP->line_size;
    ASSERT(dbitP == &(riP->line_info[ offset ]));
#endif /* ~benchmark & verify the dbit maps */
  } else {
    dbitP = NULL; /* will get reset */
  }
#else
  dbitP = NULL;
#endif
  current_clock = LocalClock;

  while (pbleft >= (long)(ADDRSEG_SIZE+ADDRPAIR_SIZE+sizeof(long))) {

    /* Check if there is data to be copied */
    if (data_len > 0) {

      /* initialize address segment, if necessary */
      if (aseg_clock != data_clock) {
	if (aseg_num > 0)
	  asegP->num = aseg_num;
	asegP = (addrseg_t *) apairP;
	asegP->segtype = DATASEG;
	asegP->clock = data_clock;
	aseg_clock = data_clock;
	if (GREATER_THAN_CLOCK(data_clock, max_clock))
	  max_clock = data_clock;
	aseg_num = 0;
	apairP = &(asegP->ap);
	pbleft -= (ADDRSEG_SIZE + ADDRPAIR_SIZE);
      } else 
	pbleft -= ADDRPAIR_SIZE;


      /* initialize address pair */
      copy_len = (data_len < pbleft) ? data_len : pbleft;
      apairP->addr = data_addr;
      apairP->len = copy_len;

      if (IS_DBG_LOCK_MLPD) {
	sprintf(smsg, "mid_pack_data: copying %lx+%ld  to 0x%lx at clock=%lu", 
		data_addr, copy_len, apairP->data, data_clock);
	StatusMsg(smsg);
      }
      ASSERT (copy_len > 0);
      ASSERT (pbleft >= copy_len);
      ASSERT ((copy_len <= data_len) && (copy_len <= amap_len));
      ASSERT ((data_addr >= amapP->addr) && 
	      (data_addr < amapP->end));
      ASSERT (((data_addr+copy_len) > amapP->addr) && 
	      ((data_addr+copy_len) <= amapP->end));

      /*
       * Copy into output buffer
       */

      bcopy (data_addr,		/* source */
	     apairP->data,	/* target */
	     copy_len);		/* length */

#ifdef STATS 
      xfer_len += copy_len;
#endif 

      data_addr += copy_len;
      data_len -= copy_len;
      amap_addr += copy_len;
      amap_len -= copy_len;
      aseg_num += 1;

      copy_len = (copy_len + 3) & ~3;
      pbleft -= copy_len;
      apairP = (addrpair_t *) &apairP->data[copy_len];
    }

    ASSERT (pbleft >= 0);
    ASSERT (data_len >= 0);
    ASSERT (amap_len >= 0);
    ASSERT (amap_len >= data_len);

#if STATS
#if STATS_DBIT
    startfrc = *midway_kern_frc;
#endif
#endif

    /* 
     * locate next cache line to copy
     */
    while (data_len == 0) {

      /* advance to next address map, if necessary */
      if (amap_len == 0) {
	if (amap_addr != 0) {	/* check if this is the 1st call */
	  amapP = amapP->next;
	  if (amapP == NULL) goto done; /* no more work to do */
	}
	amap_addr = amapP->addr;
	amap_len = amapP->len;
#if !defined(PAGE_BASED)
	/*
	 *  NOTE: I believe this code will never be called (that is,
	 *  packing using dirty bits) if page based.  If this is true,
	 *  all these routines could be compiled out.  However, JustInCase,
	 *  I am leaving things here, but have to remove x-amaps because
	 *  I only use the extra amap space if we are not page-based.
	 *  --aug 94, mjz
	 */
	/* extended AMAP info. */
	dbitP = amapP->dbit;
	/* this is an insane case, but could have happend if there
	   was a binding of an address with no dbits at bind time. */
	if (dbitP == NULL) {
	  amap_len = 0;		/* force new amap */
	  continue;
	}
	dbit_endP = amapP->dbit_end;
	line_size = ri_line_size = amapP->line_size;
	ri_line_size_in_bytes = RI_LINE_SIZE_TO_BYTES(ri_line_size);
	first_dbitP = dbitP;
#if (BENCHMARK == 0) && (VERIFY_XAMAP_DBITS != 0)
	riP = (region_info) REGION_FROM_ADDR(amap_addr);
	ASSERT(line_size == riP->line_size);
	offset = (long) amap_addr - riP->base;
	line_offset = offset >> riP->line_size;
	ASSERT(dbitP == &(riP->line_info[ line_offset ]));
#endif /* ~benchmark & verify the dbit maps */
	line_offset = amapP->line_offset;
#endif /*not PAGE_BASED*/
	ASSERT (amap_len > 0);
      }

      /* locate dirty bits for this address, if necessary 
       * Note that the requested address might not start or end 
       * on a cache line boundary.
       */
      /* do we still need this with x-amaps? 24aug94 mjz */
#if defined (PAGE_BASED) /* since no dense dbits + x-amaps, 
			     will never be true. */
      if (dbitP == NULL) {

	/* locate region descriptor & cache values from it */
	amap_region_number = ((unsigned long)amap_addr) / REGION_SIZE;
	if (amap_region_number != ri_number) {
	  riP = (region_info) REGION_FROM_ADDR(amap_addr);
	  if ((riP == NULL) || (riP->line_info == NULL)) {
	    amap_len = 0;	/* not allocated - try next addr map */
	    continue;
	  }
	  ri_number = amap_region_number;
	  ri_base = riP->base;
	  ri_line_info = riP->line_info;
	  ri_line_size = riP->line_size;
	  ri_line_size_in_bytes = RI_LINE_SIZE_TO_BYTES(ri_line_size);
	}

	/* locate dbit for 1st byte of amap_addr */
	offset = (long)amap_addr - ri_base;
	line_offset = offset >> ri_line_size;
	dbitP = &ri_line_info[ line_offset ];
	line_offset = offset - (line_offset << ri_line_size);
	ASSERT ((line_offset >= 0) && (line_offset < ri_line_size_in_bytes));
	  
	if (IS_DBG_LOCK_MLPD) {
	  sprintf(smsg, "dbit: addr=0x%x+%ld dbitP 0x%x (%d) line_off=%d size=%d",
		  amap_addr, amap_len, dbitP, *dbitP, line_offset, ri_line_size_in_bytes);
	  StatusMsg(smsg);
	}
	INCR_STATS(midLockAddrStats.pack_dbit_lookups, 1);

#if (BENCHMARK == 0)  /* removal of dense dbits */
	/* ~BENCHMARK NOTE: dbit_val is reset in the unrolled loop below, if it
	   is needed! */
	/* Check for dense dirty bits */
	dbit_val = *dbitP;
	if (LINE_UNALIGNED(dbit_val)) {
	  
	  /* Compute offset into dense dirty bit array */
	  offset = (dbit_val & LINE_DATA_VALUE) + line_offset;
	  dbitP = &vm_dense_dirty_base[ offset ];
      
	  /* determine number of contiguous dense dbits remaining in this cache line */
	  dbit_len =  ri_line_size_in_bytes - line_offset;
	  if (dbit_len > amap_len) 
	    dbit_len = amap_len;
	  dbit_endP = (dbitP + dbit_len);
	  line_size = 0;	/* line is 1 byte, no shift required */
	  line_offset = 0;

	  if (IS_DBG_LOCK_MLPD) {
	    sprintf(smsg, "dense dbit: addr=0x%x+%ld dense_dbitP 0x%x (%d), endP=0x%x",
		    amap_addr, data_len, dbitP, *dbitP, dbit_endP);
	    StatusMsg(smsg);
	  }
	  INCR_STATS(midLockAddrStats.pack_ddbit_lookups, 1);
	  INCR_STATS(midLockAddrStats.pack_ddbit_reads, dbit_len); /*we may lookup same addr twice*/

	} else  {
	  
#endif /*~BENCHMARK*/
	  /* regular dirty bits */
	  line_size = ri_line_size;
	  if (amap_len > (ri_line_size_in_bytes - line_offset)) { /* len > data in 1st line */

	    /* compute number of dirty bits covered by this amap.
	     * subtract data in the first cache line, then round up the remaining length
	     * to a cache line multiple. ie
	     * (((data_len - (data_in_first_line)) + (line_size_in_bytes-1)) >> line_size) + 1
	     * which simplifies since data_in_first_line = line_size_in_bytes - line_offset
	     */
	    dbit_len = (amap_len + line_offset - 1) >> line_size;
	    dbit_endP = dbitP + 1 + dbit_len;
	  } else
	    dbit_endP = dbitP + 1;
#if (BENCHMARK == 0)		/* removal of dense Dbits */
	} 
#endif
	first_dbitP = dbitP;	/* save ptr to 1st dbit to adjust line lengths */

	ASSERT (amap_len > 0);
	ASSERT (amap_len <= amapP->len);
	ASSERT (!(LINE_UNALIGNED(*dbitP)));
	ASSERT (dbit_endP > dbitP);
     } /* dbitP == NULL */
#endif /* PAGE_BASED */

      ASSERT(dbitP > 0);

      /* 
       * Scan dirty bits to find start of data to transfer.
       */
      { register u_long line_data_value = LINE_DATA_VALUE;
	register u_long line_unaligned_value = LINE_EXCEPTION_FLAG|LINE_UNALIGNED_FLAG;
	register clock_t o_clock = old_clock;
	register clock_t dbit_val0, dbit_val1, dbit_val2, dbit_val3;

	/*
	 * Unroll loop to optimize it.
	 * Be careful to preserve final dbit_val
	 * Note this may read past end of dbits. If so, final dbit_val not used.
	 * Also: the distance past end of dbits MUST be defined as IN_DIRTY_EXTRA_SPACE.
	 * WAS 17Feb94
	 */
	for (dbit_startP = dbitP; dbitP < dbit_endP; dbitP += 4) {
	  dbit_val0 = dbitP[0];
	  dbit_val1 = dbitP[1];
	  dbit_val2 = dbitP[2];
	  dbit_val3 = dbitP[3];
	  /*
	   * Loop until dbit_val == 0 or dbit_val > o_clock
	   * Optimize by checking for (dbit_val-1) >= o_clock
	   * WAS 17Feb94
	   */
	  if ((GREATER_THAN_OR_EQUAL_CLOCK (dbit_val0-1, o_clock)) ||
	      (GREATER_THAN_OR_EQUAL_CLOCK (dbit_val1-1, o_clock)) ||
	      (GREATER_THAN_OR_EQUAL_CLOCK (dbit_val2-1, o_clock)) ||
	      (GREATER_THAN_OR_EQUAL_CLOCK (dbit_val3-1, o_clock))) {
	    
	    if (dbit_val0 == NULL_CLOCK_VALUE)
	      goto exit1_plus0;
	    if (GREATER_THAN_CLOCK (dbit_val0, o_clock)) {
#if ADDR_ENABLE_NON_EC
	      if ((dbit_val0 & line_unaligned_value) == line_unaligned_value)
		goto exit1_plus0;;
	      dbit_val0 &= line_data_value;
	      if (GREATER_THAN_CLOCK (dbit_val0, o_clock))
#endif
		goto exit1_plus0;
	    }
	    if (dbit_val1 == NULL_CLOCK_VALUE)
	      goto exit1_plus1;
	    if (GREATER_THAN_CLOCK (dbit_val1, o_clock)) {
#if ADDR_ENABLE_NON_EC
	      if ((dbit_val1 & line_unaligned_value) == line_unaligned_value)
		goto exit1_plus1;;
	      dbit_val1 &= line_data_value;
	      if (GREATER_THAN_CLOCK (dbit_val1, o_clock))
#endif
		goto exit1_plus1;
	    }
	    if (dbit_val2 == NULL_CLOCK_VALUE)
	      goto exit1_plus2;
	    if (GREATER_THAN_CLOCK (dbit_val2, o_clock)) {
#if ADDR_ENABLE_NON_EC
	      if ((dbit_val2 & line_unaligned_value) == line_unaligned_value)
		goto exit1_plus2;;
	      dbit_val2 &= line_data_value;
	      if (GREATER_THAN_CLOCK (dbit_val2, o_clock))
#endif
		goto exit1_plus2;
	    }
	    if (dbit_val3 == NULL_CLOCK_VALUE)
	      goto exit1_plus3;
	    if (GREATER_THAN_CLOCK (dbit_val3, o_clock)) {
#if ADDR_ENABLE_NON_EC
	      if ((dbit_val3 & line_unaligned_value) == line_unaligned_value)
		goto exit1_plus3;;
	      dbit_val3 &= line_data_value;
	      if (GREATER_THAN_CLOCK (dbit_val3, o_clock))
#endif
		goto exit1_plus3;
	    }
	  }
	} /* for dbit... */
	goto exit1_done;	/* reached end of dbits -- dbit_val not used */

exit1_plus3:
	dbit_val = dbit_val3;
	dbitP += 3;
	goto exit1_done;
exit1_plus2:
	dbit_val = dbit_val2;
	dbitP += 2;
	goto exit1_done;
exit1_plus1:
	dbit_val = dbit_val1;
	dbitP++;
	goto exit1_done;
exit1_plus0:
	dbit_val = dbit_val0;
exit1_done:
	if (dbitP > dbit_endP) dbitP = dbit_endP;
      } /* register scope */

      ASSERT (dbitP >= dbit_startP);
      ASSERT (dbitP <= dbit_endP);
      ASSERT (dbitP >= first_dbitP);
      ASSERT (dbit_startP >= first_dbitP);
      ASSERT (dbit_endP  >= dbit_startP);
	
      /* compute number of bytes skipped while traversing dirty bits */
      if (dbitP > dbit_startP) {
	
	INCR_STATS(midLockAddrStats.pack_dbit_reads, (dbitP - dbit_startP));

	dbit_len = (dbitP - dbit_startP) << line_size;
	if (dbit_startP == first_dbitP)
	  dbit_len -= line_offset;
	if (dbit_len > amap_len)
	  dbit_len = amap_len;
	ASSERT (dbit_len > 0);

	amap_addr += dbit_len;
	amap_len -= dbit_len;
	ASSERT ((amap_len >= 0) && (amap_len < amapP->len));
	ASSERT ((amap_addr > amapP->addr) && 
		(amap_addr <= amapP->end));
	ASSERT (((amap_addr+amap_len) > amapP->addr) && 
		((amap_addr+amap_len) <= amapP->end));
      }
      
      if ((dbitP == dbit_endP) || (amap_len == 0)) {
	  
	ASSERT((dbitP == dbit_endP) && (amap_len == 0)); /* x-amap */
	/* nothing found  - loop again */
	dbitP = NULL;
	amap_len = 0;		/* added for x-amap BREAKS DENSE?*/
	continue;
      }

      /* Found start of data to transfer 
       * Now find end of data to transfer in this addr seg/pair
       * Also set clock value for all dirty lines.
       *
       * Note: Its also possible that we've run into a dense dirty bit
       * in which case, the loop below won't find anything to transfer.
       * Rather than checking for it above (and slowing down the common case), 
       * simply allow the loop below to exit and handle the dense dirty bit on the next iteration. 
       * WAS 10Sep93
       */
      { 
#if ADDR_ENABLE_NON_EC
	register clock_t cur_clock = current_clock & LINE_DATA_VALUE;
	register u_long line_data_value = LINE_DATA_VALUE;
	register u_long line_unaligned_value = LINE_EXCEPTION_FLAG|LINE_UNALIGNED_FLAG;
#else
	register clock_t cur_clock = current_clock;
#endif
	register clock_t dbit_val0, dbit_val1;
	register clock_t d_clock;
#ifdef STATS
	register u_long d_count = 0;
#endif

#if 0
	if (dbit_val == NULL_CLOCK_VALUE)
	  d_clock = cur_clock;
	else
#if ADDR_ENABLE_NON_EC
	  d_clock = dbit_val & line_data_value;
#else
	  d_clock = dbit_val;
#endif
	ASSERT (d_clock != NULL_CLOCK_VALUE);
	
	/*
	 * Optimize loop by splitting into two cases and unrolling each.
	 *  case 1: loop while (clock = current value or NULL)
	 *  case 2: loop while (clock = some prior value)
	 * Note that case 2 may read past end of loop, 
	 *           case 1 may NOT since it resets NULL dbit values.
	 * WAS 17Feb94
	 */
	if (d_clock == cur_clock) {
	  register clock_t *dbit_loopEndP;
	  register int dbit_count;

	  /*
	   * reset null dirtybits 
	   * Don't overrun dbits in this loop!
	   */
	  dbit_count = dbit_endP - dbitP;
	  dbit_count = (dbit_count / 4) * 4;
	  dbit_loopEndP = (dbitP + dbit_count);
	  for (dbit_startP = dbitP; dbitP < dbit_loopEndP; dbitP += 4) {
	    dbit_val0 = dbitP[0];
	    dbit_val1 = dbitP[1];
	    if (dbit_val0 == NULL_CLOCK_VALUE) {
	      dbitP[0] = cur_clock;
#if (STATS_DBIT == 0)
	      INCR_STATS (d_count, 1);
#endif
	    } else if (dbit_val0 != d_clock) {
#if ADDR_ENABLE_NON_EC
	      if ((dbit_val0 & line_data_value) != d_clock) 
		goto exit2_plus0;
	      if ((dbit_val0 & line_unaligned_value) == line_unaligned_value)
#endif
		goto exit2_plus0;
	    }
	    if (dbit_val1 == NULL_CLOCK_VALUE) {
	      dbitP[1] = cur_clock;
#if (STATS_DBIT == 0)
	      INCR_STATS (d_count, 1);
#endif
	    } else if (dbit_val1 != d_clock) {
#if ADDR_ENABLE_NON_EC
	      if ((dbit_val1 & line_data_value) != d_clock) 
		goto exit2_plus1;
	      if ((dbit_val1 & line_unaligned_value) == line_unaligned_value)
#endif
		goto exit2_plus1;
	    }
	    dbit_val0 = dbitP[2];
	    dbit_val1 = dbitP[3];
	    if (dbit_val0 == NULL_CLOCK_VALUE) {
	      dbitP[2] = cur_clock;
#if (STATS_DBIT == 0)
	      INCR_STATS (d_count, 1);
#endif
	    } else if (dbit_val0 != d_clock) {
#if ADDR_ENABLE_NON_EC
	      if ((dbit_val0 & line_data_value) != d_clock) 
		goto exit2_plus2;
	      if ((dbit_val0 & line_unaligned_value) == line_unaligned_value)
#endif
		goto exit2_plus2;
	    }
	    if (dbit_val1 == NULL_CLOCK_VALUE) {
	      dbitP[3] = cur_clock;
#if (STATS_DBIT == 0)
	      INCR_STATS (d_count, 1);
#endif
	    } else if (dbit_val1 != d_clock) {
#if ADDR_ENABLE_NON_EC
	      if ((dbit_val1 & line_data_value) != d_clock) 
		goto exit2_plus3;
	      if ((dbit_val1 & line_unaligned_value) == line_unaligned_value)
#endif
		goto exit2_plus3;
	    }
	  }
	  /* loop over final 3 dbits (if necessary) */
	  for (; dbitP < dbit_endP; dbitP++) {
	    dbit_val0 = dbitP[0];
	    if (dbit_val0 == NULL_CLOCK_VALUE) {
	      dbitP[0] = cur_clock;
#if (STATS_DBIT == 0)
	      INCR_STATS (d_count, 1);
#endif
	    } else if (dbit_val0 != d_clock) {
#if ADDR_ENABLE_NON_EC
	      if ((dbit_val0 & line_data_value) != d_clock) 
		goto exit2_plus0;
	      if ((dbit_val0 & line_unaligned_value) == line_unaligned_value)
#endif
		goto exit2_plus0;
	    }
	  }
	  goto exit2_plus0;	/* reached end of dbits */
	  
exit2_plus3:
	  dbitP++;
exit2_plus2:
	  dbitP++;
exit2_plus1:
	  dbitP++;
exit2_plus0:
	  ASSERT (dbitP <= dbit_endP);
#else /* SPLIT LOOP DIFFERENTLY */
	/*
	 * Optimize loop by splitting into two cases and unrolling each.
	 *  case 1: loop while (clock = NULL)
	 *  case 2: loop while (clock = current or some prior value)
	 * Note that case 2 may read past end of loop, 
	 *           case 1 may NOT since it resets NULL dbit values.
	 * WAS 17Feb94
	 */
	if (dbit_val == NULL_CLOCK_VALUE){
	  register clock_t *dbit_loopEndP;
	  register int dbit_count;
	  register clock_t dbit_val2, dbit_val3;
	  d_clock = cur_clock;
	  ASSERT (d_clock != NULL_CLOCK_VALUE);
	

	  /*
	   * reset null dirtybits 
	   * Don't overrun dbits in this loop!
	   */
	  dbit_count = dbit_endP - dbitP;
	  dbit_count = (dbit_count / 4) * 4;
	  dbit_loopEndP = (dbitP + dbit_count);
	  for (dbit_startP = dbitP; dbitP < dbit_loopEndP; dbitP += 4) {
	    dbit_val0 = dbitP[0];
	    dbit_val1 = dbitP[1];
	    dbit_val2 = dbitP[2];
	    dbit_val3 = dbitP[3];
	    if (dbit_val0 == NULL_CLOCK_VALUE) {
	      dbitP[0] = cur_clock;
	    } else 
	      goto exit2_plus0;
	    if (dbit_val1 == NULL_CLOCK_VALUE) {
	      dbitP[1] = cur_clock;
	    } else 
		goto exit2_plus1;
	    if (dbit_val2 == NULL_CLOCK_VALUE) {
	      dbitP[2] = cur_clock;
	    } else
		goto exit2_plus2;
	    if (dbit_val3 == NULL_CLOCK_VALUE) {
	      dbitP[3] = cur_clock;
	    } else 
		goto exit2_plus3;
	  }
	  /* loop over final 3 dbits (if necessary) */
	  for (; dbitP < dbit_endP; dbitP++) {
	    dbit_val0 = dbitP[0];
	    if (dbit_val0 == NULL_CLOCK_VALUE) {
	      dbitP[0] = cur_clock;
	    } else 
		goto exit2_plus0;
	  }
	  goto exit2_plus0;	/* reached end of dbits */
	  
exit2_plus3:
	  dbitP++;
exit2_plus2:
	  dbitP++;
exit2_plus1:
	  dbitP++;
exit2_plus0:
	  ASSERT (dbitP <= dbit_endP);
#if (STATS_DBIT == 0)
	  /* dbitP points at 1st dbit that wasn't zero */
	  INCR_STATS (d_count, dbitP-dbit_startP);
#endif
#endif

	} else {

#if 1
#if ADDR_ENABLE_NON_EC
	  d_clock = dbit_val & line_data_value;
#else
	  d_clock = dbit_val;
#endif
	  ASSERT (d_clock != NULL_CLOCK_VALUE);
#endif
	  /* 
	   * Scan dirtybits for this clock value only 
	   * Okay to read past end of dbits.
	   */
	  for (dbit_startP = dbitP; dbitP < dbit_endP; dbitP += 4) {
	    dbit_val0 = dbitP[0];
	    dbit_val1 = dbitP[1];
	    if (dbit_val0 != d_clock) {
#if ADDR_ENABLE_NON_EC
	      if ((dbit_val0 & line_data_value) != d_clock) 
		goto exit3_plus0;
	      if ((dbit_val0 & line_unaligned_value) == line_unaligned_value)
#endif
		goto exit3_plus0;
	    }
	    if (dbit_val1 != d_clock) {
#if ADDR_ENABLE_NON_EC
	      if ((dbit_val1 & line_data_value) != d_clock) 
		goto exit3_plus1;
	      if ((dbit_val1 & line_unaligned_value) == line_unaligned_value)
#endif
		goto exit3_plus1;
	    }
	    dbit_val0 = dbitP[2];
	    dbit_val1 = dbitP[3];
	    if (dbit_val0 != d_clock) {
#if ADDR_ENABLE_NON_EC
	      if ((dbit_val0 & line_data_value) != d_clock) 
		goto exit3_plus2;
	      if ((dbit_val0 & line_unaligned_value) == line_unaligned_value)
#endif
		goto exit3_plus2;
	    }
	    if (dbit_val1 != d_clock) {
#if ADDR_ENABLE_NON_EC
	      if ((dbit_val1 & line_data_value) != d_clock) 
		goto exit3_plus3;
	      if ((dbit_val1 & line_unaligned_value) == line_unaligned_value)
#endif
		goto exit3_plus3;
	    }
	  }
exit3_plus3:
	  dbitP++;
exit3_plus2:
	  dbitP++;
exit3_plus1:
	  dbitP++;
exit3_plus0:
	  if (dbitP > dbit_endP) dbitP = dbit_endP;
	}

	ASSERT (dbitP >= dbit_startP);
	ASSERT (dbitP <= dbit_endP);
	ASSERT (dbitP >= first_dbitP);
	ASSERT (dbit_startP >= first_dbitP);
	ASSERT (dbit_endP  >= dbit_startP);
	
	INCR_STATS(midLockAddrStats.pack_dbit_writes, d_count);
	INCR_STATS(midLockAddrStats.pack_dbit_wbytes, (d_count << line_size));
	INCR_STATS(midLockAddrStats.pack_dbit_reads, (dbitP - dbit_startP));

	/* 
	 * dbitP now points at first dbit past data to transfer 
	 * compute size of data to transfer.
	 */
	data_clock = d_clock;
	data_addr = amap_addr;
	data_len = (dbitP - dbit_startP) << line_size;
	if (dbit_startP == first_dbitP)
	  data_len -= line_offset;
	if (data_len > amap_len)
	  data_len = amap_len;
	
	ASSERT (data_len <= amapP->len);
	ASSERT ((data_addr >= amapP->addr) && 
		(data_addr < amapP->end));
	  /* because of new split, we COULD (and have) find that what 
	     we though was dirty is no longer.  That's OK, don't assert. */
	ASSERT (((data_addr+data_len) >= amapP->addr) && 
		((data_addr+data_len) <= amapP->end));


	if ((dbitP == dbit_endP) || (LINE_UNALIGNED(*dbitP))) {
	  ASSERT(!(LINE_UNALIGNED(*dbitP))); /* x-amap */
	  ASSERT(amap_len == data_len);		/* added for x-amap */
	  /* otherwise we would not do the right thing.
	  dbitP = NULL;		/* reset to locate dirtybit next time */
	}
      }
    } /* while data_len == 0 */
#if STATS
#if STATS_DBIT
    endfrc = *midway_kern_frc;
    /* WARNING: this assumes we can only wrap once */
    if (endfrc < startfrc)
      midway_dbit_pack_time += (FRC_MAX - startfrc) + endfrc;
    else
      midway_dbit_pack_time += endfrc - startfrc;
    startfrc = 0;		/* note we added */
#endif /*STATS_DBIT*/
#endif /*STATS*/
  } /* while pbleft >= min size */

  rc = 1;			/* 1= more work to do */

done:

#if STATS
#if STATS_DBIT
  /* since "done" can break us out of loop, count the last bit? */
  endfrc = *midway_kern_frc;
  if (startfrc) {
    if (endfrc < startfrc) {
      midway_dbit_pack_time += (FRC_MAX - startfrc) + endfrc;
    } else {
      midway_dbit_pack_time += endfrc - startfrc;
    }
  }
#endif /*STATS_DBIT*/
#endif /*STATS*/

  /* set segment count in final seg addr */
  if (asegP != NULL)
    asegP->num = aseg_num;

  /* compute size of returned buffer */
  pblen += ((char *)apairP) - pbufP;
  ASSERT (pblen <= mid_max_data_size);
  ASSERT ((pblen & 3) == 0);

  /* return values for subsequent calls */
  *amapPP = amapP;
  *amap_addrPP = amap_addr;
  *amap_lenP = amap_len;
  *max_clockP = max_clock;
  *pbufLenP = pblen;
  *data_addrPP = data_addr;
  *data_lenP = data_len;
  *data_clockP = data_clock;

#ifdef STATS 
  *xfer_lenP = xfer_len;
#endif 

  return(rc);
}


/*
 ************************************************************************
 *		     m i d _ u n p a c k _ d a t a			*
 *----------------------------------------------------------------------*
 * Used to unpack datasegs in a comm buffer -- the opposite of  	*
 * mid_pack_data.  This is used by the lock reply, release reply        *
 * and barrier reply.  Here be dirtybit updates! (dragons?)		*
 ************************************************************************
 */
#if STATS
#if STATS_DBIT
unsigned int midway_dbit_unpack_time = 0;
#endif /*STATS_DBIT*/
#endif /*STATS*/

void
mid_unpack_data (
     char **pptr,
     u_long *lptr,
     s_addr_map_t *amap_listP,	/* DEBUG only: addr maps for this lock/barrier */
     char *cname,		/* caller for debug messages */
     MUD_t ctype 		/* another way to interpret the caller */
)
{
  int		i;
  u_long	length;		/* length of packet pointed to by pptr */
  addrseg_t	*as;		/* data segments, if any */
  addrpair_t    *ap;		/* address/length/data triples, really */
  addrpair_t    *apnext;	/* where next ap will be */
  char		*source, *target; /* used to copy data */
  long          seg_len;	/* size of this data segment */
  long          data_len;	/* amount to copy in this ap */
  register clock_t *dbitP, *dbit_startP, *dbit_endP, *first_dbitP;
#if (BENCHMARK == 0)
  register clock_t  dbit_val;       /* cached value of *dbitP */
#endif
  register clock_t  clock_val;	    /* cached value of as->clock */
  int           dbit_len;
  int           line_size, line_offset, offset;
  clock_t      *ri_line_info;
  int           ri_line_size, ri_line_size_in_bytes;
  vm_offset_t   ri_base;
  int           ri_number, target_region_number;
  region_info   riP;
  kern_return_t krc;
  char255	smsg;
  extern clock_t  *midway_dirtybit_info();
#if STATS
#if STATS_DBIT
  unsigned int	startfrc = 0, endfrc = 0;
  extern unsigned int *midway_kern_frc;
#endif /*STATS_DBIT*/
#endif /*STATS*/

  as = (addrseg_t *) (*pptr);	/* we should be given start of segments */
  length = *lptr;

  if (IS_DBG_LOCK_MLPD) {	/* MLPD is now a misnomer */
    sprintf(smsg, "mid_unpack_data[%s]: incoming message at 0x%x length %u", 
	    cname, as, length);
    StatusMsg(smsg);
  }
  INCR_STATS(midLockAddrStats.unpack_calls, 1);

  ri_number = -1;

  while (length) {
    switch(as->segtype) {

    case DATASEG:
      ap = &(as->ap);
      clock_val = as->clock;
      ASSERT (length >= ((u_long)ap-(u_long)as));
      length -= ((u_long)ap-(u_long)as); /* account for addrseg header */
      for(i = 0; length && (i < as->num); i++) {

	/* addrpair_t has some of data, but not all, so just play w data
	 * +3 & ~3 round to mult of 4
	 */
	apnext = (addrpair_t *) (((u_long)ap->data + ap->len + 3) & ~3);

	/* verify the length before copying the data */
	seg_len = (u_long)apnext - (u_long)ap;
	if (length < seg_len) {
	  StatusMsg("mid_unpack_data: buffer length less than necessary!  What Gives?");
	  goto exit;
	}
	length -= seg_len;
	
	/* 3 Jul 92: only copy data that is newer than exists in our cache.
	 * A create might copy all statics, since it does not know our
	 * current data time for given locks (it doesn't even know what
	 * is covered by a lock).  If we blindly copy, and this processor
	 * owns the lock, then the next time the lock is transferred,
	 * bogus data will go with it.  If this were a lock reply, then
	 * any update of this kind is an error.
	 */
	source = (char *)ap->data;	/* from the message */
	target = ap->addr;	/* to our cache */
	data_len = ap->len;
	
	ASSERT ((data_len + ADDRPAIR_SIZE) <= seg_len);

	while (data_len > 0) {

#if STATS
#if STATS_DBIT
	  startfrc = *midway_kern_frc;
#endif /*STATS_DBIT*/
#endif /*STATS*/
	  /* locate region descriptor & cache values from it */
	  target_region_number = ((long)target) / REGION_SIZE;
	  if (target_region_number != ri_number) {
	    riP = (region_info) REGION_FROM_ADDR(target);
	    ASSERT (riP != NULL); 	/* what else can we do? */

	    /* Handle case where region was allocated remotely.
	     * We need to get the region_info from the region's owner
	     * the first time we reference data within that region.
	     * WAS 5Jan94
	     */
	    if (riP->line_size == 0) {
	      krc = get_shmalloc_info(target);
	      ASSERT (krc == KERN_SUCCESS);
	    }
	    ASSERT (riP->line_info != 0); /* gotta have it */

	    ri_number = target_region_number;
	    ri_base = riP->base;
	    ri_line_info = riP->line_info;
	    ri_line_size = riP->line_size;
	    ri_line_size_in_bytes = RI_LINE_SIZE_TO_BYTES(ri_line_size);
	  }
	    
	  /* locate dbit for 1st byte of target */
	  offset = (long)target - ri_base;
	  line_offset = offset >> ri_line_size;
	  dbitP = &ri_line_info[ line_offset ];
	  line_offset = offset - (line_offset << ri_line_size);
	  ASSERT ((line_offset >= 0) && (line_offset < ri_line_size_in_bytes));
	  
	  if (IS_DBG_LOCK_MLPD) {
	    sprintf(smsg, "unpack dbit: addr=0x%x+%ld dbitP 0x%x (%d) line_off=%d size=%d",
		    target, data_len, dbitP, *dbitP, line_offset, ri_line_size_in_bytes);
	    StatusMsg(smsg);
	  }
	  INCR_STATS(midLockAddrStats.unpack_dbit_lookups, 1);
	  
#if (BENCHMARK == 0)
	  /* Check for dense dirty bits */
	  dbit_val = *dbitP;
	  if (LINE_UNALIGNED(dbit_val)) {
	    
	    /* Compute offset into dense dirty bit array */
	    offset = (dbit_val & LINE_DATA_VALUE) + line_offset;
	    dbitP = &vm_dense_dirty_base[ offset ];
	    
	    /* determine number of contiguous dense dbits remaining in this cache line */
	    dbit_len =  ri_line_size_in_bytes - line_offset;
	    if (dbit_len > data_len) 
	      dbit_len = data_len;
	    dbit_endP = (dbitP + dbit_len);
	    line_size = 0;	/* line is 1 byte, no shift required */
	    line_offset = 0;
	    
	    if (IS_DBG_LOCK_MLPD) {
	      sprintf(smsg, "dense dbit: addr=0x%x+%ld dense_dbitP 0x%x (%d), endP=0x%x",
		      target, data_len, dbitP, *dbitP, dbit_endP);
	      StatusMsg(smsg);
	    }
	    INCR_STATS(midLockAddrStats.unpack_ddbit_lookups, 1);
	    INCR_STATS(midLockAddrStats.unpack_ddbit_writes, dbit_len);

	  } else  {
#endif /* not benchmark */
	    
	    /* regular dirty bits */
	    line_size = ri_line_size;
	    if (data_len > (ri_line_size_in_bytes - line_offset)) { /*len > data in 1st line*/

	      /* compute number of dirty bits covered by this update.
	       * subtract data in the first cache line, then round up the remaining length
               * to a cache line multiple. ie
	       * (((data_len - (data_in_first_line)) + (line_size_in_bytes-1)) >> line_size) + 1
	       * which simplifies since data_in_first_line = line_size_in_bytes - line_offset
	       */
	      dbit_len = (data_len + line_offset - 1) >> line_size;
	      dbit_endP = dbitP + 1 + dbit_len;
	    } else
	      dbit_endP = dbitP + 1;
#if (BENCHMARK == 0)
	  }
#endif
	  first_dbitP = dbitP;	/* save ptr to 1st dbit to adjust line lengths */
	  
	  ASSERT (data_len > 0);
	  ASSERT (!(LINE_UNALIGNED(*dbitP)));
	  ASSERT (dbit_endP > dbitP);
	  

	  /*
	   * Loop setting/checking dirty bits for this update.
	   */
	  { register clock_t clock_v = clock_val;
#if ADDR_ENABLE_NON_EC
	    register u_long line_data_value = LINE_DATA_VALUE;
	    register u_long line_flag_value = LINE_FLAG_VALUE;
	    register u_long line_unaligned_value = LINE_EXCEPTION_FLAG|LINE_UNALIGNED_FLAG;
#endif
	    register long copied_len;	/* amount to copy */
	    register int  copy_data;	/* =TRUE to copy data from buffer to memory */
	    register int dbit_count;
	    register clock_t *dbit_loopEndP;
#if (BENCHMARK == 0)
	    register clock_t dbit_val0, dbit_val1, dbit_val2, dbit_val3;
#endif

	    copy_data = TRUE;

	    /*
	     * Don't run past end of dbits.  WAS 17Feb94
	     */
	    dbit_count = dbit_endP - dbitP;
	    dbit_count = (dbit_count / 8) * 8;
	    dbit_loopEndP = (dbitP + dbit_count);

	    for (dbit_startP = dbitP; dbitP < dbit_loopEndP; dbitP += 8) {

#if (BENCHMARK == 0)
	      /*** WARNING: if benchmark cannot use line exception code! ***/
	      dbit_val0 = dbitP[0];
	      dbit_val1 = dbitP[1];
	      dbit_val2 = dbitP[2];
	      dbit_val3 = dbitP[3];
	      /*
	       * If the dbit == 0 or dbit > clock then its an error
	       * Simplify logic by subtracting one from dbit value.
	       * WAS 17Feb94
	       */
	      if ((GREATER_THAN_OR_EQUAL_CLOCK (dbit_val0-1, clock_v)) ||
		  (GREATER_THAN_OR_EQUAL_CLOCK (dbit_val1-1, clock_v)) ||
		  (GREATER_THAN_OR_EQUAL_CLOCK (dbit_val2-1, clock_v)) ||
		  (GREATER_THAN_OR_EQUAL_CLOCK (dbit_val3-1, clock_v))) {

		/* Move check inside error case to optimize common case
		 * This depends on clock_t being an unsigned value, 
		 * w/ hi-order bit used for exception flag.  WAS 10Feb94
		 */
		if ((dbit_val0 & line_unaligned_value) == line_unaligned_value) {
		  ASSERT (dbitP > dbit_startP);
		  goto exit4_plus0;
		}
		if ((dbit_val1 & line_unaligned_value) == line_unaligned_value)
		  goto exit4_plus1;
		if ((dbit_val2 & line_unaligned_value) == line_unaligned_value)
		  goto exit4_plus2;
		if ((dbit_val3 & line_unaligned_value) == line_unaligned_value)
		  goto exit4_plus3;

		if (dbitP > dbit_startP) goto exit4_plus0;	/* print error on next loop iteration */
		
		if ((ctype != mid_MUD_static) || IS_DBG_LOCK_WNG) {
		  sprintf(smsg, "mid_unpack_data[%s]: ERROR - updating %s address 0x%lx for %lu",
			  cname, (dbit_val0 == NULL_CLOCK_VALUE) ? "dirty" : "newer", 
			  target, data_len);
		  StatusMsg(smsg);
		  StatusMsg("This typically results from the application modifying unlocked data.");
		  StatusMsg("Verify your locking before reporting this as a bug.");
#if STOP_ON_ERROR		  
		  /* HACK until new pack/unpack is debugged */
		  ASSERT(FALSE); 
#endif
		}
		
		/* Don't copy message data */
		copy_data = FALSE;
		dbitP += 8;	/* NOTE: We are skipping all lines in this loop iteration. WAS 17Feb94 */
		INCR_STATS(midLockAddrStats.unpack_dbit_errors, 1);
		break;
	      }
	      dbit_val0 = dbitP[4];
	      dbit_val1 = dbitP[5];
	      dbit_val2 = dbitP[6];
	      dbit_val3 = dbitP[7];
	      if ((GREATER_THAN_OR_EQUAL_CLOCK (dbit_val0-1, clock_v)) ||
		  (GREATER_THAN_OR_EQUAL_CLOCK (dbit_val1-1, clock_v)) ||
		  (GREATER_THAN_OR_EQUAL_CLOCK (dbit_val2-1, clock_v)) ||
		  (GREATER_THAN_OR_EQUAL_CLOCK (dbit_val3-1, clock_v))) {

		if ((dbit_val0 & line_unaligned_value) == line_unaligned_value) {
		  ASSERT (dbitP > dbit_startP);
		  goto exit4_plus4;
		}
		if ((dbit_val1 & line_unaligned_value) == line_unaligned_value)
		  goto exit4_plus5;
		if ((dbit_val2 & line_unaligned_value) == line_unaligned_value)
		  goto exit4_plus6;
		if ((dbit_val3 & line_unaligned_value) == line_unaligned_value)
		  goto exit4_plus7;

		if (dbitP > dbit_startP) goto exit4_plus0;	/* print error on next loop iteration */
		
		if ((ctype != mid_MUD_static) || IS_DBG_LOCK_WNG) {
		  sprintf(smsg, "mid_unpack_data[%s]: ERROR - updating %s address 0x%lx for %lu",
			  cname, (dbit_val0 == NULL_CLOCK_VALUE) ? "dirty" : "newer", 
			  target, data_len);
		  StatusMsg(smsg);
		  StatusMsg("This typically results from the application modifying unlocked data.");
		  StatusMsg("Verify your locking before reporting this as a bug.");
#if STOP_ON_ERROR		  
		  /* HACK until new pack/unpack is debugged */
		  ASSERT(FALSE); 
#endif
		}
		
		/* Don't copy message data */
		copy_data = FALSE;
		dbitP += 8;	/* NOTE: We are skipping all lines in this loop iteration. WAS 17Feb94 */
		INCR_STATS(midLockAddrStats.unpack_dbit_errors, 1);
		break;
	      }
#endif /* not BENCHMARK */

	      /* update dirtybit clock value and copy the data 
	       *
	       * NOTE: This code used to preserve the line_flag_value for each dbit
	       *  dbitP[0] = (dbit_val0 & line_flag_value) | clock_v;
	       * This currently isn't necessary.  WAS 17Feb94
	       */
	      dbitP[0] = clock_v;
	      dbitP[1] = clock_v;
	      dbitP[2] = clock_v;
	      dbitP[3] = clock_v;
	      dbitP[4] = clock_v;
	      dbitP[5] = clock_v;
	      dbitP[6] = clock_v;
	      dbitP[7] = clock_v;

	    } /* for (dbit_start = dbitP; dbitP < dbit_loopEndP;... */

	    /*
	     * Handle last 7 dbits.
	     */
#if (BENCHMARK == 0)
	    dbit_val0 = dbitP[0];
#endif
	    for (; dbitP < dbit_endP; dbitP++ ) {

#if (BENCHMARK == 0)
	      if (GREATER_THAN_OR_EQUAL_CLOCK (dbit_val0-1, clock_v)) {

		/* Move check inside error case to optimize common case
		 * This depends on clock_t being an unsigned value, 
		 * w/ hi-order bit used for exception flag.  WAS 10Feb94
		 */
		if ((dbit_val0 & line_unaligned_value) == line_unaligned_value) {
		  ASSERT (dbitP > dbit_startP);
		  goto exit4_plus0;
		}
		if (dbitP > dbit_startP) goto exit4_plus0;	/* print error on next loop iteration */
		
		if ((ctype != mid_MUD_static) || IS_DBG_LOCK_WNG) {
		  sprintf(smsg, "mid_unpack_data[%s]: ERROR - updating %s address 0x%lx for %lu",
			  cname, (dbit_val0 == NULL_CLOCK_VALUE) ? "dirty" : "newer", 
			  target, data_len);
		  StatusMsg(smsg);
		  StatusMsg("This typically results from the application modifying unlocked data.");
		  StatusMsg("Verify your locking before reporting this as a bug.");

#if STOP_ON_ERROR		  
		  /* HACK until new pack/unpack is debugged */
		  ASSERT(FALSE); 
#endif
		}
		
		/* Don't copy message data */
		copy_data = FALSE;
		dbitP++;	/* skip bad dbit only */
		INCR_STATS(midLockAddrStats.unpack_dbit_errors, 1);
		break;
	      }
#endif /* not BENCHMARK */

	      /* update dirtybit clock value and copy the data */
	      dbitP[0] = clock_v;
#if (BENCHMARK == 0)
	      dbit_val0 = dbitP[1];
#endif

	    } /* for (; dbitP < dbit_endP; ... */

	    goto exit4_plus0;
exit4_plus7:
	    dbitP += 7;
	    goto exit4_plus0;
exit4_plus6:
	    dbitP += 6;
	    goto exit4_plus0;
exit4_plus5:
	    dbitP += 5;
	    goto exit4_plus0;
exit4_plus4:
	    dbitP += 4;
	    goto exit4_plus0;
exit4_plus3:
	    dbitP += 3;
	    goto exit4_plus0;
exit4_plus2:
	    dbitP++;
exit4_plus1:
	    dbitP++;
exit4_plus0:
	    ASSERT (dbitP <= dbit_endP);
	    
	    copied_len = (dbitP - dbit_startP) << line_size;
	    if (dbit_startP == first_dbitP)
	      copied_len -= line_offset;
	    if (copied_len > data_len)
	      copied_len = data_len;
	    ASSERT (copied_len > 0);

	    INCR_STATS(midLockAddrStats.unpack_dbit_writes, (dbitP - dbit_startP));

#if STATS
#if STATS_DBIT
	    endfrc = *midway_kern_frc;
	    /* WARNING: this assumes we can only wrap once */
	    if (endfrc < startfrc)
	      midway_dbit_unpack_time += (FRC_MAX - startfrc) + endfrc;
	    else
	      midway_dbit_unpack_time += endfrc - startfrc;
#endif /*STATS_DBIT*/
#endif /*STATS*/

	    if (copy_data) {
	      
	      if (IS_DBG_LOCK_MLPD) {
		sprintf(smsg, "mid_unpack_data[%s]: updating 0x%lx+%lu from 0x%lx at clock=%lu", 
			cname, target, copied_len, source, clock_v);
		StatusMsg(smsg);	/* dbg */
	      }

#ifdef DEBUG_UNPACK
	      /***
	      **** NOTE: This code can be very expensive to execute, depending on the number
	      **** of addr maps. Spam was spending 2886.12 seconds in this loop (compared
	      **** to 2.73 seconds in the rest of the routine!).    WAS 5 Jan 94
	      ***/

	      /* verify that the target is bound to the lock/barrier being transferred */
	      { register s_addr_map_t *amapP;

		if (amap_listP != NULL) {
		  for (amapP = amap_listP; amapP != NULL; amapP = amapP->next) {
		    if ((target >= amapP->addr) && (target < amapP->end)) 
		      break;
		  }
		  ASSERT (amapP != NULL);
		  ASSERT (((target+copied_len) > amapP->addr) &&
			  ((target+copied_len) <= amapP->end));
		  ASSERT (copied_len <= amapP->len);
		}
	      }
#endif /* DEBUG_UNPACK */

	      bcopy(source, target, copied_len);
	    }
	    
	    /* advance to next dirty bit */
	    INCR_STATS(midLockAddrStats.unpack_data_bytes,copied_len);
	    source += copied_len;
	    target += copied_len;
	    data_len -= copied_len;
	    ASSERT (data_len >= 0);
	  }

	} /* while (data_len > 0) */

	/* advance to next address pair in buffer */
	ap = apnext;

      } /* for i = 0; etc. */

      as = (addrseg_t *) ap;	/* next segment would follow */
      break; /* case DATASEG */

    default:
      /* bail so upper level can figure this out */
      *pptr = (char *)as;	/* note where we left off */
      *lptr = length;		/* and how much was left */
      goto quick_exit;		/* better not be under lock */
      break;
    } /* switch */
  } /* while(length) */

exit:

  /* if here, bail gracefully */
  *pptr = 0;			/* all done */
  *lptr = 0;

quick_exit:

} /* mid_unpack_data */


/*
 ************************************************************************
 *	    m i d _ l o c k a d d r _ p r i n t _ s t a t s             *
 *----------------------------------------------------------------------*
 * Print lock addr statistics                                           *
 ************************************************************************
 */
mid_lockaddr_print_stats ()
{
#ifdef STATS 
  u_long adds, ins;
  char255 smsg;
#define MAX_STR_LEN 80

  adds = midLockAddrStats.mapadds;
  ins = adds - (midLockAddrStats.mapadd_tail + midLockAddrStats.mapadd_head +
		midLockAddrStats.mapadd_empty +
		midLockAddrStats.mapadd_beforeprev + midLockAddrStats.mapadd_afterprev);

  sprintf(smsg,
  "LockAddr Adds= %lu Empty= %lu Prev= %lu/%lu tail= %lu (%0.3f) head= %lu (%0.3f) middle= %lu", 
	  adds, midLockAddrStats.mapadd_empty,
	  midLockAddrStats.mapadd_beforeprev, midLockAddrStats.mapadd_afterprev,
	  midLockAddrStats.mapadd_tail, 
	  (adds == 0) ? 0.0 :
	    (double)(((double) midLockAddrStats.mapadd_tail) / ((double) adds)),
	  midLockAddrStats.mapadd_head, 
	  (adds == 0) ? 0.0 :
	    (double)(((double) midLockAddrStats.mapadd_head) / ((double) adds)),
	  ins);
  StatusMsg (smsg);

  sprintf(smsg,
  "LockAddr #Nexts= %lu (%0.3f) between= %lu start= %lu end= %lu enclosed= %lu Purge= %lu #Deleted= %lu",
	  midLockAddrStats.mapadd_next, 
	  (ins == 0) ? 0.0 :
	    (double)(((double) midLockAddrStats.mapadd_next) / ((double) ins)),
	  midLockAddrStats.mapadd_after, midLockAddrStats.mapadd_current_start, 
	  midLockAddrStats.mapadd_current_end, midLockAddrStats.mapadd_current,
	  midLockAddrStats.map_purges, midLockAddrStats.mapadd_merged);
  StatusMsg (smsg);

  sprintf(smsg,
  "Pack: calls %lu dbit lookups %lu reads %lu writes %lu wbytes %lu DENSE dbit lookups %lu reads %lu",
	  midLockAddrStats.pack_calls, midLockAddrStats.pack_dbit_lookups,
	  midLockAddrStats.pack_dbit_reads, midLockAddrStats.pack_dbit_writes,
	  midLockAddrStats.pack_dbit_wbytes,
	  midLockAddrStats.pack_ddbit_lookups, midLockAddrStats.pack_ddbit_reads);
  StatusMsg (smsg);

  sprintf(smsg,
  "Unpack: calls %lu dbit lookups %lu writes %lu errors %lu DENSE dbit lookups %lu writes %lu DATA %lu ",
	  midLockAddrStats.unpack_calls, midLockAddrStats.unpack_dbit_lookups,
	  midLockAddrStats.unpack_dbit_writes, midLockAddrStats.unpack_dbit_errors,
	  midLockAddrStats.unpack_ddbit_lookups, midLockAddrStats.unpack_ddbit_writes,
	  midLockAddrStats.unpack_data_bytes);
  StatusMsg (smsg);
  

#endif /* STATS */
}

/*
 * Allow the gathering of dbit stats based on acquire of a certain # of
 * dbits.  Must be fn with new midway_{start,end}.o technology
 */
unsigned int
mid_lockaddr_get_db_packtime()
{
#if (STATS && STATS_DBIT)
  return(midway_dbit_pack_time);
#else
  return(0);
#endif
}
void
mid_lockaddr_reset_db_packtime()
{
#if (STATS && STATS_DBIT)
  midway_dbit_pack_time = 0;
#endif
}
unsigned int
mid_lockaddr_get_db_unpacktime()
{
#if (STATS && STATS_DBIT)
  return(midway_dbit_unpack_time);
#else
  return(0);
#endif
}
void
mid_lockaddr_reset_db_unpacktime()
{
#if (STATS && STATS_DBIT)
  midway_dbit_unpack_time = 0;
#endif
}


