/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * midway-all.h: things that will be common to everyone!
 *
 * For now, just place in conditional debugging printf stuff
 *
 * 16 Dec 91, Matthew J. Zekauskas
 * 18 Dec 91 (mjz): start adding locking stuff
 * 19 Mar 92 (mjz): add misc
 *  9 Jun 92 (mjz): add CPT for COMM
 * 18 Jun 92 (was): added midway_Port_debug & DBG_PORT_xxx flags 
 *  6 Jul 92 (mjz): add TRACE -- will place last 10 msgs per host in a buffer
 * 18 Aug 92 (was): added more DBG_PORT flags
 *  3 Oct 92 (mjz): add Xpager
 * 27 Oct 92 (was): added BENCHMARK defs
 * 29 Oct 92 (was): separated STATS from DEBUG
 *  3 Nov 92 (mjz): add WIRE_DA_THREADS
 *  5 Nov 92 (was): add stats for locks
 *  2 Dec 92 (mjz): add conditional mutex names
 * 13 Mar 93 (was): added DEBUG_USER flag
 * 19 Oct 93 (mjz): fixed ASSERT.  How did this ever work?
 *  7 Mar 94 (was): added extra flags for PAGE_BASED version.
 * 14 Jun 94 (was): added MALLOC, CALLOC & VM_ALLOCATE
 */
#ifndef _MIDWAY_DBG_H_
#define _MIDWAY_DBG_H_

#include <sys/types.h>
#include <mach.h>
#include <cthreads.h>

/*
 * always define DEBUG for now, so debugging printf's get compiled in
 * (although they *will* get compiled in with conditionals
 */
#ifdef BENCHMARK
/*
 * Compile code for benchmarking. Disable DEBUG code
 */
#undef DEBUG
#undef DEBUG_PRT
#undef DEBUG_USER
#  ifndef WIRE_DA_THREADS
#    define WIRE_DA_THREADS 1
#  endif
#  define STATS 1
#  define FAST_STATS 1
#
#  undef ASSERT
#  define ASSERT(x)
#
#  undef TRACE
#
#  define MALLOC 	malloc
#  define CALLOC 	calloc
#  define FREE          free
#  define VM_ALLOCATE	vm_allocate
#  define VM_DEALLOCATE	vm_deallocate

#else
   /* prevent duplicate messages */
#  ifndef DEBUG
#    define DEBUG 1
#  endif
   /* this one for comm.h --> cause pmsg (message xport debugging) compiled in*/
#  ifndef DEBUG_PRT
#    define DEBUG_PRT 1
#  endif
#  ifndef DEBUG_USER
#    define DEBUG_USER 1
#  endif
#  ifndef STATS    
#    define STATS 1
#  endif
#  ifndef WIRE_DA_THREADS
#    define WIRE_DA_THREADS 1
#  endif
#  undef LTIMES
#  ifndef MIDWAY_MUTEX_NAMES
#    define MIDWAY_MUTEX_NAMES 1
#  endif
#  undef ASSERT
#  define ASSERT(x) if (!(x)) botch(botch_message(__FILE__,__LINE__)); else ;
#
#  undef SET_PRIORITY
#  define SET_PRIORITY 0   /* run atm at low priority for debugging */
#
#  ifndef TRACE
#    define TRACE 1
#    define TRACE_TIME 1
#  endif
#
#  if STATS > 0
#    define MALLOC 	midway_malloc
#    define CALLOC 	midway_calloc
#    define FREE        midway_free
#    define VM_ALLOCATE	midway_vm_allocate
#    define VM_DEALLOCATE midway_vm_deallocate
#  else
#    define MALLOC 	malloc
#    define CALLOC 	calloc
#    define FREE        free
#    define VM_ALLOCATE	vm_allocate
#    define VM_DEALLOCATE vm_deallocate
#  endif

#endif


#if MIDWAY_MUTEX_NAMES
#  define MIDWAY_MUTEX_SET_NAME(a,b) mutex_set_name(a,b)
#else
#  define MIDWAY_MUTEX_SET_NAME(a,b)
#endif


/*
 * Page based version like to have the addrmaps sorted
 * to reduce the amount of diff'ing required.
 */
#ifdef PAGE_BASED
#  define SORT_ADDRMAPS
#endif



/*
 * Comm layer
 */
extern unsigned long midway_Comm_debug;
#define DBG_COMM_SCAN	0x00000001 /* Scanner thread */
#define DBG_COMM_REQ	0x00000002 /* Comm<M>Request's */
#define DBG_COMM_SEND   0x00000004 /* CommSend */
#define DBG_COMM_RCV    0x00000008 /* CommReceive */
#define DBG_COMM_MSGS   0x00000010 /* enable pmsg */
#define DBG_COMM_WORK   0x00000020 /* CommWorker (maybe break out?) */
#define DBG_COMM_WNGS	0x00000040 /* Warnings, like in process or duplicate */
#define DBG_COMM_CPT	0x00000080 /* For debugging cpt init, mainly */
#define DBG_COMM_TRACE	0x00000100 /* buffer last messages */
#define DBG_COMM_STATS  0x00000200 /* enable printing of msg stats */
#define DBG_COMM_MSGX	0x00000400 /* extended packet decoding */
#define DBG_COMM_ATM	0x00001000 /* ATM debugging - principally open */


#ifndef BENCHMARK
#define IS_DBG_COMM_SCAN (midway_Comm_debug & DBG_COMM_SCAN)
#define IS_DBG_COMM_REQ  (midway_Comm_debug & DBG_COMM_REQ)
#define IS_DBG_COMM_SEND (midway_Comm_debug & DBG_COMM_SEND)
#define IS_DBG_COMM_RCV  (midway_Comm_debug & DBG_COMM_RCV)
#define IS_DBG_COMM_MSGS (midway_Comm_debug & DBG_COMM_MSGS)
#define IS_DBG_COMM_WORK (midway_Comm_debug & DBG_COMM_WORK)
#define IS_DBG_COMM_WNGS (midway_Comm_debug & DBG_COMM_WNGS)
#define IS_DBG_COMM_CPT	 (midway_Comm_debug & DBG_COMM_CPT)
#define IS_DBG_COMM_TRACE (midway_Comm_debug & DBG_COMM_TRACE)
#define IS_DBG_COMM_STATS (midway_Comm_debug & DBG_COMM_STATS)
#define IS_DBG_COMM_MSGX (midway_Comm_debug & DBG_COMM_MSGX)
#define IS_DBG_COMM_ATM	 (midway_Comm_debug & DBG_COMM_ATM)
#else
#define IS_DBG_COMM_SCAN  0
#define IS_DBG_COMM_REQ   0
#define IS_DBG_COMM_SEND  0
#define IS_DBG_COMM_RCV   0
#define IS_DBG_COMM_MSGS  0
#define IS_DBG_COMM_WORK  0
#define IS_DBG_COMM_WNGS  0
#define IS_DBG_COMM_CPT	  0
#define IS_DBG_COMM_TRACE 0
#define IS_DBG_COMM_STATS 0
#define IS_DBG_COMM_MSGX  0
#define IS_DBG_COMM_ATM   0
#endif


/*
 * Lock layer
 */
extern unsigned long midway_Lock_debug;
#define DBG_LOCK_EXTR	0x00000001 /* mid[Ext]LockRequest, midLockReply */
#define DBG_LOCK_OPS	0x00000002 /* Note Acquire/Release */
#define DBG_LOCK_WNG	0x00000004 /* Things that are not errors, but help */
				   /* ..decode state machinery */
#define DBG_LOCK_MSGS	0x00000010 /* make it like Comm for sanity */
#define DBG_LOCK_MLPD	0x00000020 /* midLock[Pack|Decode]Data */
#define DBG_LOCK_FORK	0x00000040 /* for handlin Create: Release ReleaseReply */
#define DBG_LOCK_BIND	0x00000080 /* for bind_synch_obj */
#define DBG_LOCK_DLTA   0x00000100 /* for mid_addrmap_delta & friends */
#define DBG_LOCK_NOSTATIC 0x80000000 /* DO NOT transfer statics upon release */

#ifndef BENCHMARK
#define IS_DBG_LOCK_EXTR (midway_Lock_debug & DBG_LOCK_EXTR)
#define IS_DBG_LOCK_OPS  (midway_Lock_debug & DBG_LOCK_OPS)
#define IS_DBG_LOCK_WNG  (midway_Lock_debug & DBG_LOCK_WNG)
#define IS_DBG_LOCK_MSGS (midway_Lock_debug & DBG_LOCK_MSGS)
#define IS_DBG_LOCK_MLPD (midway_Lock_debug & DBG_LOCK_MLPD)
#define IS_DBG_LOCK_FORK (midway_Lock_debug & DBG_LOCK_FORK)
#define IS_DBG_LOCK_BIND (midway_Lock_debug & DBG_LOCK_BIND)
#define IS_DBG_LOCK_DLTA (midway_Lock_debug & DBG_LOCK_DLTA)
#define IS_DBG_LOCK_NOSTATIC (midway_Lock_debug & DBG_LOCK_NOSTATIC)
#else
#define IS_DBG_LOCK_EXTR     0
#define IS_DBG_LOCK_OPS      0
#define IS_DBG_LOCK_WNG      0
#define IS_DBG_LOCK_MSGS     0
#define IS_DBG_LOCK_MLPD     0
#define IS_DBG_LOCK_FORK     0
#define IS_DBG_LOCK_BIND     0
#define IS_DBG_LOCK_DLTA     0
#define IS_DBG_LOCK_NOSTATIC 0
#endif

/*
 * Barrier
 */
extern unsigned long midway_Barr_debug;
#define DBG_BARR_EXTR	0x00000001 /* mid[Ext]BarrierRequest,midBarrierReply */
#define DBG_BARR_OPS	0x00000002 /* Note Barrier crossing */
#define DBG_BARR_WNG	0x00000004 /* Things that are not errors, but help */
				   /* ..decode state machinery */
#define DBG_BARR_MSGS	0x00000010 /* make it like Comm for sanity */

#ifndef BENCHMARK
#define IS_DBG_BARR_EXTR (midway_Barr_debug & DBG_BARR_EXTR)
#define IS_DBG_BARR_OPS  (midway_Barr_debug & DBG_BARR_OPS)
#define IS_DBG_BARR_WNG  (midway_Barr_debug & DBG_BARR_WNG)
#define IS_DBG_BARR_MSGS (midway_Barr_debug & DBG_BARR_MSGS)
#else
#define IS_DBG_BARR_EXTR 0
#define IS_DBG_BARR_OPS  0
#define IS_DBG_BARR_WNG  0
#define IS_DBG_BARR_MSGS 0
#endif

/*
 * Miscellaneous flags (only need one or two, not lots...)
 */
extern unsigned long midway_Misc_debug;
#define DBG_MISC_INIT	0x00000001 /* Initialization, mainly in realmain.c */
#define DBG_MISC_MALL	0x00000010 /* shmalloc.c -> shmalloc/shfree */
#define DBG_MISC_MAIF	0x00000020 /* shmalloc.c -> midAllocInfo, etc */
#define DBG_MISC_MREG	0x00000040 /* shregion.c -> region allocation */

#ifndef BENCHMARK
#define IS_DBG_MISC_INIT (midway_Misc_debug & DBG_MISC_INIT)
#define IS_DBG_MISC_MALL (midway_Misc_debug & DBG_MISC_MALL)
#define IS_DBG_MISC_MAIF (midway_Misc_debug & DBG_MISC_MAIF)
#define IS_DBG_MISC_MREG (midway_Misc_debug & DBG_MISC_MREG)
#else
#define IS_DBG_MISC_INIT 0
#define IS_DBG_MISC_MALL 0
#define IS_DBG_MISC_MAIF 0
#define IS_DBG_MISC_MREG 0
#endif

/*
 * External Pager
 */
extern unsigned long midway_Xpgr_debug;
#define DBG_XPGR_OPS 	0x00000001 /* Cause xpager ops to whine */
#define DBG_XPGR_SYSL	0x00000010 /* Have errors go to syslog, not stderr */

/* NOTE: XPGR_OPS currently just sets midway_xpager_debug to one.  To catch
 * the initialization routines, you must patch this variable, since xpager
 * initialization is called from shmalloc_init, which in turn is called
 * before the arguments are parsed...
 */

#ifndef BENCHMARK
#define IS_DBG_XPGR_OPS  (midway_Xpgr_debug & DBG_XPGR_OPS)
#define IS_DBG_XPGR_SYSL (midway_Xpgr_debug & DBG_XPGR_SYSL)
#else
#define IS_DBG_XPGR_OPS  0
#define IS_DBG_XPGR_SYSL 0
#endif


/*
 * Model conversion support code
 */
extern unsigned long midway_Port_debug;
#define DBG_PORT_QUEUE  0x00000001      /* update queue adds/requests */
#define DBG_PORT_PAGE   0x00000002      /* page faults/requests */
#define DBG_PORT_LINEX  0x00000004      /* line exception code */
#define DBG_PORT_UPDATE 0x00000004      /* page updates requests */
#define DBG_PORT_STATS  0x00000200      /* print pager stats when done */

#ifndef BENCHMARK
#define IS_DBG_PORT_QUEUE  (midway_Port_debug & DBG_PORT_QUEUE)
#define IS_DBG_PORT_PAGE   (midway_Port_debug & DBG_PORT_PAGE)
#define IS_DBG_PORT_LINEX  (midway_Port_debug & DBG_PORT_LINEX)
#define IS_DBG_PORT_UPDATE (midway_Port_debug & DBG_PORT_UPDATE)
#define IS_DBG_PORT_STATS  (midway_Port_debug & DBG_PORT_STATS)
#else
#define IS_DBG_PORT_QUEUE  0
#define IS_DBG_PORT_PAGE   0
#define IS_DBG_PORT_LINEX  0
#define IS_DBG_PORT_UPDATE 0
#define IS_DBG_PORT_STATS  0
#endif


/*
 * Flags for collecting statistics
 */
extern unsigned long midway_Stat_debug;
#define DBG_STAT_COMM     0x00000001     /* communication stats */
#define DBG_STAT_PAGEMGR  0x00000002     /* Non-EC page manager */
#define DBG_STAT_UPDATEQ  0x00000004     /* Non-EC update queue */
#define DBG_STAT_LOCK     0x00000008     /* lock manager stats  */
#define DBG_STAT_XPAGER   0x00000010     /* xpager stats  */
#define DBG_STAT_SHMALLOC 0x00000020     /* shmalloc/shfree */

#ifdef STATS
#define IS_DBG_STAT_COMM      \
  ((midway_Stat_debug & DBG_STAT_COMM) || (IS_DBG_COMM_STATS))
#define IS_DBG_STAT_PAGEMGR   \
  ((midway_Stat_debug & DBG_STAT_PAGEMGR) || (IS_DBG_PORT_STATS))
#define IS_DBG_STAT_UPDATEQ   \
  ((midway_Stat_debug & DBG_STAT_UPDATEQ) || (IS_DBG_PORT_STATS))
#define IS_DBG_STAT_LOCK   \
   (midway_Stat_debug & DBG_STAT_LOCK)
#define IS_DBG_STAT_XPAGER \
   (midway_Stat_debug & DBG_STAT_XPAGER)
#define IS_DBG_STAT_SHMALLOC \
   (midway_Stat_debug & DBG_STAT_SHMALLOC)
#else
#define IS_DBG_STAT_COMM    (IS_DBG_COMM_STATS)
#define IS_DBG_STAT_PAGEMGR (IS_DBG_PORT_STATS)
#define IS_DBG_STAT_UPDATEQ (IS_DBG_PORT_STATS)
#define IS_DBG_STAT_LOCK     0
#define IS_DBG_STAT_XPAGER   0
#define IS_DBG_STAT_SHMALLOC 0
#endif



#if TRACE > 0

/*
 * Define trace table for debugging
 * Simply keep a trace of the last few events
 * to allow us to see recent history when debugging the code.
 */
#define TRACE_SIZE (256 * 1024)

extern unsigned long midTraceData[TRACE_SIZE];
extern unsigned long midTraceIndex;
extern mutex_t midTraceLock;

vm_offset_t midTraceDetailLow, midTraceDetailHigh; /* detailed tracing in this addr range */

/*
 * First Item is a ptr to trace value (ie a unique 4 character "name")
 * All other items are actual trace values.
 */
#define TRACE_I(x) \
{ midTraceData[ midTraceIndex ] = (u_long)(x); midTraceIndex = (midTraceIndex + 1) % TRACE_SIZE; }

#if TRACE_TIME > 0

/* Timestamp all trace table entries */
extern u_int *midway_kern_frc;

#  define TRACE_P(x) \
   { midTraceData[ midTraceIndex ] = *(u_long *)(x); midTraceIndex = (midTraceIndex + 1) % TRACE_SIZE; \
     midTraceData[ midTraceIndex ] = *midway_kern_frc; midTraceIndex = (midTraceIndex + 1) % TRACE_SIZE;  \
   }
#else
#  define TRACE_P(x) \
   { midTraceData[ midTraceIndex ] = *(u_long *)(x); midTraceIndex = (midTraceIndex + 1) % TRACE_SIZE; }
#endif

#define TRACE_1(x1) \
{ mutex_lock(midTraceLock); TRACE_P(x1); mutex_unlock(midTraceLock); }
#define TRACE_2(x1,x2) \
{ mutex_lock(midTraceLock); TRACE_P(x1); TRACE_I(x2); mutex_unlock(midTraceLock); }
#define TRACE_3(x1,x2,x3) \
{ mutex_lock(midTraceLock); TRACE_P(x1); TRACE_I(x2); TRACE_I(x3); mutex_unlock(midTraceLock); }
#define TRACE_4(x1,x2,x3,x4) \
{ mutex_lock(midTraceLock); TRACE_P(x1); TRACE_I(x2); TRACE_I(x3); \
  TRACE_I(x4); mutex_unlock(midTraceLock); }
#define TRACE_5(x1,x2,x3,x4,x5) \
{ mutex_lock(midTraceLock); TRACE_P(x1); TRACE_I(x2); TRACE_I(x3); \
  TRACE_I(x4); TRACE_I(x5); mutex_unlock(midTraceLock); }
#define TRACE_6(x1,x2,x3,x4,x5,x6) \
{ mutex_lock(midTraceLock); TRACE_P(x1); TRACE_I(x2); TRACE_I(x3); \
  TRACE_I(x4); TRACE_I(x5); TRACE_I(x6); mutex_unlock(midTraceLock); }

#else

#define TRACE_1(x1)
#define TRACE_2(x1,x2)
#define TRACE_3(x1,x2,x3)
#define TRACE_4(x1,x2,x3,x4)
#define TRACE_5(x1,x2,x3,x4,x5)
#define TRACE_6(x1,x2,x3,x4,x5,x6)

#endif /* TRACE */

extern void midway_init_trace(void);
extern void midway_print_trace(long how_many_entries);
extern void midway_reset_trace(void);



#endif /* _MIDWAY_DBG_H_ */
