/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * midway.c: support routines for dirty bit twiddling, also the initialization
 *	routine for all of midway.
 *
 * June, 1991 - bnb
 *
 * 13 Aug 91 (bnb): fix to generalize so not pmax specific [cmt by mjz]
 * 15 Aug 91 (mjz): midway_init now called instead of main or midway_main
 *		by crt0.
 * 22 Aug 91 (mjz): use midway_mem.h instead of midway.h
 * 18 Dec 91 (mjz): enumerate: use LAST_STATIC_SYM instead of end ('cause
 *		need &end on some systems...
 * 11 Feb 92 (mjz): add midway_addr_from_dirty_addr -- for now.
 * 19 Mar 92 (mjz): fix it!
 * 10 Jun 92 (mjz): enumerate: looks like static stuff is funny to me;
 *		try something else...
 * 16 Jul 92 (mjz): fix sanity check in midway_dirty_info_address()
 * 11 Aug 92 (was): added clock_t type
 * 28 Sep 92 (mjz): From (was): fix for no dirtybits (i.e. shared-maybe on
 *		    private area that is not on stack!)
 * 26 May 93 (was): added midway_dense_dirty_info_address
 * 14 Jul 93 (was): added midway_dirtybit_info 
 * 25 Feb 94 (was): added midway_get_pid to safely export Comm_index
 * 14 Mar 94 (was): added midTrace code 
 */
#include <sys/types.h>		/* u_long & friends */
#include <stdio.h>
#include <ctype.h>
#include "midway_types.h"
#include "smexecd_types.h"	/* ensure get char255 */
#include "midway_mem.h"
#include "midway-dbg.h"		/* for ASSERT */
#include "shmalloc.h"


extern int etext, edata, end, _ftext, _fdata, _fbss;

extern clock_t *vm_dense_dirty_base;

extern int Comm_index; 

int midway_init();
int (*midway_init_routine)() = midway_init;


#ifdef STATS

#define STAT_TABLE_SIZE 32
static struct {
    unsigned long malloc_calls;
    unsigned long malloc_size;
    unsigned long malloc_size_count[STAT_TABLE_SIZE];
    unsigned long calloc_calls;
    unsigned long calloc_size;
    unsigned long calloc_size_count[STAT_TABLE_SIZE];
    unsigned long free_calls;
} midStats;
static mutex_t statLock;

# ifdef FAST_STATS
#  define INCR_STATS(x,by) (x) += (by)
# else
#  define INCR_STATS(x,by) \
     { mutex_lock(statLock);  (x) += (by);  mutex_unlock(statLock); }
# endif 
#else
#   define INCR_STATS(x,by)
#endif


/*
 * midway_init is called *INSTEAD OF* main() by the midway version of crt0.
 */
int
midway_init(argc, argv, envp)
     int argc;
     char **argv;
     char **envp;
{
  int i;
  /* freeze at start of everything to debug init code.  bah. */
  for (i=1; i<argc; i++) {
    if (!strncmp(argv[i], "-MIDDBGff", 9)) {
      freeze_for_debugger();
      break;
    }
  }

#if STATS
  statLock = mutex_alloc ();
  MIDWAY_MUTEX_SET_NAME(statLock,"midway.c:statLock");
  bzero (&midStats, sizeof(midStats));
#endif

/*    printf("before shmalloc_init");*/ /*XXX*/
    shmalloc_init();
/*    printf("after shmalloc_init, calling midway_main");*//*XXX*/
    argc = midway_main(argc, argv, envp);
    return(argc);
}



/*
 * Define static variable to hold local clock.
 */
static clock_t midway_local_clock = MIN_CLOCK_VALUE;

    
clock_t
midway_get_local_clock()
{
    return midway_local_clock;
}


clock_t
midway_set_local_clock(c)
    clock_t c;
{
    clock_t oclock = midway_local_clock;
    if (GREATER_THAN_CLOCK (oclock, c))  {
	printf("warning: time is running backwards\n");
    }
    if (GREATER_THAN_CLOCK(c,MAX_CLOCK_VALUE) ||
	LESS_THAN_CLOCK(c,MIN_CLOCK_VALUE)) {
      printf("warning: setting time to invalid value %lu\n", c);
    }
    midway_local_clock = c;
    return oclock;
}


/***********************************************************/


int
midway_int_to_log_int(l)
    int l;
{
    register unsigned int   shiftr = l;
    register int k = 0;
    while (shiftr >>= 1)
	    k++;
    return (k);
}



/***********************************************************/

midway_enumerate_dirty_items(func, cl)
    int (*func)();
    int *cl;
{
    int i,j;
    int k;
    int region_size;
    region_info ri;
    clock_t *ddbP;
    extern region_info region_from_region_num();


    for (i = 0;; i++)  {
	ri = region_from_region_num(i);
	if (!ri) break;
	if (ri->line_info)  {
	    int linesz = RI_LINE_SIZE_TO_BYTES(ri->line_size);
	    int nlines;
	    /*
	     * Compute size of this region. dynamic regions are easy. static
	     * regions are easy unless it's the final static region.
             */
	    if (ri == end_static_ri)  {
		int rs1;
		rs1 = ((int)(LAST_STATIC_SYM))&(REGION_SIZE-1);
		region_size = ((int)LAST_STATIC_SYM) - 
		  (((int)(LAST_STATIC_SYM))&~(REGION_SIZE-1));
		if (rs1 != region_size) {
			printf("Brian and Matt agree to disagree\n");
		}
	    } else region_size = REGION_SIZE;
	    nlines = region_size / linesz;
	    
	    for (k = 0; k < nlines; k++)  {
		if (ri->line_info[k] == NULL_CLOCK_VALUE)  {
		    if ((*func)(cl, addr_from_region(ri, k)))  {
			return 0;
		    }
		} else if (LINE_UNALIGNED(ri->line_info[k])) {
		  ddbP = &vm_dense_dirty_base[ ri->line_info[k] & LINE_DATA_VALUE ];
		  for (j = 0; j < linesz; j++) {
		    if (*ddbP == NULL_CLOCK_VALUE) {
		      if ((*func)(cl, addr_from_region(ri, k)+j))
			return 0;
		    }
		    ddbP++;
		  }
		}
		
	    }
	}
    }
    return 0;
}



clock_t *
midway_dense_dirty_info_address(addr, riP, dbit)
     char *addr;
     region_info riP;		/* may be null */
     clock_t dbit;
{
    clock_t *ddbP;
    u_int ddb_offset;
    char *line_addr;

    if (riP == NULL) {
      riP = region_from_addr(addr);
      if (riP == 0)
	return 0;
      if (riP->line_info == 0)
	return 0;
    }

    line_addr = (char *)((((u_int)addr) >> riP->line_size) << riP->line_size);
    ddb_offset = (dbit & LINE_DATA_VALUE) + (addr - line_addr);
    ddbP = &vm_dense_dirty_base[ ddb_offset ];
					     
    return (ddbP);
}


clock_t *
midway_dirty_info_address(p)
    int *p;
{
    region_info ri;
    int offset;
    clock_t *db;

    ri = region_from_addr(p);
    if (ri == 0)  {
	return 0;
    }
    /*
     * Check if storage was allocated.  WAS 18Sep92
     */
    if (ri->line_info == 0)
      return 0;

#ifdef OLD_WAY
    /* 16 Jul 92, mjz:
     * Today I was trying to get the 386 version to work, and I was
     * failing on the following statement.  Why?  Because FIRST_STATIC_SYM
     * was 2f000, which is certainly not on a megabyte boundry.
     * shmalloc_init dutifully computed the number of dirtybits needed,
     * and then called region_create_map with FIRST_STATIC_SYM.
     * region_create_map copied it into region info... and there we are.
     * How come this never failed on the vax version?
     * I am going to fix this by doing the same AND for ri->base.
     * I think this works properly... why?  
     * -- All calculations are actually done with ri->base,
     *    and don't depend on megabyte boundaries.
     * -- We will never allocate any shared storage (shmalloc) before
     *	  the first region following the one containing the static
     *    data.  ==> nothing before the text, and nothing else after
     *    static in that megabyte region. ==> nothing but the static
     *    data in that region ever needs dirty bits.
     * The other alternative would be to fix FIRST_STATIC_SYM to
     * always round down to a megabyte boundry.
     */
    if (ri->base != (int*)((int)p & ~(REGION_SIZE-1)))  {
	botch("midway_dirty_info_address: botched region info");
    }
#else 
    if (((long)ri->base & ~(REGION_SIZE-1)) != ((long)p & ~(REGION_SIZE-1)))  {
	botch("midway_dirty_info_address: botched region info");
    }
#endif OLD_WAY
    offset = (int)(p) - (int)ri->base;
    offset = (offset >> ri->line_size);	/* divide by line size */
    db = &ri->line_info[offset];


    /* Check for dense dirty bits */
    if (LINE_UNALIGNED(*db)) {
      db = midway_dense_dirty_info_address(p,ri,*db);
    }

    return db;
}

int 
midway_line_size_address(p)
    int *p;
{
    region_info ri;
    int offset;
    int linesz_bytes;

    ri = region_from_addr(p);
    if (ri == 0)  {
	return 0;
    }
    linesz_bytes = RI_LINE_SIZE_TO_BYTES(ri->line_size);
    return linesz_bytes;
}

/*
 * act like addr_from_region() in shmalloc.c, but this one given
 * just dirty address.  currently used by lock.c to figure out what
 * address got stomped on if already dirty on data update.
 *  Call this one an unsupported interface that may go away...
 *
 * NOTE: This code fails if called with a dense dirty bit address. WAS 26May93 
 */
int *
midway_addr_from_dirty_addr(addr,info)
     int* addr;			/* address in region (to find ri) */
     clock_t* info;		/* dirty info address */
{
  region_info ri;
  int offset;
  char255 smsg;

  ri = region_from_addr(addr);
  offset = info - &(ri->line_info[0]);
  if (offset < 0) {
    sprintf(smsg,
      "midway_addr_from_dirty_addr: address 0x%x and dirty bit info not in same region (offset < 0)", 
      addr);
    StatusMsg (smsg);
  }

/* not necessary - "info -" does the right thing in C; if we casted 
 * info and the start of the dirtybit addresses to (int) or (u_long),
 * then we would have needed it...
  offset = offset / sizeof(ri->line_info[0]); /- * convert to index * -/
 */
  offset = offset << ri->line_size; /* convert to byte offset */
  if (offset > REGION_SIZE) {
    sprintf(smsg,
      "midway_addr_from_dirty_addr: address 0x%x and dirty bit info 0x%x not in same region (offset > REGION_SIZE)",
      addr, (int)(ri->base) + offset);
    StatusMsg (smsg);
  }
  return((int *) ((int)(ri->base) + offset));
} /* midway_addr_from_dirty_addr */

int
midway_cache_line_size(p)
    int *p;
{
    region_info ri = region_from_addr(p);
    return (ri ? RI_LINE_SIZE_TO_BYTES(ri->line_size) : 0);
}


int
midway_line_dirty(p)
    int *p;
{
    clock_t *info = midway_dirty_info_address(p);
    if (!info) return(0);	/* WAS 18sep92 */
    if (LINE_UNALIGNED(*info))
      info = midway_dense_dirty_info_address(p,NULL,*info);
    return (*info == NULL_CLOCK_VALUE);
}

void
midway_clean_line(p)
    int *p;
{
    clock_t *info = midway_dirty_info_address(p);
    if (info) {
      if (LINE_UNALIGNED(*info))
	info = midway_dense_dirty_info_address(p,NULL,*info);
      *info = midway_get_local_clock();
    }
}


int*
midway_get_start_of_data_segment()
{
    etext = etext;
    edata = edata;
    end = end;
#if mips
    _ftext = _ftext;
    _fdata = _fdata;
    _fbss = _fbss;
#endif
}

char *
midway_malloc(unsigned size)
{
  char *p;
  extern char *vm_private_base, *vm_private_free_spaceP;

#ifdef STATS
  INCR_STATS (midStats.malloc_calls, 1);
  INCR_STATS (midStats.malloc_size, size);
  { register unsigned s = size >> 3;
    register unsigned i = 2;
    while (s) { s >>= 1; i++; }
    ASSERT ((i >= 0) && (i < STAT_TABLE_SIZE));
    INCR_STATS (midStats.malloc_size_count[i], 1);
  }
#endif

  p = (char *) malloc(size);
/***************
  if ((vm_private_base != 0) && (p != 0) &&
      ((p < vm_private_base) || (p >= vm_private_free_spaceP))) {
    printf("Invalid address returned by malloc(%u) = 0x%x\n", size, p);
    botch("midway_malloc");
  }
************/
  return p;
}

char *
midway_calloc(unsigned count, unsigned size)
{
  char *p;
  extern char *vm_private_base, *vm_private_free_spaceP;

#ifdef STATS
  INCR_STATS (midStats.calloc_calls, 1);
  INCR_STATS (midStats.calloc_size, count*size);
  { register unsigned s = (count*size) >> 3;
    register unsigned i = 2;
    while (s) { s >>= 1; i++; }
    ASSERT ((i >= 0) && (i < STAT_TABLE_SIZE));
    INCR_STATS (midStats.calloc_size_count[i], 1);
  }
#endif

  p = (char *) calloc(count,size);
/***************
  if ((vm_private_base != 0) && (p != 0) &&
      ((p < vm_private_base) || (p >= vm_private_free_spaceP))) {
    printf("Invalid address returned by calloc(%u,%u) = 0x%x\n", count, size, p);
    botch("midway_calloc");
  }
***************/
  return p;
}

char *
midway_free(void *p)
{
    INCR_STATS (midStats.free_calls, 1);

    free(p);
}

int 
midway_get_pid (void)
{
    /* 
     * Return unique process id for this machine in the range [0 to P-1].
     *
     * Can't export Comm_index directly, since an application's
     * reference to it will move it into the application's 
     * static variable area (comm sbss), causing the page-based 
     * version of Midway to break.   WAS 25Feb94
     */
    return (Comm_index - 1);
}


/*
 ************************************************************************
 *	    m i d w a y _ p r i n t _ t r a c e       	                *
 *----------------------------------------------------------------------*
 * Print trace table 							*
 ************************************************************************
 */
#if TRACE
/*
 * Define trace table for debugging
 * Simply keep a trace of the last few events
 * to allow us to see recent history when debugging the code.
 */
unsigned long midTraceData[ TRACE_SIZE ];
unsigned long midTraceIndex;
mutex_t midTraceLock = 0;

vm_offset_t midTraceDetailLow, midTraceDetailHigh; /* detailed tracing in this addr range */

#endif /* TRACE */


void midway_init_trace(void)
{
#if TRACE
  if (midTraceLock == 0) {	/* allow to be call > 1ce */
    midTraceLock = mutex_alloc();
    bzero (midTraceData, sizeof(midTraceData));
    midTraceIndex = 0;
    midTraceDetailLow = 0;
    midTraceDetailHigh = midTraceDetailLow;
  }
#endif
}

void midway_reset_trace(void)
{
#if TRACE
    bzero (midTraceData, sizeof(midTraceData));
    midTraceIndex = 0;
#endif
}

void midway_print_trace(long how_many_entries)
{
#if (TRACE == 0)
    StatusMsg ("Not compiled to collect Trace data");
#else
    int i, j, start, end, end2, repeat;
    char *cP, cbuf[17];
    int not_local;		/* != 0 if uxpass was remote */
    
    if ((how_many_entries <= 0) || (how_many_entries > TRACE_SIZE))
	how_many_entries = TRACE_SIZE;
    
    not_local = midway_uxpass_local();
    printf("-=-=- Trace Buffer (oldest entry first) -=-=-\n");
    cbuf[16] = '\0';

    mutex_lock(midTraceLock);
    if (midTraceIndex >= how_many_entries) {
	start = ((midTraceIndex - how_many_entries)/4) * 4;
	end = ((midTraceIndex+3)/4)*4;
	repeat = 0;
    } else {
	how_many_entries -= midTraceIndex;
	start = ((TRACE_SIZE - how_many_entries)/4) * 4;
	end = TRACE_SIZE;
	end2 = ((midTraceIndex + 3)/4)*4;
	repeat = 1;
    }
    mutex_unlock(midTraceLock);

restart:
    for (i = start; i < end; i += 4) {
	if ((midTraceData[i] != 0) || (midTraceData[i+1] != 0) ||
	    (midTraceData[i+2] != 0) ||(midTraceData[i+3] != 0)) {
	    
	    cP = (char *) &midTraceData[i];
	    for (j = 0; j < 16; j++) {
		if (isascii((int)(*cP)) && isprint((int)(*cP)))
		    cbuf[j] = *cP;
		else
		    cbuf[j] = '.';
		cP++;
	    }
	    printf("%08lu: %08x %08x %08x %08x |%s|\n",
		    i, midTraceData[i],midTraceData[i+1],
		    midTraceData[i+2],midTraceData[i+3], cbuf);
	}
    }
    if (repeat) {
	repeat = 0;
	start = 0;
	end = end2;
	goto restart;
    }
    printf("-=-=- Trace Buffer Complete -=-=-\n");
    fflush(stdout);
    if (not_local)
      midway_uxpass_remote();
#endif
}


void midway_print_memory_map (void)
{
  kern_return_t krc;
  task_t task_self;
  vm_offset_t addr, end_addr, last_addr;
  vm_size_t len, total, last_len;
  vm_prot_t protection;
  vm_prot_t max_protection;
  vm_inherit_t inheritance;
  boolean_t sharedp;
  mach_port_t object_name;
  vm_offset_t offset;

  addr = 0;
  total = 0;
  last_len = 0;
  last_addr = 0;
  task_self = mach_task_self();

#ifdef mips
  end_addr = 0x7FFFFFFF;
#else
  end_addr = 0xFFFFFFFF;
#endif

  while (TRUE) {

      len = (vm_size_t) ((unsigned int) end_addr - (unsigned int) addr);
      krc = vm_region(task_self, &addr, &len, &protection,
		      &max_protection, &inheritance, &sharedp, &object_name,
		      &offset);
      if (krc == KERN_NO_SPACE)
	  break;			/* done */
      if (krc != KERN_SUCCESS) {
	  printf("vm_region fails, addr 0x%08lx len %u krc %d\n",
		 addr, len, krc);
	  mach_error("vm_region",krc);
	  break;
      }

      /* Combine adjacent areas */
      if (addr == (last_addr+last_len)) {
	  last_len += len;
      } else {
      
	  if (last_len > 0)  {
	      printf("memory mapped at 0x%08lx to 0x%08lx for %lu KBytes\n",
		     last_addr, last_addr+last_len, last_len/1024);
	      total += last_len;
	  }
	  last_addr = addr;
	  last_len = len;
      }
      addr += len;
  }
  if (last_len > 0) {
      printf("memory mapped at 0x%08lx to 0x%08lx for %lu KBytes\n",
	     last_addr, last_addr+last_len, last_len/1024);
      total += last_len;
  }

  printf("total memory mapped is %lu KBytes\n", total/1024);
  
} /* midway_print_memory_map */

midway_malloc_print_stats ()
{
#if !defined(STATS)
  StatusMsg ("Not compiled to collect malloc statistics");
#else
  int i, len, slen;
  char255 smsg, tmp;
#define MAX_STR_LEN 80

  sprintf(smsg, "malloc %lu (%lu KB) calloc %lu (%lu KB) free %lu",
	  midStats.malloc_calls, midStats.malloc_size/1024,
	  midStats.calloc_calls, midStats.calloc_size/1024,
	  midStats.free_calls);
  StatusMsg (smsg);

  sprintf(smsg, "malloc size(2^)");
  slen = strlen(smsg);
  for (i = 0; i < STAT_TABLE_SIZE; i++) {
    if (midStats.malloc_size_count[i] == 0) continue;
    
    sprintf(tmp, " %d=%lu", i, midStats.malloc_size_count[i]);
    len = strlen(tmp);
    if ((len + slen) > MAX_STR_LEN) {
      StatusMsg(smsg);
      sprintf(smsg, "malloc size(2^)");
      slen = strlen(smsg);
    }
    strcat (smsg, tmp);
    slen += len;
  }
  StatusMsg(smsg);

  sprintf(smsg, "calloc size(2^)");
  slen = strlen(smsg);
  for (i = 0; i < STAT_TABLE_SIZE; i++) {
    if (midStats.calloc_size_count[i] == 0) continue;
    
    sprintf(tmp, " %d=%lu", i, midStats.calloc_size_count[i]);
    len = strlen(tmp);
    if ((len + slen) > MAX_STR_LEN) {
      StatusMsg(smsg);
      sprintf(smsg, "calloc size(2^)");
      slen = strlen(smsg);
    }
    strcat (smsg, tmp);
    slen += len;
  }
  StatusMsg(smsg);
#endif
}
