/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * the file a "user" would include for Midway
 *
 *  9 may 91, mjz
 * 21 jun 91(mjz): changed from pl_t to u_long for lock_t...
 *  1 jul 92(was): added support for porting code to midway 
 * 19 aug 92(was): removed earlier porting support code
 *  3 Oct 92 (mjz): change names
 *  6 Feb 93 (was): added midway_start/stop_profile & midway_exit calls
 *  6 Jan 94 (was): added macros for fastpath acquire/release locks (FASTLOCKS)
 * 20 Feb 94 (mjz): undo inlining spin_unlock: creates a race I can't fix.
 */
#ifndef __MIDWAY_H__
#define __MIDWAY_H__

#include <sys/types.h>
#include <mach.h>
#include <cthreads.h>


/*
 * Set compiler flags
 */
#if !defined(MIDWAY_FASTLOCKS)
#  define MIDWAY_FASTLOCKS  1		/* =1 to inline lock acquire/release */
					/* =0 to subr call for each acquire/release */
#  define MIDWAY_FASTLOCKS_CACHE  1	/* =1 to inline for LOCK_RO of cached (!owned) data */
					/* =0 to subr call for this acq/rel */
#endif 

#if defined(OMITSHARED) 		/* Undefine shared/priv attributes */
#  define shared
#  define priv
#endif



/*
 * Define external types for locks and barriers
 */
typedef u_long midway_lock_t;
typedef u_long midway_barrier_t;


/*
 * Define mode flag for lock acquire
 */
typedef enum {LOCK_RO = 1, LOCK_RW = 2} midway_acquire_t;




/*
 * Define macros & types for the FastLock code.
 */

/*
 * Define "held" values for midway_lock_t's
 * Note CACHED is currently not supported.
 */
typedef enum {
  MIDWAY_LOCKHELD_FREE = 0,		/* lock is owned locally & not held */
  MIDWAY_LOCKHELD_RO = LOCK_RO,		/* MUST be same (int) value as LOCK_RO */
  MIDWAY_LOCKHELD_RW = LOCK_RW,		/* MUST be same (int) value as LOCK_RW */
  MIDWAY_LOCKHELD_CACHE = 4,		/* lock is cached locally & not held */
  MIDWAY_LOCKHELD_CACHE_RO = 8,		/* lock is ccached locally & held RO */
  MIDWAY_LOCKHELD_OTHER = 16		/* lock not owned/cached locally */
} midway_lockheld_t;

/*
 * define MACRO BEGIN/END statements (from cthreads.h)
 */
#ifndef MACRO_BEGIN

#ifdef	lint
int	NEVER;
#else	lint
#define	NEVER 0
#endif	lint

#define	MACRO_BEGIN	do {
#define	MACRO_END	} while (NEVER)

#endif	MACRO_BEGIN



#if MIDWAY_FASTLOCKS > 0


/*
 * Define internal view of lock structure.
 * Note that this code assumes a pointer can be stored in a u_long.
 */
typedef priv struct {
  struct mutex  	mutex;
  midway_lockheld_t	held;
} *midway_reallock_t;


/*
 * Optimize lock acquire/release by inlining system dependent code.
 *
 * This doesn't work (mips), disable 20Feb94, mjz.  Here's the deal:
 * the compiler insists on inserting the load of the queue pointer before
 * the store to 0 of the spinlock.  This creates a data race: if the
 * current thread is interrupted between the load & the store, then
 * some other thread could queue, and we won't notice (having already
 * completed the load), store 0, and never release the waiting thread.
 * This seems to be true, in both gcc 1.XX/mw and gcc 2.3.2, even if
 * I carefuly craft "volatile" declarations.  This does do something;
 * the store of 0 is no longer in the branch delay slot for testing
 * if the queue pointer is 0.  However, it doesn't swap the store of 0
 * and the load of the queue pointer... which just seems to be broken.
 * [I seem to be able to define the store below to be volatile, but I
 * have never been able to define the load of the queue pointer to be
 * volatile, at least as represented by the assembler ourput of gcc.]
 * CAVEAT MAINTAINER.
 */
#if defined(mips) && (MIDWAY_FIXED_CTHREADS > 0)
#  define spin_unlock(x) (*(priv int *)(x)) = 0
#endif /* mips */

/*
 * Collect statistics on FastLock usage
 */
#if defined(MIDWAY_FASTLOCKS_STATS)

/* Layout of this structure must match definition in lock.c */
extern priv struct {
  unsigned long acqRO_count;
  unsigned long acqRW_count;
  unsigned long acqCA_count;
  unsigned long relRO_count;
  unsigned long relRW_count;
  unsigned long relCA_count;
} midFastLockStats;

#  define MIDWAY_INCR_ACQS(mode) 		\
    MACRO_BEGIN					\
    if (mode == LOCK_RO)			\
        midFastLockStats.acqRO_count += 1; 	\
    else if (mode == LOCK_RW)			\
        midFastLockStats.acqRW_count += 1;	\
    else                      			\
        midFastLockStats.acqCA_count += 1;	\
    MACRO_END
#  define MIDWAY_INCR_RELS(mode)		\
    MACRO_BEGIN					\
    if (mode == LOCK_RO)			\
        midFastLockStats.relRO_count += 1; 	\
    else if (mode == LOCK_RW)			\
        midFastLockStats.relRW_count += 1;	\
    else                      			\
        midFastLockStats.relCA_count += 1;	\
    MACRO_END
#else
#  define MIDWAY_INCR_ACQS(mode)
#  define MIDWAY_INCR_RELS(mode)
#endif /* MIDWAY_FASTLOCKS_STATS */


/*
 * Note: Currently the lock_release code optimizes RO over RW.
 * They could be equal if we defined macros 
 * midway_lock_release_read and midway_lock_release_write.
 * Likewise, if we support fast access to un-owned, cached locks
 * we could optimize them be defining macros
 * midway_lock_acquire_read and midway_lock_acquire_write.
 * WAS 9Jan94
 */
#endif /* MIDWAY_FASTLOCKS */


/*
 * Define Midway entry points
 */
/*								--LOCKS */

extern midway_lock_t midway_lock_alloc();  /* allocate and initialize lock */

#if MIDWAY_FASTLOCKS == 0
#define midway_lock_acquire(lockn,mode) midway_lock_acq(lockn,mode)
extern void midway_lock_acq (
			     midway_lock_t lockn, /* which SV */
			     midway_acquire_t mode /* reader or writer */
			     );
#else
/*
 * FASTLOCKS: Define acquire as a macro for common case.
 */
#if MIDWAY_FASTLOCKS_CACHE == 0

#define midway_lock_acquire(lockn,mode)					\
    MACRO_BEGIN								\
    register midway_reallock_t lockP = (midway_reallock_t) lockn;	\
    register midway_acquire_t t_mode; 					\
    mutex_lock ((mutex_t)&(lockP->mutex));				\
    t_mode = mode;  /* force mode into reg to fill if's delay slot */	\
    if (lockP->held == MIDWAY_LOCKHELD_FREE) {				\
        lockP->held = t_mode;						\
	mutex_unlock ((mutex_t)&(lockP->mutex));			\
	MIDWAY_INCR_ACQS(t_mode);					\
    } else {								\
        midway_lock_acqX (lockn, t_mode);				\
    }									\
    MACRO_END
#else

/*
 * These are somewhat convoluted since it seems to generate better code.
 * The compiler insists on combining the two cases so that they both
 * execute the same code for mutex_unlock. Unfortunately, it does this
 * by inserting a jump in the first case. By reversing the order the 
 * fastpath case avoids the jump and performs as well as the no-cache
 * code above.    WAS 13Jan94
 */
#define midway_lock_acquire(lockn,mode)					\
    MACRO_BEGIN								\
    register midway_reallock_t lockP = (midway_reallock_t) lockn;	\
    register midway_acquire_t t_mode;					\
    register midway_lockheld_t t_held;					\
    mutex_lock ((mutex_t)&(lockP->mutex));				\
    t_held = lockP->held;						\
    t_mode = mode;  /* force mode into reg to fill if's delay slot */	\
    if (t_held != MIDWAY_LOCKHELD_FREE) { 				\
       if ((t_mode == LOCK_RO) && 					\
	   (t_held == MIDWAY_LOCKHELD_CACHE)) {				\
           lockP->held = MIDWAY_LOCKHELD_CACHE_RO;			\
	   mutex_unlock ((mutex_t)&(lockP->mutex));			\
	   MIDWAY_INCR_ACQS(MIDWAY_LOCKHELD_CACHE_RO);			\
       } else {								\
           midway_lock_acqX (lockn, t_mode);				\
       }								\
    } else {								\
       lockP->held = t_mode;						\
       mutex_unlock ((mutex_t)&(lockP->mutex));				\
       MIDWAY_INCR_ACQS(t_mode);					\
    }									\
    MACRO_END
#define midway_lock_acquire_read(lockn)					\
    MACRO_BEGIN								\
    register midway_reallock_t lockP = (midway_reallock_t) lockn;	\
    register midway_acquire_t t_mode;					\
    register midway_lockheld_t t_held;					\
    mutex_lock ((mutex_t)&(lockP->mutex));				\
    t_held = lockP->held;						\
    t_mode = LOCK_RO; /* force mode into reg to fill if's delay slot */	\
    if (t_held != MIDWAY_LOCKHELD_FREE) {				\
        if (t_held == MIDWAY_LOCKHELD_CACHE) {				\
            lockP->held = MIDWAY_LOCKHELD_CACHE_RO;			\
	    mutex_unlock ((mutex_t)&(lockP->mutex));			\
	    MIDWAY_INCR_ACQS(MIDWAY_LOCKHELD_CACHE_RO);			\
        } else {							\
            midway_lock_acqX (lockn, t_mode);				\
	}								\
    } else {								\
        lockP->held = t_mode;						\
	mutex_unlock ((mutex_t)&(lockP->mutex));			\
	MIDWAY_INCR_ACQS(t_mode);					\
    }									\
    MACRO_END
#define midway_lock_acquire_write(lockn)				\
    MACRO_BEGIN								\
    register midway_reallock_t lockP = (midway_reallock_t) lockn;	\
    register midway_acquire_t t_mode;					\
    mutex_lock ((mutex_t)&(lockP->mutex));				\
    t_mode = LOCK_RW; /* force mode into reg to fill if's delay slot */ \
    if (lockP->held == MIDWAY_LOCKHELD_FREE) {				\
        lockP->held = t_mode;						\
	mutex_unlock ((mutex_t)&(lockP->mutex));			\
	MIDWAY_INCR_ACQS(t_mode);					\
    } else {								\
        midway_lock_acqX (lockn, t_mode);				\
    }									\
    MACRO_END


#endif /* MIDWAY_FASTLOCKS_CACHE */

#endif /* MIDWAY_FASTLOCKS */

/* needs to be outside the MIDWAY_FASTLOCKS defn to typecheck the library. */ 
extern void midway_lock_acqX(
			    midway_lock_t lockn, /* which SV */
			    midway_acquire_t mode /* reader or writer */
			    );


#if MIDWAY_FASTLOCKS == 0
#define midway_lock_release(lockn) midway_lock_rel(lockn)
extern void midway_lock_rel (
			     midway_lock_t lockn /* which SV */
			     );
#else
/*
 * FASTLOCKS: Define release as a macro for common case.
 */
#if MIDWAY_FASTLOCKS_CACHE == 0

#define midway_lock_release(lockn)					\
    MACRO_BEGIN								\
    register midway_reallock_t lockP = (midway_reallock_t) lockn;	\
    register midway_lockheld_t t_held;					\
    mutex_lock ((mutex_t)&(lockP->mutex));				\
    t_held = lockP->held;						\
    if ((t_held == LOCK_RO) || (t_held == LOCK_RW)) { 			\
        lockP->held = MIDWAY_LOCKHELD_FREE;				\
        mutex_unlock ((mutex_t)&(lockP->mutex));			\
	MIDWAY_INCR_RELS(t_held);					\
    } else {								\
	midway_lock_relX (lockn);					\
    }									\
    MACRO_END
#else 

#define midway_lock_release(lockn)					\
    MACRO_BEGIN								\
    register midway_reallock_t lockP = (midway_reallock_t) lockn;	\
    register midway_lockheld_t t_held;					\
    mutex_lock ((mutex_t)&(lockP->mutex));				\
    t_held = lockP->held;						\
    if ((t_held == LOCK_RO) || (t_held == LOCK_RW)) { 			\
        lockP->held = MIDWAY_LOCKHELD_FREE;				\
        mutex_unlock ((mutex_t)&(lockP->mutex));			\
	MIDWAY_INCR_RELS(t_held);					\
    } else if (t_held == MIDWAY_LOCKHELD_CACHE_RO) {			\
        lockP->held = MIDWAY_LOCKHELD_CACHE;				\
        mutex_unlock ((mutex_t)&(lockP->mutex));			\
	MIDWAY_INCR_RELS(t_held);					\
    } else {								\
	midway_lock_relX (lockn);					\
    }									\
    MACRO_END

#define midway_lock_release_read(lockn)					\
    MACRO_BEGIN								\
    register midway_reallock_t lockP = (midway_reallock_t) lockn;	\
    register midway_lockheld_t t_held;					\
    mutex_lock ((mutex_t)&(lockP->mutex));				\
    t_held = lockP->held;						\
    if (t_held != LOCK_RO) {				 		\
        if (t_held == MIDWAY_LOCKHELD_CACHE_RO) {			\
            lockP->held = MIDWAY_LOCKHELD_CACHE;			\
            mutex_unlock ((mutex_t)&(lockP->mutex));			\
	    MIDWAY_INCR_RELS(t_held);					\
        } else {							\
	    midway_lock_relX (lockn);					\
        }								\
    } else {								\
        lockP->held = MIDWAY_LOCKHELD_FREE;				\
        mutex_unlock ((mutex_t)&(lockP->mutex));			\
	MIDWAY_INCR_RELS(t_held);					\
    }									\
    MACRO_END

#define midway_lock_release_write(lockn)				\
    MACRO_BEGIN								\
    register midway_reallock_t lockP = (midway_reallock_t) lockn;	\
    register midway_lockheld_t t_held;					\
    mutex_lock ((mutex_t)&(lockP->mutex));				\
    t_held = lockP->held;						\
    if (t_held == LOCK_RW) {					 	\
        lockP->held = MIDWAY_LOCKHELD_FREE;				\
        mutex_unlock ((mutex_t)&(lockP->mutex));			\
	MIDWAY_INCR_RELS(t_held);					\
    } else {								\
	midway_lock_relX (lockn);					\
    }									\
    MACRO_END

#endif /* MIDWAY_FASTLOCKS_CACHE */

#endif /* MIDWAY_FASTLOCKS */

/* needs to be outside the MIDWAY_FASTLOCKS defn to typecheck the library. */ 
extern void midway_lock_relX (
			      midway_lock_t lockn /* which SV */
			      );


/*								--BARRIERS */

extern midway_barrier_t midway_barrier_alloc(); /* allocate and init barrier */

extern void midway_cross_barrier( /* send changes to all nodes */
				 midway_barrier_t bname, /* which one */
				 int num	/* how many expected */
				 );

extern void midway_flush_barrier( /* send changes to master only */
				 midway_barrier_t bname, /* which one */
				 int num	/* how many expected */
				 );

/*								--ASSOCIATION*/
extern void midway_bind_synch_object(
				     midway_lock_t obj,	/* TYPE?  Wish could do e/o */
				     shared void *addr, /* start of guarded data */
				     u_long len	 /* length of guarded data */
				     );

extern void midway_rebind_synch_object(
				     midway_lock_t obj	/* TYPE?  Wish could do e/o */
				     );

/*								--THREADS */
extern void midway_create_thread(
				 void(*func)(), /* function to be performed */
				 u_long arg     /* argument to pass */
				 );

extern void midway_wait_for_end(
				u_long n /* # of threads to wait for IGNORED */
				);

/*							--DYNAMIC MEM ALLOC */
extern shared char *shmalloc(
		      unsigned n,	/* how many bytes do you want? */
		      unsigned linesz	/* cache line size, 4-2**32 */
		      );

extern void shfree(
		  shared char *mem	/* pointer to shmalloc'd storage to free */
		  );


extern int midway_start_profile (void);

extern int midway_stop_profile (void);

extern void midway_exit(
			int rc	/* application return code */
			);


#endif /* __MIDWAY_H__ */
