/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * midway_types.h:  Internal type definitions used throughout midway.
 *
 * HISTORY
 * 11 Aug 92 (was): Initial version.
 * 15 Aug 92 (was): Added defines for consistency model & addr_vector_t
 * 15 Dec 92 (was): Added LINE_NOT_ALLOCATED flag
 */

#if !defined(__MIDWAY_TYPES_H__)
#define __MIDWAY_TYPES_H__


#include "bitwiddle.h"          /* get bitarray definitions */


/*
 * Define networks used by Midway
 */
#define MIDWAY_NET_SOCKETS	0 /* Use BSD Sockets (and ethernet) */
#define MIDWAY_NET_ATM		1 /* use "nw" ATM interface for data */
extern int midway_comm_network_type; /* where to find the value */



/*
 * Define consistency models supported by midway 
 */
#define MODEL_EC    0           /* Entry consistency */
#define MODEL_RC    1           /* Release consistency */
#define MODEL_PC    2           /* Processor consistency */

#define DEFAULT_FLUSH_PERIOD 100  /* milliseconds (for PC only) */


/*
 * Define maximum number of processors supported.
 * This is only used for bit array address vectors.
 * If you change the value, you must also update the macros below.
 */
#define MAX_PROCESSORS  32


/*
 * Define bit array address vector
 */
#define ADDR_VECTOR_WORDS (midBITARRAYsize (MAX_PROCESSORS) / sizeof(BITARRAY))
typedef struct {
  BITARRAY bitArray[ ADDR_VECTOR_WORDS ];
} addr_vector_t;

#define NULL_ADDR_VECTOR  ((addr_vector_t){0})



/*
 * Define macros for quick operations on address vectors.
 * macros depend on definition of MAX_PROCESSORS and sizeof BITARRAY
 */
#if MAX_PROCESSORS == 32

#define EQUAL_ADDR_VECTOR(v1,v2) \
    ((v1).bitArray[0] == (v2).bitArray[0])

#define EQUAL_ZERO_ADDR_VECTOR(v1)   \
    ((v1).bitArray[0] == (BITARRAY) 0)

#define MERGE_ADDR_VECTOR(v1,v2) \
    (v1).bitArray[0] |= (v2).bitArray[0]

#define DIFF_ADDR_VECTOR(v1,v2) \
    (v1).bitArray[0] ^= (v2).bitArray[0] /* bitwise XOR copysets to find differences  */

#define ZERO_ADDR_VECTOR(v1) \
    (v1).bitArray[0] = 0


#elif MAX_PROCESSORS == 64

#define EQUAL_ADDR_VECTOR(v1,v2)                 \
    (((v1).bitArray[0] == (v2).bitArray[0]) &&   \
     ((v1).bitArray[1] == (v2).bitArray[1]))

#define EQUAL_ZERO_ADDR_VECTOR(v1)                \
    (((v1).bitArray[0] == (BITARRAY) 0) &&        \
     ((v1).bitArray[1] == (BITARRAY) 0))

#define MERGE_ADDR_VECTOR(v1,v2)                  \
{                                                 \
    (v1).bitArray[0] |= (v2).bitArray[0];         \
    (v1).bitArray[1] |= (v2).bitArray[1];         \
}

#define DIFF_ADDR_VECTOR(v1,v2)			  \
{                                                 \
    (v1).bitArray[0] ^= (v2).bitArray[0];         \
    (v1).bitArray[1] ^= (v2).bitArray[1];         \
}

#define ZERO_ADDR_VECTOR(v1) \
{                            \
    (v1).bitArray[0] = 0;    \
    (v1).bitArray[1] = 0;    \
}


#else

  /* Handle arbitrary size address vectors */
#define EQUAL_ADDR_VECTOR(v1,v2)                 \
    (memcmp ((v1).bitArray, (v2).bitArray, sizeof(v1)) == 0)

#define EQUAL_ZERO_ADDR_VECTOR(v1)                \
  (memcmp(&(v1), &NULL_ADDR_VECTOR, sizeof(v1)) == 0)
  /* Note: NULL_ADDR_VECTOR needs to be defined */

#define MERGE_ADDR_VECTOR(v1,v2)                                \
{ int i;                                                        \
  for (i = 0; i < sizeof(addr_vector_t); i += sizeof(BITARRAY)) \
    (v1).bitArray[i] |= (v2).bitArray[i];                       \
}

#define DIFF_ADDR_VECTOR(v1,v2)                                 \
{ int i;                                                        \
  for (i = 0; i < sizeof(addr_vector_t); i += sizeof(BITARRAY)) \
    (v1).bitArray[i] ^= (v2).bitArray[i];                       \
}

#define ZERO_ADDR_VECTOR(v1) \
    bzero (&(v1), sizeof(addr_vector_t))
  

#endif  /* MAX_PROCESSOR = unknown size */





/*
 * Define types/macros for lamport clocks
 */
typedef unsigned long clock_t;

#define MAX_CLOCK_VALUE  ((clock_t) 0x7FFFFFFF)   /* 31 bit max value */
#define MIN_CLOCK_VALUE  ((clock_t) 0x00000001)
#define NULL_CLOCK_VALUE ((clock_t) 0x0)


#ifdef DEBUG

/* Note: Don't allow clock values to wrap around. WAS 11Aug92 */
#define ADJUST_CLOCK(c,by) \
  ((((c)+(by)) > MAX_CLOCK_VALUE) ? ((clock_t) exit(100)) : ((c)+(by)))

#else /* not DEBUG */
#define ADJUST_CLOCK(c,by) ((c)+(by))
#endif


#define LESS_THAN_CLOCK(c1,c2) ((c1) < (c2))
#define LESS_THAN_OR_EQUAL_CLOCK(c1,c2) ((c1) <= (c2))
#define GREATER_THAN_CLOCK(c1,c2) ((c1) > (c2))
#define GREATER_THAN_OR_EQUAL_CLOCK(c1,c2) ((c1) >= (c2))


/*
 * Define layout of each cache line's dirty bit information.
 * Info resides in a clock_t structure
 * When using EC, the line is dirty when 0 
 *                and non-dirty when it contains a clock value.
 * When not using EC, the info is defined by the following bit patterns
 */
#define LINE_EXCEPTION_FLAG   0x80000000   /* hi order bit is 1 if line is 
					    * not on the fast path (cache-
					    * aligned EC */
#define LINE_ENQUEUE_FLAG     0x40000000   /* =1 if not EC & updates must
                                            * be added to change queue */
#define LINE_UNALIGNED_FLAG   0x20000000   /* =1 if EC & cache is objects do
					    * not align on cache boundaries */
#define LINE_CONVERSION_FLAG  0x10000000   /* =1 if model conversion tools
                                            * are collecting reference data */

#define LINE_DATA_VALUE       0x0FFFFFFF   /* Non-flag values within line info */

#define LINE_FLAG_VALUE       0xF0000000   /* Flag values within line info */

#define LINE_NOT_ALLOCATED    0x8FFFFFFF   /* this memory is not allocated
					    * used for debugging apps only */

/* define macro for checking dense dirty bits */
#define LINE_UNALIGNED(dbit) \
  ((dbit & (LINE_EXCEPTION_FLAG|LINE_UNALIGNED_FLAG)) == (LINE_EXCEPTION_FLAG|LINE_UNALIGNED_FLAG))
  


/*
 * Define macros for smashing deallocated memory
 * Once debugged these can be redefined as noops 
 */
#define SMASH_VALUE 0xDDDDDDDD
#ifdef DEBUG
#define SMASH_WORD(_x) _x = SMASH_VALUE
#define SMASH_AREA(_x,_s) (void) memset (_x, SMASH_VALUE, _s)
#else /* not DEBUG */
#define SMASH_WORD(_x)
#define SMASH_AREA(_x,_s)
#endif


#ifndef NULL
#define NULL 0
#endif

#endif /* __MIDWAY_TYPES_H__ */





