/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * pagemgr.c: Routines for managing page tables
 * 
 *
 * 15 Aug 92 (was): Initial code 
 * 20 Oct 92 (was): added code to fragment data on page replies
 *                  Also added stats
 * 29 Oct 92 (was): Puts statistics under STATS instead of DEBUG
 *  1 Nov 92 (was): Added code to check if page is entirely EC and
 *                  can be zero filled without contacting page owner.
 * 13 Jan 93 (was): use macro to check if update queue is empty
 * 19 Jan 93 (was): optimized mid_page_copyset
 */
#include <sys/types.h>		/* u_long & friends */
#include <mach.h>
#include <cthreads.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>

#include "lock.h"		/* perlock struct, lock request formats */
#include "comm.h"		/* M_xxx_OP */
#include "smexecd_types.h"	/* ensure get char255 */
#include "pbuf.h"		/* packet buffer stuff */
#include "shmalloc.h"		/* for FIRST/LAST_STATIC_SYM */
#include "midway-dbg.h"		/* filter debug messages */
#include "updateq.h"            /* prototypes for update queue routines */
#include "vm.h"                 /* ADDR_TO_PAGENO */

#define __PAGEMGR_INTERNAL__    
#include "pagemgr.h"		/* generate internal definitions  */


/*
 * Define external variables
 */
extern int Comm_index;          /* this processors "number" (1-n) */
extern task_t midway_task_self;	/* mach_task_self() */


/*
 * Define Static internal variables
 */
static page_fragment_t *pageFragListP;      /* list of page fragments */
static page_fragment_t *pageFragFreeListP;  /* free list of above */
static mutex_t pageFragMutex;               /* mutex guarding lists */



#ifdef STATS 

/*
 * Define Static variables for collecting statistics
 */
static struct {
  u_long Faults;
  u_long FaultsOnInvalidRegion;
  u_long FaultsOnAllocatedPage;
  u_long FaultsZeroFilled;
  u_long FaultsRemote;
  u_long FaultsDataSent;
  u_long FaultsFailed;
  u_long Requests;
  u_long RequestsInvalid;
  u_long RequestsZeroFilled;
  u_long RequestsUpdateCopyset;
  u_long RequestsCopysetDataSent;
  u_long RequestsDataFilled;
  u_long RequestsDataPackets;
  u_long RequestsDataSent;
  u_long Replies;
  u_long RepliesInvalid;
  u_long RepliesDuplicates;
  u_long RepliesUnrequested;
  u_long RepliesZeroFilled;
  u_long RepliesAllocated;
  u_long RepliesDuplicatePackets;
  u_long RepliesPendingUpdates;
  u_long RepliesErrors;
  u_long CopysetRequests;
  u_long CopysetRequestsInvalid;
  u_long CopysetRequestsNacked;
  u_long CopysetRequestsDataSent;
  u_long CopysetReplies;
  u_long CopysetRepliesInvalid;
  u_long CopysetRepliesNacked;
  u_long CopysetRepliesNackInvalid;
} pageStat;
static mutex_t pageStatLock;

# ifdef FAST_STATS
#  define INCR_STATS(x,by) (x) += (by)
# else 
#  define INCR_STATS(x,by)         \
    { mutex_lock(pageStatLock); (x) += (by); mutex_unlock(pageStatLock); }
# endif 

#else
#define INCR_STATS(x,by)	/* nothing to do */
#endif



#define VERIFY_FRAGMENT_LIST 0   /* =1 to verify page fragment list */


#if VERIFY_FRAGMENT_LIST > 0

#define VERIFY_FRAGS() VerifyFragList ()

static void VerifyFragList ()
{
  int count, fcount;
  page_fragment_t *fragP, *freeP;

  /* NOTE: Must be holding pageFragMutex to call this routine */

  if (pageFragListP != NULL) {

    if (pageFragListP == pageFragFreeListP) {
      botch ("VerifyFragList (fragListP == freeListP)");
    }

    for (fragP = pageFragListP; fragP != NULL; fragP = fragP->nextP) {
      count++;
      if (count > 1000)
	botch("VerifyFragList (pageFragListP looping)");
      if (fragP == fragP->nextP)
	botch("VerifyFragList (pageFragListP element pts to self)");
    
      fcount = 0;
      for (freeP = pageFragFreeListP; freeP != NULL; freeP = freeP->nextP) {
	if (fcount > 1000)
	  botch("VerifyFragList (pageFragFreeListP looping)");
	if (freeP == freeP->nextP)
	  botch("VerifyFragList (pageFragFreeListP element pts to self)");
	if (fragP == freeP)
	  botch ("VerifyFragList (element on both lists)");
      }
    }
  } else {
    fcount = 0;
    for (freeP = pageFragFreeListP; freeP != NULL; freeP = freeP->nextP) {
      if (fcount > 1000)
	botch("VerifyFragList (pageFragFreeListP looping)");
      if (freeP == freeP->nextP)
	botch("VerifyFragList (pageFragFreeListP element pts to self)");
    }
  }
}
#else

#define VERIFY_FRAGS()

#endif



/*
 ************************************************************************
 *
 * sprintf_addr_vector
 * printf_addr_vector
 *
 * Internal routines to print address vectors
 * For debugging only.
 *
 ************************************************************************
 */
char *
sprintf_addr_vector 
  (
   char255        bufP,
   addr_vector_t  addr
   )
{
  int i;
  char255 temp;

  strcpy (bufP, "");
  for (i = 0; i < ADDR_VECTOR_WORDS; i ++) {
    sprintf(temp, "%s%08x", (i == 0) ? "" : " ", addr.bitArray[i]);
    strcat (bufP, temp);
  }
  return (bufP);
}

/*
 ************************************************************************
 *
 * mid_page_init:
 *
 * Initialize the page manager for non-EC support
 * 
 ************************************************************************
 */
int 
mid_page_init ()
{
  if (IS_DBG_PORT_PAGE) {
      char255 smsg;
      sprintf(smsg,
	      "mid_page_init called, page_size=%d page_shift=%d",
	      vm_page_size, vm_page_shift);
      StatusMsg(smsg);
  }

  ASSERT (ADDR_TO_PAGENO(vm_page_size) == 1);
  ASSERT (PAGENO_TO_ADDR(1) == vm_page_size);

  pageFragMutex = mutex_alloc();
  MIDWAY_MUTEX_SET_NAME(pageFragMutex,"pagemgr.c:pageFragMutex");
  pageFragListP = NULL;
  pageFragFreeListP = NULL;

#ifdef STATS
  pageStatLock = mutex_alloc();
  MIDWAY_MUTEX_SET_NAME(pageStatLock, "pagemgr.c:pageStatLock");
  bzero (&pageStat, sizeof(pageStat));
#endif

  return (0);
}


/*
 ************************************************************************
 *
 * mid_page_status:
 *
 * Return status of page (allocated, pending allocation, free)
 * 
 ************************************************************************
 */
int 
mid_page_status
  (
   vm_address_t addr
   )
{
  int status;
  u_long offset, pageNo;
  region_info riP;
  addr_vector_t *pageEntryP;

  /* Initialize return code */
  status = PAGE_STATUS_FREE;

  /* locate region info structure for this page */
  riP = region_from_addr (addr);
  
  /* Check if page table has been allocated for this region */
  if (riP->page_tableP != NULL) {
    
    /* locate page table entry for this page */
    offset = (u_long) ((char *)addr - (char *)riP->base);
    pageNo = ADDR_TO_PAGENO(offset);
    pageEntryP = &riP->page_tableP[ pageNo ];

    /* Page is allocated if "our" bit is ON in the page's copyset */
    if (midTestBit (pageEntryP->bitArray, Comm_index))
      status = PAGE_STATUS_ALLOCATED;

    /* Check if an allocation request is pending */
    else if (midTestBit (pageEntryP->bitArray, riP->owner))
      status = PAGE_STATUS_PENDING;
  }


done:

  if (IS_DBG_PORT_PAGE) {
    char255 smsg;
    sprintf(smsg, "page at 0x%lx is %s", 
	    addr, (status == PAGE_STATUS_FREE) ? "FREE" :
	    (status == PAGE_STATUS_PENDING) ? "PENDING" :
	    (status == PAGE_STATUS_ALLOCATED) ? "ALLOCATED" : "INVALID");
    StatusMsg (smsg);
  }
  return (status);
} /* mid_page_status */


/*
 ************************************************************************
 *
 * mid_page_copyset
 *
 * Return copyset for modifications to the given address. 
 *
 * NOTE: This routine also resets the dirty bit information
 * for each cache-line effected by the modification.
 * 
 ************************************************************************
 */
void
mid_page_copyset
  (
   vm_address_t  *dataAddrP,     /* ptr to changed address */
   u_long        *dataLenP, 
   addr_vector_t *copySetP
   )
{
  int rtndLen;
  int firstTime;
  long origLen;
  u_long len, pageNo;
  u_long lineOffset, linesz;
  u_long pageSize;
  u_long line_size;
  clock_t *line_infoP, info_value;
  clock_t *lineP, *startLineP, *endLineP, *endLineLoopP;
  vm_address_t addr, lineAddr;
  vm_offset_t region_base;
  region_info riP;
  addr_vector_t *pageTableP;
  addr_vector_t *pageEntryP, *startPageP, *endPageP;


  if (IS_DBG_PORT_PAGE) {
    char255 smsg;
    sprintf(smsg, "page_copyset (addr=0x%x, len=%lu)",
	    *dataAddrP, *dataLenP);
    StatusMsg (smsg);
  }

  /* Initilize returned copyset */
  ZERO_ADDR_VECTOR (*copySetP);

  /* locate region info structure for this page */
  addr = *dataAddrP;
  riP = region_from_addr (addr);

  /* 
   * Loop until all modified data has been included in copyset
   */
  origLen = (long) *dataLenP;
  rtndLen = 0;
  firstTime = TRUE;
  pageSize = vm_page_size;
  info_value = (LINE_EXCEPTION_FLAG | LINE_ENQUEUE_FLAG);

  while (origLen > 0) {

    /* cache info out of region table */
    line_infoP = riP->line_info;
    line_size = riP->line_size;
    linesz = RI_LINE_SIZE_TO_BYTES(line_size);
    pageTableP = riP->page_tableP;
    region_base = riP->base;

    /* Verify that a page table has been allocated for this region */
    ASSERT (pageTableP != NULL);
      
    /* round requested address down to start on a cache-line boundary */
    lineAddr = (addr & ~(linesz - 1));
    lineOffset = addr - lineAddr;

    /* return starting address, first time through loop only */
    if (firstTime) {
      *dataAddrP = lineAddr;
      firstTime = FALSE;
    }
    
    /* round up length to a cache-line multiple */
    len = origLen;
    if (len > (linesz - lineOffset))    /* check for cache-line crossing */
      len += linesz;
    origLen = (len + (linesz - 1)) & ~(linesz - 1);
      
    /* Check for crossing a region boundary */
    lineOffset = (u_long) ((char *)lineAddr - (char *)region_base);
    if ((lineOffset + origLen) > REGION_SIZE) /* check for end of region */
      len = REGION_SIZE - lineOffset;
    else
      len = origLen;

    /* 
     * Build copyset from each page effected by the modification 
     */
    startPageP = &pageTableP[ ADDR_TO_PAGENO(lineOffset) ];
    endPageP = &pageTableP[ ADDR_TO_PAGENO(lineOffset + (len-1)) ];
    for (pageEntryP = startPageP; pageEntryP <= endPageP; pageEntryP++) {

	/* Copy page table entry to returned copy set */
	MERGE_ADDR_VECTOR (*copySetP, *pageEntryP);
    }

    /* 
     * Reset dirty bit line info for each line modified 
     */
    startLineP = &line_infoP[ lineOffset >> line_size ];
    endLineP = &line_infoP[ (lineOffset + (len-1)) >> line_size ];
    endLineLoopP = endLineP - 8;
    for (lineP = startLineP; lineP <= endLineLoopP; lineP += 8) {
      lineP[ 0 ] = info_value;  lineP[ 1 ] = info_value;
      lineP[ 2 ] = info_value;  lineP[ 3 ] = info_value;
      lineP[ 4 ] = info_value;  lineP[ 5 ] = info_value;
      lineP[ 6 ] = info_value;  lineP[ 7 ] = info_value;
    }
    while (lineP <= endLineP) {
      lineP[0] = info_value;
      lineP++;
    }

    /* advance to the next region */
    riP++;
    rtndLen += len;
    origLen -= len;
    addr    += len;

  } /* while (origLen > 0) */

  /* return modified length rounded up to a cache line multiple */
  *dataLenP = rtndLen;


  if (IS_DBG_PORT_PAGE) {
    char255 smsg, s1;
    sprintf(smsg, "page_copyset complete (addr=0x%x, len=%lu, copyset=0x%s)",
	    *dataAddrP, *dataLenP, 
	    sprintf_addr_vector (s1, *copySetP));
    StatusMsg (smsg);
  }

  /* return accumulated copyset */
  return;
}


/*
 ************************************************************************
 *
 * page_is_ec
 *
 * Internal routine to check the line_info bits for the given page
 * to determine if the entire page is composed of ec memory.
 *
 ************************************************************************
 */
static int
page_is_ec (region_info riP, vm_address_t pageAddr)
{
  register int i;
  u_long byte_offset;
  u_long line_offset;
  register u_long line_size_in_bytes;
  register u_long page_size;
  register clock_t *lineP, flags;

  byte_offset = (u_long)((char *) pageAddr - (char *) riP->base);
  line_offset = byte_offset >> riP->line_size;
  line_size_in_bytes = RI_LINE_SIZE_TO_BYTES(riP->line_size);
  lineP = &riP->line_info[ line_offset ];
  page_size = vm_page_size;
  flags = LINE_EXCEPTION_FLAG | LINE_ENQUEUE_FLAG;

  /* loop to check each cache line on the page */
  for (i = 0; i < page_size; i += line_size_in_bytes) {
    if ((*lineP & flags) == flags) return (0);  /* page not ec */
    lineP++;
  }

  /* page is ec */
  return (1);
}


/*
 ************************************************************************
 *
 * mid_page_fault
 *
 * Handle a page fault exception for non-EC models.
 * The contents of the page must be obtained from the owner of 
 * the page's region. If we own the region, then simply provide
 * a zero filled page.
 *
 ************************************************************************
 */
kern_return_t
mid_page_fault 
  (
   vm_address_t faultAddr,
   pointer_t	*bufP
   )
{
  region_info riP;
  kern_return_t krc;
  u_long pageNo;
  vm_address_t pageAddr;
  addr_vector_t *pageEntryP;
  m_page_request_t req;

  if (IS_DBG_PORT_PAGE) {
    char255 smsg;
    sprintf(smsg, "PageFault: addr=0x%lx, bufP=0x%x", faultAddr, bufP);
    StatusMsg (smsg);
  }

  INCR_STATS (pageStat.Faults, 1);

  /* 
   * Check for obvious programmer errors
   */
  if (faultAddr < (vm_address_t) vm_page_size) {
    char255 smsg;
    sprintf(smsg, "PAGE FAULT IN LOW MEMORY, ADDR=0x%lx", faultAddr);
    ASSERT (FALSE);
  }

  /* locate region info for this address */
  riP = region_from_addr (faultAddr);

  /* Check if the region is allocated */
  if (riP->page_tableP == NULL) {

    /* region has not been allocated in this machine 
     * Get the allocation information for this region.
     */
    krc = get_shmalloc_info (faultAddr);
    if (krc != KERN_SUCCESS) {
      INCR_STATS (pageStat.FaultsOnInvalidRegion, 1);
      return(krc);
    }
  }  /* if region not allocated */


  /* get the page table entry for this page */
  { u_long byte_offset;
    byte_offset = (u_long)((char *) faultAddr - (char *) riP->base);
    pageNo = ADDR_TO_PAGENO(byte_offset);
    pageEntryP = &riP->page_tableP[ pageNo ];
    pageAddr = (vm_address_t) (PAGENO_TO_ADDR(pageNo) + (u_long)riP->base);
  }

  /* region has been allocated
   * Check if we own the region 
   */
  if (riP->owner == Comm_index) {
    
    /* Check that the page is not already allocated */
    if (midTestBit (pageEntryP->bitArray, Comm_index)) {

      { char255 t;
	fprintf(stderr, 
		"Page already allocated, Addr=0x%x Page=0x%x Size=%d (0x%x)\n",
		faultAddr, pageAddr, vm_page_size, vm_page_size);
	fprintf(stderr, 
		"riP=0x%x base=0x%x pageP=0x%x pageNo=%lu Entry 0x%x=%s Index=%d\n",
		riP, riP->base, riP->page_tableP, pageNo, 
		 pageEntryP, sprintf_addr_vector(t, *pageEntryP), Comm_index);
	fflush (stderr);
      }
      INCR_STATS (pageStat.FaultsOnAllocatedPage, 1);
    } else {

      /* verify that the page entry is not set */
      ASSERT (EQUAL_ZERO_ADDR_VECTOR (*pageEntryP));

      /* Allocate the (zero filled) page */
      *bufP = NULL;		/* pager will allocate zfilled page */
	
      /* mark page as allocated */
      midSetBit (pageEntryP->bitArray, Comm_index);

      INCR_STATS (pageStat.FaultsZeroFilled, 1);
    }
  } else {

    /* We don't own the page, 
     * but if the entire page is EC we can go ahead and zero fill it.
     */
    if (page_is_ec (riP, pageAddr)) {

      if (IS_DBG_PORT_PAGE) {
	char255 smsg;
	sprintf(smsg, "PageFault: ZeroFill EC page at 0x%lx", pageAddr);
	StatusMsg (smsg);
      }

      /* Allocate the (zero filled) page */
      *bufP = NULL;		/* pager will allocate zfilled page */
	
      /* mark page as allocated */
      midSetBit (pageEntryP->bitArray, Comm_index);

      INCR_STATS (pageStat.FaultsZeroFilled, 1);
    } else {

      /* We don't own the page & the page is not EC. 
       * Page contents must be obtained from region's owner.
       * Set flag to indicate the request is pending.
       *  (Set owner's bit on, even though our bit (for allocation) is not.
       * This also adds us to the copy set for the page.
       */
      midSetBit (pageEntryP->bitArray, riP->owner);
      
      /* Issue request to the owner */
      req.pageSize = vm_page_size;
      req.pagerbuf = bufP;	/* so reply knows what to fill in */
      CommRequest((unsigned long) pageAddr, /* use address as "lock"  */
		  M_PAGE_DATA,    /* type */
		  riP->owner,	/* dest */
		  &req, 		/* message */
		  sizeof(req));	/* message length */
      
      INCR_STATS (pageStat.FaultsRemote, 1);
      INCR_STATS (pageStat.FaultsDataSent, sizeof(req));
    }
  }

  /* 
   * Check if page was successfully allocated 
   */
  if (midTestBit (pageEntryP->bitArray, Comm_index))
    krc = KERN_SUCCESS;
  else {
    krc = KERN_FAILURE;
    INCR_STATS (pageStat.FaultsFailed, 1);
  }

  if (IS_DBG_PORT_PAGE) {
    char255 smsg, s1;
    sprintf(smsg, 
	    "page_fault: complete, RC=%d, pageEntry=0x%s bufP=%x *(%x)\n",
	    krc, sprintf_addr_vector (s1,*pageEntryP), 
	    bufP, *bufP);
    StatusMsg (smsg);
  }

  return (krc);
}

/*
 ************************************************************************
 *
 * mid_page_request
 *
 * Handle a request from another processor for the contents of a page
 * The requestor is added to the copyset for the requested page
 * and all other members of the copyset are notified of the change.
 *
 ************************************************************************
 */
void
mid_page_request
  (
   u_long  reqAddr,             /* address of requested page */
   u_long  replyHost,	        /* host to reply to */
   u_long  replyThread,	        /* thread on host to reply to */
   u_long  replyTid,	        /* the transaction id */
   u_long *bufP,                /* ptr to message buffer */
   u_long  bufLen               /* length of message */
   )
{
  int rc;
  region_info riP;
  addr_vector_t  destAddr;
  addr_vector_t *pageEntryP;
  kern_return_t krc;
  pbuf replyBufP;
  u_long requestLen, replyLen;
  m_page_request_t *requestP;
  m_page_reply_t *replyP;
  m_copyset_request_t copysetReq;
  u_long pageSize;
  u_long packetNum;
  u_long lastPacket;
  u_long pageOffset;
  u_long bufferSize;
  u_long bufferMax;
  u_short totalPackets;
  char *bufferP;
  char *pageP;

  INCR_STATS (pageStat.Requests, 1);
  
  /* get requested data length from the request buffer */
  requestP = (m_page_request_t *) bufP;
  if (bufLen < sizeof(*requestP))
    requestLen = 0;
  else
    requestLen = requestP->pageSize;

  if (IS_DBG_PORT_PAGE) {
    char255 smsg;
    sprintf(smsg, 
	    "page_request (host=%lu thd=%lu tid=%lu), ADDR=0x%x LEN=%d",
	    replyHost, replyThread, replyTid, reqAddr, requestLen);
    StatusMsg (smsg);
  }

  /* cache page size */
  pageSize = vm_page_size;

  /* alloc & initialize buffer for reply */
  replyBufP = mid_pbuf_get();
  ASSERT (replyBufP != NULL);

  replyP = (m_page_reply_t *) replyBufP;
  replyP->pageSize = pageSize;
  replyP->pagerbuf = requestP->pagerbuf;
  replyLen = sizeof(*replyP) - sizeof(replyP->data);
  replyP->dataSize = 0;
  replyP->dataOffset = 0;

  /* lookup region info for this address */
  riP = region_from_addr ((int *) reqAddr);

  /* verify that request */
  if (riP->owner != Comm_index) {
    fprintf(stderr, 
	    "Invalid page request: from host=%lu thd=%lu addr=0x%x Owner=%d\n",
	    replyHost,replyThread, reqAddr, riP->owner);
    fflush(stderr);
    replyP->status = PAGE_REPLY_STATUS_NOT_OWNER;
    INCR_STATS (pageStat.RequestsInvalid, 1);
    goto sendReply;
  }
  if (riP->page_tableP == NULL) {
    fprintf(stderr, 
      "Invalid page request: from host=%lu thd=%lu addr=0x%x -Not Allocated\n",
	    replyHost,replyThread, reqAddr);
    fflush(stderr);
    replyP->status = PAGE_REPLY_STATUS_INVALID_ADDR;
    INCR_STATS (pageStat.RequestsInvalid, 1);
    goto sendReply;
  }
  if (requestLen == 0) {
    fprintf(stderr, 
	    "Invalid page request: from host=%lu thd=%lu addr=0x%x len=%d\n",
	    replyHost,replyThread, reqAddr, requestLen);
    fflush(stderr);
    replyP->status = PAGE_REPLY_STATUS_INVALID_PARM;
    INCR_STATS (pageStat.RequestsInvalid, 1);
    goto sendReply;
  }

    
  /*
   * Yes, we own this page.
   * Before returning the page's contents, add the requestor to 
   * the copyset & update all members 
   */


  /*
  ** NOTE: We currently don't support sending pages between machines
  ** with different page sizes. This needs to be added.  WAS 18Aug92
  */
  ASSERT (pageSize == requestLen);
  ASSERT (pageSize <= MAX_DATA_SIZE);  /* just to be sure */
  ASSERT (((u_long)reqAddr & (requestLen - 1)) == 0);


  /* get page table entry for this page 
   * This contains the page's copy set 
   */
  { u_long byte_offset, pageNo;
    byte_offset = (u_long) ((char *) reqAddr - (char *)riP->base);
    pageNo = ADDR_TO_PAGENO(byte_offset);
    pageEntryP = &riP->page_tableP[ pageNo ];
  }

  /* 
   * Check if the page has been allocated.
   */
  if (!midTestBit (pageEntryP->bitArray, Comm_index)) {

    /* Page has NOT been allocated.
     * verify that the copyset is empty 
     */
    ASSERT (EQUAL_ZERO_ADDR_VECTOR(*pageEntryP));

    /* Allocate zero filled page and tell requestor to do same. */

    /* touch page.  Pagin request should allocate, and zero fill.*/
    *(int *)reqAddr = 0;

    /* verify page as allocated */
    ASSERT (midTestBit (pageEntryP->bitArray, Comm_index));
    
    /* Add requestor to copyset */
    midSetBit (pageEntryP->bitArray, replyHost);
    
    /* return copyset to caller */
    replyP->copyset = *pageEntryP;
      
    /* tell requestor to zerofill page */
    replyP->dataPkts = 1;
    replyP->status = PAGE_REPLY_STATUS_ZEROFILL;

    INCR_STATS (pageStat.RequestsZeroFilled, 1);
    goto sendReply;
  }

  
  /*
   * Page was previously allocated.
   * Flush any queued updates to page, before updating the copyset
   * so that requester isn't send the updates unnecessary.
   */
  if (UPDATEQ_NOT_EMPTY) {
    if (IS_DBG_PORT_QUEUE) StatusMsg("PageFault flushing update queue");
    rc = mid_updateq_flush (UPDATEQ_CALLER_PAGEFAULT);
    if (rc != 0) {
      replyP->status = PAGE_REPLY_STATUS_FLUSH_ERROR;
      goto sendReply;
    }
  }


  /*
   * Check if requestor is already in copy set
   * Possibly a duplicate transmission, possibly multiply faults on same page
   */
  if (!midTestBit (pageEntryP->bitArray, replyHost)) {
    
    /* generate destination address */
    destAddr = *pageEntryP;
    midClearBit (destAddr.bitArray, Comm_index); /* remove self */
    
    /* Add requestor to copyset for page table */
    midSetBit (pageEntryP->bitArray, replyHost);

    /* Check if there are other members in the copy set */
    if (!EQUAL_ZERO_ADDR_VECTOR (destAddr)) {
	
      if (IS_DBG_PORT_PAGE) {
	char255 smsg, s1, s2;
	sprintf(smsg, "Updating copyset to 0x%s for members 0x%s",
		sprintf_addr_vector (s1,*pageEntryP), 
		sprintf_addr_vector (s2, destAddr));
	StatusMsg (smsg);
      }

      /* build request */
      copysetReq.pageSize = pageSize;
      copysetReq.copyset = *pageEntryP;
      
      /* Send a multicast request and wait for all replies */
      CommMrequest (reqAddr,             /* Use page address for lock */
		    M_PAGE_COPYSET,      /* msg type */
		    0,                   /* reply to host, 0=us */
		    0,                   /* reply to tid, 0=us */
		    NULL_CLOCK_VALUE,    /* clock */
		    destAddr.bitArray,   /* destination address list */
		    &copysetReq,         /* msg = new copy set */
		    sizeof(copysetReq)); /* size of msg */

      INCR_STATS (pageStat.RequestsUpdateCopyset, 1);
      INCR_STATS (pageStat.RequestsCopysetDataSent, sizeof(copysetReq));
    }
  } /* if (requestor was not in copyset) */

  /*
   * Loop to return page contents to the requestor.
   */
  packetNum = 0;
  lastPacket = 0;
  pageOffset = 0;
  pageP = (char *) reqAddr;
  bufferMax = mid_max_data_size - sizeof(*replyP);
  totalPackets = (pageSize + (bufferMax-1)) / bufferMax;

  INCR_STATS (pageStat.RequestsDataFilled, 1);
  INCR_STATS (pageStat.RequestsDataPackets, totalPackets);
  while (TRUE) {

    bufferSize = pageSize - pageOffset;
    if (bufferSize > bufferMax) 
      bufferSize = bufferMax;
    else
      lastPacket = 1;   /* exit loop after send */

    bcopy (&pageP[ pageOffset ], /* source */
	   replyP->data,         /* target */
	   bufferSize);          /* length */

    replyP->dataSize = bufferSize;
    replyP->dataOffset = pageOffset;
    replyP->dataPkts = totalPackets;
    replyP->copyset = *pageEntryP;    /* Also return copyset to requestor */
    replyP->status = KERN_SUCCESS;
    replyLen += bufferSize;
    
    if (IS_DBG_PORT_PAGE) {
      char255 smsg;
      sprintf(smsg, 
	      "page_request: sending %s packet %d: ADDR=0x%x Off=%lu Len=%lu",
	      (lastPacket) ? "LAST" : "", packetNum, 
	      reqAddr, pageOffset, bufferSize);
      StatusMsg (smsg);
    }

    /*
     * Send page contents to requestor 
     */
    CommReply (M_PAGE_DATA,     /* kind of message */
	       replyHost,       
	       replyThread, 
	       replyTid, 
	       reqAddr,         /* Send requested address as "lock" */
	       lastPacket,      /* 1= last packet */
	       packetNum,       /* packet number */
	       NULL_ADDR_VECTOR,  /* nacks (always 0 except for lock replies) */
	       (char *) replyP, /* reply message */
	       replyLen,        /* reply message length */
	       C_FREE);         /* free message buffer when complete */

    INCR_STATS (pageStat.RequestsDataSent, replyLen);

    if (lastPacket) goto done;

    pageOffset += bufferSize;
    packetNum++;

    /* alloc & initialize buffer for reply */
    replyBufP = mid_pbuf_get();
    ASSERT (replyBufP != NULL);

    replyP = (m_page_reply_t *) replyBufP;
    replyP->pageSize = pageSize;
    replyP->pagerbuf = requestP->pagerbuf;
    replyLen = sizeof(*replyP) - sizeof(replyP->data);

  } /* while (TRUE) */

  /* Code should never break from loop */
  

sendReply:

  if (IS_DBG_PORT_PAGE) {
    char255 smsg;
    sprintf(smsg, 
	    "page_request: complete, len=%d pb=%x ADDR=0x%x PAGE=%d RC=%d",
	    replyLen, replyP->pagerbuf, reqAddr, 
	    replyP->pageSize, replyP->status);
    StatusMsg (smsg);
  }

  /*
   * Send reply to requestor 
   */
  CommReply (M_PAGE_DATA,     /* kind of message */
	     replyHost,       
	     replyThread, 
	     replyTid, 
	     reqAddr,         /* Send requested address as "lock" */
	     1,               /* 1= last packet */
	     (u_long) 0,      /* 0= packet number */
	     NULL_ADDR_VECTOR,/* nacks */
	     (char *) replyP, /* reply message */
	     replyLen,        /* reply message length */
	     C_FREE);         /* free message buffer when complete */

    INCR_STATS (pageStat.RequestsDataSent, replyLen);

done:;
  return;

} /* mid_page_request */


/*
 ************************************************************************
 *
 * mid_page_reply
 *
 * Handle the reply to a request for the contents of a page
 *
 ************************************************************************
 */
void
mid_page_reply
  (
   u_long  reqAddr,             /* address of requested page */
   u_long  replyHost,	        /* host to reply to */
   u_long  replyThread,	        /* thread on host to reply to */
   u_long  replyTid,	        /* the transaction id */
   u_long  packetNumber,	/* reply packet number */
   u_long *bufP,                /* ptr to message buffer */
   u_long  bufLen               /* length of message */
   )
{
  region_info riP;
  kern_return_t krc;
  m_page_reply_t *replyP;
  addr_vector_t *pageEntryP;
  pbuf pageBufP;
  char *pageP;
  page_fragment_t *fragP, *prevFragP;

  if (IS_DBG_PORT_PAGE) {
    char255 smsg;
    sprintf(smsg, 
	    "Page_Reply: (addr=0x%x host=%lu thd=%lu pkt=%lu bufLen=%lu)", 
	    reqAddr, replyHost, replyThread, packetNumber, bufLen);
    StatusMsg (smsg);
  }
  
  INCR_STATS (pageStat.Replies, 1);

  /* verify the returned message */
  if (bufLen < (sizeof(*replyP) - sizeof(replyP->data))) {
    fprintf(stderr, "Invalid page reply received from host=%lu thread=%lu\n",
	    replyHost, replyThread);
    fprintf(stderr, "Message is too short, len=%d\n", bufLen);
    fflush(stderr);
    INCR_STATS (pageStat.RepliesInvalid, 1);
    goto done;
  }

  /* get status from message */
  replyP = (m_page_reply_t *) bufP;
  if ((replyP->status != PAGE_REPLY_STATUS_SUCCESS) &&
      (replyP->status != PAGE_REPLY_STATUS_ZEROFILL)) {
    fprintf(stderr, "Invalid page reply received from host=%lu thread=%lu\n",
	    replyHost, replyThread);
    fprintf(stderr, "Rejected page request, for addr=0x%x page=%lu, rc=%d\n",
	    reqAddr, replyP->pageSize, replyP->status);
    fflush(stderr);
    INCR_STATS (pageStat.RepliesInvalid, 1);
    goto done;
  }

  /* verify message length */
  if ((replyP->status == PAGE_REPLY_STATUS_SUCCESS) &&
      (bufLen != (replyP->dataSize + sizeof(*replyP) - sizeof(replyP->data)))) {
    fprintf(stderr, "Invalid page reply received from host=%lu thread=%lu\n",
	    replyHost, replyThread);
    fprintf(stderr, "Bad message length, for addr=0x%x len=%lu bufLen=%lu\n",
	    reqAddr, replyP->dataSize, bufLen);
    fflush(stderr);
    INCR_STATS (pageStat.RepliesInvalid, 1);
    goto done;
  }

  if ((replyP->status == PAGE_REPLY_STATUS_ZEROFILL) &&
      (bufLen != (sizeof(*replyP) - sizeof(replyP->data)))) {
    fprintf(stderr, "Invalid page reply received from host=%lu thread=%lu\n",
	    replyHost, replyThread);
    fprintf(stderr, "Bad zerofill length, for addr=0x%x len=%lu bufLen=%lu\n",
	    reqAddr, replyP->dataSize, bufLen);
    fflush(stderr);
    INCR_STATS (pageStat.RepliesInvalid, 1);
    goto done;
  }

  /* verify the packet number */
  if (packetNumber > MAX_PAGE_REPLY_PACKETS) {
    fprintf(stderr, "Invalid page reply received from host=%lu thread=%lu\n",
	    replyHost, replyThread);
    fprintf(stderr, "Bad packet number %d, for addr=0x%x len=%lu bufLen=%lu\n",
	    packetNumber, reqAddr, replyP->dataSize, bufLen);
    fflush(stderr);
    INCR_STATS (pageStat.RepliesInvalid, 1);
    goto done;
  }

  /*
  ** verify matching page lengths. Needs fixed. WAS 18Aug92
  */
  ASSERT (replyP->pageSize == vm_page_size);
  ASSERT ((reqAddr & (replyP->pageSize - 1)) == 0);
  

  /* lookup region info for this address */
  riP = region_from_addr ((int *) reqAddr);

  /* get page table entry for this page 
   * This contains the page's copy set 
   */
  { u_long byte_offset, pageNo;
    byte_offset = (u_long)((char *)reqAddr - (char *)riP->base);
    pageNo = ADDR_TO_PAGENO(byte_offset);
    pageEntryP = &riP->page_tableP[ pageNo ];
  }
  
  /* Check if the page is allocated.
   * Could be result of a retransmission or multiple faults on same page
   */
  if (midTestBit (pageEntryP->bitArray, Comm_index)) {
    
    if (IS_DBG_PORT_PAGE) {
      char255 smsg;
      sprintf(smsg, 
	      "Duplicate page reply for 0x%x received from host=%lu thd=%lu",
	      reqAddr, replyHost, replyThread);
      StatusMsg (smsg);
    }
    INCR_STATS (pageStat.RepliesDuplicates, 1);
    goto done;
  }

  /* verify that we requested this page */
  if (!midTestBit (pageEntryP->bitArray, riP->owner)) {
    fprintf(stderr,
	    "Unrequested page reply for addr=0x%x from host=%lu thd=%lu\n",
	    reqAddr, replyHost, replyThread);
    fflush (stderr);
    INCR_STATS (pageStat.RepliesUnrequested, 1);
    goto done;
  }
  

  if (IS_DBG_PORT_PAGE) {
    char255 smsg, s1;
    sprintf(smsg, 
	    "page_reply: requested page 0x%x size=%d %sfill pb=0x%x copyset=0x%s",
	    reqAddr, replyP->pageSize, 
	    (replyP->status == PAGE_REPLY_STATUS_SUCCESS) ? "data" : "zero",
	    replyP->pagerbuf, sprintf_addr_vector (s1, replyP->copyset));
    StatusMsg (smsg);
  }
  
  /* assume success */
  krc = KERN_SUCCESS;

  /*
   * Check if this is a zero fill page, with no pending updates.
   */
  if ((replyP->status == PAGE_REPLY_STATUS_ZEROFILL) &&
      (pending_headP == NULL)) {

    *(replyP->pagerbuf) = NULL;	/* pager will zero fill */

    INCR_STATS (pageStat.RepliesZeroFilled, 1);
    goto updateCopyset;		/* update copyset for this page */
  }


  /*
   * Buffer for this page must be allocted.
   * Check if this is the first packet for this page
   */
  prevFragP = NULL;
  mutex_lock (pageFragMutex);

  VERIFY_FRAGS();

  fragP = pageFragListP;
  while (fragP != NULL) {
    if (fragP->pageAddr == reqAddr) goto found;
    prevFragP = fragP;
    ASSERT (fragP != fragP->nextP);
    fragP = fragP->nextP;
  }

  /* first packet for this page.
   * allocate fragment descriptor from the free list.
   */
  if (pageFragFreeListP != NULL) {
    fragP = pageFragFreeListP;
    pageFragFreeListP = fragP->nextP;
  } else {
    fragP = (page_fragment_t *) MALLOC(sizeof(*fragP));
    ASSERT (fragP != NULL);
  }
  
  /* initialize descriptor */
  fragP->pageAddr = reqAddr;
  midZeroBitArray (fragP->pktsArrived, MAX_PAGE_REPLY_PACKETS);
  fragP->pktsExpected = replyP->dataPkts;
  fragP->pktsReceived = 0;
  fragP->nextP = pageFragListP;
  pageFragListP = fragP;
  prevFragP = NULL;

  VERIFY_FRAGS();

  /* allocate memory for page buffer */
  krc = VM_ALLOCATE (midway_task_self, 
		     (vm_address_t *)&(fragP->dataP), 
		     replyP->pageSize,
		     TRUE /* allocate anywhere*/ );
  if (krc != KERN_SUCCESS) {
    mutex_unlock (pageFragMutex);
    goto updateCopyset;
  }

  INCR_STATS (pageStat.RepliesAllocated, 1);
      
  if (IS_DBG_PORT_PAGE) {
    char255 smsg;
    sprintf(smsg,"page reply: alloc tmp buffer at 0x%x, data at 0x%x len=%d\n",
	    fragP->dataP, replyP->data, replyP->pageSize);
    StatusMsg (smsg);
  }

found:

  VERIFY_FRAGS();

  /* check if packet was previously received */
  if (midTestBit (fragP->pktsArrived, packetNumber)) {
    
    if (IS_DBG_PORT_PAGE) {
      char255 smsg;
      sprintf(smsg,"page alloc: duplicate packet %d for ADDR=0x%x\n",
	      packetNumber, fragP->pageAddr);
      StatusMsg (smsg);
    }
    INCR_STATS (pageStat.RepliesDuplicatePackets, 1);
    mutex_unlock (pageFragMutex);
    goto done;
  }
    
  /* copy data into buffer */
  if (replyP->status != PAGE_REPLY_STATUS_ZEROFILL) {
    bcopy((char*) replyP->data,                 /* source */
	  &(fragP->dataP[replyP->dataOffset]),  /* target */
	  replyP->dataSize);	                /* length */
  }

  midSetBit (fragP->pktsArrived, packetNumber);
  fragP->pktsReceived++;

  /* Check if all packets have been received */
  if (fragP->pktsReceived < fragP->pktsExpected) {
    mutex_unlock (pageFragMutex);
    goto done;
  }

  ASSERT (fragP->pktsReceived == fragP->pktsExpected);

  if (IS_DBG_PORT_PAGE) {
    char255 smsg;
    sprintf(smsg,"page reply: buffer 0x%x now complete for ADDR 0x%x",
	    fragP->dataP, fragP->pageAddr);
    StatusMsg (smsg);
  }
  
  /* 
   * Apply any pending updates to this page.
   * The updates are applied to the request buffer
   * so that it is mapped atomically into the requested address.
   */
  if (pending_headP != NULL) {
    mid_updateq_apply_pending (fragP->dataP, 
			       fragP->pageAddr, replyP->pageSize);
    INCR_STATS (pageStat.RepliesPendingUpdates, 1);
  }

  /* return buffer pointer to pager */
  *(replyP->pagerbuf) = (pointer_t)(fragP->dataP);


  VERIFY_FRAGS();

  /* remove from active list */
  if (prevFragP == NULL)
    pageFragListP = fragP->nextP;
  else
    prevFragP->nextP = fragP->nextP;
  
  /* return to free list */
  SMASH_AREA( fragP, sizeof(*fragP));
  fragP->nextP = pageFragFreeListP;
  pageFragFreeListP = fragP;

  VERIFY_FRAGS();

  mutex_unlock (pageFragMutex);


updateCopyset:

  if (krc == KERN_SUCCESS) {

    /* mark page as allocated 
     * Use merge instead of assignment in case copyset requests have
     * arrived before the page reply.   WAS 28Aug92
     */
    MERGE_ADDR_VECTOR (*pageEntryP, replyP->copyset);
    
    /* verify the copyset provided by the page's owner */
    ASSERT (midTestBit (pageEntryP->bitArray, Comm_index));
    ASSERT (midTestBit (pageEntryP->bitArray, riP->owner));
  } else {

    fprintf(stderr, "Error allocating page at 0x%x size=%d, rc=%d\n",
	    reqAddr, replyP->pageSize, krc);
    fflush(stderr);

    INCR_STATS (pageStat.RepliesErrors, 1);

    /* Reset page request bit */
    midClearBit (pageEntryP->bitArray, riP->owner);
    ASSERT (!midTestBit (pageEntryP->bitArray, Comm_index));
  }

done:
  return;
} /* mid_page_reply */

/*
 ************************************************************************
 *
 * mid_copyset_request
 *
 * Handle a request to update the copy set for a page
 *
 ************************************************************************
 */
void
mid_copyset_request
  (
   u_long  reqAddr,             /* address of requested page */
   u_long  replyHost,	        /* host to reply to */
   u_long  replyThread,	        /* thread on host to reply to */
   u_long  replyTid,	        /* the transaction id */
   u_long *bufP,                /* message buffer */
   u_long  bufLen               /* buffer length */
   )
{
  region_info riP;
  addr_vector_t *pageEntryP;
  m_copyset_request_t *requestP;
  m_copyset_reply_t reply;
  

  if (IS_DBG_PORT_PAGE) {
    char255 smsg;
    sprintf(smsg, 
	    "Copyset_Request (host=%lu thd=%lu tid=%lu bufLen=%d ADDR=0x%x)",
	    replyHost, replyThread, replyTid, reqAddr, bufLen);
    StatusMsg (smsg);
  }
  INCR_STATS (pageStat.CopysetRequests, 1);

  /* verify incoming request */
  if (bufLen != sizeof(*requestP)) {
    fprintf(stderr,"Invalid copyset update from %lu thd %lu for addr 0x%x\n",
	    replyHost, replyThread, reqAddr);
    fprintf(stderr, "Buffer length = %d, expecting %d bytes\n",
	    bufLen, sizeof(*requestP));
    fflush (stderr);
    
    reply.status = COPYSET_REPLY_STATUS_INVALID;
    INCR_STATS (pageStat.CopysetRequestsInvalid, 1);
    goto done;
  }

  /* cast pointer to message buffer */
  requestP = (m_copyset_request_t *) bufP;
  
  /* lookup region info for this address */
  riP = region_from_addr ((int *) reqAddr);

  /* get page table entry for this page */
  { u_long byte_offset, pageNo;
    byte_offset = (u_long)((char *)reqAddr - (char *)riP->base);
    pageNo = ADDR_TO_PAGENO(byte_offset);
    pageEntryP = &riP->page_tableP[ pageNo ];
  }

  /* Check that the page is allocated.
   * This occurs when page is dropped from local cache
   * or has been requested, but not yet received.
   */
  if (!midTestBit (pageEntryP->bitArray, Comm_index)) {
    
    /* Check if the page was requested */
    if (midTestBit (pageEntryP->bitArray, riP->owner)) {

      if (IS_DBG_PORT_PAGE) {
	char255 smsg, s1, s2;
	sprintf(smsg, 
		"Update copyset for 0x%x prior to page_reply from owner %d",
		reqAddr, riP->owner);
	StatusMsg (smsg);
	sprintf(smsg,
		"Update copyset from host=%lu thd=%lu tid=%lu, CS=%s Page=%s",
		replyHost, replyThread, replyTid, 
		sprintf_addr_vector(s1, requestP->copyset),
		sprintf_addr_vector(s2, *pageEntryP));
	StatusMsg (smsg);
      }

      /* Copyset update received before page returned by owner.
       * Handle by merging copysets. Okay to have too many members
       * in copyset (since updates will be nacked), its lethal
       * to have too few.    WAS 28Aug92
       */
      MERGE_ADDR_VECTOR (*pageEntryP, requestP->copyset);

      /* reset our bit to indicate the page is not allocated */
      midClearBit (pageEntryP->bitArray, Comm_index);
      ASSERT (midTestBit (pageEntryP->bitArray, riP->owner)); 
      reply.status = COPYSET_REPLY_STATUS_SUCCESS;

    } else {

      /* We shouldn't receive updates when there are no pending requests */
      char255 smsg, s1, s2;
      sprintf(smsg, "Update copyset error for unallocated page 0x%x", reqAddr);
      fprintf(stderr, "%s\n", smsg);
      StatusMsg (smsg);
      sprintf(smsg,
	      "Update copyset from host=%lu thd=%lu tid=%lu, CS=%s Page=%s",
	      replyHost, replyThread, replyTid, 
	      sprintf_addr_vector(s1, requestP->copyset),
	      sprintf_addr_vector(s2, *pageEntryP));
      fprintf(stderr, "%s\n", smsg);
      fflush (stderr);
      StatusMsg (smsg);
      reply.status = COPYSET_REPLY_STATUS_NACK;
      INCR_STATS (pageStat.CopysetRequestsNacked, 1);
    }

    goto done;
  }

  if (IS_DBG_PORT_PAGE) {
    char255 smsg, s1, s2;
    sprintf(smsg, "page 0x%x size=%d, copyset was 0x%s  setting to 0x%s",
	    reqAddr, requestP->pageSize,
	    sprintf_addr_vector (s1, *pageEntryP),
	    sprintf_addr_vector (s2, requestP->copyset));
    StatusMsg (smsg);
  }

  /* handle mismatched page sizes. */
  ASSERT (requestP->pageSize == vm_page_size);

  /* Copy new page table entry */
  *pageEntryP = requestP->copyset;

  /* Return success to caller */
  reply.status = COPYSET_REPLY_STATUS_SUCCESS;


done:
  
  if (IS_DBG_PORT_PAGE) {
    char255 smsg;
    sprintf(smsg, "copyset_request: complete, page 0x%x rc=%d",
	    reqAddr, reply.status);
    StatusMsg (smsg);
  }

  /* Reply to sender */
  CommReply (M_PAGE_COPYSET,  /* message type */
	     replyHost,       
	     replyThread, 
	     replyTid, 
	     reqAddr,         /* Send requested address as "lock" */
	     1,               /* 1= last packet */
	     (u_long) 0,      /* 0= packet number */
	     NULL_ADDR_VECTOR,/* nacks */
	     (char *) &reply, /* reply message */
	     (u_long) sizeof(reply), /* message length */
	     C_NOFREE);       /* no buffer to be freed */

  INCR_STATS (pageStat.CopysetRequestsDataSent, sizeof(reply));

  return;

} /* mid_copyset_request */


/*
 ************************************************************************
 *
 * mid_copyset_reply
 *
 * Handle replies to requests to update the copyset for a page 
 *
 ************************************************************************
 */
void
mid_copyset_reply
  (
   u_long  reqAddr,             /* address of requested page */
   u_long  replyHost,	        /* host to reply to */
   u_long  replyThread,	        /* thread on host to reply to */
   u_long  replyTid,	        /* the transaction id */
   u_long *bufP,                /* message buffer */
   u_long  bufLen               /* buffer length */
   )
{
  region_info riP;
  addr_vector_t *pageEntryP;
  m_copyset_reply_t *replyP;
  

  if (IS_DBG_PORT_PAGE) {
    char255 smsg;
    sprintf(smsg, "copyset_reply: from host=%lu thd=%lu tid=%lu, ADDR=0x%x",
	    replyHost, replyThread, replyTid, reqAddr);
    StatusMsg (smsg);
  }
  INCR_STATS (pageStat.CopysetReplies, 1);

  /* verify incomming request */
  if (bufLen != sizeof(*replyP)) {
    fprintf(stderr,"Invalid copyset reply from %lu thd %lu for addr 0x%x\n",
	    replyHost, replyThread, reqAddr);
    fflush (stderr);
    INCR_STATS (pageStat.CopysetRepliesInvalid, 1);
    goto done;
  }

  /* cast pointer to message buffer */
  replyP = (m_copyset_reply_t *) bufP;

  /* Check for success */
  if (replyP->status == COPYSET_REPLY_STATUS_SUCCESS)
    goto done;
  
  /* Check for hard errors */
  if (replyP->status != COPYSET_REPLY_STATUS_NACK) {
    fprintf(stderr,"Copyset reply error from host %lu thd %lu, Addr=0x%x\n",
	    replyHost, replyThread, reqAddr);
    fflush (stderr);
    INCR_STATS (pageStat.CopysetRepliesNacked, 1);
    goto done;
  }
    

  /* 
   * Host NACK'ed copyset update.
   * This means host is no longer interested in copyset updates
   * and should be removed from the copyset for this page.
   */

  /* lookup region info for this address */
  riP = region_from_addr ((int *) reqAddr);

  /* get page table entry for this page */
  { u_long byte_offset, pageNo;
    byte_offset = (u_long)((char *)reqAddr - (char *)riP->base);
    pageNo = ADDR_TO_PAGENO(byte_offset);
    pageEntryP = &riP->page_tableP[ pageNo ];
  }

  if (IS_DBG_PORT_PAGE) {
    char255 smsg, s1;
    sprintf(smsg, "Removing host %lu from copyset 0x%s for page 0x%x",
	    replyHost, sprintf_addr_vector(s1, *pageEntryP), reqAddr);
    StatusMsg (smsg);
  }

  /* Check that the page is allocated.
   * Should always be allocated?
   */
  if (!midTestBit (pageEntryP->bitArray, Comm_index)) {
    fprintf(stderr, "Copyset reply error for unallocated page 0x%x",
	    reqAddr);
    fprintf(stderr, "Copyset reply from host=%lu thd=%lu tid=%lu",
	    replyHost, replyThread, replyTid);
    fflush (stderr);
    INCR_STATS (pageStat.CopysetRepliesNackInvalid, 1);
    goto done;
  }

  /* Remove the host from this copyset */
  midClearBit (pageEntryP->bitArray, replyHost);


done:
  return;

} /* mid_copyset_reply */


/*
 ************************************************************************
 *
 * mid_page_print_stats
 *
 * Display page manager statistics
 *
 ************************************************************************
 */
void
mid_page_print_stats ()
{
#ifndef STATS
  StatusMsg("Not compiled to collect Non-EC Page Manager Statistics");
#else
  char255 smsg;

  StatusMsg("-=-=- Non-EC Page Manager Statistics -=-=-");
  
  sprintf(smsg,
    "Faults: %lu, InvReg=%lu, InvPage=%lu, Zero=%lu, Remote=%lu, Failed=%lu",
    pageStat.Faults, pageStat.FaultsOnInvalidRegion, 
    pageStat.FaultsOnAllocatedPage, pageStat.FaultsZeroFilled,
    pageStat.FaultsRemote, pageStat.FaultsFailed);
  StatusMsg (smsg);

  sprintf(smsg,
    "Requests: %lu, Inv=%lu, Zero=%lu, UpCS=%lu, DataFill=%lu, DataPkts=%lu",
    pageStat.Requests, pageStat.RequestsInvalid, 
    pageStat.RequestsZeroFilled, pageStat.RequestsUpdateCopyset,
    pageStat.RequestsDataFilled, pageStat.RequestsDataPackets);
  StatusMsg (smsg);

  sprintf(smsg,
    "Replies: %lu, Inv=%lu, Dup=%lu, UnReq=%lu Zero=%lu, Alloc=%lu, DupPkt=%lu, Pend=%lu, Err=%lu",
    pageStat.Replies, pageStat.RepliesInvalid, 
    pageStat.RepliesDuplicates, pageStat.RepliesUnrequested,
    pageStat.RepliesZeroFilled, pageStat.RepliesAllocated,  
    pageStat.RepliesDuplicatePackets, pageStat.RepliesPendingUpdates,
    pageStat.RepliesErrors);
  StatusMsg (smsg);

  sprintf(smsg,
    "Copyset Requests: %lu, Inv=%lu, Nacked=%lu, Replies %lu",
     pageStat.CopysetRequests, pageStat.CopysetRequestsInvalid,
     pageStat.CopysetRequestsNacked,  pageStat.CopysetReplies,
     pageStat.CopysetRepliesInvalid,  pageStat.CopysetRepliesNacked,
     pageStat.CopysetRepliesNackInvalid);
  StatusMsg (smsg);

  sprintf(smsg,
    "DataSent: Total=%lu Faults=%lu Replies=%lu Copysets=%lu Replies=%lu",
	  (pageStat.FaultsDataSent + pageStat.RequestsDataSent
	   + pageStat.RequestsCopysetDataSent
	   + pageStat.CopysetRequestsDataSent),
	  pageStat.FaultsDataSent, pageStat.RequestsDataSent, 
	  pageStat.RequestsCopysetDataSent, pageStat.CopysetRequestsDataSent);
  StatusMsg (smsg);
#endif

  return;
}
