/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * pagemgr.h: Types and prototypes for managing page tables
 *
 * 25 Aug 92 (was): Initial code 
 * 20 Oct 92 (was): Changed page reply to split pages into multi-packets
 *
 */
#ifndef __PAGEMGR_H__
#define __PAGEMGR_H__


#ifdef __PAGEMGR_INTERNAL__



/* 
 * Generate definitions internal to the page manager code.
 */


/*
 * Define M_PAGE_xxx message formats
 */
typedef struct {
  u_long pageSize;     /* length of data, Address sent as "lock" */
  pointer_t *pagerbuf;	/* ptr to pagein request data buffer ptr  */
} m_page_request_t;

typedef struct {
  u_long        pageSize;       /* length of entire page */
  u_long        dataSize;       /* amount of data in this packet */
  u_long        dataOffset;     /* offset of data within page */
  u_short       dataPkts;       /* number of packets to expect */
  u_short       status;
  addr_vector_t copyset;
  pointer_t	*pagerbuf;	/* ptr to pagein request data buffer ptr  */
  char          data[4];        /* actual length = dataSize */
} m_page_reply_t;

#define PAGE_REPLY_STATUS_SUCCESS       0   /* page data enclosed */
#define PAGE_REPLY_STATUS_ZEROFILL      1   /* page is full of zeros */
#define PAGE_REPLY_STATUS_NOT_OWNER     2   /* host is not owner */
#define PAGE_REPLY_STATUS_INVALID_ADDR  3   /* invalid page address */
#define PAGE_REPLY_STATUS_INVALID_PARM  4   /* invalid request parameter */
#define PAGE_REPLY_STATUS_ALLOC_ERROR   5   /* allocation error */
#define PAGE_REPLY_STATUS_FLUSH_ERROR   6   /* error flushing page updates */


typedef struct {
  u_long        pageSize;
  addr_vector_t copyset;
} m_copyset_request_t;

typedef struct {
  int           status;
} m_copyset_reply_t;

#define COPYSET_REPLY_STATUS_SUCCESS    0   /* copyset updated successfully */
#define COPYSET_REPLY_STATUS_NACK       1   /* No longer in copyset */
#define COPYSET_REPLY_STATUS_INVALID    2   /* Invalid request parameter */


/*
 * define descriptor used for reassembling fragmented pages
 */
typedef struct page_fragment_struct {
  u_long              pageAddr; 	/* address for requested page */
  char               *dataP;            /* pointer to data buffer */
  int                 pktsExpected;     /* number of pkts expected */
  int                 pktsReceived;     /* number of pkts received */
  BITARRAY            pktsArrived[1];   /* bit array for pkts received */
  struct page_fragment_struct *nextP;
} page_fragment_t;

#define MAX_PAGE_REPLY_PACKETS (BITARRAY_BITS)




#undef EXTERN
#define EXTERN			/* pagemgr routines are defined locally */
#undef INIT
#define INIT(val) = { val }     /* initialize external variables */

#else 

#undef EXTERN
#define EXTERN extern           /* pagemgr routines are external */
#undef INIT
#define INIT(val)               /* don't initialize external variables */

#endif /* __PAGEMGR_INTERNAL__ */




/*
 * Define prototypes for calling page manager routines 
 */

EXTERN char *
sprintf_addr_vector 
  (
   char255        bufP,		/* pointer to temp buffer for results */
   addr_vector_t  addr		/* address vector to be printed */
   );

#define PAGE_STATUS_FREE       0     /* page not allocated */
#define PAGE_STATUS_PENDING    1     /* page allocation is pending */
#define PAGE_STATUS_ALLOCATED  2     /* page is allocated */

EXTERN int 
mid_page_status
  (
   vm_address_t addr		/* Return page status (defined above) */
   );

EXTERN void
mid_page_copyset
  (
   vm_address_t  *dataAddrP,    /* ptr to changed address, rtnd page ptr */
   u_long        *dataLenP,	/* length of changed address, rtnd line size */
   addr_vector_t *copySetP	/* rtnd copyset for changed area */
   );

EXTERN kern_return_t
mid_page_fault 
  (
   vm_address_t  faultAddr,	/* address for faulting process */
   pointer_t	*bufP
   );

EXTERN void
mid_page_request
  (
   u_long        reqAddr,       /* address of requested page */
   u_long        replyHost,	/* host to reply to */
   u_long        replyThread,	/* thread on host to reply to */
   u_long        replyTid,	/* the transaction id */
   u_long       *bufP,          /* ptr to message buffer */
   u_long        bufLen         /* length of message */
   );

EXTERN void
mid_page_reply
  (
   u_long        reqAddr,       /* address of requested page */
   u_long        replyHost,	/* host to reply to */
   u_long        replyThread,	/* thread on host to reply to */
   u_long        replyTid,	/* the transaction id */
   u_long        packetNumber,  /* reply packet number */
   u_long       *bufP,          /* ptr to message buffer */
   u_long        bufLen         /* length of message */
   );

EXTERN void
mid_copyset_request
  (
   u_long        reqAddr,       /* address of requested page */
   u_long        replyHost,	/* host to reply to */
   u_long        replyThread,	/* thread on host to reply to */
   u_long        replyTid,	/* the transaction id */
   u_long       *bufP,          /* message buffer */
   u_long        bufLen         /* buffer length */
   );

EXTERN void
mid_copyset_reply
  (
   u_long        reqAddr,       /* address of requested page */
   u_long        replyHost,	/* host to reply to */
   u_long        replyThread,	/* thread on host to reply to */
   u_long        replyTid,      /* the transaction id */
   u_long       *bufP,          /* message buffer */
   u_long        bufLen         /* buffer length */
   );

EXTERN void
mid_page_print_stats ();


#endif /* __PAGEMGR_H__ */
