/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * pbuf: Packet BUFfers
 * 
 * 16 Jul 91 (mjz): moved pbufe to pbuf.h; added initqueue/enqueue/dequeue
 *			(mostly for barriers)
 * 26 Jul 91 (mjz): more Q stuff - for running through queue; added qfree
 * 30 Jul 91 (mjz): added _qmove
 * 20 Aug 92 (was): exported mid_pbuf_lock 
 * 15 Oct 92 (was): set pbuf size based on network packet size
 *  5 Nov 92 (was): changed ATM max pbuf size to account for next ptr
 * 19 Oct 93 (mjz): move include files around -- cthreads defines null ASSERT
 *  8 Apr 94 (was): must hold pbuf_lock when accessing pbuf_root
 */
#include <stdio.h>		/* NULL */
#include "midway-dbg.h"		/* MIDWAY_MUTEX_SET_NAME, ASSERT */
#include <cthreads.h>
#include "midway_types.h"	/* network type */
#include "smexecd_types.h"	/* ensure get char255 */

#include "pbuf.h"
#include <flow_times.h>

#define PBUF_SIG	0x50425546 /* PBUF in ascii */
pbufe_t *mid_pbuf_root;
mutex_t mid_pbuf_lock;	/* should protect the root... */

int mid_max_data_size;		/* def's for external vars */
int mid_max_pbuf_size;


#ifdef PBUF_VERIFY

/*
 * Maintain a list of all allocated pbufs.
 * Caller will provide an address & length
 * Simply search list to verify that caller's addr & len
 * reside within an allocated pbuf.
 */
static pbufe_t **pbuf_list;	/* array of pbuf ptrs */
static int pbuf_list_len;	/* number of pbufs allocated */
#define PBUF_LIST_MAX 1024 	/* max length for convenience */
#endif

#ifdef STATS

/* MUST be holding mid_pbuf_lock to update stats */
static struct  {
    unsigned long	allocated;
    unsigned long	allocated_size;
    unsigned long       gets;
    unsigned long	frees;
    unsigned long	qfrees;
    unsigned long	free_list;
    unsigned long	max_free_list;
} pbuf_stats;

#  define INCR_STATS(x,by) (x) += (by)
#else
#  define INCR_STATS(x,by)
#endif /* STATS */

void mid_pbuf_print_stats(void)
{
#ifndef STATS
    StatusMsg ("Not compiled to collect pbuf statistics");
#else 
    char255 smsg;
    
    sprintf(smsg, 
	    "pbuf: alloc %lu (%lu KB) get %lu free %lu qfree %lu list %lu/%lu",
	    pbuf_stats.allocated, pbuf_stats.allocated_size/1024,
	    pbuf_stats.gets, pbuf_stats.frees, pbuf_stats.qfrees, 
	    pbuf_stats.free_list, pbuf_stats.max_free_list);
    StatusMsg (smsg);
#endif
}


static int
pbuf_expand(num, where)
     int num;
     pbufe_t **where;		/* gets value & sets. */
{
  register char *dataP;
  register pbufe_t *p,*nextP;
  register int i;
  register int bufSize;

  
  /* verify parameters */
  ASSERT (num > 0);
  ASSERT ((mid_max_pbuf_size & (sizeof(long)-1)) == 0);

  
  /* get num packet buffers 
   * note that sizeof buffer = size of header less size of data placeholder
   *                            plus size of actual data buffer.
   */
  bufSize = (sizeof(*p) - sizeof(p->data)) +  mid_max_pbuf_size;
  dataP = (char *) MALLOC(num * bufSize);
  if (dataP == NULL)  {
    /*** OUT OF MEMORY ***/
#if STOP_ON_ERROR
    botch("pbuf_expand (out of memory)");
#endif
    return(-1);
  }

#ifdef STATS
  INCR_STATS (pbuf_stats.allocated, num);
  INCR_STATS (pbuf_stats.allocated_size, num*bufSize);
  INCR_STATS (pbuf_stats.free_list, num);
  if (pbuf_stats.free_list > pbuf_stats.max_free_list)
      pbuf_stats.max_free_list = pbuf_stats.free_list;
#endif

  /*
   * idea here is to push new blocks (that we just got in a big gulp)
   * onto the chain.  This will result in the big gulp chained "backwards" -
   * that is, the lower numbered memory blocks will be further down
   * in the chain... sort of diving in head first?
   */
  nextP = *where;		/* chain to remaining free blocks */
  for (i = 0; i < num; i++) {

    p = (pbufe_t *) dataP;	/* ptr to pbuf descriptor */
    p->sig = PBUF_SIG;		/* initialize this block */
    p->next = nextP;		/* chain in other blocks */

    nextP = p;			/* save ptr to old descriptor */
    dataP += bufSize;           /* advance to next buffer */
  }
  *where = p;			/* finally, point to new chain */
  return 0;
} /* pbuf_expand */

#if 0
/* was small with old net 'cause of overflowing fifos on barrier */
int mid_pbuf_atm_size = 1500-NETWORK_MSG_OVERHEAD; /* stupid calc. */
#endif
int mid_pbuf_atm_size = MAX_PBUF_SIZE;

#if FAST_UDP
/* small 'cause jumbo-grams are reassembled by UX, not in fast path */
int mid_pbuf_ether_size = 1500-NETWORK_MSG_OVERHEAD;
#else
int mid_pbuf_ether_size = MAX_PBUF_SIZE;
#endif

int
mid_pbuf_init()
{
  mid_pbuf_lock = mutex_alloc();
  MIDWAY_MUTEX_SET_NAME(mid_pbuf_lock,"mid_pbuf_lock");
  
  /* Set maximum packet size */
  /* We should use an ioctl to the network to determine the max
   * packet size that it supports efficiently. But currently
   * we just hardcode the standard IP max packet size.   WAS 15Oct92
   */
  /* with new integrated NW, need to check global to see if ATM enabled */
  if (midway_comm_network_type == MIDWAY_NET_ATM) {
    mid_max_pbuf_size = mid_pbuf_atm_size;
#if FLOW_TIMES
    mid_max_pbuf_size += sizeof(u_int);
#endif
  } else {
    mid_max_pbuf_size = mid_pbuf_ether_size;
  }

  /* round up pbuf size, just to be sure */
  mid_max_pbuf_size = ((mid_max_pbuf_size 
			  + (sizeof(long)-1)) / sizeof(long)) * sizeof(long);
  mid_max_data_size = mid_max_pbuf_size - MIDWAY_MSG_OVERHEAD;

  if (midway_comm_network_type == MIDWAY_NET_ATM) {
    mid_max_data_size -= sizeof(u_long);  /* atm stores buf len in pbuf */
#if FLOW_TIMES
    mid_max_data_size -= sizeof(u_int); /* store timestamp here too */
#endif
  }

#ifdef PBUF_VERIFY
  pbuf_list_len = 0;
  pbuf_list = (pbufe_t **) malloc(sizeof(pbufe_t *) * PBUF_LIST_MAX);
  ASSERT (pbuf_list != NULL);
#endif

#ifdef STATS
  bzero (&pbuf_stats, sizeof(pbuf_stats));
#endif

  return(pbuf_expand(32,&mid_pbuf_root));
} /* pbuf_init */

pbuf 
mid_pbuf_get()
{
  pbufe_t *p;

  mutex_lock(mid_pbuf_lock);
  INCR_STATS(pbuf_stats.gets, 1);

  if (!mid_pbuf_root) {
    if (pbuf_expand(8,&mid_pbuf_root)) {
      mutex_unlock(mid_pbuf_lock);
      return(0);
    }
  }
  INCR_STATS (pbuf_stats.free_list, -1);
  p = mid_pbuf_root;
  mid_pbuf_root = p->next;

#ifdef PBUF_VERIFY
  ASSERT (pbuf_list_len < PBUF_LIST_MAX);
  pbuf_list[ pbuf_list_len ] = p;
  pbuf_list_len++;
#endif

  mutex_unlock(mid_pbuf_lock);
  p->next = 0;
  return( p->data );
} /* mid_pbuf_get */

static void pbuf_free (pbufe_t *p)
{

    /* 
     * CALLER MUST HOLD PBUF MUTEX! 
     */

#ifdef PBUF_VERIFY
    { int _i;
      ASSERT (pbuf_list_len > 0);
      for (_i = 0; _i < pbuf_list_len; _i++) 
	  if (pbuf_list[_i] == p) break;
      ASSERT (_i < pbuf_list_len);
      pbuf_list_len--;
      pbuf_list[_i] = pbuf_list[pbuf_list_len];
    }
#endif
#ifdef STATS
    INCR_STATS(pbuf_stats.frees, 1);
    INCR_STATS (pbuf_stats.free_list, 1);
    if (pbuf_stats.free_list > pbuf_stats.max_free_list)
	pbuf_stats.max_free_list = pbuf_stats.free_list;
#endif

    p->next = mid_pbuf_root;
    mid_pbuf_root = p;
}

int
mid_pbuf_free(x)
     pbuf x;
{
  int off;
  pbufe_t *p;
  
  /* get ptr to buffer descriptor */
  off = sizeof(*p) - sizeof(p->data);
  p = (pbufe_t *) (((char *) x) - off);

  ASSERT(p->sig == PBUF_SIG);	/* for now */
  ASSERT((((unsigned long)p) & 0x3) == 0);

  mutex_lock(mid_pbuf_lock);
  pbuf_free(p);
  mutex_unlock(mid_pbuf_lock);
  return(0);
} /* mid_pbuf_free */

/* could macro-ize the following */
void
mid_pbuf_qinit( q )
     pbuf_q q;
{
  q->head =  0;
  q->last = &(q->head);		/* where next packet will be enqueued */
} /* mid_pbuf_qinit */

/* place p on queue q */
void
mid_pbuf_enqueue( p, q )
     pbuf p;
     pbuf_q q;
{
  int off;
  pbufe_t *x;

  /* get ptr to buffer descriptor */
  off = sizeof(*x) - sizeof(x->data);
  x = (pbufe_t *) (((char *) p) - off);

  ASSERT(x->sig == PBUF_SIG);	/* for now */

  x->next = 0;			/* just in case */
  *q->last = x;			/* save actual item */
  q->last = (&(x->next));	/* where next one will go */
  return;			/* done! */
} /* mid_pbuf_enqueue */

/* dequeue a pbuf from queue and return it */
pbuf
mid_pbuf_dequeue( q )
     pbuf_q q;
{
  pbuf p;
  pbufe_t *x;

  if (q->head) {			/* that is, there was something */
    x = q->head;
    p = x->data;		/* return pbuf */
    q->head = x->next;		/* dequeue it */
    x->next = 0;		/* Just In Case */
    if (q->head == 0)		/* need to reset last */
      q->last = &(q->head);
  }
  else {
    p = (pbuf) 0;		/* nothing to return */
  }

  return( p );
} /* mid_pbuf_dequeue */

/* free all queued pbuf's */
void
mid_pbuf_qfree( q )
     pbuf_q q;
{
  pbufe_t *x;

  /* rather than lock/unlock, going to do it once.  usually short, anyhow */
  mutex_lock(mid_pbuf_lock);
  while (q->head) {		/* there is something on Q */
    x = q->head->next;
    ASSERT(q->head->sig == PBUF_SIG);	/* for now */
    INCR_STATS(pbuf_stats.qfrees, 1);
    pbuf_free(q->head);
    q->head = x;
  }
  mutex_unlock(mid_pbuf_lock);

  /* reset Q */
  q->last = &(q->head);

  return;
} /* mid_pbuf_qfree */

/* move q1 to q2, freeing anything that was on q2. */
void
mid_pbuf_qmove( q1, q2 )
     pbuf_q q1;			/* source */
     pbuf_q q2;			/* dest */
{
  mid_pbuf_qfree(q2);
  q2->head = q1->head;
  q2->last = q1->last;
  mid_pbuf_qinit(q1);
  return;
} /* mid_pbuf_qmove */

/* Ya know, the rest of these could be macros.   I'll do it in upper case! */
/* --26jul91,mjz */
/* get top without disturbing q; returns pbufe_t NOT pbuf! */
pbufe_t *
mid_pbuf_qtop( q )
     pbuf_q q;
{
  return(q->head);
} /* mid_pbuf_qtop */

/* get bottom without disturbing q; (he he he) - this time returns pbuf */
/* NOTE: cannot take tail of empty Q */
pbuf
mid_pbuf_qtail( q )
     pbuf_q q;
{
  pbufe_t *x;

  /* q->last points to address of next field of last pbufe.  However, that
     is also the beginning of the pbufe (the next field is the 1st field) */
  x = (pbufe_t *) (q->last);
  return(x->data);
} /* mid_pbuf_qtail */

/* get next pbuf queue entry given one pbuf queue entry.  returns 0 at EOQ */
pbufe_t *
mid_pbuf_qnext( e )
     pbufe_t *e;
{
  return(e->next);
} /* mid_pbuf_qnext */

#ifdef PBUF_VERIFY
void mid_pbuf_verify (vm_offset_t addr, vm_size_t len)
{
    register int i, max;
    register pbufe_t *p, *z;
    register char *s = (char *) addr;
    register char *e = s + len;
    int min, diff;
    pbufe_t *minp;
    
    max = mid_max_pbuf_size;
    mutex_lock(mid_pbuf_lock);
    for (i = 0; i < pbuf_list_len; i++) {
	p = pbuf_list[i];
	if ((s >= p->data) && (s < &p->data[max]) &&
	    (e >= p->data) && (e < &p->data[max])) {
	    z = pbuf_list[0];
	    pbuf_list[0] = p;	/* move up in list */
	    pbuf_list[i] = z;
	    break;
	}
    }
    if (i >= pbuf_list_len) {

	/* Check if pbuf is on the free list */
	for (p =  mid_pbuf_root; p != NULL; p = p->next) {
	    if ((s >= p->data) && (s < &p->data[max]) &&
		(e >= p->data) && (e < &p->data[max]))
		break;
	}
	ASSERT (p == NULL);
	    
	/* Addr not allocated or on the free list 
	 * Find distance to closest pbuf 
	 */
	minp = NULL;
	min = 0;
	for (i = 0; i < pbuf_list_len; i++) {
	    p = pbuf_list[i];
	    if (s >= p->data) {
		diff = s - p->data;
		if ((minp == NULL) || (diff < min)) {
		    min = diff; 
		    minp = p;
		}
	    }
	}
	for (p =  mid_pbuf_root; p != NULL; p = p->next) {
	    if (s >= p->data) {
		diff = s - p->data;
		if ((minp == NULL) || (diff < min)) {
		    min = diff; 
		    minp = p;
		}
	    }
	}
	
	fprintf(stderr, 
		"Invalid pbuf addr 0x%x len %d, nearest pbuf 0x%x to 0x%x, %d diff\n",
		addr, len, 
		(minp == NULL) ? NULL : minp->data, 
		(minp == NULL) ? NULL : &minp->data[mid_max_pbuf_size], min);
	ASSERT (FALSE);
    }
    mutex_unlock(mid_pbuf_lock);
}
#endif
