/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/* 
 * ipatch.c:
 *
 * System dependent code to 
 * copy a dirtybit update template to the start of a region
 * and patch the instructions as needed.
 *
 *  4 Oct 93 (was): created
 * 17 Feb 94 (was): added special case for aligned 8 doubleword write
 */

#include <sys/types.h>		/* for comm.h */
#include <signal.h>
#define KERNEL 1
#include <sys/time.h>
#undef KERNEL
#include <sys/resource.h>
#include <mach.h>
#include <mach/kern_return.h>
#include <stdio.h>
#include <cthreads.h>
#include "midway_types.h"
#include "midway_mem.h"
#include "shmalloc.h"
#include "vm.h"
#include "smexecd_types.h"	/* char255 */
#include "midway-dbg.h"		/* DBG_IS_MISC_MAIF */
#include "midway.h"		/* prototypes */
#include "midway_comp.h"	/* template offsets */


/*
 * Define macros for patching MIPS R2000/R3000 instructions
 */
#define PATCH_INSTR_LOADUPPER(addr,val)                      \
{                                                            \
  /* MIPS: 31   26 25  21 20  16 15          0               \
   *        +-----+------+------+------------+               \
   *        | LUI |  rs  |  rt  |   offset   |               \
   *        +-----+------+------+------------+               \
   */                                                        \
  register u_long *instrP = (u_long *)(addr);                \
  *instrP = (*instrP & 0xFFFF0000) | ((u_long)(val) >> 16);  \
  ASSERT ((*instrP & 0xFC000000) == 0x3C000000);             \
}

#define PATCH_INSTR_LOADLOWER(addr,val)                            \
{                                                                  \
  /* MIPS: 31   26 25  21 20  16 15          0                     \
   *        +-----+------+------+------------+                     \
   *        | ORI |  rs  |  rt  |   immed    |                     \
   *        +-----+------+------+------------+                     \
   */                                                              \
  register u_long *instrP = (u_long *)(addr);                      \
  *instrP = (*instrP & 0xFFFF0000) | ((u_long)(val) & 0x0000FFFF); \
  ASSERT ((*instrP & 0xFC000000) == 0x34000000);                   \
}

#define PATCH_INSTR_SHIFTRIGHT(addr,val)                   \
{                                                          \
  /* MIPS: 31   26 25  21 20  16 15  11 10  6 5   0        \
   *        +-----+------+------+------+-----+-----+       \
   *        |SPECL|   0  |  rt  |  rd  |shamt| SRL |       \
   *        +-----+------+------+------+-----+-----+       \
   */                                                      \
  register u_long *instrP = (u_long *)(addr);              \
  ASSERT(((val) > 0) && ((val) < 32));                     \
  *instrP = (*instrP & 0xFFFFF83F) | ((u_long)(val) << 6); \
  ASSERT ((*instrP & 0xFFE0003F) == 0x00000002);           \
}


/*
 * Define external template routines & end labels for each
 */
extern void   __midway_builtin_unshared_word();
extern void   __midlabel_builtin_unshared_word_end();
extern void   __midway_builtin_unshared_half();
extern void   __midlabel_builtin_unshared_half_end();
extern void   __midway_builtin_unshared_byte();
extern void   __midlabel_builtin_unshared_byte_end();
extern void   __midway_builtin_unshared_double();
extern void   __midlabel_builtin_unshared_double_end();
extern void   __midway_builtin_unshared_area();
extern void   __midlabel_builtin_unshared_area_end();

extern void   __midway_builtin_aligned4_word();
extern void   __midlabel_builtin_aligned4_word_end();
extern void   __midway_builtin_aligned4_half();
extern void   __midlabel_builtin_aligned4_half_end();
extern void   __midway_builtin_aligned4_byte();
extern void   __midlabel_builtin_aligned4_byte_end();
extern void   __midway_builtin_aligned4_double();
extern void   __midlabel_builtin_aligned4_double_end();
extern void   __midway_builtin_aligned4_area();
extern void   __midlabel_builtin_aligned4_area_end();

extern void   __midway_builtin_aligned_word();
extern void   __midlabel_builtin_aligned_word_end();
extern void   __midway_builtin_aligned_half();
extern void   __midlabel_builtin_aligned_half_end();
extern void   __midway_builtin_aligned_byte();
extern void   __midlabel_builtin_aligned_byte_end();
extern void   __midway_builtin_aligned_double();
extern void   __midlabel_builtin_aligned_double_end();
extern void   __midway_builtin_aligned_area();
extern void   __midlabel_builtin_aligned_area_end();

extern void   __midway_builtin_aligned8_double();
extern void   __midlabel_builtin_aligned8_double_end();

extern void   __midway_builtin_unaligned_word();
extern void   __midlabel_builtin_unaligned_word_end();
extern void   __midway_builtin_unaligned_half();
extern void   __midlabel_builtin_unaligned_half_end();
extern void   __midway_builtin_unaligned_byte();
extern void   __midlabel_builtin_unaligned_byte_end();
extern void   __midway_builtin_unaligned_double();
extern void   __midlabel_builtin_unaligned_double_end();
extern void   __midway_builtin_unaligned_area();
extern void   __midlabel_builtin_unaligned_area_end();

extern void   __midway_builtin_mixed_word();
extern void   __midlabel_builtin_mixed_word_end();
extern void   __midway_builtin_mixed_half();
extern void   __midlabel_builtin_mixed_half_end();
extern void   __midway_builtin_mixed_byte();
extern void   __midlabel_builtin_mixed_byte_end();
extern void   __midway_builtin_mixed_double();
extern void   __midlabel_builtin_mixed_double_end();
extern void   __midway_builtin_mixed_area();
extern void   __midlabel_builtin_mixed_area_end();

/*
 * Define external template instruction labels
 */
extern void   __midlabel_al4_word_patch_upper_1();
extern void   __midlabel_al4_half_patch_upper_1();
extern void   __midlabel_al4_byte_patch_upper_1();
extern void   __midlabel_al4_double_patch_upper_1();
extern void   __midlabel_al4_area_patch_upper_1();
extern void   __midlabel_al4_area_patch_upper_2();
extern void   __midlabel_al4_area_patch_lower_1();

extern void   __midlabel_al_word_patch_upper_1();
extern void   __midlabel_al_word_patch_shift_1();
extern void   __midlabel_al_half_patch_upper_1();
extern void   __midlabel_al_half_patch_shift_1();
extern void   __midlabel_al_byte_patch_upper_1();
extern void   __midlabel_al_byte_patch_shift_1();
extern void   __midlabel_al_double_patch_upper_1();
extern void   __midlabel_al_double_patch_shift_1();
extern void   __midlabel_al_area_patch_upper_1();
extern void   __midlabel_al_area_patch_shift_1();
extern void   __midlabel_al_area_patch_upper_2();
extern void   __midlabel_al_area_patch_lower_1();

extern void   __midlabel_al8_double_patch_upper_1();

extern void   __midlabel_un_word_patch_upper_1();
extern void   __midlabel_un_word_patch_shift_1();
extern void   __midlabel_un_word_patch_upper_2();
extern void   __midlabel_un_word_patch_lower_1();
extern void   __midlabel_un_half_patch_upper_1();
extern void   __midlabel_un_half_patch_shift_1();
extern void   __midlabel_un_half_patch_upper_2();
extern void   __midlabel_un_half_patch_lower_1();
extern void   __midlabel_un_byte_patch_upper_1();
extern void   __midlabel_un_byte_patch_shift_1();
extern void   __midlabel_un_byte_patch_upper_2();
extern void   __midlabel_un_byte_patch_lower_1();
extern void   __midlabel_un_double_patch_upper_1();
extern void   __midlabel_un_double_patch_shift_1();
extern void   __midlabel_un_double_patch_shift_2();
extern void   __midlabel_un_double_patch_upper_2();
extern void   __midlabel_un_double_patch_lower_1();
extern void   __midlabel_un_area_patch_upper_1();
extern void   __midlabel_un_area_patch_shift_1();
extern void   __midlabel_un_area_patch_upper_2();
extern void   __midlabel_un_area_patch_lower_1();

extern void   __midlabel_mix_word_patch_upper_1();
extern void   __midlabel_mix_word_patch_shift_1();
extern void   __midlabel_mix_word_patch_upper_2();
extern void   __midlabel_mix_word_patch_lower_1();
extern void   __midlabel_mix_half_patch_upper_1();
extern void   __midlabel_mix_half_patch_shift_1();
extern void   __midlabel_mix_half_patch_upper_2();
extern void   __midlabel_mix_half_patch_lower_1();
extern void   __midlabel_mix_byte_patch_upper_1();
extern void   __midlabel_mix_byte_patch_shift_1();
extern void   __midlabel_mix_byte_patch_upper_2();
extern void   __midlabel_mix_byte_patch_lower_1();
extern void   __midlabel_mix_double_patch_upper_1();
extern void   __midlabel_mix_double_patch_shift_1();
extern void   __midlabel_mix_double_patch_shift_2();
extern void   __midlabel_mix_double_patch_upper_2();
extern void   __midlabel_mix_double_patch_lower_1();
extern void   __midlabel_mix_area_patch_upper_1();
extern void   __midlabel_mix_area_patch_shift_1();
extern void   __midlabel_mix_area_patch_upper_2();
extern void   __midlabel_mix_area_patch_lower_1();


extern task_t midway_task_self;	/* mach_task_self() */


/*
 * copy a dirtybit template into a new region
 */
void
ipatch_template
  (
   region_info      riP
   )
{
    register kern_return_t krc;
    register int type = riP->type;
    register int line_size = riP->line_size;
    register clock_t *line_info = riP->line_info;
    register vm_offset_t base = riP->base;

    if (IS_DBG_MISC_MAIF)
      printf("ipatch_template: addr 0x%x line_size %d type %s\n",
	     base, RI_LINE_SIZE_TO_BYTES(line_size),
	     (type == REGION_TYPE_PRIVATE) ? "PRIVATE" : 
	     (type == REGION_TYPE_WORD_EC) ? "ALIGNED-WORD-EC" :
	     (type == REGION_TYPE_ALIGNED_EC) ? "ALIGNED-EC" :
	     (type == REGION_TYPE_UNALIGNED_EC) ? "UNALIGNED-EC" :
	     (type == REGION_TYPE_MIXED) ? "MIXED" : "INVALID");


    ASSERT ((base & (REGION_SIZE-1)) == 0);


#ifdef LATER
    /*
     * Make the page writeable
     */
    krc = vm_protect (midway_task_self, base, vm_page_size,
		      (boolean_t) FALSE, VM_PROT_READ+VM_PROT_WRITE);
    if (krc != KERN_SUCCESS)
      panic ("template: could not vm_protect(WRITE) address 0x%x: %s\n",
	     riP->base, mach_error_string(krc));
#endif
    

    /*
     * Copy template to 1st page of region
     * Then patch instructions as needed.
     */
    switch (type) {
      
    case REGION_TYPE_PRIVATE:
      bcopy ((char *)__midway_builtin_unshared_word, 
	     riP->base+MIDWAY_TEMPLATE_WORD_OFFSET,
	     (char *)__midlabel_builtin_unshared_word_end 
	       - (char *)__midway_builtin_unshared_word);
      bcopy ((char *)__midway_builtin_unshared_double, 
	     riP->base+MIDWAY_TEMPLATE_DOUBLE_OFFSET,
	     (char *)__midlabel_builtin_unshared_double_end 
	       - (char *)__midway_builtin_unshared_double);
      bcopy ((char *)__midway_builtin_unshared_half, 
	     riP->base+MIDWAY_TEMPLATE_HALF_OFFSET,
	     (char *)__midlabel_builtin_unshared_half_end 
	       - (char *)__midway_builtin_unshared_half);
      bcopy ((char *)__midway_builtin_unshared_byte, 
	     riP->base+MIDWAY_TEMPLATE_BYTE_OFFSET,
	     (char *)__midlabel_builtin_unshared_byte_end 
	       - (char *)__midway_builtin_unshared_byte);
      bcopy ((char *)__midway_builtin_unshared_area, 
	     riP->base+MIDWAY_TEMPLATE_AREA_OFFSET,
	     (char *)__midlabel_builtin_unshared_area_end 
	       - (char *)__midway_builtin_unshared_area);
      break;

    case REGION_TYPE_WORD_EC:
      bcopy ((char *)__midway_builtin_aligned4_word, 
	     riP->base+MIDWAY_TEMPLATE_WORD_OFFSET,
	     (char *)__midlabel_builtin_aligned4_word_end 
	       - (char *)__midway_builtin_aligned4_word);
      bcopy ((char *)__midway_builtin_aligned4_double, 
	     riP->base+MIDWAY_TEMPLATE_DOUBLE_OFFSET,
	     (char *)__midlabel_builtin_aligned4_double_end 
	       - (char *)__midway_builtin_aligned4_double);
      bcopy ((char *)__midway_builtin_aligned4_half, 
	     riP->base+MIDWAY_TEMPLATE_HALF_OFFSET,
	     (char *)__midlabel_builtin_aligned4_half_end 
	       - (char *)__midway_builtin_aligned4_half);
      bcopy ((char *)__midway_builtin_aligned4_byte, 
	     riP->base+MIDWAY_TEMPLATE_BYTE_OFFSET,
	     (char *)__midlabel_builtin_aligned4_byte_end 
	       - (char *)__midway_builtin_aligned4_byte);
      bcopy ((char *)__midway_builtin_aligned4_area, 
	     riP->base+MIDWAY_TEMPLATE_AREA_OFFSET,
	     (char *)__midlabel_builtin_aligned4_area_end 
	       - (char *)__midway_builtin_aligned4_area);
      
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_WORD_OFFSET +
			     ((char *)__midlabel_al4_word_patch_upper_1
			      - (char *)__midway_builtin_aligned4_word),
			     line_info);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_HALF_OFFSET +
			     ((char *)__midlabel_al4_half_patch_upper_1
			      - (char *)__midway_builtin_aligned4_half),
			     line_info);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_BYTE_OFFSET +
			     ((char *)__midlabel_al4_byte_patch_upper_1
			      - (char *)__midway_builtin_aligned4_byte),
			     line_info);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_DOUBLE_OFFSET +
			     ((char *)__midlabel_al4_double_patch_upper_1
			      - (char *)__midway_builtin_aligned4_double),
			     line_info);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_AREA_OFFSET +
			     ((char *)__midlabel_al4_area_patch_upper_1
			      - (char *)__midway_builtin_aligned4_area),
			     line_info);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_AREA_OFFSET +
			     ((char *)__midlabel_al4_area_patch_upper_2
			      - (char *)__midway_builtin_aligned4_area),
			     riP);
      PATCH_INSTR_LOADLOWER (base + MIDWAY_TEMPLATE_AREA_OFFSET +
			     ((char *)__midlabel_al4_area_patch_lower_1
			      - (char *)__midway_builtin_aligned4_area),
			     riP);
      break;

    case REGION_TYPE_ALIGNED_EC:
      bcopy ((char *)__midway_builtin_aligned_word, 
	     base+MIDWAY_TEMPLATE_WORD_OFFSET,
	     (char *)__midlabel_builtin_aligned_word_end 
	       - (char *)__midway_builtin_aligned_word);
      if (line_size == 3)
	bcopy ((char *)__midway_builtin_aligned8_double, 
	       base+MIDWAY_TEMPLATE_DOUBLE_OFFSET,
	       (char *)__midlabel_builtin_aligned8_double_end 
	       - (char *)__midway_builtin_aligned8_double);
      else
	bcopy ((char *)__midway_builtin_aligned_double, 
	       base+MIDWAY_TEMPLATE_DOUBLE_OFFSET,
	       (char *)__midlabel_builtin_aligned_double_end 
	       - (char *)__midway_builtin_aligned_double);
      bcopy ((char *)__midway_builtin_aligned_half, 
	     base+MIDWAY_TEMPLATE_HALF_OFFSET,
	     (char *)__midlabel_builtin_aligned_half_end 
	       - (char *)__midway_builtin_aligned_half);
      bcopy ((char *)__midway_builtin_aligned_byte, 
	     base+MIDWAY_TEMPLATE_BYTE_OFFSET,
	     (char *)__midlabel_builtin_aligned_byte_end 
	       - (char *)__midway_builtin_aligned_byte);
      bcopy ((char *)__midway_builtin_aligned_area, 
	     base+MIDWAY_TEMPLATE_AREA_OFFSET,
	     (char *)__midlabel_builtin_aligned_area_end 
	       - (char *)__midway_builtin_aligned_area); 

      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_WORD_OFFSET +
			     ((char *)__midlabel_al_word_patch_upper_1
			      - (char *)__midway_builtin_aligned_word),
			     line_info);
      PATCH_INSTR_SHIFTRIGHT(base + MIDWAY_TEMPLATE_WORD_OFFSET +
			     ((char *)__midlabel_al_word_patch_shift_1
			      - (char *)__midway_builtin_aligned_word),
			     line_size);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_HALF_OFFSET +
			     ((char *)__midlabel_al_half_patch_upper_1
			      - (char *)__midway_builtin_aligned_half),
			     line_info);
      PATCH_INSTR_SHIFTRIGHT(base + MIDWAY_TEMPLATE_HALF_OFFSET +
			     ((char *)__midlabel_al_half_patch_shift_1
			      - (char *)__midway_builtin_aligned_half),
			     line_size);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_BYTE_OFFSET +
			     ((char *)__midlabel_al_byte_patch_upper_1
			      - (char *)__midway_builtin_aligned_byte),
			     line_info);
      PATCH_INSTR_SHIFTRIGHT(base + MIDWAY_TEMPLATE_BYTE_OFFSET +
			     ((char *)__midlabel_al_byte_patch_shift_1
			      - (char *)__midway_builtin_aligned_byte),
			     line_size);
      if (line_size == 3) {
	PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_DOUBLE_OFFSET +
			       ((char *)__midlabel_al8_double_patch_upper_1
				- (char *)__midway_builtin_aligned8_double),
			       line_info);
      } else {
	PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_DOUBLE_OFFSET +
			       ((char *)__midlabel_al_double_patch_upper_1
				- (char *)__midway_builtin_aligned_double),
			       line_info);
	PATCH_INSTR_SHIFTRIGHT(base + MIDWAY_TEMPLATE_DOUBLE_OFFSET +
			       ((char *)__midlabel_al_double_patch_shift_1
				- (char *)__midway_builtin_aligned_double),
			       line_size);
      }
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_AREA_OFFSET +
			     ((char *)__midlabel_al_area_patch_upper_1
			      - (char *)__midway_builtin_aligned_area),
			     line_info);
      PATCH_INSTR_SHIFTRIGHT(base + MIDWAY_TEMPLATE_AREA_OFFSET +
			     ((char *)__midlabel_al_area_patch_shift_1
			      - (char *)__midway_builtin_aligned_area),
			     line_size);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_AREA_OFFSET +
			     ((char *)__midlabel_al_area_patch_upper_2
			      - (char *)__midway_builtin_aligned_area),
			     riP);
      PATCH_INSTR_LOADLOWER (base + MIDWAY_TEMPLATE_AREA_OFFSET +
			     ((char *)__midlabel_al_area_patch_lower_1
			      - (char *)__midway_builtin_aligned_area),
			     riP);
      break;

    case REGION_TYPE_UNALIGNED_EC:
      bcopy ((char *)__midway_builtin_unaligned_word, 
	     base+MIDWAY_TEMPLATE_WORD_OFFSET,
	     (char *)__midlabel_builtin_unaligned_word_end 
	       - (char *)__midway_builtin_unaligned_word);
      bcopy ((char *)__midway_builtin_unaligned_double, 
	     base+MIDWAY_TEMPLATE_DOUBLE_OFFSET,
	     (char *)__midlabel_builtin_unaligned_double_end 
	       - (char *)__midway_builtin_unaligned_double);
      bcopy ((char *)__midway_builtin_unaligned_half, 
	     base+MIDWAY_TEMPLATE_HALF_OFFSET,
	     (char *)__midlabel_builtin_unaligned_half_end 
	       - (char *)__midway_builtin_unaligned_half);
      bcopy ((char *)__midway_builtin_unaligned_byte, 
	     base+MIDWAY_TEMPLATE_BYTE_OFFSET,
	     (char *)__midlabel_builtin_unaligned_byte_end 
	       - (char *)__midway_builtin_unaligned_byte);
      bcopy ((char *)__midway_builtin_unaligned_area, 
	     base+MIDWAY_TEMPLATE_AREA_OFFSET,
	     (char *)__midlabel_builtin_unaligned_area_end 
	       - (char *)__midway_builtin_unaligned_area); 

      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_WORD_OFFSET +
			     ((char *)__midlabel_un_word_patch_upper_1
			      - (char *)__midway_builtin_unaligned_word),
			     line_info);
      PATCH_INSTR_SHIFTRIGHT(base + MIDWAY_TEMPLATE_WORD_OFFSET +
			     ((char *)__midlabel_un_word_patch_shift_1
			      - (char *)__midway_builtin_unaligned_word),
			     line_size);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_WORD_OFFSET +
			     ((char *)__midlabel_un_word_patch_upper_2
			      - (char *)__midway_builtin_unaligned_word),
			     riP);
      PATCH_INSTR_LOADLOWER (base + MIDWAY_TEMPLATE_WORD_OFFSET +
			     ((char *)__midlabel_un_word_patch_lower_1
			      - (char *)__midway_builtin_unaligned_word),
			     riP);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_HALF_OFFSET +
			     ((char *)__midlabel_un_half_patch_upper_1
			      - (char *)__midway_builtin_unaligned_half),
			     line_info);
      PATCH_INSTR_SHIFTRIGHT(base + MIDWAY_TEMPLATE_HALF_OFFSET +
			     ((char *)__midlabel_un_half_patch_shift_1
			      - (char *)__midway_builtin_unaligned_half),
			     line_size);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_HALF_OFFSET +
			     ((char *)__midlabel_un_half_patch_upper_2
			      - (char *)__midway_builtin_unaligned_half),
			     riP);
      PATCH_INSTR_LOADLOWER (base + MIDWAY_TEMPLATE_HALF_OFFSET +
			     ((char *)__midlabel_un_half_patch_lower_1
			      - (char *)__midway_builtin_unaligned_half),
			     riP);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_BYTE_OFFSET +
			     ((char *)__midlabel_un_byte_patch_upper_1
			      - (char *)__midway_builtin_unaligned_byte),
			     line_info);
      PATCH_INSTR_SHIFTRIGHT(base + MIDWAY_TEMPLATE_BYTE_OFFSET +
			     ((char *)__midlabel_un_byte_patch_shift_1
			      - (char *)__midway_builtin_unaligned_byte),
			     line_size);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_BYTE_OFFSET +
			     ((char *)__midlabel_un_byte_patch_upper_2
			      - (char *)__midway_builtin_unaligned_byte),
			     riP);
      PATCH_INSTR_LOADLOWER (base + MIDWAY_TEMPLATE_BYTE_OFFSET +
			     ((char *)__midlabel_un_byte_patch_lower_1
			      - (char *)__midway_builtin_unaligned_byte),
			     riP);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_DOUBLE_OFFSET +
			     ((char *)__midlabel_un_double_patch_upper_1
			      - (char *)__midway_builtin_unaligned_double),
			     line_info);
      PATCH_INSTR_SHIFTRIGHT(base + MIDWAY_TEMPLATE_DOUBLE_OFFSET +
			     ((char *)__midlabel_un_double_patch_shift_1
			      - (char *)__midway_builtin_unaligned_double),
			     line_size);
      PATCH_INSTR_SHIFTRIGHT(base + MIDWAY_TEMPLATE_DOUBLE_OFFSET +
			     ((char *)__midlabel_un_double_patch_shift_2
			      - (char *)__midway_builtin_unaligned_double),
			     line_size);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_DOUBLE_OFFSET +
			     ((char *)__midlabel_un_double_patch_upper_2
			      - (char *)__midway_builtin_unaligned_double),
			     riP);
      PATCH_INSTR_LOADLOWER (base + MIDWAY_TEMPLATE_DOUBLE_OFFSET +
			     ((char *)__midlabel_un_double_patch_lower_1
			      - (char *)__midway_builtin_unaligned_double),
			     riP);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_AREA_OFFSET +
			     ((char *)__midlabel_un_area_patch_upper_1
			      - (char *)__midway_builtin_unaligned_area),
			     line_info);
      PATCH_INSTR_SHIFTRIGHT(base + MIDWAY_TEMPLATE_AREA_OFFSET +
			     ((char *)__midlabel_un_area_patch_shift_1
			      - (char *)__midway_builtin_unaligned_area),
			     line_size);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_AREA_OFFSET +
			     ((char *)__midlabel_un_area_patch_upper_2
			      - (char *)__midway_builtin_unaligned_area),
			     riP);
      PATCH_INSTR_LOADLOWER (base + MIDWAY_TEMPLATE_AREA_OFFSET +
			     ((char *)__midlabel_un_area_patch_lower_1
			      - (char *)__midway_builtin_unaligned_area),
			     riP);
      break;

    case REGION_TYPE_MIXED:
      bcopy ((char *)__midway_builtin_mixed_word, 
	     base+MIDWAY_TEMPLATE_WORD_OFFSET,
	     (char *)__midlabel_builtin_mixed_word_end 
	       - (char *)__midway_builtin_mixed_word);
      bcopy ((char *)__midway_builtin_mixed_double, 
	     base+MIDWAY_TEMPLATE_DOUBLE_OFFSET,
	     (char *)__midlabel_builtin_mixed_double_end 
	       - (char *)__midway_builtin_mixed_double);
      bcopy ((char *)__midway_builtin_mixed_half, 
	     base+MIDWAY_TEMPLATE_HALF_OFFSET,
	     (char *)__midlabel_builtin_mixed_half_end 
	       - (char *)__midway_builtin_mixed_half);
      bcopy ((char *)__midway_builtin_mixed_byte, 
	     base+MIDWAY_TEMPLATE_BYTE_OFFSET,
	     (char *)__midlabel_builtin_mixed_byte_end 
	       - (char *)__midway_builtin_mixed_byte);
      bcopy ((char *)__midway_builtin_mixed_area, 
	     base+MIDWAY_TEMPLATE_AREA_OFFSET,
	     (char *)__midlabel_builtin_mixed_area_end 
	       - (char *)__midway_builtin_mixed_area); 

      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_WORD_OFFSET +
			     ((char *)__midlabel_mix_word_patch_upper_1
			      - (char *)__midway_builtin_mixed_word),
			     line_info);
      PATCH_INSTR_SHIFTRIGHT(base + MIDWAY_TEMPLATE_WORD_OFFSET +
			     ((char *)__midlabel_mix_word_patch_shift_1
			      - (char *)__midway_builtin_mixed_word),
			     line_size);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_WORD_OFFSET +
			     ((char *)__midlabel_mix_word_patch_upper_2
			      - (char *)__midway_builtin_mixed_word),
			     riP);
      PATCH_INSTR_LOADLOWER (base + MIDWAY_TEMPLATE_WORD_OFFSET +
			     ((char *)__midlabel_mix_word_patch_lower_1
			      - (char *)__midway_builtin_mixed_word),
			     riP);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_HALF_OFFSET +
			     ((char *)__midlabel_mix_half_patch_upper_1
			      - (char *)__midway_builtin_mixed_half),
			     line_info);
      PATCH_INSTR_SHIFTRIGHT(base + MIDWAY_TEMPLATE_HALF_OFFSET +
			     ((char *)__midlabel_mix_half_patch_shift_1
			      - (char *)__midway_builtin_mixed_half),
			     line_size);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_HALF_OFFSET +
			     ((char *)__midlabel_mix_half_patch_upper_2
			      - (char *)__midway_builtin_mixed_half),
			     riP);
      PATCH_INSTR_LOADLOWER (base + MIDWAY_TEMPLATE_HALF_OFFSET +
			     ((char *)__midlabel_mix_half_patch_lower_1
			      - (char *)__midway_builtin_mixed_half),
			     riP);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_BYTE_OFFSET +
			     ((char *)__midlabel_mix_byte_patch_upper_1
			      - (char *)__midway_builtin_mixed_byte),
			     line_info);
      PATCH_INSTR_SHIFTRIGHT(base + MIDWAY_TEMPLATE_BYTE_OFFSET +
			     ((char *)__midlabel_mix_byte_patch_shift_1
			      - (char *)__midway_builtin_mixed_byte),
			     line_size);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_BYTE_OFFSET +
			     ((char *)__midlabel_mix_byte_patch_upper_2
			      - (char *)__midway_builtin_mixed_byte),
			     riP);
      PATCH_INSTR_LOADLOWER (base + MIDWAY_TEMPLATE_BYTE_OFFSET +
			     ((char *)__midlabel_mix_byte_patch_lower_1
			      - (char *)__midway_builtin_mixed_byte),
			     riP);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_DOUBLE_OFFSET +
			     ((char *)__midlabel_mix_double_patch_upper_1
			      - (char *)__midway_builtin_mixed_double),
			     line_info);
      PATCH_INSTR_SHIFTRIGHT(base + MIDWAY_TEMPLATE_DOUBLE_OFFSET +
			     ((char *)__midlabel_mix_double_patch_shift_1
			      - (char *)__midway_builtin_mixed_double),
			     line_size);
      PATCH_INSTR_SHIFTRIGHT(base + MIDWAY_TEMPLATE_DOUBLE_OFFSET +
			     ((char *)__midlabel_mix_double_patch_shift_2
			      - (char *)__midway_builtin_mixed_double),
			     line_size);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_DOUBLE_OFFSET +
			     ((char *)__midlabel_mix_double_patch_upper_2
			      - (char *)__midway_builtin_mixed_double),
			     riP);
      PATCH_INSTR_LOADLOWER (base + MIDWAY_TEMPLATE_DOUBLE_OFFSET +
			     ((char *)__midlabel_mix_double_patch_lower_1
			      - (char *)__midway_builtin_mixed_double),
			     riP);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_AREA_OFFSET +
			     ((char *)__midlabel_mix_area_patch_upper_1
			      - (char *)__midway_builtin_mixed_area),
			     line_info);
      PATCH_INSTR_SHIFTRIGHT(base + MIDWAY_TEMPLATE_AREA_OFFSET +
			     ((char *)__midlabel_mix_area_patch_shift_1
			      - (char *)__midway_builtin_mixed_area),
			     line_size);
      PATCH_INSTR_LOADUPPER (base + MIDWAY_TEMPLATE_AREA_OFFSET +
			     ((char *)__midlabel_mix_area_patch_upper_2
			      - (char *)__midway_builtin_mixed_area),
			     riP);
      PATCH_INSTR_LOADLOWER (base + MIDWAY_TEMPLATE_AREA_OFFSET +
			     ((char *)__midlabel_mix_area_patch_lower_1
			      - (char *)__midway_builtin_mixed_area),
			     riP);
      break;
    default:
      panic("invalid template type");
    } /* switch */

#ifdef LATER
    /*
     * Protect the page
     */
    krc = vm_protect (midway_task_self, base, vm_page_size,
		      (boolean_t) FALSE, VM_PROT_EXECUTE);
    if (krc != KERN_SUCCESS)
      panic ("template: could not vm_protect(EXECUTE) address 0x%x: %s\n",
	     base, mach_error_string(krc));
#endif
  
}
