/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * midway_start.s
 *
 * This file contains the system dependent code for allocating a dirtybit template
 * in the first page of the static data. On the MIPS, the static data starts
 * with the read-only data section ("rdata"). This file must be the first object 
 * file included by all midway applications.
 *
 *  12Nov93 (was): created.
 *  22Feb94 (was): added globals for start of bss, sbss, data and sdata
 */

#include <mips/asm.h>
#ifndef ASSEMBLER
#define ASSEMBLER 1 /* for midway_comp.h */ /*XXX 14feb94 mjz */
#endif
#include "../midway_comp.h"

#define PAGE_BASED 1	/* standard gcc doesn't pass command line -D's to asm */


	.rdata
	.align	2

/*
 * Define variable to reside at base of static region.
 * The variable is modified during midway initialization
 * to contain a dirtybit template for setting dirtybits
 * for shared data within the static region.
 *
 * The variable is given an initial value to allow the midway
 * initialization to verify that it is the first variable in the
 * that static region.
 */
	.globl	midway_static_template
midway_static_template:
	.ascii	MIDWAY_TEMPLATE_VALUE
	.space	MIDWAY_TEMPLATE_ZERO_SIZE

#if defined(PAGE_BASED)

/*
 * Define variables at the start of each of the data regions
 * mips has 4: 
 *              .data     - Initialized statics & externs
 *              .sdata    - "small" data 
 *              .bss      - Uninitialized statics & externs
 *              .sbss     - "small" bss
 *
 * Unfortunately, the mips assembler splits the bss & sbss 
 * into internal and external symbols, giving 6 regions.
 *
 * Allocate a full page at each boundary edge
 * to make twin/diffing easier.  WAS 24Feb94
 *
 * Assembler uses values of -G flag to determine if a variable
 * goes into "small" or "regular" area. Data which is smaller
 * than the -G value goes into "small". This code depends on
 * -G being set to 4095 in the command line.  WAS 24Feb94
 *
 * It also doesn't assign "common" data to the bss or sbss
 * until the final relinking. Thus all data explicitly assigned
 * to the bss/sbss is assigned addresses, before any common data.
 * Furthermore, the -G trick above doesn't work on the linker.
 * Anything larger than 8 bytes goes to the bss, 8 bytes or less
 * to the sbss.  WAS 25Feb94
 *
 * No longer allocate a full page, since I couldn't get the xpager
 * to give my faults on the static region. Make all the tags a known
 * size to allow midway to combine adjacent regions. WAS 25Feb94
 * 
 * Added tags to determine size of user text segment.  WAS 25May94
 *
 */
#define MIPS_TAG_SIZE  16	/* MUST match _end.s & vm_paged.c */
#define MIPS_STAG_SIZE  4

/* define start of text region */
	.text
	.align 2
	.ent	midway_start_text_data,0
midway_start_text_data:
	.frame	sp,0,ra
	j	ra		# size = 4 bytes
	.end 	midway_start_text_data

/* define start of bss region */
	.globl	midway_start_bss_data
	.lcomm	midway_start_bss_data, MIPS_TAG_SIZE

/* define start of sbss region */
	.globl	midway_start_sbss_data
	.lcomm	midway_start_sbss_data, MIPS_STAG_SIZE

/* define variables in static regions */
	.lcomm	midway_start_static_bss_data, MIPS_TAG_SIZE
	.lcomm	midway_start_static_sbss_data, MIPS_STAG_SIZE

/* define variables in common regions */
	.comm	midway_start_comm_data, MIPS_TAG_SIZE
	.comm	midway_start_scomm_data, MIPS_STAG_SIZE

/* define start of data region */
	.data
	.align	2
	.globl midway_start_data_data
midway_start_data_data:
	.space	MIPS_TAG_SIZE

/* define start of small data region */
	.sdata
	.align	2

/* 
 * Declare global pointers to start of each region
 * (1st address past boundary tag)
 */
	.globl	midway_start_text
midway_start_text:
	.word	midway_start_text_data+4

	.globl	midway_start_data
midway_start_data:
	.word	midway_start_data_data+MIPS_TAG_SIZE

	.globl	midway_start_sdata
midway_start_sdata:
	.word 	midway_start_sdata_data+MIPS_STAG_SIZE

	.globl midway_start_bss
midway_start_bss:
	.word	midway_start_bss_data+MIPS_TAG_SIZE

	.globl midway_start_sbss
midway_start_sbss:
	.word	midway_start_sbss_data+MIPS_STAG_SIZE

	.globl midway_start_static_bss
midway_start_static_bss:
	.word	midway_start_static_bss_data+MIPS_TAG_SIZE

	.globl midway_start_static_sbss
midway_start_static_sbss:
	.word	midway_start_static_sbss_data+MIPS_STAG_SIZE

	.globl	midway_start_comm
midway_start_comm:
	.word	midway_start_comm_data+MIPS_TAG_SIZE

	.globl	midway_start_scomm
midway_start_scomm:
	.word	midway_start_scomm_data+MIPS_STAG_SIZE

/* Declare flag for start of application's sdata */	
	.globl midway_start_sdata_data
midway_start_sdata_data:
	.space	MIPS_STAG_SIZE

#endif /* PAGE_BASED */



