/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * template.s
 *
 * This file contains the "template" code stored at the base of each region
 * which sets the dirty bits for data within the region. There are different
 * templates for different types of memory. 
 *
 *    unshared
 *    aligned 4-byte cache line ec
 *    aligned ec
 *    unaligned ec
 *    mixed model
 *
 * Each template has an entry point for each of the following compiler generated calls.
 * The entry points begin at a known location within the template.  The entry points are:
 *
 *    __midway_builtin_shared_word
 *    __midway_builtin_shared_double
 *    __midway_builtin_shared_byte
 *    __midway_builtin_shared_half
 *    __midway_builtin_shared_area
 *
 *  The compiler generates the call to each entry point as follows:
 *
 *  To call the first entry point (_shared_word)
 *
 *    st   *, 0(rx)			; original store instruction
 *    lui  at, template_mask            ; mask off lower bits of address
 *    and  at, at, rx			; generate address for template 
 *    jalr at				; jump and link
 *    sub  a0, rx, at 			; put address offset in arg register
 *
 *  To call the other entry points (rx != a0)
 *
 *    st   *, 0(rx)			; original store instruction
 *    lui  at, template_mask            ; mask off lower bits of address
 *    and  a0, at, rx			; generate address for template
 *    or   at, a0, entry_pointX_mask	; generate address for entry point w/in template
 *    jalr at				; jump and link
 *    sub  a0, rx, a0 			; put address offset in arg register
 *
 *
 *
 * Regions are layed out as follows:
 *
 *    +----------------------+
 *    |  template            |
 *    +----------------------+
 *    |                      |
 *    |  data                |
 *    |                      |
 *    |                      |
 *    +----------------------+
 *    |                      |
 *    |  dirty bits          |
 *    |                      |
 *    |                      |
 *    |                      |
 *    +----------------------+
 *
 * Summer 1993, (WAS)
 *
 *  1Nov93 (was): Only clobber registers a0, a1, v0 and AT.
 * 17Feb94 (was): Aligned_double now assumes both words in same cache line
 *                Also added special case for aligned 8byte double
 */

#include <mips/asm.h>

#define FAST_ENQUEUE 		/* define to handle updateq_add inline */
				/* undef to call C routine to perform */


#define PATCH_DBIT_ADDR    0xFFFF
#define PATCH_LINE_SHIFT   0x001F
#define PATCH_UPPER_REGION 0xFFFF
#define PATCH_LOWER_REGION 0xFFFF




/*
 * gcc assumes that these routines only clobber registers at, v0, a0 and a1
 * If we call up to 'C' code, we must save all temp registers that are not
 * saved by the 'C' code.     WAS 2 Nov 93
 */

#define DECLARE_TEMP_REGS								 \
	.mask (M_RA|M_T9|M_T8|M_T7|M_T6|M_T5|M_T4|M_T3|M_T2|M_T1|M_T0|M_A3|M_A2|M_V1),-4 \
	.frame	sp,80,ra	# 16 saved regs, 4 arguments

#define SAVE_TEMP_REGS			\
	subu	sp,80;			\
	sw	ra,76(sp);		\
	sw	t9,72(sp);		\
	sw	t8,68(sp);		\
	sw	t7,64(sp);		\
	sw	t6,60(sp);		\
	sw	t5,56(sp);		\
	sw	t4,52(sp);		\
	sw	t3,48(sp);		\
	sw	t2,44(sp);		\
	sw	t1,40(sp);		\
	sw	t0,36(sp);		\
	sw	a3,32(sp);		\
	sw	a2,28(sp);		\
	sw	v1,24(sp)

#define I_RESTORE_TEMP_REGS		\
	lw	ra, 76(sp);		\
	lw	v1, 24(sp);		\
	lw	a2, 28(sp);		\
	lw	a3, 32(sp);		\
	lw	t0, 36(sp);		\
	lw	t1, 40(sp);		\
	lw	t2, 44(sp);		\
	lw	t3, 48(sp);		\
	lw	t4, 52(sp);		\
	lw	t5, 56(sp);		\
	lw	t6, 60(sp);		\
	lw	t7, 64(sp);		\
	lw	t8, 68(sp);		\
	lw	t9, 72(sp)

#define RESTORE_TEMP_REGS		\
	I_RESTORE_TEMP_REGS;		\
	addu	sp,80

#define RESTORE_TEMP_REGS_AND_RETURN	\
	I_RESTORE_TEMP_REGS;		\
	j	ra;			\
	addu	sp,80


/*
 * Save a few more caller-saved registers when calling gprof
 */
#define SAVE_USER_REGS			\
	sw	a1,20(sp);		\
	sw	a0,16(sp)

#define RESTORE_USER_REGS		\
	lw	a1,20(sp);		\
	lw	a0,16(sp)

/*
 * Define macros for profiling
 * Address used for gprof is template's original address
 * not the shared region address.
 */
#ifdef PROFILE
#  define IF_PROFILE2(entrypoint)	\
	SAVE_USER_REGS;			\
					\
	.set	macro;			\
	la    a1, entrypoint;		\
	la    v0, mcountaux;		\
	move  a0, ra;			\
	jal   v0;			\
	addi  a1, a1, 4;  		\
	.set	nomacro;		\
					\
	RESTORE_USER_REGS;		\
	move	zero, zero

#  define IF_PROFILE(entrypoint)	\
	SAVE_TEMP_REGS;			\
	IF_PROFILE2(entrypoint);	\
	RESTORE_TEMP_REGS

#else
#  define IF_PROFILE(entrypoint)
#  define IF_PROFILE2(entrypoint)
#endif




	.text
	.align	2
	.set	noreorder
	.set	nomacro
	.set	noat


/************************************************************************
*									*
*	Template for UNSHARED memory 					*
*									*
************************************************************************/

LEAF(__midlabel_builtin_unshared_start)

XLEAF(__midway_builtin_unshared_word)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_unshared_word)
	j 	ra			# nothing to do - just return
	nop

EXPORT(__midlabel_builtin_unshared_word_end)

XLEAF(__midway_builtin_unshared_half)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_unshared_half)
	j 	ra			# nothing to do - just return
	nop

EXPORT(__midlabel_builtin_unshared_half_end)

XLEAF(__midway_builtin_unshared_byte)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_unshared_byte)
	j 	ra			# nothing to do - just return
	nop

EXPORT(__midlabel_builtin_unshared_byte_end)

XLEAF(__midway_builtin_unshared_double)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_unshared_double)
	j 	ra			# nothing to do - just return
	nop

EXPORT(__midlabel_builtin_unshared_double_end)

XLEAF(__midway_builtin_unshared_area)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_unshared_area)
	j 	ra			# nothing to do - just return
	nop

EXPORT(__midlabel_builtin_unshared_area_end)



END(__midlabel_builtin_unshared_start)



/************************************************************************
*									*
*	Template for ALIGNED EC memory for cache line = 4 bytes 	*
*									*
************************************************************************/

LEAF(__midlabel_builtin_aligned4_start)

XLEAF(__midway_builtin_aligned4_word)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_aligned4_word)

	EXPORT(__midlabel_al4_word_patch_upper_1)
	lui	AT, PATCH_DBIT_ADDR	# load address of start of dbits for this region
	addu    AT, a0, AT		# address of dbit
        j 	ra			# and return
	sw	zero, 0(AT)		# zero dirty bit and return
EXPORT(__midlabel_builtin_aligned4_word_end)


XLEAF(__midway_builtin_aligned4_half)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_aligned4_half)

	EXPORT(__midlabel_al4_half_patch_upper_1)
	lui	AT, PATCH_DBIT_ADDR	# load address of start of dbits for this region
	addu    AT, a0, AT		# address of dbit
	srl	AT, AT, 2		# zero low order 2 bits to get a word boundary
	sll     AT, AT, 2		# 
        j 	ra			# and return
	sw	zero, 0(AT)		# zero dirty bit and return
EXPORT(__midlabel_builtin_aligned4_half_end)


XLEAF(__midway_builtin_aligned4_byte)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_aligned4_byte)

	EXPORT(__midlabel_al4_byte_patch_upper_1)
	lui	AT, PATCH_DBIT_ADDR	# load address of start of dbits for this region
	addu    AT, a0, AT		# address of dbit
	srl	AT, AT, 2		# zero low order 2 bits to get a word boundary
	sll     AT, AT, 2		# 
        j 	ra			# and return
	sw	zero, 0(AT)		# zero dirty bit and return
EXPORT(__midlabel_builtin_aligned4_byte_end)


XLEAF(__midway_builtin_aligned4_double)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_aligned4_double)

	EXPORT(__midlabel_al4_double_patch_upper_1)
	lui	AT, PATCH_DBIT_ADDR	# load address of start of dbits for this region
	addu    AT, a0, AT		# address of 1st dbit
	sw	zero, 0(AT)		# zero 1st dbit  
        j 	ra			# and return
	sw	zero, 4(AT)		# zero 2nd dirty bit and return
EXPORT(__midlabel_builtin_aligned4_double_end)


XLEAF(__midway_builtin_aligned4_area)

	DECLARE_TEMP_REGS
	SAVE_TEMP_REGS
	IF_PROFILE2(__midway_builtin_aligned4_area)

	EXPORT(__midlabel_al4_area_patch_upper_1)
	lui	AT, PATCH_DBIT_ADDR	# load address of start of dbits for this region
	addu    a2, a0, AT		# address of dbit
	srl	a2, a2, 2		# zero low order 2 bits to get a word boundary
	sll     a2, a2, 2		# 

	EXPORT(__midlabel_al4_area_patch_upper_2)
	lui	a3, PATCH_UPPER_REGION
	.set	macro
	la	v0, mid_line_exception  # MUST load addr into reg for jump > 28bits
	.set	nomacro
	jal	v0                	# mid_line_exc(offset,len,dbitP,riP)
	EXPORT(__midlabel_al4_area_patch_lower_1)
	ori	a3, a3, PATCH_LOWER_REGION

	RESTORE_TEMP_REGS_AND_RETURN

EXPORT(__midlabel_builtin_aligned4_area_end)


END(__midlabel_builtin_aligned4_start)



/************************************************************************
*									*
*	Template for ALIGNED EC memory for cache line != 4 bytes 	*
*									*
************************************************************************/

LEAF(__midlabel_builtin_aligned_start)

XLEAF(__midway_builtin_aligned_word)
XLEAF(__midway_builtin_aligned_half)
XLEAF(__midway_builtin_aligned_byte)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_aligned_word)

	EXPORT(__midlabel_al_word_patch_upper_1)
	EXPORT(__midlabel_al_half_patch_upper_1)
	EXPORT(__midlabel_al_byte_patch_upper_1)
	lui	AT, PATCH_DBIT_ADDR	# load address of start of dbits for this region
	EXPORT(__midlabel_al_word_patch_shift_1)
	EXPORT(__midlabel_al_half_patch_shift_1)
	EXPORT(__midlabel_al_byte_patch_shift_1)
	srl	a1, a0, PATCH_LINE_SHIFT	# shift offset by cache line size
	sll     a1, a1, 2		# times 4 to get dbit offset in region
	addu    AT, a1, AT		# address of dbit
        j 	ra			# and return
	sw	zero, 0(AT)		# zero dirty bit and return

EXPORT(__midlabel_builtin_aligned_word_end)
EXPORT(__midlabel_builtin_aligned_half_end)
EXPORT(__midlabel_builtin_aligned_byte_end)


XLEAF(__midway_builtin_aligned_double)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_aligned_double)

	/*
	 * Code now assume both words are in same cache line. WAS 17Feb94
	 */
	EXPORT(__midlabel_al_double_patch_upper_1)
	lui	AT, PATCH_DBIT_ADDR		# load address of start of dbits for this region
	EXPORT(__midlabel_al_double_patch_shift_1)
	srl	a1, a0, PATCH_LINE_SHIFT	# shift offset by cache line size
	sll     a1, a1, 2			# times 4 to get dbit offset in region
	addu	AT, a1, AT			# generate address of dbit
        j 	ra			# and return
	sw	zero, 0(AT)		# zero dbit on way out

EXPORT(__midlabel_builtin_aligned_double_end)


XLEAF(__midway_builtin_aligned8_double)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_aligned8_double)

	EXPORT(__midlabel_al8_double_patch_upper_1)
	lui	AT, PATCH_DBIT_ADDR		# load address of start of dbits for this region
	srl	a1, a0, 1			# divide by 2 to get dbit offset
	addu	AT, a1, AT			# generate address of dbit
        j 	ra			# and return
	sw	zero, 0(AT)		# zero dbit on way out

EXPORT(__midlabel_builtin_aligned8_double_end)


XLEAF(__midway_builtin_aligned_area)

	DECLARE_TEMP_REGS
	SAVE_TEMP_REGS
	IF_PROFILE2(__midway_builtin_aligned_area)

	EXPORT(__midlabel_al_area_patch_upper_1)
	lui	AT, PATCH_DBIT_ADDR	# load address of start of dbits for this region
	EXPORT(__midlabel_al_area_patch_shift_1)
	srl	a2, a0, PATCH_LINE_SHIFT	# shift offset by cache line size
	sll     a2, a2, 2		# times 4 to get dbit offset in region
	addu    a2, a2, AT		# address of dbit

	EXPORT(__midlabel_al_area_patch_upper_2)
	lui	a3, PATCH_UPPER_REGION
	.set	macro
	la	v0, mid_line_exception  # MUST load addr into reg for jump > 28bits
	.set	nomacro
	jal	v0                	# mid_line_exc(offset,len,dbitP,riP)
	EXPORT(__midlabel_al_area_patch_lower_1)
	ori	a3, a3, PATCH_LOWER_REGION

	RESTORE_TEMP_REGS_AND_RETURN
	
EXPORT(__midlabel_builtin_aligned_area_end)

END(__midlabel_builtin_aligned_start)





/************************************************************************
*									*
*	Template for UNALIGNED EC memory				*
*									*
************************************************************************/

LEAF(__midlabel_builtin_unaligned_start)

XLEAF(__midway_builtin_unaligned_word)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_unaligned_word)

	EXPORT(__midlabel_un_word_patch_upper_1)
	lui	AT, PATCH_DBIT_ADDR	# load address of start of dbits for this region
	EXPORT(__midlabel_un_word_patch_shift_1)
	srl	a1, a0, PATCH_LINE_SHIFT	# shift offset by cache line size
	sll     a1, a1, 2		# times 4 to get dbit offset in region
	addu    AT, a1, AT		# address of dbit
	lw	v0, 0(AT)		# read dbit
	addi	a1, zero, 4		# delay slot: set length arg to mid_line_exception
	bltz    v0,$L4000		# branch if hi order bit is 1
	nop				# unused delay slot
	
        j 	ra			# and return
	sw	zero, 0(AT)		# zero dirty bit and return

$L4000:
	SAVE_TEMP_REGS

	move	a2, AT
	EXPORT(__midlabel_un_word_patch_upper_2)
	lui	a3, PATCH_UPPER_REGION	#  compute region_info arg
	.set	macro
	la	v0, mid_line_exception  # MUST load addr into reg for jump > 28bits
	.set	nomacro
	jal	v0                	# mid_line_exc(offset,len,dbitP,riP)
	EXPORT(__midlabel_un_word_patch_lower_1)
	ori	a3, a3, PATCH_LOWER_REGION

	RESTORE_TEMP_REGS_AND_RETURN

EXPORT(__midlabel_builtin_unaligned_word_end)


XLEAF(__midway_builtin_unaligned_half)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_unaligned_half)

	EXPORT(__midlabel_un_half_patch_upper_1)
	lui	AT, PATCH_DBIT_ADDR	# load address of start of dbits for this region
	EXPORT(__midlabel_un_half_patch_shift_1)
	srl	a1, a0, PATCH_LINE_SHIFT	# shift offset by cache line size
	sll     a1, a1, 2		# times 4 to get dbit offset in region
	addu    AT, a1, AT		# address of dbit
	lw	v0, 0(AT)		# read dbit
	addi	a1, zero, 2		# delay slot: set length arg to mid_line_exception
	bltz    v0,$L4100		# branch if hi order bit is 1
	nop				# unused delay slot
	
        j 	ra			# and return
	sw	zero, 0(AT)		# zero dirty bit and return

$L4100:
	SAVE_TEMP_REGS

	move	a2, AT
	EXPORT(__midlabel_un_half_patch_upper_2)
	lui	a3, PATCH_UPPER_REGION	#  compute region_info arg
	.set	macro
	la	v0, mid_line_exception  # MUST load addr into reg for jump > 28bits
	.set	nomacro
	jal	v0                	# mid_line_exc(offset,len,dbitP,riP)
	EXPORT(__midlabel_un_half_patch_lower_1)
	ori	a3, a3, PATCH_LOWER_REGION

	RESTORE_TEMP_REGS_AND_RETURN

EXPORT(__midlabel_builtin_unaligned_half_end)


XLEAF(__midway_builtin_unaligned_byte)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_unaligned_byte)

	EXPORT(__midlabel_un_byte_patch_upper_1)
	lui	AT, PATCH_DBIT_ADDR	# load address of start of dbits for this region
	EXPORT(__midlabel_un_byte_patch_shift_1)
	srl	a1, a0, PATCH_LINE_SHIFT	# shift offset by cache line size
	sll     a1, a1, 2		# times 4 to get dbit offset in region
	addu    AT, a1, AT		# address of dbit
	lw	v0, 0(AT)		# read dbit
	addi	a1, zero, 1		# delay slot: set length arg to mid_line_exception
	bltz    v0,$L4200		# branch if hi order bit is 1
	nop				# unused delay slot
	
        j 	ra			# and return
	sw	zero, 0(AT)		# zero dirty bit and return

$L4200:
	SAVE_TEMP_REGS

	move	a2, AT
	EXPORT(__midlabel_un_byte_patch_upper_2)
	lui	a3, PATCH_UPPER_REGION	# delay slot: compute region_info arg
	.set	macro
	la	v0, mid_line_exception  # MUST load addr into reg for jump > 28bits
	.set	nomacro
	jal	v0                	# mid_line_exc(offset,len,dbitP,riP)
	EXPORT(__midlabel_un_byte_patch_lower_1)
	ori	a3, a3, PATCH_LOWER_REGION

	RESTORE_TEMP_REGS_AND_RETURN

EXPORT(__midlabel_builtin_unaligned_byte_end)


XLEAF(__midway_builtin_unaligned_double)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_unaligned_double)

	EXPORT(__midlabel_un_double_patch_upper_1)
	lui	AT, PATCH_DBIT_ADDR		# load address of start of dbits for this region
	EXPORT(__midlabel_un_double_patch_shift_1)
	srl	a1, a0, PATCH_LINE_SHIFT	# shift offset by cache line size
	sll     v0, a1, 2			# times 4 to get dbit offset in region
	addu    AT, v0, AT			# address of 1st dbit
	lw	v0, 0(AT)			# read dbit
	addi	a0, a0, 4			# offset to 2nd word
	bltz    v0,$L4300			# branch if hi order bit is 1
	EXPORT(__midlabel_un_double_patch_shift_2)
	srl     v0, a0, PATCH_LINE_SHIFT	# shift 2nd offset by cache line size

	bne     v0, a1, $L4301		# branch if not in same cache line
	sw	zero, 0(AT)		# zero 1st dbit
	j	ra			# return (both in same cache line)
	nop

$L4301:	
	lw	v0, 4(AT)
	addi	a1, zero, 4		# delay slot: set length parm to mid_line_exc
	bltz    v0,$L4302		# branch if hi order bit is 1
	addi	AT, AT, 4		# advance dbit ptr (in case we call mid_line_exc)
	j	ra			# return
	sw	zero, 0(AT)		# zero 2nd dbit	

$L4300:
	addi	a0, a0, -4		# restore data offset to 1st word
	addi	a1, zero, 8		# length of update
$L4302:
	SAVE_TEMP_REGS

	move	a2, AT
	EXPORT(__midlabel_un_double_patch_upper_2)
	lui	a3, PATCH_UPPER_REGION	# delay slot: compute region_info arg
	.set	macro
	la	v0, mid_line_exception  # MUST load addr into reg for jump > 28bits
	.set	nomacro
	jal	v0                	# mid_line_exc(offset,len,dbitP,riP)
	EXPORT(__midlabel_un_double_patch_lower_1)
	ori	a3, a3, PATCH_LOWER_REGION

	RESTORE_TEMP_REGS_AND_RETURN

EXPORT(__midlabel_builtin_unaligned_double_end)


XLEAF(__midway_builtin_unaligned_area)

	DECLARE_TEMP_REGS
	SAVE_TEMP_REGS
	IF_PROFILE2(__midway_builtin_unaligned_area)

	EXPORT(__midlabel_un_area_patch_upper_1)
	lui	AT, PATCH_DBIT_ADDR	# load address of start of dbits for this region
	EXPORT(__midlabel_un_area_patch_shift_1)
	srl	a2, a0, PATCH_LINE_SHIFT	# shift offset by cache line size
	sll     a2, a2, 2		# times 4 to get dbit offset in region
	addu    a2, a2, AT		# address of dbit

	EXPORT(__midlabel_un_area_patch_upper_2)
	lui	a3, PATCH_UPPER_REGION
	.set	macro
	la	v0, mid_line_exception  # MUST load addr into reg for jump > 28bits
	.set	nomacro
	jal	v0                	# mid_line_exc(offset,len,dbitP,riP)
	EXPORT(__midlabel_un_area_patch_lower_1)
	ori	a3, a3, PATCH_LOWER_REGION

	RESTORE_TEMP_REGS_AND_RETURN

EXPORT(__midlabel_builtin_unaligned_area_end)

END(__midlabel_builtin_unaligned_start)



/************************************************************************
*									*
*	Template for MIXED MODEL memory					*
*									*
************************************************************************/

LEAF(__midlabel_builtin_mixed_start)

XLEAF(__midway_builtin_mixed_word)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_mixed_word)

	EXPORT(__midlabel_mix_word_patch_upper_1)
	lui	AT, PATCH_DBIT_ADDR	# load address of start of dbits for this region
	EXPORT(__midlabel_mix_word_patch_shift_1)
	srl	a1, a0, PATCH_LINE_SHIFT	# shift offset by cache line size
	sll     a1, a1, 2		# times 4 to get dbit offset in region
	addu    AT, a1, AT		# address of dbit
	lw	v0, 0(AT)		# read dbit
	addi	a1, zero, 4		# delay slot: set length arg to mid_line_exception
	bltz    v0,$L5000		# branch if hi order bit is 1
	nop				# unused delay slot
	
        j 	ra			# and return
	sw	zero, 0(AT)		# zero dirty bit and return

$L5000:
	SAVE_TEMP_REGS

	move	a2, AT			# dbit addr parm to mid_line_exc
	EXPORT(__midlabel_mix_word_patch_upper_2)
	lui	a3, PATCH_UPPER_REGION	#  compute region_info arg
	.set	macro
	la	v0, mid_line_exception  # MUST load addr into reg for jump > 28bits
	.set	nomacro
	jal	v0                	# mid_line_exc(offset,len,dbitP,riP)
	EXPORT(__midlabel_mix_word_patch_lower_1)
	ori	a3, a3, PATCH_LOWER_REGION

	RESTORE_TEMP_REGS_AND_RETURN

EXPORT(__midlabel_builtin_mixed_word_end)


XLEAF(__midway_builtin_mixed_half)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_mixed_half)

	EXPORT(__midlabel_mix_half_patch_upper_1)
	lui	AT, PATCH_DBIT_ADDR	# load address of start of dbits for this region
	EXPORT(__midlabel_mix_half_patch_shift_1)
	srl	a1, a0, PATCH_LINE_SHIFT	# shift offset by cache line size
	sll     a1, a1, 2		# times 4 to get dbit offset in region
	addu    AT, a1, AT		# address of dbit
	lw	v0, 0(AT)		# read dbit
	addi	a1, zero, 2		# delay slot: set length arg to mid_line_exception
	bltz    v0,$L5100		# branch if hi order bit is 1
	nop				# unused delay slot
	
        j 	ra			# and return
	sw	zero, 0(AT)		# zero dirty bit and return

$L5100:
	SAVE_TEMP_REGS

	move	a2, AT			# dbit addr parm to mid_line_exc
	EXPORT(__midlabel_mix_half_patch_upper_2)
	lui	a3, PATCH_UPPER_REGION	# delay slot: compute region_info arg
	.set	macro
	la	v0, mid_line_exception  # MUST load addr into reg for jump > 28bits
	.set	nomacro
	jal	v0                	# mid_line_exc(offset,len,dbitP,riP)
	EXPORT(__midlabel_mix_half_patch_lower_1)
	ori	a3, a3, PATCH_LOWER_REGION

	RESTORE_TEMP_REGS_AND_RETURN

EXPORT(__midlabel_builtin_mixed_half_end)


XLEAF(__midway_builtin_mixed_byte)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_mixed_byte)

	EXPORT(__midlabel_mix_byte_patch_upper_1)
	lui	AT, PATCH_DBIT_ADDR	# load address of start of dbits for this region
	EXPORT(__midlabel_mix_byte_patch_shift_1)
	srl	a1, a0, PATCH_LINE_SHIFT	# shift offset by cache line size
	sll     a1, a1, 2		# times 4 to get dbit offset in region
	addu    AT, a1, AT		# address of dbit
	lw	v0, 0(AT)		# read dbit
	addi	a1, zero, 1		# delay slot: set length arg to mid_line_exception
	bltz    v0,$L5200		# branch if hi order bit is 1
	nop				# unused delay slot
	
        j 	ra			# and return
	sw	zero, 0(AT)		# zero dirty bit and return

$L5200:
	SAVE_TEMP_REGS

	move	a2, AT			# dbit addr parm to mid_line_exc
	EXPORT(__midlabel_mix_byte_patch_upper_2)
	lui	a3, PATCH_UPPER_REGION	# delay slot: compute region_info arg
	.set	macro
	la	v0, mid_line_exception  # MUST load addr into reg for jump > 28bits
	.set	nomacro
	jal	v0                	# mid_line_exc(offset,len,dbitP,riP)
	EXPORT(__midlabel_mix_byte_patch_lower_1)
	ori	a3, a3, PATCH_LOWER_REGION

	RESTORE_TEMP_REGS_AND_RETURN

EXPORT(__midlabel_builtin_mixed_byte_end)


XLEAF(__midway_builtin_mixed_double)

	DECLARE_TEMP_REGS
	IF_PROFILE(__midway_builtin_mixed_double)

	EXPORT(__midlabel_mix_double_patch_upper_1)
	lui	AT, PATCH_DBIT_ADDR		# load address of start of dbits for this region
	EXPORT(__midlabel_mix_double_patch_shift_1)
	srl	a1, a0, PATCH_LINE_SHIFT	# shift offset by cache line size
	sll     v0, a1, 2			# times 4 to get dbit offset in region
	addu    AT, v0, AT			# address of 1st dbit
	lw	v0, 0(AT)			# read dbit
	addi	a0, a0, 4			# offset to 2nd word
	bltz    v0,$L5300			# branch if hi order bit is 1
	EXPORT(__midlabel_mix_double_patch_shift_2)
	srl     v0, a0, PATCH_LINE_SHIFT	# shift 2nd offset by cache line size

	bne     v0, a1, $L5301		# branch if not in same cache line
	sw	zero, 0(AT)		# zero 1st dbit
	j	ra			# return (both in same cache line)
	nop

$L5301:	
	lw	v0, 4(AT)
	addi	a1, zero, 4		# delay slot: set length parm to mid_line_exc
	bltz    v0,$L5302		# branch if hi order bit is 1
	addi	AT, AT, 4		# advance dbit ptr (in case we call mid_line_exc)
	j	ra			# return
	sw	zero, 0(AT)		# zero 2nd dbit	

$L5300:
	addi	a0, a0, -4		# restore data offset to 1st word
	addi	a1, zero, 8		# length of update
$L5302:
	SAVE_TEMP_REGS

	move	a2, AT
	EXPORT(__midlabel_mix_double_patch_upper_2)
	lui	a3, PATCH_UPPER_REGION	# delay slot: compute region_info arg
	.set	macro
	la	v0, mid_line_exception  # MUST load addr into reg for jump > 28bits
	.set	nomacro
	jal	v0                	# mid_line_exc(offset,len,dbitP,riP)
	EXPORT(__midlabel_mix_double_patch_lower_1)
	ori	a3, a3, PATCH_LOWER_REGION

	RESTORE_TEMP_REGS_AND_RETURN

EXPORT(__midlabel_builtin_mixed_double_end)


XLEAF(__midway_builtin_mixed_area)

	DECLARE_TEMP_REGS
	SAVE_TEMP_REGS
	IF_PROFILE2(__midway_builtin_mixed_area)

	EXPORT(__midlabel_mix_area_patch_upper_1)
	lui	AT, PATCH_DBIT_ADDR	# load address of start of dbits for this region
	EXPORT(__midlabel_mix_area_patch_shift_1)
	srl	a2, a0, PATCH_LINE_SHIFT	# shift offset by cache line size
	sll     a2, a2, 2		# times 4 to get dbit offset in region
	addu    a2, a2, AT		# address of dbit

	EXPORT(__midlabel_mix_area_patch_upper_2)
	lui	a3, PATCH_UPPER_REGION
	.set	macro
	la	v0, mid_line_exception  # MUST load addr into reg for jump > 28bits
	.set	nomacro
	jal	v0                	# mid_line_exc(offset,len,dbitP,riP)
	EXPORT(__midlabel_mix_area_patch_lower_1)
	ori	a3, a3, PATCH_LOWER_REGION

	RESTORE_TEMP_REGS_AND_RETURN

EXPORT(__midlabel_builtin_mixed_area_end)

END(__midlabel_builtin_mixed_start)



