/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * pmsg: Print MSG - decode messages between MIDWAY lock modules...
 *
 * 11 jul 91 (mjz): add retrans bit
 * 17 Sep 91 (mjz): add allocate, allocInfo
 * 15 Dec 91 (mjz): add host number (Comm_index) to messages
 * 16 Dec 91 (mjz): add dest to outcall's (src is in msg in incalls (at least
 *			as much as we know, even for fwd'd))
 * 30 Jan 92 (mjz): do more work here, so we can call write instead of
 *	piecemeal printf's.  These were turning into lots of little mach
 *	messages, which besides being overly wasteful could get interleaved.
 *  8 Jun 92 (mjz)(per wsawdon): put local includes in "" rather than <>
 *  6 Jul 92 (mjz): add support for keeping a trace buffer: log the last
 *	ten messages sent to or received from a particular processor.
 * 15 Jul 92 (mjz): supress duplicates.  This requires scanning all
 *	entries;  I tried just the last one, but it's not enough data.
 *	if # of entries becomes significant, should delete this support.
 * 18 Aug 92 (was): added printfs for M_PAGE_xxx ops.
 * 16 Aug 92 (mjz): add lock_newown; fix seppuku
 *  4 May 94 (was): changed comm msg format 
 */
#include "comm.h"
#include "lock.h"
#include <stdio.h>
#include "midway-dbg.h"		/* IS_DBG_COMM_MSGX */
#include "midway.h"

extern int Comm_index;		/* number of this processor/host */
extern int Comm_hnum;		/* number in list of names... */

/*
 *  if (incall)
 *   (*incall)(&sin, rc, msgheader, 0, 0, buf, buflen);
 *  if (outcall)
 *    (*outcall)(&sin, &msgheader, 0, 0, msg, mlen);
 *               sock    our hdr   extra extralen msg messagelen 
 *                       COMMMSG
 */

static void
midPmsgLock(pbi, pbl, buf, buflen)
     char *pbi;			/* current place in output buf */
     int  pbl;			/* amt left in output buf */
     char *buf;
     u_long buflen;
{
  int		i;
  m_lock_req_t	*lr;
  addrseg_t	*as;
  addrpair_t	*ap, *apnext;
  int		len;

  if (buflen < sizeof(m_lock_req_t)) {
    strcpy(pbi, "\t bogus lock request!");
    return;
  }

  lr = (m_lock_req_t *)buf;
  sprintf(pbi, " ((@%lu %s i %d))", lr->dataClock,
	  (lr->mode == LOCK_RO) ? "RO" :
	  (lr->mode == LOCK_RW) ? "RW" : "INVALID",
	  lr->incarnation);

  if (IS_DBG_COMM_MSGX) {	/* extended message translation */
    len = strlen(pbi);		/* adjust lengths from above */
    pbi += len;
    pbl -= len;
    buflen -= sizeof(m_lock_req_t);
    as = (addrseg_t *) (lr+1);	/* point after lockrequest */
    while (buflen) {		/* something to do */
      switch (as->segtype) {
      case DATASEG:
#ifdef PAGE_BASED
      case DIFFSEG:
#endif
	ap = &(as->ap);		/* look at addrpair */
	buflen -= ((u_long)ap-(u_long)as); /* account for header */
	for (i=0; len && (i< as->num); i++) {
	  if (pbl < 20) {
	    return;		/* no more room */
	  }
	  sprintf(pbi, " %lu: %lx+%lu", as->clock, ap->addr, ap->len);
	  len = strlen(pbi);
	  pbi += len;
	  pbl -= len;
	  apnext = (addrpair_t *)
	    (((u_long)ap->data + ap->len + 3) & ~3); /* word boundaries only */
	  if (buflen < ((u_long) apnext - (u_long) ap)) {
	    /* this is really an error... sigh. */
	    buflen = 0;
	  } else {
	    buflen -= ((u_long)apnext - (u_long)ap);
	  }
	  ap = apnext;
	} /* for i */
	as = (addrseg_t *) ap; /* next segment would follow */
	break;			/* case DATASEG */
      default:
	sprintf(pbi, "?Seg %lx - %d left", as->segtype, buflen);
	break;
      } /* switch */
    } /* while buflen */
  } /* extended message translation */

  return;
} /* midPmsgLock */

static void
midPmsg(pb,pbi,pbl,sin, msgh, buf, buflen, reallen)
     char	*pb;		/* output buffer */
     char	*pbi;		/* current spot in buffer */
     int	pbl;		/* amount left in buffer */
     struct sockaddr_in *sin;
     COMMMSG 	*msgh;
     char	*buf;
     u_long	buflen;
     u_long	reallen;
{
  m_lock_req_t *mlr;
  int ok;
  int len;			/* for figuring out intermediate lengths */

  struct ipst {			/* to translate IP addrs on fly */
    unsigned ip1:8;
    unsigned ip2:8;
    unsigned ip3:8;
    unsigned ip4:8;
  } *ipaddr;
  int i, j;
  u_long mask;

/* avoid
  ipaddr = (struct ipst *) sin->sin_addr.s_addr;
  fprintf(stderr, "%u.%u.%u.%u[%u] ",ipaddr->ip1,ipaddr->ip2,ipaddr->ip3,
	  ipaddr->ip4, sin->sin_port);
*/



#define TESTOP(x,y) (((y) & M_OP_BITS) == (x))
#define STRADD(y) strcpy(pbi,y); pbi += 4; pbl -=4
  ok = 0;
  if    (TESTOP(M_ACK_OP, msgh->type))	       {STRADD("ACK "); ok++;}
  if 	(TESTOP(M_NACK_OP, msgh->type))	       {STRADD("NAK "); ok++;}
  if	(TESTOP(M_HOST_OP, msgh->type))	       {STRADD("HST "); ok++;}
  if 	(TESTOP(M_BARRIER_OP, msgh->type))     {STRADD("BAR "); ok++;}
  if 	(TESTOP(M_LOCK_OP, msgh->type))	       {STRADD("LCK "); ok++;}
  if 	(TESTOP(M_LOCK_INV_OP, msgh->type))    {STRADD("LIN "); ok++;}
  if 	(TESTOP(M_INVALIDATE_OP, msgh->type))  {STRADD("INV "); ok++;}
  if 	(TESTOP(M_LOCK_NEWOWN, msgh->type))    {STRADD("OWN "); ok++;}
  if 	(TESTOP(M_NOTIFY_OP, msgh->type))      {STRADD("NOT "); ok++;}
  if 	(TESTOP(M_SCHED_OP, msgh->type))       {STRADD("SCH "); ok++;}
  if 	(TESTOP(M_ALLOCATE, msgh->type))       {STRADD("ALC "); ok++;}
  if 	(TESTOP(M_ALLOC_INFO, msgh->type))     {STRADD("AIF "); ok++;}
  if 	(TESTOP(M_PAGE_DATA, msgh->type))      {STRADD("PDA "); ok++;}
  if 	(TESTOP(M_PAGE_UPDATE, msgh->type))    {STRADD("PUP "); ok++;}
  if 	(TESTOP(M_PAGE_COPYSET, msgh->type))   {STRADD("PCS "); ok++;}
  if 	(TESTOP(M_SEPPUKU, msgh->type))        {STRADD("DIE "); ok++;}

  if (!ok) {
    sprintf(pbi, "?OP?(%lx) ", msgh->type);
    len = strlen(pbi);
    pbi += len;
    pbl -= len;
  } 

  if (msgh->type & M_REPLY_BIT)	{
    strcpy(pbi, "RPLY ");
    pbi += 5;
    pbl -= 5;
  }

  sprintf(pbi, "%lu(%lu)@%lu", msgh->source, msgh->id, msgh->clock);
  len = strlen(pbi);
  pbi += len;
  pbl -= len;

  if (msgh->type & M_REPLY_BIT) {
    sprintf(pbi, " D %lu", msgh->rid);
  } else {
    sprintf(pbi, " R %lu(%lu)", msgh->rhost, msgh->rid);
  }
  len = strlen(pbi);
  pbi += len;
  pbl -= len;

  sprintf(pbi, " L %lx {%lu} #%lu", msgh->lock, msgh->tid, msgh->pacnum);
  len = strlen(pbi);
  pbi += len;
  pbl -= len;

  if (msgh->flags & M_RETRANS) {
    *pbi++ = '!';		/* note retransmissions */
    *pbi = '\0';
    pbl--;
  }

  if (msgh->flags & M_LASTPACK) {
    strcpy(pbi, " [LAST]");
    pbi += 7;
    pbl -= 7;
  }

  if (msgh->type & M_REPLY_BIT) {
      *pbi++ = ' '; *pbi++ = 'A'; *pbi++ = '[';	*pbi = '\0'; /* " A[" */
      pbl -= 3;
      for (i=0,j=0,mask=0x80000000; i<=Comm_hnum; i++) {
	  if (mask & msgh->nacks.bitArray[j]) {	/* host is in list */
	      sprintf(pbi, " %lu", i);
	      len = strlen(pbi);
	      pbi += len;
	      pbl -= len;
	  }
	  if (!(mask >>= 1)) {	/* shift right one */
	      mask = 0x80000000;	/* if fallen off end, restart */
	      j++;
	  }
      }
      *pbi++ = ' '; *pbi++ = ']'; *pbi = '\0'; /* " ]" */
      pbl -= 2;
  }

  /* length of (non-header) part of message */
  sprintf(pbi, " l %lu", reallen);
  len = strlen(pbi);
  pbi += len;
  pbl -= len;

  switch(msgh->type) {
  case M_LOCK_OP:
    midPmsgLock(pbi, pbl, buf, buflen);
    break;
  case M_LOCK_OP|M_REPLY_BIT:
  case M_LOCK_INV_OP|M_REPLY_BIT:
    if (IS_DBG_COMM_MSGX)
      midPmsgLock(pbi, pbl, buf, buflen);
    break;
  case M_INVALIDATE_OP:		/* print out incarnation # too */
    sprintf(pbi, " i %lu", *((int *) buf));
    /* length figured out after... */
    break;
  default:
    break;
  } /* switch on type */

  len = strlen(pbi);
  pbi += len;
  pbl -= len;
  *pbi++ = '\n';
  *pbi = '\0';
  len = pbi - pb + 1; /* +1 for \0 */
  write(2/*stderr*/, pb, len);
  return;
} /* midPmsg */

/*
 * these can be called by either the trace or print routines
 */
int
midPmsgI( header, sin, rc, msgh, eaddr, elen, buf, buflen, reallen ) 
     char *header; /* print if available */
     struct sockaddr_in *sin;
     int rc;
     COMMMSG *msgh;
     char *eaddr;
     u_long elen;
     char *buf;
     u_long buflen;
     u_long reallen;
{
  char pb[256];			/* what we will make this be... */
  int len;

  if (header) {
    sprintf(pb, "%s<%d+ ",header, Comm_index);
  } else {
    sprintf(pb, "<%d+ ",Comm_index);
  }
  len = strlen(pb);
  midPmsg(pb,pb+len, sizeof(pb)-len, sin, msgh, buf, buflen, reallen);
  return(0);
} /* midPmsgI */

int
midPmsgO( header, destid, sin, msgh, eaddr, elen, buf, buflen, reallen)
     char * header;
     u_long destid;		/* to whom is this being delivered? */
     struct sockaddr_in *sin;
     COMMMSG *msgh;
     char *eaddr;
     u_long elen;
     char *buf;
     u_long buflen;
     u_long reallen;
{
  char pb[256];
  int len;

  if (header) {
    sprintf(pb, "%s+%d>%d> ", header,Comm_index,destid);
  } else {
    sprintf(pb, "+%d>%d> ", Comm_index,destid);
  }
  len = strlen(pb);
  midPmsg(pb,pb+len,sizeof(pb)-len,sin, msgh, buf, buflen, reallen);
  return(0);
} /* midPmsgO */

/*
 * called by the comm routines to print
 */
int
midPmsgIn( sin, rc, msgh, eaddr, elen, buf, buflen) /* for incalls */
     struct sockaddr_in *sin;
     int rc;
     COMMMSG *msgh;
     char *eaddr;
     u_long elen;
     char *buf;
     u_long buflen;
{
  midPmsgI((char*)0, sin, rc, msgh, eaddr, elen, buf, buflen, buflen ) ;
  return(0);
} /* midPmsgIn */

int
midPmsgOut( destid, sin, msgh, eaddr, elen, buf, buflen) /* for outcalls */
     u_long destid;		/* to whom is this being delivered? */
     struct sockaddr_in *sin;
     COMMMSG *msgh;
     char *eaddr;
     u_long elen;
     char *buf;
     u_long buflen;
{
  midPmsgO( (char*) 0, destid, sin, msgh, eaddr, elen, buf, buflen, buflen);
  return(0);
} /* midPmsgOut */

/*
 * TRACE BUFFER SUPPORT
 */
static struct sockaddr_in mid_pmsg_sin;	/* junk to give to print routines */
typedef struct {
  u_long	destid;		/* destination (0 if incoming) */
  COMMMSG	msgh;		/* msg header info */
  m_lock_req_t	msg;		/* enuf of msg to hold for mlockreqt */
  u_long	len;		/* length of original buffer */
#ifndef PMSG_ALLOW_DUPLICATES
#ifndef PMSG_LAST_DUP_ONLY
  u_long	dupcount;	/* # of duplicates */
  u_long	lastseqno;	/* last duplicate seqno */
#endif PMSG_LAST_DUP_ONLY
#endif PMSG_ALLOW_DUPLICATES
} pmsg_trace_entry;

/* enough memory will be allocated to hold 10 entries for each host. */
pmsg_trace_entry *mid_pmsg_tracebuf;
#define PMSG_TRACE_ENTRIES	20

/* to tell where everything is, we need circular buffer control for each. */
typedef struct {
  u_long	start;		/* start index for this host */
  u_long	end;		/* end index */
  u_long	p;		/* current pointer (place next) */
} pmsg_trace_control;

pmsg_trace_control *mid_pmsg_tracectl = 0;

mutex_t mid_pmsg_lock;		/* so trace's don't stomp on each other */

extern Comm_hnum;		/* how many "processors" exist? */

void
midPmsgTraceInit()
{
  int i;

/*  for (i=0; i<0x2000000; i++) ; */

  mid_pmsg_tracebuf = (pmsg_trace_entry *) CALLOC(Comm_hnum*PMSG_TRACE_ENTRIES,
						  sizeof(pmsg_trace_entry));
  if (!mid_pmsg_tracebuf) {
    fprintf(stderr, "midPmsgTraceInit: error allocating tracebuf\n");
    return;
  }

  mid_pmsg_tracectl = (pmsg_trace_control *) CALLOC(Comm_hnum,
						    sizeof(pmsg_trace_control));
  if (!mid_pmsg_tracectl) {
    fprintf(stderr, "midPmsgTraceInit: error allocating tracectl, although tracebuf made it OK.\n");
    FREE (mid_pmsg_tracebuf);
    mid_pmsg_tracebuf = 0;
    return;
  }

  for (i=0; i<Comm_hnum; i++) {	/* set up tracectl */
    mid_pmsg_tracectl[i].start = i*PMSG_TRACE_ENTRIES;
    mid_pmsg_tracectl[i].end = mid_pmsg_tracectl[i].start+PMSG_TRACE_ENTRIES-1;
    mid_pmsg_tracectl[i].p = mid_pmsg_tracectl[i].start;
  }

  mid_pmsg_lock = mutex_alloc();
  MIDWAY_MUTEX_SET_NAME(mid_pmsg_lock, "mid_pmsg_lock");
  return;
} /* midPmsgTraceInit */

void
midPmsgTrace( destid, msgh, buf, buflen )
     u_long	destid;
     COMMMSG	*msgh;
     char	*buf;
     u_long	buflen;
{
  int i;
  int j,dup;			/* for previous */
  int host;

  if (mid_pmsg_tracectl == 0)
    return;			/* no trace data */

  /* figure out host */
  if (destid) {
    host = ((int)destid);	/* whew */
  } else {
    host = ((int)(msgh->source));
    if (host == 0) {		/* HOST_OP messages could have this */
      return;			/* sorry, no spot in table for them */
    }
  }
  host--;			/* convert to index */

  mutex_lock(mid_pmsg_lock);

  i = mid_pmsg_tracectl[host].p; /* where to put data */
#ifndef PMSG_ALLOW_DUPLICATES
  /* simple trick to hopefully increase the s/n ratio by removing
     retransmissions... they will quickly "flush the cache" of all
     useful info...
   */
#ifdef PMSG_RETRANS_DUP_ONLY	/* 4sep92 disable this misfeature */
  if (msgh->flags & M_RETRANS) { /* only for retransmissions */
#endif
#ifdef PMSG_LAST_DUP_ONLY
    if (i == mid_pmsg_tracectl[host].start)
      j = mid_pmsg_tracectl[host].end; /* wrap backwards */
    else
      j = i-1;			/* just look at previous */
    
    /* hmm... this will not work for multi-packet messages, either. sigh. */
    if ((destid == mid_pmsg_tracebuf[j].destid) &&
	(msgh->type == mid_pmsg_tracebuf[j].msgh.type) && /* add 9sep92 */
	(msgh->tid == mid_pmsg_tracebuf[j].msgh.tid)){	/* trans ids match */
      /* now check source! */
      if (msgh->type & M_REPLY_BIT) {
	dup = msgh->source == mid_pmsg_tracebuf[j].msgh.source;
      } else {			/* use reply-to for request message */
	dup = msgh->rhost == mid_pmsg_tracebuf[j].msgh.rhost;
      }
      if (dup) {
	mutex_unlock(mid_pmsg_lock);
	return;
      }
    } /* DESTID & TID MATCH */
#else /*PMSG_LAST_DUP_ONLY*/ /* try catching *all* duplicates */
    j = i;			/* will start 1 back of here */
    do {
      /* work backwards */
      if (j == mid_pmsg_tracectl[host].start)
	j = mid_pmsg_tracectl[host].end; /* wrap backwards */
      else
	j = j-1;			/* just look at previous */

      if ((destid == mid_pmsg_tracebuf[j].destid) &&
	  (msgh->type == mid_pmsg_tracebuf[j].msgh.type) && /* add 9sep92 */
	  (msgh->tid == mid_pmsg_tracebuf[j].msgh.tid)){
	if (msgh->type & M_REPLY_BIT) { /* stupid, replies are never retrans */
	  dup = msgh->source == mid_pmsg_tracebuf[j].msgh.source;
	} else {
	  dup = msgh->rhost == mid_pmsg_tracebuf[j].msgh.rhost;
	}
	if (dup) {
	  mid_pmsg_tracebuf[j].dupcount++;	/* note # duplicates */
	  mid_pmsg_tracebuf[j].lastseqno = msgh->clock;
	  mutex_unlock(mid_pmsg_lock);
	  return;
	}
      } /* DESTID & TID MATCH */
    } while (j != i);
#endif /*PMSG_LAST_DUP_ONLY*/
#ifdef PMSG_RETRANS_DUP_ONLY
  } /* RETRANS */
#endif /*PMSG_RETRANS_DUP_ONLY */
#endif /*PMSG_ALLOW_DUPLICATES*/
  mid_pmsg_tracebuf[i].destid = destid;
  mid_pmsg_tracebuf[i].len = buflen;
  bcopy((char *) msgh, (char *) &(mid_pmsg_tracebuf[i].msgh), sizeof(COMMMSG));
  if (buflen< sizeof(mid_pmsg_tracebuf[i].msg))
    bzero(&(mid_pmsg_tracebuf[i].msg), sizeof(mid_pmsg_tracebuf[i].msg));
  if (buflen) {
    if (buflen < sizeof(mid_pmsg_tracebuf[i].msg))
      bcopy(buf, (char *)&(mid_pmsg_tracebuf[i].msg), buflen);
    else
      bcopy(buf, (char *)&(mid_pmsg_tracebuf[i].msg),
	    sizeof(mid_pmsg_tracebuf[i].msg));
  }
#ifndef PMSG_ALLOW_DUPLICATES
#ifndef PMSG_LAST_DUP_ONLY
    mid_pmsg_tracebuf[i].dupcount = 0; /* no dups yet */
    mid_pmsg_tracebuf[i].lastseqno = 0; /* yaya */
#endif PMSG_LAST_DUP_ONLY
#endif PMSG_ALLOW_DUPLICATES

  /* inc "i" */
  mid_pmsg_tracectl[host].p++;
  if (mid_pmsg_tracectl[host].p > mid_pmsg_tracectl[host].end)
    mid_pmsg_tracectl[host].p = mid_pmsg_tracectl[host].start;

  mutex_unlock(mid_pmsg_lock);

  return;
} /* midPmsgTrace */

void
midPmsgTracePrt( host )
     u_long host;		/* print this host's trace entries old->new */
{
  int i = (int)(host-1);
  int j;
  char msg[48];			/* for prepending stuff */
  
  if (mid_pmsg_tracectl == 0)
    return;			/* no trace data */

  if ((host == 0) || (host > Comm_hnum)) {
    fprintf(stderr, 
	    "--?-> midPmsgTracePrt: processor %lu not from 1 to %lu!\n",
	    host, Comm_hnum);
    return;
  }


  fprintf(stderr, "\nLast %d messages to and from processor %lu:\n",
	  PMSG_TRACE_ENTRIES, host);
  /* don't lock for now;  so can print even if task stopped w/in trace */
  j = mid_pmsg_tracectl[i].p;	/* start here */
  do {
    msg[0] = '\0';		/* no message yet */
    if (mid_pmsg_tracebuf[j].msgh.version != 0) { /* valid entry */
#ifndef PMSG_ALLOW_DUPLICATES
#ifndef PMSG_LAST_DUP_ONLY
      if (mid_pmsg_tracebuf[j].dupcount) {
	sprintf(msg, "[%d dup, lastseqno %d] ",
		mid_pmsg_tracebuf[j].dupcount,
		mid_pmsg_tracebuf[j].lastseqno);
#ifdef oldway
	write(2/*stderr*/,msg,strlen(msg));
#endif
      }
#endif PMSG_LAST_DUP_ONLY
#endif PMSG_ALLOW_DUPLICATES
      if (mid_pmsg_tracebuf[j].destid) { /* out */
	(void)midPmsgO( msg, 
		          mid_pmsg_tracebuf[j].destid,
			  &mid_pmsg_sin, /* junk */
			  &(mid_pmsg_tracebuf[j].msgh),
			  0, 0,
			 /* following only used for lockreq, so ok to add
			    always */
			  &(mid_pmsg_tracebuf[j].msg),
			 sizeof(mid_pmsg_tracebuf[j].msg),
		          mid_pmsg_tracebuf[j].len
			 );
      } else {			/* in */
	(void)midPmsgI(  msg,
			  &mid_pmsg_sin, /* junk */
			  0,
			  &(mid_pmsg_tracebuf[j].msgh),
			  0, 0,
			 /* following only used for lockreq, so ok to add
			    always */
			  &(mid_pmsg_tracebuf[j].msg),
			 sizeof(mid_pmsg_tracebuf[j].msg),
			  mid_pmsg_tracebuf[j].len
			 );
      }
    } /* valid */
    /* move on to next */
    j++;
    if (j > mid_pmsg_tracectl[i].end)
      j = mid_pmsg_tracectl[i].start; /* wraparound */
  } while (j != mid_pmsg_tracectl[i].p); /* not looped around */

  return;
} /* midPmsgTracePrt */
