/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * queue.c: queue routines for midway
 *
 * 27 Jun 91 (mjz): add prt, null, and dup for extq.
 *		Remove ->last entries; circular stuff was busted.
 *  8 Jun 92 (mjz): add extq_printall; check for integrity
 * 11 Aug 92 (was): added clock_t type
 *  9 Mar 94 (was): added incarn to queued elements
 */
#include <sys/types.h>		/* for u_long */
#include <mach.h>          
#include <stdio.h>		/* for EXTQ_PARANOIA printing */

#include "midway_types.h"       /* global midway type definitions */
#include "midway-dbg.h"
#include "queue.h" 

#define EXTQ_PARANOIA 1

extern task_t midway_task_self;	/* mach_task_self() */

int
mid_threadq_add( q, tid )
     threadq_t *q;		/* to this queue */
     u_long	tid;		/* add this element */
{
  if (q->ld->next == q->q) {	/* queue needs expansion */
    if((q->ld->next = (ourqe_t *) CALLOC(1,sizeof(threadqe_t))) == 0)
      return(-1);		/* memory exhaustion */
    q->ld->next->next = q->q;	/* link it in */
  } 
  q->ld = q->ld->next;	/* point to free element */
  ((threadqe_t *)(q->ld))->tid = tid;	/* fill in data */
  return(0);			/* done! */
} /* mid_threadq_add */

int
mid_threadq_pop( q, tid )
     threadq_t	*q;		/* from this q , remove the next item*/
     u_long	*tid;		/* and place it here */
{
  if(q->q == q->ld) {		/* gosh, we're empty */
    return (-1);
  }
  *tid = ((threadqe_t *)(q->q->next))->tid;	/* get value */
  q->q = q->q->next;		/* and circle queue around */
  return(0);			/* bye */
} /* mid_threadq_pop */

u_long
mid_threadq_top( q )		/* return top w/o disturbing */
     threadq_t	*q;		/* from this q , remove the next item*/
{
  if(q->q == q->ld) {		/* gosh, we're empty */
    return (0xffffffff);	/* -1, unsigned :-) */
  }
  return( ((threadqe_t *)(q->q->next))->tid );	/* get value */
} /* mid_threadq_top */


/*
 * push is not a really good op with this data structure...
 *  However, for now I will use it as is... I use it only for
 *  queuing invalidates (which have to come before new lock requests...
 *  and it will only matter if this node is a reader, currently requesting
 *  a write lock, and gets other requests before it can process the
 *  invalidate.  Perhaps change DS...  (to avoid linear lookup in case
 *  of partially full q). --23may91 mjz
 */
int
mid_extq_push( q, host, tid, tranid, clock, incarn )
     extq_t	*q;		/* to this queue */
     u_long	host;		/* add... */
     u_long	tid;
     u_long	tranid;		/* transaction id */
     clock_t	clock;
     clock_t	incarn;
{
  register extqe_t	*eq;
  ourqe_t *p;

#ifdef EXTQ_PARANOIA
  if (mid_extq_check(q)){
    mid_extq_prt_all(q);
    fprintf(stderr, "queue %lx inconsistent -- HALTING PID %d!!", 
	    q, getpid());
    task_suspend(midway_task_self);	/* mach-ism */
  }
#endif EXTQ_PARANOIA
  if (q->ld->next == q->q) {	/* queue needs expansion */ /* treat like q*/
    if((q->ld->next = (ourqe_t *) CALLOC(1,sizeof(extqe_t))) == 0)
      return(-1);		/* memory exhaustion */
    q->ld->next->next = q->q;	/* link it in */
    eq = (extqe_t *) q->q;	/* use old head for data */
    q->q = q->ld->next;		/* new element is new head */
  } else {			/* q does not need expansion */
    if (q->q == q->ld) {	/* empty, treat like queue */
      q->ld = q->ld->next;	/* point to free element */
      eq = (extqe_t *) q->ld;	/* prepare to fill */
    } else {			/* not empty, must find ptr to last elt */
      for (p = q->ld; p->next != q->q; p = p->next) ; /* got it! */
      /* error in ds could cause above to spin */
      eq = (extqe_t *) q->q;	/* use top element */
      q->q = p;			/* circle head back */
    }
  } /* q does not need expansion */
  eq->tid = tid;		/* fill in data */
  eq->host = host;
  eq->clock = clock;
  eq->incarn = incarn;
  eq->tranid = tranid;
  return(0);			/* done! */
} /* mid_extq_push */
/* could distill common parts of these two... */
int
mid_extq_add( q, host, tid, tranid, clock, incarn )
     extq_t	*q;		/* to this queue */
     u_long	host;		/* add... */
     u_long	tid;
     u_long	tranid;		/* transaction id */
     clock_t	clock;
     clock_t	incarn;
{
  register extqe_t	*eq;

#ifdef EXTQ_PARANOIA
  if (mid_extq_check(q)){
    mid_extq_prt_all(q);
    fprintf(stderr, "queue %lx inconsistent -- HALTING PID %d!!", 
	    q, getpid());
    task_suspend(midway_task_self);	/* mach-ism */
  }
#endif EXTQ_PARANOIA
  if (q->ld->next == q->q) {	/* queue needs expansion */
    if((q->ld->next = (ourqe_t *) CALLOC(1,sizeof(extqe_t))) == 0)
      return(-1);		/* memory exhaustion */
    q->ld->next->next = q->q;	/* link it in */
  }
  q->ld = q->ld->next;		/* point to free element */
  
  (eq = (extqe_t *)(q->ld))->tid = tid;	/* fill in data */
  eq->host = host;
  eq->clock = clock;
  eq->incarn = incarn;
  eq->tranid = tranid;
  return(0);			/* done! */
} /* mid_extq_add */

int
mid_extq_pop( q, host, tid, tranid, clock, incarn )
     threadq_t	*q;		/* from this q , remove the next item*/
     u_long	*host;
     u_long	*tid;		/* and place it here */
     u_long	*tranid;
     clock_t	*clock;
     clock_t	*incarn;
{
  register extqe_t	*eq;

#ifdef EXTQ_PARANOIA
  if (mid_extq_check(q)){
    mid_extq_prt_all(q);
    fprintf(stderr, "queue %lx inconsistent -- HALTING PID %d!!", 
	    q, getpid());
    task_suspend(midway_task_self);	/* mach-ism */
  }
#endif EXTQ_PARANOIA
  if(q->q == q->ld) {		/* gosh, we're empty */
    return (-1);
  }
  *tid = (eq = (extqe_t *)(q->q->next))->tid;	/* get value */
  *host = eq->host;
  *tranid = eq->tranid;
  *clock = eq->clock;
  *incarn = eq->incarn;
  q->q = q->q->next;		/* and circle queue around */
  return(0);			/* bye */
} /* mid_extq_pop */

/*
 ************************************************************************
 *		      m i d _ e x t q _ d u p				*
 *----------------------------------------------------------------------*
 * Given a host, thread, transid triple, see if it's already in the Q.  *
 ************************************************************************
 */
int
mid_extq_dup( q, host, thread, tranid )
     extq_t	*q;
     u_long	host;
     u_long	thread;
     u_long	tranid;
{
  register extqe_t	*eq;

#ifdef EXTQ_PARANOIA
  if (mid_extq_check(q)){
    mid_extq_prt_all(q);
    fprintf(stderr, "queue %lx inconsistent -- HALTING PID %d!!", 
	    q, getpid());
    task_suspend(midway_task_self);	/* mach-ism */
  }
#endif EXTQ_PARANOIA
  if (q->q == q->ld) {		/* empty, no duplicate */
    return(0);
  }

  eq = (extqe_t *) (q->q); /* get (just before) first elt */
  while ((ourqe_t *) eq != q->ld) { /* still elements to search */
    eq = eq->next;		/* move to next elment */
    if ((eq->host == host) && (eq->tid == thread) && (eq->tranid == tranid)) {
      return(1);		/* got one! */
    }
  } /* while stuff to search for */

  return(0);			/* didn't find */
} /* mid_extq_dup */

/*
 ************************************************************************
 *		      m i d _ e x t q _ n u l l				*
 *----------------------------------------------------------------------*
 * Empty the queue.  With this scheme, it's really simple-- set the last*
 * data pointer to q->q.  Voila!					*
 ************************************************************************
 */
void
mid_extq_null( q )
     extq_t	*q;
{
#ifdef EXTQ_PARANOIA
  if (mid_extq_check(q)){
    mid_extq_prt_all(q);
    fprintf(stderr, "queue %lx inconsistent -- HALTING PID %d!!", 
	    q, getpid());
    task_suspend(midway_task_self);	/* mach-ism */
  }
#endif EXTQ_PARANOIA
  q->ld = q->q;
  return;
} /* mid_extq_null */

/*
 ************************************************************************
 *			m i d _ e x t q _ n u m				*
 *----------------------------------------------------------------------*
 * Return the number of elements in the queue.				*
 * Future: keep in head of list (mjz10sep92)				*
 ************************************************************************
 */
int
mid_extq_num( q )
     extq_t	*q;
{
  register ourqe_t	*oq;
  register int num = 0;

  if (q->q == q->ld) {		/* empty */
    return(0);
  }

  oq = (q->q);			/* get (just before) first elt */
  while (oq != q->ld) {		/* still elements to search */
    oq = oq->next;		/* move to next elment */
    num++;
  } /* while stuff to search for */

  return num;
} /* mid_extq_num */

/*
 ************************************************************************
 *			m i d _ e x t q _ p r t				*
 *----------------------------------------------------------------------*
 * Print out the ext q given.  To status server?			*
 ************************************************************************
 */
void
mid_extq_prt( q )
     extq_t	*q;
{
  register extqe_t	*eq;

  if (q->q == q->ld) {		/* empty, no duplicate */
    printf("-extq->[queue empty]\n");
    return;
  }

  eq = (extqe_t *) (q->q); /* get (just before) first elt */
  while ((ourqe_t *) eq != q->ld) { /* still elements to search */
    eq = eq->next;		/* move to next elment */
    printf("-extq-> host %lu thread %lu tid %lu clock %lu incarn %lu\n",
	   eq->host, eq->tid, eq->tranid, eq->clock, eq->incarn);
  } /* while stuff to search for */
  printf("-extq-> [end of queue]\n");

  return;
} /* mid_extq_prt */
/*
 ************************************************************************
 *		    m i d _ e x t q _ p r t _ a l l			*
 *----------------------------------------------------------------------*
 * Print out the ext q given.  To status server?			*
 ************************************************************************
 */
int
mid_extq_prt_all( q )
     extq_t	*q;
{
  register extqe_t	*eq;
  char *empty;
  int broken = 0;

  if (q->q == q->ld) {		/* empty, no duplicate */
    printf("-extq(all)->[queue empty]\n");
    empty = "{empty}";
  } else
    empty = "";

  if (q->ld == 0) {
    printf("-extq(all)->[q->ld is zero]\n");
    broken = 1;
  }

  if (q->q == 0) {
    printf("-extq(all)->[q->q is zero!]\n");
    return(1);			/* cannot continue -- nothing to check */
  }

  eq = (extqe_t *) (q->q); /* get (just before) first elt */
  do {
    eq = eq->next;		/* move to next elment */
    if (eq != 0) {
      printf("-extq-> %s host %lu thread %lu tid %lu clock %lu incarn %lu\n",
	     empty, eq->host, eq->tid, eq->tranid, eq->clock, eq->incarn);
    } else {
      printf("-extq-> [ZERO next pointer found!]\n");
      broken = 1;
    }
    if ((ourqe_t *)eq == q->ld) { /* this was last with data */
      empty = "{empty}";
    }
  } while ((ourqe_t *) eq != q->q && eq != 0);
  printf("-extq-> [end of queue]\n");

  return(broken);
} /* mid_extq_prt_all  */

int
mid_extq_check( q )
     extq_t *q;
{
  register  extqe_t *eq;
  int seenlast = 0;

  if (q == 0)
    return 1;			/* really broken */

  if ((q->q == 0) || (q->ld == 0))
    return 1;			/* broken */

  eq = (extqe_t *) (q->q);
  do {
    if ((ourqe_t *) eq == q->ld)
      seenlast = 1;
    eq = eq->next;
    if (eq == 0)
      return 1;			/* zero next pointer is broken */
  } while ((ourqe_t *) eq != q->q);

  if (seenlast)
    return 0;
  else
    return 1;			/* q->ld doesn't point into loop */
} /* mid_extq_check */
