/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * realmain: the real main for the Midway shared memory system.
 *
 * We do setup, and then call Main().
 *
 * 15 Dec 90, Matthew J. Zekauskas
 *  2 May 91 (mjz): integrate with Comm
 * 21 Jun 91 (mjz): New master/slave startup; remote cthread_fork...
 * 31 Jul 91 (mjz): need mid_pbuf_init!
 *  6 Aug 91 (mjz): add ux_port; rename main to midway_main (need crt0 mods)
 *		call main() instead of Main().
 * 28 Feb 92 (mjz): use netnameserver instead of envmgr; the latter is not
 *		officially supported under Mach 3.0, and it duplicates
 *		name server function anyway...
 * 19 Mar 92 (mjz): guard debugging printf with IS_DBG_LOCK_FORK in .._main.
 * 22 Jun 92 (was): added version info & midway_Port_debug code.
 * 16 Jul 92 (was): added assertion checks on sizeof(BITARRAY)  
 * 16 Jul 92 (mjz): add assert.h
 * 30 Jul 92 (mjz): add midway_Barr_debug
 * 11 Aug 92 (was): added clock_t type
 * 19 Aug 92 (was): add MIDMODEL_xx command line parms, init code for
 *                  update queue. removed PORT code
 *  1 Sep 92 (mjz): merge in brian's frozen/xpager updates; no longer use
 *		mach messages but rpc to get ss & ux ports.
 * 23 Sep 92 (was): add MIDSTATIC_xxx to determine handling of static data
 *  1 Oct 92 (was): added calls to gprof_mon_start & _stop
 * 15 Oct 92 (was): moved FAST_UDP stuff to CommInit.
 * 19 Oct 92 (was): Fixed bug in getting slaves to generate gmon.out profile
 *                  Changed profile name to <HostName>-<HostIndex>.gmon.out
 * 20 Oct 92 (was): Added midway_task_self to replace mach_task_self() calls.
 * 29 Oct 92 (was): Puts statistics under STAT instead of DEBUG
 * 28 Dec 92 (was): Added call to vm_dirty_page_init if default model changes
 *  6 Feb 93 (was): cleaned up termination a little. Added midway_start/stop_profile
 * 10 Mar 93 (was): Added counter to gmon.out file name to allow >1 to be gen'd.
 * 29 Jun 93 (was): Added vm_shared_term & vm_dirty_page_term to erase paging file.
 * 13 Oct 93 (was): Added midway_proc_number
 * 11 Nov 93 (was): start application's main thread via a call to rsched_move_stack()
 * 17 Dec 93 (was): Start profiling even if not compiled -pg (to get pc sample info)
 * 23 May 94 (mjz): add general trace table initialization
 */

#define DEFINE_BITWIDDLE_H	/* define external routines for bit manipulation here */

#include <sys/types.h>
#include <limits.h>             /* for CHAR_BIT used in sizeof assertions */
#include <mach.h>
#include "midway_types.h"
#include "bitwiddle.h"
#include "smexecd_types.h"
#include <mach/message.h>
#include <mach/msg_type.h>
#include <mach/mig_errors.h>
#include <cthreads.h>
#include <stdio.h>
#include "midway-dbg.h"		/* for IS_DBG... */
#include "lock.h"
#include "updateq.h"		/* prototypes for calling update queue rtns */

extern char *midway_version_string;
extern char *midway_model_string;
extern char *midway_option_string;

int midway_profiling = 0;	/* =1 if gprof profiling is enabled */
int midway_vm_profiling = 0;	/* =1 if gprof profiling of page faults */
int midway_prof_count = 0;      /* incremented for each gmon.out file written */
int midway_prof_wait = 0;	/* time in seconds to wait for profiling to complete */
/* prof_wait is not sufficient for slaves. */
int midway_prof_dump_in_progress = 0; /* =1 if gprof profiling is in the
					 process of being disabled */
int midway_prof_dump_waiters = 0; /* if anyone waiting for progress->0 */
struct mutex midway_prof_mutex;	/* so we don't have to spin for above */
struct condition midway_prof_cond;
int midway_proc_number =  0;	/* processor number */

/* forward */
static int death_thread();
static int idle_thread();
kern_return_t midway_start_profile(void);
kern_return_t midway_stop_profile(void);
void midway_exit (int rc);

extern void main();

/*
 ************************************************************************
 *			m i d w a y _ m a i n				*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
unsigned long midway_Comm_debug = 0; /* global debug definition */
unsigned long midway_Lock_debug = 0;
unsigned long midway_Barr_debug = 0;
unsigned long midway_Misc_debug = 0;
unsigned long midway_Port_debug = 0;
unsigned long midway_Stat_debug = 0;
unsigned long midway_Xpgr_debug = 0; /* for midway-dbg */
unsigned long midway_xpager_debug = 0; /* internal to xpager/panic.c */
unsigned long midway_xpager_use_syslog = 0; /* xpager/panic.c */
extern int midway_comm_network_standalone;

task_t midway_task_self;	/* cache results of mach_task_self() */

midway_main(argc,argv,envp)
     int argc;
     char **argv;
     char **envp;
{
  char255	smsg;		/* buffer to send to status server */
  char255	hname;		/* our host name */
  mach_port_t	ss_port;	/* status server port */
  mach_port_t    ux_port;	/* I/O (UniX) server port */
  mach_port_t parent_port;
  u_long	rhost;		/* IP address of rendesvous host */
  u_short	rport;		/* UDP port of rendesvous host */
  char	*hlist;			/* list of hosts */
  unsigned int hlistCnt;	/* sizeof(hlist) */
  int		hcnt;		/* number of hosts */
  u_long	hindex;		/* index of this host in hlist */
  u_long	incarn;		/* incarnation number */
  int		argc2;		/* count to pass to main() */
  char		**argv2;	/* pointer to pass to main() */
  int		done;		/* for option finding loop */
  int		gotone;		/* also for option finding loop */
  int		midway_soak = 0;/* start low(er) priority thread to soak cpu*/
  int		prof_init = 0;	/* ensure all profiling initialized early
				   so it's not charged to the process */
  kern_return_t krc;
  char *idbuf;
  
  extern void uxpass_go_remote(); /* to redirect I/O to I/O server */
  
  
  
#ifdef CTHREAD_DEBUG
  extern int cthread_debug; /* to get voluminous info */
#endif CTHREAD_DEBUG

#if WIRE_DA_THREADS
  cthread_wire();		/* under 3.0 wire down for now */
#endif
  
  
  /* pullup the first arg into the second arg as the second arg
   * contains the name to use when accessing the nameserver */
  idbuf = argv[1];
  argv[1] = argv[0];
  argv++;
  argc--;
  
  /* Provide version information */
  printf("Running MIDWAY/%s/%s%s under %s\n", 
	 midway_model_string,
	 midway_version_string,
	 midway_option_string,
	 idbuf);
  

  /*
   * Perform assertion checks on assumed sizes of types
   */
  ASSERT (sizeof(BITARRAY) == BITARRAY_CHARS);
  ASSERT ((sizeof(BITARRAY)*CHAR_BIT) == BITARRAY_BITS);

  
  /* cache task id */
  midway_task_self = mach_task_self();

  if((krc = netname_look_up(name_server_port, "", idbuf, &parent_port)) !=
     KERN_SUCCESS) {
    mach_error("netname_look_up returned ",krc); /* can we do better? */
    return(1);
  }

  /* note: to get IS_DBG_MISC_INIT to be true before the command line is
   *       processed, the binary has to be patched, e.g. with
   *   adb -w <linked-binary-name>
   *   midway_Misc_debug?W 1
   *   $q
   */
  if (IS_DBG_MISC_INIT) {
    fprintf(stderr, "midway_main: pinging parent\n"); /***/    
    krc = smexec_ping(parent_port, &krc);
    if (krc != KERN_SUCCESS)  {
      mach_error("realmain: ping is broken", krc);
    }
  }
  
  krc = smexec_getargs(parent_port,
		       &ss_port,
		       &ux_port,
		       &rhost,
		       &rport,
		       &hlist,
		       &hlistCnt,
		       &hcnt,
		       &hindex,
		       &incarn);
  if (krc != KERN_SUCCESS)  {
    /* gotta exit, not much else we can do. shouldn't write here...*/
    mach_error("realmain: smexec_getargs ",krc);
    fprintf(stderr,"child process barfola.\n");
    exit(krc);
  }
  
  if (IS_DBG_MISC_INIT) {
    fprintf(stderr,"I'm holding onto an uxport: %d\n", ux_port);
    fprintf(stderr,"ss_port =%d\tux_port=%d\trhost=%d\trport=%d\thindex=%d\n",
	    ss_port, ux_port, rhost, rport, hindex);
  }
  
  gethostname(hname,sizeof(hname));
  
  if (IS_DBG_MISC_INIT) {
    fprintf(stderr,"\n\thlist: %s\n",hlist);
    ss_msg(ss_port,(long)0,hname,hindex,"slave started!");
    ss_msg(ss_port,(long)0,hname,hindex,hlist);
  }
  
  uxpass_go_remote(ux_port); /* wubba */
  if (IS_DBG_MISC_INIT) {
    ss_msg(ss_port,(long)0,hname,hindex,"I/O redirected.");
    write(1,"test of I/O redirect by write(1,blah)\n  ",40);
  }
  
  CommInit(ss_port, NULL_CLOCK_VALUE, hindex, hcnt, hlist, rhost, rport, 
	   incarn);
  if (IS_DBG_MISC_INIT) {
    StatusMsg("CommInit finished");
  }

  cthread_detach(cthread_fork(death_thread, (any_t) ss_port));

  /*
   * parse the arguments: Look for midway-specific debugging info.
   * do after CommInit so StatusMsg works.
   */
  argv2 = argv; /* save old argv */
  argc2 = argc; /* save old argc */
  done = gotone = 0; /* not done, and don't have one */
  for (argv2++,argc2--; /* skip program name */
       (argc2 > 0) && (**argv2 == '-'); /* while an potential option */
       argv2++, argc2--) { /* skip to next */
    if (IS_DBG_MISC_INIT) {
      fprintf(stderr,"in switch loop, argv=%s, argc=%d\n", *argv2, argc2);
    }
    if (!strncmp((*argv2)+1, "MIDDBG",6)) {
      /* found one */
      switch((*argv2)[7]) { /* type of debug option */
	
      case 'c': /* comm */
	sscanf(((*argv2)+8), "%x", &midway_Comm_debug); /* get one */
	if (IS_DBG_MISC_INIT) {
	  fprintf(stderr,"setting midway_Comm_debug to %lx\n", midway_Comm_debug);
	}
	gotone = 1;
	break;
	
	
      case 'l': /* lock */
	sscanf(((*argv2)+8), "%x", &midway_Lock_debug); /* get one */
	if (IS_DBG_MISC_INIT) {
	  fprintf(stderr,"setting midway_Lock_debug to %lx\n", midway_Lock_debug);
	}
	gotone = 1;
	break;
	
      case 'b': /* barrier */
	sscanf(((*argv2)+8), "%x", &midway_Barr_debug); /* get one */
	if (IS_DBG_MISC_INIT) {
	  fprintf(stderr,"setting midway_Barr_debug to %lx\n", midway_Barr_debug);
	}
	gotone = 1;
	break;
	
      case 'm': /* miscellaneous */
	sscanf(((*argv2)+8), "%x", &midway_Misc_debug); /* get flags */
	if (IS_DBG_MISC_INIT) {
	  fprintf(stderr,"setting midway_Misc_debug to %lx\n", midway_Misc_debug);
	}
	gotone = 1;
	break;
	
      case 'p': /* porting support */
	sscanf(((*argv2)+8), "%x", &midway_Port_debug); /* get flags */
	if (IS_DBG_MISC_INIT) {
	  fprintf(stderr,"setting midway_Port_debug to %lx\n", midway_Port_debug);
	}
	gotone = 1;
	break;

      case 'x': /* xpager */
	sscanf(((*argv2)+8), "%x", &midway_Xpgr_debug); /* get flags */
	if (IS_DBG_MISC_INIT) {
	  fprintf(stderr,"setting midway_Xpgr_debug to %lx\n", midway_Xpgr_debug);
	}
	gotone = 1;
	break;
	
      case 's': /* statistics */
	sscanf(((*argv2)+8), "%x", &midway_Stat_debug); /* get flags */
	if (IS_DBG_MISC_INIT) {
	  fprintf(stderr,"setting midway_Stat_debug to %lx\n", midway_Stat_debug);
	}
	gotone = 1;
	break;
	
      case 'f': /* freeze until gdb */
	StatusMsg("child process will freeze until resumed");
	freeze_for_debugger();
	gotone = 1;
	break;
	
      default:
	sprintf(smsg,"Found probable malformed Midway option %s in arglist",
		*argv2);
	StatusMsg(smsg);
	StatusMsg("...continuing; will pass to main()...");
	done = 1; /* let it act as a stopper */
      } /* switch */
      if (done)
	break; /* don't want to move to next argument */
    }
    else if (!strncmp((*argv2)+1, "MIDMODEL_EC",11)) {
      midway_default_model = MODEL_EC;
      midway_default_flush_period = 0;
      fprintf(stderr,"setting midway default model to EC\n");
      gotone = 1;
    }
    else if (!strncmp((*argv2)+1, "MIDMODEL_RC",11)) {
      midway_default_model = MODEL_RC;
      midway_default_flush_period = 0;
      fprintf(stderr,"setting midway default model to RC\n");
      gotone = 1;
    }
    else if (!strncmp((*argv2)+1, "MIDMODEL_PC",11)) {
      midway_default_model = MODEL_PC;
      midway_default_flush_period = DEFAULT_FLUSH_PERIOD;
      fprintf(stderr,
	      "setting midway default model to PC, flush period=%lu ms\n",
	      midway_default_flush_period);
      gotone = 1;
    }
    else if (!strncmp((*argv2)+1, "MIDMODEL_FLUSH",14)) {
      sscanf(((*argv2)+15), "%lu", &midway_default_flush_period);
      fprintf(stderr,
	      "setting midway default flush period to %lu ms\n",
	      midway_default_flush_period);
      gotone = 1;
    }
    else if (!strncmp((*argv2)+1, "MIDSTATIC_SHARED",16)) {
      fprintf(stderr, "treating static data as shared.\n");
      gotone = 1;
    }
    else if (!strncmp((*argv2)+1, "MIDSTATIC_PRIVATE",17)) {
      fprintf(stderr, "treating static data as private.\n");
      gotone = 1;
    }
    else if (!strncmp((*argv2)+1, "MIDPROFILE_INIT",15)) {
      prof_init = 1;
      gotone = 1;
    }
    else if (!strncmp((*argv2)+1, "MIDPROFILE_WAIT",15)) {
      sscanf(((*argv2)+16), "%lu", &midway_prof_wait);
      gotone = 1;
    }
    else if (!strncmp((*argv2)+1, "MIDPROFILE_VM",13)) {
      midway_profiling = 1;
      midway_vm_profiling = 1;
      gotone = 1;
    }
    else if (!strncmp((*argv2)+1, "MIDPROFILE",10)) {
#if (PROFILE == 0)
      StatusMsg("Midway not compiled for profiling; you will only get instruction count info.");
#endif
      midway_profiling = 1;
      gotone = 1;
    }
    else if (!strncmp((*argv2)+1, "MIDNETWORK_ATM",14)) {
      midway_comm_network_type = MIDWAY_NET_ATM;
      gotone = 1;
    }
    else if (!strncmp((*argv2)+1, "MIDNETWORK_SA",13)) {
      midway_comm_network_standalone = 1;
      gotone = 1;
    }
    else if (!strncmp((*argv2)+1, "MIDSOAK_CPU",11)) {
      midway_soak = 1;
      gotone = 1;
    }
    else if ((*argv2)[1] == '-') { /* -- */
      gotone = 1;	/* don't let -- through */
      break; /* force end, let -- act as stopper */
    }
    else {
      break; /* unrecognized, stop here */
    }
  } /* for */
  if (gotone) {
    /* gotta fix up argstring so argv[0] is name */
    argc2++; /* going to stick in name */
    argv2--; /* point at previous arg */
    *argv2 = *argv;	/* reset pointer to name */
  } else {
    /* use original */
    argc2 = argc;
    argv2 = argv;
  }

  /* initialize profiling mutex & condition */
  mutex_init(&midway_prof_mutex);
  condition_init(&midway_prof_cond);
  

  /* re-initialize page data for dirty bit region, if model has changed */
  if (midway_default_model != MODEL_EC) {
    vm_dirty_page_init();
  }
  
  timer_init();
  if (IS_DBG_MISC_INIT) {
    StatusMsg("timer_init() finished");
  }
  
  mid_pbuf_init();
  if (IS_DBG_MISC_INIT) {
    StatusMsg("mid_pbuf_init() finished");
  }
  
  mid_lock_init(); /* init lock layer */
  if (IS_DBG_MISC_INIT) {
    StatusMsg("mid_lock_init() finished");
  }
  mid_rsched_init();
  if (IS_DBG_MISC_INIT) {
    StatusMsg("mid_rsched_init() finished");
  }

#if TRACE
  midway_init_trace();
  if (IS_DBG_MISC_INIT) {
    StatusMsg("midway_init_trace() finished");
  }
#endif /*TRACE*/
  
  
  /*
   * Perform assertion checks on assumed sizes of types
   */
  ASSERT (sizeof(BITARRAY) == BITARRAY_CHARS);
  ASSERT ((sizeof(BITARRAY)*CHAR_BIT) == BITARRAY_BITS);
  
  
  /*
   * set xpager debug flags
   */
  if (IS_DBG_XPGR_OPS) {
    midway_xpager_debug = 1;	/* causes mumble() to output stuff */
    /* to catch initialization routines, need to patch
       midway_xpager_debug because xpager init is called from shmalloc_init,
       well before we get here...
       */
  }

  if (IS_DBG_XPGR_SYSL) {
    /* we probably don't need this for Midway, but What The Heck */
    midway_xpager_use_syslog = 1;
    if (midway_xpager_debug)
      StatusMsg("realmain: turned on midway_xpager_use_syslog, but midway_xpager_debug is on, so fret calls will go to stderr...");
  }


  /* finish initialize of shmalloc.
   *  1. give it the processor # so it knows what initial region to allocate
   *  2. initialize the catching of exceptions
   */
  midway_proc_number = hindex;
  midway_set_numprocs( hindex, hcnt ); /* proc x of y */
  
#ifdef notdef    /* now uses external pager; that is set up as part of */
  /* ...shmalloc_init() */
  midway_setup_exception_handler();
#endif    
  if (IS_DBG_MISC_INIT) {
    StatusMsg("shmalloc initialization finished.");
  }
  
  /* Initialize support for non-EC models */
  mid_updateq_init ();
  if (IS_DBG_MISC_INIT) {
    StatusMsg("mid_updateq_init() finished");
  }
  mid_page_init ();
  if (IS_DBG_MISC_INIT) {
    StatusMsg("mid_page_init() finished");
  }
  
  CommOpen();
  if (IS_DBG_MISC_INIT) {
    StatusMsg("CommOpen finished");
  }
  
#ifdef CTHREAD_DEBUG
  cthread_debug = 1; /* prepare the shovels! */
#endif CTHREAD_DEBUG

  /* initialize profiling.  We want to crank the thread, and allocate
     all the space for the histogram and callarcs early, so it's not
     charged to the wall clock time for the program. */
  if (prof_init) {
    int onoff = 0;
    krc = gprof_mon_start();
    if (krc == KERN_SUCCESS) {
      /* turn off profiling without dumping profile! */
      bsd_mon_switch(&onoff);
    }
  }

  /* Start profiling */
  if (midway_profiling) {
    krc = midway_start_profile ();
    if (krc != KERN_SUCCESS)
      exit(krc);
  }

  /* soak cpu */
  if (midway_soak) {
    cthread_detach(cthread_fork(idle_thread, (any_t) 0));
  }

  do { /* forever */

    if (hindex == 1) { /* main processor */

      /* We must move the thread's stack to a known region
       * in case compiler generates a dirtybit write for data on the stack.
       * WAS 11Nov93
       */
      mid_rsched_move_stack (main, argc2, argv2, envp);

    } else { /* non-main procesor */
      if (IS_DBG_LOCK_FORK) {
	StatusMsg("Calling mid_rsched_do()");
      }
      mid_rsched_do(); /* find work, and do it */
    }
  
    /* Stop profiling */
    if (midway_profiling) {
      krc = midway_stop_profile();
    }

  } while (hindex != 1);   /* let master exit, slaves continue looping */


#if !defined(LIMBO)

  if (midway_prof_wait > 0) {
    sprintf(smsg, "Main returned, pausing %d seconds before exiting.", midway_prof_wait);
    StatusMsg(smsg);
    timer_pause_ms(midway_prof_wait*1000);
  } else {
    StatusMsg("Main returned  -- exiting.");
  }
  midway_exit(0);
#else
  StatusMsg("Main returned -- going into limbo.");
  /* we will now wait for all threads to exit... may be a Very Long Time */
  /*
  ** This doesn't erase the paging files.   WAS 29Jun93 
  ** Exit should be via midway_exit.
  */
#endif
}   /* main */


int midway_frozen = 1;
freeze_for_debugger()
{
  int pid  = getpid();
  int i = 0;
  while (midway_frozen)  {
    if (i % 50 == 0)  {
      fprintf(stderr,"midway child (%d) frozen\n", pid);
    }
    i++;
    timer_pause_ms(100);
  }
}

static int
death_thread(ss_port)
     mach_port_t ss_port;	/* who is our status server? */
{
  kern_return_t krc;

/*#if WIRE_DA_THREADS*/		/* always wire? */
  cthread_wire();		/* under 3.0 wire down for now */
/*#endif*/

  krc = ss_waitfor_exit(ss_port);
  if (krc != KERN_SUCCESS) {

    /*
     * Error return from the netmsgserver
     * Don't kill the process, just terminate this thread.
     * Also route all further I/O locally.
     */
    (void) midway_uxpass_local();
    mach_error("ss_waitforexit:",krc);
    cthread_exit(krc);
  }

  /* dump profile, if profiling was enabled */
  if (midway_profiling) {
    krc = midway_stop_profile();
  }
  if (midway_prof_dump_in_progress) {
    mutex_lock(&midway_prof_mutex);
    midway_prof_dump_waiters++;	/* let dump know we exist */
    while (midway_prof_dump_in_progress) {
      condition_wait(&midway_prof_cond, &midway_prof_mutex);
    }
    midway_prof_dump_waiters--;
  }

  /* erase paging files */
  vm_shared_term();
  vm_dirty_term();

  /* don't try to print in case ss hung/dead? */
  _exit(0);			/* emergency exit */
}

int midway_soak_priority = 30; /*XXX*/

unsigned int midway_idle_count = 0;

static int
idle_thread()
{
  kern_return_t krc;
  int i = 0;

  cthread_wire();

  midway_set_good_priority(midway_soak_priority);

  while(1) {
    midway_idle_count++;
#if 0
    cthread_yield();
#else
    if ((midway_idle_count % 100000) == 0) {
      cthread_yield();
    }
#endif
  }
}/* idle(soak)_thread*/

void
midway_exit(int rc)
{
  kern_return_t krc;
  extern mach_port_t ss_port;

  ss_exit(ss_port);		/* tell everyone to die */

  if (midway_profiling) {
    krc = midway_stop_profile();
  }
  if (midway_prof_dump_in_progress) {
    mutex_lock(&midway_prof_mutex);
    midway_prof_dump_waiters++;	/* let dump know we exist */
    while (midway_prof_dump_in_progress) {
      condition_wait(&midway_prof_cond, &midway_prof_mutex);
    }
    midway_prof_dump_waiters--;
  }

  /* erase paging files */
  vm_shared_term();
  vm_dirty_term();

  _exit(rc); /* test _exit for ATM 31jan94,mjz */
}

kern_return_t
midway_start_profile(void)
{
  kern_return_t krc;
  extern int gprof_sample_type;

  if (IS_DBG_MISC_INIT)
#ifdef PROFILE
    StatusMsg("enabling profiling for this process");
#else
    StatusMsg("MIDWAY was not compiled for profiling, instr. counts only!");
#endif

  if (midway_vm_profiling) {
      StatusMsg("enabling VM Fault profiling for this process");
      gprof_sample_type = SAMPLED_PC_VM_FAULTS;      
  }

  krc = gprof_mon_start ();
  if (krc != KERN_SUCCESS)
    mach_error("from gprof_mon_start ",krc);
  else
    midway_profiling = 1;

  return (krc);
}

kern_return_t
midway_stop_profile(void)
{
  char *cp;
  char255 fileName, hostName;
  kern_return_t krc;
  extern int Comm_index;


  if (!midway_profiling)
    return (KERN_SUCCESS);

  /* move this set here from at the end, because there can be two calls
     to midway_stop_profile running essentially concurrently at the end
     if both the exit code & the death thread are running.
     Try to stop the death thread (or exit code) from exiting until
     the profile dump is done, though...
     */

  midway_profiling = 0;
  mutex_lock(&midway_prof_mutex);
  midway_prof_dump_in_progress = 1;
  mutex_unlock(&midway_prof_mutex);

  if (IS_DBG_MISC_INIT)
    StatusMsg("ending profiling for this process");

  gethostname(hostName,sizeof(hostName));
  cp = (char *)index(hostName,'.');
  if (cp != NULL) *cp = '\0';

  if (midway_prof_count == 0) {
    midway_prof_count = 1;
    sprintf(fileName, "%s-%d.%s.out", 
	    hostName, Comm_index,
	    (midway_vm_profiling) ? "vmon" : "gmon");
  } else {
    midway_prof_count++;
    sprintf(fileName, "%s-%d.%s.out-%d", 
	    hostName, Comm_index, 
	    (midway_vm_profiling) ? "vmon" : "gmon",
	    midway_prof_count);
  }

  krc = gprof_mon_stop (fileName);
  if (krc != KERN_SUCCESS)
    mach_error("from gprof_mon_stop ",krc);

  mutex_lock(&midway_prof_mutex);
  midway_prof_dump_in_progress = 0;
  mutex_unlock(&midway_prof_mutex);
  if (midway_prof_dump_waiters) {
    condition_broadcast(&midway_prof_cond);
  }

  return(krc);
}
