/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */

/*
 * rsched.c: how to do a remote cthread_fork()...
 *
 * 19 Jun 91 (mjz)
 * 12 Jul 91 (mjz): add WAIT_FOR_END(numprocs) support
 * 19 Mar 92 (mjz): fix debugging statements; for now leverage off lock::fork
 *  3 Oct 92 (mjz): the grand renaming.  Again 10/4 create->create_thread
 * 11 Nov 93 (was): Added mid_rsched_move_stack to move a thread's stack to a known region.
 *  8 Jan 94 (was): Limit size of application's stack
 * 22 Feb 94 (was): Record application's thread id
 */
#include <sys/types.h>		/* u_long, etc. */
#include "rsched.h"
#include <mach.h>		/* for cthreads */
#include <cthreads.h>		/* for locking */
extern vm_size_t cthread_stack_size;
#include "comm.h"		/* for M_SCHED_OP */
#include "smexecd_types.h"	/* char255 */
#include "vm.h"			/* for vm_private_xxx vars */
#include "shmalloc.h"		/* for MAX_STACK_SIZE */
#include "timer.h"
#include "midway-dbg.h"
#ifndef shared			/* so I can compile with normal compilers */
#define shared
#endif
#include "midway.h"		/* prototypes, lock types... */


extern int Comm_hnum;		/* number of hosts participating */
extern int Comm_index;		/* number of this host */
extern u_long LocalClock;	/* for global timing */
extern void StatusMsg();	/* to report errors */
extern task_t midway_task_self;	/* cache results of mach_task_self() */

static sched_info_t *si = 0;
static u_long sched_ptr = 0;	/* wiper for RR external schedule policy */
static mutex_t sched_lock;	/* just in case */
static int si_max;		/* Comm_hnum + 1... so modulo fun works */

static int join_num = 0;	/* for silly Wait request...  */
static condition_t join_c;	/* so a processor can wait */

static void (*saveproc)();	/* for waiting */
static u_long savearg;		/* likewise */
static mutex_t waitlock;	/* to enforce policy of one waiter/processor */
static int mid_rsched_waiting;	/* likewise */

/*
 * for 3.0 cthreads, ensure all threads wired for now 2sep92mjz
 */
typedef struct {
  void(*proc)();
  u_long arg;
  struct mutex m;
  struct condition c;
} forking_arg_t;


/*
 * Move a thread's stack to a pre-allocated region
 * which has a dirtybit template stored at the base.
 * This is necessary in case the compiler generates a
 * dirtybit update call for a write to stack space.
 */
void
mid_rsched_move_stack (void (*proc)(), u_long arg0, u_long arg1, u_long arg2)
{
  register void (*saved_proc)() = proc;
  register u_long saved_arg0 = arg0;
  register u_long saved_arg1 = arg1;
  register u_long saved_arg2 = arg2;
  vm_size_t stack_size;
  kern_return_t krc;

#ifdef PAGE_BASED_RC
  /* Save application thread's id to allow suspend/resume */
  { extern mach_port_t paged_appl_thread_id;
    paged_appl_thread_id = mach_thread_self();
  }
#endif

  /* Allocate a stack from private memory.
   * Allocate double the stack size, to allow 
   * stack base to be rounded up to start on a known boundary.
   *
   * NOTE: Limit stack size to half of a region so we can allocate
   * private space for it. Stacks larger than this are typically 
   * unnecessary and should be reduced via the limit command. 
   * WAS 8Jan94
   */
  if (vm_private_stackP == 0) {
    stack_size = 2 * cthread_stack_size;
    if (stack_size > MAX_STACK_SIZE) stack_size = MAX_STACK_SIZE;
    krc = VM_ALLOCATE (midway_task_self, &vm_private_stackP, 
		       stack_size, 1); /* anywhere = TRUE */
    ASSERT (krc == KERN_SUCCESS);
  }

  /* Copy this thread's stack & change the stack pointer! 
   * 
   * NOTE! Currently, this code assumes there is exactly 1 application
   * thread which uses this stack space. If more than 1 thread is ever 
   * used this must be handle stack allocation/deallocation (instead 
   * of simply reusing the same stack space each time).   WAS 11Nov93
   */
  imove_stack(vm_private_stackP);

  /* invoke the thread's routine */
  saved_proc (saved_arg0, saved_arg1, saved_arg0);

  /* We should free this thread's stack... */

#ifdef PAGE_BASED_RC
  /* Reset application thread's id now that it has terminated */
  { extern mach_port_t paged_appl_thread_id;
    paged_appl_thread_id = MACH_PORT_NULL;
  }
#endif
}

void
mid_rsched_wrapper(fa)
     forking_arg_t *fa;
{
  void (*proc)();
  u_long arg;
  extern int midway_user_priority;

#if WIRE_DA_THREADS
  cthread_wire();
#endif

#if SET_PRIORITY
  if (midway_user_priority) { /* defined in comm.c.  ugh. */
    midway_set_good_priority(midway_user_priority);
  }
#endif

  mutex_lock(&(fa->m));
  proc = fa->proc;
  arg = fa->arg;
  mutex_unlock(&(fa->m));
  condition_signal(&(fa->c));

  mid_rsched_move_stack (proc, arg, 0, 0);

#ifdef PAGE_BASED_RC
  /*
   * Be sure all updates are flushed before terminating this thread.
   */
  paged_flush_updates();
#endif

  cthread_exit(0);
} /* mid_rsched_wrapper */

cthread_t
mid_wired_cthread_fork(proc, arg)
     void (*proc)();
     u_long arg;
{
  forking_arg_t fa;
  cthread_t ct;

  fa.proc = proc;
  fa.arg = arg;
  mutex_init(&(fa.m));
  MIDWAY_MUTEX_SET_NAME(&(fa.m), "rsched_wrapper_mutex");
  condition_init(&(fa.c));
  mutex_lock(&(fa.m));
  ct = cthread_fork(mid_rsched_wrapper, &fa);
  /* wait so fa doesn't dissapear before child can read */
  condition_wait(&(fa.c), &(fa.m)); 
  mutex_unlock(&(fa.m));
  return(ct);
} /* mid_wired_cthread_fork */

/*
 * try to wire everything in sight
 */
#if MIDWAY_WIRE_MEMORY
  /* attempt to wire everything now in low memory; code, data, etc. */
/*
  doesn't work on mc16; we wire down 29 meg, and croak.
*/
#ifdef large /* won't work. for one thing, writes all over ATM memory */
#define START_ADDR 0
#define END_ADDR vm_shared_base
#endif

/* try just text and statics */
/*#define START_ADDR (&__start) too big with chec*//* mips, at least */
/*#ifdef small*/
#define START_ADDR (0x400000)	/* MIPS SPECIFIC!!! */
#define END_ADDR (&etext) /* end still too far for mc16 */
/*#endif*/

static wired_it_all = 0;
static wire_it_all(name)
     char *name;
{ 
  kern_return_t krc;
  extern task_t midway_task_self;
  extern mach_port_t midway_host_priv;
  extern vm_offset_t vm_shared_base;
  vm_offset_t addr;
  vm_size_t len;
  vm_prot_t protection;
  vm_prot_t max_protection;
  vm_inherit_t inheritance;
  boolean_t shared;
  mach_port_t object_name;
  vm_offset_t offset;
  char255 smsg;
/*  extern int __start; too big with chec. */
  extern int etext;
  extern int end;

  if(wired_it_all)		/* not clear... maybe should do again... */
    return(0);
  wired_it_all = 1;

  for (addr = START_ADDR; addr < END_ADDR; addr += len) {
    len = (vm_size_t) ((unsigned int) END_ADDR - (unsigned int) addr);
    krc = vm_region(midway_task_self, &addr, &len, &protection,
		    &max_protection, &inheritance, &shared, &object_name,
		    &offset);
    if (krc == KERN_NO_SPACE)
      break;			/* done */
    if (krc != KERN_SUCCESS) {
      sprintf(smsg,"%s vm_region fails, addr 0x%lx len %u",
	      name, addr, len);
      StatusMsg(smsg);
      mach_error("vm_region",krc);
    }
    
    krc = vm_wire(midway_host_priv, midway_task_self, addr, 
		  len, protection); /*VM_PROT_READ|VM_PROT_WRITE*/
    if (krc != KERN_SUCCESS) {
      sprintf(smsg,"%s vm_wire fails, addr 0x%lx len %u",
	      name, addr, len);
      StatusMsg(smsg);
      mach_error("vm_wire",krc);
    }
  }
} /* wire_it_all */
#endif

/*
 ************************************************************************
 *		    m i d _ r s c h e d _ i n i t			*
 *----------------------------------------------------------------------*
 * If called by  main thread, initialize the rsched data structure.	*
 * else, initialize waitlock only.					*
 * FOR NOW, ALL CREATE CALLS CAN ONLY BE MADE FROM THREADS ON MAIN PROC *
 ************************************************************************
 */
int
mid_rsched_init()
{
  kern_return_t krc;

  if (si) {
    StatusMsg("rsched_init called more than once! Ignoring...");
    return(-1);
  }

  if (Comm_index != 1) {	/* other processors... */
    waitlock = mutex_alloc();
    MIDWAY_MUTEX_SET_NAME(waitlock, "rsched.c:waitlock");
    mid_rsched_waiting = 0;
  }
  else {			/* main processor... */
    si_max = Comm_hnum + 1;
    si = (sched_info_t *) CALLOC(si_max, sizeof(sched_info_t));

    if (si == 0) {		/* rats */
      StatusMsg("rsched_init: out of memory of sched_info table :-(. Barfing.");
      ASSERT(si == 0);		/* XXX */
    }

    sched_lock = mutex_alloc();	/* check for valid return? --mjz */
    MIDWAY_MUTEX_SET_NAME(sched_lock, "rsched.c:sched_lock");

    join_c = condition_alloc();
  }

  return(0);
} /* mid_rsched_init */


/*
 ************************************************************************
 *		m i d _ r s c h e d _  w a i t _ a r g			*
 *----------------------------------------------------------------------*
 * called by the locking layer to set up the arguments necessary for    *
 * mid_rsched_do to complete (in particular the procedure & argument)	*
 ************************************************************************
 */
mid_rsched_wait_arg(proc, arg)
     void (*proc)();
     u_long arg;
{
  char255 smsg;
  if (mid_rsched_waiting) {
    saveproc = proc;
    savearg = arg;
    if (IS_DBG_LOCK_FORK) {
      sprintf(smsg,"rsched_wait_arg: saved proc %x and arg %x for _do",
	      proc, arg);
      StatusMsg(smsg);
    }
  }
  else StatusMsg("rsched_wait_arg: no _do thread waiting, dropping...");
} /* mid_rsched_wait */

/*
 ************************************************************************
 *		    m i d _ r s c h e d _  d o				*
 *----------------------------------------------------------------------*
 * Called by non-main-processor when waiting for work...		*
 * (was mid_rsched_wait; but we now do the given task too!		*
 * For now, only one outstanding call per processor... and so we can    *
 * easily simulate a join with this brain damaged implementation, we 	*
 * also don't allow another wait until this work is done.		*
 ************************************************************************
 */
void
mid_rsched_do()
{
  cthread_t thread_id;
  char smsg[255];

  if (Comm_hnum == 1) {		/* uh, we've been called on the main proc */
    StatusMsg("mid_rsched_do called on main processor -- ignoring.");
    return;
  }

  mutex_lock(waitlock);		/* enforce policy */
  mid_rsched_waiting = 1;	/* allow args to be saved */
  CommRequest(0, M_SCHED_OP, 1 /*"best guess"*/, 0, 0); /* request work */
  mid_rsched_waiting = 0;	/* done! */

  if (IS_DBG_LOCK_FORK) {
    sprintf(smsg,"rsched_do: about to c_fork(%x,%x)", saveproc, savearg);
    StatusMsg(smsg);
  }
  thread_id = mid_wired_cthread_fork(saveproc, savearg); /* VA VA Voom! */
  if (IS_DBG_LOCK_FORK) {
    sprintf(smsg,"rsched_do: about to c_join(%x)", thread_id);
    StatusMsg(smsg);
  }
  cthread_join(thread_id);	/* wait for completion */
  if (IS_DBG_LOCK_FORK) {
    StatusMsg("rsched_do: after cthread_join, returning");
  }

  mutex_unlock(waitlock);

#if XXMIDWAY_WIRE_MEMORY
  { kern_return_t krc;
    extern task_t midway_task_self;
    extern mach_port_t midway_host_priv;
    extern vm_offset_t vm_shared_base;
    /* attempt to wire everything now in low memory; code, data, etc. */
    krc = vm_wire(midway_host_priv, midway_task_self, 0, 
		  vm_shared_base, VM_PROT_READ|VM_PROT_WRITE);
    if (krc != KERN_SUCCESS) {
      sprintf(smsg,"rsched do vm_wire fails, addr 0x%lx len %u",
	      0, vm_shared_base);
      StatusMsg(smsg);
      mach_error("rsched_do:vm_wire",krc);
    }
  }
#endif
#if MIDWAY_WIRE_MEMORY
  wire_it_all("rsched_do");
#endif
} /* mid_rsched_do */

/*
 ************************************************************************
 *		    m i d _ r s c h e d _ r w a i t			*
 *----------------------------------------------------------------------*
 * called by locking routines... when a remote processor is initiated,  *
 * or an entire thread tree dies...					*
 ************************************************************************
 */
comm_request_t
mid_rsched_rwait(host,thread,tid)
     u_long	host;		/* host wants to wait */
     u_long	thread;		/* the thread id */
     u_long	tid;		/* the transaction id */
{
  char smsg[255];
  comm_request_t rc = M_R_UNCHANGED;

  if (Comm_index != 1) {
    StatusMsg("rsched_wait called on non-main processor, dropping request");
    return(M_R_DROPPED);
  }

  mutex_lock(sched_lock);
  if (si[host].tid == tid) {	/* sigh, duplicate */
    if (si[host].proc != 0) {	/* gotta resend */
      if (IS_DBG_LOCK_WNG) {	/* don't babble, may not be an error */
	sprintf(smsg,
 	  "rwait: Duplicate request from hst=%d thd=%d tid=%d, resending answer.",
	  host, thread, tid);
	StatusMsg(smsg);
      }
      midRemoteRelease(host,thread,tid,si[host].owakeupt,
		       si[host].proc,si[host].arg);
      rc = M_R_DONE;
    }
    /* else just waiting */
  }
  else {			/* set it up */
    if (IS_DBG_LOCK_FORK) {
      sprintf(smsg,"rwait: about to enter in waiter table h=%lu t=%lu tid=%lu",
	      host, thread,tid);
      StatusMsg(smsg);
    }

    if (si[host].proc && (join_num > 0)) {
      /* si[host].proc ensures we don't decrement the INITIAL time through. */
      /* (could happen if Create is called, and a while later the sched */
      /*  packet comes in, which is then assigned to the waiting create) */

      /* someone is waiting for procs to end */
      join_num--;		/* reduce */
      if (IS_DBG_LOCK_FORK) {
	sprintf(smsg,"rwait: join_num now %d", join_num);
	StatusMsg(smsg);
      }
      if (join_num <= 0) {	/* should never be negative */
	condition_broadcast(join_c); /* signal the waiter */
      }
    }

    si[host].thread = thread;
    si[host].tid = tid;
    si[host].proc = 0;
    si[host].arg = 0;
    rc = M_R_QUEUED;
  } /* set up new waiting processor */

  mutex_unlock(sched_lock);
  return(rc);
} /*mid_rsched_rwait*/



static int
fork_on_idle_processor(func, arg)
    void (*func)();
    u_long arg;
{
    u_long p;
    int i;
    char255 smsg;
    
    mutex_lock(sched_lock);

    for (i = 0; i < si_max; i++)  {
	p = (sched_ptr + i) % si_max;
	if (si[p].tid && !(si[p].proc)) { 
	    sched_ptr = (p+1)%si_max; /* update */
	    break; 
	}
    } 

    if (i == si_max)  {	/* no free processors */
	mutex_unlock(sched_lock);
	return 0;
    } else {
	if (IS_DBG_LOCK_FORK) {
	    sprintf(smsg,"Create: releasing proc %u th %u tid %u time %u func %x(%x)",
		    p, si[p].thread, si[p].tid, si[p].wakeuptime, func, arg);
	    StatusMsg(smsg);
	}	
	si[p].proc = func;
	si[p].arg = arg;
	si[p].owakeupt = si[p].wakeuptime;
	si[p].wakeuptime = LocalClock;
	midRemoteRelease(p /*host*/, si[p].thread, si[p].tid, si[p].owakeupt,
			 func, arg);
	CommSetTranState(p, si[p].thread, si[p].tid, M_R_DONE);
	join_num++; /* for now, inc on create/dec on wait */
	mutex_unlock(sched_lock);
	return 1;
    }
}



/*
 ************************************************************************
 *		m i d w a y _ c r e a t e _ t h r e a d			*
 *----------------------------------------------------------------------*
 * Called by a thread on the main processor to create a thread on a     *
 * (possibly) remote processor						*
 ************************************************************************
 */
void
midway_create_thread(func,arg)
     void (*func)();		/* function to be performed */
     u_long arg;		/* argument */
{
    int trys;
    char255 smsg;
    
    if (si == 0) {
	StatusMsg("Create called, but rsched_init not done...");
	return;
    }

#if XXMIDWAY_WIRE_MEMORY
    { kern_return_t krc;
      extern task_t midway_task_self;
      extern mach_port_t midway_host_priv;
      extern vm_offset_t vm_shared_base;
      /* attempt to wire everything now in low memory; code, data, etc. */
      krc = vm_wire(midway_host_priv, midway_task_self, 0, 
		    vm_shared_base, VM_PROT_READ|VM_PROT_WRITE);
      if (krc != KERN_SUCCESS) {
	sprintf(smsg,"rsched do vm_wire fails, addr 0x%lx len %u",
		0, vm_shared_base);
	StatusMsg(smsg);
	mach_error("rsched_do:vm_wire",krc);
      }
    }
#endif
#if MIDWAY_WIRE_MEMORY
    wire_it_all("create_thread");
#endif

    if (IS_DBG_LOCK_FORK) {
	sprintf(smsg, "Create: func=%x, arg=%x", func, arg);
	StatusMsg(smsg);
    }

    if (Comm_index != 1) {
	StatusMsg("rsched_wait called on non-main processor, local fork.");
	goto local_fork;
    }

    if (si_max > 1)  {
	/* find next available processor */
	for (trys = 0; trys < 25; trys++)  {
	    if (fork_on_idle_processor(func, arg))  {
		return;
	    } else {
#ifdef VERBOSITY
		fprintf(stderr,"Create - no procs, sleeping\n");
		fflush(stderr);
		StatusMsg("Create - no procs, sleeping");
#endif	
		timer_pause_ms(50); /* wait 50 ms */
	    } 
	}
    }
    
 local_fork:
    StatusMsg("Create: Could not find free remote processor."); 
    StatusMsg("Check that your smexec file defined sufficient processors for this execution.");
    StatusMsg("You must have 1 for the master +1 for each slave started via midway_create_thread");
    /* XXX this is broken, since WaitForEnd will not work --2sep92 mjz */
    botch("local_fork");	/* enforce botchedness. 19dec92 */
    cthread_detach(mid_wired_cthread_fork(func, arg));
    return;    

}   /* midway_create */


/*
 ************************************************************************
 *		m i d w a y _ w a i t _ f o r _ e n d			*
 *----------------------------------------------------------------------*
 * Simulate the "WAIT_FOR_END" ANL macro call.  Wait for n "remote"	*
 * processes to finish...						*
 * 12jul: well, I looked at one ANL implementation, and they didn't even*
 * use "n", they just waited for any that had been created to finish.	*
 * Let's just do that for now...					*
 ************************************************************************
 */
void
midway_wait_for_end(n)
     u_long n;			/* number of procs to wait for */
{
  char255 smsg;

  if (Comm_index != 1) {
    StatusMsg("WaitForEnd called on non-main processor -- ignoring.");
    return;
  }
  if (si == 0) {
    StatusMsg("WaitForEnd called, but rsched_init not done...");
    return;
  }

  mutex_lock(sched_lock);
#ifdef FIRST_WAY
  if (join_num != 0) {
    sprintf(smsg,"WaitForEnd called, but join_num was %d not 0!", join_num);
    StatusMsg(smsg);
    mutex_unlock(sched_lock);
    return;
  }
#endif

  if (IS_DBG_LOCK_FORK) {
    sprintf(smsg,"WaitForEnd: waiting for %u procs to finish.",n);
    StatusMsg(smsg);
  }

#ifdef FIRST_WAY
  join_num = (int) n;
  condition_wait(join_c, sched_lock);

  if (join_num != 0) {
    sprintf(smsg, "WaitForEnd: join_c signalled, but join_num was %d not 0!",
	    join_num);
    StatusMsg(smsg);
  }
#endif

  while (join_num > 0) {
    if (IS_DBG_LOCK_FORK) {
      StatusMsg("WaitForEnd: about to condition_wait(join_c, sched_lock)");
    }
    condition_wait(join_c, sched_lock);
    if (IS_DBG_LOCK_FORK) {
      sprintf(smsg,"WaitForEnd: condition signalled, join_num now %d", join_num);
      StatusMsg(smsg);
    }
  }

  mutex_unlock(sched_lock);
  return;
} /* midway_wait_for_end */
