/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/* 
 * shmalloc/shfree:
 *  Dynamic memory allocation for MIDWAY.
 *
 *  The basic approach used here is to use a malloc-like algorithm for
 *  allocating memory from large regions.  All data in a region is
 *  considered to have the same "line size" for the software managed
 *  dirty bits.  Regions are allocated on aligned REGION_SIZE boundaries
 *  and are of size REGION_SIZE.  Each region has a region descriptor which is
 *  kept in a contiguous, statically allocated table.  A simple masking function
 *  can be used to map from any address in a region to its region descriptor.
 *  The region descriptor includes a reference to the dirty bit array, of which
 *  there is one word per line in the region.
 *
 *  Some of the information maintained internally is kept in log form and
 *  pre-shifted to make the access paths faster.
 *
 *  This file was derived from the GNU malloc distributed with gnu-emacs.
 *
 *  To do:
 *	Allocate cache information lazily.
 *
 * 22 Aug 91 (mjz): get latest exception handling code.  Use midway_mem
 *		instead of midway header file.
 * 17 Sep 91 (mjz): add comm interface code midAlloc{,Info}{Request,Reply}
 * 12 Dec 91 (mjz): midway_set_up_exeception from midway_set_region_proc().
 * 16 Dec 91 (mjz): fix region_find to only look in *our* regions!
 * 13 Jul 92 (mjz): add more debug info to catch_exception_raise
 * 16 Jul 92 (mjz): spin rather than exit in botch so we can debug
 * 		    use SHMALLOC_ALLOCATION_BASE for initial ria_info...
 * 30 Jul 92 (mjz): finally put some allocate info printfs under conditionals
 *  3 Aug 92 (was): changed shmalloc to allocate on line size boundaries
 *  3 Aug 92 (mjz): need to handle multiple page faults for region gracefully
 *  4 Aug 92 (was): fixed bug in sizeof(region_info->nextf) array
 * 11 Aug 92 (was): added clock_t type
 * 15 Aug 92 (was): added page tables to region_info desc.
 * 19 Aug 92 (bnb): removed exception handling to separate module. lots more.
 * 13 Oct 92 (mjz): fix allocinfo to return ri->base v. addr; remote needs
 *		to know real base to be able to set up ri.
 * 15 Oct 92 (was): added assertion checks after mid_pbuf_get
 * 16 Oct 92 (was): Optimized region_info_allocate (via loop unrolling)
 * 20 Oct 92 (was): No longer preallocate regions with cacheline sizes 8 & 32
 * 27 Oct 92 (was): Changed shmalloc_debug to a macro for benchmarking
 *  9 Nov 92 (was): Minor corrections for allocating 8MB regions.
 * 15 Dec 92 (was): Lazy initialization of dirtybits from dirtybit region
 *                  Also made 8MB regions the default.
 * 28 Dec 92 (was): malloc from dirtybit region must be in page_size multiples
 * 19 Jan 93 (was): use ADDR_TO_PAGENO instead of /vm_page_size
 * 24 Feb 93 (was): Lazy allocation of regions at slaves & alloc 1 at a time.
 *                  Also bumped max number of regions to 64
 * 20 Sep 93 (was): changed region allocation to place dbit template at top
 * 10 Oct 93 (was): broke out shmalloc/shfree from region allocate/free 
 *                  revised shmalloc to use boundary tags instead of buddy-system
 */

#include <sys/types.h>		/* for comm.h */
#include <signal.h>
#define KERNEL 1
#include <sys/time.h>
#undef KERNEL
#include <sys/resource.h>
#include <mach.h>
#include <mach/kern_return.h>
#include <stdio.h>
#include <cthreads.h>
#include "midway_types.h"
#include "midway_mem.h"
#include "shmalloc.h"
#include "pbuf.h"
#include "vm.h"
#include "comm.h"		/* barf.  should have interface include */
#include "smexecd_types.h"	/* char255 */
#include "pagemgr.h"		/* prototypes for calling page mgr rtns */
#include "midway-dbg.h"		/* DBG_IS_MISC_MAIF */
#include "midway.h"		/* prototypes */


#ifdef BENCHMARK
  #define SHMALLOC_DEBUG  0	/* =1 to compile in debugging code, =0 for none */
  #define SHMALLOC_STATS  0     /* =1 to collect usage stats, =0 for none */
#else
#ifdef STATS
  #define SHMALLOC_DEBUG  1	/* =1 to compile in debugging code, =0 for none */
  #define SHMALLOC_STATS  1     /* =1 to collect usage stats, =0 for none */
#else
  #define SHMALLOC_DEBUG  1	/* =1 to compile in debugging code, =0 for none */
  #define SHMALLOC_STATS  0     /* =1 to collect usage stats, =0 for none */
#endif 
#endif 


#if SHMALLOC_DEBUG > 0
  #define MASSERT ASSERT
  #define MASSIGN(x) x
#else
#ifdef DEBUG_USER
  #define MASSERT 
  #define MASSIGN(x) x
#else
  #define MASSERT 
  #define MASSIGN
#endif
#endif



/*
 * Define boundary tags for each allocated block of shared memory.
 * The tags are stored in the cache line before and after the block.
 * If the tags indicate both the prior and next blocks are allocated
 * then the tag exactly matches the mhead defined below.
 *
 * If the tag indicates the next block is free, then the pointers for
 * the free list for the next blocks are stored in the first 8 bytes
 * of the block.
 *
 * If the tag indicates the previous block is free, then a pointer
 * to the mhead describing that block is stored in the first 4 bytes
 * before the mhead. 
 *
 * If the tag indicates the previous block is of the minimum size, 
 * then the previous block is free, and the mhead for the block is 
 * MIN_ALLOC bytes before the mhead. The minimum size tag is necessary
 * to allow doubly-linked list of free pointers to be stored in the
 * block without increasing the size of the mhead or the min block.
 */
#define MIN_USER_ALLOC   8	/* minimum size allocation 
				 * (MUST be >= sizeof(mhead) */
#define MIN_ALLOC       (MIN_USER_ALLOC + sizeof(mhead_t))
#define MIN_ALLOC_INDEX  0
#define MIN_ALLOC_SHIFT	 4 	/* power of 2 = MIN_ALLOC */
#define MAX_ALLOC_SHIFT  REGION_SHIFT /* power of 2 = MAX_REGION */
#define MAX_ALLOC_INDEX  4


typedef struct mhead_s {
  u_char  m_magic;		/* validation byte */
  u_char  m_prev_tag;		/* boundary tag for previous piece */
  u_char  m_next_tag;		/* boundary tag for next piece */
  u_char  m_next_index;		/* power of 2 index for next piece */
  u_long  m_next_size;		/* size of next piece */
} mhead_t;

#define ISMAGIC ((u_char) 0xb9) /* magic number for validation checking */
#define ISALLOC ((u_char) 0xf7) /* magic number that implies allocated */
#define ISFREE  ((u_char) 0x54) /* magic number that implies free block */
#define ISMINSZ ((u_char) 0x85) /* magic number implies free block size = MIN_ALLOC */


/*
 * When the tag indicates the next piece is free
 * the words immediately following the mhead is used to hold free list ptrs.
 */
#define MFREE_NEXT(mP) (*(mhead_t **)(((int)mP) + sizeof(*mP)))
#define MFREE_PREV(mP) (*(mhead_t **)(((int)mP) + sizeof(*mP) + sizeof(mP)))


/*
 * Define macros to manipulate items on the free list
 */
#define MFREE_ENQUEUE(rP,mP)					\
{ register mhead_t *tprevP;					\
  register mhead_t *tnextP = MFREE_FIRST(rP,mP->m_next_index);	\
  if (tnextP == NULL) {						\
    MFREE_PREV(mP) = MFREE_NEXT(mP) = mP;			\
  } else {							\
    tprevP = MFREE_PREV(tnextP);				\
    MFREE_PREV(tnextP) = MFREE_NEXT(tprevP) = mP;		\
    MFREE_NEXT(mP) = tnextP;					\
    MFREE_PREV(mP) = tprevP;					\
  }								\
  MFREE_FIRST(rP,mP->m_next_index) = mP;			\
}

#define MFREE_DEQUEUE(rP,mP)					\
{ register mhead_t *tprevP;					\
  register mhead_t *tnextP = MFREE_NEXT(mP);			\
  if (tnextP == mP) {						\
    MASSERT (MFREE_FIRST(rP, mP->m_next_index) == mP);		\
    MFREE_FIRST(rP, mP->m_next_index) = NULL;			\
 } else {							\
    tprevP = MFREE_PREV(mP);					\
    MFREE_PREV(tnextP) = tprevP;				\
    MFREE_NEXT(tprevP) = tnextP;				\
    if (MFREE_FIRST(rP, mP->m_next_index) == mP)		\
      MFREE_FIRST(rP, mP->m_next_index) = tnextP;		\
  }								\
}


/*
 * When the tag indicates the previous block is free (and not minimum size)
 * the word immediately before the mhead holds a ptr to the prev mhead
 */
#define MBLOCK_PREV(mP) (*(mhead_t **)(((int)mP) - (sizeof(mhead_t *))))

/*
 * When the tag indicates the previous block is free
 * and is the minimum size, then the a prev ptr is not stored.
 */
#define MBLOCK_MIN_PREV(mP) ((mhead_t *)(((int)mP) - MIN_ALLOC))


/*
 * The next block is always at the current block + size 
 */
#define MBLOCK_NEXT(mP) ((mhead_t *)(((int)mP) + (mP)->m_next_size))

#define MBLOCK_SIZE_NEXT(mP,sz) ((mhead_t *)(((int)mP) + sz))



/* 
 * Map from mhead to pointer 
 * returned ptr must start on line size boundary
 */
#define MHEAD_TO_PTR(mhead,ri,ptr)                                  \
{ register u_long line_size = RI_LINE_SIZE_TO_BYTES(ri->line_size); \
  if (line_size < MIN_USER_ALLOC) line_size = MIN_USER_ALLOC;                 \
  ptr = ((char *)(((int)(mhead)) + line_size));                     \
}

/* 
 * Map from pointer to its mhead
 * mhead resides on previous line size boundary
 */
#define PTR_TO_MHEAD(ptr,ri,mhead)                                  \
{ register u_long line_size = RI_LINE_SIZE_TO_BYTES(ri->line_size); \
  if (line_size < MIN_USER_ALLOC) line_size = MIN_USER_ALLOC;                 \
  mhead = ((mhead_t *) (((int)(ptr)) - line_size));            \
}



/*
 * A list of free blocks is maintained per region.
 * The region descriptor contains a pointer to a list of free block ptrs.
 * The free blocks are maintained in a circular doubly-linked list.
 * free_block[i] points to the head of the list.
 */
#define MFREE_FIRST(rP,index)  ((mhead_t *)(riP->free_block[index]))


/*
 * Define free block index macros
 * There are exactly 4 indexes covering the range of [MIN_ALLOC_SIZE, MAX_ALLOC_SIZE]
 * For example if MIN_ALLOCS_SIZE is 16 (2^4) bytes, MAX is 16 MB (2^24)
 *  index[0] covers [2^4,  2^7 )  or [  16,  127]
 *  index[1] covers [2^7,  2^14)  or [ 128,  16K-1]
 *  index[2] covers [2^14, 2^19)  or [ 16K, 512K-1]
 *  index[3] covers [2^19, 2^24]  or [512K,  16M]
 */
#define HALF_ALLOC_SHIFT  ((MAX_ALLOC_SHIFT + MIN_ALLOC_SHIFT) / 2)
#define HALF_ALLOC_SIZE   (1 << HALF_ALLOC_SHIFT)
#define QUARTER_ALLOC_SHIFT  ((HALF_ALLOC_SHIFT + MIN_ALLOC_SHIFT) / 2)
#define QUARTER_ALLOC_SIZE   (1 << QUARTER_ALLOC_SHIFT)
#define THREEQUARTER_ALLOC_SHIFT  ((MAX_ALLOC_SHIFT + HALF_ALLOC_SHIFT) / 2)
#define THREEQUARTER_ALLOC_SIZE   (1 << THREEQUARTER_ALLOC_SHIFT)

#define FREE_BLOCK_INDEX(bytes)  ((bytes < HALF_ALLOC_SIZE)				\
				   ? ((bytes < QUARTER_ALLOC_SIZE) ? 0 : 1 )		\
				   : (((bytes >> HALF_ALLOC_SHIFT) < QUARTER_ALLOC_SIZE) ? 2 : 3 ))

   

/*
 * Synchronization is handled about as badly as possible.
 * This module is a giant monitor. One thread through at
 * a time.  Finer grained synchronization to come...
 */

static struct mutex shmalloc_mutex = MUTEX_INITIALIZER;
static mutex_t shmu = &shmalloc_mutex;




/* 
 * Define external variables
 */
extern struct region_info_entry region_info_table[];
extern region_info region_info_index[];
extern int midway_proc_number;
extern region_info shregion_allocate ();



/*
 * External routines to acquire/release shmalloc's mutex 
 */
void acquire_shmalloc_mutex ()
{
  mutex_lock (shmu);
}
void release_shmalloc_mutex ()
{
  mutex_unlock (shmu);
}




#if SHMALLOC_STATS > 0
	
#define STAT_TABLE_SIZE 32
static struct {
  unsigned long allocRegions;	/* number of allocatable regions */
  unsigned long shmallocs;	/* number of calls to shmalloc */
  unsigned long shmallocBytes;	/* number of bytes user requested to shmalloc */
  unsigned long shmalloc_size_count[STAT_TABLE_SIZE];
  unsigned long shfrees;	/* number of calls to shfree */
  unsigned long freeSpace; 	/* bytes of free space (including tags) */
  unsigned long freeTags; 	/* Number of free tags (m_next_tag == ISFREE) */
  unsigned long allocSpace; 	/* bytes of allocated space (including tags) */
  unsigned long allocTags; 	/* Number of allocted tags (m_next_tag == ISALLOC)*/
  unsigned long searchBlocks;	/* Number of free blocks searched */
  unsigned long searchIndexes;	/* Number of free blocks indexes searched */
  unsigned long searchRegions;	/* Number of regions searched for free space */
  unsigned long userErrors;    	/* Number of user/application errors */
} shStats;
static mutex_t statLock;		/* just in case */


/* 
 * Define macro to conditionally maintain stat info
 */

#ifdef FAST_STATS
#  define INCR_STATS(x,by) (x) += (by)
#else
#  define INCR_STATS(x,by) \
    { mutex_lock(statLock);  (x) += (by);  mutex_unlock(statLock); }
#endif 

#define IF_STATS(x) x

#else /* not STATS */
#define INCR_STATS(x,by)   /* nothing to do  */
#define IF_STATS(x)
#endif


/*
 ************************************************************************
 *			s h m a l l o c _ i n i t			*
 *----------------------------------------------------------------------*
 ************************************************************************
 */

void
shmalloc_init ()
{

    /* verify the size of a mhead */
    MASSERT (sizeof(mhead_t) <= MIN_USER_ALLOC);
    
    /* Verify max allocation index matches the length of the free block */
    MASSERT (MAX_ALLOC_INDEX == FREE_BLOCK_INDEXES);

    /* initialize shmalloc mutex */
    MIDWAY_MUTEX_SET_NAME(shmu,"shmalloc_mutex");

#if SHMALLOC_STATS > 0
    statLock = mutex_alloc ();
    MIDWAY_MUTEX_SET_NAME(statLock,"shmalloc.c:statLock");
    bzero(&shStats, sizeof(shStats));
#endif

    /* initialize region allocation code */
    shregion_init();
}


/*
 ************************************************************************
 *		s h m a l l o c _ r e g i o n _ i n i t			*
 *----------------------------------------------------------------------*
 ************************************************************************
 */

void
shmalloc_region_init (region_info riP)
{
  mhead_t *mfirstP, *mlastP;
  int tag_size;

  if (IS_DBG_MISC_MALL)
    printf("shmalloc_region_init: (base = 0x%x)\n",  riP->base);

#ifdef PAGE_BASED
  /*
   * Nothing to do for page-based version.
   */
  return;
#endif


  MASSERT (riP->owner == midway_proc_number);


  /*
   * Write boundary tags in the first and last cache lines in the region.
   * Note the 1st allocatable line is in the 2nd page in the region
   * The first page contains the dirty bit update code
   */

  /* get pointer to first allocatable cache line in region */
  mfirstP = (mhead_t *)(((u_long) riP->base) + MIDWAY_TEMPLATE_SIZE);

  /* get pointer to last cache line in the region */
  tag_size = RI_LINE_SIZE_TO_BYTES(riP->line_size);
  if (tag_size < sizeof(mhead_t)) tag_size = sizeof(mhead_t);
  mlastP = (mhead_t *)(((u_long) riP->base) + REGION_SIZE - tag_size);

  /* determine allocatable region size 
   * space for 1st tag is included in free space for the block
   */
  { register unsigned long block_size = REGION_SIZE - (MIDWAY_TEMPLATE_SIZE + tag_size);
    register int block_index = FREE_BLOCK_INDEX(block_size);

    /* write 1st mhead */
    mfirstP->m_prev_tag = ISALLOC;
    mfirstP->m_next_tag = ISFREE;
    mfirstP->m_next_index = block_index;
    mfirstP->m_next_size = block_size;
    MASSIGN(mfirstP->m_magic = ISMAGIC);

    INCR_STATS (shStats.freeSpace, block_size);
    INCR_STATS (shStats.freeTags, 1);
  }

  /* Add 1st element to the free list for this region */
  MFREE_ENQUEUE(riP, mfirstP);
  
  /* write last mhead */
  mlastP->m_prev_tag = ISFREE;
  mlastP->m_next_tag = ISALLOC;
  mlastP->m_next_index = MIN_ALLOC_INDEX;
  mlastP->m_next_size = MIN_ALLOC;
  MBLOCK_PREV(mlastP) = mfirstP;
  MASSIGN(mlastP->m_magic = ISMAGIC);
  INCR_STATS (shStats.allocTags, 2);


  /* Add region to allocation index table */
  mutex_lock(shmu);
  riP->nextP = region_info_index[ riP->line_size ];
  region_info_index[ riP->line_size ] = riP;
  mutex_unlock(shmu);

  INCR_STATS (shStats.allocRegions, 1);
}


/*
 ************************************************************************
 *			    s h m a l l o c				*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
char *
shmalloc(n, line_size_in_bytes)		/* get a block */
    unsigned int n;
    unsigned int line_size_in_bytes;
{
    register mhead_t *mheadP, *mprevP, *mnextP, *mfirstP, *mfreeP;
    register unsigned long nbytes;
    register unsigned long region_line_size;
    register unsigned long region_index, alloc_index;
    register region_info riP;
    unsigned int block_size;
    unsigned int line_size;
    unsigned long extra;
    int i, once;
    char *resultP;
    IF_STATS(register unsigned long blockCount);
    IF_STATS(register unsigned long indexCount);
    IF_STATS(register unsigned long regionCount);

    INCR_STATS (shStats.shmallocs, 1);
    INCR_STATS (shStats.shmallocBytes, n);
#if SHMALLOC_STATS > 0
    { register unsigned _s = n >> 3;
      register unsigned _i = 2;
      while (_s) { _s >>= 1; _i++; }
      ASSERT ((_i >= 0) && (_i < STAT_TABLE_SIZE));
      INCR_STATS (shStats.shmalloc_size_count[_i], 1);
    }
#endif

#ifdef PAGE_BASED
    /*
     * Since we are not allowed to write into the shared space
     * we can't maintain tags, like shmalloc normally does.
     * So for now, use a brain dead implementation that
     * can malloc, but never free.   WAS 23Feb94
     */
    { vm_size_t size;

#ifdef PAGE_BASED_PAGE_ALIGNED
      /* round size up to a page multiple */
      size = PAGENO_TO_ADDR (ADDR_TO_PAGENO (n + (vm_page_size-1)));
#else 
      /* Try to match Midway's memory allocation */
      /* 
       * Figure out how many bytes are required, 
       * rounding up to the nearest multiple of the cache line size
       * and allocate extra space for the mhead.
       */
      if (n < line_size_in_bytes) line_size_in_bytes = n;
      { register unsigned long line_sz, sz;
	line_sz = 4;
	while (line_size_in_bytes > line_sz) {
	    line_sz <<= 1;
	}
	sz = (line_sz < MIN_USER_ALLOC) ? MIN_USER_ALLOC-1 : line_sz-1;
	size = ((n + sz) & ~sz) + (sizeof(*mheadP) + sz) & ~sz;
       }
#endif

      /* make sure we haven't run outa space */
      ASSERT (size <= vm_paged_free);

      resultP =  (char *)vm_paged_free_spaceP;
      vm_paged_free_spaceP += size;
      vm_paged_free -= size;
  
      return (resultP);
    }
#endif

    if (n < line_size_in_bytes)  {
	/*
	 * You can't allocate a little object in a region with
	 * a larger line size.  It doesn't make any sense.
	 * Quietly adjust.
	 */
	line_size_in_bytes = n;
    }

#ifdef DEBUG_USER
    if (n >= REGION_SIZE) {
      printf("shmalloc: invalid size requested %u bytes\n", n);
      INCR_STATS (shStats.userErrors, 1);
      return (NULL);
    }
    if (line_size_in_bytes >= REGION_SIZE) {
      printf("shmalloc: invalid line_size requested %u bytes\n", 
	     line_size_in_bytes);
      INCR_STATS (shStats.userErrors, 1);
      return (NULL);
    }
#endif    

    /* 
     * convert line size from bytes to an index into allocatable regions array
     * (ie index into power-of-two sizes) Note: line_size must be rounded up!
     */
    region_index = (line_size_in_bytes < (1 << 10)) ? 2 : 10;
    region_line_size = 1 << region_index;
    while (line_size_in_bytes > region_line_size) {
      region_line_size <<= 1;
      region_index++;
    }
    
    /* 
     * Figure out how many bytes are required, 
     * rounding up to the nearest multiple of the cache line size
     */
    { register unsigned long sz;
      sz = (region_line_size < MIN_USER_ALLOC) ? MIN_USER_ALLOC-1 : region_line_size-1;
      nbytes = ((n + sz) & ~sz) + (sizeof(*mheadP) + sz) & ~sz;
    }
    
    /*
     * Determine the allocation size index for nbytes
     * Note: This value is rounded down to the nearest power of two.
     */
    alloc_index = FREE_BLOCK_INDEX(nbytes);
    MASSERT ((alloc_index >= MIN_ALLOC_INDEX) &&
	     (alloc_index < MAX_ALLOC_INDEX));

    /*
     * serialize access to the allocation tables
     */
    mutex_lock(shmu);
    
    /* Locate an allocatable region with the requested line size. */
    riP = region_info_index[ region_index ];

    IF_STATS(blockCount = indexCount = regionCount = 0);

    /*
     * Loop until an allocatable region is found.
     */
    once = 0;
    while (TRUE) {

      /* Now locate a region with sufficient free space for this request */
      while (riP != NULL) {
	IF_STATS(regionCount++);

	{ register unsigned long m_next_size;
	  register mhead_t **first_indexP = (mhead_t **) &riP->free_block[ alloc_index ];
	  register mhead_t **last_indexP = (mhead_t **) &riP->free_block[ MAX_ALLOC_INDEX ];
	   
	  do {

	    mfirstP = *first_indexP++;
	    IF_STATS(indexCount++);
	    
	    /* Scan free list for a block of sufficient size */
	    if (mfirstP != NULL) {
	      mheadP = mfirstP;
	      do {
		IF_STATS (blockCount++);
		MASSERT (mheadP->m_magic == ISMAGIC);
		MASSERT (mheadP->m_next_tag == ISFREE);
		m_next_size = mheadP->m_next_size;
		mnextP = MFREE_NEXT(mheadP);
		if (m_next_size >= nbytes) goto found;
		mheadP = mnextP;
	      } while (mheadP != mfirstP);
	    } /* if mfirstP != NULL */

	  } while (first_indexP < last_indexP);
	}
	
	/* advance to next allocatable region with the line size */
	riP = riP->nextP;

      } /* while riP != NULL */

      /* Halt if we can't allocate a region */
      if (once++ != 0) break;

      /* No region with sufficient space was found.
       * Allocate a new region for this request
       */
      mutex_unlock(shmu);
      riP = shregion_allocate (region_line_size, 0);
      mutex_lock(shmu);
    }

    botch ("shmalloc: unable to allocate region with line=%d size=%d\n",
	   region_line_size, nbytes);
    return(NULL);		/* allow compiler to kill registers */

found:

    INCR_STATS (shStats.searchBlocks, blockCount);
    INCR_STATS (shStats.searchIndexes, indexCount);
    INCR_STATS (shStats.searchRegions, regionCount);

    /*
     * A free block of sufficient size has been found!
     * mheadP points to mhead for this block
     * mprevP points to mhead of previous block (or NULL)
     * riP points to region being allocated from
     */
	
    /* get the block off the free list */
    MFREE_DEQUEUE (riP, mheadP);

    /* mark this block as allocated */
    mheadP->m_next_tag = ISALLOC;

    /* get pointer to next (adjacent) block */
    block_size = mheadP->m_next_size;
    mnextP = MBLOCK_SIZE_NEXT(mheadP,block_size);
    MASSERT ((mnextP->m_prev_tag == ISFREE) || 
	     (mnextP->m_prev_tag == ISMINSZ));
    MASSERT (mnextP->m_next_tag == ISALLOC);
    
    /* Split block if necessary */
    extra = block_size - nbytes;
    if (extra >= (region_line_size * 4)) {

      /* min region line size is 4 bytes */
      MASSERT (extra >= MIN_ALLOC);

      /* subtract free space from the size of the allocated block */
      mheadP->m_next_size = nbytes;
      mheadP->m_next_index = alloc_index;

      /* Determine position of mhead for new free block */
      mfreeP = MBLOCK_SIZE_NEXT(mheadP, nbytes);

      /* initialize new mhead */
      mfreeP->m_prev_tag = ISALLOC;
      mfreeP->m_next_tag = ISFREE;
      mfreeP->m_next_size = extra;
      MASSIGN(mfreeP->m_magic = ISMAGIC);
      
      /* update boundary tag of next block
       * so that it points to the new mhead
       */
      MBLOCK_PREV(mnextP) = mfreeP;
      if (extra == MIN_ALLOC)
	mnextP->m_prev_tag = ISMINSZ;

      /* insert into free list for this size block */
      mfreeP->m_next_index = FREE_BLOCK_INDEX(extra);
      MFREE_ENQUEUE (riP, mfreeP);

    } else {

      /* Not necessary to split this block
       * Update boundary tag of next block to indicate this one is allocated
       */
      mnextP->m_prev_tag = ISALLOC;
      INCR_STATS (shStats.freeTags, -1);
    }

    /* Get pointer to returned space */
    MHEAD_TO_PTR(mheadP,riP,resultP);	/* sets resultP */

    /* verify the returned ptr is on a cache line boundary */
    MASSERT (((u_int)resultP) == ((((u_int)resultP) >> riP->line_size) << riP->line_size));


    /* Mark the page as allocated within this region 
     * This should be optimized to set more than 1 bit at a time. WAS 15Dec92
     */
    { register unsigned long offset, pageNo, lastPageNo;
      offset = ((char *)resultP) - ((char *)riP->base);
      pageNo = ADDR_TO_PAGENO (offset);
      offset += (nbytes - 1);
      lastPageNo = ADDR_TO_PAGENO (offset);

      if (pageNo == lastPageNo) {
	midSetBit (riP->alloc_mapP, pageNo);
      } else {
	midSetBitRange(riP->alloc_mapP, pageNo, lastPageNo);
      }
    }

    if (IS_DBG_MISC_MALL)
      printf("shmalloc: (n=%d,linesz=%d) alloc=%d at 0x%x\n", 
	     n, line_size_in_bytes, nbytes, resultP);

    INCR_STATS (shStats.freeSpace, -mheadP->m_next_size);
    INCR_STATS (shStats.allocSpace, mheadP->m_next_size-sizeof(mhead_t));
    INCR_STATS (shStats.allocTags, 1);

#if MIDWAY_WIRE_MEMORY
    { kern_return_t krc;
      char255 smsg;
      extern task_t midway_task_self;
      extern mach_port_t midway_host_priv;
      vm_size_t length;

      length = nbytes;

      krc = vm_wire(midway_host_priv, midway_task_self, resultP, 
		    length, VM_PROT_READ|VM_PROT_WRITE);

      if (krc != KERN_SUCCESS) {
	sprintf(smsg,"shmalloc:vm_wire fails, addr 0x%lx len %u",
		resultP, length);
	StatusMsg(smsg);
	mach_error("shmalloc:vm_wire",krc);
      }
    }
#endif
    
    mutex_unlock(shmu);
    return (resultP);
}

/*
 ************************************************************************
 *			    s h f r e e					*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
void
shfree (mem)
     char *mem;
{
    register mhead_t *mheadP, *mprevP, *mnextP;
    register u_long mhead_next_size;
    region_info riP;
    int index_invalid;
    IF_STATS (register unsigned long freeTags);

    INCR_STATS (shStats.shfrees, 1);

#ifdef PAGE_BASED
    /* 
     * Sorry, can't reclaim paged based space. 
     * WAS 23Feb94
     */
    return;
#endif

    if (mem == 0) {
      INCR_STATS (shStats.userErrors, 1);
      return;
    }

    /*
     * Get a pointer to the region descriptor for the returned block
     */
    riP = (region_info) &region_info_table[ (int) mem / REGION_SIZE ];
    
    /*
     * Locate mhead for the returned block.
     * mhead is in previous cache line
     */

    PTR_TO_MHEAD(mem,riP,mheadP);	/* sets mheadP */

    if (IS_DBG_MISC_MALL)
      printf("shfree(mem=0x%x) head=0x%x tag=%s size=%lu owner=%lu\n", 
	     mem, mheadP, 
	     (mheadP->m_next_tag == ISALLOC) ? "ALLOC" :
	     (mheadP->m_next_tag == ISFREE) ? "FREE" : "UNKNOWN",
	     mheadP->m_next_size, riP->owner);
    
#ifdef DEBUG_USER
    /*
     * Validate the returned block
     */
    if ((mheadP->m_magic != ISMAGIC) ||
	(mheadP->m_next_tag != ISALLOC)) {
      fprintf(stderr, "shfree: boundary tags before 0x%x have been overwritten\n", mem);
      INCR_STATS (shStats.userErrors, 1);
      return;			/* botch()? */
    }
    if (riP->owner == 0) {
      fprintf(stderr, "shfree: buffer 0x%x was not allocated\n", mem);
      INCR_STATS (shStats.userErrors, 1);
      return;			/* botch()? */
    }
#endif


    /* 
     * Currently, user is not allowed to free memory at a node different from 
     * the node it was allocated.
     */
    if (riP->owner != midway_proc_number) {
      
      /* if (IS_DBG_MISC_MALL) PRINT MSG UNTIL WE ADD THE CODE TO SUPPORT THIS */
      printf("shmalloc: Notifying owner %lu that block at 0x%x is free\n",
	     riP->owner, mem);
      INCR_STATS (shStats.userErrors, 1);
      return;
    }

    /* initialize flags */
    index_invalid = FALSE;

    /* 
     * Get a pointer to the next block 
     */
    mnextP = MBLOCK_NEXT(mheadP);

#ifdef DEBUG_USER
    if ((mnextP->m_magic != ISMAGIC) ||
	(mnextP->m_prev_tag != ISALLOC)) {
      fprintf(stderr, "shfree: boundary tags after 0x%x (%d bytes) have been overwritten\n", 
	      mem, mheadP->m_next_size - sizeof(*mheadP));
      INCR_STATS (shStats.userErrors, 1);
      return;			/* botch()? */
    }
#endif

    INCR_STATS (shStats.freeSpace, mheadP->m_next_size);
    INCR_STATS (shStats.allocSpace, -(mheadP->m_next_size-sizeof(mhead_t)));
    INCR_STATS (shStats.allocTags, -1);
    IF_STATS (freeTags = 1);


    /*
     * serialize access to the allocation tables
     */
    mutex_lock(shmu);

    /* cache value from mhead */
    mhead_next_size = mheadP->m_next_size;

    /* check if next block is free */
    if (mnextP->m_next_tag == ISFREE) {

      /* remove next block from its free list */
      MFREE_DEQUEUE (riP, mnextP);

      /* merge it with our original block */
      mhead_next_size += mnextP->m_next_size;
      mheadP->m_next_size = mhead_next_size;
      mnextP = MBLOCK_NEXT(mnextP);
      MASSERT (mnextP->m_magic == ISMAGIC);
      MASSERT ((mnextP->m_prev_tag == ISFREE) ||
	       (mnextP->m_prev_tag == ISMINSZ));
      mnextP->m_prev_tag = ISFREE;
      MBLOCK_PREV(mnextP) = mheadP;
      index_invalid = TRUE;
      IF_STATS (freeTags -= 1);
    } else {
      
      /* Set prev_tag in next header to free */
      if (mhead_next_size > MIN_ALLOC) {
	mnextP->m_prev_tag = ISFREE; 
	MBLOCK_PREV(mnextP) = mheadP;
      } else {
	mnextP->m_prev_tag = ISMINSZ;
      }
    }

    /* 
     * Get a pointer to the previous block 
     */
    if (mheadP->m_prev_tag == ISFREE)
      mprevP = MBLOCK_PREV(mheadP);
    else if (mheadP->m_prev_tag == ISMINSZ) 
      mprevP = MBLOCK_MIN_PREV(mheadP);
    else
      mprevP = NULL;
    
    /* check if previous block is free */
    if (mprevP != NULL) {

      MASSERT (mprevP->m_magic == ISMAGIC);
      MASSERT (mprevP->m_next_tag == ISFREE);

      /* remove previous block from its free list */
      MFREE_DEQUEUE (riP, mprevP);

      /* merge our original block with it */
      mhead_next_size += mprevP->m_next_size;
      mprevP->m_next_size = mhead_next_size;

      /* Set backpointer from next block to new head */
      mnextP->m_prev_tag = ISFREE;
      MBLOCK_PREV(mnextP) = mprevP;

      /* swap pointers */
      mheadP = mprevP;
      index_invalid = TRUE;
      IF_STATS (freeTags -= 1);
    } else {

      /* previous block is allocated
       * Since we didn't merge with the previous block
       * we need to mark this one as free.
       */
      mheadP->m_next_tag = ISFREE;
    }

    INCR_STATS (shStats.freeTags, freeTags);


    /*
     * Compute new index for this block
     * if we merged it with an adjacent block.
     */
    if (index_invalid) {

      /* store new index for this block */
      mheadP->m_next_index = FREE_BLOCK_INDEX(mhead_next_size);
    }

    /*
     * Insert returned block into its free list
     */
    MFREE_ENQUEUE(riP, mheadP);

    mutex_unlock(shmu);


}

/*
 ************************************************************************
 *	m i d _ s h m a l l o c _ p r i n t _ s t a t s			*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
void mid_shmalloc_print_stats(void)
{
#if SHMALLOC_STATS == 0
  StatusMsg ("Not compiled to collect shmalloc statistics");
#else
  int i, len, slen;
  char255 smsg, tmp;
#define MAX_STR_LEN 80

  StatusMsg("-=-=- shmalloc Statistics -=-=-");

  sprintf(smsg,
	  "shmalloc %lu shfree %lu Search blocks %lu index %lu reg %lu Errors %lu",
	  shStats.shmallocs, shStats.shfrees, shStats.searchBlocks,  
	  shStats.searchIndexes, shStats.searchRegions,  shStats.userErrors);
  StatusMsg (smsg);

  sprintf(smsg, "shmalloc size(2^)");
  slen = strlen(smsg);
  for (i = 0; i < STAT_TABLE_SIZE; i++) {
    if (shStats.shmalloc_size_count[i] == 0) continue;
    
    sprintf(tmp, " %d=%lu", i, shStats.shmalloc_size_count[i]);
    len = strlen(tmp);
    if ((len + slen) > MAX_STR_LEN) {
      StatusMsg(smsg);
      sprintf(smsg, "shmalloc size(2^)");
      slen = strlen(smsg);
    }
    strcat (smsg, tmp);
    slen += len;
  }
  StatusMsg(smsg);

  sprintf(smsg,
	  "Regions %lu (%lu bytes) free %lu tags %lu alloc %lu tags %lu user %lu",
	  shStats.allocRegions, (shStats.allocRegions * REGION_SIZE),
	  shStats.freeSpace,  shStats.freeTags, shStats.allocSpace,  
	  shStats.allocTags, shStats.shmallocBytes);
  StatusMsg (smsg);

  midway_malloc_print_stats();

#endif /* SHMALLOC_STATS */
}
