/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */

/*
 *  Created by bnb sometime in june...
 *  8 aug 91 (mjz): add owner to region info
 * 16 Dec 91 (mjz): delete bogus [0=us if linesize] comment for owner
 *  7 May 92 (mjz): add i386
 * 16 Jul 92 (mjz): add SHMALLOC_ALLOCATION_BASE because of 386
 * 11 Aug 92 (was): added clock_t type
 * 15 Aug 92 (was): added page tables to region_info desc.
 * 18 Oct 92 (was): Increased REGION_SIZE to 2MB. 
 *  2 Jun 93 (mjz): incr shared space from 128 to 512M
 *  8 Jan 94 (was): Define max size for application's stack
 */

#include <mach.h>
#include "midway_types.h"               /* for clock_t */
#include "midway_comp.h"	/* for REGION_SIZE */



/*
 * Define region types based on the region's contents.
 * This type defines the dirty bit update code stored
 * in the first page of the region.
 */

#define REGION_TYPE_UNKNOWN         0		/* region type not yet known */
#define REGION_TYPE_PRIVATE         1		/* data in region is private */
#define REGION_TYPE_WORD_EC         2		/* data is aligned, word size cache, EC */
#define REGION_TYPE_ALIGNED_EC      3 	        /* data is aligned, EC */
#define REGION_TYPE_UNALIGNED_EC    4           /* data is unaligned, EC */
#define REGION_TYPE_MIXED           5     	/* data is unaligned, mixed model */


/*
 * WARNING: If you change the size or layout of a region_info_entry, make
 * sure that you  check that you have not broken any machine dependent
 * assembler code.
 * NOTE: The size of this structure MUST be a power of 2.
 */
typedef struct region_info_entry {
     unsigned char  line_size;	/* how big each line */
     unsigned char  type;	/* type of dbit template for this region */
     unsigned short unused_pad_to_word;
     clock_t   *line_info;	/* dirty bit information */
     char     **free_block;	/* for shmalloc: free block index */
     vm_offset_t base;		/* where in memory are we */
     unsigned long owner;	/* who "owns" this region */
     struct region_info_entry *nextP; /* ptr to next allocatable region
				       * with the same line size */
     addr_vector_t *page_tableP;  /* ptr to region's page table */
     BITARRAY   *alloc_mapP;    /* ptr to region's allocation map */
} *region_info, region_info_t;

/*
 * region_info->line_size is kept in log form to make the division in
 * the update code run fast.  
 * These macros allow you to easily go from log form to native form.
 */
#define BYTES_TO_RI_LINE_SIZE(b)  ( midway_int_to_log_int(b) )
#define RI_LINE_SIZE_TO_BYTES(l)  ( ((1)) << (l) )

/*
 * Define length of free_block index
 */
#define FREE_BLOCK_INDEXES    4

extern region_info end_static_ri;
extern clock_t* compute_dirty_bit_address();
extern region_info region_from_addr();
extern kern_return_t get_shmalloc_info();
extern kern_return_t shregion_allocate_private ();


#if mips
extern int _fdata;
extern int end;
#define FIRST_STATIC_SYM	 ((char*)&_fdata)
#define LAST_STATIC_SYM		 ((char*)&end)
extern int __start();
extern int etext;
#define FIRST_TEXT_SYM	 ((char*)__start)
#define LAST_TEXT_SYM	 ((char*)&etext)
#define SHMALLOC_ALLOCATION_BASE 0x20000000	/* start at 1G */
#endif

#if vax || m88k || i386
/*
 * On these machines, the loader won't let us
 * force data to begin on a fresh region,
 * so we end  up allocating region_info for text.
 */
extern int etext;
extern int end;
#define FIRST_STATIC_SYM	 ((vm_offset_t)&etext)
#define LAST_STATIC_SYM		 ((vm_offset_t)&end)
/* following at 10 meg.  I found that after the initial 1meg allocation
   at 0x100000 for the 386, lots of goop was allocated up to 0x600000.
   This should give us plenty extra.  (I thought about starting with
   the mips number (which will start at 0x10100000 (fdata + 1meg)).
   However, I couldn't bear to "waste" the "limited" space covered
   by dirty bits (up to 0x40000000 today (7/92)).
   Wish I knew what the goop was.
   Wish I had a method that will work in all cases... --mjz */
#define SHMALLOC_ALLOCATION_BASE (0x08000000)
#endif


/*
 * Define region size 
 */
#define REGION_SHIFT    MIDWAY_REGION_SHIFT
#define REGION_SIZE	MIDWAY_REGION_SIZE

extern int mid_region_shift;
extern int mid_region_size;



/*
 * MAX_SHARED_SPACE defines the max size of the shared space
 *
 * MAXREGION defines the maximum number of shared regions
 * 
 * MAXVREGION includes regions information records allocated for the
 * nonshared portion of the address space.  These are not used but
 * are there simply to occupy space to allow the computation of
 * the region_info from an address to move more quickly.
 */

#define MAX_SHARED_SPACE (512*1024*1024)
#define MAXREGION	 (MAX_SHARED_SPACE / REGION_SIZE)
#define MAXVREGION	 ((SHMALLOC_ALLOCATION_BASE / REGION_SIZE) \
			     + (2 * MAXREGION))

/*
 * Define maximum size for application's stack
 */
#define MAX_STACK_SIZE  (REGION_SIZE / 2)

/*
 * Define size of each dense dirty bit.
 */
#define DENSE_DIRTYBIT_SIZE 1	/* =1 for 1 byte, =4 for 4 bytes */

#define DIRTYBIT_ALIGNMENT  (64*1024) /* must allocate on 64K boundary */

/*
 * A region_from_addr() macro: the real function (shregion.c) has 
 * additional safety checks.
 */
extern struct region_info_entry region_info_table[];
/* sizeof(pointer) == sizeof(long) dependency: */
#define REGION_FROM_ADDR(p) ( &region_info_table[ (((unsigned long) (p)) / REGION_SIZE ) ] )

