/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/* 
 *  Dynamic region allocation for MIDWAY
 *
 * 10 Oct 93 (was): broke out shmalloc/shfree from region allocate/free 
 * 20 Mar 94 (mjz): think about adding support for allocinfo with lock replies.
 *		..end up with ALLOC_INFO_HACK: farm out request to other thr.
 */

#define ALLOC_INFO_HACK 1

#include <sys/types.h>		/* for comm.h */
#include <signal.h>
#define KERNEL 1
#include <sys/time.h>
#undef KERNEL
#include <sys/resource.h>
#include <mach.h>
#include <mach/kern_return.h>
#include <stdio.h>
#include <cthreads.h>
#include "midway_types.h"
#include "midway_mem.h"
#include "shmalloc.h"
#include "pbuf.h"
#include "vm.h"
#include "comm.h"		/* barf.  should have interface include */
#include "smexecd_types.h"	/* char255 */
#include "pagemgr.h"		/* prototypes for calling page mgr rtns */
#include "midway-dbg.h"		/* DBG_IS_MISC_MAIF */
#include "midway.h"		/* prototypes */


#if (ALLOC_INFO_HACK)
void mid_aihack_init(void);
#endif




/*
 * The regions from MAXVREGION-MAXREGION to MAXVREGION
 * contin descriptors for dynamically shared data.
 * Some of the regions in 0.. MAXVREGION-MAXREGION contain
 * static data.
 */
static int lowest_region = MAXVREGION - MAXREGION;
static int highest_region = 0;	/* make our life easier on searches */

struct region_info_entry region_info_table[MAXVREGION];

/*
 * Define an index over the allocatable regions
 * based on the region's line size
 */
#define REGION_LINE_SIZES   32				/* 1,2,4,8,16,32,...,2^31 */
region_info region_info_index[REGION_LINE_SIZES];	/* list for each line size */


static region_info region_init();
region_info shregion_allocate();


region_info end_static_ri = 0;
static int num_static_regions;


static vm_offset_t get_valid_region_for_this_processor();
static vm_offset_t r_base = 0;	/* region base */
static vm_offset_t r_bound = 0;	/* region bound */


/* global allocator used only on main proc */
void midway_global_allocator();	/* entry point for the global allocator */
static vm_offset_t global_pointer;	/* base of next global allocation */
static void new_region_please(); /* who to call to get another */

unsigned long midway_default_allocation_size 
  = MIDWAY_DEFAULT_ALLOCATION_REGIONS * REGION_SIZE;
static unsigned long proc_num;  /* this processors "number" (1-n) */
static unsigned long max_proc_num; /* how many are there? */
void midway_set_numprocs();	/* set for allocator */

/*
 * Define default memory consistency model
 */
extern int midway_default_model;


/* secondary mutex for global allocation -- used only on main processor */
static struct mutex global_mutex = MUTEX_INITIALIZER;
static mutex_t globlock = &global_mutex;

static struct mutex info_reply_mutex = MUTEX_INITIALIZER;
static mutex_t info_reply_lock = &info_reply_mutex;


/*
 * other forward declarations
 */
void botch();			/* now defined in xpager/panic.c */
region_info region_from_addr();


/*
 * Comm Related
 *   overload "lock" to be address we care about
 */
typedef struct m_alloc_reply {
    unsigned long success; /* just in case we could fail... */
    vm_offset_t start_addr; /* start addr of region to use */
    vm_offset_t end_addr; /* end addr of region to use */
} m_alloc_reply_t;

/*
 * called by comm layer to service a remote allocate request
 */
typedef struct arcache_s {
    unsigned long tranid; /* the transaction id of last req */
    vm_offset_t   start; /* starting address of allocation */
    vm_offset_t   end; /* ending address of allocation */
} arcache_t;

static arcache_t *arcache;	/* mem alloc'd by set_num_proc */
  
void midAllocateRequest(lock, replyhost, replythread, tranid, retrans)
     unsigned long lock;		/* useless, is 0 */
     unsigned long replyhost;		/* host to reply to */
     unsigned long replythread;	/* thread on host to reply to */
     unsigned long tranid;		/* the transaction id */
     int    retrans;		/* 1 if a retransmission */
{
    /* NOT IDEMPOTENT! */
    /* know 1/host, though... so if prev req not sat, we won't get another
       req of same type.  do a tranid check? */

    vm_offset_t sss; /* start of newly allocated memory */
    vm_offset_t eee; /* end of newly allocated memory */
    m_alloc_reply_t *ar;

    /*  FIRST VERIFY WE ARE THE MAIN PROCESSOR? --19nov91mjz */
    /*
     * do check to see if this is retrans; if so, then forward old data.
     * else, call midway_global_allocator to get new memory for this
     * guy, remember it, and send reply.
     */
    mutex_lock(globlock); /* protect concurrent acc to arcache */
    if (arcache[replyhost].tranid == tranid) {
	/* DOESNOT WORK FOR MULTIPLE THREADS MAKING REQUESTS.
	   WE ALREADY HAVE OLD REQUESTS FILTERED OUT BY COMM LAYER,
	   SO ONLY HAVE TO WORRY ABOUT DUPLICATES!
	   */
	if ((arcache[replyhost].start == 0) &&
	    (arcache[replyhost].end == 0)) {
	    /* still working on request.  drop this duplicate */
	    /* could reply pending */
	    mutex_unlock(globlock);
	    return;
	}
	/* reply using info from cache.  unlock first? */
	sss = arcache[replyhost].start;
	eee = arcache[replyhost].end;
    } else { /* is a new request! */
	arcache[replyhost].tranid = tranid;
	arcache[replyhost].start = 0;
	arcache[replyhost].end = 0;
	/* since globlock used by global allocator, must release.
	   [actually, could just lock here and not there, or use
	   two different locks.]  For now just don't forward if no info */
	mutex_unlock(globlock);
	midway_global_allocator(replyhost, &sss, &eee);
	mutex_lock(globlock);
	arcache[replyhost].start = sss;
	arcache[replyhost].end = eee;
    }
    mutex_unlock(globlock);

    ar = (m_alloc_reply_t *) mid_pbuf_get();
    ASSERT (ar != NULL);
    ar->success = 1;
    ar->start_addr = sss;
    ar->end_addr = eee;

    CommReply(M_ALLOCATE, replyhost, replythread, tranid, lock, 1 /*last*/,
	      0 /* packnum */, NULL_ADDR_VECTOR /*acks*/, 
	      (char *) ar, sizeof(m_alloc_reply_t), 1 /*freepack*/);

}   /* midAllocateRequest */

/* routine that asks the master processor for a new region */
static void
new_region_please( thisProcNum, baseAddr, boundAddr )
     unsigned long thisProcNum;
     vm_offset_t   *baseAddr;
     vm_offset_t   *boundAddr;
{
    if (proc_num == 1) { /* master */
	midway_global_allocator(thisProcNum, baseAddr, boundAddr);
    } else {
	/* not master, but */
	/* forward to master */
	CommRequest(0, /* use address as "lock in request */
		    M_ALLOCATE,	/* type */
		    1, /* bletch constant *//* dest */
		    0, /* no extra message */
		    0);	/* so length is 0 */
    }

    return;
}   /* new_region_please */

void midAllocateReply(lock, source, thread, tranid, bufp, buflen)
     unsigned long lock;		/* useless, is 0 */
     unsigned long source;		/* source host */
     unsigned long thread;		/* source thread */
     unsigned long tranid;		/* the transaction id */
     char   *bufp;		/* the reply data */
     unsigned long buflen;		/* length of reply data */
{
    m_alloc_reply_t *m;

    m = (m_alloc_reply_t *) bufp;
    if (buflen != sizeof(m_alloc_reply_t)) {
	fret("midAllocateReply: size of buffer (%d) is not %d! Dropping.",
		buflen, sizeof(m_alloc_reply_t));
    }
    if (m->success) {
	r_base = m->start_addr;	/* set global values for this processor */
	r_bound = m->end_addr;
    }
    /* complain if failure *//* permanent damage -- abort? */
}   /* midAllocateReply */


kern_return_t
#if (ALLOC_INFO_HACK == 0)
get_shmalloc_info( addr )
#else
midway_get_shmalloc_info_internal( addr )
#endif
     char * addr;
{
  region_info ri;

  ri = region_from_addr( (int *)addr );

  
  if (ri->line_size)  {
      /* no need to continue */
      return KERN_SUCCESS;
  }

  /*
   * master & owner exists -> send to owner
   *   else fail
   * not master -> send to master
   *
   * check if now have info... then success else fail
   */
  if (proc_num == 1) {		/* master */
    if (ri->owner != 0) {	/* owner exists */
      /* send to owner */
      CommRequest((unsigned long) addr, /* use address as "lock" in request */
		  M_ALLOC_INFO, /* type */
		  ri->owner,	/* dest */
		  0,		/* no message */
		  0);		/* so length is 0 */
    } else {
      /* reply with fail -- segment never allocated */
      /* don't need to do anything, will fail below */
    }
  } else {			/* not master */
    /* forward to master */
    CommRequest((unsigned long) addr, /* use address as "lock in request */
		M_ALLOC_INFO,	/* type */
		1,		/* bletch constant *//* dest */
		0,		/* no extra message */
		0);		/* so length is 0 */
  }

  if (ri->line_size)  {
      return KERN_SUCCESS;
  } else {
      return KERN_FAILURE;
  }
} /* get_shmalloc_info */

typedef struct m_ainfo_reply {
  unsigned long	success;	/* 1 if successful, else 0 (boo hoo) */
  unsigned long	linesize;	/* the answer! */
  unsigned long	owner;		/* owner's name (here in case not sender) */
  unsigned long type;		/* region's type */
  BITARRAY      alloc_map[1];   /* bit map of allocated pages
				 * actual size 
				 *   = midBITARRAYsize(region_size/vm_page_size) */
} m_ainfo_reply_t;

/* called to service request */
void midAllocInfoRequest(addr, replyhost, replythread, tranid, retrans)
     vm_offset_t  addr;       /* is address need info about, really char* */
     unsigned long replyhost;		/* host to reply to */
     unsigned long replythread;	/* thread on host to reply to */
     unsigned long tranid;		/* the transaction id */
     int    retrans;		/* 1 if a retransmission */
{
    region_info ri;
    unsigned long airlen;
    m_ainfo_reply_t *air;

    ri = region_from_addr((int *) addr);
    /*
     * if ours & linesize, reply affermative
     * else, if we are the master
     *           if know owner, forward
     *           if don't know owner, nack
     *       else (must have been forwarded erroneously -- nack
     *
     * So we only have to allocate in one place...
     * If master & know owner, and not the owner,
     *	Forward
     * Else
     *	alloc packet
     *	if owner & linesize, ack with info
     *    else
     *	  if (!master & !owner), bitch (probably programming error)
     *	  reply with nack (will kill us)
     *		- some form of problem.  State State (owner is known/unknown,
     *			linesize is known/unknown)
     */
    if ((proc_num == 1) && (ri->owner != 1) && (ri->owner != 0)) {
	CommForward(addr, M_ALLOC_INFO, ri->owner, replyhost, replythread,
		    tranid, retrans, (char *) 0, 0);
    } else {
	/* get packet buffer for reply */	
	air= (m_ainfo_reply_t *) mid_pbuf_get(); 
	ASSERT (air != NULL);
	if ((ri->owner == proc_num) && (ri->line_info)) {
	    air->success = 1; /* success */
	    air->linesize = ri->line_size; /* NOTE: in log format */
	    air->owner = ri->owner;
	    air->type = ri->type;
	    bcopy (ri->alloc_mapP,                             /* source */
		   air->alloc_map,                             /* target */
		   midBITARRAYsize(ADDR_TO_PAGENO (REGION_SIZE))); /* length */
	    airlen = (sizeof(*air)-sizeof(air->alloc_map)) 
 	                + midBITARRAYsize(ADDR_TO_PAGENO (REGION_SIZE));

	    if (IS_DBG_MISC_MAIF) {
		dbg("midAllocInfoReq: success! linesize %u, owner %u",
		    air->linesize, air->owner);
	    }
	} else { /* there is a problem here */
	    if ((proc_num != 1) && (ri->owner != proc_num)) {
		fret("I'm not the master, and I don't own this block???");
	    }
	    fret("midAllocInfoRequest: block unallocated -- nacking; ri %lx lsz %x linfo %lx base %lx owner %lu",
		 ri, ri->line_size, ri->line_info, ri->base, ri->owner);
	    air->success = 0; /* sigh */
	    airlen = sizeof(*air);
	}
	CommReply(M_ALLOC_INFO, replyhost, replythread, tranid,
		  ri->base,	/* was addr, but we need base to alloc */
		  1	/*last*/,
		  0 /* packnum */, NULL_ADDR_VECTOR /*acks*/, 
		  (char *) air, airlen, 1 /*freepack*/);
    }
    return;
}   /* midAllocInfoRequest */

/* called when reply to request arrives */
void midAllocInfoReply(addr, source, thread, tranid, bufp, buflen)
     vm_offset_t   addr;
     unsigned long source;		/* source host */
     unsigned long thread;		/* source thread */
     unsigned long tranid;		/* the transaction id */
     char   *bufp;		/* the reply data */
{
  region_info ri;
  m_ainfo_reply_t *air;
  unsigned long airlen;

  if (buflen < sizeof(m_ainfo_reply_t)) {
      fret("midAllocInfoReply: size of buffer (%d) is < %d!  Dropping. (addr = %lx)",
	   buflen, sizeof(m_ainfo_reply_t), addr);
      goto done;
  }

  air = (m_ainfo_reply_t *) bufp;
  if (air->success) {

      airlen = (sizeof(*air)-sizeof(air->alloc_map))
		      + midBITARRAYsize(ADDR_TO_PAGENO (REGION_SIZE));
      if (buflen !=  airlen) {
	fret("midAllocInfoReply: size of buffer (%d) is not %d!  Dropping. (addr = %lx)",
	   buflen, airlen, addr);
	goto done;
      }

      /*
       * 5 sep 94, mjz: it's possible that there could be > 1 outstanding
       * request for the same region.  These should be resolved
       * locally, but for now, just supress a double and note it.
       * (SIGH, this is a race!)
       */
      mutex_lock(info_reply_lock);
      ri = region_from_addr( addr );
      if (ri->line_size != 0) {
	if (IS_DBG_MISC_MAIF) {
	  dbg("midAllocInfoReply: duplicate reply for region 0x%x addr 0x%x discarded",
	      ri->base, addr);
	}
	goto done_ri_setup;
      }


      /* region_allocate will map region and allocate dirty bits */
      ri = region_init(addr,
		       REGION_SIZE,
		       RI_LINE_SIZE_TO_BYTES(air->linesize));
      ri->owner = air->owner;
      ri->type = air->type;
      
      /* copy allocation map to the region descriptor */
      bcopy (&air->alloc_map,					/* source */
	     ri->alloc_mapP,					/* target */
	     midBITARRAYsize(ADDR_TO_PAGENO (REGION_SIZE)));	/* length */

      /* initialize the dirtybit template */
      ipatch_template(ri);
done_ri_setup:;
      mutex_unlock(info_reply_lock);
      
  } else {
    fret("midAllocInfoReply: Request for addr 0x%x failed, host=%d thd=%d.", 
	 addr, source, thread);
  }

done:;
  return;
} /* midAllocInfoReply */

#if 0
/* This is in preparation for doing things the "right" way. */
/*
 * midway_set_shmalloc_info: for EC lock xfers, set on the fly so we
 *   do not have to make a call back to the sender.
 *   THIS IS EC ONLY.
 *	1. No allocation map (ri->alloc_mapP)
 *	2. Pure EC memory (??)
 *	3. Owner will not be set (shouldn't make a difference...)
 */
void midway_set_shmalloc_info( vm_offset_t addr, /* random address in region */
			        int  lin_sz      /* region line size */
			     )
{
  vm_offset_t base = addr & ~(REGION_SIZE-1);  /*true except for 386 static*/
  int linesz_bytes = RI_LINE_SIZE_TO_BYTES(lin_sz);

  /* region_allocate will map region and allocate dirty bits */
  ri = region_init(addr,
		   REGION_SIZE,
		   linesz_bytes);

  ri->owner = 0; /*XXX*/

  if (midway_default_model != MODEL_EC)
    ri->type = REGION_TYPE_MIXED;
  else if (linesz_bytes == sizeof(clock_t))
    ri->type = REGION_TYPE_WORD_EC;
  else
    ri->type = REGION_TYPE_ALIGNED_EC;
  
  /* initialize the dirtybit template */
  ipatch_template(ri);
} /* midway_set_shmalloc_info */
#endif


/*
 * utility functions to support partitioning of space on various procs
 */
void
midway_print_region_debuginfo()
{
    printf("==>r_base=%lx r_bound=%lx \n",  r_base, r_bound);
    printf("==>global_pointer=%lx\n", global_pointer);
}/* midway print region debuginfo */


void
midway_set_numprocs( pnum, max )
    unsigned long pnum;
    unsigned long max;
{
    region_info ri;
    int i, j, numregions;

    if ((r_base != r_bound) || (r_base != 0)) {
	botch("midway_set_numprocs called twice!");
    }

    proc_num = pnum;
    max_proc_num = max;

    /* set initial base & bounds for this processor */
    /* (index origin 0) */     
    r_base = (vm_shared_base +
	      (pnum - 1) * midway_default_allocation_size);
    r_bound = ((vm_offset_t)r_base +  midway_default_allocation_size);

    /* alloc memory for 'arcache' -- a structure to make
       the requests for new regions to allocate out of
       idempotent.
       */
    if (pnum == 1) { /* only necessary on main processor */
	arcache = (arcache_t *) CALLOC(max,sizeof(arcache_t));
	if (arcache == 0) {
	    botch("midway_set_numprocs: unable to allocate arcache struct");
	}
	/*
	 * also need to note what processors have pre-allocated segments
	 */
	ri = region_from_addr(vm_shared_base + midway_default_allocation_size);
	numregions = midway_default_allocation_size / REGION_SIZE;
	/* ought to be MIDWAY_DEFAULT_ALLOCATION_REGIONS, unless someone
	 * patched it */
	for (i = 2; i<=max; i++) {
	  for (j = 0; j < numregions; j++) {
	    ri->owner = i;
	    ri++;		/* look at next region */
	  }
	}
    }

#if ALLOC_INFO_HACK
    mid_aihack_init();
#endif
    return;
}   /* midway_set_numprocs */



void
midway_global_allocator(pnum, where, end_where)
    unsigned long pnum;	/* which processor wants to know? */
    vm_offset_t *where;		/* the assigned address */
    vm_offset_t *end_where;		/* bound */
{
    region_info ri;
    vm_offset_t p;
    unsigned long alloc_size;

    mutex_lock(globlock);
    if (global_pointer == 0) {
	global_pointer = (vm_offset_t) (vm_shared_base +
			    (max_proc_num * midway_default_allocation_size));
    }

    /* ignore processor number in allocation for now... */
    
    /*
     * Allocate 1 region at a time.
     */
    alloc_size = REGION_SIZE;
    *where = global_pointer;
    global_pointer = global_pointer + alloc_size;
    if (global_pointer > vm_shared_end) {
	botch("global_allocator: out of shared memory");
    }
    *end_where = global_pointer;

    if (IS_DBG_MISC_MREG) {
      printf("midway_global_allocator called - lo %lx hi %lx sz %lx\n",
	    *where, *end_where, alloc_size); /*DEBUG*/
    }

    /* note owner  -- have to do for each region in allocation_size */
    for( p = *where; p < *end_where; p += REGION_SIZE) { 
	ri = region_from_addr(*where);
	ri->owner = pnum;
    }

    mutex_unlock(globlock);
    return;
} /* midway_global_allocator */



static vm_offset_t
get_valid_region_for_this_processor()
{
    vm_offset_t p;
    region_info ri;

    if (r_base >= r_bound) {
	if (r_base == 0)
	    botch("get_valid_region_for_this_processor: never initialized");
	new_region_please( proc_num, &r_base, &r_bound );
    }

    p = r_base;			/* return current base pointer */
    r_base += REGION_SIZE;	/* bump */

    ri = region_from_addr(p);
    ri->owner = proc_num;	/* owned by us now... */

    return(p);
} /* get valid region for this processor */


region_info
region_from_addr(p)
    vm_offset_t p;
{
    int i = (int)p / REGION_SIZE;
    if (i > highest_region)  {
	highest_region = i;
	if (i > MAXVREGION)  {
	    botch("OUT OF MEMORY:region info table is too small");
	}
    } else if (i < lowest_region)
      lowest_region = i;

    return &region_info_table[i];
}

region_info
region_from_region_num(i)
    int i;
{
    if (i > MAXVREGION) return 0; else return &region_info_table[i];
}

    

vm_offset_t
addr_from_region(ri, k)
    region_info ri;
    int k;
{
    int linesz = RI_LINE_SIZE_TO_BYTES(ri->line_size);
    int offset = k * linesz;
    vm_offset_t base = (ri - &region_info_table[0]) * REGION_SIZE;
    return (base + offset);
}

/*
 * Allocate space from the dirty bit region.
 * This space is never freed or reused.
 */
clock_t *
dbmalloc (int size, int align)
{
  clock_t *rtndP;
  vm_offset_t wasted, newP;

  ASSERT ((size & 0x03) == 0);   /* word multiples only */
  ASSERT (size <= vm_dirty_free);

  /* round size up to a page multiple */
  size = PAGENO_TO_ADDR (ADDR_TO_PAGENO (size + (vm_page_size-1)));

  /* Align dirtybits on 64K boundary, if necessary 
   * Dirtybits assigned to a region must have low 16 bits = 0
   * to patch instruction in dirtybit template code.   WAS 29Oct93
   */
  if (align) {

    newP = (vm_dirty_free_spaceP + (DIRTYBIT_ALIGNMENT-1)) 
	       & ~(DIRTYBIT_ALIGNMENT-1);
    wasted = newP - vm_dirty_free_spaceP;
    vm_dirty_free_spaceP = newP;
    vm_dirty_free -= wasted;
  }

  rtndP = (clock_t *) vm_dirty_free_spaceP;
  
  vm_dirty_free_spaceP += size;
  vm_dirty_free -= size;

  ASSERT((((unsigned long)rtndP) & (vm_page_size-1)) == 0);
  
  return (rtndP);
}


#ifdef INIT_DIRTYBITS
/*
 * shregion_allocate_memory: init the dirty bits for the specified memory only
 */    
void
shregion_allocated_memory
  (
   region_info riP, 
   void *memP, 
   int mem_size
   )
{
    int off, incr;
    register clock_t clock;
    register clock_t *line_infoP, *end_infoP;

    if (IS_DBG_MISC_MREG)
      printf("shRegion_allocated_memory: Addr=0x%x length=%d\n", memP, mem_size);


    /* locate region descriptor, if caller hasn't already */
    if (riP == NULL) {
      riP = region_from_addr (memP);
      ASSERT(riP != NULL);
    }

    /*
     * Initialize dirty bit information based on the default consistency model
     * Entry consistency stores a clock value in each entry.
     * Processor & Release consistency record changes to shared data
     * via an update queue. Mark each cache line to indicate changes
     * must be added to the update queue.
     *
     * NOTE: Changes made to line_info initialization must also be
     * made in midway_bind_synch_object, which re-initializes a the
     * dirty bits when memory changes from Non-EC to EC.
     * WAS 29Oct92
     */

    /*
     * Dirty bit info is now initialized as it is faulted in.
     * We step through the pages of the dirty bits to fault in
     * each page that backs the memory allocated by the caller.
     * WAS 17Dec92
     */
    clock = vm_dirty_value; 	/*  cache value */

    /* get pointer to region info at the starting & ending memory addresses */
    off = ((char *)memP) - ((char *)riP->base);
    line_infoP = &riP->line_info[ off >> riP->line_size ];

    off += (mem_size-1); /* -1 so we don't step into the next line */
    end_infoP = &riP->line_info[ off >> riP->line_size ];

    /* write 1st word in each page to incur a copy-on-write fault */
    incr = vm_page_size / sizeof(*line_infoP);
    while (line_infoP <= end_infoP) {
      line_infoP[0] = clock;
      line_infoP += incr;
    }
    
} /* shregion_allocated_memory */

#endif /* INIT_DIRTYBITS */


static region_info
region_init(cp, region_size, linesz)
    vm_offset_t cp;
    int region_size;
    int linesz;
{
    int nPages, nBytes;
    region_info ri;
    int info_size;


    /*
     * initialize info struct for this region
     */
    ri = region_from_addr(cp);
    ri->base = cp;
    ri->line_size = BYTES_TO_RI_LINE_SIZE(linesz);
    ri->type = REGION_TYPE_UNKNOWN;

    /*
     * Allocate & initialize free pointers for this region
     */
    ri->free_block = (char **) MALLOC( FREE_BLOCK_INDEXES * sizeof(char *) );
    ASSERT (ri->free_block != NULL);
    bzero (ri->free_block, FREE_BLOCK_INDEXES * sizeof(char *));

    /*
     * Allocate & initialize page allocation tables
     * Used by non-ec models
     */
    nPages = ADDR_TO_PAGENO (region_size + (vm_page_size-1));
    ri->alloc_mapP = (BITARRAY *) MALLOC( midBITARRAYsize(nPages) );
    ASSERT (ri->alloc_mapP != NULL);
    midZeroBitArray (ri->alloc_mapP, nPages);

#ifdef PAGE_BASED
    /*
     * Don't allocate dbits for page-based version
     */
    info_size = 0;
    ri->line_info = NULL;
#else
    /*
     * Allocate dirty bits for this region 
     * These are malloc'ed from the dirtybit region
     * and initialized as they are faulted in.
     */
    info_size = (region_size >> ri->line_size) * sizeof(clock_t);
    ri->line_info = (clock_t *) dbmalloc(info_size, TRUE);
    ASSERT (ri->line_info != NULL);
    ASSERT (((u_long)ri->line_info & (DIRTYBIT_ALIGNMENT-1)) == 0);
#endif /* not PAGE_BASED */

    /*
     * allocate the page tables for this region
     * page tables contain the copy set for each page which is 
     * represented as a bit array with 1 bit per processor
     */
    nBytes = nPages * sizeof(addr_vector_t);
    ri->page_tableP = (addr_vector_t *) MALLOC( nBytes );
    ASSERT (ri->page_tableP != (addr_vector_t *) 0);
    
    /* initialize page tables (all copy sets are empty) */
    bzero (ri->page_tableP, nBytes);

    if (IS_DBG_MISC_MREG)
      printf("region_init: 0x%x line=%d (%d) info=0x%x (%d bytes) page=0x%x (%d bytes)\n",
	     ri->base, linesz, ri->line_size, ri->line_info, info_size,
	     ri->page_tableP, nBytes);
    
    return ri;
}

region_info
shregion_allocate(linesz_bytes, addr)
    int linesz_bytes;
    vm_offset_t addr;
{
    kern_return_t e;
    vm_offset_t cp;
    region_info ri;

    if (!addr)  {
	addr = get_valid_region_for_this_processor();	
    } 
    cp = addr;	

    /*
     * If the address is not aligned properly, the following vm_map
     * call seems to ask for REGION_SIZE starting at the address,
     * but we might get more back because of the mask.  i.e. if
     * cp is 1.5m, and we ask for 1m on 1m boundry (which we are),
     * then 1-3 is allocated instead of 1-2.
     */
    if (IS_DBG_MISC_MREG)
      printf("region_allocate:cp was 0x%x\n", cp); 
   
    /* align to REGION_SIZE boundry */    
    cp &= ~(REGION_SIZE-1);
    if (cp != addr)  {
	/* sanity check */
	botch("region_allocate -- address not on region boundary.\n");
    }
	
    if (IS_DBG_MISC_MREG) {
	fprintf(stderr,
		"region_allocate:cp is 0x%x, linesize=%d\n", cp,
			linesz_bytes);
    }

    /*
     * Allocate region_info structure for the new region
     */
    ri = region_init(cp, REGION_SIZE, linesz_bytes);


#ifdef PAGE_BASED
    /*
     * Page-based version of midway does not require dbit templates.
     * leave as is (ie all 0's), to cause an error if executed.
     */
#else

    /* determine the region's type */
    if (midway_default_model != MODEL_EC)
      ri->type = REGION_TYPE_MIXED;
    else if (linesz_bytes == sizeof(clock_t))
      ri->type = REGION_TYPE_WORD_EC;
    else
      ri->type = REGION_TYPE_ALIGNED_EC;

    /*
     * Initialize the dbit template 
     * stored in the 1st page of the region
     */
    ipatch_template (ri);

    /* 
     * if we own this region, then initialize it for allocation.
     */
    if (ri->owner == proc_num)
      shmalloc_region_init (ri);

#endif /* not PAGE_BASED */


    return ri;
}


/*
 ************************************************************************
 *			s h r e g i o n _ i n i t			*
 *----------------------------------------------------------------------*
 ************************************************************************
 */

/* Cause reinitialization based on job parameters;
  also declare where the end of pure storage is. */
void
shregion_init ()
{
    vm_offset_t p;
    region_info ri;
    kern_return_t e;
    int region_size;

    MIDWAY_MUTEX_SET_NAME(globlock,"shregion.c:global_mutex");

    bzero(region_info_table, sizeof(region_info_table));
    bzero(region_info_index, sizeof(region_info_index));

    e = vm_shared_init(REGION_SIZE * MAXREGION);
    ASSERT (e == KERN_SUCCESS);

    /* initialize dirty bit regions */
    e = vm_dirty_init(REGION_SIZE * MAXREGION);
    ASSERT (e == KERN_SUCCESS);

    
    if (midway_default_static_linesize == 0)  {
	midway_default_static_linesize = MIDWAY_DEFAULT_STATIC_LINESIZE;
    }

#ifdef PAGE_BASED
    /*
     * Nothing left to do for page-base version of midway.
     */
    return;
#endif

    /*
     * create region information for the static data.  Currently,
     * all static data gets region information, even though we only
     * need it for shared static data.  We eventually need to hack the
     * loader to separate shared static data from private static data
     * but hacking the loader isn't much fun.
     *
     * NOTE: In the best case, the loader will place FIRST_STATIC_SYM
     *	     on a 1 megabyte boundary, otherwise we will allocate
     *	     region_info for the text before the first static symbol.
     *
     */
    num_static_regions = 0; /*  */
    for (p = (vm_offset_t)FIRST_STATIC_SYM;
	 p < (vm_offset_t)LAST_STATIC_SYM;
	 p += REGION_SIZE)  {
	/* 
	 * Don't bother allocating region info for unused chunk 
	 * of static memory 
	 */
	region_size = (vm_offset_t)LAST_STATIC_SYM - p;
	if (region_size > REGION_SIZE) {
	    region_size = REGION_SIZE;
	}

	ri = region_init(p, region_size, midway_default_static_linesize);
#ifdef INIT_DIRTYBITS
	shregion_allocated_memory (ri, (void *)p, region_size);
#endif
	
	/* 
	 * Verify that a dirtybit template is stored at the start of the region
	 * Then initialize it.
	 */
	if (!MIDWAY_TEMPLATE_VALID_P(p)) {
	  fprintf(stderr,
		  "Midway dirtybit template is not stored at statics 0x%x\n", p);
	  botch("shregion_init");
	}

	if (midway_default_static_linesize == sizeof(int))
	  ri->type = REGION_TYPE_WORD_EC;
	else
	  ri->type = REGION_TYPE_ALIGNED_EC;
	ipatch_template (ri);

	num_static_regions++;
    }
    end_static_ri = ri;

    if ((IS_DBG_MISC_MAIF) || (IS_DBG_MISC_MALL)) {
      printf("ini: at end, num static regions now %lu\n", num_static_regions);
    }
    printf("statics: base = 0x%x end =  0x%x, length = %d KB\n",
	   FIRST_STATIC_SYM, LAST_STATIC_SYM, 
	   (LAST_STATIC_SYM - FIRST_STATIC_SYM) /1024);

    if (num_static_regions > 1) {
      fprintf(stderr,"Static initialization failed. Static size > region size\n");
      botch("shregion_init");
    }

}


#if ALLOC_INFO_HACK
/*
 * Hack to allow alloc info request from reply context.
 *
 * Right now, if you call get_shmalloc_info() from a reply context,
 * you will stomp on the cpt being used for the original request.
 *
 * The Correct Solution (tm) is to send this information with
 * lock bindings.  This should be done, and integrated with lock
 * bindings (or deltas thereof) being transferred as well.
 *
 * It would also be nice to have requests that are not tied to
 * a particular thread...
 *
 * 20 Mar 94, mjz
 */
mutex_t mid_aihack_mutex;
condition_t mid_aihack_c;
int midway_aihack_initialized = 0;

typedef struct aihack {
  struct aihack *next;		/* so can deal with > 1 */
  struct condition waiting;	/* where thread that needs this will wait */
  vm_offset_t addr;		/* address in question */
/*  int thread_id;	not easy to get */	/* fyi */
  int done;
} aihack_t;

aihack_t *aihack_anchor;
aihack_t *aihack_last;

void
mid_aihack_loop( int arg )
{
  aihack_t *aihack_p;
  vm_offset_t cur_addr;
  int cur_thread;

  while(1) {
    mutex_lock(mid_aihack_mutex);

    /* 1. wait for work */
    while(aihack_anchor == 0) {
      condition_wait(mid_aihack_c, mid_aihack_mutex);
    }
    /* 2. cache stuff, release mutex */
    aihack_p = aihack_anchor;
    cur_addr = aihack_p->addr;
    mutex_unlock(mid_aihack_mutex);

    /* 3. issue request */
    midway_get_shmalloc_info_internal(cur_addr);

    /* 4. reacquire mutex, delete item from q */
    mutex_lock(mid_aihack_mutex);
    ASSERT(aihack_p == aihack_anchor);

    aihack_anchor = aihack_p->next;
    if (aihack_anchor == 0)
      aihack_last = 0;

    /* 5. release waiter (it will clean up memory, if necc) */
    aihack_p->done = 1;		/* know done */
    mutex_unlock(mid_aihack_mutex);
    condition_signal(&(aihack_p->waiting));
  }
} /* mid_aihack_loop */

void
mid_aihack_do( vm_offset_t addr )
{
  aihack_t hack;

  condition_init(&(hack.waiting));
  hack.next = 0;
  hack.addr = addr;
  hack.done = 0;

  mutex_lock(mid_aihack_mutex);	/* grab queue exclusively */
  if (aihack_last == 0) {
    ASSERT(aihack_anchor == 0);
    aihack_anchor = &hack;
  } else {
    aihack_last->next = &hack;
  }
  aihack_last = &hack;

  condition_signal(mid_aihack_c); /* just in case */

  while (!(hack.done)) {
    condition_wait(&(hack.waiting), mid_aihack_mutex);
  }

  mutex_unlock(mid_aihack_mutex);
} /* mid_aihack_do */

void
mid_aihack_init( void )
{
  if (midway_comm_network_type != MIDWAY_NET_ATM) {
    /* no need for the hack if we are not running ATM */
    return;
  }

  mid_aihack_mutex = mutex_alloc();
  MIDWAY_MUTEX_SET_NAME(mid_aihack_mutex, "aihack");
  mid_aihack_c = condition_alloc();

  aihack_anchor = 0;
  aihack_last = 0;

  midway_aihack_initialized = 1; /* gag */

  cthread_detach(cthread_fork((cthread_fn_t)mid_aihack_loop, 0));
} /* mid_aihack_init */

kern_return_t
get_shmalloc_info( addr )
     char * addr;
{
  region_info ri;
  extern int midway_atm_wantreply_active;

  if (midway_atm_wantreply_active == 0) {
    /* just do it the usual way */
    return( midway_get_shmalloc_info_internal( addr ) );
  }

  /* ensure we have to go through this mess */
  ri = region_from_addr( (int *)addr );

  if (ri->line_size)  {
      /* no need to continue */
      return KERN_SUCCESS;
  }

  ASSERT(midway_aihack_initialized != 0);

  mid_aihack_do( (vm_offset_t) addr ); /* get another thread to do it. Yuck. */

  return(KERN_SUCCESS);		/* assume success.  Yuck. */
} /* get_shmalloc_info under hack */

#endif /*ALLOC_INFO_HACK*/
