/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * sys_stat: a file for throwing in routines to do general system statistics
 * so we can try and do some performance debugging.
 *
 * 9 Jun 93 (mjz)
 */
#include <mach.h>

static struct vm_statistics last_stat = {0}; /* so we can do deltas */
extern task_t midway_task_self; /* mach_task_self() */

void midway_reset_sys_vmfaults( void )
{
  kern_return_t krc;

  krc = vm_statistics(midway_task_self, &last_stat);
  if (krc != KERN_SUCCESS) {
    mach_error("midway_reset_sys_vmfaults: vm_statistics", krc);
  }

  return;
} /* midway_reset_sys_vmfaults */

/*integer_t: only for > MK80 systems.... */
long midway_get_sys_vmfaults( void )
{
  long result;
  kern_return_t krc;
  struct vm_statistics stats;

  krc = vm_statistics(midway_task_self, &stats);
  if (krc != KERN_SUCCESS) {
    mach_error("midway_get_sys_vmfaults: vm_statistics", krc);
    return(0);			/* something safe */
  }

  /* for now, just return # faults.  In the future, return more? */
  result = stats.faults - last_stat.faults;

  last_stat = stats;		/* remember for next time */

  return( result );
} /* midway_get_sys_vmfaults */

void midway_get_sys_e_vmfaults(long *faults, /* general vm faults */
			       long *zfill,  /* # that were zero filled */
			       long *react,  /* # that were reactivated */
			       long *pagein, /* # that were page-ins */
			       long *pageout, /* # that were ? page-out */
			       long *cow /* # that were copy-on-write */
			       )
{
  kern_return_t krc;
  struct vm_statistics stats;

  krc = vm_statistics(midway_task_self, &stats);
  if (krc != KERN_SUCCESS) {
    mach_error("midway_get_sys_e_vmfaults: vm_statistics", krc);
    return;			/* something safe */
  }

  *faults = stats.faults - last_stat.faults;
  *zfill = stats.zero_fill_count - last_stat.zero_fill_count;
  *react = stats.reactivations - last_stat.reactivations;
  *pagein = stats.pageins - last_stat.pageins;
  *pageout = stats.pageouts - last_stat.pageouts;
  *cow = stats.cow_faults - last_stat.cow_faults;

  last_stat = stats;		/* remember for next time */

  return;
} /* midway_get_sys_e_vmfaults */			       


/*
 * Get Midway VM stats too.  This strictly does not belong here, but
 * it is the fastest way to get it into the system... --mjz 17jun93
 */
#include <midway-dbg.h>		/* for STATS/FAST_STATS */
#include <cthreads.h>		/* xpager needs mutex_t */
#include <xpager.h>
/*
 * gak.  this structure matches the one in xpager/xpager.h
 */
#if STATS
static struct{
  unsigned long data_reqs;
  unsigned long data_writes;
  unsigned long data_returns;
  unsigned long data_zeroed;
  unsigned long shared_zeroed;
  unsigned long shared_filled;
  unsigned long dirty_filled;
  unsigned long backing_reads;
  unsigned long backing_writes;
  unsigned long backing_releases;
} last_xpager_stat = {0};
#endif /*STATS*/

void midway_reset_vmfaults( void )
{
#if STATS
  /*  last_xpager_stat = xpager_stat; CANNOT DO THIS -- TYPES DO NOT MATCH */
  last_xpager_stat.data_reqs = xpager_stat.data_reqs;
  last_xpager_stat.data_writes = xpager_stat.data_writes;
  last_xpager_stat.data_returns = xpager_stat.data_returns;
  last_xpager_stat.data_zeroed = xpager_stat.data_zeroed;
  last_xpager_stat.shared_zeroed = xpager_stat.shared_zeroed;
  last_xpager_stat.shared_filled = xpager_stat.shared_filled;
  last_xpager_stat.dirty_filled = xpager_stat.dirty_filled;
  last_xpager_stat.backing_reads = xpager_stat.backing_reads;
  last_xpager_stat.backing_writes = xpager_stat.backing_writes;
  last_xpager_stat.backing_releases = xpager_stat.backing_releases;
#endif /*STATS*/

  return;
} /* midway_reset_vmfaults */

void midway_get_vmfaults(unsigned long *data_reqs,
			 unsigned long *data_writes,
			 unsigned long *data_returns,
			 unsigned long *data_zeroed,
			 unsigned long *shared_zeroed,
			 unsigned long *shared_filled,
			 unsigned long *dirty_filled,
			 unsigned long *backing_reads,
			 unsigned long *backing_writes,
			 unsigned long *backing_releases
			 )
{
#if STATS
  *data_reqs = xpager_stat.data_reqs - last_xpager_stat.data_reqs;
  *data_writes = xpager_stat.data_writes - last_xpager_stat.data_writes;
  *data_returns = xpager_stat.data_returns - last_xpager_stat.data_returns;
  *data_zeroed = xpager_stat.data_zeroed - last_xpager_stat.data_zeroed;
  *shared_zeroed = xpager_stat.shared_zeroed - last_xpager_stat.shared_zeroed;
  *shared_filled = xpager_stat.shared_filled - last_xpager_stat.shared_filled;
  *dirty_filled = xpager_stat.dirty_filled - last_xpager_stat.dirty_filled;
  *backing_reads = xpager_stat.backing_reads - last_xpager_stat.backing_reads;
  *backing_writes = xpager_stat.backing_writes - last_xpager_stat.backing_writes;
  *backing_releases = xpager_stat.backing_releases - last_xpager_stat.backing_releases;

  /*  last_xpager_stat = xpager_stat;*/
  last_xpager_stat.data_reqs = xpager_stat.data_reqs;
  last_xpager_stat.data_writes = xpager_stat.data_writes;
  last_xpager_stat.data_returns = xpager_stat.data_returns;
  last_xpager_stat.data_zeroed = xpager_stat.data_zeroed;
  last_xpager_stat.shared_zeroed = xpager_stat.shared_zeroed;
  last_xpager_stat.shared_filled = xpager_stat.shared_filled;
  last_xpager_stat.dirty_filled = xpager_stat.dirty_filled;
  last_xpager_stat.backing_reads = xpager_stat.backing_reads;
  last_xpager_stat.backing_writes = xpager_stat.backing_writes;
  last_xpager_stat.backing_releases = xpager_stat.backing_releases;
#endif

  return;
} /* midway_get_vmfaults */
