/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * updateq.c: Routines to maintain/flush the update queue
 * 
 *
 * 13 Jun 92 (was): Initial code 
 *  2 Sep 92 (mjz): remove mach 2.5 isms
 * 21 Sep 92 (was): Fixed bug in freeing packet buffer
 * 23 Sep 92 (was): Check if client change was to static/private data.
 * 15 Oct 92 (was): changed MAX_DATA_SIZE to mid_max_data_size
 * 20 Oct 92 (was): added midway_task_self for mach_task_self()
 * 22 Oct 92 (was): added stats, changed queue elements to have >1 entry
 * 27 Oct 92 (was): try to merge updates w/ last element on update queue.
 * 29 Oct 92 (was): Puts statistics under STATS instead of DEBUG
 * 19 Dec 92 (was): malloc histogram tables
 * 12 Jan 92 (was): Optimized enqueue of updates
 */
#include <sys/types.h>		/* u_long & friends */
#include <mach.h>
#include <mach/kern_return.h>
#include <cthreads.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>		/* NULL */

#ifdef mach_2_5_way
#include <mach/message.h>
#include <mach/exception.h>
#endif

#include "midway_types.h"       /* global type definitions */
#include "lock.h"		/* perlock struct, lock request formats */
#include "comm.h"		/* M_xxx_OP */
#include "smexecd_types.h"	/* ensure get char255 */
#include "pbuf.h"		/* packet buffer stuff */
#include "shmalloc.h"		/* memory region definitions */
#include "vm.h"
#include "midway-dbg.h"		/* filter debug messages */
#include "pagemgr.h"            /* prototypes for calling page mgr rtns */

#define __UPDATEQ_INTERNAL__
#include "updateq.h"            /* generate internal defs for updateq */


/*
 * Define external variables
 */
extern int Comm_index;          /* this processors "number" (1-n) */
extern int Comm_hnum;           /* total number of processors */
extern task_t midway_task_self;	/* mach_task_self() */

extern vm_offset_t vm_shared_base;
extern vm_offset_t vm_shared_end;


#ifdef STATS

/*
 * Define Static variables for collecting statistics
 */
#define STAT_DATA_PER_BUCKET  32
#define STAT_MAX_DATA   (64 * 1024)
#define STAT_DATA_TABLE_SIZE (STAT_MAX_DATA / STAT_DATA_PER_BUCKET)

#define STAT_ITEMS_PER_BUCKET 4
#define STAT_MAX_ITEMS  (STAT_MAX_DATA / 8)
#define STAT_ITEMS_TABLE_SIZE (STAT_MAX_ITEMS / STAT_ITEMS_PER_BUCKET)

#define STAT_QUEUE_PER_BUCKET    4
#define STAT_MAX_QUEUE         256
#define STAT_QUEUE_TABLE_SIZE (STAT_MAX_QUEUE / STAT_QUEUE_PER_BUCKET)

static struct {
  u_long  FlushCalls;
  u_long  FlushCallers[ UPDATEQ_CALLERS ];
  u_long  FlushEmptyQueue;
  u_long  FlushTotalItems;
  u_long  FlushItemRecipients[MAX_PROCESSORS];
  u_long *FlushItemsEachP;
  u_long  FlushItemsByCaller[ UPDATEQ_CALLERS ];
  u_long  FlushRequests;
  u_long  FlushRequestRecipients[MAX_PROCESSORS];
  u_long  FlushTotalData;
  u_long *FlushDataEachP;
  u_long  FlushDataByCaller[ UPDATEQ_CALLERS ];
  u_long  FlushErrors;

  u_long  Requests;
  u_long  RequestUpdates;
  u_long  RequestUpdateData;
  u_long  RequestApplied;
  u_long  RequestAppliedData;
  u_long  RequestIgnored;
  u_long  RequestPending;
  u_long  RequestNacked;
  u_long  RequestErrors;

  u_long  Replies;
  u_long  ReplyErrors;
  u_long  ReplyNacked;

  u_long  PendingQueueCalls;
  u_long  PendingQueueData;

  u_long  QueueAdds;
  u_long  QueueUnshared;
  u_long  QueueMergedPrev;
  u_long  QueueMergedNext;
  u_long  QueueTotalData;
  u_long *QueueDataEachP;
  
} upStat;
static mutex_t upStatLock;

# ifdef FAST_STATS
#  define INCR_STATS(x,by) (x) += (by)
#  define IF_STATS(x) x
#  define LOCK_STATS()
#  define FAST_INCR_STATS(x,by) (x) += (by)
#  define UNLOCK_STATS()
# else
#  define INCR_STATS(x,by)         \
     { mutex_lock(upStatLock); (x) += (by); mutex_unlock(upStatLock); }
#  define IF_STATS(x) x
#  define LOCK_STATS() mutex_lock(upStatLock)
#  define FAST_INCR_STATS(x,by) (x) += (by)
#  define UNLOCK_STATS() mutex_unlock(upStatLock)
# endif

#else /* not compiled to collect statistics */

#define INCR_STATS(x,by)	/* nothing to do */
#define IF_STATS(x)
#define LOCK_STATS()
#define FAST_INCR_STATS(x,by)
#define UNLOCK_STATS()

#endif /* STATS */



/*
 ************************************************************************
 *
 * alloc_updateq_elem
 *
 * Internal routines to allocate/free an update queue element
 *
 ************************************************************************
 */
static q_elem_t *alloc_updateq_elem ()
{
  q_elem_t *elemP;

  /* Allocate a new element */
  elemP = updateq_freeP;
  if (elemP != NULL) 
    updateq_freeP = elemP->nextP;
  else {
    elemP = (q_elem_t *) MALLOC (sizeof(*elemP));
    ASSERT (elemP != NULL);
    
    elemP->null_entry.dataAddr = (vm_address_t) 0xFFFFFFFF;
    elemP->null_entry.dataLen = 0;
  }

  return (elemP);
}

static void enqueue_updateq_elem (q_elem_t *elemP)
{
  /* insert element at tail of update queue */
  if (updateq_tailP != NULL)
    updateq_tailP->nextP = elemP;
  else
    updateq_headP = elemP;

  updateq_tailP = elemP;
  elemP->nextP = NULL;

  mid_updateq_prev_entryP = &elemP->null_entry;
  mid_updateq_last_entryP = &elemP->entry[ QENTRIES_PER_ELEM-1 ];
}


/*
 ************************************************************************
 *
 * mid_updateq_init
 *
 * Perform one time initialization of update queue.
 * Start slave thread to periodically flush the update queue.
 *
 ************************************************************************
 */
/*
 * start_slave
 * 
 * internal routine to start/restart the update queue slave thread
 */
static
start_slave ()
{
  char255 smsg;
  static void updateq_server ();   /* forward declaration */

  if (IS_DBG_PORT_QUEUE) {
    sprintf(smsg, "mid_updateq: starting slave thread, period=%d ms",
	    midway_default_flush_period);
    StatusMsg (smsg);
  }
    
  /* spawn thread */
  cthread_detach(cthread_fork(updateq_server, 0)); /* kick 'er off */
}


kern_return_t 
  mid_updateq_init ()
{
  kern_return_t krc;
  q_elem_t *elemP;
  char255 smsg;

  if (IS_DBG_PORT_QUEUE) {
    sprintf(smsg, "mid_updateq_init, update_period = %d ms",
	    midway_default_flush_period);
    StatusMsg(smsg);
  }

  /* initialize external variables */
  pending_headP = NULL;
  updateq_headP = NULL;
  updateq_tailP = NULL;
  elemP = alloc_updateq_elem();
  enqueue_updateq_elem( elemP );

  /* Allocate mutex to guard update queue */
  mid_updateq_mutex = mutex_alloc();
  MIDWAY_MUTEX_SET_NAME(mid_updateq_mutex,"mid_updateq_mutex");
  
  /* Allocate mutex to serialize flushing of the update queue */
  updateq_flushing = mutex_alloc();
  MIDWAY_MUTEX_SET_NAME(updateq_flushing,"updateq_flushing");

  /* Allocate mutex to guard pending update queue */
  pending_mutex = mutex_alloc();
  MIDWAY_MUTEX_SET_NAME(pending_mutex,"updateq.c:pending_mutex");

#ifdef STATS
  /* Initialize stats */
  upStatLock = mutex_alloc();
  MIDWAY_MUTEX_SET_NAME(upStatLock,"updateq.c:upStatLock");
  bzero (&upStat, sizeof(upStat));

  /* malloc space for histogram tables, add +1 for overflows */
  upStat.FlushItemsEachP = (u_long *) CALLOC( (STAT_ITEMS_TABLE_SIZE + 1),
					       sizeof(u_long));
  ASSERT (upStat.FlushItemsEachP != NULL);

  upStat.FlushDataEachP = (u_long *) CALLOC( (STAT_DATA_TABLE_SIZE + 1),
					      sizeof(u_long));
  ASSERT (upStat.FlushDataEachP != NULL);

  upStat.QueueDataEachP = (u_long *) CALLOC( (STAT_QUEUE_TABLE_SIZE + 1),
					      sizeof(u_long));
  ASSERT (upStat.QueueDataEachP != NULL);

#endif 

#ifdef mach_2_5_way  
  /* allocate port for update queue thread to use for timing */
  krc = port_allocate(midway_task_self, &updateq_port);
  if (krc != KERN_SUCCESS) {
    sprintf(smsg,"Error in mid_updatq_init from port_allocate: %s (%d)",
	    mach_error_string(krc), krc);
    StatusMsg(smsg);
    goto done;
  }
	
  if (IS_DBG_PORT_QUEUE) {
    sprintf(smsg,"mid_updateq_init: port=%d", updateq_port);
    StatusMsg (smsg);
  }
#else
  krc = KERN_SUCCESS;
#endif
	
  /* Check if the period for flushing the queue is non-zero */
  if (midway_default_flush_period != 0)
    start_slave ();
    

done:

  if (IS_DBG_PORT_QUEUE) {
    sprintf(smsg, "mid_updateq_init complete, rc=%d", krc);
    StatusMsg (smsg);
  }


#ifdef mach_2_5_way
  /* since caller does not check return codes */
  ASSERT (krc == KERN_SUCCESS);
  
  /* and return */
  return (krc);
#else
  return( KERN_SUCCESS );
#endif

} /* mid_updateq_init */

/*
 ************************************************************************
 *
 * updateq_server
 *
 * Starting point for slave thread.
 * Thread loops forever waiting for elements to be put on 
 * the update queue. When found, it dequeues them and 
 * sends them to the necessary processors.
 *
 ************************************************************************
 */
static void
updateq_server()
{
  s_addr_map_t *listP;          /* ptr to queue element */
#ifdef mach_2_5_way
  msg_header_t msgh;		/* never really used here */
  msg_timeout_t timeout;	/* milliseconds */
  msg_return_t mrc;
#endif
  int rc;
  char255 smsg;		        /* for status messages */


  /* we are cthread_fork'd; for now wire all cthreads with 3.0 threads*/
  cthread_wire();

  if (IS_DBG_PORT_QUEUE)
    StatusMsg("->->UpdateQ_Server: starting");

  /* indicate that we are running */
  mutex_lock(mid_updateq_mutex);
  ASSERT (updateq_server_running == FALSE);
  updateq_server_running = TRUE;
  mutex_unlock(mid_updateq_mutex);


#ifdef mach_2_5_way
  /* set up message */
  msgh.msg_local_port = updateq_port;
  msgh.msg_size = sizeof(msg_header_t);	/* only message header */
#endif

  /* Loop forever */
  while (1) {

    /*
     * Check if we were told to terminate
     */
#ifdef mach_2_5_way
    timeout = (msg_timeout_t) midway_default_flush_period;
    if (timeout <= 0)
#else
    if (midway_default_flush_period <= 0)
#endif
      goto done;
    

    /*
    ** Wait for updateq_period milliseconds
    */
#ifdef mach_2_5_way
    mrc = msg_receive(&msgh, RCV_TIMEOUT, timeout); /* wait for timeout */
    if ((mrc != RCV_TIMED_OUT) && 
	((mrc != RCV_SUCCESS) || (msgh.msg_id != UPDATEQ_WAKEUP_MSGID))) {
      sprintf(smsg, 
	      "UpdateQ_Server: unexpected return from msg_receive: %s (%d)",
	      mach_error_string(mrc), mrc);
      StatusMsg(smsg);
      goto done;		/* die! */
    } /* error */
#endif
    timer_pause_ms(midway_default_flush_period);


    /*
    ** Flush update queue 
    */
    if (updateq_headP != NULL) {

      if (IS_DBG_PORT_QUEUE)
	StatusMsg("->->UpdateQ_Server: flushing update queueg");

      rc = mid_updateq_flush (UPDATEQ_CALLER_TIMER);
      if (rc != 0) {
	fprintf(stderr, "Error flushing update queue, rc=%d\n", rc);
	fflush (stderr);
	goto done;
      }
    }

  } /* while (1) */

done:

  /* indicate that we are no longer running */
  mutex_lock(mid_updateq_mutex);
  ASSERT (updateq_server_running != FALSE);
  updateq_server_running = FALSE;
  mutex_unlock(mid_updateq_mutex);

} /* updateq_server */



/*
 ************************************************************************
 *
 * add_to_pending_queue
 *
 * Internal routine to queue a page update to a page that is 
 * in the process of being allocated. Once the page is received
 * the pending updates may be applied.
 *
 ************************************************************************
 */
static void
add_to_pending_queue
  (
   vm_address_t   dataAddr,	/* address of update */
   u_long         dataLen,	/* length of update */
   char          *dataP		/* ptr to data for update */
   )
{
  pending_elem_t *pendP;


  if (IS_DBG_PORT_QUEUE) {
    char255 smsg;
    sprintf(smsg, "add_update_to_pending_queue (addr=0x%lx len=%lu data=0x%x)",
	    dataAddr, dataLen, dataP);
    StatusMsg(smsg);
  }
  INCR_STATS (upStat.PendingQueueCalls, 1);
  INCR_STATS (upStat.PendingQueueData, dataLen);

  /* Allocate element to store the update */
  pendP = (pending_elem_t *) MALLOC (dataLen + 
				     (sizeof(*pendP) - sizeof(pendP->data)));
  ASSERT (pendP != NULL);

  /* Copy data into queue element */
  pendP->nextP = NULL;
  pendP->dataAddr = dataAddr;
  pendP->dataLen = dataLen;
  bcopy (dataP,        /* source */
	 pendP->data,  /* target */
	 dataLen);     /* length */

  /* Add element to pending update queue */
  mutex_lock(pending_mutex);
  if (pending_tailP != NULL)
    pending_tailP->nextP = pendP;
  else
    pending_headP = pendP;
  pending_tailP = pendP;
  mutex_unlock(pending_mutex);

  return;
}


/*
 ************************************************************************
 *
 * mid_updateq_apply_pending
 *
 * External routine to apply pending updates to an allocated page.
 * Note that the page's data does not yet reside in the allocated page
 * but in the page buffer provided. All updates are done based on offsets.
 *
 ************************************************************************
 */
void
mid_updateq_apply_pending
  (
   char          *pageP,        /* ptr to page buffer to update */
   vm_address_t   pageAddr,	/* address of page to update */
   u_long         pageLen 	/* length of page to update */
   )
{
  pending_elem_t *pendP;
  pending_elem_t *prevP;
  pending_elem_t *nextP;
  vm_address_t nextPageAddr;
  u_long pageOffset;
  

  if (IS_DBG_PORT_QUEUE) {
    char255 smsg;
    sprintf(smsg, "Apply Pending Updates (pageP=0x%x pageAddr=0x%lx len=%lu)",
	    pageP, pageAddr, pageLen);
    StatusMsg(smsg);
  }

  /* compute address for next page */
  nextPageAddr = pageAddr + pageLen;

  /* gain exclusive access to pending update queue */
  mutex_lock (pending_mutex);

  /*
   * Loop to check each element on the queue
   */
  prevP = NULL;
  for (pendP = pending_headP; pendP != NULL; pendP = nextP) {

    /* get pointer to next element */
    nextP = pendP->nextP;
    
    /* Check if the update applies to the requested page */
    if ((pendP->dataAddr >= pageAddr) && 
	(pendP->dataAddr < nextPageAddr)) {

      if (IS_DBG_PORT_QUEUE) {
	char255 smsg;
	sprintf(smsg, "Applying update to addr=0x%lx off=%lu len=%lu",
		pendP->dataAddr, 
		(u_long)((char *)pendP->dataAddr - (char *)pageAddr),
		pendP->dataLen);
	StatusMsg(smsg);
      }

      /* verify that pending updates don't span pages */
      ASSERT ((pendP->dataAddr + pendP->dataLen) <= nextPageAddr);

      /* compute update's offset into page */
      pageOffset = (u_long)((char *)pendP->dataAddr - (char *)pageAddr);

      /* apply update to same offset in page buffer */
      bcopy (pendP->data,            /* source */
	     &pageP[ pageOffset ],   /* target */
	     pendP->dataLen);        /* length */

      /* remove the element from the pending queue */
      if (prevP != NULL)
	prevP->nextP = nextP;
      else
	pending_headP = nextP;

      if (nextP == NULL)
	pending_tailP = prevP;
      
      /* free the element, destroy its contents for debugging only */
      SMASH_AREA (pendP, 
		  pendP->dataLen + ((sizeof(*pendP) - sizeof(pendP->data))));
      FREE (pendP);

    } else 
      prevP = pendP;   /* update pointer to previous element */
  }

  /* release mutex before returning */
  mutex_unlock (pending_mutex);

  return;
}
/*
 ************************************************************************
 *
 * count_recipients
 *
 * Internal routine to count number of recipients in a copyset
 * Only used for collecting statistics on amount of sharing.
 *
 ************************************************************************
 */
static int
count_recipients (addr_vector_t addr)
{
  register int i, c;
  register u_long x;

  c = 0;
  for (i = 0; i < ADDR_VECTOR_WORDS; i++) {
    x = addr.bitArray[i];
    while (x) {
      if (x & 0x80000000) c++;
      if (x & 0x40000000) c++;
      if (x & 0x20000000) c++;
      if (x & 0x10000000) c++;
      x <<= 4;
    }
  }
  return (c);
}


/*
 ************************************************************************
 *
 * mid_updateq_flush
 *
 * Externally callable routine to transmist list of updates
 * on the update queue to all members of the copyset. The caller
 * must waits for the transmissions to complete & be acked AND
 * wait for all previous transmissions to complete & be acked.
 *
 ************************************************************************
 */
int
mid_updateq_flush  (int caller)
{
  q_elem_t *listP;          /* ptr to head of update list */
  q_elem_t *tailP;          /* ptr to tail of update list */
  q_elem_t *freeP;          /* ptr to element about to be freed */
  q_entry_t *itemP;	    /* ptr to entry within current element */
  pbuf bufP;                /* ptr to packet buffer */
  u_long bufOffset;         /* length of data in packet buffer */
  addrpair_t *aP;           /* ptr to address/length pair in buffer */
  int rc;                   /* return code */
  u_long max_data;          /* cache external variable mid_max_data */
  int comm_index;           /* cache of external variable Comm_index */
  addr_vector_t destAddr;   /* message destination address vector */
  addr_vector_t copySet;    /* copyset address vector */
  char *sprintf_addr_vector();
  register vm_address_t dataAddr;    /* cache of itemP->dataAddr */
  register vm_address_t dataLen;     /* cache of itemP->dataLen */
  u_long copyLen;

  IF_STATS (int statRecip);     /* variables used for collecting stats */
  IF_STATS (u_long statDataCount = 0);
  IF_STATS (u_long statItemCount = 0);


  /* initialize return code */
  rc = 0; 

  /* Serialize the flushing of the update queue.
   * Each thread wanting to flush the queue MUST wait
   * for all previous flushes to complete.
   */
  mutex_lock (updateq_flushing);

  if (IS_DBG_PORT_QUEUE)
    StatusMsg("->->UpdateQ_Flush: flush update queue starting");

  INCR_STATS (upStat.FlushCalls, 1);
  INCR_STATS (upStat.FlushCallers[caller], 1);


  /* allocate a free queue element */
  freeP = alloc_updateq_elem ();

  /* 
   * Remove all items from update queue.
   */
  mutex_lock(mid_updateq_mutex);

  listP = updateq_headP;
  ASSERT (listP != NULL);
  ASSERT (updateq_tailP != NULL);
  ASSERT ((mid_updateq_prev_entryP >= &updateq_tailP->null_entry) &&
	  (mid_updateq_prev_entryP 
	     <= &updateq_tailP->entry[ QENTRIES_PER_ELEM-1 ]));
  ASSERT (mid_updateq_last_entryP 
	     == &updateq_tailP->entry[ QENTRIES_PER_ELEM-1 ]);
  

  /* mark end of list */
  if (mid_updateq_prev_entryP < mid_updateq_last_entryP) {
    itemP = mid_updateq_prev_entryP + 1;
    itemP->dataAddr = 0;
  }

  /* remove old elements from list,
   * put new element at head of list.
   */
  tailP = updateq_tailP;	/* save ptr to tail for later */
  updateq_tailP = NULL;
  enqueue_updateq_elem (freeP);

  mutex_unlock(mid_updateq_mutex);


  /*
   * Special case for stand-alone execution
   */
  if (Comm_hnum == 1) {

    /* move all elements to the free list */
    mutex_lock (mid_updateq_mutex);
    tailP->nextP = updateq_freeP;
    updateq_freeP = listP;
    mutex_unlock (mid_updateq_mutex);
    goto done;
  }

    
  /*
  ** NOTE: Currently, all updates are merged into a single buffer
  ** and sent to the union of the copysets. This should be changed
  ** to send each node its own buffer of the updates that concern it.
  ** WAS 16Aug92
  */

     
  /* zero destination address vector */
  ZERO_ADDR_VECTOR (destAddr);

  /* allocate packet buffer */
  bufP = mid_pbuf_get ();
  ASSERT (bufP != NULL);
  bufOffset = 0;
  max_data = mid_max_data_size;   /* cache externs in a register */
  comm_index = Comm_index;

  /*
   * Loop through update queue & construct an update message
   * Must also construct list of recipients.
   * NOTE: Updates must appear in message in same order as they were written.
   */
  do {  /* while (listP != NULL) */

    /* loop over each entry in the elem's table */
    itemP = listP->entry; 
    do {

      /* cache item values */
      dataAddr = itemP->dataAddr;
      dataLen = itemP->dataLen;

      /* check for end of list */
      if (dataAddr == 0) break;

      /* verify entry */
      ASSERT (dataLen > 0);
      ASSERT (dataAddr > (vm_address_t) vm_page_size);
      
      /* verify that the change doesn't cross shared/private boundary */
      ASSERT ((((char *)dataAddr) 
  	       >= (char *)SHMALLOC_ALLOCATION_BASE) ||
	      ((((char *)dataAddr)+dataLen)
	       < (char *) SHMALLOC_ALLOCATION_BASE));

      /*  
       * Obtain copy set for the page & round address/size to line multiples
       * This also resets the dirty bits on the cache line info.
       * Move register vars to memory so they can be modified by this call.
       */
      { vm_address_t tmpAddr = dataAddr;
	u_long tmpLen = dataLen;
	mid_page_copyset (&tmpAddr, &tmpLen, &copySet);
	dataAddr = tmpAddr;
	dataLen = tmpLen;
      }

      IF_STATS (statItemCount += 1);
      IF_STATS (statRecip = count_recipients(copySet));
      IF_STATS (ASSERT (statRecip > 0));
      INCR_STATS (upStat.FlushItemRecipients[statRecip-1], 1);
      
      /* Check if we are the only member of the copyset */
      midClearBit (copySet.bitArray, comm_index);
      if (EQUAL_ZERO_ADDR_VECTOR (copySet))
	goto nextItem;

      /* Construct destination address by ORing copysets */
      MERGE_ADDR_VECTOR (destAddr, copySet);
      

      
sameItem:
      
				
      if (IS_DBG_PORT_QUEUE) {
	char255 smsg, s1;        
	sprintf(smsg, 
		"->->UpdateQ_Flush: 0x%x update to 0x%x for %d, CS=0x%s CI=%d, off=%d",
		itemP, dataAddr, dataLen, 
		sprintf_addr_vector (s1, copySet), comm_index, bufOffset);
	StatusMsg (smsg);
      }
      
      /* Verify the buffer offset is on a word boundary */
      ASSERT ((bufOffset & 0x03) == 0);
      ASSERT ((dataLen & 0x03) == 0);   /* must be a word multiple */
      
      /* construct update buffer
       * Buffer consists of series of addrpair_t records.
       */
      aP = (addrpair_t *) &bufP[ bufOffset ];
      bufOffset += (sizeof(*aP) - sizeof(aP->data));
      
      /* Check if entire update will fit in the buffer */
      if (dataLen <= (max_data - bufOffset))
	copyLen = dataLen;
      else
	copyLen = max_data - bufOffset;
      
      /* Copy data into buffer */
      aP->addr = (char *)dataAddr;
      aP->len = copyLen;
      bcopy (aP->addr,        /* source */
	     aP->data,        /* target */
	     copyLen);        /* length */

      /* update pointers */
      dataAddr  += copyLen;
      dataLen   -= copyLen;
      bufOffset += copyLen; 
      ASSERT (bufOffset <= max_data);

      IF_STATS (statDataCount += copyLen);
      

      /* Check if the buffer is full */
      if ((max_data - bufOffset) < sizeof(*aP)) {
	
	/* No need to send update to ourself */
	midClearBit (destAddr.bitArray, comm_index);
	
	/* Send multicast request & wait for replies */
	rc = CommMrequest ((u_long) 0,         /* lock */
			   M_PAGE_UPDATE,      /* packet type */
			   (u_long) 0,         /* reply to host (0 = us) */
			   (u_long) 0,         /* reply to tid (0 = us) */
			   NULL_CLOCK_VALUE,   /* clock value */
			   destAddr.bitArray,  /* destination addresses */
			   bufP,               /* ptr to message data */
			   bufOffset);         /* length of message data */

	LOCK_STATS ();
	FAST_INCR_STATS (upStat.FlushRequests, 1);
	IF_STATS (statRecip = count_recipients(destAddr));
	IF_STATS (ASSERT (statRecip > 0));
	FAST_INCR_STATS (upStat.FlushRequestRecipients[statRecip-1], 1);
	UNLOCK_STATS ();
	
	/* reset buffer offset, before checking return code  */
	bufOffset = 0;

	/* Exit loop if error was encountered */
	if (rc != 0) goto done;

	/* reset destination address */
	if (dataLen > 0)
	  destAddr = copySet;
	else
	  ZERO_ADDR_VECTOR (destAddr);
      }


      /* Repeat loop, if there is more data in this item */
      if (dataLen > 0)
	goto sameItem;


nextItem:
      
      /* advance to next item in the entry table */
      itemP++;

    } while (itemP < &listP->entry[QENTRIES_PER_ELEM]);
      

    /* free the current element and advance to the next */
    freeP = listP;
    listP = listP->nextP;

    /* smash contents of element before freeing (for debugging only) */
    SMASH_AREA (freeP->entry, sizeof(freeP->entry));

    /* put element on the free list */
    mutex_lock (mid_updateq_mutex);
    freeP->nextP = updateq_freeP;
    updateq_freeP = freeP;
    mutex_unlock (mid_updateq_mutex);
    
  } while (listP != NULL);

  
  /* 
   * Send contents of final buffer & wait for replies
   */
  if (bufOffset > 0) {
    midClearBit (destAddr.bitArray, comm_index);
    rc = CommMrequest ((u_long) 0,         /* lock */
		       M_PAGE_UPDATE,      /* packet type */
		       (u_long) 0,         /* reply to host (0 = us) */
		       (u_long) 0,         /* reply to tid (0 = us) */
		       NULL_CLOCK_VALUE,   /* clock value */
		       destAddr.bitArray,  /* destination addresses */
		       bufP,               /* ptr to message data */
		       bufOffset);         /* length of message data */

    INCR_STATS (upStat.FlushRequests, 1);
    IF_STATS (statRecip = count_recipients(destAddr));
    IF_STATS (ASSERT (statRecip > 0));
    INCR_STATS (upStat.FlushRequestRecipients[statRecip-1], 1);
	
  }

  if (IS_DBG_PORT_QUEUE) {
    char255 smsg;
    sprintf(smsg, "->->UpdateQ_Flush: complete, rc=%d", rc);
    StatusMsg (smsg);
  }
  if (rc) {
    fprintf(stderr, "Error flushing update queue, rc=%d\n", rc);
    fflush (stderr);
    INCR_STATS (upStat.FlushErrors, 1);
  }

  /* free packet buffer */
  mid_pbuf_free (bufP);  

  LOCK_STATS();
  FAST_INCR_STATS (upStat.FlushTotalItems, statItemCount);
  FAST_INCR_STATS (upStat.FlushItemsByCaller[ caller ], statItemCount);
  FAST_INCR_STATS (upStat.FlushTotalData, statDataCount);
  FAST_INCR_STATS (upStat.FlushDataByCaller[ caller ], statDataCount);

  IF_STATS (statItemCount /= STAT_ITEMS_PER_BUCKET);
  IF_STATS (if (statItemCount > STAT_ITEMS_TABLE_SIZE)
	      statItemCount = STAT_ITEMS_TABLE_SIZE);
  FAST_INCR_STATS (upStat.FlushItemsEachP[statItemCount], 1);

  IF_STATS (statDataCount /= STAT_DATA_PER_BUCKET);
  IF_STATS (if (statDataCount > STAT_DATA_TABLE_SIZE)
              statDataCount = STAT_DATA_TABLE_SIZE);
  FAST_INCR_STATS (upStat.FlushDataEachP[statDataCount], 1);
  UNLOCK_STATS();


done:  

  /* Unlock serialization mutex before leaving */
  mutex_unlock (updateq_flushing);

  return (rc);
} /* mid_updateq_flush */



/*
 ************************************************************************
 *
 * mid_updateq_request
 *
 * Externally callable routine to perform list of updates 
 * in a request. Thread MUST have exlusive access to shared
 * memory while updates are being performed.
 *
 ************************************************************************
 */
extern mutex_t lockmutex;         /* external vars needed temporarily */
extern mutex_t mid_pbuf_lock;

void
mid_updateq_request
  (
   u_long host,		/* host sending update info */
   u_long thread,	/* thread@host sending update */
   u_long tid,		/* transaction id for this request */
   u_long *bufP,        /* ptr to request buffer */
   u_long bufLen 	/* length of data in msg */
   )
{
  int rc;
  int status;               /* status of updated page */
  char *dataP;              /* ptr to update data for the page */
  u_long dataLen;           /* length this update */
  u_long bufOffset;         /* length of data in packet buffer */
  addrpair_t *aP;           /* ptr to address/length pair in buffer */
  region_info riP;          /* ptr to region info for updated address */
  u_long pageLen;           /* length of update on current page */
  vm_address_t pageAddr;    /* page address for update */
  vm_size_t page_size; 	    /* cache vm_page_size */
  m_updateq_reply_t rep;    /* reply message buffer */

  IF_STATS(int statIncr);   /* define variable used for stats */


  if (IS_DBG_PORT_QUEUE) {
    char255 smsg;
    sprintf(smsg,"->->UpdateQ_Request: (host=%lu thr=%lu tid=%lu len=%lu)",
	    host, thread, tid, bufLen);
    StatusMsg(smsg);
  }
  INCR_STATS (upStat.Requests, 1);

  /*
  ** NOTE: Updates must either be applied atomically, or
  ** applied in operational order (with the correct value).
  ** Currently, we do not record all writes to shared memory,
  ** only 1 write to each cache line and the values transmitted
  ** in the update buffer are the final values (not the intermediate).
  ** Thus the update must be applied atomically.  Currently, this is
  ** not done, since the update may split into multiple request buffers.
  ** WAS 16Aug92
  */

  /*
   * cache global vars
   */
  page_size = vm_page_size;

  /* lock everything that we can */
  mutex_lock (lockmutex);
  acquire_shmalloc_mutex ();
  mutex_lock (mid_pbuf_lock);
  mutex_lock (mid_updateq_mutex);

  /*
  ** We should also suspend all other threads. WAS 16Aug92
  */

  
  /* Initialize return code to indicate no updates were applied */
  rep.status = UPDATEQ_REPLY_STATUS_NACK;

  /*
   * Loop, applying each update in the buffer 
   */
  bufOffset = 0;
  while (bufOffset < bufLen) {

    /* get pointer to update record */
    aP = (addrpair_t *) &bufP[ bufOffset / sizeof(bufP[0]) ];

    if (IS_DBG_PORT_QUEUE) {
      char255 smsg;
      sprintf(smsg,"->->UpdateQ_Request: update 0x%x for %d bytes, off=%d",
	      aP->addr, aP->len, bufOffset);
      StatusMsg(smsg);
    }

    /* verify that sufficient data was sent */
    if ((bufLen - bufOffset) < (aP->len + (sizeof(*aP) - sizeof(aP->data)))) {
      
      fprintf(stderr, "Invalid Updateq_request: from host=%lu thd=%lu\n",
	      host, thread);
      fprintf(stderr, 
	      "Insufficient data for update, bufLen=%d Left=%d Reqd=%d\n",
	      bufLen, bufLen - bufOffset, 
	      (aP->len + (sizeof(*aP) - sizeof(aP->data))));
      fflush(stderr);
      INCR_STATS (upStat.RequestErrors, 1);
      break;
    }


    INCR_STATS(upStat.RequestUpdates, 1);
    IF_STATS(statIncr = 1);
    

    /* Advance buffer offset to next record */
    bufOffset += (aP->len + (sizeof(*aP) - sizeof(aP->data)));
    ASSERT ((bufOffset & 0x03) == 0);      /* must be word aligned */
    ASSERT (bufOffset <= mid_max_data_size);  /* must stay within buffer */

    /*
     * Break the original update into page updates
     */
    dataP = (char *)aP->data;
    while (aP->len > 0) {

      /* determine amount that changed on this page */
      pageAddr = PAGENO_TO_ADDR( ADDR_TO_PAGENO( (vm_address_t)aP->addr ));

      pageLen  = page_size - ((char *)aP->addr - (char *)pageAddr);
      dataLen = (aP->len < pageLen) ? aP->len : pageLen;

      INCR_STATS (upStat.RequestUpdateData, dataLen);
      
      /* Check if page has been allocated on this processor */
      status = mid_page_status ((vm_address_t)aP->addr);
      if (status == PAGE_STATUS_ALLOCATED) {

	/* apply update to the page */
	bcopy (dataP,        /* source */
	       aP->addr,     /* target */
	       dataLen);     /* length */
	rep.status = UPDATEQ_REPLY_STATUS_SUCCESS;  /* update applied */

	INCR_STATS (upStat.RequestApplied, statIncr);
	INCR_STATS (upStat.RequestAppliedData, dataLen);
	
      } else if (status == PAGE_STATUS_PENDING) {
	
	/* Save update on the pending update queue */
	add_to_pending_queue ((vm_address_t)aP->addr, dataLen, dataP);
	rep.status = UPDATEQ_REPLY_STATUS_SUCCESS;  /* update saved */

	INCR_STATS (upStat.RequestPending, statIncr);
      } else { 
	INCR_STATS (upStat.RequestIgnored, statIncr);
      }
	
      IF_STATS (statIncr = 0);

      /* update pointers */
      aP->len  -= dataLen;
      dataP    += dataLen;
      aP->addr += dataLen;

    } /* while (aP->len > 0) */
  } /* while (bufOffset < bufLen) */

  /* unlocking would be nice, too */
  mutex_unlock (lockmutex);
  release_shmalloc_mutex ();
  mutex_unlock (mid_pbuf_lock);
  mutex_unlock (mid_updateq_mutex);

  IF_STATS (if (rep.status == UPDATEQ_REPLY_STATUS_NACK) 
	       INCR_STATS(upStat.RequestNacked, 1));

  /*
  ** NOTE: If we NACK this update, we need to return the request buffer
  ** so that the caller can determine which page copysets we need to be 
  ** removed from.  WAS 28Aug92
  */

  /* ACK the update request */
  rc = CommReply (M_PAGE_UPDATE, 
		  host, thread, tid, 
		  (u_long) 0,          /* lock */
		  TRUE,                /* last packet */
		  (u_long) 0,          /* packet number */
		  NULL_ADDR_VECTOR,    /* number of acks required */
		  (char *) &rep,       /* ptr to message buffer */
		  (u_long) sizeof(rep),/* message length */
		  C_NOFREE);           /* don't free buffer */
  if (rc) {
    fprintf(stderr, "Error sending UPDATE reply to %lu %lu, tid=%lu, rc=%d\n",
	    host, thread, tid, rc);
    fflush (stderr);
    INCR_STATS (upStat.RequestErrors, 1);
  }

  /* and return */
  return;
}  /* mid_updateq_request */




/*
 ************************************************************************
 *
 * mid_updateq_reply
 *
 * Handle reply message to an updateq_request
 *
 ************************************************************************
 */
void
mid_updateq_reply
  (
   u_long host,		/* host sending update info */
   u_long thread,	/* thread@host sending update */
   u_long tid,		/* transaction id for this request */
   u_long *bufP,        /* ptr to request buffer */
   u_long bufLen 	/* length of data in msg */
   )
{
  region_info riP;         /* ptr to region info for updated address */
  m_updateq_reply_t *repP;


  if (IS_DBG_PORT_QUEUE) {
    char255 smsg;
    sprintf(smsg,"->->UpdateQ_Reply: (host=%lu thr=%lu tid=%lu len=%lu)",
	    host, thread, tid, bufLen);
    StatusMsg(smsg);
  }
  INCR_STATS (upStat.Replies, 1);
  
  /* verify the length of the reply */
  if (bufLen != sizeof(*repP)) {
    fprintf(stderr, "Invalid Updateq_reply: from host=%lu thd=%lu, Len=%d\n",
	    host, thread, bufLen);
    fflush(stderr);
    INCR_STATS (upStat.ReplyErrors, 1);
    goto done;
  }

  /* cast pointer to message buffer */
  repP = (m_updateq_reply_t *) bufP;

  /* Check for success */
  if (repP->status == UPDATEQ_REPLY_STATUS_SUCCESS)
    goto done;

  /* Check for hard errors */
  if (repP->status != UPDATEQ_REPLY_STATUS_NACK) {
    fprintf(stderr, "Error in Updateq_reply: from host=%lu thd=%lu\n",
	    host, thread);
    fflush (stderr);
    INCR_STATS (upStat.ReplyErrors, 1);
    goto done;
  }

  /*
   * Handle NACKs from host
   * NACK indicates the host did not apply any of the updates
   * and should be removed from the copyset from all updated pages.
   *
   * Needs supported.  19Aug92 WAS
   */
#ifdef PRINT_WARNINGS
  fprintf(stderr, "Updateq_reply NACK from host=%lu thread=%lu\n",
	  host, thread);
  fflush (stderr);
#endif

  INCR_STATS (upStat.ReplyNacked, 1);

  /* and return */
done:
  return;
}  /* mid_updateq_reply */




/*
 ************************************************************************
 *
 * updateq_setperiod
 *
 * routine to set period for slave to flush queue
 *
 ************************************************************************
 */
u_long 
mid_updateq_setperiod
  (
   u_long newPeriod
   )
{
  char255 smsg;
  u_long oldPeriod;
#ifdef mach_2_5_way
  msg_header_t wakeupMsg;
#endif
  
  if (IS_DBG_PORT_QUEUE) {
    sprintf(smsg, "UpdateQ_set_period to %d from %d",
	    newPeriod, midway_default_flush_period);
    StatusMsg(smsg);
  }

  /* set the new period */
  mutex_lock(mid_updateq_mutex);
  oldPeriod = midway_default_flush_period;
  midway_default_flush_period = newPeriod;

  /* Check if the server is already running */
  if ((newPeriod > 0) &&
      (updateq_server_running == FALSE)) {
    
    /* spawn a new slave thread */
    start_slave ();
  } 
#ifdef mach_2_5_way
  else if (updateq_server_running != FALSE) {

    /* tell the server to change the period */
    wakeupMsg.msg_local_port = PORT_NULL;
    wakeupMsg.msg_remote_port = updateq_port;
    wakeupMsg.msg_size = sizeof(wakeupMsg);     /* header only */
    wakeupMsg.msg_type = MSG_TYPE_NORMAL;
    wakeupMsg.msg_simple = TRUE;
    wakeupMsg.msg_id = UPDATEQ_WAKEUP_MSGID;

    msg_send(&wakeupMsg, MSG_OPTION_NONE, 0);
  }
#endif

  /* unlock update queue before returning */
  mutex_unlock(mid_updateq_mutex);

  return (oldPeriod);
} /* mid_updateq_set_period */


/*
 ************************************************************************
 * 
 * mid_updateq_add
 *
 * Add an element to the global update queue
 *
 ************************************************************************
 */
void 
mid_updateq_add 
  (
   vm_address_t dataAddr,       /* address of shared storage being written */
   u_long       dataLen         /* length being written */
   )
{
  int index;
  q_elem_t *elemP;
  q_entry_t *entryP;
  vm_address_t prevAddr;
  vm_address_t nextAddr;
  
  char255 smsg;
  
  if (IS_DBG_PORT_QUEUE) {
    sprintf(smsg, "updateq_add (addr=0x%x, len=%d)", 
	    dataAddr, dataLen);
    StatusMsg(smsg);
  }
  INCR_STATS (upStat.QueueAdds, 1);


  /* verify the update */
  ASSERT (dataLen > 0);
  ASSERT (dataAddr > (vm_address_t) vm_page_size);
  ASSERT ((char *) dataAddr <= (char *) vm_shared_end);

  /* verify that the change doesn't cross shared/private boundary */
  ASSERT ((((char *)dataAddr) >= (char *)SHMALLOC_ALLOCATION_BASE) ||
	  ((((char *)dataAddr)+dataLen) < (char *)SHMALLOC_ALLOCATION_BASE));


  /*
   * Check if change was to non-shared static/malloc'ed memory.
   */
  if ((midway_shared_static_data == FALSE) &&
      (((char *)dataAddr) < (char *)vm_shared_base)) {
    INCR_STATS (upStat.QueueUnshared, 1);
    goto done;
  }

  /*
   * Acquire mutex guarding update queue
   */
  mutex_lock(mid_updateq_mutex);
  
  /*
   * Check if we can merge update with the previous entry
   */
  entryP = mid_updateq_prev_entryP;
  ASSERT (entryP != NULL);
  ASSERT (((mid_updateq_last_entryP - entryP) >= 0) &&
	  ((mid_updateq_last_entryP - entryP) <= QENTRIES_PER_ELEM));

  prevAddr = entryP->dataAddr + entryP->dataLen;
  if (prevAddr == dataAddr) {
    entryP->dataLen += dataLen;
    INCR_STATS (upStat.QueueMergedPrev, 1);
    goto merged;
  }
  nextAddr = dataAddr + dataLen;
  if (nextAddr == entryP->dataAddr) {
    entryP->dataAddr = dataAddr;
    entryP->dataLen += dataLen;
    INCR_STATS (upStat.QueueMergedNext, 1);
    goto merged;
  }

  /*
   * Couldn't merge update with previous entry
   * Check for a free entry in this queue element
   */
  if (entryP != mid_updateq_last_entryP) {

    /* use next entry within this element */
    entryP++;

  } else {

    /* Allocate a new element */
    elemP = alloc_updateq_elem();
    ASSERT (elemP != NULL);

    /* Add element to head of update queue */
    enqueue_updateq_elem (elemP);
    entryP = &elemP->entry[0];
  }

  /* Store update in this entry */
  entryP->dataAddr = dataAddr;
  entryP->dataLen  = dataLen;
  mid_updateq_prev_entryP = entryP;

merged:

  /* release mutex and return */
  mutex_unlock(mid_updateq_mutex);

  INCR_STATS (upStat.QueueTotalData, dataLen);
  IF_STATS (dataLen /= STAT_QUEUE_PER_BUCKET);
  IF_STATS (if (dataLen > STAT_MAX_QUEUE) dataLen = STAT_MAX_QUEUE);
  INCR_STATS (upStat.QueueDataEachP[ dataLen ], 1);

 done:
  return;
} /* mid_updateq_add */



/*
 ************************************************************************
 * 
 * mid_updateq_print_stats
 *
 * Debugging routine to print usage statistics
 *
 ************************************************************************
 */
void 
mid_updateq_print_stats ()
{
#ifndef STATS
  StatusMsg("Not compiled to collect Update Queue statistics");
#else

  int i, len, slen;
  u_long calls, calla, callb, callr, callt, callp, callx;
  char255 smsg, tmp;
#define MAX_STR_LEN 80

  calls = upStat.FlushCalls - upStat.FlushEmptyQueue;
  calla = upStat.FlushCallers[UPDATEQ_CALLER_ACQUIRE];
  callr = upStat.FlushCallers[UPDATEQ_CALLER_RELEASE];
  callb = upStat.FlushCallers[UPDATEQ_CALLER_BARRIER];
  callt = upStat.FlushCallers[UPDATEQ_CALLER_TIMER];
  callp = upStat.FlushCallers[UPDATEQ_CALLER_PAGEFAULT];
  callx = upStat.FlushCallers[UPDATEQ_CALLER_LOCKXFER];
  
  StatusMsg("-=-=- Non-EC Update Queue Statistics -=-=-");
  
  sprintf(smsg,
  "Flush: Tot=%lu, Acq=%lu Rel=%lu Bar=%lu Tmr=%lu PF=%lu XF=%lu Empty=%lu Err=%lu",
	  upStat.FlushCalls, calla, callr, callb, callt, callp, callx,
	  upStat.FlushEmptyQueue, upStat.FlushErrors);
  StatusMsg (smsg);

  sprintf(smsg,
  "   Items=%lu, #Copies:", 
	  upStat.FlushTotalItems);
  slen = strlen(smsg);
  for (i = 0; i < MAX_PROCESSORS; i++) {
    if (upStat.FlushItemRecipients[i] == 0) continue;
    
    sprintf(tmp, " %d=%lu", i+1, upStat.FlushItemRecipients[i]);
    len = strlen(tmp);
    if ((len + slen) > MAX_STR_LEN) {
      StatusMsg(smsg);
      sprintf(smsg, "   ItemRecip:");
      slen = strlen(smsg);
    }
    strcat (smsg, tmp);
    slen += len;
  }
  StatusMsg(smsg);
	  
  sprintf(smsg,
  "   Items/Call=%.2f, Each:",
	  (calls == 0) ? 0.0 : 
  	    (float)(((float)upStat.FlushTotalItems)/((float)calls)));
#ifdef PRINT_HISTOGRAMS
  slen = strlen(smsg);
  for (i = 0; i < STAT_ITEMS_TABLE_SIZE+1; i++) {
    if (upStat.FlushItemsEachP[i] == 0) continue;
    
    sprintf(tmp, " %d=%lu", i*STAT_ITEMS_PER_BUCKET, upStat.FlushItemsEachP[i]);
    len = strlen(tmp);
    if ((len + slen) > MAX_STR_LEN) {
      StatusMsg(smsg);
      sprintf(smsg, "   Each:");
      slen = strlen(smsg);
    }
    strcat (smsg, tmp);
    slen += len;
  }
#endif
  StatusMsg(smsg);

  sprintf(smsg,
  "   Items/Caller: Acq=%lu (%.2f) Rel=%lu (%.2f) Bar=%lu (%.2f) Tmr=%lu (%.2f) PF=%lu (%.2f) XF=%lu (%.2f)",
	  upStat.FlushItemsByCaller[UPDATEQ_CALLER_ACQUIRE], 
	  (calla == 0) ? 0.0 :
	  (float)(((float)upStat.FlushItemsByCaller[UPDATEQ_CALLER_ACQUIRE])/
	          ((float)(calla))), 
	  upStat.FlushItemsByCaller[UPDATEQ_CALLER_RELEASE], 
	  (callr == 0) ? 0.0 :
	  (float)(((float)upStat.FlushItemsByCaller[UPDATEQ_CALLER_RELEASE])/
	          ((float)(callr))), 
	  upStat.FlushItemsByCaller[UPDATEQ_CALLER_BARRIER], 
	  (callb == 0) ? 0.0 :
	  (float)(((float)upStat.FlushItemsByCaller[UPDATEQ_CALLER_BARRIER])/
	          ((float)(callb))),  
	  upStat.FlushItemsByCaller[UPDATEQ_CALLER_TIMER], 
	  (callt == 0) ? 0.0 :
	  (float)(((float)upStat.FlushItemsByCaller[UPDATEQ_CALLER_TIMER])/
	          ((float)(callt))),
	  upStat.FlushItemsByCaller[UPDATEQ_CALLER_PAGEFAULT],
	  (callp == 0) ? 0.0 :
	  (float)(((float)upStat.FlushItemsByCaller[UPDATEQ_CALLER_PAGEFAULT])/
		  ((float)(callp))),
	  upStat.FlushItemsByCaller[UPDATEQ_CALLER_LOCKXFER],
	  (callx == 0) ? 0.0 :
	  (float)(((float)upStat.FlushItemsByCaller[UPDATEQ_CALLER_LOCKXFER])/
		  ((float)(callx))));
  StatusMsg(smsg);
	    
  sprintf(smsg,
  "   Requests=%lu, #Recip:", 
	  upStat.FlushRequests);
#ifdef PRINT_HISTOGRAMS
  slen = strlen(smsg);
  for (i = 0; i < MAX_PROCESSORS; i++) {
    if (upStat.FlushRequestRecipients[i] == 0) continue;
    
    sprintf(tmp, " %d=%lu", i+1, upStat.FlushRequestRecipients[i]);
    len = strlen(tmp);
    if ((len + slen) > MAX_STR_LEN) {
      StatusMsg(smsg);
      sprintf(smsg, "   #ReqRecip:");
      slen = strlen(smsg);
    }
    strcat (smsg, tmp);
    slen += len;
  }
#endif
  StatusMsg(smsg);

  sprintf(smsg,
  "   Data=%lu /Call=%.2f, Each:",
	  upStat.FlushTotalData, 
	  (calls == 0) ? 0.0 : 
  	    (float)(((float)upStat.FlushTotalData)/((float)calls)));
#ifdef PRINT_HISTOGRAMS
  slen = strlen(smsg);
  for (i = 0; i < STAT_DATA_TABLE_SIZE+1; i++) {
    if (upStat.FlushDataEachP[i] == 0) continue;
    
    sprintf(tmp, " %d=%lu", i*STAT_DATA_PER_BUCKET,  upStat.FlushDataEachP[i]);
    len = strlen(tmp);
    if ((len + slen) > MAX_STR_LEN) {
      StatusMsg(smsg);
      sprintf(smsg, "   Each:");
      slen = strlen(smsg);
    }
    strcat (smsg, tmp);
    slen += len;
  }
#endif
  StatusMsg(smsg);

  sprintf(smsg,
  "   Data/Caller: AQ=%lu (%.2f) RL=%lu (%.2f) BR=%lu (%.2f) TR=%lu (%.2f) PF=%lu (%.2f) XF=%lu (%.2f)",
	  upStat.FlushDataByCaller[UPDATEQ_CALLER_ACQUIRE], 
	  (calla == 0) ? 0.0 : 
	  (float)(((float)upStat.FlushDataByCaller[UPDATEQ_CALLER_ACQUIRE])/ 
		  ((float)(calla))), 
	  upStat.FlushDataByCaller[UPDATEQ_CALLER_RELEASE], 
	  (callr == 0) ? 0.0 :
	  (float)(((float)upStat.FlushDataByCaller[UPDATEQ_CALLER_RELEASE])/
	          ((float)(callr))), 
	  upStat.FlushDataByCaller[UPDATEQ_CALLER_BARRIER], 
	  (callb == 0) ? 0.0 :
	  (float)(((float)upStat.FlushDataByCaller[UPDATEQ_CALLER_BARRIER])/
	          ((float)(callb))),  
	  upStat.FlushDataByCaller[UPDATEQ_CALLER_TIMER], 
	  (callt == 0) ? 0.0 :
	  (float)(((float)upStat.FlushDataByCaller[UPDATEQ_CALLER_TIMER])/
	          ((float)(callt))),
	  upStat.FlushDataByCaller[UPDATEQ_CALLER_PAGEFAULT], 
	  (callp == 0) ? 0.0 :
	  (float)(((float)upStat.FlushDataByCaller[UPDATEQ_CALLER_PAGEFAULT])/
	          ((float)(callp))),
	  upStat.FlushDataByCaller[UPDATEQ_CALLER_LOCKXFER], 
	  (callx == 0) ? 0.0 :
	  (float)(((float)upStat.FlushDataByCaller[UPDATEQ_CALLER_LOCKXFER])/
	          ((float)(callx))));
  StatusMsg(smsg);

  sprintf(smsg,
  "Requests %lu, Nacked=%lu Errors=%lu, Total Updates=%lu for %lu bytes",
	  upStat.Requests, upStat.RequestNacked, upStat.RequestErrors,
	  upStat.RequestUpdates, upStat.RequestUpdateData);
  StatusMsg(smsg);

  sprintf(smsg,
  "   Updates: Applied=%lu (%lu bytes) Pending=%lu Ignored=%lu",
	  upStat.RequestApplied, upStat.RequestAppliedData, 
	  upStat.RequestPending, upStat.RequestIgnored);
  StatusMsg(smsg);

  sprintf(smsg,
  "Replies: %lu, Nacked=%lu Errors=%lu",
	  upStat.Replies, upStat.ReplyNacked, upStat.ReplyErrors);
  StatusMsg(smsg);

  sprintf(smsg,
  "Pending Queue Calls=%lu Data=%lu",
	  upStat.PendingQueueCalls, upStat.PendingQueueData);
  StatusMsg(smsg);

  sprintf(smsg,
  "Update Queue Calls=%lu Unshared=%lu Merged=%lu/%lu Total Data=%lu",
	  upStat.QueueAdds, upStat.QueueUnshared, 
	  upStat.QueueMergedPrev, upStat.QueueMergedNext,
	  upStat.QueueTotalData);
  StatusMsg(smsg);

  calls = upStat.QueueAdds - upStat.QueueUnshared;
  sprintf(smsg,
  "   Data/Call=%.2f, Each:",
	  (calls == 0) ? 0.0 : 
	  (float)(((float)upStat.QueueTotalData) /
		  ((float)calls)));
#ifdef PRINT_HISTOGRAMS
  slen = strlen(smsg);
  for (i = 0; i < STAT_QUEUE_TABLE_SIZE+1; i++) {
    if (upStat.QueueDataEachP[i] == 0) continue;
    
    sprintf(tmp, " %d=%lu", i*STAT_QUEUE_PER_BUCKET, upStat.QueueDataEachP[i]);
    len = strlen(tmp);
    if ((len + slen) > MAX_STR_LEN) {
      StatusMsg(smsg);
      sprintf(smsg, "   Each:");
      slen = strlen(smsg);
    }
    strcat (smsg, tmp);
    slen += len;
  }
#endif
  StatusMsg(smsg);

#endif	    
}
