/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * updateq.h: Routines to maintain/flush the update queue
 * 
 * 25 Jun 92 (was): Initial code 
 * 23 Sep 92 (was): Added flag for handling of static data.
 * 12 Jan 93 (was): Optimized enqueue of updates
 */
#ifndef __UPDATEQ_H__
#define __UPDATEQ_H__


/* 
 * Generate definitions for both internal & external callers
 */


/*
 * Define types
 */
#define QENTRIES_PER_ELEM 255

/****
***** WARNING: Do NOT change q_entry_t structure without also changing
*****          code.s assembler routines which depend on this layout.
*****          WAS 12Jan93
****/
typedef struct q_entry {
  vm_address_t   dataAddr;	/* starting address of changed region */
  u_long         dataLen;	/* length := end - addr */
} q_entry_t;

typedef struct q_elem {
  q_entry_t      null_entry;	/* MUST preceed entry array! WAS 12Jan92 */
  q_entry_t      entry[ QENTRIES_PER_ELEM ]; /* array of entries */
  struct q_elem *nextP;                      /* ptr to next element */
} q_elem_t;




#ifdef __UPDATEQ_INTERNAL__

/* 
 * Define constants 
 */
#ifdef mach_2_5_way
#define UPDATEQ_WAKEUP_MSGID	0xb00ff00e
#endif


/* 
 * Generate definitions internal to the update queue code.
 */

typedef struct p_elem {
  struct p_elem *nextP;		/* pointer to next element in list */
  vm_address_t   dataAddr;	/* starting address of changed region */
  u_long         dataLen;	/* length of changed region */
  char           data[4];	/* data, actual length = dataLen */
} pending_elem_t;


/*
 * Define static variables
 */
#ifdef mach_2_5_way
static port_t updateq_port;		/* scanner waits on this */
#endif
static int updateq_server_running;
static mutex_t updateq_flushing;        /* serializes flushing of update q */
static q_elem_t *updateq_freeP = NULL;

static mutex_t pending_mutex;           /* protects pending_list ptrs */
static pending_elem_t *pending_tailP = NULL;


/*
 * Define request/reply message formats 
 */
typedef struct {
  int           status;
} m_updateq_reply_t;

#define UPDATEQ_REPLY_STATUS_SUCCESS   0
#define UPDATEQ_REPLY_STATUS_NACK      1    /* remove host from copyset */
#define UPDATEQ_REPLY_STATUS_INVALID   2    /* invalid request parameter */



#undef EXTERN
#define EXTERN			/* routines are defined locally */
#undef INIT
#define INIT(val) = { val }     /* initialize external variables */

#else 


#undef EXTERN
#define pending_elem_t void     /* external view of internal type */
#define EXTERN extern           /* routines are external */
#undef INIT
#define INIT(val)               /* don't initialize external variables */

#endif /* __UPDATEQ_INTERNAL__ */




/*
 * Define external variables
 */

/* ptr to head of pending update queue, Non-NULL if queue has data */
EXTERN pending_elem_t *pending_headP INIT(NULL);

/* flush period for update queue in msecs */
EXTERN u_long midway_default_flush_period INIT(0);

/* default memory consistency model */
EXTERN int midway_default_model INIT(MODEL_EC);

/* flag to indicate if static data is private (=0) or shared (!=0) */
EXTERN int midway_shared_static_data INIT(0);

/*
 * Define update queue
 */
EXTERN mutex_t   mid_updateq_mutex;
EXTERN q_elem_t *updateq_headP;
EXTERN q_elem_t *updateq_tailP; 
EXTERN q_entry_t *mid_updateq_prev_entryP;
EXTERN q_entry_t *mid_updateq_last_entryP;


/*
 * Define caller id flag for calls to mid_updateq_flush
 */
#define UPDATEQ_CALLER_TIMER     0      /* Timer called flush queue */
#define UPDATEQ_CALLER_BARRIER   1      /* Barrier called */
#define UPDATEQ_CALLER_ACQUIRE   2      /* Acquire called */
#define UPDATEQ_CALLER_RELEASE   3      /* Release called */
#define UPDATEQ_CALLER_PAGEFAULT 4      /* PageFault called */
#define UPDATEQ_CALLER_LOCKXFER  5      /* Lock Transfer called */

#define UPDATEQ_CALLERS          6


#define UPDATEQ_FLUSH_AGGRESSIVELY   0   /* =1 to flush on lock acq/rel */
#define UPDATEQ_FLUSH_ON_DEMAND      1   /* =1 to flush on lock xfers */

/* 
 * Define macro to check if the update queue is empty
 */
#define UPDATEQ_NOT_EMPTY \
  ((mid_updateq_prev_entryP != &updateq_tailP->null_entry) || \
   (updateq_headP != updateq_tailP))


/*
 * Define prototypes for calling page manager routines 
 */
EXTERN kern_return_t 
  mid_updateq_init (void);

EXTERN int
  mid_updateq_flush  
  (
   int caller		/* id of calling routine (acq/rel/bar/timer) */
   );

EXTERN void
  mid_updateq_request
  (
   u_long host,		/* host sending update info */
   u_long thread,	/* thread@host sending update */
   u_long tid,		/* transaction id for this request */
   u_long *bufP,        /* ptr to request buffer */
   u_long bufLen 	/* length of data in msg */
   );

EXTERN void
mid_updateq_reply
  (
   u_long host,		/* host sending update info */
   u_long thread,	/* thread@host sending update */
   u_long tid,		/* transaction id for this request */
   u_long *bufP,        /* ptr to request buffer */
   u_long bufLen 	/* length of data in msg */
   );

EXTERN u_long 
  mid_updateq_setperiod
  (
   u_long newPeriod		/* period between flushes in milliseconds */
   );

EXTERN void 
  mid_updateq_add 
  (
   vm_address_t dataAddr,       /* address of shared storage being written */
   u_long       dataLen         /* length being written */
   );

EXTERN void
  mid_updateq_apply_pending
  (
   char          *pageP,        /* ptr to page buffer to update */
   vm_address_t   pageAddr,	/* address of page to update */
   u_long         pageLen 	/* length of page to update */
   );


#endif /* __UPDATEQ_H__ */




