/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */

/*
 * uxpass: pass Unix I/O to IO server
 *
 * 17 Sep 91 (mjz): undefine VERBOSITY -- way too chatty!
 *  3 Oct 92 (mjz): add open, close & lseek. (& uxpass_go_local!)
 *  4 Oct 92 (mjz): system routines, notably gethostbyname() use open/close
 *			locally.  We probably can't catch 'em all.  (Also,
 *			multiple opens to the same file return the same fd,
 *			and the first one that closes wins.  Ugh.
 *			Try to filter out closes for things that we haven't
 *			opened.  Probably sockets.
 * 14 Oct 92 (mjz): update for arbitrary length transfers
 * 20 Oct 92 (was): added midway_task_self for mach_task_self()
 * 17 Nov 92 (mjz): add midway_uxpass_local & midway_uxpass_remote
 *			...so you can go back and forth
 *  9 Jul 93 (mjz): make work again with Chris' fast udp stuff
 * 10 Oct 93 (mjz): fix again... need to have fd's be locally consistent.
 */
#include <stdio.h>
#include <mach.h>
#include <mach_error.h>
#include <mig_errors.h>
#include <mach/message.h>
#include <mach/notify.h>	/* for NOTIFY_PORT_DELETED */
#include <strings.h>
#include <syscall.h>
#include "midway-dbg.h"
#include "midwayux.h"
#include "uxpass.h"
#include <cthreads.h>		/* for mutexes..., sigh */
#include <sys/file.h>		/* O_RDONLY */
#include <sys/socket.h>		/* sockaddr */


extern int errno;
extern unsigned long Comm_index; /* who we are */
extern task_t midway_task_self;	/* mach_task_self() */

/*
 * values for ux_remote
 */
#define UXPASS_LOCAL	0	/* everything local */
#define UXPASS_REMOTE	1	/* everything remote */
#define UXPASS_FLUSH	2	/* everything to bit bucket */

int ux_remote = 0;
mach_port_t ux_server_port = 0;
int uxpass_local_count = 0;

int ux_nfds;
typedef struct u_opentab {
  int isremote;
  int remotefd;
} u_opentab_t;
u_opentab_t *uxpass_opentab;
mutex_t uxpass_lock = 0;

/* #define VERBOSITY 1 */

#if FAST_UDP
#  define WRITE	__write
#  define READ	__read
   /* open not in Chris' library */
#  define OPEN	open
#  define LSEEK	lseek
#  define CLOSE	__close
#else /* not FAST_UDP */
#  define WRITE	write
#  define READ	read
#  define OPEN	open
#  define LSEEK	lseek
#  define CLOSE	close
#endif /*FAST_UDP*/

int
WRITE(fd, buf, len)
    int fd;
    char *buf;
    int len;
{
    int uxerr;
    int cc;
    kern_return_t krc;
    register int uxr = ux_remote;

    if ((uxr == UXPASS_REMOTE) && (uxpass_opentab[fd].isremote == 0)) {
      /* not ours. */
      uxr = UXPASS_LOCAL;
    }

    switch(ux_remote) {
    case UXPASS_LOCAL:
      cc = syscall(SYS_write,  fd,  buf, len);
      break;


    case UXPASS_REMOTE:
#ifdef VERBOSITY
			  /*01234567890123456789012345678901234567890*/
      syscall(SYS_write, 2, "About to call midwayux_write...",32);
      syscall(SYS_write, fd, buf, len);
#endif /* VERBOSITY */
    	krc = midwayux_write(ux_server_port, Comm_index, 
			     uxpass_opentab[fd].remotefd, buf, len, len, 
			     &uxerr, &cc);
        if (krc != KERN_SUCCESS) {
#ifndef UXPASS_QUIET_ON_MACH_ERROR
	  int len;
	  char message[256];
	  sprintf(message,"midwayux_write: %s\n", mach_error_string(krc));
	  len = strlen(message);
	  syscall(SYS_write,2,message,len);
#endif /*UXPASS_(not)QUIET_ON_MACH_ERROR*/
	  cc = -1;		/* probably should be -1 */
	  errno = krc;		/* will errno get stomped? */
	} else {		/* rpc succeeded */
	  if (cc < 0) {
	    errno = uxerr;
	  }
	}
#ifdef UXPASS_DEBUG
        {
	  char smsg[255];
	  sprintf(smsg,"write(fd %d l %d):%d (uxerr %d)(krc %d)",
		  fd, len, cc, uxerr, krc);
	  StatusMsg(smsg);
	}
#endif
      break;

    case UXPASS_FLUSH:
      /* do nothing */
      cc = 0;
      errno = 0;
      break;

    default:
      {
	char smsg[255];
	sprintf(smsg,"write(fd %d l %d): bad ux_remote value %d",
		fd, len, ux_remote);
      }
      break;
    } /* switch */

    return cc;
} /* write */


int
READ(fd, buf, len)
    int fd;
    char *buf;
    int len;
{
    int uxerr;
    int cc;
    kern_return_t krc;

    if ((!ux_remote) || (!uxpass_opentab[fd].isremote)) {
	cc = syscall(SYS_read,  fd,  buf, len);
    } else {
      char *p = buf; /* save */
#ifdef VERBOSITY
			  /*01234567890123456789012345678901234567890*/
      syscall(SYS_write, 2, "About to call midwayux_read...\n  ",33);
#endif /* VERBOSITY */
    	krc = midwayux_read(ux_server_port, Comm_index, 
			    uxpass_opentab[fd].remotefd, &p, &len, len,
			    &uxerr, &cc);
        if (krc != KERN_SUCCESS) {
	  cc = -1;		/* probably should be -1 */
	  errno = krc;		/* will errno get stomped? */
	} else {		/* rpc succeeded */
	  if (cc < 0) {
	    errno = uxerr;
	  }
	  if (p != buf) {	/* data arrived out-of-line */
	    if (cc > 0) {	/* to preserve semantics, copy data */
	      bcopy(p,buf,cc);
	    }
	    VM_DEALLOCATE(midway_task_self, p, len);
	  } /* p != buf */
	}
#ifdef UXPASS_DEBUG
        {
	  char smsg[255];
	  sprintf(smsg,"read(fd %d):%d (uxerr %d)",fd, cc, uxerr);
	  StatusMsg(smsg);
	}
#endif /*UXPASS_DEBUG*/
    }
    return cc;
} /* read */

int
OPEN(path, flags, mode)
    char *path;
    int flags;
    int mode;
{
    int fd;
    int uxerr;
    int len;
    kern_return_t krc;
    int localfd;

    if (!ux_remote) {
	fd = syscall(SYS_open, path, flags, mode);
    } else {
        len = strlen(path) + 1;	/* +1 to pass \0 at end!!! */
    	krc = midwayux_open(ux_server_port, Comm_index, path, len, flags, mode,
		       &uxerr, &fd);
        if (krc != KERN_SUCCESS) {
	  fd = -1;		/* probably should be -1 */
	  errno = krc;		/* will errno get stomped? */
	} else {		/* rpc succeeded */
	  if (fd < 0) {
	    errno = uxerr;
	  } else {
	    /* find local fd to give back */
	    localfd = syscall(SYS_open, "/dev/null", O_RDONLY, 0);
	    if (localfd < 0) {
	      char smsg[255];
	      int cc;
	      sprintf(smsg,"open localfd for %s, uxerr %d\n", path, errno);
	      StatusMsg(smsg);
	      krc = midwayux_close(ux_server_port, Comm_index, fd,
				   &uxerr, &cc);
	      /* throw away close info */
	    } else {
	      mutex_lock(uxpass_lock);
	      uxpass_opentab[localfd].isremote = 1;
	      uxpass_opentab[localfd].remotefd = fd;
	      mutex_unlock(uxpass_lock);
	    }
	    fd = localfd;
	  }
	}
#ifdef UXPASS_DEBUG
        {
	  char smsg[255];
	  sprintf(smsg,"open(%s,%x 0%o):%d (uxerr %d)",path, flags, mode,
		  		fd, uxerr);
	  StatusMsg(smsg);
	}
#endif /*UXPASS_DEBUG*/
    }
    return fd;
} /* open */

int
LSEEK(fd, offset, whence)
    int fd;
    int offset;
    int whence;
{
    int uxerr;
    int pos;
    int ours;
    kern_return_t krc;

    if (ux_remote) {
      mutex_lock(uxpass_lock); /* why under lock? all serialized anyway */
      ours = uxpass_opentab[fd].isremote;
      mutex_unlock(uxpass_lock);
    }
    if ((!ux_remote) || (!ours)) {
	pos = syscall(SYS_lseek,  fd,  offset, whence);
    } else {
    	krc = midwayux_lseek(ux_server_port, Comm_index, 
			     uxpass_opentab[fd].remotefd, offset, whence,
			     &uxerr, &pos);
        if (krc != KERN_SUCCESS) {
	  pos = -1;		/* probably should be -1 */
	  errno = krc;		/* will errno get stomped? */
	} else {		/* rpc succeeded */
	  if (pos == -1) {
	    errno = uxerr;
	  }	
	}
#ifdef UXPASS_DEBUG
        {
	  char smsg[255];
	  sprintf(smsg,"lseek(fd %d %d):%d (uxerr %d)",fd, offset, whence,
		  		pos, uxerr);
	  StatusMsg(smsg);
	}
#endif /*UXPASS_DEBUG*/
    }
    return pos;
} /* lseek */

int
CLOSE(fd)
    int fd;
{
    int uxerr;
    int cc;
    int ours;
    kern_return_t krc;
    int rfd;

    if (ux_remote) {
      mutex_lock(uxpass_lock);
      ours = uxpass_opentab[fd].isremote;
      rfd = uxpass_opentab[fd].remotefd;
      /* going to close anyway */
      /*.. 14 oct 93(mjz): for now don't turn off 0/1/2 */
      if (fd >= 3) {
	uxpass_opentab[fd].isremote = 0;
	uxpass_opentab[fd].remotefd = 0;
      }
      mutex_unlock(uxpass_lock);
    }
    
    if ((!ux_remote) || (fd < 3) || (!ours)) {
	cc = syscall(SYS_close,  fd);
    } else {
    	krc = midwayux_close(ux_server_port, Comm_index, rfd,
		       &uxerr, &cc);
        if (krc != KERN_SUCCESS) {
	  cc = -1;		/* probably should be -1 */
	  errno = krc;		/* will errno get stomped? */
	} else {		/* rpc succeeded */
	  if (cc < 0) {
	    errno = uxerr;
	  }	
	}
#ifdef UXPASS_DEBUG
        {
	  char smsg[255];
	  sprintf(smsg,"close(fd %d):%d (uxerr %d)",fd, cc, uxerr);
	  StatusMsg(smsg);
	}
#endif /*UXPASS_DEBUG*/
    }
    return cc;
} /* close */

#if FAST_UDP
#include <sys/uio.h> /* for iovec */
/*
 * These are functions that are defined in Chris' library, but we
 * do not usurp.  Just do normal system calls.
 * from ux/sockets/sock_kern/unix_syscalls.c
 */
/* an osockaddr is defined in ux/sockets/sock_sys/socket.h, it is a 4.3 style
 * sockaddr (old for NET2, I guess).
 */
#define osockaddr sockaddr
/* 
 * Mach Operating System
 * Copyright (c) 1993 Carnegie Mellon University
 * All rights reserved.  The CMU software License Agreement specifies
 * the terms and conditions for use and redistribution.
 */
/*
 * Author:	Chris Maeda
 * Date:	Summer 1993
 *
 * Fallback functions for each library entry point.
 * User code can replace these at link time.
 */



int
__socket(int domain, int type, int protocol)
{
	return syscall(SYS_socket, domain, type, protocol);
}


int
__bind(int s, struct osockaddr *name, int namelen)
{
	return syscall(SYS_bind, s, name, namelen);
}


int
__getsockname(int s, struct osockaddr *name, int *namelen)
{
	return syscall(SYS_getsockname, s, name, namelen);
}	

int
__connect(int s, struct osockaddr *name, int namelen)
{
	return syscall(SYS_connect, s, name, namelen);
}

int
__send(int s, caddr_t buf, int len, int flags)
{
	return syscall(SYS_send, s, buf, len, flags);
}

int
__sendto(int s, caddr_t buf, int len, int flags, caddr_t to, int tolen)
{
	return syscall(SYS_sendto, s, buf, len, flags, to, tolen);
}

int
__sendmsg(int s, caddr_t msgp, int flags)
{
	return syscall(SYS_sendmsg, s, msgp, flags);
}

int
__recvfrom(int s, caddr_t buf, int len, int flags, caddr_t from, int *fromlen)
{
	return syscall(SYS_recvfrom,s, buf, len, flags, from, fromlen);
}

int
__recv(int s, caddr_t buf, int len, int flags)
{
	return syscall(SYS_recv, s, buf, len, flags);
}

int
__recvmsg(int s, caddr_t msgp, int flags)
{
	return syscall(SYS_recvmsg, s, msgp, flags);
}

int
__getpeername(int fdes, caddr_t asa, int *alen)
{
	return syscall(SYS_getpeername, fdes, asa, alen);
}

int
__accept(int s, caddr_t name, int *namelen)
{
	return syscall(SYS_accept, s, name, namelen);
}

int
__listen(int s, int backlog)
{
	return syscall(SYS_listen, s, backlog);
}

int
__setsockopt(int s, int level, int name, caddr_t val, int valsize)
{
	return syscall(SYS_setsockopt, s, level, name, val, valsize);
}

int
__getsockopt(int s, int level, int name, caddr_t val, int *valsize)
{
	return syscall(SYS_setsockopt, s, level, name, val, valsize);
}

int
__shutdown(int s, int how)
{
	return syscall(SYS_shutdown, s, how);
}

#if DEFINED_ABOVE
int
__read(int fdes, char *cbuf, unsigned count)
{
	return syscall(SYS_read, fdes, cbuf, count);
}
#endif

int
__readv(int fdes, const struct iovec *iov, int iovcnt)
{
	return syscall(SYS_readv, fdes, iov, iovcnt);
}

#if DEFINED_ABOVE
int
__write(int fdes, char *cbuf, unsigned count)
{
	return syscall(SYS_write, fdes, cbuf, count);
}
#endif

int
__writev(int fdes, const struct iovec *iov, int iovcnt)
{
	return syscall(SYS_writev, fdes, iov, iovcnt);
}

int
__ioctl(int fdes, int cmd, caddr_t cmdarg)
{
	return syscall(SYS_ioctl, fdes, cmd, cmdarg);
}
#if DEFINED_ABOVE
int
__close(int s)
{
	return syscall(SYS_close, s);
}
#endif
#endif /*FAST_UDP*/

/*
 *---------------------------------------------------------------
 */

#ifdef STANDALONE
extern char **gargv;
unsigned long Comm_index = 1;

void
uxpass_go_remote()
{
    kern_return_t k;

    ux_nfds = getdtablesize();	/* get number of fd's allowed in kernel */
    uxpass_opentab = (u_opentab_t *) CALLOC(ux_nfds, sizeof(u_opentab_t));
    /* init stdin/stdout/stderr, allow to go remote */
    uxpass_opentab[0].isremote = 1; /*extra to not have stdin test on every r/w*/
    uxpass_opentab[1].isremote = 1; uxpass_opentab[1].remotefd = 1;
    uxpass_opentab[2].isremote = 1; uxpass_opentab[2].remotefd = 2;
    uxpass_lock = mutex_alloc();
    MIDWAY_MUTEX_SET_NAME(uxpass_lock,"uxpass_lock");
    
    k = netname_look_up(name_server_port, gargv[1],
			"MidwayUxServer", &ux_server_port);
    if (k != KERN_SUCCESS)  {
	fprintf(stderr,"uxpass_go_remote[standalone]: Could not find server port\n");
    } else {   
      ux_remote = 1;
    }
}
#else /* not STANDALONE */

void
uxpass_go_remote(ux_port)
    mach_port_t ux_port;
{
  if (ux_server_port) {
    fprintf(stderr,"uxpass_go_remote called, but ux_server_port already set.  Ignoring.\n");
    return;
  }
  
  ux_nfds = getdtablesize();	/* get number of fd's allowed in kernel */
  uxpass_opentab = (u_opentab_t *) CALLOC(ux_nfds, sizeof(u_opentab_t));
  /* init stdin/stdout/stderr, allow to go remote */
  uxpass_opentab[0].isremote = 1; /*extra to not have stdin test on every r/w*/
  uxpass_opentab[1].isremote = 1; uxpass_opentab[1].remotefd = 1;
  uxpass_opentab[2].isremote = 1; uxpass_opentab[2].remotefd = 2;
  uxpass_lock = mutex_alloc();

  ux_server_port = ux_port;	/* set up ux port */
  ux_remote = 1;		/* and redirect the world */

  return;
} /* uxpass_go_remote */
#endif /* STANDALONE */

/* DEPRECATED */
void
uxpass_go_local()
{
  ux_remote = 0;		/* gee, that was easy */
  /* hmmm. uxpass_opentab could get hosed.  Better close ONLY what you open. */

  return;
} /* uxpass_go_local */

/*
 * New, improved, better interface
 */
int
midway_uxpass_local()
{
  register int old;
  int setup = (int)uxpass_lock;

  if (setup) {
    mutex_lock(uxpass_lock);
    uxpass_local_count++;	/* won't go remote until this is 0 */
    mutex_unlock(uxpass_lock);
  }
/*  old = ux_remote; */
  ux_remote = 0;		/* ensure we are local */

  return(1);			/* always appear we were remote */
} /* midway_uxpass_local */

int
midway_uxpass_remote()
{
  register int old;
  int setup = (int)uxpass_lock;
  int count;

  if (setup) {
    mutex_lock(uxpass_lock);
  }
  uxpass_local_count--;
  if (uxpass_local_count < 0)
    uxpass_local_count = 0;

  if (uxpass_local_count) {
    old = -2;
  } else {
    if (ux_server_port) {	/* only if remote server has been set */
      old = ux_remote;
      ux_remote = 1;		/* ensure remote */
    } else {
      /* no port */
      old = -1;			/* better be the case that ux_remote == 0 */
    }
  }

  if (setup) {
    mutex_unlock(uxpass_lock);
  }

  return(old);
} /* midway_uxpass_remote */

int
midway_uxpass_vaporize()
{
  register int old;

  old = ux_remote;
  ux_remote = UXPASS_FLUSH;	/* ensure we throw (at least writes) away */

  return(old);
} /* midway_uxpass_vaporize */


#ifdef PROFILE

void bcopy(char *src, char *dst, int length)
{
  (void) memcpy(dst, src, length);
}
#endif    
