/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
#include <mach.h>
#include <cthreads.h>
#include <stdio.h>
#include "shmalloc.h"
#include "backing.h"
#include "xpager.h"
#include "xpager_obj.h"
#include "XPager.h"		/* mig generated */
#include "midway_types.h"	/* MODEL_EC */
#include "midway-dbg.h"
#include "vm.h"

extern int midway_default_model;
extern task_t midway_task_self;	/* mach_task_self() */

/*
 * Shared memory starts at a hardcoded address.
 * Currently, it has a maximum size, but this should be changed to 
 * allow it to grow to consume the rest of the virtual address space.
 * WAS 16Dec93
 * It should also be changed to allow the application to set it
 * at via a runtime flag.  WAS 8Jan94
 */
vm_offset_t vm_shared_base = SHMALLOC_ALLOCATION_BASE;
vm_offset_t vm_shared_end = 0xffffffff;
static mach_port_t shared_memory_pager_port = MACH_PORT_NULL;

/*
 * The dirtybits currently start at a fixed offset from shared memory.
 * This imposes a max size on shared memory.
 *
 * The dirtybits SHOULD start at MAX_ALLOCATION and grow downward towards
 * the shared data. That would allow us to allocate additional shared
 * memory as needed by the application.   WAS 16Dec93
 */
vm_offset_t vm_dirty_base = SHMALLOC_ALLOCATION_BASE + MAX_SHARED_SPACE;
vm_offset_t vm_dirty_end = 0xffffffff;
vm_offset_t vm_dirty_buffer;
clock_t     vm_dirty_value;
vm_offset_t vm_dirty_free_spaceP;
vm_offset_t vm_dirty_free;
clock_t    *vm_dense_dirty_base;

/*
 * Private memory is allocated in the gap between the end of the application's 
 * data segment and the start of shared memory. It is allocated in REGION_SIZE chunks 
 * with a dirtybit template in the first page of each chunk. The first region allocated
 * is just below the start of shared memory and it grows down towards the application's
 * data segment. 
 */
vm_offset_t vm_private_base = 0;
vm_offset_t vm_private_end = 0;
vm_offset_t vm_private_free_spaceP;
vm_offset_t vm_private_free;
vm_offset_t vm_private_stackP = 0; 		/* ptr to base of application's stack */

/* 
 * Maintain a free list for private buffers which are exactly 1 page in size.
 */
typedef struct priv_free_list_struct {
    struct priv_free_list_struct *nextP;
} priv_free_list_t;

static priv_free_list_t *vm_private_free_listP = NULL;
static int               vm_private_free_list_entries = 0;


/*
 * The lock table starts after the application's data segment 
 * and it grows up towards the start of shared memory. If the 
 * lock table collides with private memory, then the library
 * needs to be rebuilt with a higher SHMALLOC_ALLOCATION_BASE.
 * (Actually, this should be setable via a flag at runtime. WAS 8Jan94)
 */
vm_offset_t vm_locktable_base = 0;
vm_offset_t vm_locktable_end = 0;
vm_offset_t vm_locktable_free;



static mach_port_t dirty_memory_pager_port = MACH_PORT_NULL;

vm_size_t   vm_page_shift = 0;	/* "X / 4096" ==> "X >> 14" */



/*
 * xpager_request functions
 */


static kern_return_t
shared_memory_data_request(arg, offset, len, access_prot, buf, precious)
        char	   *arg;
	vm_offset_t offset;
	vm_size_t   len;
	vm_prot_t   *access_prot;
	pointer_t *buf;
	boolean_t   *precious;
{
    kern_return_t krc;

    mumble("$ shared_memory_data_request(0x%x 0x%x %lu %d)",
	   arg, offset, len, access_prot);
    
    offset += (vm_offset_t)arg;
    if (!IN_SHARED(offset)) {
	fret("shared memory data request for non-shared space");
	return KERN_FAILURE;
    }
    
    krc = get_shmalloc_info(offset);
    if (krc == KERN_SUCCESS) {
        if (midway_default_model == MODEL_EC) {
	  *buf = 0;		/* supply a zero page */
	  INCR_STAT(xpager_stat.shared_zeroed, 1);
	} else {
	  /* supply the right page */
	  krc = mid_page_fault (offset, buf);
	  if (krc != KERN_SUCCESS)
	    fret("shared memory data request could not get page data");
	  INCR_STAT(xpager_stat.shared_filled, 1);
	}
    } else {
	fret("shared memory data request could not get shmalloc info for address 0x%x + %u",
	     offset, len);
	fret("Your program generated an invalid address 0x%x into the shared region.", offset);
#ifdef HALT_ON_ERROR
	fret("SPINNING!");
	while(krc != KERN_SUCCESS) {
	  cthread_yield();
	}
#endif
    }
    
    return krc;
}


static kern_return_t
dirty_memory_data_request(arg, offset, len, access_prot, buf, precious)
        char	   *arg;
	vm_offset_t offset;
	vm_size_t   len;
	vm_prot_t   *access_prot;
	pointer_t *buf;
	boolean_t   *precious;
{
    mumble("$ ~dirty_memory_data_request(0x%x 0x%x %lu %d)",
	   arg, offset, len, access_prot);
    
    offset += (vm_offset_t)arg;
    if (!IN_DIRTY(offset)) {
	fret("dirty memory data request 0x%x for non-dirty space",
	     offset);
	return KERN_FAILURE;
    }

    if (offset >= (((vm_offset_t)(vm_dirty_free_spaceP)) + IN_DIRTY_EXTRA_SPACE)) {
      fret("Your program generated an invalid address 0x%x into the dirty bit region.\n(max dirtybit address 0x%x)", 
	   offset, vm_dirty_free_spaceP);
#ifdef HALT_ON_ERROR
      fret("SPINNING!");
      while(1) {
	cthread_yield();
      }
#endif
      return KERN_FAILURE;
    }
    
    /* supply pre-initialized page */
    *buf = vm_dirty_buffer;
    INCR_STAT(xpager_stat.dirty_filled, 1);
    
    return KERN_SUCCESS;
}


static kern_return_t
shared_memory_databuf_free(arg,buf)
    char *arg;
    pointer_t buf;
{
    kern_return_t krc = KERN_SUCCESS;

    if (buf)  {
	krc = VM_DEALLOCATE(midway_task_self, buf, vm_page_size); /* XXX */
	if (krc)
	  if (midway_task_self != mach_task_self())
	    botch ("midway_task_self != mach_task_self");
    }
    return(krc);
}


static kern_return_t
shared_memory_data_write(arg, offset, len, buf,discarded)
        char  *arg;
	vm_offset_t offset;
	vm_size_t   len;
	pointer_t *buf;
	boolean_t   *discarded;
{
    offset += (vm_offset_t)arg;    
    if (!IN_SHARED(offset) && !IN_DIRTY(offset)) {
	fret("shared memory data write for non-shared space");
	return KERN_FAILURE;
    }    
    *discarded = FALSE;
    return KERN_SUCCESS;
}


static kern_return_t
shared_memory_data_access_request(arg, offset, len)
        char *arg;
	vm_offset_t offset;
	vm_size_t   len;
{
    offset += (vm_offset_t)arg;
    if (!IN_SHARED(offset) && !IN_DIRTY(offset)) {
	fret("shared memory data write for non-shared space");
	return KERN_FAILURE;
    }        
    return KERN_FAILURE;
}


struct xpager_request xpager_rq = {
    0, 
    shared_memory_data_request,
    shared_memory_databuf_free,
    shared_memory_data_write,
    shared_memory_data_access_request
};


kern_return_t
vm_shared_init(size)
    vm_size_t size;
{
    kern_return_t krc;
    vm_offset_t addr;
    xpager_request_t rq;


#ifdef PAGE_BASED
    /* initialize page based code only */
    krc = vm_paged_init(size);
    return krc;
#endif    

    if ((vm_offset_t)LAST_STATIC_SYM > vm_shared_base)  {
	botch("vm_shared_init: statics overflowing into shared region. Call a guru.");
	exit(-1);
    }
    
    xpager_init(0);
    
    krc = do_xpager_create(MACH_PORT_NULL, &shared_memory_pager_port);
    if (krc != KERN_SUCCESS)  {
	fret("vm_shared_init failure: do_xpager_create:%s\n",
	     mach_error_string(krc));
	return krc;
    }

    rq = &xpager_rq;
    xpager_rq.data_any = (char*)vm_shared_base;
    krc = set_xpager_request(shared_memory_pager_port, &rq);
    if (krc != KERN_SUCCESS)  {
	fret("vm_shared_init failure: set_xpager_request:%s\n",
		mach_error_string(krc));
	return krc;
    }

    /*
     * ignore control channel
     */
  
    
    /*
     * Now map the whole range in.
     */
    addr = vm_shared_base;
    krc = vm_map(mach_task_self(),
	  	 &addr,
		 size,
		 0,
		 FALSE,
		 shared_memory_pager_port,
		 0,
		 FALSE,
		 VM_PROT_DEFAULT,
		 VM_PROT_ALL,
		 VM_INHERIT_COPY);
    if (addr != vm_shared_base)  {
	botch("vm_shared_init");
    }
    if (krc != KERN_SUCCESS)  {
	fret("vm_map failure in vm_shared_init: %s\n",
		mach_error_string(krc));
	return krc;
    }
    vm_shared_end = vm_shared_base + size - 1;

    /* Convert page size to a shift amount to eliminate divides */
    if (vm_page_shift == 0)
      vm_page_shift = midway_int_to_log_int(vm_page_size);

    fret("vm_shared_init: base = 0x%x end = 0x%x", /*  */
	    vm_shared_base, vm_shared_end);
    return krc;
}


void 
vm_shared_term(void)
{
  kern_return_t krc;

  if (shared_memory_pager_port != MACH_PORT_NULL) {
    krc = do_xpager_destroy(shared_memory_pager_port);
    if (krc != KERN_SUCCESS)  {
      fret("vm_shared_term failure: do_xpager_destroy:%s\n",
	   mach_error_string(krc));
    }
    shared_memory_pager_port = MACH_PORT_NULL;
  }
}


struct xpager_request xpager_dirty_rq = {
    0, 
    dirty_memory_data_request,
    0,				/* shared_memory_databuf_free */
    shared_memory_data_write,
    shared_memory_data_access_request
};


kern_return_t
vm_dirty_init(size)
    vm_size_t size;
{
    kern_return_t krc;
    vm_offset_t addr;
    xpager_request_t rq;
    clock_t clock, *cP, *endP;

#ifdef PAGE_BASED
    vm_dirty_end = vm_dirty_base;
    vm_dirty_free_spaceP = vm_dirty_base;
    vm_dirty_free = 0;
    vm_dense_dirty_base = (clock_t *) vm_dirty_base;
    return KERN_SUCCESS;
#endif    


    ASSERT (vm_shared_end < vm_dirty_base);


    krc = do_xpager_create(MACH_PORT_NULL, &dirty_memory_pager_port);
    if (krc != KERN_SUCCESS)  {
	fret("vm_dirty_init failure: do_xpager_create:%s\n",
	     mach_error_string(krc));
	return krc;
    }

    rq = &xpager_dirty_rq;
    xpager_dirty_rq.data_any = (char*)vm_dirty_base;
    krc = set_xpager_request(dirty_memory_pager_port, &rq);
    if (krc != KERN_SUCCESS)  {
	fret("vm_dirty_init failure: set_xpager_request:%s\n",
		mach_error_string(krc));
	return krc;
    }

    /*
     * ignore control channel
     */
  
    
    /*
     * Now map the whole range in.
     */
    addr = vm_dirty_base;
    krc = vm_map(mach_task_self(),
	  	 &addr,
		 size,
		 0,
		 FALSE,
		 dirty_memory_pager_port,
		 0,
		 FALSE,
		 VM_PROT_DEFAULT,
		 VM_PROT_ALL,
		 VM_INHERIT_COPY);
    if (addr != vm_dirty_base)  {
	botch("vm_dirty_init");
    }
    if (krc != KERN_SUCCESS)  {
	fret("vm_map failure in vm_dirty_init: %s\n",
		mach_error_string(krc));
	return krc;
    }
    vm_dirty_end = vm_dirty_base + size - 1;
    vm_dirty_free_spaceP = vm_dirty_base;
    vm_dirty_free = size;
    vm_dense_dirty_base = (clock_t *) vm_dirty_base;

    /* preallocate a page to be used to initialize new pages */
    krc = VM_ALLOCATE (mach_task_self(), &vm_dirty_buffer, vm_page_size, 1);
    if (krc != KERN_SUCCESS) {
	fret("vm_dirty_init failure: vm_allocate:%s\n",
	     mach_error_string(krc));
	return krc;
    }

    /* initialize page based on the default memory model */
    vm_dirty_page_init();


    fret("vm_dirty_init: base = 0x%x end = 0x%x",
	    vm_dirty_base, vm_dirty_end);

    return krc;
}

void
vm_dirty_term(void)
{
  kern_return_t krc;

  if (dirty_memory_pager_port != MACH_PORT_NULL) {
    krc = do_xpager_destroy(dirty_memory_pager_port);
    if (krc != KERN_SUCCESS)  {
      fret("vm_dirty_term failure: do_xpager_destroy:%s\n",
	   mach_error_string(krc));
    }
    dirty_memory_pager_port = MACH_PORT_NULL;
  }
}


vm_dirty_page_init()
{
    register clock_t clock, *cP, *endP;

    /* initialize dirty page data based on the default memory model */
    if (midway_default_model == MODEL_EC)
      clock = midway_get_local_clock();
    else
      clock = (clock_t)(LINE_EXCEPTION_FLAG | LINE_ENQUEUE_FLAG);
    vm_dirty_value = clock;

    endP = (clock_t *)(vm_dirty_buffer + vm_page_size);
    for (cP = (clock_t *) vm_dirty_buffer; cP < endP; cP += 16) {
      cP[0] = clock; cP[1] = clock; cP[2] = clock; cP[3] = clock;
      cP[4] = clock; cP[5] = clock; cP[6] = clock; cP[7] = clock;
      cP[8] = clock; cP[9] = clock; cP[10] = clock; cP[11] = clock;
      cP[12] = clock; cP[13] = clock; cP[14] = clock; cP[15] = clock;
    }
    ASSERT (cP == endP);
}

/*
 ************************************************************************
 *		v m _ p r i v a t e _ a l l o c a t e 			*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
kern_return_t 
vm_private_allocate (void)
{
  kern_return_t krc;
  vm_offset_t addr;
  region_info_t private_ri;

  addr = vm_private_base - REGION_SIZE;

  mumble ("allocate private region 0x%x, base 0x%x end 0x%x freeP 0x%x free %d", 
	  addr, vm_private_base, vm_private_end, 
	  vm_private_free_spaceP, vm_private_free);

  if (addr <= vm_locktable_end) {
    krc = KERN_NO_SPACE;
    fprintf(stderr, 
	    "vm_allocate: Private memory collided with locktable at 0x%x, rc=%d\n", 
	    addr, krc);
    return krc;
  }
  
  krc = VM_ALLOCATE (mach_task_self(),
		     &addr, REGION_SIZE, 0); /* anywhere = FALSE */
  if (krc != KERN_SUCCESS) {
    
    /*
     * NOTE: We should try to find additional space.
     * There is a gap between the end of the application's text and the
     * start of its static data. We could also get space out of the shared
     * region, if necessary.  WAS 16Dec93
     */

    fprintf(stderr, 
	    "Unable to vm_allocate space for private data at 0x%x, rc=%d\n", 
	    addr, krc);
    return krc;
  }
  ASSERT (addr == (vm_private_base-REGION_SIZE));

  vm_private_base = addr;
  vm_private_free_spaceP = addr + MIDWAY_TEMPLATE_SIZE;
  vm_private_free = REGION_SIZE - MIDWAY_TEMPLATE_SIZE;

  /* initialize dirtybit template at base of region */
  bzero (&private_ri, sizeof(private_ri));
  private_ri.type = REGION_TYPE_PRIVATE;
  private_ri.base = addr;
  ipatch_template (&private_ri);

  return KERN_SUCCESS;
}


/*
 * One time initialization for private memory
 */
kern_return_t
vm_private_init(void)
{
    kern_return_t krc;
    vm_offset_t addr;

    /*
     * Find the start of the locktable
     * It starts at the first page boundary past the end of the application's
     * data segment.
     */
    addr = ((unsigned)(LAST_STATIC_SYM + (vm_page_size-1)) >> vm_page_shift) << vm_page_shift;
    addr += vm_page_size;	/* leave an unmapped page for protection from application errors */
    vm_locktable_base = addr;
    vm_locktable_end = addr - 1;
    vm_locktable_free = addr;

    /*
     * Compute starting address of private memory
     * It starts at the first region beneath shared memory
     * and grows downwards towards the application's data segment
     * by allocating region size chunks.
     */
    addr = ((unsigned)(vm_shared_base - REGION_SIZE) >> REGION_SHIFT) << REGION_SHIFT;
    vm_private_base = addr + REGION_SIZE;
    vm_private_end = vm_private_base - 1;

    /*
     * allocate region for application's private data (ie stack/heap)
     * This will subtract REGION_SIZE from vm_private_base.
     */
    krc = vm_private_allocate();
    ASSERT (krc == KERN_SUCCESS);
    ASSERT (vm_private_free >= 0);

    fret("vm_private_init: base = 0x%x end = 0x%x",
	    vm_private_base, vm_private_end);

    return krc;
}

/*
 * vm_allocate:
 *	Allocates virtual memory space for this process.
 * 
 *      NOTE: This routine assumes that if "anywhere" is FALSE
 *            that the caller wants unshared memory and it is allocated
 *            from a private region. If "anywhere" is TRUE, the caller
 *            must initialize the dirtybit template as necessary.
 *            WAS 23Nov93
 */

kern_return_t
vm_allocate(task, addr, size, anywhere)
	task_t		task;
	vm_offset_t	*addr;
	vm_size_t	size;
	boolean_t	anywhere;
{
    kern_return_t krc;

#ifdef PROFILE
    Routine vm_allocate cannot be compiled -pg!
    It results in a cyclic dependency between mcount & vm_allocate.
    WAS 6 Nov 92
#endif


    mumble("vm_allocate(task=0x%x,addr=0x%x,size=%ld,anywhere=%d)",
	   task, *addr, size, anywhere);
    INCR_STAT(xpager_stat.alloc_calls, 1);
    INCR_STAT(xpager_stat.alloc_space, size);
#ifdef STATS
      { register unsigned s = size >> (STAT_TABLE_MIN_SHIFT + 1);
	register unsigned i = 0;
	while (s) { s >>= 1; i++; }
	ASSERT ((i >= 0) && (i < STAT_TABLE_SIZE));
	INCR_STAT (xpager_stat.alloc_size_count[i], 1);
      }
#endif
    
    /*
     * If the caller didn't specify a starting address
     * Then assume the caller wants private memory.
     */
    if (anywhere != FALSE) { 

      /* Initialize private memory on the first call. */
      if (vm_private_base == 0) {

	/* Convert page size to a shift amount to eliminate divides */
	if (vm_page_shift == 0)
	  vm_page_shift = midway_int_to_log_int(vm_page_size);

	krc = vm_private_init();
	if (krc != KERN_SUCCESS)
	  return krc;
      }
	
      /* round size up to a page multiple */
      size = ((size + (vm_page_size-1)) >> vm_page_shift) << vm_page_shift;
      
      /* Allocate space in a private region */
      if (size > vm_private_free) {
	
	/* allocate another region of private space.
	 * Note the space will not be contiguous due to 
	 * the dbit template in the new region.
	 */
	krc = vm_private_allocate();
	if (krc != KERN_SUCCESS)
	  return krc;

	if (size > vm_private_free) {
	  fret("vm_allocate: Requested size %lu is greater than max allowed %lu",
	       size, vm_private_free);
	  return KERN_NO_SPACE;	
	}
      }
      
      /* 
       * If the request is for 1 page, 
       * try to reuse a page from the free list
       */
      if ((size == vm_page_size) &&
	  (vm_private_free_listP != NULL)) {
	  
	  ASSERT (IN_PRIVATE((vm_offset_t)vm_private_free_listP));
	  ASSERT (vm_private_free_list_entries > 0);
	  
	  *addr = (vm_size_t) vm_private_free_listP;
	  vm_private_free_listP = vm_private_free_listP->nextP;
	  ASSERT (IN_PRIVATE((vm_offset_t)vm_private_free_listP));
	  vm_private_free_list_entries--;
	  
	  /* must return a zerofilled page! */
	  bzero (*addr, vm_page_size);

	  INCR_STAT(xpager_stat.alloc_free_list, 1);
	  
      } else {
      
	  /* Allocate new space for this request */
	  *addr = vm_private_free_spaceP;
	  vm_private_free_spaceP += size;
	  vm_private_free -= size;
      }

      INCR_STAT(xpager_stat.alloc_priv_calls, 1);
      INCR_STAT(xpager_stat.alloc_priv_space, size);
#ifdef STATS
      { register unsigned s = size >> (STAT_TABLE_MIN_SHIFT + 1);
	register unsigned i = 0;
	while (s) { s >>= 1; i++; }
	ASSERT ((i >= 0) && (i < STAT_TABLE_SIZE));
	INCR_STAT (xpager_stat.alloc_priv_size_count[i], 1);
      }
#endif
      krc = KERN_SUCCESS;
      
    } else { 

      /* check that the caller didn't request reserved memory */
      if (IN_SHARED(*addr) || IN_SHARED(*addr+size-1) ||
	  IN_DIRTY(*addr) || IN_DIRTY(*addr+size-1) ||
	  IN_PRIVATE(*addr) || IN_PRIVATE(*addr+size-1) ||
	  IN_LOCKTABLE(*addr) || IN_LOCKTABLE(*addr+size-1)) {
	fret("Not allowed to vm_allocate from shared/dirty region (0x%x+%ld)", *addr, size);
	return KERN_NO_SPACE;	
      }

      /* map the request into the caller's address space  */
      krc = vm_map(task, 
		   addr, size, (vm_offset_t)0, anywhere,
		   MEMORY_OBJECT_NULL, (vm_offset_t)0, FALSE,
		   VM_PROT_DEFAULT, VM_PROT_ALL, VM_INHERIT_COPY);
      if (krc == KERN_SUCCESS)  {
	
	if (IN_SHARED(*addr) || IN_SHARED(*addr+size-1) ||
	    IN_DIRTY(*addr) || IN_DIRTY(*addr+size-1) ||
	    IN_PRIVATE(*addr) || IN_PRIVATE(*addr+size-1) ||
	    IN_LOCKTABLE(*addr) || IN_LOCKTABLE(*addr+size-1)) {
	  fret("vm_allocate failed to allocate outside reserved regions (0x%x+%ld)", 
	       *addr, size);
	  (void)VM_DEALLOCATE(task, *addr, size);
	  return KERN_NO_SPACE;
	  
	  /* We could try again, with a different address */
	}
      }
    }
  
    mumble("vm_allocate complete(addr=0x%x), krc=%d", *addr, krc);
    ASSERT ((krc != KERN_SUCCESS) ||
	    (((((unsigned)*addr) >> vm_page_shift) << vm_page_shift) == ((unsigned)*addr)));
    
    return krc;
}


kern_return_t
midway_vm_allocate(task, addr, size, anywhere)
	task_t		task;
	vm_offset_t	*addr;
	vm_size_t	size;
	boolean_t	anywhere;
{
    kern_return_t krc;

    INCR_STAT(xpager_stat.alloc_lib_calls, 1);
    INCR_STAT(xpager_stat.alloc_lib_space, size);
#ifdef STATS
    { register unsigned s = size >> (STAT_TABLE_MIN_SHIFT + 1);
      register unsigned i = 0;
      while (s) { s >>= 1; i++; }
      ASSERT ((i >= 0) && (i < STAT_TABLE_SIZE));
      INCR_STAT (xpager_stat.alloc_lib_size_count[i], 1);
    }
#endif

    krc = vm_allocate (task, addr, size, anywhere);
    return krc;
}

kern_return_t
midway_vm_deallocate (
	       task_t		task,
	       vm_offset_t	addr,
	       vm_size_t	size
	       )
{
    kern_return_t krc;

    INCR_STAT(xpager_stat.dealloc_lib_calls, 1);
    INCR_STAT(xpager_stat.dealloc_lib_space, size);
#ifdef STATS
    { register unsigned s = size >> (STAT_TABLE_MIN_SHIFT + 1);
      register unsigned i = 0;
      while (s) { s >>= 1; i++; }
      ASSERT ((i >= 0) && (i < STAT_TABLE_SIZE));
      INCR_STAT (xpager_stat.dealloc_lib_size_count[i], 1);
    }
#endif

    /* Call kernel to really deallocate the memory */
    krc = vm_deallocate (task, addr, size);
    return (krc);
}
