/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * vm_paged.c:  Routines for handling external pager requests 
 *              for paged based execution of midway.
 *
 *  6 Feb 94 (was): Initial code
 *  5 Mar 94 (was): converted to EC instead of RC
 *  6 Sep 94 (was): cleaned up EC version, for performance and better stats
 */
#include <sys/types.h>		/* u_long & friends */
#include <mach.h>
#include <cthreads.h>
#include <stdio.h>
#include <ctype.h>
#include "midway_types.h" 
#include "smexecd_types.h"

/*undef TRACE*/			/* disable tracing for now */
/*define TRACE 0*/

#define DETAILS 1

#include "midway-dbg.h"
#include "lock.h"
#include "vm.h"
#include "shmalloc.h"
#include "pbuf.h"
#include "backing.h"
#include "xpager.h"
#include "xpager_obj.h"
#include "XPager.h"		/* mig generated */


/**
*** In theory, we could map the paged area twice, once with 
*** write protected for the application and once with write
*** access for us to update it. Unfortunately, MACH doesn't
*** allow xpagers to have multiple mappings. The protection
*** set by vm_map/vm_protect merely determines which access faults
*** are sent to the xpager, the others cause the process to 
*** terminate and dump core. Once in the xpager, protection
*** set on the memory object itself, change the protectiion for
*** all mappings. Thus preventing a multiple mapping with 
*** different levels of protection.   WAS 8Mar94
***
*** We can "fix" this by catching signals. Actually, once
*** we catch signals, do we really need the xpager? WAS 17Aug94
***
*** #define MULTIPLE_VM_MAPS 1 
**/


#ifdef BENCHMARK
#  define PSTATS 1
#  undef  PTRACE_PAGE
#else /* not BENCHMARK */

#ifndef PSTATS
#  define PSTATS 1		/* =1 to collect histograms */
				/* =0 to disable histogram collection */
#endif
#if TRACE > 0
#  define PTRACE_PAGE 1
#endif

#endif /* BENCHMARK */



/*
 * Define page states.
 *   FREE = page desc on free list
 *   CLEAN = page allocated, its clean, its mapped read-only. twin & diffs are empty
 *   DIRTY = page has been written, its dirty & mapped read-write. twin contains previous data
 *           diffs are unused.
 *   DIFFED = page has been written, but only partially cleaned. Its mapped read-only
 *            The diffs are valid, the twin contains data, but must be updated w/ cleaned data.
 */
typedef enum { PAGE_FREE, PAGE_CLEAN, PAGE_DIRTY, PAGE_DIFFED } page_state_t;


/*
 * Define a diff descriptor to store a page diff
 */
typedef struct {
    unsigned short	offset;	/* offset w/in page of a difference in data  */
    unsigned short	length;	/* length of changed data  */
} diff_desc_t;

/*
 * Max number of diffs per page occurs when every other word in the page is dirty.
 * A 4K page has 1024 words, and would require 512 diff_desc's 
 */
#define DIFF_PAGE_SIZE      4096 	/* hack to make life easier */
#define MAX_DIFFS_PER_PAGE (DIFF_PAGE_SIZE / (2 * sizeof(int)))


/*
 * Define page descriptor to store page state
 */
typedef struct page_desc {
    vm_offset_t        page_addr; 	/* real address of this page */
    vm_offset_t        twin_addr; 	/* copy of page or buffer used for page data */
    page_state_t       state;		/* page state */
    unsigned short     diff_reuse; 	/* collect stats on #times diff is reused */
    unsigned short     dirty_len; 	/* number of dirty areas ("diffs") on the page */
    unsigned short     dirty_data;	/* total dirty data on the page */
    unsigned short     dirty_unsorted;	/* =1 if diffs are NOT sorted by addr, =0 otherwise */
    unsigned short     clean_len; 	/* number of cleaned areas on the page */
    unsigned short     clean_data;	/* total cleaned data on the page */
    diff_desc_t	      *diffP;		/* ptr to diffs describing changes to this page */
    diff_desc_t	      *cleanP;		/* ptr to diffs describing cleaned diffs */
    struct page_desc  *nextP;		/* single-linked list pointer for free list */
} page_desc_t;

/*
 * Define free lists for the page descriptors 
 */
#define PAGED_INIT_FREE_LIST   0
#define PAGED_MAX_FREE_LIST  256

static page_desc_t *free_list;
static int          free_list_entries;

/*
 * Define "constant" for space needed for each page desc 
 */
static vm_size_t page_twin_buffer_size;





/*
 * Define page table used to determine free/allocated pages
 * Allocate 1 entry per page. If the page is dirty, store ptr to 
 * page_desc in the table.
 */
typedef union {
    unsigned long  marked;
    page_desc_t   *descP;
} page_table_t;

static page_table_t *page_table; /* ptr to per-page data array */

#define PAGE_MARKED_FREE  0x0
#define PAGE_MARKED_CLEAN 0x1

#define PAGE_TABLE(_addr)	  page_table[ ADDR_TO_PAGENO(_addr-vm_paged_base) ]
#define PAGE_DESC(_addr)          PAGE_TABLE(_addr).descP

#define IS_PAGE_ALLOCATED(_addr)  (PAGE_TABLE(_addr).marked != PAGE_MARKED_FREE)
#define IS_PAGE_CLEAN(_addr)      (PAGE_TABLE(_addr).marked == PAGE_MARKED_CLEAN)
#define IS_PAGE_DIRTY(_addr)      (PAGE_TABLE(_addr).marked > PAGE_MARKED_CLEAN)

#define MARK_PAGE_CLEAN(_addr)       	PAGE_TABLE(_addr).marked = PAGE_MARKED_CLEAN
#define MARK_PAGE_DIRTY(_addr,_desc)	PAGE_DESC(_addr) = _desc





/*
 * Define our static & external variables
 */
static mach_port_t paged_memory_pager_port = MACH_PORT_NULL;
static mutex_t paged_mutex;

static vm_offset_t last_page_created = 0;


vm_offset_t vm_static_base;
vm_offset_t vm_static_end;

vm_offset_t vm_paged_base;
vm_offset_t vm_paged_end;
vm_offset_t vm_paged_free_spaceP;
vm_offset_t vm_paged_free;

#ifdef MULTIPLE_VM_MAPS

vm_offset_t vm_paged_write_base; 	/* R/W mapping for paged region  */
vm_offset_t vm_paged_write_end;

#  define WRITE_ADDR(_addr) (vm_paged_write_base + (((vm_offset_t)(_addr)) - vm_paged_base))
#else
#  define WRITE_ADDR(_addr) (_addr) /* just write directly to memory */
#endif

/*
 * Declare external variables we reference
 */
extern task_t midway_task_self;	/* mach_task_self() */




/*
 * forward declarations 
 */
static kern_return_t make_twin (vm_offset_t page_addr, vm_offset_t page_data);
static void apply_clean_diffs (page_desc_t *pageP);

typedef enum {NO_ACCESS, READONLY_ACCESS, READWRITE_ACCESS} page_access_t;
static void page_protect (vm_offset_t page_addr, page_access_t access_allowed);



/*
 * Define useful macros 
 */
#ifndef MIN
#  define MIN(a,b) (a < b) ? a : b
#endif
#ifndef MAX
#  define MAX(a,b) (a >= b) ? a : b
#endif

#define PAGE_ADDR(_addr)	((vm_offset_t)((((vm_offset_t)_addr) >> vm_page_shift) << vm_page_shift))
#define NEXT_PAGE_ADDR(_addr)	((vm_offset_t)(PAGE_ADDR(_addr) + vm_page_size))
#define PAGE_OFFSET(_addr)	(((vm_offset_t)_addr) - PAGE_ADDR(_addr))
#define PAGE_REMAINS(_addr)	(NEXT_PAGE_ADDR(_addr) - ((vm_offset_t)_addr))



/*
 * Define structure & macros for collecting statistics
 */
#ifdef STATS

#define STAT_DIFFCOUNT_PER_BUCKET  1
#define STAT_DIFFCOUNT_TABLE_SIZE  256
#define STAT_DIFFSIZE_PER_BUCKET   4
#define STAT_DIFFSIZE_TABLE_SIZE   1024
#define STAT_DIFFPAGE_PER_BUCKET   4
#define STAT_DIFFPAGE_TABLE_SIZE   1024
#define STAT_DIFFTOTAL_PER_BUCKET  1024
#define STAT_DIFFTOTAL_TABLE_SIZE  4096
#define STAT_FLUSHDIFFS_PER_BUCKET 4
#define STAT_FLUSHDIFFS_TABLE_SIZE 1024

#define STAT_DIFFREUSE_TABLE_SIZE  64
#define STAT_DIFFREUSE_PER_BUCKET   1

static struct {
    u_long	data_reqs;		/* # of data requests */
    u_long	data_reqsRead;		/* # of data requests for read access */
    u_long	data_reqsWrite;		/* # of data requests for write access */
    u_long	data_reqsZeroed;	/* # of data requests for local zeroed pages */
    u_long      data_free;		/* # of databuf_free calls */
    u_long      data_write;		/* # of databuf_write calls */
    u_long      data_access;		/* # of databuf_access calls */
    u_long      data_accessClean;	/* # of databuf_access calls for a clean buffer */
    u_long      data_accessDiffed;	/* # of databuf_access calls for a diffed buffer */
    u_long      data_accessDirty;	/* # of databuf_access calls for a dirty buffer */
    u_long	malloc_calls;		/* # of calls to malloc_page_desc */
    u_long	malloc_max;		/* Max page descriptor's allocated at any time */
    u_long	malloc_zeroed;		/* # of pages bzeroed */
    u_long	free_calls;		/* # of calls to free_page_desc */
    u_long	free_max;  		/* Max page descriptor's on the free list at any time*/
    u_long	protect_NOcalls;	/* # of calls to page_protect (no access) */
    u_long	protect_ROcalls;	/* # of calls to page_protect (r-o access) */
    u_long	protect_RWcalls;	/* # of calls to page_protect (r/w access) */
    u_long	protect_mmROcalls;	/* # of calls due to lack of multiple vm mapping*/
    u_long	protect_mmRWcalls;	/* # of calls  "  "   "   "     "     "     "   */
    u_long	protect_vmNOcalls;	/* # of calls to page_protect (no access) */
    u_long	protect_vmROcalls;	/* # of calls to page_protect (r-o access) */
    u_long	protect_vmRWcalls;	/* # of calls to page_protect (r/w access) */
    u_long	page_twins;		/* # of page twins made */
    u_long	page_twinsCopied;	/* # of page twins requiring a data copy */

    u_long	make_diffs;		/* # of page diffs made */
    u_long	make_diffsCount; 	/* total number of diffs generated  */
    u_long	make_diffsSize; 	/* total number of different bytes */
    u_long     *make_diffsSizeH;  	/* histogram of sizeof each diff */
    u_long	make_diffsSizePage;	/* should = make_diffsSize */
    u_long     *make_diffsSizePageH;  	/* histogram of sum of diff sizes per page */

    u_long	 diff_reuseCleanTotal;	/* total diffs reused before completely cleaned */
    u_long	*diff_reuseCleanH;	/* hisogram of reuse */
    u_long	 diff_reuseFaultTotal;	/* total diffs reused before a fault occured */
    u_long	*diff_reuseFaultH;	/* hisogram of reuse */
    u_long	 diff_reuseUpdateTotal;	/* total diffs reused before an update applied*/
    u_long	*diff_reuseUpdateH;	/* hisogram of reuse */ 

    u_long	copy_data;		/* # of calls to copy raw data */
    u_long	copy_dataSize;		/* length of raw data copied */
    u_long	copy_dirty;		/* # of copy diffs calls */
    u_long	copy_dirtyCount;	/* # of diffs copied actually copied */
    u_long	copy_dirtySize;		/* length of diff data copied */
    u_long	copy_dirtyCountComp; 	/* # of dirty diffs completed copied */
    u_long	copy_dirtyCountFirst;	/* # of dirty diffs copied front portion */
    u_long	copy_dirtyCountLast; 	/* # of dirty diffs copied last portion */
    u_long	copy_dirtyCountMiddle;	/* # of dirty diffs copied middle portion */
    u_long	apply_clean;		/* # of apply clean diff calls */
    u_long	apply_cleanDiffs;	/* # of diffs actually applied to twin */
    u_long	apply_cleanSize;	/* length of diff data applied */
    u_long     *apply_cleanSizeH; 	/* histogram of size of cleaned data */
    u_long	apply_cleanSizePage;	/* should = apply_cleanSize */
    u_long     *apply_cleanSizePageH;	/* histogram of total applied to each page */

    u_long	pack_diffCalls;		/* # of calls to pack diff data */
    u_long	pack_dataCalls;		/* # of calls to pack all data */
    u_long	pack_diffSize; 		/* bytes of diff data packed */
    u_long	pack_dataSize; 		/* bytes of raw data packed */
    u_long	pack_unallocSize;	/* size of diffs for unallocated pages */
    u_long	pack_cleanSize;		/* size of diffs for clean pages */
    u_long	pack_dirtySize;		/* size of diffs for dirty pages */
    u_long	pack_diffedSize;	/* size of diffs for diffed pages */
    u_long	pack_diffsMade;		/* # of pages diffed */
    u_long	pack_diffsReused;	/* # of diffs reused */
    u_long	pack_cleaned;  		/* # of pages cleaned */

    u_long	unpack_calls;		/* # of calls to unpack */
    u_long	unpack_staticCount; 	/* # of diffs for static data  */
    u_long	unpack_staticSize; 	/* bytes of data copied to statics */
    u_long	unpack_pagesAlloc;	/* # of page allocates */
    u_long	unpack_cleanCount;	/* # of diffs for clean pages */
    u_long	unpack_cleanSize;	/* bytes copied to clean pages */
    u_long	unpack_dirtyCount;	/* # of diffs for dirty pages */
    u_long	unpack_dirtySize;	/* bytes copied to dirty pages */
    u_long	unpack_diffedCount;	/* # of diffs for dirty pages */
    u_long	unpack_diffedSize;	/* bytes copied to dirty pages */
    u_long	unpack_errors;		/* application errors overwriting live data */

    u_long	system_errors;   	/* # of o/s, xpager, comm errors */
} PagedStats;
static mutex_t statLock;

# ifdef FAST_STATS
#  define INCR_STATS(x,by) (x) += (by)
# else
#  define INCR_STATS(x,by) \
    { mutex_lock(statLock);  (x) += (by);  mutex_unlock(statLock); }
# endif 

#define INCR_HISTO(total,histoP,val,hSize,bSize) \
{ register u_long h_index = (val);	\
  INCR_STATS(total, h_index);		\
  h_index /= bSize;   			\
  if (h_index > hSize) h_index = hSize; \
  INCR_STATS(histoP[h_index], 1);	\
}

#else /* not STATS */
#define INCR_STATS(x,by)   /* nothing to do  */
#define INCR_HISTO(total,histoP,val,hSize,bSize)
#endif /* STATS */



/*
 ************************************************************************
 *									*
 *	Internal routines 						*
 *									*
 ************************************************************************
 */

/*
 * Internal routine to allocate a page_desc structure
 * Caller must hold paged_mutex
 */
static page_desc_t *malloc_page_desc (int zero_page)
{
    page_desc_t *pageP;
    kern_return_t krc;

    INCR_STATS(PagedStats.malloc_calls, 1);
#ifdef STATS
    if ((PagedStats.malloc_calls - PagedStats.free_calls) > PagedStats.malloc_max) 
	PagedStats.malloc_max = PagedStats.malloc_calls - PagedStats.free_calls;
#endif
    
    if (free_list == NULL) {
	ASSERT (free_list_entries == 0);
	pageP = (page_desc_t *)MALLOC(sizeof(*pageP));
	if (pageP == NULL) {
	    fret("malloc_page_desc: malloc failed");
	    goto done;
	}

	krc = VM_ALLOCATE (midway_task_self, &pageP->twin_addr, page_twin_buffer_size, TRUE);
	if (krc != KERN_SUCCESS) {
	    fret("malloc_page_desc: vm_allocate failed, krc=%s", mach_error_string(krc));
	    FREE(pageP);
	    pageP = NULL;
	    goto done;
	}

	pageP->diffP = (diff_desc_t *)(pageP->twin_addr + vm_page_size);
	pageP->cleanP = (diff_desc_t *)(((vm_offset_t)pageP->diffP) 
					   + (MAX_DIFFS_PER_PAGE * sizeof(diff_desc_t)));

    } else {
	ASSERT (free_list_entries > 0);
	pageP = free_list;
	free_list = pageP->nextP;
	free_list_entries--;
	
	/* zero page before reusing it, if necessary */
	if (zero_page) {
	    bzero (pageP->twin_addr, vm_page_size);
	    INCR_STATS(PagedStats.malloc_zeroed, 1);
	}
    }

    pageP->page_addr = 0; 
    pageP->state = PAGE_FREE;
    pageP->diff_reuse = 0;
    pageP->dirty_len = 0;
    pageP->dirty_data = 0;
    pageP->dirty_unsorted = 0;
    pageP->clean_len = 0;
    pageP->clean_data = 0;
    pageP->nextP = NULL;
    
done:
    return(pageP);
} /* malloc_page_desc */

/* Extern routine used for measuring diff times */
extern void * midway_malloc_page_desc (vm_offset_t page_addr)
{
    page_desc_t *pageP;

    pageP = malloc_page_desc(1);
    pageP->page_addr = page_addr;
    return ((void *)pageP);
}

/*
 * Internal routine to free a page_desc structure
 * Caller must hold paged_mutex
 */
static void free_page_desc ( 
			     page_desc_t *pageP
			    )
{
    kern_return_t krc;

    INCR_STATS(PagedStats.free_calls, 1);

    if (free_list_entries < PAGED_MAX_FREE_LIST) {
	pageP->state = PAGE_FREE;
	pageP->nextP = free_list;
	free_list = pageP;
	free_list_entries++;
#ifdef STATS
	if (free_list_entries > PagedStats.free_max)
	    PagedStats.free_max = free_list_entries;
#endif
    } else {
	krc = VM_DEALLOCATE (midway_task_self, pageP->twin_addr, page_twin_buffer_size);
	if (krc != KERN_SUCCESS) {
	    fret("free_page_desc: vm_deallocate failed, krc=%s", mach_error_string(krc));
	}

	SMASH_AREA(pageP, sizeof(*pageP));
	FREE(pageP);
    }
} /* free_page_desc */



/*
 ************************************************************************
 *	    x p a g e r _ r e q u e s t _ f u n c t i o n s             *
 *----------------------------------------------------------------------*
 * Xpager request functions. Called indirectly by the xpager only.	*
 ************************************************************************
 */

static kern_return_t 
paged_memory_data_request (
			   vm_offset_t arg,
			   vm_offset_t offset,
			   vm_size_t   len,
			   vm_prot_t   *access_prot,
			   pointer_t   *buf,
			   boolean_t   *precious
			   )
{
    kern_return_t krc;
    vm_prot_t prot = *access_prot;
    page_desc_t *twinP, *pageP;
    vm_offset_t page_addr = offset + arg;

    if (IS_DBG_PORT_PAGE) {
	char255 smsg;
	sprintf(smsg, "paged_memory_data_request(0x%x 0x%x %lu %s)",
		arg, offset, len, 
		(prot == VM_PROT_NONE) ? "none" :
		(prot == VM_PROT_READ) ? "read" :
		(prot == VM_PROT_WRITE) ? "write" :
		(prot == VM_PROT_READ+VM_PROT_WRITE) ? "read+write" :
		(prot == VM_PROT_EXECUTE) ? "exec" : "unknown");
	StatusMsg(smsg);
    }
    INCR_STATS(PagedStats.data_reqs, 1);
#ifdef STATS
    if (prot & VM_PROT_WRITE) {
	INCR_STATS(PagedStats.data_reqsWrite, 1);
    } else {
	INCR_STATS(PagedStats.data_reqsRead, 1);
    }
#endif
    
    if (!IN_SHARED(page_addr)) {
	fret("paged memory data request for non-shared space 0x%x", page_addr);
	INCR_STATS(PagedStats.system_errors,1);
	return KERN_FAILURE;
    }
    ASSERT (((vm_offset_t) arg) == vm_paged_base);

    /* Obtain exclusive access to page data structures */
    mutex_lock (paged_mutex);
    TRACE_3("XREQ", page_addr, prot);

    /*
     * Check if the page has been allocated 
     */
    if (!IS_PAGE_ALLOCATED(page_addr)) {

	/*
	 * Page was never allocated
	 * supply a zero filled page 
	 */
	*buf = 0;	
	TRACE_2("XReZ", page_addr);

	/* Check if we need to create a zero filled twin to match it */
	if (prot & VM_PROT_WRITE)
	    krc = make_twin (page_addr, 0);
	else {
	    MARK_PAGE_CLEAN(page_addr);
	    krc = KERN_SUCCESS;
	}

#ifdef MULTIPLE_VM_MAPS
	if (last_page_created != 0)
	    vm_page_protect(WRITE_ADDR(last_page_created), READWRITE_ACCESS);
	last_page_created = page_addr;
#endif


    } else {

	/*
	 * Page already allocated
	 */
	fret ("paged_data_request for allocated page 0x%x access %d\n",
	      page_addr, prot);
	INCR_STATS(PagedStats.system_errors, 1);
	TRACE_3("XReE", page_addr, prot);
	krc = KERN_FAILURE;
    }
    
    /* release access to page data structures */
    mutex_unlock (paged_mutex);
    
    return krc;
} /* paged_memory_data_request */


static kern_return_t
paged_memory_databuf_free (
			   vm_offset_t arg,
			   pointer_t   buf
			   )
{
    kern_return_t krc = KERN_SUCCESS;
    page_desc_t *pageP;
    
    if (IS_DBG_PORT_PAGE) {
	char255 smsg;
	sprintf(smsg, "paged_memory_databuf_free(0x%x 0x%x)",
		arg, buf);
	StatusMsg(smsg);
    }
    INCR_STATS(PagedStats.data_free, 1);
    TRACE_2("XFRE", buf);
    
    if (buf)  {

	/* We never provide a buffer */
	fret ("paged_databuf_free request for 0x%x\n", buf);
	INCR_STATS(PagedStats.system_errors, 1);
	krc = KERN_FAILURE;
    }
    return(krc);
} /* paged_memory_databuf_free */


static kern_return_t
paged_memory_data_write (
			 vm_offset_t arg,
			 vm_offset_t offset,
			 vm_size_t   len,
			 pointer_t *buf,
			 boolean_t   *discarded
			)
{
    vm_offset_t page_addr = offset + arg;

    if (IS_DBG_PORT_PAGE) {
	char255 smsg;
	sprintf(smsg, "paged_memory_data_write(0x%x offset=0x%x len=%ld buf=0x%x)",
		arg, offset, len, buf);
	StatusMsg(smsg);
    }
    INCR_STATS(PagedStats.data_write, 1);
    TRACE_2("XWRT", page_addr);
    
    if (!IN_SHARED(page_addr)) {
	fret("paged memory data write for non-shared space 0x%x", page_addr);
	INCR_STATS(PagedStats.system_errors, 1);
	return KERN_FAILURE;
    }    
    
    *discarded = FALSE;
    return KERN_SUCCESS;
} /* paged_memory_data_write */


static kern_return_t
paged_memory_data_access_request (
				  vm_offset_t   arg,
				  vm_offset_t 	offset,
				  vm_size_t   	len,
				  vm_prot_t	access_prot
				 )
{
    kern_return_t krc;
    vm_offset_t page_addr = offset + arg;
    page_desc_t *pageP;
    page_access_t access;
    
    if (IS_DBG_PORT_PAGE) {
	char255 smsg;
	sprintf(smsg, "paged_memory_data_access_request(0x%x 0x%x %lu %s)",
		arg, offset, len, 
		(access_prot == VM_PROT_NONE) ? "none" :
		(access_prot == VM_PROT_READ) ? "read" :
		(access_prot == VM_PROT_WRITE) ? "write" :
		(access_prot == VM_PROT_READ+VM_PROT_WRITE) ? "read+write" :
		(access_prot == VM_PROT_EXECUTE) ? "exec" : "unknown");
	StatusMsg(smsg);
    }
    INCR_STATS(PagedStats.data_access, 1);
    
    if (!IN_SHARED(page_addr)) {
	fret("paged memory data access request for non-shared space 0x%x", page_addr);
	INCR_STATS(PagedStats.system_errors, 1);
	return KERN_FAILURE;
    }

    mutex_lock(paged_mutex);

    /* Verify that the page was allocated */
    ASSERT (IS_PAGE_ALLOCATED(page_addr));
    ASSERT (access_prot & VM_PROT_WRITE);

    /* Check if the page is marked clean */
    if (IS_PAGE_CLEAN(page_addr)) {
	
	INCR_STATS(PagedStats.data_accessClean,1);

	/* make a twin & return */
	krc = make_twin (page_addr, page_addr);
	if (krc == KERN_SUCCESS) {

	    /* Grant the requested page access */
	    page_protect (page_addr, READWRITE_ACCESS);
	}
    } else {

	/* verify that the page is allocated */
	ASSERT (IS_PAGE_DIRTY(page_addr));

	pageP = PAGE_DESC(page_addr);
	ASSERT (pageP != NULL);

	/*
	 * Check if the page has been diffed
	 */
	if (pageP->state == PAGE_DIFFED) {

	    INCR_STATS(PagedStats.data_accessDiffed,1);
	    INCR_HISTO (PagedStats.diff_reuseFaultTotal,PagedStats.diff_reuseFaultH,
			pageP->diff_reuse, STAT_DIFFREUSE_TABLE_SIZE, 
			STAT_DIFFREUSE_PER_BUCKET);

	    /*
	     * The page was dirty, has been diffed, but is only partially clean.
	     * The pending write will invalidate the saved diff.
	     * Update the twin to contain the cleaned data
	     */
	    apply_clean_diffs( pageP );

	    /* Mark the page as dirty and unprotect it. */
	    pageP->state = PAGE_DIRTY;
	    page_protect (page_addr, READWRITE_ACCESS);
	    krc = KERN_SUCCESS;
	    
	} else {

	    ASSERT (pageP->state == PAGE_DIRTY);
	    INCR_STATS(PagedStats.data_accessDirty,1);

	    /*
	     * Hmmm...
	     * Write access has already been restore to the page. 
	     * This happens when the application tries to write to a page
	     * we are cleaning. The cleaning code holds the mutex until complete.
	     * If the page was clean, its marked clean & left read-only.
	     * If the page still had dirty data on it, its marked as "diffed"
	     * and left read-only. So, a subsequent update must have been applied
	     * to the diff, which made the page dirty again and restored write-access.
	     */
	    krc = KERN_SUCCESS;
	}
    }

    /* release access to page data structures */
    mutex_unlock (paged_mutex);
    
    return krc;
} /* paged_memory_data_access_request */


/*
 ************************************************************************
 *									*
 *	Initialization and Termination routines 			*
 *									*
 ************************************************************************
 */

struct xpager_request paged_xpager_rq = {
    0, 
    paged_memory_data_request,
    paged_memory_databuf_free,
    paged_memory_data_write,
    paged_memory_data_access_request
    };

kern_return_t
vm_paged_init (
	       vm_size_t size
	      )
{
    int i, j, k, changed;
    kern_return_t krc;
    vm_offset_t addr;
    vm_size_t table_size, static_size;
    xpager_request_t rq;
    page_desc_t *pageP, *listP;

#ifdef WAIT
    printf("vm_paged_init\n");
    for (i = 0; i < 10000000; i++) {
	if ((i % 1000000) == 0) printf("waiting\n");
    }
#endif


    if ((vm_offset_t)LAST_STATIC_SYM > vm_shared_base)  {
	botch("vm_paged_init: statics overflowing into shared region. Call a guru.");
	exit(-1);
    }
    
    /* Convert page size to a shift amount to eliminate divides */
    if (vm_page_shift == 0)
	vm_page_shift = midway_int_to_log_int(vm_page_size);

#ifdef TRACE
    midway_init_trace();
#endif

#if PSTATS > 0
    statLock = mutex_alloc();
    bzero (&PagedStats, sizeof(PagedStats));
    PagedStats.make_diffsSizeH = (u_long *) CALLOC (STAT_DIFFSIZE_TABLE_SIZE+1, 
						    sizeof(u_long));
    ASSERT(PagedStats.make_diffsSizeH != NULL);
    PagedStats.make_diffsSizePageH = (u_long *) CALLOC (STAT_DIFFPAGE_TABLE_SIZE+1, 
							 sizeof(u_long));
    ASSERT(PagedStats.make_diffsSizePageH != NULL);
    PagedStats.apply_cleanSizeH = (u_long *) CALLOC (STAT_DIFFPAGE_TABLE_SIZE+1, 
							 sizeof(u_long));
    ASSERT(PagedStats.apply_cleanSizeH != NULL);
    PagedStats.apply_cleanSizePageH = (u_long *) CALLOC (STAT_DIFFPAGE_TABLE_SIZE+1, 
							 sizeof(u_long));
    ASSERT(PagedStats.apply_cleanSizePageH != NULL);

    PagedStats.diff_reuseCleanH = (u_long *) CALLOC (STAT_DIFFREUSE_TABLE_SIZE+1, 
						     sizeof(u_long));
    ASSERT(PagedStats.diff_reuseCleanH != NULL);
    PagedStats.diff_reuseFaultH = (u_long *) CALLOC (STAT_DIFFREUSE_TABLE_SIZE+1, 
						     sizeof(u_long));
    ASSERT(PagedStats.diff_reuseFaultH != NULL);
    PagedStats.diff_reuseUpdateH = (u_long *) CALLOC (STAT_DIFFREUSE_TABLE_SIZE+1, 
						      sizeof(u_long));
    ASSERT(PagedStats.diff_reuseUpdateH != NULL);
#endif
    TRACE_2("INIT", size);

    /* 
     * initialize paged base region boundaries 
     */
    vm_paged_base = vm_shared_base;
    vm_paged_free_spaceP = vm_paged_base;
    vm_paged_free = size;

    /*
     * Initialize internal data structures
     */
    paged_mutex = mutex_alloc();
    free_list = NULL;
    free_list_entries = 0;
    page_twin_buffer_size = ((vm_page_size 						/* twin */
				+ (MAX_DIFFS_PER_PAGE * sizeof(diff_desc_t) * 2)	/* diffs */
				+ (vm_page_size - 1)) >> vm_page_shift) << vm_page_shift;

    ASSERT (MAX_DATA_SIZE > vm_page_size);
    ASSERT (DIFF_PAGE_SIZE == vm_page_size);

    /*
     * Allocate per-page data
     * This is just malloc'ed from private storage.
     */
    table_size = ((MAX_SHARED_SPACE) >> vm_page_shift) * sizeof(*page_table);
    page_table = (page_table_t *) MALLOC(table_size);
    ASSERT (page_table != NULL);
    
    /* To be safe, we should zero this array.
     * But I don't want to zero 512K of space,
     * when most of it has not been allocated and won't be used.
     * Solution: Zero the first few pages, Don't sweat the rest.
     */
    if (table_size > (vm_page_size*16))
	bzero (page_table, vm_page_size*16);
    else
	bzero (page_table, table_size);
    
    /*
     * Initialize external pager
     */
    xpager_init(0);
    
    /* 
     * Set up a pager thread for the paged/shared memory
     */
    krc = do_xpager_create(MACH_PORT_NULL, &paged_memory_pager_port);
    if (krc != KERN_SUCCESS)  {
	fret("vm_paged_init failure: do_xpager_create:%s\n",
	     mach_error_string(krc));
	return krc;
    }
    rq = &paged_xpager_rq;
    paged_xpager_rq.data_any = (char*)vm_paged_base;
    krc = set_xpager_request(paged_memory_pager_port, &rq);
    if (krc != KERN_SUCCESS)  {
	fret("vm_paged_init failure: set_xpager_request: paged_memory: %s\n",
	     mach_error_string(krc));
	return krc;
    }


    /*
     * ignore control channel
     */
  
    
    /*
     * Now map the whole range in.
     */
    addr = vm_paged_base;
    krc = vm_map(mach_task_self(),
	  	 &addr,
		 size,
		 0,
		 FALSE,
		 paged_memory_pager_port,
		 0,
		 FALSE,
		 VM_PROT_DEFAULT,
		 VM_PROT_ALL,
		 VM_INHERIT_COPY);
    if (addr != vm_paged_base)  {
	botch("vm_paged_init");
    }
    if (krc != KERN_SUCCESS)  {
	fret("vm_map failure in vm_paged_init: %s\n",
	     mach_error_string(krc));
	return krc;
    }
    vm_shared_end = vm_shared_base + size - 1;
    vm_paged_end = vm_shared_end;

    fret("vm_paged_init: base = 0x%x end = 0x%x", /*  */
	 vm_paged_base, vm_shared_end);

    /*
     * Allocate a few initial twin buffers
     */
    if (PAGED_INIT_FREE_LIST > 0) {
	fret("vm_paged_init: initializing %d page descriptors",  
	     PAGED_INIT_FREE_LIST);
	listP = NULL;
	for (i = 0; i < PAGED_INIT_FREE_LIST; i++) {
	     pageP = malloc_page_desc(1);
	     ASSERT (pageP != NULL);
	     pageP->nextP = listP;
	     *(int *)pageP->twin_addr = 0; /* allocate the twin */
	     listP = pageP;
	 }
	while (listP != NULL) {
	    pageP = listP;
	    listP = pageP->nextP;
	    free_page_desc(pageP);
	}
    }
	 


#ifdef MULTIPLE_VM_MAPS
    /*
     * Now map the whole range in with write access
     */
    vm_paged_write_base = vm_paged_base + size;
    vm_paged_write_end = vm_paged_write_base + size -1;
    addr = vm_paged_write_base;
    krc = vm_map(mach_task_self(),
	  	 &addr,
		 size,
		 0,		/* mask */
		 FALSE,		/* anywhere */
		 paged_memory_pager_port,
		 0,		/* offset */
		 FALSE,		/* Don't copy data! */
		 VM_PROT_DEFAULT,
		 VM_PROT_ALL,
		 VM_INHERIT_COPY);
    if (addr != vm_paged_write_base)  {
	botch("vm_paged_init(write)");
    }
    if (krc != KERN_SUCCESS)  {
	fret("vm_map(write) failure in vm_paged_init: %s\n",
	     mach_error_string(krc));
	return krc;
    }
#endif

    return krc;
} /* vm_paged_init */


void 
vm_paged_term (void)
{
    kern_return_t krc;
    
    if (paged_memory_pager_port != MACH_PORT_NULL) {
	krc = do_xpager_destroy(paged_memory_pager_port);
	if (krc != KERN_SUCCESS)  {
	    fret("vm_paged_term failure: do_xpager_destroy:%s\n",
		 mach_error_string(krc));
	}
	paged_memory_pager_port = MACH_PORT_NULL;
    }
}


/*
 ************************************************************************
 *									*
 *	Internal routines 						*
 *									*
 ************************************************************************
 */

/*
 * Internal routine to set the VM page protection
 */
static void page_protect (vm_offset_t page_addr, page_access_t access_allowed)
{
    kern_return_t krc;
    vm_offset_t page_offset;
    vm_prot_t lock_value;

    TRACE_3("PROT", 
	    (access_allowed == READONLY_ACCESS) ? 0x64616572 : 		/* "read" */
	    (access_allowed == READWRITE_ACCESS) ? 0x74697277 : 	/* "writ" */
	    (access_allowed == NO_ACCESS) ? 0x656e6f6e : access_allowed, /* "none" */
	    page_addr);
	    
    switch (access_allowed) {
    case NO_ACCESS:
	lock_value = VM_PROT_ALL;
	INCR_STATS(PagedStats.protect_NOcalls,1);
	break;
    case READONLY_ACCESS:
	lock_value = VM_PROT_WRITE;
	INCR_STATS(PagedStats.protect_ROcalls,1);
	break;
    case READWRITE_ACCESS:
	lock_value = VM_PROT_NONE;
	INCR_STATS(PagedStats.protect_RWcalls,1);
	break;
    default:
	ASSERT(FALSE);
    }
    
    /* Set the page protection */
    page_addr = PAGE_ADDR(page_addr);
    if ((page_addr >= vm_paged_base) &&
	(page_addr < vm_paged_end))
	page_offset = page_addr - vm_paged_base;

#ifdef MULTIPLE_VM_MAPS
    else if ((page_addr >= vm_paged_write_base) &&
	     (page_addr < vm_paged_write_end))
	page_offset = page_addr - vm_paged_write_base;
#endif
    else
	botch("Invalid address 0x%x to page_protect", page_addr);

    krc = xpager_protect_data (paged_memory_pager_port, 
			       page_offset, vm_page_size, 
			       lock_value);
    if (krc != KERN_SUCCESS) {
	fret("vm_paged: error in xpager_protect_data at 0x%x for lock %d, krc=%s",
	     page_addr, lock_value, mach_error_string(krc));
	INCR_STATS(PagedStats.system_errors, 1);
	TRACE_3("PRER", page_addr, krc);
    }
}

#ifdef MULTIPLE_VM_MAPS
static void vm_page_protect (vm_offset_t page_addr, page_access_t access_allowed)
{
    kern_return_t krc;
    vm_prot_t prot_allowed;

    TRACE_3("VPRO", 
	    (access_allowed == READONLY_ACCESS) ? 0x64616572 : 		/* "read" */
	    (access_allowed == READWRITE_ACCESS) ? 0x74697277 : 	/* "writ" */
	    (access_allowed == NO_ACCESS) ? 0x656e6f6e : access_allowed, /* "none" */
	    page_addr);
	    
    switch (access_allowed) {
    case NO_ACCESS:
	prot_allowed = VM_PROT_NONE;
	INCR_STATS(PagedStats.protect_vmNOcalls,1);
	break;
    case READONLY_ACCESS:
	prot_allowed = VM_PROT_READ;
	INCR_STATS(PagedStats.protect_vmROcalls,1);
	break;
    case READWRITE_ACCESS:
	prot_allowed = VM_PROT_READ | VM_PROT_WRITE;
	INCR_STATS(PagedStats.protect_vmRWcalls,1);
	break;
    default:
	ASSERT(FALSE);
    }
    
    /* Set the page protection */
    krc = vm_protect (midway_task_self,
		      page_addr, vm_page_size,
		      0, /* 0 = set current protection */
		      prot_allowed);
    if (krc != KERN_SUCCESS) {
	fret("vm_paged: error in vm_protect at 0x%x for prot %d, krc=%s",
	     page_addr, prot_allowed, mach_error_string(krc));
	INCR_STATS(PagedStats.system_errors, 1);
	TRACE_3("VPER", page_addr, krc);
    }
}
#endif /* MULTIPLE_VM_MAPS */
	    


/*
 * Internal routine to make a twin of an allocated page
 * Caller must hold paged_mutex
 */
static kern_return_t make_twin (
				vm_offset_t page_addr,
				vm_offset_t page_data
			       )
{
    page_desc_t *pageP;
    
    if (IS_DBG_PORT_PAGE) {
	char255 smsg;
	sprintf(smsg, "make_twin(addr=0x%x, data=0x%x)", page_addr, page_data);
	StatusMsg(smsg);
    }
    INCR_STATS(PagedStats.page_twins, 1);
    TRACE_3("TWIN", page_addr, page_data);
    
    /* get free page desc structure
     * have malloc zero page, only if necessary 
     */
    pageP = malloc_page_desc( (page_data == 0) ? TRUE : FALSE );
    if (pageP == NULL) return KERN_FAILURE;
    
    /* 
     * Make twin 
     */
    pageP->page_addr = page_addr;
    pageP->state = PAGE_DIRTY;
    if (page_data != 0) {
	bcopy (page_data,		/* source */
	       pageP->twin_addr,	/* target */
	       vm_page_size);		/* length */
	INCR_STATS(PagedStats.page_twinsCopied, 1);
    }
    
    /* insert descriptor into page table  */
    ASSERT (!IS_PAGE_DIRTY(page_addr));
    MARK_PAGE_DIRTY(page_addr, pageP);

    return KERN_SUCCESS;
} /* make_twin */


/*
 * Internal routine to make a diff describing the changes to a page
 * Caller must hold paged_mutex
 */
static void make_diff (
		       page_desc_t	*pageP
		      )
{
    register long page_data0, page_data1, page_data2, page_data3;
    register long twin_data0, twin_data1, twin_data2, twin_data3;
    register long *page_dataP, *twin_dataP, *page_endP, *loop_endP;
    register vm_offset_t page_addr;
    register vm_size_t actual_data; 
    register diff_desc_t *diffP;
    register u_long diff_offset, diff_length;
#if PSTATS > 0
    register u_long diff_count = 0;
#endif

    /*
     * Macro to compute ptr to end of loop.
     *    count1 = endP - startP;  count4 = (count1/4)*4;  rtnd_endP = startP + count4;
     * => (startP + (((endP-startP)/4)*4))
     */
#define LOOP_END_ADDR(_start,_end) (((long *)(_start)) + ((((int)(((long *)(_end))-((long *)(_start))))/4)*4))
    

    if (IS_DBG_PORT_PAGE) {
	char255 smsg;
	sprintf(smsg, "make_diff(addr=0x%x, twin=0x%x)",
		pageP->page_addr, pageP->twin_addr);
	StatusMsg(smsg);
    }
    INCR_STATS(PagedStats.make_diffs, 1);
    TRACE_3("DIFF", pageP->page_addr, pageP->twin_addr);

    
    /* 
     * Initialize local variables 
     */
    page_addr = (vm_offset_t) pageP->page_addr;
    diffP = pageP->diffP;
    page_dataP = (long *) page_addr;
    twin_dataP = (long *) pageP->twin_addr;
    page_endP  = (long *) (page_addr + vm_page_size);
    actual_data = 0; 
    
  
    /*
     * Outer loop checks for differences in data
     */
    loop_endP = LOOP_END_ADDR(page_dataP, page_endP);
    while (page_dataP < loop_endP) {

	page_data3 = page_dataP[3];
	twin_data3 = twin_dataP[3];
diff_outer_3:
	page_data2 = page_dataP[2];
	twin_data2 = twin_dataP[2];
diff_outer_2:
	page_data1 = page_dataP[1];
	twin_data1 = twin_dataP[1];
diff_outer_1:
	page_data0 = page_dataP[0];
	twin_data0 = twin_dataP[0];

	if ((page_data0 == twin_data0) &&
	    (page_data1 == twin_data1) &&
	    (page_data2 == twin_data2) &&
	    (page_data3 == twin_data3)) {
	    page_dataP += 4;
	    twin_dataP += 4;
	    continue;
	}

	/*
	 * Pages differ.
	 * Determine first word that has changed.
	 */
	if (page_data0 == twin_data0) {
	    if (page_data1 == twin_data1) {
		if (page_data2 == twin_data2) {
		    page_dataP += 3;
		    twin_dataP += 3;
		} else {
		    page_dataP += 2;
		    twin_dataP += 2;
		}
	    } else {
		page_dataP += 1;
		twin_dataP += 1;
	    }
	}

#if PSTATS > 0
	INCR_STATS(diff_count, 1);
#endif
	
	/* 
	 * Inner loop runs while data is different
	 */
	diff_offset = ((vm_offset_t) page_dataP) - page_addr;
	diffP->offset = diff_offset;
	loop_endP = LOOP_END_ADDR(page_dataP, page_endP);
	while (page_dataP < loop_endP) {
	    
	    page_data3 = page_dataP[3];
	    twin_data3 = twin_dataP[3];
diff_inner_3:
	    page_data2 = page_dataP[2];
	    twin_data2 = twin_dataP[2];
diff_inner_2:
	    page_data1 = page_dataP[1];
	    twin_data1 = twin_dataP[1];
diff_inner_1:
	    page_data0 = page_dataP[0];
	    twin_data0 = twin_dataP[0];

	    if ((page_data1 == twin_data1) ||
		(page_data2 == twin_data2) ||
		(page_data3 == twin_data3) ||
		(page_data0 == twin_data0))
		goto diff_inner_done;

	    page_dataP += 4;
	    twin_dataP += 4;

	} /* INNER LOOP */

	/* 
	 * Restart inner loop to handle loop unrolling
	 */
	if (page_dataP < page_endP) {
	    register int extra;

	    extra = page_endP - page_dataP;
	    ASSERT ((extra > 0) && (extra < 4));
	    loop_endP = page_endP;
	    twin_data3 = page_data3;
	    if (extra == 3) goto diff_inner_3;
	    twin_data2 = page_data2;
	    if (extra == 2) goto diff_inner_2;
	    twin_data1 = page_data1;
	    goto diff_inner_1;
	}

diff_inner_done:

	/*
	 * Data wasn't modified or we reached the end of the page.
	 * Determine first word that hasn't changed.
	 */
	if (page_dataP >= page_endP)
	    page_dataP = page_endP;
	else { 
	    if (page_data0 != twin_data0) {
		if (page_data1 != twin_data1) {
		    if (page_data2 != twin_data2) {
			page_dataP += 3;
			twin_dataP += 3;
		    } else {
			page_dataP += 2;
			twin_dataP += 2;
		    }
		} else {
		    page_dataP += 1;
		    twin_dataP += 1;
		}
	    }
	}

	/* determine length of this diff */
	diff_length = (((vm_offset_t)page_dataP) - page_addr) - diff_offset;
	diffP->length = diff_length;
	actual_data += diff_length;
	
	TRACE_3("Diff", page_addr + diff_offset, diff_length);
	
#ifdef PTRACE_PAGE
	if ((page_addr >= midTraceDetailLow) && (page_addr < midTraceDetailHigh)) {
	    TRACE_4("DPAG", page_addr + diff_offset, diff_length, 
		    *(long *)(page_addr + diff_offset));
	}
#endif

	ASSERT ((diff_length > 0) && (diff_length <= vm_page_size));
	ASSERT (diff_offset < vm_page_size);
	ASSERT ((char *)(page_addr + diff_offset) < (char *) page_endP);
	ASSERT ((char *)(page_addr + diff_offset + diff_length) <= (char *) page_endP);
	
	INCR_HISTO (PagedStats.make_diffsSize,PagedStats.make_diffsSizeH,
		    diff_length, STAT_DIFFSIZE_TABLE_SIZE, STAT_DIFFSIZE_PER_BUCKET);


	/* update diff pointer */
	diffP++;
	ASSERT (diffP <= &pageP->diffP[MAX_DIFFS_PER_PAGE]);
    
	/* recompute loop endP for outer loop */
	loop_endP = LOOP_END_ADDR(page_dataP, page_endP);

    } /* while page_dataP < loop_endP */

    /* 
     * Restart loop to handle loop unrolling
     */
    { register int extra;
      extra = page_endP - page_dataP;
      if (extra > 0) {
	  loop_endP = page_endP;
	  twin_data3 = page_data3;
	  if (extra == 3) goto diff_outer_3;
	  twin_data2 = page_data2;
	  if (extra == 2) goto diff_outer_2;
	  twin_data1 = page_data1;
	  goto diff_outer_1;
      }
    }

    INCR_HISTO (PagedStats.make_diffsSizePage,PagedStats.make_diffsSizePageH,
		actual_data, STAT_DIFFSIZE_TABLE_SIZE, STAT_DIFFSIZE_PER_BUCKET);


    
    /* return amount of data in buffer & total size of buffer */
    pageP->dirty_len = diffP - pageP->diffP;
    pageP->dirty_data = actual_data;
    pageP->dirty_unsorted = 0;
    pageP->clean_len = 0;
    pageP->clean_data = 0;
    pageP->diff_reuse = 0;
    TRACE_4("DiSz", page_addr, pageP->dirty_data, pageP->dirty_len);
    
} /* make_diff */

extern void midway_make_diff (page_desc_t *pageP)
{
    make_diff (pageP);
}

int
mid_debug_diff_pages (vm_offset_t page_addr, vm_offset_t twin_addr)
{
    long i, *dataP;
    vm_size_t buf_size, buf_data;
    page_desc_t page;
    diff_desc_t *diffP, diff_buf[MAX_DIFFS_PER_PAGE];

    page.page_addr = page_addr;
    page.twin_addr = twin_addr;
    page.state = PAGE_FREE;
    page.dirty_len = 0;
    page.dirty_data = 0;
    page.dirty_unsorted = 0;
    page.clean_len = 0;
    page.clean_data = 0;
    page.diffP = diff_buf;
    page.cleanP = NULL;
    page.nextP = NULL;

    make_diff(&page);

    printf("make_diff(0x%x, 0x%x) returns size=%lu data=%lu\n", 
	   page_addr, twin_addr, page.dirty_len, page.dirty_data);
    for (diffP = page.diffP; diffP < &page.diffP[ page.dirty_len ]; diffP++) {
	printf("Diff: off=0x%x len=%u\n", diffP->offset, diffP->length);
	dataP = (long *)(page_addr + diffP->offset);
	for (i = 0; i < diffP->length; i += 4) {
	    printf(" %08x", *dataP);
	    if ((i % 32) == 0) printf("\n");
	    dataP++;
	}
	printf("\n");
    }
    printf("\n");
}

static void
update_diff (page_desc_t *pageP,	/* ptr to page desc */
	     diff_desc_t *diffP,	/* ptr to dirty diff being cleaned */
	     vm_offset_t clean_addr,	/* starting address being cleaned */
	     vm_offset_t clean_size	/* length being cleaned */
	     )
{
    int clen, dlen;
    vm_offset_t page_addr;
    vm_offset_t clean_end;
    vm_offset_t diff_start, diff_end;

    /*
     * Mark this part of the page "clean".
     * This is by updating the clean list in the page descriptor
     * and removing this diff from the dirty list.
     */
    
    /* Add diff to clean list */
    clen = pageP->clean_len;
    page_addr = pageP->page_addr;
    ASSERT (clen < MAX_DIFFS_PER_PAGE);
    pageP->cleanP[clen].offset = clean_addr - page_addr;
    pageP->cleanP[clen].length = clean_size;
    pageP->clean_len = clen + 1;
    pageP->clean_data += clean_size;
    ASSERT (pageP->clean_data <= DIFF_PAGE_SIZE);

    /* read original dirty diff */
    diff_start = page_addr + diffP->offset;
    diff_end = diff_start + diffP->length;

    ASSERT ((clean_addr >= diff_start) &&
	    (clean_addr < diff_end));
    ASSERT (((clean_addr+clean_size) > diff_start) &&
	    ((clean_addr+clean_size) <= diff_end));
    ASSERT (diffP->length >= clean_size);
    ASSERT (pageP->dirty_data >= clean_size);
    ASSERT (pageP->dirty_len > 0);

    /* Remove diff from dirty list */
    pageP->dirty_data -= clean_size;
    clean_end = clean_addr + clean_size;

    if (clean_addr == diff_start) {
	if (clean_end == diff_end) {
	    diffP->length = 0;		/* entire diff has been copied */
	    INCR_STATS(PagedStats.copy_dirtyCountComp, 1);
	} else {
	    diffP->offset = clean_end - page_addr;	/* first part of diff has been copied */ 
	    diffP->length -= clean_size;
	    INCR_STATS(PagedStats.copy_dirtyCountFirst, 1);
	}
    } else {
	if (clean_end == diff_end) {
	    diffP->length -= clean_size;	/* last part of diff has been copied */ 
	    INCR_STATS(PagedStats.copy_dirtyCountLast, 1);
	} else {
	    /* Middle of diff has been copied.
	     * Add a new diff at the end & shorten this diff 
	     */
	    dlen = pageP->dirty_len;
	    ASSERT (dlen < MAX_DIFFS_PER_PAGE);
	    pageP->diffP[dlen].offset = clean_end - page_addr;
	    pageP->diffP[dlen].length = diff_end - clean_end;
	      
	    /* By adding to end, we've made the diff list unsorted */
	    if (diffP != &pageP->diffP[ dlen-1 ]) {
		pageP->dirty_unsorted = 1;
	    }
	    pageP->dirty_len = dlen + 1;

	    /* shorten the current diff */
	    diffP->length = clean_addr - diff_start;
	    ASSERT ((diffP->length > 0) && (diffP->length < vm_page_size));
	    INCR_STATS(PagedStats.copy_dirtyCountMiddle, 1);

	    TRACE_6("DCsp", pageP->dirty_len, pageP->dirty_data, 
		    pageP->clean_len, pageP->clean_data, pageP->dirty_unsorted);
	    TRACE_5("dcsp", diffP->offset+page_addr, diffP->length,
		    pageP->diffP[dlen].offset+page_addr, pageP->diffP[dlen].length);
	}
    }

}    /* update_diff */

static void 
locate_diffs (page_desc_t *pageP,
	      vm_offset_t data_addr,
	      vm_size_t   data_size
	      )
{
    vm_offset_t page_addr;
    vm_offset_t end_addr;
    diff_desc_t *diffP, *diff_endP;
    vm_offset_t diff_start, diff_end;
    vm_offset_t low_addr, high_addr;
    vm_size_t size;

    /* 
     * Locate & clean all diffs for the specified area 
     * within the given page.
     */
    if (IS_DBG_LOCK_MLPD) {
	char255 smsg;
	sprintf(smsg, "locate diff for 0x%x+%d, dlen=%d ddata=%d",
		data_addr, data_size, pageP->dirty_len, pageP->dirty_data);
	StatusMsg(smsg);
    }
    TRACE_6("DRld", data_addr, data_size, 
	    pageP->dirty_len, pageP->dirty_data, pageP->dirty_unsorted);
    ASSERT (pageP->state == PAGE_DIFFED);

    /* 
     * Loop over each diff descriptor in this page's diff buffer
     */
    diffP = pageP->diffP;
    diff_endP =  &diffP[ pageP->dirty_len ];
    page_addr = pageP->page_addr;
    end_addr = data_addr + data_size;

    while (diffP < diff_endP) {

	diff_start = page_addr + diffP->offset; 
	diff_end = diff_start + diffP->length;

	/* check if addrpair is within requested data range */
	if (diff_start >= end_addr) {
	    /* Verify that the diffs are sorted before exiting */
	    if (pageP->dirty_unsorted == 0) break;
	    goto next_diff;
	}

	/* Check if the diff is before the requested address */
	if (diff_end <= data_addr)
	    goto next_diff;

	/* check for a deleted diff */
	if (diff_start == diff_end)
	    goto next_diff;

	/* Diff falls within requested range
	 * Find exact range that it covers 
	 */
	low_addr = MAX(data_addr, diff_start);
	high_addr = MIN(end_addr, diff_end);
	ASSERT (high_addr > low_addr);

	/* Mark this part of the page "clean". */
	size = high_addr - low_addr;
	update_diff (pageP, diffP, low_addr, size);
	
	/* reload diff_endP, just in case it changed */
	diff_endP = &pageP->diffP[ pageP->dirty_len ];

next_diff:
	/* advance to next diff */
	diffP++;
    } /* while diffP < diff_endP */

} /* locate_diffs */

static void 
copy_raw_data (
	       page_desc_t	*pageP,		/* ptr to page desc or NULL */
	       vm_offset_t  data_addr,	/* starting address for copy */
	       vm_size_t    data_len,		/* length */
	       int          mark_clean,		/* =1 if data is to be marked clean */
	       addrpair_t **bufPP,		/* in/out ptr to data buffer */
	       vm_size_t   *buf_lenP,		/* in/out free space in data buffer */
	       vm_size_t   *copy_lenP,	/* out: actual data copied */
	       int         *aseg_cntP)	/* out: count of addrpairs added */
{
    register vm_size_t buf_len = *buf_lenP;
    register addrpair_t *out_pairP = *bufPP;
    vm_size_t word_len;
    vm_offset_t twin_addr;


    if (IS_DBG_PORT_PAGE) {
	char255 smsg;
	sprintf(smsg, "copy_raw_data(addr=0x%x+%lu clean=%d OutBufP=0x%x+%lu)",
		data_addr, data_len, mark_clean, out_pairP, *buf_lenP);
	StatusMsg(smsg);
    }
    INCR_STATS(PagedStats.copy_data, 1);
    ASSERT (buf_len > (sizeof(*out_pairP) - sizeof(out_pairP->data)));
    ASSERT ((pageP == NULL) || (pageP == PAGE_DESC(data_addr)));
    ASSERT ((pageP == NULL) || (pageP->page_addr == PAGE_ADDR(data_addr)));
    ASSERT (data_len <= PAGE_REMAINS(data_addr));
    

    /* Determine how much data will fit in the out buffer */
    buf_len -= (sizeof(*out_pairP) - sizeof(out_pairP->data));
    if (data_len > buf_len) data_len = buf_len;
    
    out_pairP->addr = (char *) data_addr;
    out_pairP->len = data_len;

    if (IS_DBG_LOCK_MLPD) {
	char255 smsg;
	sprintf(smsg, "paged_copy_raw_data: copying %lx+%ld  to 0x%lx", 
		data_addr, data_len, out_pairP->data);
	StatusMsg(smsg);
    }
    TRACE_4("DRaw", data_addr, data_len, mark_clean);
    INCR_STATS(PagedStats.copy_dataSize, data_len);

    bcopy (data_addr, 		/* source */
	   out_pairP->data, 	/* target */
	   data_len); 		/* length */


    /*
     * Set return values for parms 
     */
    *copy_lenP = data_len;
    word_len = (data_len + 3) & ~3; /* round up to word boundary */
    *buf_lenP = (buf_len - word_len);
    *bufPP    = (addrpair_t *) &out_pairP->data[word_len];
    *aseg_cntP = 1;


    /*
     * Check if caller wants to mark data as clean
     */
    if (mark_clean == 0) return;


    /*
     * Update page descriptor for the data just copied.
     * If the page is marked dirty, we must mark the
     * copied area as clean
     * Note: Caller must check if a previously dirty page
     * is completely cleaned by this call.
     */
	
    /* check if the page is or was dirty */
    if (IS_PAGE_DIRTY(data_addr)) {
	
	/* locate page descriptor, if necessary */
	if (pageP == NULL) 
	    pageP = PAGE_DESC(data_addr);
	
	ASSERT (pageP->page_addr == PAGE_ADDR(data_addr));
	
	if (pageP->state == PAGE_DIRTY) {
	    
	    /* 
	     * Page is dirty.
	     * mark area clean by updating the twin 
	     */
	    if (IS_DBG_LOCK_MLPD) {
		char255 smsg;
		sprintf(smsg, "raw: updating twin data for 0x%x+%d",
			data_addr, data_len);
		StatusMsg(smsg);
	    }
	    TRACE_3("DRtw", data_addr, data_len);
	    INCR_STATS(PagedStats.copy_dataSize, data_len);

	    twin_addr = pageP->twin_addr + 
		(((vm_offset_t)data_addr) - ((vm_offset_t) PAGE_ADDR(data_addr)));
	    
	    bcopy (data_addr,	/* source */
		   twin_addr, 	/* target */
		   data_len);	/* length */
	    
	} else {
	    
	    /* 
	     * Page has been diffed.
	     * Mark area clean by updating dirty & clean lists
	     */
	    locate_diffs (pageP, data_addr, data_len);
	}
    } /* if (IS_PAGE_DIRTY()) */

    return;
} /* copy_raw_data */
		

static void 
copy_dirty_data (
		 page_desc_t  *pageP, 		/* ptr to page descriptor */
		 vm_offset_t  data_addr,	/* starting address for copy */
		 vm_size_t    data_len,		/* length */
		 addrpair_t **bufPP,		/* in/out ptr to data buffer */
		 vm_size_t   *buf_lenP,		/* in/out free space in data buffer */
		 vm_size_t   *addr_space_copiedP,	/* out: address space "copied" */
		 vm_size_t   *user_data_copiedP,	/* out: actual user data copied */
		 int	     *aseg_cntP)	/* out: count of addrpairs added */
{
    register vm_offset_t page_addr = pageP->page_addr;
    register vm_offset_t end_addr = data_addr + data_len;
    register vm_size_t buf_len = *buf_lenP;
    register diff_desc_t *diffP = pageP->diffP;
    register diff_desc_t *diff_endP = &diffP[ pageP->dirty_len ];
    register addrpair_t *out_pairP = *bufPP;
    register addrpair_t *out_endP  = (addrpair_t *)(((vm_offset_t)out_pairP) + *buf_lenP);
    register vm_offset_t diff_start, diff_end;
    register vm_offset_t low_addr, high_addr;
    register vm_size_t size, user_data_copied, addr_space_copied;
    register vm_offset_t orig_data_addr = data_addr;
    register int aseg_cnt;

    if (IS_DBG_PORT_PAGE) {
	char255 smsg;
	sprintf(smsg, "copy_dirty_data(addr=0x%x+%lu Dirty=%lu Clean=%lu OutBufP=0x%x+%lu)",
		data_addr, data_len, pageP->dirty_data, pageP->clean_data,
		out_pairP, *buf_lenP);
	StatusMsg(smsg);
    }
    INCR_STATS(PagedStats.copy_dirty, 1);
    TRACE_6("DCpy", data_addr, data_len, pageP->dirty_len,
	    pageP->dirty_data, pageP->dirty_unsorted);

    ASSERT (pageP->diffP != NULL);
    ASSERT (pageP->cleanP != NULL);

    /* 
     * Loop over each diff descriptor in this page's diff buffer
     */
    user_data_copied = 0;
    addr_space_copied = data_len; /* assume all diff data will fit in buffer */
    aseg_cnt = 0;
    while (diffP < diff_endP) {

	diff_start = page_addr + diffP->offset; 
	diff_end = diff_start + diffP->length;

	/* check if the diff is past the requested address  */
	if (diff_start >= end_addr) {

	    /* Verify that the diffs are sorted before exiting */
	    if (pageP->dirty_unsorted == 0) break;
	    goto next_diff;
	}

	/* Check if the diff is before the requested address */
	if (diff_end <= data_addr)
	    goto next_diff;

	/* check for a deleted diff */
	if (diff_start == diff_end)
	    goto next_diff;

	/* Diff falls within requested range
	 * Find exact range that it covers 
	 */
	low_addr = MAX(data_addr, diff_start);
	high_addr = MIN(end_addr, diff_end);
	ASSERT (high_addr > low_addr);
	    
	/* Check if there is space in the output buffer */
	if (buf_len < sizeof(*out_pairP)) {

	    /* Check if the diffs were sorted */
	    if (pageP->dirty_unsorted == 0) {

		/* Yes, diffs were sorted.
		 * Advance data addr as far as possible 
		 */
		addr_space_copied = low_addr - orig_data_addr;
	    } else {

		/*
		 * Diffs were not in sorted order.
		 * Can't advance address space covered
		 * since there may be later diffs with 
		 * a lower address. Just return a 0 length.
		 * The next call will search over the same range,
		 * but only return dirty data that hasn't yet 
		 * been cleaned.   WAS 3Sep94
		 */
		addr_space_copied = 0;
	    }
	    break;
	}
	buf_len -= (sizeof(*out_pairP) - sizeof(out_pairP->data));
	    
	size = high_addr - low_addr;
	if (size > buf_len) size = buf_len;
	ASSERT (size > 0);
	
	out_pairP->addr = (char *) low_addr;
	out_pairP->len = size;
	aseg_cnt++;
	
	if (IS_DBG_LOCK_MLPD) {
	    char255 smsg;
	    sprintf(smsg, "mid_paged_copy_dirty_data: copying %lx+%ld  to 0x%lx", 
		    low_addr, size, out_pairP->data);
	    StatusMsg(smsg);
	}
	TRACE_3("DCdf", low_addr, size);
	
	bcopy (low_addr, 		/* source */
	       out_pairP->data, 	/* target */
	       size); 		/* length */
	
	buf_len -= size;
	user_data_copied += size;
	out_pairP = (addrpair_t *) &out_pairP->data[size];
	
	INCR_STATS(PagedStats.copy_dirtyCount, 1);
	INCR_STATS(PagedStats.copy_dirtySize, size);
	
	/***
	 *** Mark this part of the page "clean".
	 *** This is by updating the clean list in the page descriptor
	 *** and removing this diff from the dirty list.
	 ***/
	update_diff (pageP, diffP, low_addr, size);

	/* Check for a full buffer */
	if (buf_len < sizeof(*out_pairP)) {
	    if (pageP->dirty_unsorted == 0)
		addr_space_copied = (low_addr+size) - orig_data_addr;
	    else
		addr_space_copied = 0;
	    break;
	}
	    
	/* reload ptr to end of diffs, in case it changed */
	diff_endP = &pageP->diffP[ pageP->dirty_len ];
	

next_diff:
	/* advance to next diff */
	diffP++;
    } /* while diffP < diff_endP */

    TRACE_6("dcpy", buf_len, data_len, user_data_copied, 
	    pageP->dirty_data, pageP->clean_data);

    *buf_lenP = buf_len;
    *bufPP    = out_pairP;
    *addr_space_copiedP = addr_space_copied;
    *user_data_copiedP = user_data_copied;
    *aseg_cntP = aseg_cnt;
    return;
}
		

static void apply_clean_diffs (page_desc_t *pageP)
{
    register int clen;
    register vm_offset_t page_addr, twin_addr, offset, length;
    register diff_desc_t *diffP, *endP;

    if (IS_DBG_PORT_PAGE) {
	char255 smsg;
	sprintf(smsg, "apply_clean_diffs(page=0x%x twin=0x%x len=%d data=%d)",
		pageP->page_addr, pageP->twin_addr, pageP->clean_len, pageP->clean_data);
	StatusMsg(smsg);
    }
    INCR_STATS(PagedStats.apply_clean, 1);
    TRACE_5("DApl", pageP->page_addr, pageP->twin_addr, pageP->clean_len, pageP->clean_data);
    INCR_HISTO (PagedStats.apply_cleanSizePage,PagedStats.apply_cleanSizePageH,
		pageP->clean_data, STAT_DIFFSIZE_TABLE_SIZE, STAT_DIFFSIZE_PER_BUCKET);


    
    /* NOTE:
     * If there is no dirty data on the page
     * it may be cheaper to simply copy the entire page
     * than to apply the "cleaned" diffs to the old twin.
     */

    clen = pageP->clean_len;
    if (clen > 0) {
	
	diffP = pageP->cleanP;	/* apply the "clean" diffs */
	endP = &diffP[clen];
	page_addr = pageP->page_addr;
	twin_addr = pageP->twin_addr;
	    
	/* Loop applying each "clean" diff to the twin */
	while (diffP < endP) {
	    
	    offset = diffP->offset;
	    length = diffP->length;
	    
#ifdef DETAILS
	    if (IS_DBG_LOCK_MLPD) {
		char255 smsg;
		sprintf(smsg, "mid_paged_apply_clean_diff: copying %lx+%ld  to 0x%lx", 
			page_addr+offset, length, twin_addr+offset);
		StatusMsg(smsg);
	    }
	    TRACE_3("DAcd", page_addr+offset, length);
#endif

	    bcopy (page_addr+offset,	/* source */
		   twin_addr+offset,  	/* target */
		   length); 		/* length */
	    
	    INCR_STATS(PagedStats.apply_cleanDiffs, 1);
	    INCR_HISTO (PagedStats.apply_cleanSize,PagedStats.apply_cleanSizeH,
			length, STAT_DIFFSIZE_TABLE_SIZE, STAT_DIFFSIZE_PER_BUCKET);

	    diffP++;
	}

	/* reset clean lists before leaving */
	pageP->clean_len = 0;
	pageP->clean_data = 0;
    }
    
    /* invalidate the saved dirty diff */
    pageP->dirty_len = 0;
    pageP->dirty_data = 0;
    pageP->dirty_unsorted = 0;

    return;
}



/*
 ************************************************************************
 *									*
 *	Externally callable routines              			*
 *									*
 ************************************************************************
 */

/* return 1 if ran out of buffer before addrmap, 0 if done */
int
mid_paged_pack_data 
    (
     u_long		synch_obj,	/* used for debugging */
     int		pack_diffs, 	/* =1 to pack diffs, =0 to pack all data  */
     int                mark_clean,     /* =1 to mark data as clean */
     s_addr_map_t	**amapPP,	/* the map in/out -- in case more needed */
     char         	**amap_addrPP,	/* in/out -- ptr to uncopied portion of map */
     long          	*amap_lenP,  	/* in/out -- uncopied length of map */
     clock_t		 clock_val,     /* in -- lock's incarnation number */
     char		*pbufP,		/* pointer into packetbuffer */
     long		*pbufLenP,	/* length of stuff already in buffer */
     long		*xfer_lenP 	/* actual amount xferred */
     )
{
    int           rc;
    s_addr_map_t *amapP;
    vm_offset_t   amap_addr;
    long          amap_len;
    addrseg_t    *asegP;
    addrpair_t   *apairP;
    long          pblen, pbleft;
    vm_size_t     addr_space_copied, total_user_data_copied, user_data_copied;
    int           aseg_num, aseg_cnt;
    vm_offset_t   data_addr;
    long          data_len;
    page_desc_t  *pageP;
    

    /* cache args */
    amapP = *amapPP;
    amap_addr = (vm_offset_t) *amap_addrPP;
    amap_len = *amap_lenP;
    pblen = *pbufLenP;
    
    if (IS_DBG_PORT_PAGE) {
	char255 smsg;
	sprintf(smsg,"mid_paged_pack_data(%s,%s,0x%x,0x%x,%lu,%lu,0x%x,%lu)", 
		(pack_diffs) ? "diffs" : "data",
		(mark_clean) ? "clean" : "raw", 
		amapP, amap_addr, amap_len, clock_val, pbufP, pblen);
	StatusMsg(smsg);
    }
    TRACE_5("PACK", synch_obj, pack_diffs, clock_val, mark_clean);
    TRACE_4("AMAP", synch_obj, amap_addr, amap_len);
#ifdef STATS
    if (pack_diffs)
	INCR_STATS(PagedStats.pack_diffCalls, 1);
    else
	INCR_STATS(PagedStats.pack_dataCalls, 1);
#endif

    rc = 0;			/* 0 = no more work to do */
    asegP = NULL;
    apairP = (addrpair_t *) pbufP;
    total_user_data_copied = 0;
    
    if (amapP == NULL)
	goto done;

    /*
     * Acquire exclusive access to page structures
     */
    mutex_lock(paged_mutex);

    pbleft = (mid_max_data_size - pblen) & ~3;

    /* initialize address segment. There is only 1 per buffer */
    asegP = (addrseg_t *) pbufP;
    asegP->segtype = DIFFSEG;
    asegP->clock = clock_val;
    asegP->num = -1;
    aseg_num = 0;
    aseg_cnt = 0;
    apairP = &(asegP->ap);
    pbleft -= ADDRSEG_SIZE;
    pageP = NULL;
    data_len = 0;
    data_addr = 0;

    /* 
     * Check data covered by amaps
     */
    while (pbleft >= sizeof(addrpair_t)) {

	/* 
	 * locate next page to copy
	 */
	while (data_len == 0) {

	    /* advance to next address map, if necessary */
	    if (amap_len == 0) {
		if (amap_addr != 0) {	/* check if this is the 1st call */
		    amapP = amapP->next;
		    if (amapP == NULL) goto done; /* no more work to do */
		}
		amap_addr = (vm_offset_t) amapP->addr;
		amap_len = amapP->len;
		TRACE_4("AMAP", synch_obj, amap_addr, amap_len);
		ASSERT (amap_len > 0);
	    }

	    /* determine the amount of data on this page */
	    pageP = NULL;
	    data_len = PAGE_REMAINS(amap_addr);
	    if (amap_len < data_len) data_len = amap_len;
	    
	    /* Check if the amap is within the paged region
	     * Static data is packed by this code as well.
	     */
	    if (amap_addr < vm_paged_base) {
		
		/* Handle statics */
		ASSERT ((amap_addr >= vm_static_base) && 
			(amap_addr < vm_static_end));
		ASSERT (((amap_addr+amap_len) > vm_static_base) && 
			((amap_addr+amap_len) <= vm_static_end));
		ASSERT (!pack_diffs);

		/* Just move it */
		break;
	    }

	    /* amap is within paged region */
	    ASSERT ((amap_addr >= vm_paged_base) && 
		    (amap_addr < vm_paged_end));
	    ASSERT (((amap_addr+amap_len) > vm_paged_base) && 
		    ((amap_addr+amap_len) <= vm_paged_end));

	    
	    /* check if the page has ever been allocated */
	    if (!IS_PAGE_ALLOCATED(amap_addr)) {
		
		/* Page is not allocated -- nothing to copy */
		INCR_STATS(PagedStats.pack_unallocSize, data_len);
		amap_addr += data_len;
		amap_len -= data_len;
		data_len = 0;

		continue;
	    }
		
	    /* Page is allocated, check if it is clean */
	    if (IS_PAGE_CLEAN(amap_addr)) {
		INCR_STATS(PagedStats.pack_cleanSize, data_len);
		
		/* No need to diff a clean page */
		if (pack_diffs) {
		    amap_addr += data_len;
		    amap_len -= data_len;
		    data_len = 0;
		}
		continue;
	    }

	    /* Page is or was dirty */
	    ASSERT (IS_PAGE_DIRTY(amap_addr));
	    
	    pageP = PAGE_DESC(amap_addr);
	    ASSERT (pageP->page_addr == PAGE_ADDR(amap_addr));

#ifdef STATS
	    if (pageP->state == PAGE_DIRTY) {
		INCR_STATS(PagedStats.pack_dirtySize, data_len);
	    } else {
		INCR_STATS(PagedStats.pack_diffedSize, data_len);
	    }
#endif
	    
	} /* while data_len == 0 */ 

	
	/*
	 * We now have a data address & length
	 * Send diffs or all of the data
	 */
	data_addr = amap_addr;
	if (pack_diffs == 0) {

	    /* Copy all data for the range [data_addr to data_addr+data_len]  */
	    copy_raw_data (pageP, data_addr, data_len, mark_clean,
			   &apairP, &pbleft, &addr_space_copied, &aseg_cnt);
	    total_user_data_copied += addr_space_copied; /* = user_data_copied for raw copies */

	} else {

	    /* verify that we are cleaning the page */
	    ASSERT (mark_clean != FALSE);

	    /* check if page needs to be diffed */
	    if (pageP->state == PAGE_DIRTY) {

		/* Write protect the page */
		page_protect (PAGE_ADDR(amap_addr), READONLY_ACCESS);
		    
		/* diff the page */
		make_diff (pageP);

		pageP->state = PAGE_DIFFED;
		INCR_STATS(PagedStats.pack_diffsMade, 1);
	    } else {
		ASSERT (pageP->state == PAGE_DIFFED);
		INCR_STATS(PagedStats.pack_diffsReused, 1);
	    }
	    INCR_STATS(pageP->diff_reuse, 1);

	    if (pageP->dirty_data > 0) {

		/* Copy diff's for the range [data_addr to data_addr+data_len] 
		 * This will also adjust the page's dirty_data counter.
		 */
		copy_dirty_data (pageP,
				 data_addr, data_len, 
				 &apairP, &pbleft,
				 &addr_space_copied, &user_data_copied, &aseg_cnt);
		total_user_data_copied += user_data_copied;
	    } else {
		
		addr_space_copied = data_len;
		aseg_cnt = 0;
	    }
	} /* else pack_diffs was not 0 */
	
	/* Check if we have transferred all of the dirty data from the page */
	if ((pageP != NULL) &&
	    (pageP->state == PAGE_DIFFED) &&
	    (pageP->dirty_data == 0)) {
	    
	    /* Mark the page as clean & free the page descriptor */
	    if (IS_DBG_PORT_PAGE) {
		char255 smsg;
		sprintf(smsg, "clean_page(page=0x%x)",  pageP->page_addr);
		StatusMsg(smsg);
	    }
	    TRACE_2("ClPg", pageP->page_addr);
	    INCR_STATS(PagedStats.pack_cleaned, 1);
	    INCR_HISTO (PagedStats.diff_reuseCleanTotal,PagedStats.diff_reuseCleanH,
			pageP->diff_reuse, STAT_DIFFREUSE_TABLE_SIZE, 
			STAT_DIFFREUSE_PER_BUCKET);
	    
	    /* free this page descriptor */
	    MARK_PAGE_CLEAN (pageP->page_addr);
	    free_page_desc (pageP);
	    
	    /* find new page descriptor */
	    ASSERT (data_len >= addr_space_copied);
	    data_len = addr_space_copied;
	    pageP = NULL;
	}
		
	/* update data/amap address */
	data_addr += addr_space_copied;
	data_len -= addr_space_copied;
	amap_addr += addr_space_copied;
	amap_len -= addr_space_copied;
	aseg_num += aseg_cnt;

	/*verify that we are making progress */
	ASSERT ((addr_space_copied > 0) ||
		(pbleft < sizeof(addrpair_t)));

	if (addr_space_copied == 0) break;
	
	ASSERT (pbleft >= 0);
	ASSERT (data_len >= 0);
	ASSERT (amap_len >= 0);
	ASSERT (amap_len >= data_len);

    } /* while pbleft >= min size */

    rc = 1;			/* 1= more work to do */

done:     

    /* Release mutex */
    mutex_unlock(paged_mutex);
    
    
    /* set segment count in final seg addr */
    if (asegP != NULL)
	asegP->num = aseg_num;
    
    /* compute size of returned buffer */
    pblen += ((char *)apairP) - pbufP;
    ASSERT (pblen <= mid_max_data_size);
    ASSERT ((pblen & 3) == 0);
    
    /* return values for subsequent calls */
    *amapPP = amapP;
    *amap_addrPP = (char *) amap_addr;
    *amap_lenP = amap_len;
    *pbufLenP = pblen;
    *xfer_lenP = total_user_data_copied;

#ifdef STATS 
    if (pack_diffs) 
	INCR_STATS(PagedStats.pack_diffSize, total_user_data_copied);
    else 
	INCR_STATS(PagedStats.pack_dataSize, total_user_data_copied);
#endif 
    TRACE_5("pack", synch_obj, rc, pblen, total_user_data_copied);
    
    return(rc);
}


/*
 ************************************************************************
 *		m i d _ p a g e d _ u n p a c k _ d a t a		*
 *----------------------------------------------------------------------*
 * Used to unpack diffsegs in a comm buffer -- the opposite of  	*
 * mid_pack_data.  This is used by the lock reply, release reply        *
 * and barrier reply.  Here be dirtybit updates! (dragons?)		*
 ************************************************************************
 */
void
mid_paged_unpack_data (
		       u_long	synch_obj,	/* used for debugging only */
		       char 	**pptr,
		       u_long 	*lptr,
		       s_addr_map_t *amap_listP,	/* DEBUG only: addr maps for this lock/barrier */
		       char 	*cname,		/* caller for debug messages */
		       MUD_t 	ctype 		/* another way to interpret the caller */
		       )
{
    int		i;
    u_long	length;		/* length of packet pointed to by pptr */
    addrseg_t	*segP;		/* data segments, if any */
    addrpair_t    *pairP;		/* address/length/data triples, really */
    addrpair_t    *nextP;
    long          seg_len;	/* size of this data segment */
    

    segP = (addrseg_t *) (*pptr);	/* we should be given start of segments */
    length = *lptr;
    
    if (IS_DBG_PORT_PAGE) {
	char255	smsg;
	sprintf(smsg, "mid_paged_UNPACK_data[%s]: buf at 0x%x length %u", 
		cname, segP, length);
	StatusMsg(smsg);
    }
    INCR_STATS(PagedStats.unpack_calls, 1);
    
    /* acquire exclusive access to page tables*/
    mutex_lock (paged_mutex);


#ifdef MULTIPLE_VM_MAPS
	if (last_page_created != 0) {
	    vm_page_protect(WRITE_ADDR(last_page_created), READWRITE_ACCESS);
	    last_page_created = 0;
	}
#endif


    while (length) {
	if (segP->segtype != DIFFSEG) {	/* bail so upper level can figure this out */

	    *pptr = (char *)segP;	/* note where we left off */
	    *lptr = length;		/* and how much was left */
	    goto done;
	}

	/* Handle a DIFFSEG */
	pairP = &(segP->ap);
	ASSERT (length >= ((u_long)pairP - (u_long)segP));
	length -= ((u_long)pairP - (u_long)segP); /* account for addrseg header */

	TRACE_4("UNPK", synch_obj, segP->clock, pairP->addr);

	for(i = 0; i < segP->num; i++) {
	    
	    /* verify the length before copying the data */
	    nextP = (addrpair_t *) &pairP->data[ pairP->len ];
	    seg_len = (u_long)nextP - (u_long)pairP;
	    ASSERT (length >= seg_len);
	    length -= seg_len;
	    
	    
	    { register vm_offset_t data_addr = (vm_offset_t) pairP->addr;
	      register vm_size_t data_len = pairP->len;
	      register vm_offset_t source_addr = (vm_offset_t) pairP->data;
	      register vm_offset_t twin_addr;
	      register vm_size_t copy_len;
	      register page_desc_t *pageP;
	      register int rc;
	      
	      while (data_len > 0) {
		  
		  /* Check if update is for static data */
		  if (data_addr < vm_paged_base) {
		      
		      ASSERT ((data_addr >= vm_static_base) &&
			      (data_addr < vm_static_end));
		      ASSERT (((data_addr+data_len) > vm_static_base) &&
			      ((data_addr+data_len) <= vm_static_end));
		      TRACE_3("USta", data_addr, data_len);
		      
		      bcopy (source_addr,     	/* source */
			     data_addr,		/* target */
			     data_len);		/* length */

		      INCR_STATS(PagedStats.unpack_staticCount, 1);
		      INCR_STATS(PagedStats.unpack_staticSize, data_len);
		      break;
		  }
		  
		  /* Has this page been allocated */
		  if (!IS_PAGE_ALLOCATED(data_addr)) {
		      
		      /* Page not yet allocated.
		       * Release mutex & read from the page
		       */
		      TRACE_2("UALL", data_addr);
		      mutex_unlock(paged_mutex);
		      
		      rc = *(volatile int *)data_addr; /* read fault on page to allocate it */
		      
		      mutex_lock (paged_mutex);
		      TRACE_2("Uall", data_addr);
		      ASSERT (IS_PAGE_ALLOCATED(data_addr));
		      INCR_STATS(PagedStats.unpack_pagesAlloc, 1);
		  }
		  
		  copy_len = PAGE_REMAINS(data_addr);
		  if (data_len < copy_len) copy_len = data_len;
		  
#ifdef PTRACE_PAGE
		  if ((data_addr >= midTraceDetailLow) && (data_addr < midTraceDetailHigh)) {
		      TRACE_4("UNPA", WRITE_ADDR(data_addr), copy_len, *(long *)source_addr);
		  }
#endif
		  
		  /* Check if the page is clean */
		  if (IS_PAGE_CLEAN(data_addr)) {
		      
		      if (IS_DBG_LOCK_MLPD) {
			  char255 smsg;
			  sprintf(smsg, "mid_paged_UNPACK_data: copying to %lx+%ld from 0x%lx", 
				  WRITE_ADDR(data_addr), copy_len, source_addr);
			  StatusMsg(smsg);
		      }

#if !defined(MULTIPLE_VM_MAPS)
		      /* allow write to page so we can update it.
		       * Not a problem for EC programs
		       * Application is stopped waiting for lock xfer
		       */
		      page_protect (WRITE_ADDR(PAGE_ADDR(data_addr)), READWRITE_ACCESS);
		      INCR_STATS(PagedStats.protect_mmRWcalls,1);
#endif 

		      TRACE_3("Ucln", WRITE_ADDR(data_addr), copy_len);

		      /* copy data to page, using read/write mapping */
		      bcopy (source_addr,     		/* source */
			     WRITE_ADDR(data_addr),	/* target's write mapped page */
			     copy_len);			/* length */

#if !defined(MULTIPLE_VM_MAPS)
		      page_protect (WRITE_ADDR(PAGE_ADDR(data_addr)), READONLY_ACCESS);
		      INCR_STATS(PagedStats.protect_mmROcalls,1);
#endif 

		      INCR_STATS(PagedStats.unpack_cleanCount, 1);
		      INCR_STATS(PagedStats.unpack_cleanSize, copy_len);
		      
		  } else {
		      
		      /* get page descriptor for dirty buffer */
		      ASSERT (IS_PAGE_DIRTY(data_addr));
		      pageP = PAGE_DESC(data_addr);
		      
		      ASSERT (pageP != NULL);
		      ASSERT (pageP->page_addr == PAGE_ADDR(data_addr));

		      /*
		       * Check if page was diffed.
		       */
		      if (pageP->state == PAGE_DIFFED) {

			  INCR_HISTO (PagedStats.diff_reuseUpdateTotal,
				      PagedStats.diff_reuseUpdateH,
				      pageP->diff_reuse, STAT_DIFFREUSE_TABLE_SIZE, 
				      STAT_DIFFREUSE_PER_BUCKET);

			  /*
			   * Since the page is marked as "diffed", there is still
			   * modified data on the primary and we cannot discard
			   * the twin. Apply the clean diffs to the twin, then change
			   * the page's status to dirty and apply the update.
			   */
			  apply_clean_diffs (pageP);

			  /* Change the page to dirty & writable */
			  pageP->state = PAGE_DIRTY;
			  page_protect (pageP->page_addr, READWRITE_ACCESS);
			  INCR_STATS(PagedStats.protect_mmRWcalls,1);

			  INCR_STATS(PagedStats.unpack_diffedCount, 1);
			  INCR_STATS(PagedStats.unpack_diffedSize, copy_len);
		      } else {
			  INCR_STATS(PagedStats.unpack_dirtyCount, 1);
			  INCR_STATS(PagedStats.unpack_dirtySize, copy_len);
		      }

		      /*
		       * Page is dirty
		       * Apply the data to both the original and the twin
		       */
		      ASSERT (pageP->state == PAGE_DIRTY);
		      
		      /* get address of twin */
		      twin_addr = pageP->twin_addr;
		      ASSERT (twin_addr != NULL);
		  
		      twin_addr += data_addr - PAGE_ADDR(data_addr);
		  
		      if (IS_DBG_LOCK_MLPD) {
			  char255 smsg;
			  sprintf(smsg, "mid_paged_UNPACK_data: copying to 0x%x+%ld and 0x%x from 0x%lx", 
				  data_addr, copy_len, twin_addr, source_addr);
			  StatusMsg(smsg);
		      }
		      TRACE_3("Udty", data_addr, copy_len);
		      
#ifdef DEBUG
		      /* Compare twin to current page contents for the area we are about to change.
		       * If twin & current differ, then we are about to overwrite live data.
		       * This is a bug in the application. Issue a warning.  WAS 27Feb94
		       */
		      if (bcmp (data_addr, twin_addr, copy_len) != 0) {
			  fret("WARNING: Overwriting live data at 0x%x to 0x%x (%d bytes)",
			       data_addr, data_addr+copy_len, copy_len);
			  INCR_STATS(PagedStats.unpack_errors, 1);
#ifdef STOP_ON_ERROR		
			  botch("application error");
#endif
		      }
#endif /* DEBUG */
		      /* applying diffs to BOTH the data page & the twin! */
		      bcopy (source_addr,		/* source */
			     WRITE_ADDR(data_addr),	/* target's write mapped page */
			     copy_len);			/* length */
		      
		      bcopy (source_addr,     	/* source */
			     twin_addr,		/* target */
			     copy_len);		/* length */
		  }
		  
		  /* update loop counters */
		  data_len -= copy_len;
		  data_addr += copy_len;
		  source_addr += copy_len;
		  
	      } /* while data_len > 0 */
	    } /* register... */

	    /* advance to next addrpair */
	    pairP = nextP;
	    
	} /* for i = */

	/* advance to next addrseg */
	segP = (addrseg_t *) pairP;

    } /* while (length) */

    /* nothing left in buffer */
    *pptr = 0;			/* all done */
    *lptr = 0;
    
done:

    TRACE_4("unpk", synch_obj, *pptr, *lptr);

    /* release  page tables*/
    mutex_unlock (paged_mutex);

    return;

} /* mid_unpack_data */

u_long 
mid_paged_live_data (m_lock_req_t *lockReqP,
		     u_long        lockReqLen)
{
    register int		i;
    register u_long      live_data;
    register long	length;		/* length of packet pointed to by pptr */
    register addrseg_t	*segP;		/* data segments, if any */
    register addrpair_t    *pairP;		/* address/length/data triples, really */
    register addrpair_t    *nextP;
    

    live_data = 0;
    segP = (addrseg_t *) (lockReqP+1);
    length = lockReqLen - sizeof(*lockReqP);

    while (length > sizeof(*segP)) {
	if (segP->segtype != DIFFSEG)
	    break;
	
	pairP = &(segP->ap);
	length -= ((u_long)pairP - (u_long)segP); /* account for addrseg header */
	
	for(i = 0; i < segP->num; i++) {

	    live_data += pairP->len;
	    ASSERT (live_data < lockReqLen);

	    nextP = (addrpair_t *) &pairP->data[ pairP->len ];
	    length -= ((u_long)nextP - (u_long)pairP);
	    pairP = nextP;
	    ASSERT (length >= 0);
	    
	} /* for i = */

	/* advance to next segment descriptor */
	segP = (addrseg_t *) pairP;

    } /* while (length) */

    ASSERT (live_data < lockReqLen);
    return live_data;

} /* mid_paged_live_data */


#if PSTATS > 0
static void print_summary (char *nameP, u_long *dataP, int size, int bucket)
{
    int i;
    u_long sum, val, count, count1, data_count;
    u_long maxval, maxcount;
    double modepct, mean;
    double dcount, dsum;
    double dval, sumsq, var, stdev;
    double dcount1, mean1, var1, stdev1;
    char overflow[80];
    char255 smsg;
    extern double sqrt();

    /*
     * data[0..size-1] contains the actual counts
     * data[size] contains the overflow count
     */
    
    sum = 0;
    count = 0;
    sumsq = 0.0;
    maxval = 0;
    maxcount = 0;
    for (i = 0; i < size; i++) {
	data_count = dataP[i];
	if (data_count != 0) {
	    if (data_count > maxcount) {
		maxcount = data_count;
		maxval = i;
	    }
	    count += data_count;
	    val = data_count * (i * bucket);
	    sum += val;
	    dval = (double) val;
	    sumsq += (dval * dval);
	}
    }
    if (dataP[size] != 0) {
	sprintf(overflow, "Ovfl %lu at %lu, ", 
		dataP[size], size*bucket);
    } else
	overflow[0] = 0;
    if (count == 0) {
	sprintf(smsg, "%s: %scnt = 0", nameP, overflow);
	StatusMsg (smsg);
    } else {
	dsum = (double) sum;
	dcount = (double) count;
	mean = dsum / dcount;
	modepct = (((double) maxcount) * 100.0) / dcount;
	if (count == 1) {
	    var = 0.0;
	    stdev = 0.0;
	} else {
	    var = (sumsq - ((dsum * dsum) / dcount)) / (dcount - 1.0);
	    stdev = sqrt(var);
	    count1 = count - dataP[0];
	    dcount1 = count1;
	    if (count1 == 0) {
		mean1 = var1 = stdev1 = 0.0;
	    } else if (count1 == 1) {
		mean1 = dsum / dcount1;
		var1 = stdev1 = 0.0;
	    } else {
		mean1 = dsum / dcount1;
		var1 = (sumsq - ((dsum * dsum) / dcount1)) / (dcount1 - 1.0);
		stdev1 = sqrt(var1);
	    }
	}
	sprintf(smsg,"%s: %scnt %lu sum %lu mode %lu (%lu %2.1f%%)",
	       nameP, overflow, count, sum, maxval*bucket, maxcount, modepct);
	StatusMsg (smsg);
	sprintf(smsg,"AVG0 %.3f var %.3f stdev %.3f, AVG1 %.3f var %.3f stdev %.3f",
		mean, var, stdev, mean1, var1, stdev1);
	StatusMsg (smsg);
    }
}

static void print_histogram (char *nameP, u_long *dataP, int size, int bucket)
{
    int i, j, len, slen;
    char255 smsg, tmp;
#define MAX_STR_LEN 80

    /*
     * data[0..size-1] contains the actual counts
     * data[size] contains to overflow count
     */
    
#if PSTATS_DETAILED
    sprintf(smsg, "%s[%d]:", nameP, bucket);
    slen = strlen(smsg);
    for (i = 0; i <= size; i++) {
	if (dataP[i] == 0) continue;
	
	sprintf(tmp, " %d=%lu", i*bucket, dataP[i]);
	len = strlen(tmp);
	if ((len + slen) > MAX_STR_LEN) {
	    StatusMsg(smsg);
	    sprintf(smsg, "   %s:", nameP);
	    slen = strlen(smsg);
	}
	strcat (smsg, tmp);
    slen += len;
    }
    StatusMsg(smsg);
#else
    /* show lowest non-zero buckets */
    sprintf(smsg, "%s[%d]:", nameP, bucket);
    slen = strlen(smsg);
    for (i = 0; i <= size; i++) {
	if (dataP[i] == 0) continue;
	
	sprintf(tmp, " %d=%lu", i*bucket, dataP[i]);
	len = strlen(tmp);
	if ((len + slen) > MAX_STR_LEN) {
	    strcat (smsg, ", ...,");
	    StatusMsg(smsg);
	    goto high;
	}
	strcat (smsg, tmp);
	slen += len;
    }
    StatusMsg(smsg);
    goto done;
high:
    /* determine highest non-zero buckets */
    sprintf(smsg, "   %s:", nameP);
    slen = strlen(smsg);
    for (j = size; j >= i; j--) {
	if (dataP[j] == 0) continue;
	
	sprintf(tmp, " %d=%lu", j*bucket, dataP[j]);
	len = strlen(tmp);
	if ((len + slen) > MAX_STR_LEN) {
	    j++;
	    goto print_high;
	}
	strcat (smsg, tmp);
	slen += len;
    }
print_high:
    /* print highest non-zero buckets */
    sprintf(smsg, "   %s:", nameP);
    slen = strlen(smsg);
    for (; j <= size; j++) {
	if (dataP[j] == 0) continue;
	
	sprintf(tmp, " %d=%lu", j*bucket, dataP[j]);
	len = strlen(tmp);
	if ((len + slen) > MAX_STR_LEN)
	    break;
	strcat (smsg, tmp);
	slen += len;
    }
    StatusMsg(smsg);
done:;
#endif
}
#endif /* PSTATS > 0 */

void mid_paged_print_stats (void)
{
#if !defined(PAGE_BASED)
    return;			/* no errors, and no worries */
#else
#if !defined(STATS)
    StatusMsg ("Not compiled to collect Page-Based statistics");
#else
    int i, len, slen;
    u_long acqs, reqs;
    char255 smsg, tmp;
    
    StatusMsg("-=-=- Page-Based Statistics -=-=-");
    
    sprintf(smsg,
	    "PgReqs %lu (R %lu W %lu Z %lu) free %lu write %lu acc %lu (C %lu D %lu Df %lu)",
	    PagedStats.data_reqs, PagedStats.data_reqsRead,
	    PagedStats.data_reqsWrite, PagedStats.data_reqsZeroed, 
	    PagedStats.data_free, PagedStats.data_write, 
	    PagedStats.data_access, PagedStats.data_accessClean,
	    PagedStats.data_accessDirty, PagedStats.data_accessDiffed);
    StatusMsg (smsg);

    sprintf(smsg,
	    "PgDesc %lu (Z %lu) MaxAlloc %lu Free %lu cur %lu max %lu",
	    PagedStats.malloc_calls, PagedStats.malloc_zeroed, PagedStats.malloc_max,
	    PagedStats.free_calls, free_list_entries, PagedStats.free_max);
    StatusMsg (smsg);

    sprintf(smsg,
	    "PgProt (N %lu R %lu W %lu - MultMap R %lu W %lu) RealProt (R %lu W %lu)",
	    PagedStats.protect_NOcalls, PagedStats.protect_ROcalls, 
	    PagedStats.protect_RWcalls, PagedStats.protect_mmROcalls, 
	    PagedStats.protect_mmRWcalls,
	    PagedStats.protect_ROcalls - PagedStats.protect_mmROcalls,
	    PagedStats.protect_RWcalls - PagedStats.protect_mmRWcalls);
    StatusMsg (smsg);

    sprintf(smsg,
	    "PgTwins %lu (Copy %lu) Diffs %lu cnt %lu siz %luK raw: %lu siz %luK ",
	    PagedStats.page_twins, PagedStats.page_twinsCopied,  
	    PagedStats.make_diffs, PagedStats.make_diffsCount, 
	    PagedStats.make_diffsSize/1024, 
	    PagedStats.copy_data, PagedStats.copy_dataSize/1024);
    StatusMsg (smsg);

    sprintf(smsg,
	    "PgCopy %lu cnt %lu siz %luK (all %lu f %lu l %lu m %lu)",
	    PagedStats.copy_dirty, PagedStats.copy_dirtyCount, 
	    PagedStats.copy_dirtySize/1024, PagedStats.copy_dirtyCountComp,
	    PagedStats.copy_dirtyCountFirst, PagedStats.copy_dirtyCountLast,
	    PagedStats.copy_dirtyCountMiddle);
    StatusMsg (smsg);

    sprintf(smsg,
	    "PgApply %lu cnt %lu siz %luK",
	    PagedStats.apply_clean, PagedStats.apply_cleanDiffs, 
	    PagedStats.apply_cleanSize/1024); 
    StatusMsg (smsg);

    sprintf(smsg,
	    "PgPack: Diff %lu (%luK) Data %lu (%luK) AMap(C %luK D %luK Df %luK)",
	    PagedStats.pack_diffCalls, PagedStats.pack_diffSize/1024,
	    PagedStats.pack_dataCalls, PagedStats.pack_dataSize/1024,
	    PagedStats.pack_cleanSize/1024,  PagedStats.pack_dirtySize/1024, 
	    PagedStats.pack_diffSize/1024);
    StatusMsg (smsg);

    sprintf(smsg,
	    "PgDiff: Made %lu Reuse %lu Cln %lu Used(C %lu F %lu D %lu)",
	    PagedStats.pack_diffsMade, PagedStats.pack_diffsReused,
	    PagedStats.pack_cleaned,   PagedStats.diff_reuseCleanTotal,
	    PagedStats.diff_reuseFaultTotal, PagedStats.diff_reuseUpdateTotal);
    StatusMsg (smsg);

    sprintf(smsg,
	    "PgUnpack: %lu Stat %lu (%lu b) Alloc %lu Cln %lu (%luK) Drty %lu (%luK) Diff %lu (%luK)",
	    PagedStats.unpack_calls, PagedStats.unpack_staticCount, 
	    PagedStats.unpack_staticSize, PagedStats.unpack_pagesAlloc, 
	    PagedStats.unpack_cleanCount, PagedStats.unpack_cleanSize/1024,
	    PagedStats.unpack_dirtyCount, PagedStats.unpack_dirtySize/1024,
	    PagedStats.unpack_diffedCount, PagedStats.unpack_diffedSize/1024);
    StatusMsg (smsg);

    sprintf(smsg,
	    "PgErrors: Unpack %lu System %lu",
	    PagedStats.unpack_errors, 
	    PagedStats.system_errors);
    StatusMsg (smsg);

    mid_paged_lock_stats();
    
    
#if PSTATS > 0
    /* WARNING: print_summary dies with dataP == 0, trying to reference
       dataP[i] if it is called and we are not page_based. */
    print_summary ("bytes/diff", PagedStats.make_diffsSizeH, 
		   STAT_DIFFSIZE_TABLE_SIZE, STAT_DIFFSIZE_PER_BUCKET);
    print_histogram ("df SIZ", PagedStats.make_diffsSizeH, 
		   STAT_DIFFSIZE_TABLE_SIZE, STAT_DIFFSIZE_PER_BUCKET);

    print_summary ("bytes/page", PagedStats.make_diffsSizePageH, 
		   STAT_DIFFPAGE_TABLE_SIZE, STAT_DIFFPAGE_PER_BUCKET);
    print_histogram ("df PAG", PagedStats.make_diffsSizePageH, 
		     STAT_DIFFPAGE_TABLE_SIZE, STAT_DIFFPAGE_PER_BUCKET);
    
    print_summary ("bytes/clean", PagedStats.apply_cleanSizeH, 
		   STAT_DIFFSIZE_TABLE_SIZE, STAT_DIFFSIZE_PER_BUCKET);
    print_histogram ("Cl SIZ", PagedStats.apply_cleanSizeH, 
		   STAT_DIFFSIZE_TABLE_SIZE, STAT_DIFFSIZE_PER_BUCKET);

    print_summary ("clean/page", PagedStats.apply_cleanSizePageH, 
		   STAT_DIFFPAGE_TABLE_SIZE, STAT_DIFFPAGE_PER_BUCKET);
    print_histogram ("Cl PAG", PagedStats.apply_cleanSizePageH, 
		     STAT_DIFFPAGE_TABLE_SIZE, STAT_DIFFPAGE_PER_BUCKET);
    
    print_summary ("dfRe-Clnd", PagedStats.diff_reuseCleanH,
		   STAT_DIFFREUSE_TABLE_SIZE, STAT_DIFFREUSE_PER_BUCKET);
    print_histogram ("dfRe CLN", PagedStats.diff_reuseCleanH, 
		     STAT_DIFFREUSE_TABLE_SIZE, STAT_DIFFREUSE_PER_BUCKET);
    print_summary ("dfRe-Fltd", PagedStats.diff_reuseFaultH,
		   STAT_DIFFREUSE_TABLE_SIZE, STAT_DIFFREUSE_PER_BUCKET);
    print_histogram ("dfRe FLT", PagedStats.diff_reuseFaultH, 
		     STAT_DIFFREUSE_TABLE_SIZE, STAT_DIFFREUSE_PER_BUCKET);
    print_summary ("dfRe-Updt", PagedStats.diff_reuseUpdateH,
		   STAT_DIFFREUSE_TABLE_SIZE, STAT_DIFFREUSE_PER_BUCKET);
    print_histogram ("dfRe UPD", PagedStats.diff_reuseUpdateH, 
		     STAT_DIFFREUSE_TABLE_SIZE, STAT_DIFFREUSE_PER_BUCKET);
    
#endif /* PSTATS>0 */
#endif /*STATS*/
#endif /*PAGE_BASED*/
} /* mid_paged_print_stats */

void mid_paged_reset_stats (void)
{
#if defined(STATS)
    u_long *p1, *p2, *p3, *p4, *p5, *p6, *p7;
#if PSTATS > 0
    bzero (PagedStats.make_diffsSizeH, (STAT_DIFFSIZE_TABLE_SIZE+1)*sizeof(u_long));
    bzero (PagedStats.make_diffsSizePageH, (STAT_DIFFPAGE_TABLE_SIZE+1)*sizeof(u_long));
    bzero (PagedStats.apply_cleanSizeH, (STAT_DIFFPAGE_TABLE_SIZE+1)*sizeof(u_long));
    bzero (PagedStats.apply_cleanSizePageH, (STAT_DIFFPAGE_TABLE_SIZE+1)*sizeof(u_long));
    p1 = PagedStats.make_diffsSizeH;
    p2 = PagedStats.make_diffsSizePageH;
    p3 = PagedStats.apply_cleanSizeH;
    p4 = PagedStats.apply_cleanSizePageH;
    bzero (PagedStats.diff_reuseCleanH, (STAT_DIFFREUSE_TABLE_SIZE+1)*sizeof(u_long));
    bzero (PagedStats.diff_reuseFaultH, (STAT_DIFFREUSE_TABLE_SIZE+1)*sizeof(u_long));
    bzero (PagedStats.diff_reuseUpdateH, (STAT_DIFFREUSE_TABLE_SIZE+1)*sizeof(u_long));
    p5 = PagedStats.diff_reuseCleanH;
    p6 = PagedStats.diff_reuseFaultH;
    p7 = PagedStats.diff_reuseUpdateH;
#endif  

    bzero (&PagedStats, sizeof(PagedStats));

#if PSTATS > 0
    PagedStats.make_diffsSizeH = p1;
    PagedStats.make_diffsSizePageH = p2;
    PagedStats.apply_cleanSizeH = p3;
    PagedStats.apply_cleanSizePageH = p4;
    PagedStats.diff_reuseCleanH = p5;
    PagedStats.diff_reuseFaultH = p6;
    PagedStats.diff_reuseUpdateH = p7;
#endif  
#endif
} /* mid_paged_reset_stats */
