/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */

#include <mach.h>
#include <mach/kern_return.h>
#include <cthreads.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <errno.h>
#include <syscall.h>
#include "midway_types.h"
#include "midway_mem.h"
#include "bitwiddle.h"
#include "shmalloc.h"
#include "smexecd_types.h"	/* char255 */
#include "backing.h"
#include "xpager.h"
#include "midway-dbg.h"		/* MIDWAY_MUTEX_SET_NAME */


#define PAGE_TO_DISK 1		/*  = 0 to page to virtual memory */
                                /*  = 1 to page to file on disk */



#if PAGE_TO_DISK > 0

#define PAGEMAP_LENGTH   (MAX_SHARED_SPACE / vm_page_size) /* 1 bit/page */

struct backing_store_data {
    mutex_t mu;
    int  fd;			/* stdio file descriptor for backing store file */
    BITARRAY *pagemap;		/* bit map indicating data written to backing store */
    char fname[64];		/* filename for debugging */
};

static char backing_store_filename[] = { "/usr/tmp/midway.backing" };
static int  backing_store_number = 0;

#else /* not PAGE_TO_DISK */

#define HASH_TABLE_SIZE 2019

struct backing_store_data {
    mutex_t mu;
    backing_page_t pages[HASH_TABLE_SIZE];
};

#endif /* PAGE_TO_DISK */




/*
 * Free list machinery
 */

backing_page_t free_list = 0;
int free_list_len = 0;
mutex_t free_list_mu = 0;


static
backing_page_t
new_backing_page()
{
    backing_page_t p;
    kern_return_t krc;
    vm_address_t pageAddr;
    
    mutex_lock(free_list_mu);
    p = free_list;
    if (p)  {
	free_list_len--;
	free_list = p->next;
    }
    mutex_unlock(free_list_mu);
    if (!p)  {
	p = (backing_page_t) MALLOC(sizeof(struct backing_page_data));

	/* Data buffer MUST be page aligned. WAS 21 Oct 92  */
	krc = VM_ALLOCATE (mach_task_self(),
			   &pageAddr, 
			   (vm_size_t)vm_page_size, 
			   (boolean_t) 1); /* alloc anywhere */
	if (krc != KERN_SUCCESS) {
  	    panic("xpager: could not allocate backing store: %s\n",
		  mach_error_string(krc));
	}
	p->data = (char*)pageAddr;
    }
    return p;
}

static void
free_backing_page(p)
    backing_page_t p;
{
    kern_return_t krc;
  
    if (free_list_len > 16)  {
        krc = VM_DEALLOCATE (mach_task_self(),
			     (vm_address_t) p->data, vm_page_size);
	if (krc != KERN_SUCCESS) {
	    panic("xpager: could not deallocate backing store: %s\n",
		  mach_error_string(krc));
        }
	FREE(p);
    } else {
	mutex_lock(free_list_mu);
	p->next = free_list;
	free_list = p;
	free_list_len++;
        mutex_unlock(free_list_mu);
    }
}

/*
 * backing store manipulation
 */


    
backing_store_t
backing_store_create()
{
  int cc;
    backing_store_t b;

    if (!free_list_mu)  {
	free_list_mu = mutex_alloc();
	MIDWAY_MUTEX_SET_NAME(free_list_mu,"backing.c:free_list_mu");
    }

    b = (backing_store_t) MALLOC(sizeof(struct backing_store_data));
    if (!b)  {
	fprintf(stderr,"backing_store_create: malloc failed\n");
	exit(-1);
    }
    b->mu = mutex_alloc();
    MIDWAY_MUTEX_SET_NAME(b->mu,"backing_store_mutex");

#if PAGE_TO_DISK > 0
    b->pagemap = (BITARRAY *) MALLOC(midBITARRAYsize(PAGEMAP_LENGTH));
    if (!b->pagemap) {
	fprintf(stderr,"backing_store_create: pagemap malloc failed\n");
	exit(-1);
    }
    midZeroBitArray (b->pagemap, midBITARRAYsize(PAGEMAP_LENGTH));
    
    { char255 filename;
      int pid;

      pid = getpid();
      backing_store_number++;
      sprintf(filename, "%s.%d.%d", 
	      backing_store_filename, pid, backing_store_number);
      mumble("creating backing store in file %s", filename);
      
      b->fd = syscall (SYS_open, filename, 
		       O_RDWR | O_CREAT | O_TRUNC, 
		       0666); /* PERMS = R/W/ for everyone */
      if (b->fd == -1) {
	fprintf(stderr,"backing_store_create: failed to open file %s, errno=%d\n",
		filename, errno);
	exit(-1);
      }
      strncpy(b->fname, filename, sizeof(b->fname)); /* save filename for debugging */
    }
#else
    bzero(b->pages, sizeof(b->pages));
#endif

    return b;
}


void
backing_store_release(b)
    backing_store_t b;
{
    int i;
    int cc;
    backing_page_t p;
    backing_page_t next;    
    
    mutex_lock(b->mu);

#if PAGE_TO_DISK > 0
    if (b->pagemap) { 
      FREE (b->pagemap);
      b->pagemap = NULL;
    }
    if (b->fd != -1) {
      cc = syscall (SYS_close, b->fd);
      if (cc != 0) {
	fprintf(stderr,"backing_store_release: failed to close file %s, errno=%d, cc=%d\n",
		b->fname, errno, cc);
	exit(-1);
      }
      b->fd = -1;

      cc = remove(b->fname);
      if (cc != 0) {
	fprintf(stderr,"backing_store_release: failed to remove file %s, errno=%d, cc=%d\n",
		b->fname, errno, cc);
	exit(-1);
      }
    }
#else
    for (i = 0; i < HASH_TABLE_SIZE; i++)  {
	p = b->pages[i];
	while (p)  {
	    next = p->next;
	    free_backing_page(p);
	    p = p->next;
	}
    }
#endif
    mutex_unlock(b->mu);
    mutex_free(b->mu);
}




#if PAGE_TO_DISK == 0

static int
hash(offset)
    vm_offset_t offset;
{
    int h;
    h = (offset / vm_page_size) % HASH_TABLE_SIZE;
/****
    mumble("hash: hindex is %d\n", h);
****/
    return h;
}



static void
insert_backing_page(b, p)
    backing_store_t b;
    backing_page_t p;
{
    int h;

    h = hash(p->offset);
    p->next = b->pages[h];
    b->pages[h] = p;
}

	
    
static void
remove_backing_page(b, p)
    backing_store_t b;
    backing_page_t p;
{
    int h;
    backing_page_t cur;
    backing_page_t prev = 0;    

    
    h = hash(p->offset);
    cur = b->pages[h];
    while (cur)  {
	if (cur->offset == p->offset)  {
	    if (prev)  {
		prev->next = cur->next;
	    } else {
		b->pages[h] = cur->next;
	    }
	    return;
	}
	prev = cur;
	cur = cur->next;
    }
    fret("warning: remove backing page page not present: 0x%x\n",p->offset);
}
#endif  /* not PAGE_TO_DISK */

static backing_page_t
find_backing_page(b, offset)
    backing_store_t b;
    vm_offset_t offset;
{
    int h;
    int cc;
    backing_page_t p;    
    vm_offset_t pageno;
    struct stat statbuf;

#if PAGE_TO_DISK > 0    
    pageno = offset / vm_page_size;
    if (midTestBit (b->pagemap, pageno)) {

      p = new_backing_page();
      p->offset = offset;

      cc = fstat (b->fd, &statbuf);
      if (cc == -1) {
	fprintf(stderr, 
		"backing_store_read_page: error %d stating file %d, cc=%d\n", 
		errno, b->fd, cc);
	exit(-1);
      }
      cc = syscall (SYS_lseek, b->fd, offset, L_SET);
      if (cc != offset) {
	fprintf(stderr, 
		"backing_store_read_page: error %d seeking to offset 0x%x in %s\n",
		errno, offset, b->fname);
	exit(-1);
      }
      cc = fstat (b->fd, &statbuf);
      if (cc == -1) {
	fprintf(stderr, 
		"backing_store_read_page: error %d stating file %d, cc=%d\n", 
		errno, b->fd, cc);
	exit(-1);
      }
      errno = 0;
#ifdef READS
      cc = syscall (SYS_read, b->fd, p->data, vm_page_size);
#else
      { struct iovec iov;
	iov.iov_base = p->data;
	iov.iov_len = vm_page_size;
	cc = syscall (SYS_readv, b->fd, &iov, 1);
      }
#endif
      if (cc != vm_page_size) {
	fprintf(stderr, 
		"backing_store_read_page: error %d reading addr 0x%x at offset 0x%x in %s\n",
		errno, p->data, offset, b->fname);
/***********
	exit(-1);
**************/
      }
      cc = fstat (b->fd, &statbuf);
      if (cc == -1) {
	fprintf(stderr, 
		"backing_store_read_page: error %d stating file %d, cc=%d\n", 
		errno, b->fd, cc);
	exit(-1);
      }
    } else {
      p = NULL;
    }
#else
    h = hash(offset);
    p = b->pages[h];
    while (p)  {
	if (p->offset == offset)  {
	    break;
	}
	p = p->next;
    }
#endif

    return p;
}



backing_store_write_page(b, offset, addr)
    backing_store_t b;
    vm_offset_t offset;
    vm_offset_t addr;
{
    backing_page_t p;
    vm_offset_t pageno;
    int cc;

    mutex_lock(b->mu);

#if PAGE_TO_DISK > 0
    cc = syscall (SYS_lseek, b->fd, offset, L_SET);
    if (cc != offset) {
      fprintf(stderr, 
	      "backing_store_write_page: error %d seeking to offset 0x%x in %s\n",
	      errno, offset, b->fname);
      exit(-1);
    }
    cc = syscall (SYS_write, b->fd, (char *)addr, vm_page_size);
    if (cc != vm_page_size) {
      fprintf(stderr, 
	      "backing_store_write_page: error %d writing addr 0x%x to offset 0x%x in %s\n",
	      errno, addr, offset, b->fname);
      exit(-1);
    }
    pageno = offset / vm_page_size;
    midSetBit (b->pagemap, pageno);
#else
    p = find_backing_page(b, offset);
    if (p)  {
	fret("warning. backing_store_write_page writing backed page: %x\n",
	     offset);
    } else {
	p = new_backing_page();
    }
    bcopy(addr, p->data, vm_page_size);    
    p->offset = offset;
    insert_backing_page(b, p);
#endif

    mutex_unlock(b->mu);
}

    
void
backing_store_write(b, offset, addr, len)
    backing_store_t b;
    vm_offset_t addr;
    vm_size_t len;
{
    while (len > 0)  {
	backing_store_write_page(b, offset, addr);
	len  -= vm_page_size;
	addr += vm_page_size;
	offset += vm_page_size;
    }
}


backing_page_t
backing_store_read_page(b, offset)
    backing_store_t b;
    vm_offset_t offset;
{
    backing_page_t p;
    mutex_lock(b->mu);
    p = find_backing_page(b, offset);
    mutex_unlock(b->mu);
    return p;
}



void
backing_store_release_page(b, p)
    backing_store_t b;
    backing_page_t p;
{
    mutex_lock(b->mu);

#if PAGE_TO_DISK > 0
    { vm_offset_t pageno;
      pageno = p->offset / vm_page_size;
      midClearBit (b->pagemap, pageno);
    }
#else
    remove_backing_page(b, p);
#endif

    mutex_unlock(b->mu);
    free_backing_page(p);
}


	
    
