/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/* 
 * Mach Operating System
 * Copyright (c) 1989 Carnegie-Mellon University
 * All rights reserved.  The CMU software License Agreement specifies
 * the terms and conditions for use and redistribution.
 */
/*
 * HISTORY
 * $Log: xpager.c,v $
 * Revision 1.14  1994/10/19  21:03:27  wsawdon
 * Minor changes for page based code.
 *
 * Revision 1.13  1994/06/16  03:12:03  wsawdon
 * Updated stats on malloc/vm_alloc/etc.
 *
 * Revision 1.12  1994/06/15  03:03:45  wsawdon
 * Collect stats on vm_alloc to check memory usage.
 *
 * Revision 1.11  1994/06/07  04:28:15  wsawdon
 * Changes for page based version of midway.
 *
 * Revision 1.10  1994/02/18  14:53:50  wsawdon
 * Initial changes to support page based version.
 *
 * Revision 1.9  1993/07/02  04:48:18  wsawdon
 * Page shared/dirty memory to disk rather than to non-shared memory.
 * Also fixed bug in freeing temp buffers used by backing store.
 *
 * Revision 1.8  1992/12/22  21:47:00  wsawdon
 * Added lazy allocation of dirty bits. Dirty bits are allocated
 * out of their own region. Page faults in this region are handled
 * by the xpager and filled with 1's (for EC memory) or Non-EC flags
 * (for Non-EC memory). Also changed STATS to malloc histograms
 * and to compiled out if STATS are not defined.
 *
 * Revision 1.7  1992/12/15  02:34:56  mattz
 * Use precious pages so that we do not have to keep copies of pages that
 * are in memory.
 *
 * Revision 1.6  1992/12/02  20:49:52  mattz
 * Add mutex name support.
 *
 * Revision 1.5  1992/11/08  04:25:40  mattz
 * Add xpager stat. (So we see number of pagin requests.)
 *
 * Revision 1.4  1992/10/24  21:43:17  wsawdon
 * Fixed 2 bugs in external pager.
 *   Page data passed back on a data_request must be page aligned.
 *   Always give R/W access even if R/O requested.
 *
 * Revision 1.3  1992/10/14  21:24:50  wsawdon
 * Changes to support fast udp library.
 *
 * Revision 1.2  1992/10/04  23:42:14  mattz
 * Remove verbosity.  debug flags changed name, now placed under cmd line flag
 * control.
 *
 * Revision 1.1  1992/10/01  20:00:44  mattz
 * External Pager for the Midway runtime.  Keeps track of shared storage.
 *
 */
#define _XPAGER_INTERNAL_  	/* def external vars here */

#include <mach.h>
#include <cthreads.h>
#include "midway-dbg.h"		/* for STATS, etc. */
#include "backing.h"
#include "xpager.h"
#include "xpager_obj.h"

#define USING_PRECIOUS_PAGES 1


#ifdef PAGE_BASED
static vm_offset_t paged_zero_buf = 0; /* pre-allocate a buffer full of zeros */
#define PTIME
#endif

#ifdef PTIME
/*
 * Define free running counter
 */
#define TICKS_PER_USEC 25	/* 1 tick every 40 nanoseconds */
#define MAX_FRC ((unsigned int) 0xffffffff)
#define DELTA_FRC(efrc,sfrc) (efrc >= sfrc) ? (efrc - sfrc) : ((MAX_FRC - sfrc) + efrc)

extern u_int *midway_kern_frc;
#define GET_TIME(x) x = *midway_kern_frc
#define DIFF_TIME(end,start) DELTA_FRC(end,start)
#define USEC_TIME(diff) (diff / TICKS_PER_USEC)

static u_long ptime_no_count = 0;
static u_long ptime_no_total = 0;
static u_long ptime_ro_count = 0;
static u_long ptime_ro_total = 0;
static u_long ptime_rw_count = 0;
static u_long ptime_rw_total = 0;
#endif


/*
 * return the "protection value indicating those forms of access that
 * should not be permitted to the specified cached data given the
 * desired access requested."
 * Ugh.
 */

static vm_prot_t
lock_value_from_desired_access(desired_access)
    vm_prot_t desired_access;
{
    if (desired_access & VM_PROT_WRITE)  {
	return VM_PROT_NONE;
    } else {
	return VM_PROT_WRITE;
    }
}


static object_t
verify_and_lookup(memory_object, memory_control, offset, length)
    mach_port_t memory_object;
    mach_port_t memory_control;
    vm_offset_t offset;
    vm_size_t length;
{
    object_t object;

    object = object_lookup(memory_object);
    if (object == OBJECT_NULL) {
	fret("verify_and_lookup: invalid memory_object\n");
	return OBJECT_NULL;
    }
    if (memory_control == MACH_PORT_NULL) {
	fret("verify_and_lookup: null memory_control\n");
	return OBJECT_NULL;
    }
    if (object->memory_control != memory_control) {
	fret("verify_and_lookup: invalid memory_control\n");
	return OBJECT_NULL;
    }
    if (length % vm_page_size) {
	fret("verify_and_lookup: invalid length %d\n", length);
	return OBJECT_NULL;
    }
    if (offset % vm_page_size) {
	fret("verify_and_lookup: invalid offset %d\n", offset);
	return OBJECT_NULL;
    }
    return object;
}

kern_return_t
do_memory_object_init(memory_object, memory_control, memory_object_name,
	page_size)
    mach_port_t memory_object;
    mach_port_t memory_control;
    mach_port_t memory_object_name;
    vm_size_t page_size;
{
    object_t object;

    mumble("$ memory_object_init(%d, %d, %d, 0x%x)\n",
	memory_object, memory_control, memory_object_name, page_size);
    if (page_size != vm_page_size) {
	fret("memory_object_init: invalid page size %d (v. %d)\n",
	    page_size, vm_page_size);
	return (KERN_SUCCESS);
    }
    object = object_lookup(memory_object);
    
    if (object == OBJECT_NULL) {
	fret("memory_object_init: invalid memory_object\n");
	return (KERN_SUCCESS);
    }
    if (object->memory_control != MACH_PORT_NULL) {
	fret("memory_object_init: duplicate call\n");
	return;
    }
    object->memory_control = memory_control;
    memory_object_set_attributes(memory_control, TRUE, 
				 midway_xpager_may_cache, 0);

#ifdef PAGE_BASED
    { kern_return_t krc;
      if (paged_zero_buf == 0) {
	krc = VM_ALLOCATE(mach_task_self(),
			  &paged_zero_buf,
			  vm_page_size,
			  TRUE); /* anywhere=TRUE */
	if (krc != KERN_SUCCESS)
	  fret("memory_object_init: vm_allocate failed, krc=%s",
	       mach_error_string(krc));

	/* Zero the page to DIRTY it.
	 * Otherwise kernel will not instantiate it
	 * and the 1st write to the page is very expensive
	 */
	bzero(paged_zero_buf, vm_page_size);
      }
    }
#endif    

    return (KERN_SUCCESS);
}



kern_return_t
do_memory_object_copy(old_memory_object, old_memory_control, offset, length,
	new_memory_object)
    memory_object_t old_memory_object;
    memory_object_control_t old_memory_control;
    vm_offset_t offset;
    vm_size_t length;
    memory_object_t new_memory_object;
{
    fret("$ memory_object_copy(%d, %d, 0x%x, 0x%x, %d)\n",
	old_memory_object, old_memory_control, offset, length,
	new_memory_object);
    fret("memory_object_copy: invalid call");
    return (KERN_SUCCESS);
}



kern_return_t
do_memory_object_terminate(memory_object, memory_control, memory_object_name)
    mach_port_t memory_object;
    mach_port_t memory_control;
    mach_port_t memory_object_name;
{
    object_t object;

    mumble("$ memory_object_terminate(%d, %d, %d)\n",
	memory_object, memory_control, memory_object_name);
    mach_port_deallocate(mach_task_self(), memory_control);
    mach_port_deallocate(mach_task_self(), memory_object_name);
    object = object_lookup(memory_object);
    if (object == OBJECT_NULL) {
	fret("memory_object_terminate: invalid memory_object\n");
	return (KERN_SUCCESS);
    }
    object->memory_control = MACH_PORT_NULL;
    return (KERN_SUCCESS);
}

kern_return_t
do_memory_object_data_request(memory_object, memory_control, offset, length,
	desired_access)
    mach_port_t memory_object, memory_control;
    vm_offset_t offset;
    vm_size_t length;
    vm_prot_t desired_access;
{
    object_t object;
    backing_page_t p;
    vm_prot_t lock_value;
    kern_return_t krc;
    int precious; /* IGNORED */    
    char *buf = 0;    

    mumble("$ memory_object_data_request(%d, %d, 0x%x, 0x%x, %d)",
	memory_object, memory_control, offset, length, desired_access);

    INCR_STAT(xpager_stat.data_reqs,1);

    object = verify_and_lookup(memory_object, memory_control, offset, length);
    if (object == OBJECT_NULL) {
	return (KERN_SUCCESS);
    }

    while (length > 0)  {
	buf = 0;
	p = backing_store_read_page(object->backing_store, offset);
	if (p)  {
	    buf = p->data;
	    krc = KERN_SUCCESS;
	    INCR_STAT(xpager_stat.backing_reads,1);
	  } else if (object->request && object->request->data_request)  {
	    krc = (*(object->request->data_request))(object->request->data_any,
						     offset,
						     length,
						     &desired_access,
						     &buf,
						     &precious);
	    if (krc != KERN_SUCCESS)  {
		memory_object_data_error(memory_control,
					 offset,
					 vm_page_size,
					 krc);
		buf = 0;
	    }
	}
	if (buf)  { /* supply some data */
#if defined(READONLY_ACCESS_TO_SHARED_REGION) || defined(PAGE_BASED)
	    lock_value = lock_value_from_desired_access(desired_access);
#else
	    lock_value = VM_PROT_NONE; /* give full access every time */
#endif
#ifdef USING_PRECIOUS_PAGES
	    memory_object_data_supply(memory_control,
					offset,
					(vm_offset_t)buf,
					vm_page_size,
					FALSE, /* don't deallocate buf */
					lock_value,
					TRUE, /* precious */
					MACH_PORT_NULL);
	    if (p)  {
		backing_store_release_page(object->backing_store, p);
		INCR_STAT(xpager_stat.backing_releases,1);
	    } else {
		if (object->request->databuf_free)  {
		    krc = (*(object->request->databuf_free))(
					     object->request->data_any,
					     buf);
		}
	    }
#else
	    memory_object_data_provided(memory_control,
					offset,
					buf,
					vm_page_size,
					lock_value);
#endif
	} else if (krc == KERN_SUCCESS) {
#ifdef PAGE_BASED
	    /* The xpager interface does NOT have a mechanism to 
	     * zerofill a page AND set the page protection. 
	     * Golly, why would you ever want to do that?
	     * So, if this request is for read access, supply
	     * a page of zeros and let mach copy it. WAS 23Feb94
	     */
	    lock_value = lock_value_from_desired_access(desired_access);
	    if (lock_value != VM_PROT_NONE) {
#ifdef USING_PRECIOUS_PAGES
	      memory_object_data_supply(memory_control,
					offset,
					paged_zero_buf,
					vm_page_size,
					FALSE, /* don't deallocate buf */
					lock_value,
					TRUE, /* precious */
					MACH_PORT_NULL);
#else
	      memory_object_data_provided(memory_control,
					  offset,
					  paged_zero_buf,
					  vm_page_size,
					  lock_value);
#endif
	    } else {
	      /* provide zeros and full access to the page */
	      memory_object_data_unavailable(memory_control,
					     offset,
					     vm_page_size);
	    }
#else /* not PAGE_BASED */
	    /* make up some data (zeros) */
	    memory_object_data_unavailable(memory_control,
					   offset,
					   vm_page_size);
#endif /* PAGE_BASED */
	    INCR_STAT(xpager_stat.data_zeroed,1);
	} 
	length -= vm_page_size;
	offset += vm_page_size;
    }
    return (KERN_SUCCESS);
}



kern_return_t
do_memory_object_data_write(memory_object, memory_control,
			    offset, data, length)
    mach_port_t memory_object;
    mach_port_t memory_control;
    vm_offset_t offset;
    char *data;
    int length;
{
    kern_return_t krc;

    mumble("$ memory_object_data_write(%d, %d, 0x%x, 0x%x, 0x%x)\n",
	memory_object, memory_control, offset, data, length);

    INCR_STAT(xpager_stat.data_writes,1);

    krc = do_memory_object_data_return(memory_object, memory_control,
				       offset, data, length,
				       TRUE, /* data is dirty */
				       FALSE /* kernel kept a copy */
				       );
}


/*
 * kernel could call this if it wants to flush the page, or if we request
 * that the page be given back to us.  It is the new form of data_write.
 * Since we will never request that a page should be given to us, for
 * now the action would be as if data_write:  write the page to backing
 * store, and vm_deallocate it.  If we did request this, then we could
 * only place it in backing store if the kernel did not keep a copy.
 * However, we do not mark all pages as precious; only those non-zero ones
 * we supply back to the kernel.  For non-precious pages kernel_copy is
 * a "useless" hint.  Bottom line: until we make everything precious,
 * or change our usage, keep anything the kernel gives us.
 */
kern_return_t
do_memory_object_data_return(memory_object, memory_control,
			    offset, data, length, dirty, kernel_copy)
    mach_port_t memory_object;
    mach_port_t memory_control;
    vm_offset_t offset;
    char *data;
    int length;
    boolean_t dirty;		/* was returned data dirty? */
    boolean_t kernel_copy;	/* did the kernel keep a copy? */
{
    object_t object;
    kern_return_t krc;
    boolean_t discarded = FALSE;

    mumble("$ memory_object_data_return(%d, %d, 0x%x, 0x%x, 0x%x %d %d)\n",
	memory_object, memory_control, offset, data, length, dirty, 
	   kernel_copy);
    INCR_STAT(xpager_stat.data_returns,1);

    object = verify_and_lookup(memory_object, memory_control, offset, length);
    if (object == OBJECT_NULL) {
	return (KERN_SUCCESS);
    }
    /* Save the data somewhere... */
    if (object->request && object->request->data_write)  {
         krc = (*(object->request->data_write))(object->request->data_any,
						offset,
					        length,
					        data,
					        &discarded);

	 /* what do we do with krc ? */
     }
    if (!discarded)  {
	backing_store_write(object->backing_store, offset, data, length);
	INCR_STAT(xpager_stat.backing_writes,1);
    }
    VM_DEALLOCATE(mach_task_self(), (vm_offset_t)data, length);
    return (KERN_SUCCESS);
}


kern_return_t
do_memory_object_data_unlock(memory_object, memory_control, offset, length,
	desired_access)
    mach_port_t memory_object, memory_control;
    vm_offset_t offset;
    vm_size_t length;
    vm_prot_t desired_access;
{
    object_t object;
    vm_prot_t lock_value;
    kern_return_t krc = KERN_SUCCESS;

    mumble("$ memory_object_data_unlock(%d, %d, 0x%x, 0x%x, %d)",
	   memory_object, memory_control, offset, length, desired_access);
    INCR_STAT(xpager_stat.data_access_reqs,1);

    object = verify_and_lookup(memory_object, memory_control, offset, length);
    if ((object != OBJECT_NULL) &&
	(object->request && object->request->data_access_request)) {

      while (length > 0)  {
	krc = (*(object->request->data_access_request))(object->request->data_any,
							offset,
							length,
							desired_access);
	if (krc != KERN_SUCCESS) break;
	length -= vm_page_size;
	offset += vm_page_size;
      }

    } else {

      lock_value = lock_value_from_desired_access(desired_access);
      krc = xpager_protect_data(memory_object, offset, length, lock_value);
    }
    return (krc);
}

kern_return_t
do_memory_object_lock_completed(memory_object, memory_control, offset, length)
    mach_port_t memory_object;
    mach_port_t memory_control;
    vm_offset_t offset;
    vm_size_t length;
{
    object_t object;

    fret("$ memory_object_lock_completed(%d, %d, 0x%x, 0x%x)\n",
	memory_object, memory_control, offset, length);
    fret("memory_object_lock_completed: invalid call");
    /*
     * call appropriate request
     */    
    return (KERN_SUCCESS);
}

kern_return_t
do_memory_object_supply_completed(memory_object, memory_control, offset, 
				  length, result, error_offset)
    mach_port_t memory_object;
    mach_port_t memory_control;
    vm_offset_t offset;
    vm_size_t length;
    kern_return_t result;
    vm_offset_t error_offset;
{
    object_t object;

    fret("$ memory_object_supply_completed(%d, %d, 0x%x, 0x%x, %d, 0x%x)\n",
	memory_object, memory_control, offset, length, result, error_offset);
    fret("memory_object_supply_completed: invalid call");
    /*
     * call appropriate request
     */    
    return (KERN_SUCCESS);
}






static void
xpager_loop(object)
    object_t object;
{
    extern int memory_object_server();
    
    int old_state;


    /* we are cthread_fork'd; for now wire all cthreads with 3.0 threads*/
    cthread_wire();

#ifdef STATS
    xpager_stat_lock = mutex_alloc();
    MIDWAY_MUTEX_SET_NAME(xpager_stat_lock,"xpager_stat_lock");
    bzero (&xpager_stat, sizeof(xpager_stat));
#endif

#ifdef PTIME
    /* ensure FRC is enabled */
    (void) midway_setup_mapped_frc();
#endif

    mach_msg_server(memory_object_server, sizeof(max_msg_t),
		    object->memory_object);
    object_destroy(object);
    cthread_exit(0);
}



kern_return_t
do_xpager_create(xpager_port, memory_object)
    mach_port_t xpager_port;
    mach_port_t *memory_object;
{
    object_t object;
    kern_return_t krc = KERN_SUCCESS;

    object = object_create();
    if (object == OBJECT_NULL) {
	mumble("xpager_open: object_create failed\n");
	return KERN_FAILURE;
    }
    cthread_detach(cthread_fork(xpager_loop, object));
    
    *memory_object = object->memory_object;
    if (xpager_port == MACH_PORT_NULL)  {
	/* local call.  Insert a send right */
	krc = mach_port_insert_right(mach_task_self(),
			       *memory_object,
			       *memory_object,
			       MACH_MSG_TYPE_MAKE_SEND);
	if (krc != KERN_SUCCESS)  {
	    panic("xpager_open: could not make send right: %s\n",
		    mach_error_string(krc));
	}
    }
	    
    return krc;
}

kern_return_t
do_xpager_destroy(memory_object)
    mach_port_t memory_object;
{
  object_t object;
  kern_return_t krc = KERN_SUCCESS;
  
  object = object_lookup (memory_object);
  if (object == OBJECT_NULL) {
    fret("do_xpager_destroy: object_lookup failed\n");
    return KERN_FAILURE;
  }
  
  krc = object_destroy(object);
  if (krc != KERN_SUCCESS)  {
    fret("vm_xpager_destroy failure: object_destroy:%s\n",
	 mach_error_string(krc));
  }
  
  return krc;
}



/*
 * functions for local use of an external pager
 */


kern_return_t
xpager_protect_data(memory_object, offset, len, lock_value)
    mach_port_t memory_object;
    vm_offset_t offset;
    vm_size_t len;
    vm_prot_t lock_value;
{
    kern_return_t krc;
    object_t object;

    /*
     * NOTE: The "lock_value" parameter specifies the the forms
     * of access that are NOT permittted to the specified data.
     * Thus VM_PROT_WRITE means there is NO write access to the page.
     * and  VM_PROT_ALL means there is NO access to the page.
     */
    object = object_lookup(memory_object);
    if (object == OBJECT_NULL) {
        fret("xpager_protect_data: Invalid memory_object 0x%x", memory_object);
	return KERN_FAILURE;
    }
    if (object->memory_control == MACH_PORT_NULL) {
	fret("xpager_protect_data: invalid memory_control for object 0x%x", memory_object);
	return KERN_FAILURE;
    }

#ifdef PTIME
    { unsigned long _start, _end, _diff;
      GET_TIME(_start);
#endif
      krc = memory_object_lock_request (object->memory_control,
					offset, len,
					FALSE, /* should_clean */
					FALSE, /* should_flush */
					lock_value,
					MACH_PORT_NULL);
#ifdef PTIME
      GET_TIME(_end);
      _diff = DIFF_TIME(_end, _start);
      if (lock_value == VM_PROT_WRITE) {
	  ptime_ro_count++;
	  ptime_ro_total += USEC_TIME(_diff);
      } else if (lock_value == VM_PROT_NONE) {
	  ptime_rw_count++;
	  ptime_rw_total += USEC_TIME(_diff);
      } else {
	  ptime_no_count++;
	  ptime_no_total += USEC_TIME(_diff);
      }
    }
#endif

    return krc;
}

kern_return_t
xpager_clean_data(memory_object, offset, len)
    mach_port_t memory_object;
    vm_offset_t offset;
    vm_size_t len;
{
    return KERN_FAILURE;	/* unimplemented */
}

kern_return_t
xpager_flush_data(memory_object, offset, len)
    mach_port_t memory_object;
    vm_offset_t offset;
    vm_size_t len;
{
    return KERN_FAILURE;	/* unimplemented */
}





struct xpager_control xp_control = {
    xpager_protect_data,
    xpager_clean_data,
    xpager_flush_data
};


kern_return_t
set_xpager_request(memory_object, rq)
    memory_object_t memory_object;
    xpager_request_t *rq;
{
    object_t object;
    xpager_request_t nrq;

    object = object_lookup(memory_object);
    if (object == OBJECT_NULL) {
	return KERN_FAILURE;
    }
    object_lock(object);
    nrq = *rq;
    *rq = object->request;
    object->request = nrq;
    object_unlock(object);
    return KERN_SUCCESS;
}

kern_return_t
get_xpager_control(memory_object, xpc)
    mach_port_t memory_object;
    xpager_control_t xpc;
{
    bcopy(&xp_control, xpc, sizeof(xp_control));
    return KERN_SUCCESS;
}


/*
 * Print XPAGER statistics
 */
void
mid_xpager_print_stats()
{
#ifdef STATS
  int i, len, slen;
  char smsg[256], tmp[256];
#define MAX_STR_LEN 80

  StatusMsg("-=-=- External-Pager Statistics -=-=-");
  sprintf(smsg, "%lu data reqs, %lu zeroed, %lu data writes, %lu returns, %lu access reqs",
	  xpager_stat.data_reqs, xpager_stat.data_zeroed,
	  xpager_stat.data_returns, xpager_stat.data_writes,
	  xpager_stat.data_access_reqs);
  StatusMsg(smsg);
  sprintf(smsg, "%lu shared pages zeroed, %lu filled, %lu dirty pages filled",
	  xpager_stat.shared_zeroed, xpager_stat.shared_filled,
	  xpager_stat.dirty_filled);
  StatusMsg(smsg);
  sprintf(smsg, "%lu backing reads, %lu writes, %lu releases",
	  xpager_stat.backing_reads, xpager_stat.backing_writes,
	  xpager_stat.backing_releases);
  StatusMsg(smsg);
#ifdef PTIME
  sprintf(smsg, "mem_obj_lock: RO %lu (%lu) = %f, RW %lu (%lu) = %f, NO %lu (%lu) = %f",
	  ptime_ro_count, ptime_ro_total,
	  (ptime_ro_count == 0) ? 0.0 : 
	    (double)(((double)ptime_ro_total)/((double)ptime_ro_count)),
	  ptime_rw_count, ptime_rw_total,
	  (ptime_rw_count == 0) ? 0.0 : 
	    (double)(((double)ptime_rw_total)/((double)ptime_rw_count)),
	  ptime_no_count, ptime_no_total,
	  (ptime_no_count == 0) ? 0.0 : 
	    (double)(((double)ptime_no_total)/((double)ptime_no_count)));
  StatusMsg(smsg);
#endif
#ifdef MEM_STATS
  sprintf(smsg, 
	  "vm_alloc: %lu (%luKB) Priv %lu (%luKB) MidLib %lu (%luKB) Dealloc %lu (%luKB)",
	  xpager_stat.alloc_calls, xpager_stat.alloc_space/1024, 
	  xpager_stat.alloc_priv_calls, xpager_stat.alloc_priv_space/1024,
	  xpager_stat.alloc_lib_calls, xpager_stat.alloc_lib_space/1024, 
	  xpager_stat.dealloc_lib_calls, xpager_stat.dealloc_lib_space/1024);
  StatusMsg(smsg);

  sprintf(smsg, "vm_alloc size(2^):");
  slen = strlen(smsg);
  for (i = 0; i < STAT_TABLE_SIZE; i++) {
    if (xpager_stat.alloc_size_count[i] == 0) continue;
    
    sprintf(tmp, " %d=%lu", i+STAT_TABLE_MIN_SHIFT, 
	    xpager_stat.alloc_size_count[i]);
    len = strlen(tmp);
    if ((len + slen) > MAX_STR_LEN) {
      StatusMsg(smsg);
      sprintf(smsg, "vm_alloc size(2^):");
      slen = strlen(smsg);
    }
    strcat (smsg, tmp);
    slen += len;
  }
  StatusMsg(smsg);

  sprintf(smsg, "vm_alloc(priv):");
  slen = strlen(smsg);
  for (i = 0; i < STAT_TABLE_SIZE; i++) {
    if (xpager_stat.alloc_priv_size_count[i] == 0) continue;
    
    sprintf(tmp, " %d=%lu", i+STAT_TABLE_MIN_SHIFT, 
	    xpager_stat.alloc_priv_size_count[i]);
    len = strlen(tmp);
    if ((len + slen) > MAX_STR_LEN) {
      StatusMsg(smsg);
      sprintf(smsg, "vm_alloc(priv):");
      slen = strlen(smsg);
    }
    strcat (smsg, tmp);
    slen += len;
  }
  StatusMsg(smsg);

  sprintf(smsg, "vm_alloc(midlib):");
  slen = strlen(smsg);
  for (i = 0; i < STAT_TABLE_SIZE; i++) {
    if (xpager_stat.alloc_lib_size_count[i] == 0) continue;
    
    sprintf(tmp, " %d=%lu", i+STAT_TABLE_MIN_SHIFT, 
	    xpager_stat.alloc_lib_size_count[i]);
    len = strlen(tmp);
    if ((len + slen) > MAX_STR_LEN) {
      StatusMsg(smsg);
      sprintf(smsg, "vm_alloc(midlib):");
      slen = strlen(smsg);
    }
    strcat (smsg, tmp);
    slen += len;
  }
  StatusMsg(smsg);

  sprintf(smsg, "vm_dealloc(midlib):");
  slen = strlen(smsg);
  for (i = 0; i < STAT_TABLE_SIZE; i++) {
    if (xpager_stat.dealloc_lib_size_count[i] == 0) continue;
    
    sprintf(tmp, " %d=%lu", i+STAT_TABLE_MIN_SHIFT, 
	    xpager_stat.dealloc_lib_size_count[i]);
    len = strlen(tmp);
    if ((len + slen) > MAX_STR_LEN) {
      StatusMsg(smsg);
      sprintf(smsg, "vm_dealloc(midlib):");
      slen = strlen(smsg);
    }
    strcat (smsg, tmp);
    slen += len;
  }
  StatusMsg(smsg);
#endif /* MEMSTATS */


#else
  StatusMsg("Not compiled to collect XPAGER statistics.");
#endif /* STATS */
  return;
} /* mid_xpager_print_stats */
