/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/* 
 * Mach Operating System
 * Copyright (c) 1989 Carnegie-Mellon University
 * All rights reserved.  The CMU software License Agreement specifies
 * the terms and conditions for use and redistribution.
 */

#include "midway-dbg.h"   /* get defs for STATS & BENCHMARK */


kern_return_t xpager_init();
extern boolean_t	midway_xpager_may_cache;
extern mach_port_t		xpager_port;



/*
 * xpager_requests can be registered to force the external pager
 * to invoke higher level code on an xmm request that can not be satisfied
 * by the external pager's local cache.
 *
 * kern_return_t
 * data_request(offset, len, prot, &buf, &precious):
 *	vm_offset_t offset;
 *	vm_size_t   len;
 *	vm_prot_t   *prot;	
 *	vm_pointer_t *buf;
 *	boolean_t   *precious;
 *		called when xpager needs some data.
 *		callee should supply the data in buf.
 *		precious should be TRUE if notification is needed
 *		when the data is removed from the kernel,even if it is
 *		clean.
 *
 * kern_return_t 
 * databuf_free(buf)
 *	vm_pointer_t buf;
 *		called when xpager releases buffer.
 *
 * kern_return_t 
 * data_write(address, len, buf,discarded);
 *	vm_offset_t offset;
 *	vm_size_t   len;
 *	vm_pointer_t *buf;
 *	boolean_t   *discarded;
 *		called when xpager has some data that has been discarded
 *		by the kernel. *discard should be returned as true if the
 *		callee has properly discarded of the data. Otherwise,
 *		xpager will hang on to the data.
 *
 * kern_return_t 
 * data_access_request(offset, len)
 *	vm_offset_t offset;
 *	vm_size_t   len;
 *		called when xpager needs to allow some access to page.
 * 
 */


typedef struct xpager_request {
    char		*data_any;		/* passed on calls */
    kern_return_t 	(*data_request)();	/* xmm needs data */
    kern_return_t	(*databuf_free)();	/* free buffer */
    kern_return_t	(*data_write)();	/* data written out */
    kern_return_t	(*data_access_request)();  /* client needs access */
} *xpager_request_t;



/*
 * xpager_control is used to allow clients of the external pager
 * to control the behavior of the underlying mapped object.
 * These operations all take the memory_object returned by xpager_create.
 *
 * protect_data(memory_object, offset, len, protection)
 *	mach_port_t memory_object;
 *	vm_offset_t offset;
 *	vm_size_t   len;
 *	vm_prot_t   protection;
 *		Call to set protection to range of data.
 *
 * clean_data(memory_object, offset, len)
 *	mach_port_t memory_object;
 *	vm_offset_t offset;
 *	vm_size_t   len; 
 *		Remove the data from kernel's cache.  Will call
 *		data_write if the data has been dirtied, or
 *		is precious.
 *
 * flush_data(memory_object, offset, len)
 *	mach_port_t memory_object;
 *	vm_offset_t offset;
 *	vm_size_t   len; 
 *		Remove the data from kernel's cache. If the data
 *		is dirty, it is thrown away.
 */
 
typedef struct xpager_control {
    kern_return_t	(*protect_data)();  /* set access to data */
    kern_return_t	(*clean_data)();    /* write data  */
    kern_return_t	(*flush_data)();    /* throw it away */
} *xpager_control_t;





/*
 * set the request functions used by xpager. Return the previous
 * set.
 */
kern_return_t
set_xpager_request(/*mach_port_t mach_port_t,
		     xpager_request_t *new_and_old*/  );


/*
 * get the control functions to be used for the external pager
 */
kern_return_t get_xpager_control(/* mach_port_t, xpager_control_t */);



/*
 * Compile is tracing code when debugging only!
 */
#ifdef BENCHMARK
# define mumble (void)
#endif


#ifdef STATS

#ifdef _XPAGER_INTERNAL_
#  define EXTERN 
#else
#  define EXTERN extern
#endif

#define STAT_TABLE_MIN_SHIFT	12
#define STAT_TABLE_MIN_SIZE    (1 << STAT_TABLE_MIN_SHIFT) /* 2^12 = 4096 */
#define STAT_TABLE_MAX_SHIFT    32
#define STAT_TABLE_SIZE        (STAT_TABLE_MAX_SHIFT - STAT_TABLE_MIN_SHIFT)

EXTERN struct {
  unsigned long data_reqs;
  unsigned long data_writes;
  unsigned long data_returns;
  unsigned long data_zeroed;
  unsigned long data_access_reqs;
  unsigned long shared_zeroed;
  unsigned long shared_filled;
  unsigned long dirty_filled;
  unsigned long backing_reads;
  unsigned long backing_writes;
  unsigned long backing_releases;

  unsigned long	alloc_calls;		/* # of calls to vm_allocate */
  unsigned long	alloc_space;		/* total size of space allocated */
  unsigned long alloc_size_count[STAT_TABLE_SIZE]; /* power of 2 table of sizes */
  unsigned long	alloc_priv_calls; 	/* # of calls to vm_allocate for private space */
  unsigned long	alloc_priv_space;	/* total size of private space allocated */
  unsigned long alloc_priv_size_count[STAT_TABLE_SIZE]; /* power of 2 table of sizes */
  unsigned long	alloc_free_list;	/* # of calls for 1 page of private space */
  unsigned long	alloc_lib_calls;	/* # of calls to vm_allocate from midway library */
  unsigned long	alloc_lib_space;	/* total size of space allocated from midway lib */
  unsigned long alloc_lib_size_count[STAT_TABLE_SIZE]; /* power of 2 table of sizes */
  unsigned long dealloc_lib_calls;  	/* # of calls to vm_deallocate from midway library */
  unsigned long dealloc_lib_space;  	/* total size of space deallocated from midway lib */
  unsigned long dealloc_lib_size_count[STAT_TABLE_SIZE]; /* power of 2 table of sizes */
} xpager_stat;
EXTERN mutex_t xpager_stat_lock;


#ifdef FAST_STATS 
#  define INCR_STAT(x,by) (x) += (by)
#else
#  define INCR_STAT(x,by) \
   { mutex_lock(xpager_stat_lock); (x) += (by); mutex_unlock(xpager_stat_lock); }
#endif

#else  /* not STATS */
#  define INCR_STAT(x,by)
#endif


     
    
    



			   


