/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/* 
 * Mach Operating System
 * Copyright (c) 1989 Carnegie-Mellon University
 * All rights reserved.  The CMU software License Agreement specifies
 * the terms and conditions for use and redistribution.
 */
/*
 * HISTORY
 * $Log: xpager_obj.c,v $
 * Revision 1.7  1994/06/16  03:12:11  wsawdon
 * Updated stats on malloc/vm_alloc/etc.
 *
 * Revision 1.6  1994/06/15  03:03:55  wsawdon
 * Collect stats on vm_alloc to check memory usage.
 *
 * Revision 1.5  1993/07/02  04:48:23  wsawdon
 * Page shared/dirty memory to disk rather than to non-shared memory.
 * Also fixed bug in freeing temp buffers used by backing store.
 *
 * Revision 1.4  1992/12/02  20:49:54  mattz
 * Add mutex name support.
 *
 * Revision 1.3  1992/10/24  21:43:23  wsawdon
 * Fixed 2 bugs in external pager.
 *   Page data passed back on a data_request must be page aligned.
 *   Always give R/W access even if R/O requested.
 *
 * Revision 1.2  1992/10/04  23:42:22  mattz
 * Remove verbosity.  debug flags changed name, now placed under cmd line flag
 * control.
 *
 * Revision 1.1  1992/10/01  20:00:58  mattz
 * External Pager for the Midway runtime.  Keeps track of shared storage.
 *
 */


#include <cthreads.h>
#include "backing.h"
#include "xpager.h"
#include "xpager_obj.h"
#include "midway-dbg.h"		/* MIDWAY_MUTEX_SET_NAME */


static object_t object_list = OBJECT_NULL;

#ifdef STANDALONE
int midway_xpager_debug = 1;
int midway_xpager_use_syslog = 0;
#endif

int midway_xpager_may_cache = 0;

static void
handle_port_death(port_name)
    mach_port_t port_name;
{
    mumble(". port_death %d\n", port_name);
}

object_t
object_lookup(memory_object)
    mach_port_t memory_object;
{
    object_t object;

    for (object = object_list; object; object = object->next) {
	if (object->memory_object == memory_object) {
	    return object;
	}
    }
    return OBJECT_NULL;
}

object_t
object_create(filename)
    char *filename;
{
    object_t object;
    kern_return_t status;
    mach_port_t memory_object;

    status = mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &memory_object);
    if (status != KERN_SUCCESS) {
	fret("object_create: port_allocate: %s\n", ErrString(status));
	return OBJECT_NULL;
    }
    object = (object_t) MALLOC(sizeof(*object));
    if (! object) {
	fret("object_create: malloc failed\n");
	return OBJECT_NULL;
    }
    object->backing_store = backing_store_create();    
    object->state = O_STATE_READY;
    object->memory_object = memory_object;
    object->memory_control = MACH_PORT_NULL;
    object->buf = (char *) 0;
    object->r_head = 0;
    object->r_tail = 0;
    object->max_req = 0;
    object->request = 0;
    object->control = 0;
    object->mu = mutex_alloc();
    MIDWAY_MUTEX_SET_NAME(object->mu, "xpager_obj.c:object_mu");
    object->next = object_list;
    object_list = object;
    
    return object;
}

object_destroy(object)
    object_t object;
{
    mutex_t mu = object->mu;

    mutex_lock(mu);
    
    mach_port_deallocate(mach_task_self(), object->memory_object);
    mach_port_deallocate(mach_task_self(), object->memory_control);
    if (object->buf) {
	FREE(object->buf);
    }
    backing_store_release(object->backing_store);
    /* XXX need mutex before able to free object  ?? */    
    FREE(object);
    
    mutex_unlock(mu);
    mutex_free(mu);

    return (KERN_SUCCESS);
}

void
object_lock(object)
    object_t object;
{
    mutex_lock(object->mu);
}

void
object_unlock(object)
    object_t object;
{
    mutex_unlock(object->mu);
}


