/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * IO Server for Midway
 *
 * 14 Dec 90: Matthew J. Zekauskas
 * 21 Jun 91 (mjz): add host number
 *  6 Aug 91 (bnb&mjz): convert status server to io redirection server
 *  3 Oct 92 (mjz): add open,close & lseek; fix write bug (no bufCnt)
 * 14 Oct 92 (mjz): update for arbitrary sized buf passing (array[])
 * 18 Nov 92 (mjz): add locking because we're losing printfs.  Sigh.
 */


#include <stdio.h>
#include <mach.h>
#include <mig_errors.h>
#include <mach/message.h>
#include <servers/netname.h>
#include <strings.h>
#include "midwayux_types.h"
#include <cthreads.h>

extern boolean_t	midwayux_server();	/* MIG supplied */
extern int errno;
vm_task_t mytask = (vm_task_t) 0; /* init to zero */

extern mutex_t midway_ss_stderr_lock;


/*
 * ------------------- the supplied functions ------------------------
 */

kern_return_t
midwayux_write(server_port, hostid, fd, buf, bufCnt, len, uxerrno, cnt)
    mach_port_t server_port;
    long hostid;		/* (index of) host making request */
    int fd;
    char *buf;
    int bufCnt;			/* magic MiG length for v-length array */
    int len;
    int *uxerrno;
    int *cnt;
{
    if ((fd == 1) || (fd == 2)) {
      mutex_lock(midway_ss_stderr_lock);
    }
    *cnt = write(fd, buf, len);
    if (*cnt < 0)  {
	*uxerrno = errno;
    }
    if ((fd == 1) || (fd == 2)) {
      mutex_unlock(midway_ss_stderr_lock);
    }
    return(KERN_SUCCESS);
} /* midwayux_write */


kern_return_t
midwayux_read(server_port, hostid, fd, buf, bufCnt, bufDlc, len, uxerrno, cnt)
    mach_port_t server_port;
    long hostid;		/* (index of) host making request */
    int fd;
    char **buf;
    int *bufCnt;		/* magic MiG length for v-length out array */
    boolean_t *bufDlc;		/* magic boolean: true, dealloc buf on reply */
    int len;
    int *uxerrno;    
    int *cnt;
{
    kern_return_t krc;
    char * savebuf;

    *bufDlc = FALSE;		/* assume inline */
    if (len > *bufCnt) {	/* won't fit in-line */
      if (mytask == (vm_task_t) 0) {
	mytask = mach_task_self(); /* cache */
      }
      savebuf = *buf;		/* save old pointer */
      *buf = 0;			/* start at bottom */
      krc = vm_allocate(mytask, buf, len, TRUE /* anywhere */);
      if (krc != KERN_SUCCESS) {
	*uxerrno = krc;		/* pass back to caller */
	*bufCnt = 0;
	*buf = savebuf;
	return(KERN_SUCCESS);
      }
      *bufDlc = TRUE;
    }
    *cnt = read(fd, *buf, len);
    if (*cnt < 0)  {
	*uxerrno = errno;
	*bufCnt = 0;
	if (*bufDlc == TRUE) {
	  vm_deallocate(mytask, *buf, len); /* ensure all deallocated */
	  *bufDlc = FALSE;
	  *buf = savebuf;
	}
    } else {
      if (*bufDlc == TRUE) {
	if (*bufCnt > *cnt) {	/* would actually fit in line.  Copy away... */
	  bcopy(*buf, savebuf, *cnt);
	  vm_deallocate(mytask, *buf, len);
	  *buf = savebuf;
	  *bufDlc = FALSE;
	  *bufCnt = *cnt;
	} else {
	  *bufCnt = len;	/* want to deallocate all, but may send more */
	}
      }
      else
        *bufCnt = *cnt;		/* ...but not copy */
    }
    return(KERN_SUCCESS);
} /* midwayux_read */

kern_return_t
midwayux_open(server_port, hostid, path, pathCnt, flags, mode, uxerrno, fd)
    mach_port_t server_port;
    long hostid;		/* (index of) host making request */
    char *path;
    int pathCnt;		/* magic MiG length for v-length array */
    int flags;
    int mode;
    int *uxerrno;    
    int *fd;
{
    *fd = open(path, flags, mode);
    if (*fd < 0)  {
	*uxerrno = errno;
    } 
    return(KERN_SUCCESS);
} /* midwayux_open */

kern_return_t
midwayux_lseek(server_port, hostid, fd, offset, whence,  uxerrno, pos)
    mach_port_t server_port;
    long hostid;		/* (index of) host making request */
    int fd;
    int offset;
    int whence;
    int *uxerrno;    
    int *pos;
{
    *pos = lseek(fd, offset, whence);
    if (*pos == -1)  {
	*uxerrno = errno;
    }
    return(KERN_SUCCESS);
} /* midwayux_lseek */

kern_return_t
midwayux_close(server_port, hostid, fd, uxerrno, rc)
    mach_port_t server_port;
    long hostid;		/* (index of) host making request */
    int fd;
    int *uxerrno;    
    int *rc;
{
    *rc = close(fd);
    if (*rc == -1)  {
	*uxerrno = errno;
    }
    return(KERN_SUCCESS);
} /* midwayux_close */

/*
 * ------------------- the support functions ------------------------
 */







/*
 * ------------------- standalone test function ------------------------
 * (with uxcl.c)
 */


#ifdef STANDALONE
main(argc, argv)
    int argc;
    char **argv;
{
    mach_port_t server;
    mach_port_t server_set;
    kern_return_t krc;
    int maxmsglen = 2048+1024; /* 2K for in-line Vlength array + 1k for slop */
    /* 1024 was definitely too small. */
    
    fprintf(stderr,"starting up io/server\n");
    allocate_server_ports(&server, &server_set, "MidwayUxServer");
    krc = mach_msg_server( midwayux_server, maxmsglen, server_set);
    if (krc != KERN_SUCCESS)  {
	mach_error("mach_msg_server returns:", krc);
	exit(krc);
    }
}

#endif
  








kern_return_t
midwayux_ping(server_port)
    mach_port_t server_port;
{
    fprintf(stderr,"MIDWAYUX_PONG\n");
    return(KERN_SUCCESS);
}


