/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */

/*
 * comm.c: communications routines for Midway
 *
 *
 * StatusMsg - send a status message to the status server; depends on
 *		CommInit being issued, and is used by comm routines
 *
 * External Interfaces:
 * ***AS OF 17 MAR 91, MOST OF THESE DESCRIPTIONS ARE OBSOLETE***
 * CommInit(port_t ss_port) - initialize static variables/mutex/condition
 * CommOpen() - open a socket listening on unused UDP port
 * CommRequest(?) - send request, wait for answer
 * CommGetReq(?) - get next request
 * CommSendReply(?) - reply to request
 *
 * Internal Interfaces (but are exported anyway):
 * CommSend - send a request or reply
 * CommScanner - does retransmission strategy
 * CommRecv - receive a packet from net
 * CommListener - loop, recieving; getting partial packets,
 *		performing acks if necessary, and then dispatching full
 *		requests or replies to either waiting requestors or
 *		worker threads (to handle requests) (via outstanding
 *		CommGetReq's)
 * "upper" layers provide the worker processes to handle requests
 *
 * HISTORY
 * 16 Dec 90: Matthew J. Zekauskas
 *  9 Oct 93 (mjz): updated for new integrated kernel [MK83+UX41]
 *		- new FAST_UDP interface (june/july 83)
 *		- have workers use pbufs instead of static storage
 * 18 Oct 93 (was): added call to shmalloc_stats
 * 25 Oct 93 (mjz): removed serialization of sends, so atm can do it if
 *		necessary (or (in particular) not, if not)
 *  4 May 94 (was): changed CommReply to take use a bitarray for the nacks
 *		rather than passing it as variable length part of the data.
 *              Also changed format of COMMMSG and optimized CommSend
 */
#include <sys/types.h>		/* u_long & friends */

#include <errno.h>
#include <stdio.h>
#include "midway_types.h"       /* global midway type definitions */
#include "comm.h"		/* includes bitwiddle.h--bit twiddling rtns */
#include "pbuf.h"               /* for pbuf definitions */
#include "sm.h"		/* for UDPPORTSTART - move to comm.h? */
#ifndef NOSS			/* no Status Server */
#  include "ss.h"		/* for char255  + ss_msg defn; MiG generated */
#else /*NOSS*/
#  include "smexecd_types.h"	/* for char255 */
#  define ss_msg(a1,a2,a3,a4) fprintf(stderr,"-?>ss_msg() called!\n")
#endif /*NOSS*/
#include "midway-dbg.h"		/* conditional debugging */
#include "updateq.h"            /* prototypes for calling update queue rtns */
#include "pagemgr.h"            /* prototypes for calling page manager rtns */
#include <flow_times.h>
#include "atm_iface.h"		/* atm fn defns */
#include "socket_iface.h"	/* sockets fn defs */
#include <ctype.h>

#define TSTATS 0
#if TSTATS
#define KERNEL 1		/* HACK */
#include "timing.h"		/* for TSTATS */
#undef KERNEL
#endif /*TSTATS*/

/*
 * for scanner
 */
/*
 * scanflag states:
 *	0 - am sleeping waiting to be woken up
 *	1 - am scanning, and will go to sleep if I find no more work
 *	2 - will scan again after timeout
 */
#define	SCAN_SLEEP	0
#define	SCAN_WORK	1
#define SCAN_AGAIN	2

extern char *mach_error_string();

/* locally global vars */

static initdone = 0;		/* for sanity */
/* un-static for ss_debug_threat 29 jun 92 mjz */
mach_port_t ss_port = 0;		/* status server port */
static char255 hname;		/* local host name for messages */
static char *hlist;		/* list of names of hosts */
htblent_t *htbl;		/* gonna point to an array of them */
int midway_udpsock;		/* Sigh. Keep here too... Why pass? */

#ifdef DEBUG_PRT
static int (*outcall)();	/* to print outgoing messages */
static int (*incall)();		/* to print incoming messages */
#endif /*DEBUG_PRT*/

/* I was going to ifdef COMM_LOCK_RCV these, but they could be used
 * by CommGetReq instead of CommRecv */
mutex_t commlock;		/* so only one receive outstanding */
condition_t commsignal;		/* so others can wait */
int listener_active;		/* so we know we're waiting */

#ifdef SERIALIZE_SENDS
mutex_t scommlock;		/* ditto for senders; really need? *mjz* */
condition_t scomm_signal;
int sender_active;		/* not sure if we need this, but? *mjz* */
#endif /*SERIALIZE_SENDS*/


u_long	CommThreadMax = 100;	/* *MAGIC*, but at least redefinable */
cpt_t	*cpt;
u_long	cptnext;		/* next entry to alloc in cpt array */
mutex_t cptlock;		/* since multiple threads update cptnext */
mutex_t scan_mu;		/* for scanner of CPT table */
condition_t scan_cv;
int	scanflag = SCAN_SLEEP;	/* say something to do; always under lock */

u_long	CommWorkerNum = 16;	/* *MAGIC*: number of worker threads */

/*
 * transaction table; records outstanding and completed transactions.
 * This will be used for requests only.  Since replies go to a particular
 * cpt entry, replies to completed transactions can always be dropped.
 * If Outstanding, we will drop as "in process"; if completed, we will
 * forward to the lock layer with a new "already done" classification,
 * so it knows *only* to retransmit, never to treat as a new request.
 * This could be changed if we keep reply buffers for outstanding
 * transactions.
 */
typedef struct {
#if (BENCHMARK == 0)		/* check sanity */
  u_long	signature;
#define TRANTAB_SIG	0xfeedface
#endif        
  u_long	tranid;		/* transaction id */
  union {
    u_long	state;		/* of this transaction if hi order bit on */
    u_long	fwd;		/* if forwarded, where to (hi order bit off)*/
    /* see comm.h, M_R_xxxx for possible values */
  } u;
  u_long	waiter;		/* if state=M_R_WAITING who is waiting */
  u_long	pstart;
  u_long	plist;
  u_long	pl_shift;	/* shift to test for all if lastpacket. could overload waiter? */
  u_long	inproc_max;
  u_long	inproc;		/* SH*T. I need a connection record! */
} trantab_t;
trantab_t *trantab;
mutex_t trantab_lock;		/* protect access */
condition_t trantab_c;		/* Barf */
/* how to get to trantab entry, or DIY array references */
#define TRANTAB(h,t) (trantab+(((h)*CommThreadMax)+(t)))

/* really global vars */
int  Comm_hnum;			/* number in list of names... */
int  Comm_index;		/* index of this host in htbl/hlist */
u_long Comm_incarn;		/* incarnation # of this execution */
int midway_comm_reply_ackafter = 8; /* HACK HACK HACK */
int midway_comm_reply_ackafter_atm = 100000; /* large  -- reset for ATM*/
int midway_comm_send_ackafter = 8;
int midway_comm_send_ackafter_atm = 16;	/* fix code so not <= 32, matt */

/*
 * priorities, see CommOpen, and others, below. (SET_PRIORITY)
 */
int midway_comm_worker_prio = 0;
int midway_comm_scanner_prio = 0;
int midway_user_priority = 0;
int midway_gprof_priority = 0;	/* used in gprof_support.c */

clock_t LocalClock;		/* local clock -- maintained by us? */
mutex_t ClockLock;		/* he he */
/* lock clock before update?  multiple threads????? */

/* types defined in midway_types.h */
int midway_comm_network_type = MIDWAY_NET_SOCKETS;

/* if standalone (isolated network) do any hacks necessary */
int midway_comm_network_standalone = 0;

typedef struct comm_net_fns {
  int (*rcv)    (char **buf, int *len, u_int *ftstart, 
		 struct sockaddr *sin);
  int (*sendto) (int s, char *msg, int len, int flags,
		 struct sockaddr *to, int tolen, int dest);
  int (*sendmsg)(int s, struct msghdr msg[], int flags, int dest);
  int (*free)   (pbuf x);
  int (*wantreply)(comm_filter filt, int cpti);
} comm_net_fns_t;

comm_net_fns_t midway_sock_fns = { 
  midway_socket_rcv, 
  midway_socket_sendto, 
  midway_socket_sendmsg,
  mid_pbuf_free,
  0				/* not yet available */
  };

comm_net_fns_t midway_atm_fns = {
  midway_atm_rcv,
  midway_atm_sendto,
  midway_atm_sendmsg,
  midway_atm_pbuf_free,
  midway_atm_wantreply
  };

comm_net_fns_t *midway_net_fns = &midway_sock_fns;



extern int	errno;
extern char	*sys_errlist[];	/* printable errno messages */
extern int	sys_nerr;	/* # of those messages */
extern char *MALLOC();

extern task_t midway_task_self;	/* mach_task_self() */


/* 
 * Define macro to conditionally maintain stat info
 */
#ifdef STATS

#define MSTATS 1		/* more detailed 'message' stats  */

static struct {
  u_long	snd;		/* # packets sent */
  u_long	ack;		/* # of those that were acks */
  u_long        multireqs;      /* # of multicast requests */
  u_long        multisnds;      /* # of snds that were multicasts */
  u_long	retran;		/* # of snd  that were retransmissions */
  u_long	retranack;	/* # of those that were Ack Requests */
  u_long	retranmulti;	/* # of those that were mulitcast reqs */
  u_long	lastpack;	/* largest packet burst sent */
  u_long	fwd;		/* packet was forwarded */
  u_long	rcv;		/* # packets received */
  u_long	badclk;		/* dropped due to time going backwards */
  u_long	badtid;		/* those dropped due to bad tid */
  u_long	badtidretrans;	/* bad tid due to retransmission */
  u_long	ipp;		/* in process packet dropped */
  u_long	again;		/* dropped; multi-pack reply & here */
  u_long	mpseqfail;	/* # of multi-packet sequence failures */
  				/* (this probably implies a dropped net pkt)*/
  u_long        sndBytes;       /* # bytes sent */
  u_long        rcvBytes;       /* # bytes received */
  u_long        fwdBytes;       /* # bytes forwarded */
  u_long        multiBytes;     /* # bytes multicast (multisnds*sizeof(pkt) */
  u_long	useless;	/* # of requests that were serviced, but the
				   transaction id was updated by the time they
				   finished, so the work was redundant.*/
  u_long        ackReqFilter1;  /* number of acks sent from CommRequestFilter */
  u_long        ackReqFilter2;
  u_long        ackReqFilter3;
  u_long        ackRecv1;	/* number of acks sent from CommRecv */
  u_long        ackRecv2;
  u_long        ackRecv3;
  u_long        ackWorker1;	/* number of acks sent from CommWorker */
  u_long        ackWorker2;
  u_long        ackWorker3;

#if MSTATS
  /*lock*/
  u_long	reqlout;	/* # of request "msgs" thru commsend */
  u_long	replout;	/* # of reply "messages" thru commsend */
  u_long	reqlin;		/* # of request "msgs" thru worker */
  u_long	replin;		/* # of reply messages thru worker */
  /*barrier*/
  u_long	reqbout;	/* # of request "msgs" thru commsend */
  u_long	repbout;	/* # of reply "messages" thru commsend */
  u_long	reqbin;		/* # of request "msgs" thru worker */
  u_long	repbin;		/* # of reply messages thru worker */
  /* other */
  u_long	reqout;		/* # of request "msgs" thru commsend */
  u_long	repout;		/* # of reply "messages" thru commsend */
  u_long	reqin;		/* # of request "msgs" thru worker */
  u_long	repin;		/* # of reply messages thru worker */
  u_long        csreqin;        /* # of copyset messages */
  u_long        csrepin;        /* # of copyset messages */
  u_long        pfreqin;        /* # of page fault requests */
  u_long        pfrepin;        /* # of page fault replies */
  u_long        pureqin;        /* # of page update requests */
  u_long        purepin;        /* # of page update replies */
#endif

#if TSTATS
  double	copytime;	/* usecs spent copying data in lock replies */
  u_long	copynum;
#endif

/*#if ATM_IFACE*/
  u_long	replfilt;	/* # of filter calls */
  u_long	fastreply;	/* # of acks w/o context switching */
  u_long	repfset;	/* # of filter sets */
  u_long	repfsetint;	/* # where intr satisfied before spinning */
  u_long	repflmiss;
  u_long	reqffilt;
  u_long	reqflmiss;	/* missed comm lock */
  u_long	reqftmiss;	/* missed tid after locking (rare) */
  u_long	reqfbmiss;	/* missed due to call would block  */
  u_long	reqfomiss;	/* missed due to op being wrong */
  u_long	reqfmmiss;	/* missed due to not what we expected  */
  u_long	reqfmtype;	/* last type that missed */
  u_long	reqffast;	/* worked */
/*#endif*/

} mid_comm_stats;
static mutex_t statlock;	/* just in case */


# ifdef FAST_STATS
#  define INCR_STAT(x,by) (x) += (by)
# else 
#  define INCR_STAT(x,by) \
    { mutex_lock(statlock);  (x) += (by);  mutex_unlock(statlock); }
# endif 

#else /* not STATS */
#define INCR_STAT(x,by)   /* nothing to do  */
#endif

#if FLOW_TIMES
DEFINE_FT;			/* put definitions here */
#endif


/*
 * forward references
 */
void midPlistClear();

/*
 ************************************************************************
 *			S t a t u s M s g				*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
#include <syscall.h>		/* for error case below */
static int StatusMsgCalls = 0;
StatusMsg(msg)
     char	*msg;		/* message to send to status server */
{
  register kern_return_t krc;
  if (ss_port) {
    INCR_STAT(StatusMsgCalls,1); /* protect with hammer */
    krc = ss_msg(ss_port,LocalClock,hname,Comm_index,msg);
    if (krc != KERN_SUCCESS) {
      int len;
      char xmsg[256];
      len = strlen(msg);
      syscall(SYS_write,2/*stderr*/,msg,len); /* get it out */
      sprintf(xmsg,"Error from ss_msg: %s",mach_error_string(krc));
      len = strlen(xmsg);
      syscall(SYS_write,2,xmsg,len);
    }
  }
  else {
    fprintf(stderr,msg);
    putc('\n',stderr);
  }
} /* StatusMsg */

/*
 ************************************************************************
 *		m i d _ c o m m _ p r i n t _ s t a t s			*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
int midway_print_flow_times_ok = 1; /* default OK to print flow_times */
mid_comm_print_stats()
{
#ifndef STATS
  StatusMsg("Not compiled to collect Communication Statistics");
#else

  char255 smsg;

  StatusMsg("-=-=- Communication Statistics -=-=-");
  sprintf(smsg, "%lu packets sent: %lu ack %lu fwd %lu retrans %lu inv ack req retrans; largest burst %lu",
	  mid_comm_stats.snd, mid_comm_stats.ack,
	  mid_comm_stats.fwd, mid_comm_stats.retran, 
	  mid_comm_stats.retranack, mid_comm_stats.lastpack);
  StatusMsg(smsg);
  sprintf(smsg,"%Multicast %lu requests, %lu packets (%lu retrans), %lu bytes",
	  mid_comm_stats.multireqs, mid_comm_stats.multisnds, 
	  mid_comm_stats.retranmulti, mid_comm_stats.multiBytes);
  StatusMsg(smsg);
  sprintf(smsg, "%lu packets rcvd: %lu badclk %lu badtid %lu in process %lu mpack dups %lu mp seqfail",
	  mid_comm_stats.rcv, mid_comm_stats.badclk, mid_comm_stats.badtid,
	  mid_comm_stats.ipp, mid_comm_stats.again, mid_comm_stats.mpseqfail);
  StatusMsg(smsg);
  sprintf(smsg, "%lu bytes sent.  %lu bytes received.  %lu bytes forwarded",
	  mid_comm_stats.sndBytes, mid_comm_stats.rcvBytes, 
	  mid_comm_stats.fwdBytes);
  StatusMsg(smsg);
  sprintf(smsg, "%lu positively redundant requests handled.",
	  mid_comm_stats.useless);
  StatusMsg(smsg);

  mid_pbuf_print_stats();


#if MSTATS
  sprintf(smsg, "lock: %lu rq in %lu rp in %lu rq out %lu rp out",
	  mid_comm_stats.reqlin, mid_comm_stats.replin,
	  mid_comm_stats.reqlout, mid_comm_stats.replout);
  StatusMsg(smsg);
  sprintf(smsg, "barrier: %lu rq in %lu rp in %lu rq out %lu rp out",
	  mid_comm_stats.reqbin, mid_comm_stats.repbin,
	  mid_comm_stats.reqbout, mid_comm_stats.repbout);
  StatusMsg(smsg);
  sprintf(smsg, "other: %lu rq in %lu rp in %lu rq out %lu rp out",
	  mid_comm_stats.reqin, mid_comm_stats.repin,
	  mid_comm_stats.reqout, mid_comm_stats.repout);
  StatusMsg(smsg);
  sprintf(smsg, "other2: %lu pfrq %lu pfrp %lu purq %lu purp %lu cprq %lu cprp",
	  mid_comm_stats.pfreqin, mid_comm_stats.pfrepin,
	  mid_comm_stats.pureqin, mid_comm_stats.purepin, 
	  mid_comm_stats.csreqin, mid_comm_stats.csrepin);
  StatusMsg(smsg);
  sprintf(smsg,"acks: ReqFltr %lu %lu %lu  Recv %lu %lu %lu  Worker %lu %lu %lu  Other %lu",
	  mid_comm_stats.ackReqFilter1, mid_comm_stats.ackReqFilter2, mid_comm_stats.ackReqFilter3,
	  mid_comm_stats.ackRecv1, mid_comm_stats.ackRecv2, mid_comm_stats.ackRecv3,
	  mid_comm_stats.ackWorker1, mid_comm_stats.ackWorker2, mid_comm_stats.ackWorker3,
	  mid_comm_stats.ack - (
 	  mid_comm_stats.ackReqFilter1 + mid_comm_stats.ackReqFilter2 + mid_comm_stats.ackReqFilter3
	   + mid_comm_stats.ackRecv1 + mid_comm_stats.ackRecv2 + mid_comm_stats.ackRecv3
	   + mid_comm_stats.ackWorker1 + mid_comm_stats.ackWorker2 + mid_comm_stats.ackWorker3));
  StatusMsg(smsg);
  sprintf(smsg, "%lu badtidretrans", 
	  mid_comm_stats.badtidretrans);
#endif
#if TSTATS
  sprintf(smsg, "%d lock requests for a total of %f seconds.",
	  mid_comm_stats.copynum, mid_comm_stats.copytime);
  StatusMsg(smsg);
#endif
#if ATM_IFACE
  sprintf(smsg, "crpf: %u set %u intrhit %d calls %d fast reply hits",
	  mid_comm_stats.repfset, mid_comm_stats.repfsetint,
	  mid_comm_stats.replfilt,
	  mid_comm_stats.fastreply);
  StatusMsg(smsg);
  sprintf(smsg, "crqf: %d calls %d hits %d lockmiss %d tidmiss %d blockmiss %d opmiss %d matchmiss 0x%lx matchmisstype",
	  mid_comm_stats.reqffilt, mid_comm_stats.reqffast,
	  mid_comm_stats.reqflmiss, mid_comm_stats.reqftmiss,
	  mid_comm_stats.reqfbmiss, mid_comm_stats.reqfomiss,
	  mid_comm_stats.reqfmmiss, mid_comm_stats.reqfmtype);
  StatusMsg(smsg);

  mid_atm_print_stats();
#endif
  if (midway_comm_network_type == MIDWAY_NET_ATM)
    mid_atm_print_stats();

#ifdef STATUSMSG_TEST
  fprintf(stderr,"%d: local StatusMsg calls: %d\n", Comm_index, 
	  StatusMsgCalls);
  ss_stats(ss_port);		/* force status server to acknowledge. */
#endif

#if CTHREAD_DEBUG
  {
    extern mutex_t midway_lock_head;
    mutex_t m = midway_lock_head;
    printf("%d: locks with contention: ",Comm_index);
    if (m == 0) {
      printf("none!\n");
      fflush(stdout);
    } else {
      while(m) {
	if (m->name != 0)
	  printf("%s %d cpc 0x%lx ",m->name,m->count,m->countpc);
	else
	  printf("[adr 0x%lx pc 0x%lx] %d cpc 0x%lx ",m,m->whopc,m->count,
		 m->countpc);
	m = m->next;
      }
      printf("\n");
      fflush(stdout);
    }
      
  }
#endif

#endif /* STATS */

#if FLOW_TIMES
  if (midway_print_flow_times_ok)
    print_flow_times();
#endif

  return 0;
} /* mid_comm_print_stats */

mid_comm_reset_stats()
{
#ifdef STATS
  bzero((char *)(&mid_comm_stats), sizeof(mid_comm_stats));
#endif
} /* mid_comm_reset_stats */

unsigned int
mid_comm_get_stats(int idx)	/* get stat value at index i */
{
#if STATS
  unsigned int *statlist = (unsigned long *) &mid_comm_stats;
  int max = (sizeof(mid_comm_stats) / sizeof(unsigned long));
  char255 smsg;

  if ((idx < 0) || (idx > max)) {
    sprintf(smsg,"mid_comm_get_stats index out of range [0,%d]: %d",
	    max, idx);
    StatusMsg(smsg);
#if STOP_ON_ERROR
    botch("mid_comm_get_stats");
#endif
    return(0);
  }
  return((unsigned int)statlist[idx]);
#else
  return(0);
#endif
} /* mid_comm_get_stats */

#ifdef NOTDEF
/*
 ************************************************************************
 *			C o m m F i n d S t a b				*
 *----------------------------------------------------------------------*
 * use this or something like it if keeping tids per host/lock or host  *
 * & remote tid (how I was doing it when this was written)		*
 * note:								*
 *	servertab_t **s_tab						*
 *	typedef struct servertab {					*
 *		u_long host;						*
 *		u_long it; / * remote thread * /			*
 *		u_long tid; / * current transaction id * /		*
 *		struct servertab *next;					*
 *	} servertab_t *servertab_pt;					*
 ************************************************************************
 */
static servertab_t *
CommFindStab(host, thread)
     u_long	host;		/* client host */
     u_long	thread;		/* client thread@host  */
{
  servertab_t ** p;

  mutex_lock(s_tab_lock);

  p = &(s_tab[host]);
  while (*p) {			/* search through s_tab entries */
    if ((*p)->thread == thread)  /* got it */
      break;
    *p = (*p)->next;		/* check next */
  }
  /* ok, either found bucket or at end of chain */
  if (*p == 0) {		/* not found -- new */
    *p = (servertab_t *) CALLOC(1, sizeof(servertab_t));
    ASSERT(*p != 0);		/* BARF MALLOC OUT OF SPACE --MJZ */
    (*p)-> host = host;		/* why is this here? -mjz */
    (*p)-> thread = thread;
  }

  mutex_unlock(s_tab_lock);

  return (*p);
} /* CommFindStab */
#endif /*NOTDEF*/

/*
 ************************************************************************
 *			    C o m m I n i t				*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
CommInit(ssp,clk,hindex,hnm,hlst,rhost,rport,incarn)
     mach_port_t	ssp;		/* status server port */
     clock_t	clk;		/* clock value */
     int	hindex;		/* this host's position in htbl/hlist */
     int	hnm;		/* entries in ... */
     char	*hlst;		/* list of host names we talk to */
     u_long	rhost;		/* ip addr and */
     u_short	rport;		/* udp port of host we talk to to get ports */
     u_long	incarn;		/* incarnation # of this execution */
{
  kern_return_t krc;
  char255	smsg;		/* just in case */
  int		i;
#if TSTATS
  extern int setup_mapped_timer();
#endif

#ifdef DEBUG_PRT
  extern int midPmsgIn();
  extern int midPmsgOut();
#endif

  if (initdone) {		/* Ack! */
    StatusMsg("error - duplicate CommInit call!");
    return(-1);
  }
  initdone = 1;


  /*
   * Validate typedef sizes 
   */
  ASSERT (MAX_COMM_HOSTID >= MAX_PROCESSORS);
  ASSERT (MAX_COMM_THREADID >= CommThreadMax);


#ifdef DEBUG_PRT
  incall = midPmsgIn;
  outcall = midPmsgOut;
#endif

  ss_port = ssp;		/* do it! */
  LocalClock = clk;
  ClockLock = mutex_alloc();

  MIDWAY_MUTEX_SET_NAME(ClockLock, "comm.c:ClockLock");
  gethostname(hname,sizeof(hname));
  hlist = hlst;
  Comm_index = hindex;
  Comm_hnum = hnm;
  Comm_incarn = incarn;

  htbl = (htblent_t *) CALLOC((hnm+1), sizeof(htblent_t));
  htbl[0].ip = rhost;		/* 0 is who we get ip/port list from */
  htbl[0].uport = rport;

  listener_active = 0;		/* nobody home yet */
  commlock = mutex_alloc();	/* for listeners */
  MIDWAY_MUTEX_SET_NAME(commlock, "commlock");
  commsignal = condition_alloc();

#ifdef SERIALIZE_SENDS
  sender_active = 0;
  scommlock = mutex_alloc();
  MIDWAY_MUTEX_SET_NAME(scommlock, "scommlock");
  scomm_signal = condition_alloc();
#endif 

#ifdef STATS
  statlock = mutex_alloc();
  MIDWAY_MUTEX_SET_NAME(statlock, "statlock");
  mid_comm_reset_stats();
#endif

  cpt = (cpt_t *) CALLOC(CommThreadMax, sizeof(cpt_t));
  ASSERT (cpt != NULL);
  cptnext = 1;			/* alloc starting at 1, 0 not allowed */
  cptlock = mutex_alloc();
  MIDWAY_MUTEX_SET_NAME(cptlock, "cptlock");
  scan_mu = mutex_alloc();	/* one for scanner of cpt table */
  MIDWAY_MUTEX_SET_NAME(scan_mu, "comm.c:scan_mu");
  scan_cv = condition_alloc();
  scanflag = 0;

  trantab = (trantab_t *) CALLOC((hnm+1)*CommThreadMax, sizeof(trantab_t));
  ASSERT(trantab != 0);
#if (BENCHMARK == 0)
  for (i=0; i < ((hnm+1)*CommThreadMax); i++) {
      trantab[i].signature = TRANTAB_SIG;
  }
#endif /*~BENCHMARK*/
  trantab_lock = mutex_alloc();
  MIDWAY_MUTEX_SET_NAME(trantab_lock, "comm.c:trantab_lock");
  trantab_c = condition_alloc();

#if TSTATS
  setup_mapped_timer();
#endif;

#if FLOW_TIMES
  INITIALIZE_FT;
  /* try wiring down the sucker (about 1.5MB) */
  {
    /* shouldn't duplicate this code so many times */
    mach_port_t host_priv;
    kern_return_t krc;

    host_priv = task_by_pid(-1); /* back door to get priv host port */
    if (host_priv < 0) {	/* Brian did this.  Not sure if proper. */
      perror("bsd_task_by_pid");
      exit(1);
    }
    if (host_priv == 0) {	/* see this empirically */
      StatusMsg("Can't get priv host port [not grp=kmem or uid=0?], will not set receive thread priority.");
    } else {
      /* valid host_priv port */
      StatusMsg("wiring packet flow times buffer.");
      krc = vm_wire(host_priv, mach_task_self(), ft_times, sizeof(ft_times),
		    VM_PROT_READ|VM_PROT_WRITE);
      if (krc != KERN_SUCCESS) {
	char smsg[130];
	sprintf(smsg,"vm_wire of packet flow buffer fails, addr 0x%08x len %u",
		ft_times, sizeof(ft_times));
	StatusMsg(smsg);
	mach_error("vm_wire", krc);
	/* just continue */
      }
    } /* valid host_priv port */
  }
#endif

  return(0);
} /* CommInit */


DoCommScanner()
{
  int i,j; 		/* into cpt */
  int morework = 0;
  int r; 		/* into AcksRequired... */
  u_long mask;
  u_long redoHosts;   /* also for AcksRequired, etc. */
  u_long midGetLockIncarnTime(); /* in lock.c */
  char255	smsg;		/* just in case */    
  /*
   * For now, run through all cpt entries linearly.
   * Look for ones with SCANNABLE bit set in retries
   *	if RETRANSMIT bit set, 
   *		if ALLPACKETS bit set, then just looking for Acks
   *			retransmit invalidate's
   *			(if no invalidates, clear scannable bit? is error.)
   *		else retransmit packet
   *		clear RETRANSMIT bit
   *	else set RETRANSMIT bit so will get on next go round
   *  set morework flag
   */
  for (i=1; i<cptnext; i++) {	/* for each entry in cpt, */
    if (cpt[i].retries & M_SCANNABLE) { /* fast check */
      mutex_lock(&(cpt[i].m));
      if (cpt[i].retries & M_SCANNABLE) { /* try again... COMPILER OPT? */
	if (cpt[i].retries & M_RETRANSMIT) {
	  /*** NOTE NOTE NOTE NOTE RETHINK, and DELETE if already DONE ***/
	  if (cpt[i].retries & M_NEWTARGET) { /* target shifted */
	    cpt[i].c__mflags &= ~M_RETRANS; /* so ensure retrans bit off */
	    cpt[i].retries &= M_NEWTARGCLR; /* ...but only once */
	  } else {		/* same old target, so... */
	    cpt[i].c__mflags |= M_RETRANS; /* note pkt is retransmission */
	  }
	  if(midAllpackets(i)) { /* Then no need to retrans request, but.. */

	    /* retransmit invalidates for LOCK requests */
	    /* can stomp on cpt, 'cause of "ALLPACKETS" */
	    if (cpt[i].c__optype == M_LOCK_OP) {
	      cpt[i].c__optype = M_INVALIDATE_OP; /* stomp! */
	      /* INCEST ALERT */
	      if ((cpt[i].msg == (char *)0) ||
		  (cpt[i].mlen < sizeof(u_long))) {
		sprintf(smsg,"CommScanner: cannot send invalidates for lock %lu, no message available.", cpt[i].c__lock);
		StatusMsg(smsg);
#if STOP_ON_ERROR
		botch("CommScanner: no invalidate msg");
#endif
		continue;
	      }
	      /* STOMP */
	      *((u_long *)(cpt[i].msg)) = midGetLockIncarnTime(cpt[i].c__lock);
	      cpt[i].mlen = sizeof(u_long);
	    } /* op wasn't invalidate */

	    for (j=1,r=0,mask=0x40000000,
		 redoHosts=(cpt[i].acksRequired[r]&
			    (~cpt[i].acksReceived[r]));
		 j<=Comm_hnum; j++) {
#ifdef DEBUG
	      if (IS_DBG_COMM_SCAN) {
		sprintf(smsg, "->scan: areq %lx arcv %lx redo %lx mask %lx",
			cpt[i].acksRequired[r], cpt[i].acksReceived[r],
			redoHosts, mask);
		StatusMsg(smsg);
	      }
#endif /*DEBUG*/
	      if (mask & redoHosts) { /* host is in list */
		cpt[i].dest = j;
		CommSend((u_long)i);

		if (cpt[i].c__optype == M_INVALIDATE_OP) {
		  INCR_STAT (mid_comm_stats.retranack,1);
		} else {	
		  INCR_STAT (mid_comm_stats.retranmulti,1);
		}
	      }
	      if (!(mask >>= 1)) { /* shift right one */
		mask = 0x80000000; /* if fallen off end, restart */
		r++;
		redoHosts=(cpt[i].acksRequired[r] &
			   (~cpt[i].acksReceived[r]));
	      }
	    }
	  } else {
	    CommSend((u_long)i); /* WHAT ABOUT THAT PESKY PACKET LIST? *mjz* */
	  }
	  cpt[i].retries &= M_RETRANSCLR; /* deed is done for this round */
	}
	else {
	  cpt[i].retries |= M_RETRANSMIT; /* do next time */
	}
	morework = 1;
      } /* scannable entry */
      mutex_unlock(&(cpt[i].m));
    } /* inital scannable fast check */
  }
  return morework;
}


/*
 ************************************************************************
 *			C o m m S c a n n e r				*
 *----------------------------------------------------------------------*
 ************************************************************************
 */


/*
 * scanflag states:
 *	0 - am sleeping waiting to be woken up
 *	1 - am scanning, and will go to sleep if I find no more work
 *	2 - will scan again after timeout
 */
/*#define	SCAN_SLEEP	0
 *#define	SCAN_WORK	1
 *#define SCAN_AGAIN	2
 * --defined above so can initialize scanflag statically
 */
signal_scanner()
{
  mutex_lock(scan_mu); /* tell scanner */
  if (scanflag == SCAN_SLEEP)  {
    condition_signal(scan_cv);
  }
  scanflag = SCAN_AGAIN; /* definitely won't sleep now */
  mutex_unlock(scan_mu);
  
}



/*
 * message type to wake up scanner
 */
int midway_comm_scanner_interval = 500;	/* mS */

int midway_comm_scanner_interval_atm = 50;	/* mS */
int midway_comm_scanner_interval_fastudp = 100; /* mS */




CommScanner()
{
  int morework = 0;

#if WIRE_DA_THREADS
  /* we are cthread_fork'd; for now wire all cthreads with 3.0 threads*/
  cthread_wire();
#endif

#if SET_PRIORITY
  if (midway_comm_scanner_prio) {
    midway_set_good_priority(midway_comm_scanner_prio);
  }
#endif
  
  mutex_lock(scan_mu);
/*  scanflag = SCAN_SLEEP; -- now initialized statically */
  
  for (;;) {
    while (scanflag == SCAN_SLEEP)  {
      condition_wait(scan_cv, scan_mu);
    }
    scanflag = SCAN_WORK;
    mutex_unlock(scan_mu);
    
#ifdef DEBUG
    if (IS_DBG_COMM_SCAN) {
      StatusMsg("->Scanner: starting scan");
    }
#endif /*DEBUG*/
    
    /* Our working is: wait for 50ms, do the work, check the
     * world.
     */
    timer_pause_ms(midway_comm_scanner_interval);
    
    morework = DoCommScanner();
    mutex_lock(scan_mu);
    if (morework == 0 && scanflag == SCAN_WORK)  {
      scanflag = SCAN_SLEEP;
    } 
  }
}   /* CommScanner */



/*
 ************************************************************************
 *			    C o m m O p e n				*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
CommOpen()
{
  char255	smsg;		/* for error messages */
  void		CommWorker();
  int		i;
  struct sockaddr_in sin;	/* to bind port value */
  u_short	uport;
  int		rc;
  HREPLYMSG	req;		/* really request here */
  struct hostent *hp;		/* for name service */
  extern int	h_errno;	/* for name service errors */
  char		*emsg;		/* for name service errors */
  long		sbsize;
  int	        sock;		/* socket, returned */

/*
 * called with ux_pass probably in effect.  Problem: now we pass opens
 * and closes...  and a close of our socket is likely to close the
 * host_op socket on the originator.  This would be bad. I also see
 * that we try to read & write from /etc/resolv.conf & /etc/hosts...
 * but these necessarily have to go across.  How did this ever work?
 * I'm gonna find out.  For now, make close DoTheRightThing.
 */
#if !defined(FAST_UDP)
#include <syscall.h>
#define close(f) syscall(SYS_close, (f))
#endif

  if (!initdone) {
    StatusMsg("error - CommOpen called before CommInit");
    return(-1);
  }

#if SET_PRIORITY
  if (midway_comm_network_type == MIDWAY_NET_ATM) {

    /* attempt to hog machine */

    if (!midway_comm_worker_prio)
      midway_comm_worker_prio = 6; /* was 3*/

    if (!midway_user_priority)
      midway_user_priority = 7;  /* was 5 */

    if(!midway_comm_scanner_prio)
      midway_comm_scanner_prio = 7; /* was 5 */

    if(!midway_gprof_priority)
      midway_gprof_priority = 7; /* was 5 */
  }
#endif

#ifdef FAST_UDP
  /*
   * Initialize network library routines
   */
  {
    extern void network_init(char *);
    extern int use_device_syscall;

    if (IS_DBG_MISC_INIT)
      fprintf(stderr, "CommInit: initializing fast udp library\n");

    /* set global flag to enable device syscall in library */
    use_device_syscall = 1;

    /* set resend timer */
    midway_comm_scanner_interval = midway_comm_scanner_interval_fastudp;

    /* initialize library */
#if 0
    network_init ("/parameter_file_doesnt_exist");
#endif
    /* need to figure out a better way to specify the file, perhaps
       search ourselves, first local, then std place?  what about
       per-machine?  maybe give as a parameter?  I dunno...
       */
    network_init (".sockets.config");
  }
#if SET_PRIORITY
  if (midway_user_priority) {
    midway_set_good_priority(midway_user_priority);
  }
#endif
#else
#if SET_PRIORITY
  /* not sure if fast_udp should change all thread priorities... */
  if (midway_user_priority) {	/* defined in rsched.c */
    /* note: sets xpager, etc. too. */
    midway_set_good_priority_all(midway_user_priority);
  }
#endif
#endif
  
  sock = NET_SOCKET(AF_INET, SOCK_DGRAM, 0); /* attempt to open UDP socket */
  if (sock == -1) {
    sprintf(smsg,"error - socket(AF_INET,SOCK_DGRAM,0): %s (%d)",
	    ((errno>sys_nerr)?"unknown":sys_errlist[errno]), errno);
    StatusMsg(smsg);
    return(-1);
  }

  /* 4 Aug 92, mjz: attempt to maximize buffering in the kernel so
   * we don't drop messages when doing big transfers.  According
   * to the source, the maximum I can set this value to is
   * SB_MAX * CLBYTES / (2 * MSIZE + CLBYTES)
   * from sys/socketvar.h: SB_MAX = 65535
   * from sys/param.h:	 CLBYTES = CLSIZE*NBPG
   * from machine/machparam.h (for mips) CLSIZE = 1, NBPG = 4096
   * from sys/mbuf.h: MSIZE = 128
   * ==> (for mips at least) 61680.  This sets (from sys/socketvar.h,
   * also see bsd/uipc_socket{,2}.c for the code) sb_hiwat to this value,
   * and sb_mbmax to MIN(#*2, SB_MAX).
   */
  sbsize = 32767;
  i = NET_SETSOCKOPT(sock, SOL_SOCKET, SO_SNDBUF, &sbsize, sizeof(sbsize));
  if (i) {
    sprintf(smsg,"setsockopt(%d, SOL_SOCKET, SO_SNDBUF, %d, %d): %s (%d)",
	    sock, sbsize, sizeof(sbsize),
	    ((errno>sys_nerr)?"unknown":sys_errlist[errno]), errno);
    StatusMsg(smsg);
    /* not fatal */
  }
  i = NET_SETSOCKOPT(sock, SOL_SOCKET, SO_RCVBUF, &sbsize, sizeof(sbsize));
  if (i) {
    sprintf(smsg,"setsockopt(%d, SOL_SOCKET, SO_RCVBUF, %d, %d): %s (%d)",
	    sock, sbsize, sizeof(sbsize),
	    ((errno>sys_nerr)?"unknown":sys_errlist[errno]), errno);
    StatusMsg(smsg);
    /* not fatal */
  }
    

  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = INADDR_ANY; /* don't bind to particular intfce */
  bzero(sin.sin_zero,sizeof(sin.sin_zero)); /* I've seen picky kernels */
  for (uport = UDPPORTSTART; uport <= UDPPORTEND; uport++) { /* find port */
    sin.sin_port = uport;
    rc = NET_BIND(sock,(struct sockaddr *)&sin,sizeof(sin));
    if (rc == 0) break;		/* we're done */
    if (errno != EADDRINUSE) {	/* failure */
      sprintf(smsg,"failure binding UDP socket to port %u: %s (%d)",
	      uport,((errno>sys_nerr)?"unknown":sys_errlist[errno]), errno);
      if (ss_port)
	ss_msg(ss_port,LocalClock,hname,Comm_index,smsg);
      else
	fprintf(stderr,smsg);
      NET_CLOSE(sock);
      return(-1);
    }
  }/* for uport */
  if (rc != 0) {		/* looped to UDPPORTEND w/o finding */
    sprintf(smsg,"failure - all UDP ports from %d to %d in use!",
	    UDPPORTSTART, UDPPORTEND);
    if (ss_port)
      ss_msg(ss_port,LocalClock,hname,Comm_index,smsg);
    else
      fprintf(stderr,smsg);
    NET_CLOSE(sock);
    return(-1);
  }
  midway_udpsock = sock;		/* note for other Comm routines */
				/* THINK ABOUT THIS AGAIN *mjz* */

  if (midway_comm_network_type == MIDWAY_NET_ATM) {
    /*
     * atm.  Start it up.  No HOST_OP, wired for now :-( 
     */
    if (midway_atm_open(hlist)) {
      /* error -- fallback */
      StatusMsg("CommOpen: error opening ATM, falling back to sockets");
      midway_comm_network_type = MIDWAY_NET_SOCKETS;
    }
  }


  /* get scanner set up */
  cthread_detach(cthread_fork((cthread_fn_t)CommScanner,(void *)0)); /* crank it up */

  /* crank one worker for open only.  This will allow us to switch
     network protocols after host_op. */
  cthread_detach(cthread_fork((cthread_fn_t)CommWorker,(void *)CommWorkerNum));

  if (IS_DBG_COMM_TRACE) {
    midPmsgTraceInit();
  }

  /* make initial request . must first get IP addr */
  /* gethostbyname reads files.  Avoid overhead & inconsistency... read
     the local one -10oct93 mjz (clash with cmaedas new fast_udp; they
     just stuck a local resolv.conf on this weekend.  shit. */
  (void) midway_uxpass_local();
/*   ---#if USE_ULTRIX_LIBRARY ---*/
if (midway_comm_network_standalone) {
    char255 lname;		/* try making lower case */
    char *src, *dst;
    for (dst = lname, src=hname; *src; src++,dst++) {
      if (isupper(*src))
	*dst = tolower(*src);
      else
	*dst = *src;
    }
    *dst = '\0';	/* add trailing null */
    printf("going to gethostbyname(%s)\n", lname);
#if USE_ULTRIX_LIBRARY
    hp = gethostbyname(lname);
#else
    { extern struct hostent *_gethtbyname();
      hp = _gethtbyname(lname); /* cmu-cs specific */
    }
#endif
  } else {
    hp = gethostbyname(hname);
  }
  (void) midway_uxpass_remote();
  if (!hp) {			/* error will robinson */
    switch(h_errno) {
    case HOST_NOT_FOUND:   emsg = "HOST_NOT_FOUND"; break;
    case TRY_AGAIN:	   emsg = "TRY_AGAIN"; break;
    case NO_RECOVERY:	   emsg = "NO_RECOVERY"; break;
    case NO_ADDRESS:	   emsg = "NO_ADDRESS"; break;
    default: emsg = "unknown!"; break;
    } /* switch */
    sprintf(smsg,"Error from gethostbyname(%s): %s (%d)",
	    hname, emsg, h_errno);
    StatusMsg(smsg);
    botch("gethostbyname");
    NET_CLOSE(sock);
    midway_udpsock = -1;
    return(-1);
  }
  if (hp->h_addrtype != AF_INET) { /* sanity failure; commit suicide */
    sprintf(smsg,"gethostbyname(%s) returned unknown address family %d\n",
	    hname,hp->h_addrtype);
    StatusMsg(smsg);
    NET_CLOSE(sock);
    midway_udpsock = -1;
    return(-1);
  }
  bcopy((char *)hp->h_addr, (char *)&req.ipaddr, sizeof(req.ipaddr)); 

  if (IS_DBG_COMM_TRACE) {
    sprintf(smsg, "CommOpen sending HOST_OP(index=%d port=%d)\n",
	    Comm_index, uport);
    StatusMsg(smsg);
  }


  /*
   * Send a request to "smexecd"
   * Destination 0 contains its address.
   */
  req.index = Comm_index;
  req.udpport = uport;
  CommRequest((u_long)0,(u_long)M_HOST_OP,0,(char*) &req, (u_long)sizeof(req));
  /* once we return, the htbl will be filled in */

  if (midway_comm_network_type == MIDWAY_NET_ATM) {
    /* switch fns */
    /* could twiddle with retrans interval */
    /* ***note, depend on retrans if we throw away queuers!!! */
    midway_net_fns = &midway_atm_fns;
    
    midway_comm_scanner_interval = midway_comm_scanner_interval_atm;
    midway_comm_reply_ackafter = midway_comm_reply_ackafter_atm;
    midway_comm_send_ackafter = midway_comm_send_ackafter_atm;
  }

  sprintf(smsg,"Using a retransmission interval of %d ms.", 
	  midway_comm_scanner_interval);
  StatusMsg(smsg);


  /* spawn workers */ /* n */
  for (i=0; i<CommWorkerNum; i++) {
    /* NOTE: with just one, we will deadlock on a write req following a
      read req; 'cause one worker is hung waiting for the write req
	to finish, while there is a second "request" that must come
	  in to complete the write request -- the invalidate reply!! */
    /* any other cases??? */
    cthread_detach(cthread_fork((cthread_fn_t)CommWorker,(void *)i));
  }

  if (IS_DBG_COMM_TRACE) {
    sprintf(smsg, "CommOpen: index=%d socket=%d port=%d IP=%x name=\"%s\"",
	    Comm_index, midway_udpsock, uport, htbl[Comm_index].ip, hname);
    StatusMsg(smsg);
  }

  return(0);

#if !defined(FAST_UDP)
#undef close
#endif
} /* CommOpen */

/*
 ************************************************************************
 *			    c p t _ i n i t				*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
/*static*/ cthread_t cpt_last_alloc = 0; /* cache last value */
u_long
cpt_init()
{
  u_long cpti;			/* new c-per-thread index */
  cpt_t *cptP;
  char *mname;
  cthread_t me;
  char255 smsg;			/* for errors */

  me = cthread_self();
  mutex_lock(cptlock);
  if (me == cpt_last_alloc) {
    cpti = cptnext - 1;		/* last allocated, so must be last one */
    mutex_unlock(cptlock);	/* quick exit */
    return(cpti);
  }
  if (cptnext >= CommThreadMax) {
    mutex_unlock(cptlock);
    sprintf(smsg,"error - too many threads.  Re-compile with extern Comm\
ThreadMax (now %lu) set higher.", CommThreadMax);
    StatusMsg(smsg);
    if (IS_DBG_COMM_CPT) {
      task_suspend(mach_task_self());	/* XXXXXYYYYZZZZZ 9jun92mjz */
    }
    return(-1);
  }
  cpti = cptnext++;
  cpt_last_alloc = me;
  if (IS_DBG_COMM_CPT) {
    sprintf(smsg,"XXXX cpt_init called for thread %x, new index=%lu",
	    me, cpti);
    StatusMsg(smsg);
  }
  mutex_unlock(cptlock);

  cptP = &cpt[cpti];
  cthread_set_data(me,(any_t) cpti);

  mutex_init(&(cptP->m));
#if MIDWAY_MUTEX_NAMES
  mname = (char *) MALLOC(12);
  ASSERT(mname != 0);
  sprintf(mname, "CPT%.6d",cpti);
  mutex_set_name(&(cptP->m),mname);
#endif
  condition_init(&(cptP->c));
  ZERO_ADDR_VECTOR(cptP->acks_received);
  ZERO_ADDR_VECTOR(cptP->acks_required);
  cptP->maxlinear = 0;
  cptP->retries = 0;
  cptP->lastpacket = 0;


  /*
   * Initialize message header
   */
  cptP->msgheader.version = M_PROTOVER;
  cptP->msgheader.type = 0;
  cptP->msgheader.flags = 0;
  cptP->msgheader.unused = 0;
  cptP->msgheader.incarnation = Comm_incarn;
  cptP->msgheader.source = (c_hostid_t) Comm_index;	/* index of this host */
  cptP->msgheader.id = (c_threadid_t) cpti;
  cptP->msgheader.rhost = 0;
  cptP->msgheader.rid = 0;
  cptP->msgheader.lock = 0;
  cptP->msgheader.clock = 0;
  cptP->msgheader.tid = 0;
  cptP->msgheader.pacnum = 0;
  ZERO_ADDR_VECTOR(cptP->msgheader.nacks);

  cptP->iov[0].iov_base = (caddr_t)&cptP->msgheader;
  cptP->iov[0].iov_len = sizeof(cptP->msgheader);
  cptP->iov[1].iov_base = 0;
  cptP->iov[1].iov_len = 0;

  bzero(&cptP->sin,sizeof(cptP->sin));
  cptP->sin.sin_family = AF_INET;

  cptP->msgh.msg_name = (caddr_t) &cptP->sin;	/* where it goes */
  cptP->msgh.msg_namelen = sizeof(cptP->sin);
  cptP->msgh.msg_iov = cptP->iov;
  cptP->msgh.msg_iovlen = 0;
  cptP->msgh.msg_accrights = (caddr_t) 0; /* no rights */
  cptP->msgh.msg_accrightslen = 0;

  /* copy yourself faster than scat/gath? */
  /* message could be too big...  next iteration *mjz* */
  /* no - not any longer; messages packaged to max size by requestor */


  return(cpti);
} /* cpt_init */

/*
 ************************************************************************
 *			C o m m R e q u e s t				*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
#if ATM_IFACE  /* old ATM interface */
/* return 0: not us, 1: us, and we expect more; -1: us, and we're done */
/* this name is horrible.  This should be CommReplyFilter, and CR's
   should be CommAckFilter, since that is what they are looking for */
CommRequestFilter(char *msg, u_long len, any_t opaque)
{
  COMMMSG *cm;
  u_long cpti;
  cpt_t *cptP;
  char255 smsg;
  int done;
  int *ip;
  int numnacks;
  int buflen;
  u_long pnum;
  int pindex;
  u_long updt_maxlinear = 0;
  clock_t updt_clock = NULL_CLOCK_VALUE;
  comm_request_t rc;
  int rv;
  register u_char msg_type;
#if FLOW_TIMES
  u_int fti;
  u_int ftclk;
  u_int *ftp;
#endif

  INCR_STAT (mid_comm_stats.reqffilt, 1);

  cm = (COMMMSG *) msg;
  cpti = (u_long) opaque;
  cptP = &cpt[cpti];
  msg_type = cm->type;

  if ((cm->version == M_PROTOVER) &&
      (cm->incarnation == Comm_incarn) &&
      (cm->rid == cpti) &&	/* destination id */
      (cm->tid == cptP->c__tid) &&
      /* pretty sure is us, check op too? */
      (msg_type & M_REPLY_BIT)
      )
    {
      /* Filters need to always update the Lamport clock! */
      if (mutex_try_lock(ClockLock) == 0) {
	mid_comm_stats.reqflmiss++; /* not quite right */
	return(0);
      }
      if (GREATER_THAN_CLOCK (cm->clock, LocalClock))
	LocalClock = ADJUST_CLOCK (cm->clock, 1);
      else
	LocalClock = ADJUST_CLOCK (LocalClock, 1);
      mutex_unlock(ClockLock);

      if (mutex_try_lock(&(cptP->m)) == 0) {
	/* lock miss */
	INCR_STAT (mid_comm_stats.reqflmiss, 1);
	return(0);		/* punt packet */
      }
      /* double check packet is for us, now that we have the lock */
      if (cm->tid != cptP->c__tid) {
	INCR_STAT (mid_comm_stats.reqftmiss, 1);
	mutex_unlock(&(cptP->m));
	return(0);
      }
#if FLOW_TIMES
      ftclk = FT_READ_TIMER;
#endif
      cptP->retries &= M_RETRANSCLR; /* got a packet */
      ip = (int *) (cm + 1);	/* point after header */
      buflen = len - sizeof(COMMMSG); /* length left */

      /* obtain any AcksRequired entries */
      cptP->acks_required = cm->nacks; /* NOTE: This sets value for cptP->acksRequired */

      /*
       * Handle multicasts.
       * Expect a reply from all recipients of the request 
       */
      if ((cptP->multicast != 0) ||
	  (msg_type == (M_LOCK_INV_OP|M_REPLY_BIT)) ||
	  (msg_type == (M_INVALIDATE_OP|M_REPLY_BIT))) { /* this is an invalidate ack */

	if (IS_DBG_COMM_WORK) {
	  sprintf(smsg,"MulticastReply: adding %lu to acksReceived", cm->source);
	  StatusMsg(smsg);
	  sprintf(smsg,"allpackets %d areq=arcv %d areq %lx arcv %lx",
		  midAllpackets(cpti),
		  midCompareBit(cptP->acksRequired,
				cptP->acksReceived,
				Comm_hnum),
		  cptP->acksRequired[0], cptP->acksReceived[0]);
	  StatusMsg(smsg);
	}

	midSetBit(cptP->acksReceived, cm->source);
	if (midAllpackets(cpti) &&
	    midCompareBit(cptP->acksRequired,cptP->acksReceived,
			  Comm_hnum)) {

	  if (IS_DBG_COMM_WORK) {
	    sprintf(smsg,
		    "...all pkts+acks rcvd. writers=%d", cptP->writers);
	    StatusMsg(smsg);
	  }

	  cptP->retries &= M_SCANCLEAR; /* no more retrans necessary */

	  if ((msg_type == (M_INVALIDATE_OP|M_REPLY_BIT)) && /* this is an invalidate ack */
	      (cptP->writers == 0)) {
	    mutex_unlock(&(cptP->m));
#if MSTATS
	    INCR_STAT (mid_comm_stats.replin, 1);
#endif /*MSTATS*/
	    INCR_STAT (mid_comm_stats.reqffast, 1); /* fast hit too */
	    return(-1);		/* all done! */
	  }
	}
      } /* "multicast" */
      if (((msg_type & M_OP_BITS) == M_LOCK_OP) ||
	  ((msg_type & M_OP_BITS) == M_LOCK_INV_OP) ||
	  ((msg_type & M_OP_BITS) == M_BARRIER_OP) ||
	  ((msg_type & M_OP_BITS) == M_HOST_OP) ||
	  ((msg_type & M_OP_BITS) == M_ALLOCATE) ||
	  ((msg_type & M_OP_BITS) == M_ALLOC_INFO) ||
	  ((msg_type & M_OP_BITS) == M_SCHED_OP) ||
	  ((msg_type & M_OP_BITS) == M_PAGE_DATA)) {

	/* LOCK, BARRIER and other non-multicasted replies that release
	   the sender. */

	/*process packet number info. */

	if (cm->pacnum > cptP->lastpacket)
	  cptP->lastpacket = cm->pacnum;

	pnum = cm->pacnum << 1;
	if (pnum == cptP->maxlinear) { /* next in sequence */
	  updt_maxlinear = 1;	/* note we did this, so can undo :-( */
	  cptP->maxlinear += 2; /* bump */
	} else {
	  pindex = cm->pacnum - (cptP->maxlinear >> 1);
	  /* check for duplicates.  It's a duplicate if (1) pindex is
	   * less than 0, since that means it was one that arrived in
	   * sequence, or (2) it is in the .plist bitvector.
	   */
	  if ((pindex < 0) || midPlistTest(&(cptP->plist),pindex)) {
	    if (IS_DBG_COMM_WNGS) {
	      sprintf(smsg,
		      "comm_worker: Packet %lu tid %lu dropped as duplicate.",
		      cm->pacnum, cm->tid);
	      StatusMsg(smsg);
	    }
	    /* This one might actually be asking for an ACK.  Test theory */
	    if (((cm->pacnum+1) % midway_comm_reply_ackafter) == 0) {
	      /* good shot */
	      if (cptP->ackpnum == 0) {
		/* Well, there's no ack pending.  Must be a retrans. Ack! */
		/* may block on sendlock? */
		CommAck(cm->source, cm->id, cm->tid, cm->lock, cm->pacnum, 0);
		INCR_STAT (mid_comm_stats.ackReqFilter1, 1);
	      }
	    }
	    mutex_unlock(&(cptP->m));
	    INCR_STAT (mid_comm_stats.again, 1); /* XXX */
	    return(1);		/* drop this packet as duplicate */
	  } else {
	    midPlistSet(&(cptP->plist), pindex);
	  }
#if STATS			/* keep track when we get out of order */
	  if ((cptP->maxlinear & 1) == 0) {
	    cptP->maxlinear |= 1;
	    INCR_STAT (mid_comm_stats.mpseqfail,1);
	  }
#else
	  cptP->maxlinear |= 1; /* ensure will never match again */
#endif
	}
	if (cm->flags & M_LASTPACK){ /*this IS the largest pacnum we will see*/
	  cptP->retries |= M_ALLPACKETS;
	  if (midAllpackets(cpti) &&
	      midCompareBit(cptP->acksRequired,cptP->acksReceived,
			    Comm_hnum)){
	    cptP->retries &= M_SCANCLEAR;	/* no need to scan */
	    updt_clock = cptP->c__tid;
	    /* ensure retrans of replies die */
	    cptP->c__tid = ADJUST_CLOCK(cptP->c__tid,1);
	    /* (perhaps due to retrans of request, get retransed reply
	       but got first reply too... */
	  }
	}
	if (((cm->pacnum+1)%midway_comm_reply_ackafter) == 0) {
	  /* We should ack after this packet arrives. Note that... */
	  if (cptP->writers <= (midway_comm_reply_ackafter / 2)) {
	    CommAck(cm->source, cm->id, cm->tid, cm->lock, cm->pacnum, 0);
	    INCR_STAT (mid_comm_stats.ackReqFilter2, 1);
	  } else {
	    cptP->ackpnum = cm->pacnum;
	    cptP->ackthread = cm->id;
	  }
	}

	cptP->writers++;	/* thread will write to memory. */
      } /* type not invalidate */ /* lock & barrier only */

      if (cptP->multicast) {
	/* non-Invalidate multicast requests... these expect a reply
	 * that is handled at a higher level.  Increment writer count
	 * so that the release check after the switch will release
	 * the waiter.  (It should never be true for multicast
	 * that multiple packets can become outstanding... so strictly
	 * speaking, we don't need the writer check if we are multicast.)
	 * (I.e. it could be fixed below :-) --mjz27aug92)
	 */
	cptP->writers++;
      }

#if MSTATS     
      if (((msg_type & M_OP_BITS) != M_HOST_OP) &&
	  ((msg_type & M_OP_BITS) != M_SCHED_OP) &&
	  (cm->flags & M_LASTPACK)) {
	/* know acks, nacks already filtered out */
	/* duplicates should be filtered too, no?  Only for multipacket,
	   I guess. */
	switch(msg_type) {
	case M_LOCK_OP|M_REPLY_BIT:
	case M_LOCK_INV_OP|M_REPLY_BIT:
	case M_INVALIDATE_OP|M_REPLY_BIT: /* ? separate for inv? */
	  INCR_STAT (mid_comm_stats.replin, 1);
	  break;
	case M_BARRIER_OP|M_REPLY_BIT:
	  INCR_STAT (mid_comm_stats.repbin, 1);
	  break;
	case M_PAGE_DATA | M_REPLY_BIT:
	  INCR_STAT (mid_comm_stats.pfrepin, 1);
	  break;
	case M_PAGE_UPDATE | M_REPLY_BIT:
	  INCR_STAT (mid_comm_stats.purepin, 1);
	  break;
	case M_PAGE_COPYSET | M_REPLY_BIT:
	  INCR_STAT (mid_comm_stats.csrepin, 1);
	  break;
	default:
	  INCR_STAT (mid_comm_stats.repin, 1);
	  break;
	}
      }
#endif /* MSTATS */

/*      mutex_unlock(&(cptP->m)); don't unlock... we can't fail to lock again*/
      /* perform upcall to lock layer.  Note: we could die because of lock
	 entry mutex held... */

      switch(msg_type) {
      case M_LOCK_OP | M_REPLY_BIT:		/* lock reply */
      case M_LOCK_INV_OP | M_REPLY_BIT: /* no one was reading */
	if (IS_DBG_COMM_WORK) {
	  StatusMsg("about to midLockReply");
	  sprintf(smsg,"writers=%lu allpackets=%d Required==Received=%d",
		  cptP->writers, midAllpackets(cpti),
		  midCompareBit(cptP->acksRequired,
				cptP->acksReceived, Comm_hnum));
	  StatusMsg(smsg);
	}
	
	rc =
	midLockReply(cm->lock,cm->source,cm->id,cm->tid,
		     (char *)ip,(u_long)buflen, 1 /*don't block*/);

	if (IS_DBG_COMM_WORK)
	  StatusMsg("done midLockReply");

	if (rc == M_R_COULDBLOCK) {
	  /* undo */
	  ASSERT(cptP->writers > 0);
	  cptP->writers--;
	  if (updt_maxlinear) {
	    cptP->maxlinear -= 2;
	  } else {
	    midPlistClearBit(&(cptP->plist), pindex);
	  }
	  if (updt_clock != NULL_CLOCK_VALUE)
	    cptP->c__tid = updt_clock;
	  mutex_unlock(&(cptP->m));
	  INCR_STAT (mid_comm_stats.reqfbmiss, 1);
	  return(0);
	}
	break;

      case M_INVALIDATE_OP | M_REPLY_BIT:
	break;			/* nothing to do here */

      default:
	/* undo */
	if (((msg_type & M_OP_BITS) == M_LOCK_OP) ||
	    ((msg_type & M_OP_BITS) == M_LOCK_INV_OP) ||
	    ((msg_type & M_OP_BITS) == M_BARRIER_OP) ||
	    ((msg_type & M_OP_BITS) == M_HOST_OP) ||
	    ((msg_type & M_OP_BITS) == M_ALLOCATE) ||
	    ((msg_type & M_OP_BITS) == M_ALLOC_INFO) ||
	    ((msg_type & M_OP_BITS) == M_SCHED_OP) ||
	    ((msg_type & M_OP_BITS) == M_PAGE_DATA) ||
	    (cptP->multicast != 0)) { 
	  ASSERT(cptP->writers > 0);
	  cptP->writers--;	/* don't need since hold lock whole time now */
	}
	if (updt_maxlinear) {
	  cptP->maxlinear -= 2;
	} else {
	  midPlistClear(&(cptP->plist), pindex);
	}
	if (updt_clock != NULL_CLOCK_VALUE)
	  cptP->c__tid = updt_clock;
	mutex_unlock(&(cptP->m));
	INCR_STAT (mid_comm_stats.reqfomiss, 1);
	return(0);
      } /* switch on type */

      INCR_STAT (mid_comm_stats.reqffast, 1);

      if (((msg_type & M_OP_BITS) == M_LOCK_OP) ||
	  ((msg_type & M_OP_BITS) == M_LOCK_INV_OP) ||
	  ((msg_type & M_OP_BITS) == M_BARRIER_OP) ||
	  ((msg_type & M_OP_BITS) == M_HOST_OP) ||
	  ((msg_type & M_OP_BITS) == M_ALLOCATE) ||
	  ((msg_type & M_OP_BITS) == M_ALLOC_INFO) ||
	  ((msg_type & M_OP_BITS) == M_SCHED_OP) ||
	  ((msg_type & M_OP_BITS) == M_PAGE_DATA) ||
	  (cptP->multicast != 0)) { 
	  ASSERT(cptP->writers > 0);
	  cptP->writers--;	/* don't need since hold lock whole time now */
      }
      /* see if ack required.  Should never happen here... since we are
	 going to loop back for more, if we are here we should ack now,
	 since otherwise the condition will never be satisfied... 
	 and we know one packet made it through. */
      if (cptP->ackpnum) {
	/* real test also checks to ensure not many writers... */
	CommAck(cm->source, cptP->ackthread, cm->tid, cm->lock,
		cptP->ackpnum, 0);
	INCR_STAT (mid_comm_stats.ackReqFilter3, 1);
	cptP->ackpnum = 0;
	cptP->ackthread = 0;
      }
#if FLOW_TIMES
      fti = cptP->pflow;
      if (cm->pacnum <= FT_MAXPAKS) {
	ftp = &(ft_times[fti][FT_REPPAKS+(2*cm->pacnum)]);
	*ftp++ = ftclk;
	*ftp = FT_READ_TIMER;
      }
      if ((cm->flags & M_LASTPACK) || (cm->pacnum == FT_MAXPAKS))
	ft_times[fti][FT_REPPNUM] = cm->pacnum;
#endif
      if ((cptP->writers == 0) &&
	  (midAllpackets(cpti)) && 
	  (midCompareBit(cptP->acksRequired,cptP->acksReceived,
			 Comm_hnum))) {
	rv = -1;		/* all done */
#if FLOW_TIMES
	ft_times[fti][FT_REPSIGNAL] = FT_READ_TIMER;
#endif
      }
      else 
	rv = 1;			/* we want more */
      mutex_unlock(&(cptP->m));
      return(rv);
    }
  else {
    /*
    sprintf(smsg,"crf: packet not for us received, type 0x%x shst %d sthrd %d",
	    msg_type, cm->source, cm->id);
    StatusMsg(smsg);
    */
    INCR_STAT (mid_comm_stats.reqfmmiss, 1); /* "match" miss */
#ifdef STATS
    mid_comm_stats.reqfmtype = msg_type;
#endif
    return(0);
  }
} /* CommRequestFilter */
#else
/* NEW ATM INTERFACE.  Allow blocking. Fastpath?  Handle all comers. */
/* return 0: not us, 1: us, and we expect more; -1: us, and we're done */
/* this name is horrible.  This should be CommReplyFilter, and CR's
   should be CommAckFilter, since that is what they are looking for */
CommRequestFilter(char *msg, u_long len, any_t opaque, u_int rcvtime)
{
  COMMMSG *cm;
  u_long cpti;
  cpt_t *cptP;
  char255 smsg;
  int save_reqd;
  int save_recvd = 0;
  int not_done = 1;
  int xor_res;			/* DEBUG */

  INCR_STAT (mid_comm_stats.reqffilt, 1);

  /* need to test for version #, epoch? --18nov93 mjz */
  INCR_STAT(mid_comm_stats.rcv, 1);
  INCR_STAT(mid_comm_stats.rcvBytes, len); 


  cm = (COMMMSG *) msg;
/*  cpti = (u_long) opaque;*/
  cpti = cm->rid;		/* destination id */
  cptP = &cpt[cpti];


  /*
   * if we want to verify anything in the incoming packet, we should
   * do it here...
   */
#if FLOW_TIMES
  cm->FT_STASH_INMSG = rcvtime; /* stomps on version */
#endif

  /* Filters need to always update the Lamport clock! */
  mutex_lock(ClockLock);
  if (GREATER_THAN_CLOCK (cm->clock, LocalClock))
    LocalClock = ADJUST_CLOCK (cm->clock, 1);
  else
    LocalClock = ADJUST_CLOCK (LocalClock, 1);
  mutex_unlock(ClockLock);

#ifdef DEBUG
  if (IS_DBG_COMM_TRACE) {
    midPmsgTrace(0, cm, msg+sizeof(COMMMSG), len - sizeof(COMMMSG));
  }
#endif /*DEBUG*/
#ifdef DEBUG_PRT
  if (incall && IS_DBG_COMM_MSGS) {
    struct sockaddr_in sin;
    (*incall)(&sin, len, cm, 0, 0, msg+sizeof(COMMMSG), len - sizeof(COMMMSG));
  }
#endif /*DEBUG_PRT*/

  CommProcessPacket(msg, len);
  /* if we just completed a function (whether or not for us), we
     should die so we don't hang.  Possible exception: if all that
     is left is an InvalidateAck needed from us (which will come
     via the loopback mechanism, not through the reply mechanism) */

  mutex_lock(&(cptP->m));	/* ??? */
  /* getting rid of testing for us is really annoying */
  save_reqd = midTestBit(cptP->acksRequired,Comm_index);
  if (save_reqd) {
    save_recvd = midTestBit(cptP->acksReceived,Comm_index);
#if 0
    this should no longer be necessary because ProcessPacket invals self;
    if (!(save_recvd)) {
      if (midTestBit(cptP->acksRequired,0) == 0) {/* XXX */
	BITARRAY *i_reqd = midGetLockInvReqd(cm->lock);
	if (i_reqd) {
	  midOrBitArray(cptP->acksRequired,i_reqd,Comm_hnum);
	  midSetBit(cptP->acksReceived,0); /* so know done */
	}
      }
    }
#endif
    midClearBit(cptP->acksRequired,Comm_index);
    midClearBit(cptP->acksReceived,Comm_index);
  }
  if (midAllpackets(cpti) &&
      midCompareBit(cptP->acksRequired,cptP->acksReceived,
		     Comm_hnum)) {
    not_done = 0;
  }
  if (save_reqd) {
    midSetBit(cptP->acksRequired,Comm_index);
  }
  if (save_recvd) {
    midSetBit(cptP->acksReceived,Comm_index);
  }
#if 0
  /*DEBUG*/
  xor_res = cptP->acksRequired[0] ^ cptP->acksReceived[0];
  if ((xor_res != 0) && (not_done == 0)) {
    sprintf(smsg,"done, but xor_res=0x%08x, req 0x%08x rcv 0x%08x",
	    xor_res, cptP->acksRequired[0], cptP->acksReceived[0]);
  }
  /*ENDDEBUG*/
#endif
  mutex_unlock(&(cptP->m));
  
#if 0
  /*DEBUG, AGAIN*/
  if (xor_res && !(not_done)) {
    StatusMsg(smsg);
  }
#endif
    
  return(not_done);
} /* CommRequestFilter */
#endif /* ATM_IFACE */

CommRequest(lock, type,dest,msg,mlen)
     u_long	lock;		/* le object 'd art */
     u_long	type;		/* kind of message */
     int	dest;		/* index into host table */
     char	*msg;		/* type dependent message info */
     u_long	mlen;		/* size of message */
{
    /*  char255	smsg;	*/	/* for error messages */
    u_long	cpti; /* client-per-thread index */
    cpt_t	*cptP;
    u_long	clk;
#if FLOW_TIMES
    u_int	fti;
    extern u_int ft_send_time_start, ft_send_time_alloc, ft_send_time_end;
#endif

    ASSERT (initdone);

    /* see if client-per-thread entry exists yet */
    cpti = (u_long)cthread_data(cthread_self());
    if (!cpti) { /* not yet allocated */
	cpti = cpt_init();
    }
    cptP = &cpt[cpti];

    clk = LocalClock;
#if FLOW_TIMES
    NEXT_FT;
    ft_times[fti][FT_TYPE] = type | M_REPLY_BIT; /* reply bit marks "client" */
    ft_times[fti][FT_LOCK] = lock;
    ft_times[fti][FT_TID] = clk;
    ft_times[fti][FT_SENDSTART] = FT_READ_TIMER;
#endif
    mutex_lock(&(cptP->m));	/* too restrictive? *//* vrfy order */

    /* cpt exists, so fill it in, and then send message & sleep */
    cptP->dest = dest;
    cptP->msg = msg;
    cptP->mlen = mlen;
    cptP->c__rhost = Comm_index;	/* do these belong here? *mjz* */
    cptP->c__rid = cpti;
    cptP->c__tid = clk;
    cptP->c__lock = lock;
    cptP->c__optype = type;
    cptP->c__mflags = M_LASTPACK;	/* this is only packet for request **MJZ** */
    cptP->c__pnum = 0;			/* for now, first and only */
    ZERO_ADDR_VECTOR(cptP->c__nacks);
    cptP->maxlinear = 0;
    cptP->lastpacket = 0;
    cptP->multicast = 0;
    midPlistClear(&(cptP->plist)); /* no replies arrived yet */
    ZERO_ADDR_VECTOR(cptP->acks_required);
    ZERO_ADDR_VECTOR(cptP->acks_received);

#if FLOW_TIMES
    cptP->pflow = fti;	/* so can get replies */
    ft_times[fti][FT_SENDREQ] = FT_READ_TIMER;
#endif

#ifdef DEBUG
    if (IS_DBG_COMM_REQ) {
	fprintf(stderr, "about to commsend\n");
    }
#endif /*DEBUG*/

    CommSend(cpti);

#ifdef DEBUG
    if (IS_DBG_COMM_REQ) {
	fprintf(stderr, "back from commsend\n");
    }
#endif /*DEBUG*/

#if FLOW_TIMES
    ft_times[fti][FT_SENDDONE] = FT_READ_TIMER;
    ft_times[fti][FT_SENDALLOC] = ft_send_time_alloc; /* XXX global */
    ft_times[fti][FT_SENDCOPY] = ft_send_time_start;
    ft_times[fti][FT_SENDSEND] = ft_send_time_end;
#endif

    cptP->retries = M_SCANNABLE; /* can retransmit... condition? DO LAST */

    mutex_unlock(&(cptP->m));

    signal_scanner();


#ifdef DEBUG
    if (IS_DBG_COMM_REQ) {
	fprintf(stderr,"CommRequest: about to wait for condition\n");
    }
#endif /*DEBUG*/
    /* now must wait for reply */
    if  ((((type & M_OP_BITS) == M_LOCK_OP) ||
	  ((type & M_OP_BITS) == M_LOCK_INV_OP) ||
	  ((type & M_OP_BITS) == M_INVALIDATE_OP)) &&
	 (midway_net_fns->wantreply)) {
      /* try fast path */
      /* try to get reply without context switching */
      midway_net_fns->wantreply((comm_filter)CommRequestFilter, (int) cpti);
    }
    mutex_lock(&(cptP->m));
    while(!(midAllpackets(cpti) &&
	    midCompareBit(cptP->acksRequired,cptP->acksReceived,
			  Comm_hnum) &&
	    cptP->writers == 0)) {
#if ATM_IFACE
      midway_atm_willblock();
#endif
      condition_wait(&(cptP->c),&(cptP->m)); /* do in while? */
      /* NOTE: if we are signalled, the condition is satisfied because
	 we are now the only ones who can 'unsatisfy' it.  However, we
	 might have gotten real unlucky and been preempted while the
	 condition was satisfied; without the check, we will "miss" the
	 signal (it was actually never sent since no one was queued
	 at the time), and we will hang forever.
	   */
#if ATM_IFACE
      midway_atm_doneblock();
#endif
    }
    cptP->retries &= M_SCANCLEAR; /* ensure */
    cptP->msg = 0;
    mutex_unlock(&(cptP->m));
#ifdef DEBUG
    if (IS_DBG_COMM_REQ) {
	fprintf(stderr,"CommRequest: released from condition\n");
    }
#endif /*DEBUG*/
#if FLOW_TIMES
    ft_times[fti][FT_REQDONE] = FT_READ_TIMER;
#endif

    /* do stuff */ /* nothing to do! - already done by LockReply calls */

    return(0);
}   /*CommRequest */

/*
 ************************************************************************
 *			C o m m M r e q u e s t				*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
/*
 * send identical request to multiple destinations; if reply-to host is
 * us, then wait for replies.
 * 
 * NOTE: Caller is not allowed to send multi-packet messages.
 */
CommMrequest(lock, type, replh, replt, clk, tolist, msg,mlen)
     u_long	lock;
     u_long	type;		/* kind of message */
     u_long	replh;		/* reply-to host, 0 for us */
     u_long	replt;		/* reply-to tid, 0 for us */
     clock_t	clk;		/* clock for message, 0 for current */
     BITARRAY	*tolist;	/* places to send */
     char	*msg;		/* type dependent message info */
     u_long	mlen;		/* size of message */
{
/*  char255	smsg;	*/	/* for error messages */
    u_long	cpti;		/* client-per-thread index */
    cpt_t	*cptP;
    int		i,j;
    u_long	mask;		/* for checking hosts...  */
#if FLOW_TIMES
    u_int	fti;
#endif


  if (IS_DBG_COMM_REQ) {
    fprintf(stderr, 
      "CommMrequest: Lock=%lu Type=0x%x rep=%lu-%lu clk=%lu TO=0x%x len=%lu\n",
	    lock, type, replh, replt, clk, *tolist, mlen);
  }
  ASSERT (initdone);

  /* see if client-per-thread entry exists yet */
  cpti = (u_long)cthread_data(cthread_self());
  if (!cpti) {			/* not yet allocated */
    cpti = cpt_init();
  }
  cptP = &cpt[cpti];

#if FLOW_TIMES
    NEXT_FT;
    ft_times[fti][FT_TYPE] = type | M_REPLY_BIT; /* reply bit marks "client" */
    ft_times[fti][FT_LOCK] = lock;
    ft_times[fti][FT_TID] = (clk != NULL_CLOCK_VALUE)? clk : LocalClock;
    ft_times[fti][FT_SENDREQ] = FT_READ_TIMER;
#endif

  /* count number of multicast requests */
  INCR_STAT (mid_comm_stats.multireqs, 1);

  mutex_lock(&(cptP->m));	/* too restrictive? *//* vrfy order */

  /* cpt exists, so fill it in, and then send message & sleep */
  cptP->c__lock = lock;
  cptP->c__optype = type;
  cptP->c__mflags = M_LASTPACK; /* this is only packet for request **MJZ** */
  cptP->c__rhost = (replh)? replh : Comm_index;
  cptP->c__rid = (replh)? replt : cpti;
  cptP->c__tid = (clk != NULL_CLOCK_VALUE)? clk : LocalClock; /* VS. TID? */
  cptP->c__pnum = 0;
  ZERO_ADDR_VECTOR(cptP->c__nacks);
  cptP->msg = msg;
  cptP->mlen = mlen;
  cptP->retries = M_ALLPACKETS;
  cptP->maxlinear = 2;    /* 1 << 1 */
  cptP->lastpacket = 0;
  cptP->multicast = (type == M_INVALIDATE_OP) ? 0 : 1;
  cptP->pflow = 0;
  midPlistClear(&(cptP->plist)); /* no replies arrived yet */
  ZERO_ADDR_VECTOR(cptP->acks_required);
  ZERO_ADDR_VECTOR(cptP->acks_received);
#if FLOW_TIMES
    cptP->pflow = fti;	/* so can get replies */
#endif

  /* remember who we must get answers from */
  /* NOTE: we must turn off the "VALID" bit (BAH_USED, 0x80000000)
     that lock may (must) use in tolist; we do straight compares of
     required with received and don't use valid bit.  tolist must be
     valid -- it's all we use to know what hosts to send to.
   */
  if (replh == 0) {   /* replh == 0, then replies come to us */
    bcopy((char *)tolist, (char *)(cptP->acksRequired), 
	  midBITARRAYsize(Comm_hnum));
    *(cptP->acksRequired) &= 0x7FFFFFFF; /* turn off high order bit */
  }

  /*
  ** NOTE: maxlinear must be set to 1<<1 for Invalidate requests, since
  ** the reply is processed only by the comm layer, and no lock reply
  ** is expected.  [comm says that the request is finished when
  ** -- the reply packet has come in (actually all packets)
  ** -- all invalidate replies have come in
  ** -- no writers
  **
  ** In addition, the retries should only have M_ALLPACKETS if this
  ** is an invalidate.
  **
  ** (A	lso set scannable (+ acksRequired OK too)
  ** if necessary (replh == 0) before releasing cptP->m)
  */

  /* no one else should touch entry sinced retries not set... */
  mutex_unlock(&(cptP->m));

  /* following is probably awful for 1000's of hosts... tho is minimal
     in information storage size, if have > 200 or so... */
  /* certainly fine for 32 hosts */
  /* for all dests, send message */
  for (i=1,j=0,mask=0x40000000; i<=Comm_hnum; i++) { /**note <= **//* never0*/
    if (mask & tolist[j]) {	/* host is in list */
      cptP->dest = i;
      CommSend(cpti);
      INCR_STAT (mid_comm_stats.multisnds, 1);     /*count of multicast pkts*/
      INCR_STAT (mid_comm_stats.multiBytes, mlen); /*count of multicast bytes*/
    }
    if (!(mask >>= 1)) {	/* shift right one */
      mask = 0x80000000;	/* if fallen off end, restart */
      j++;
    }
  }

#if FLOW_TIMES
  ft_times[fti][FT_SENDDONE] = FT_READ_TIMER;
#endif

  if (replh)			/* replies not to us */
    return(0);			/* done! */

  mutex_lock(&(cptP->m));
  /* here, replies are coming to us.  We will have to sleep. */
  /* gotta ensure that the request is not resent; can use just
     Invalidate resends implicit in the AcksRequired bits */

  cptP->retries |= M_SCANNABLE;
  mutex_unlock(&(cptP->m));

  /* check this guy out, but ALLPACKETS->received reply to op (which we 
     don't need) */

  signal_scanner();

  /* now must wait for replies */

  if ((((type & M_OP_BITS) == M_LOCK_OP) ||
       ((type & M_OP_BITS) == M_LOCK_INV_OP) ||
       ((type & M_OP_BITS) == M_INVALIDATE_OP)) &&
      (midway_net_fns->wantreply)) {
    /* try fast path */
    /* try to get reply without context switching */
    midway_net_fns->wantreply((comm_filter)CommRequestFilter, (int) cpti);
  }

  mutex_lock(&(cptP->m));
    /* strictly speaking, we should only need to test required==received,
       since all others are set for us above... */
    while(!(midAllpackets(cpti) &&
	    midCompareBit(cptP->acksRequired,cptP->acksReceived,
			  Comm_hnum) &&
	    cptP->writers == 0)) {
#if ATM_IFACE
      midway_atm_willblock(); /* XXX copied, not in old ATM code */
#endif

      if (IS_DBG_COMM_REQ) {
	char255 smsg;
	sprintf(smsg,"CommMrequest: waiting for reply\n");
	StatusMsg(smsg);
	sprintf(smsg,"writers=%lu allpackets=%d Required==Received=%d",
		cptP->writers, midAllpackets(cpti),
		midCompareBit(cptP->acksRequired,
			      cptP->acksReceived, Comm_hnum));
	StatusMsg(smsg);
      }

      condition_wait(&(cptP->c),&(cptP->m));
#if ATM_IFACE
      midway_atm_doneblock(); /*XXX see above */
#endif
    }

  cptP->msg = 0;		/* for sanity */
  cptP->retries &= M_SCANCLEAR;
  mutex_unlock(&(cptP->m));

#if FLOW_TIMES
    ft_times[fti][FT_REQDONE] = FT_READ_TIMER;
#endif

  return(0);
} /*CommMrequest */
/*
 ************************************************************************
 *			C o m m P M r e q u e s t			*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
/*
 * send identical request to multiple destinations.  This is assumed
 * to be called from a reply context, so we check to see if this
 * thread is the one making the request.  If so, we re-use the cpt
 * entry carefully... if not, we pass it off to CommMrequest.
 * (it's a Protected Multi-request)
 * 
 * NOTE: Caller is not allowed to send multi-packet messages.
 */
CommPMrequest(lock, type, replh, replt, clk, tolist, msg, mlen)
     u_long	lock;
     u_long	type;		/* kind of message */
     u_long	replh;		/* reply-to host, 0 for us */
     u_long	replt;		/* reply-to tid, 0 for us */
     clock_t	clk;		/* clock for message, 0 for current */
     BITARRAY	*tolist;	/* places to send */
     char	*msg;		/* type dependent message info */
     u_long	mlen;		/* size of message */
{
/*  char255	smsg;	*/	/* for error messages */
    u_long	cpti;		/* client-per-thread index */
    cpt_t	*cptP;
  int		i,j;
  u_long	mask;		/* for checking hosts...  */
  u_long	save_type;
  char		*save_msg;
  u_long	save_mlen;
  u_long	save_dest;
#if 0				/*FLOW_TIMES*/
    u_int	fti;
#endif


  if (IS_DBG_COMM_REQ) {
    fprintf(stderr, 
      "CommPMrequest: Lock=%lu Type=0x%x rep=%lu-%lu clk=%lu TO=0x%x len=%lu\n",
	    lock, type, replh, replt, clk, *tolist, mlen);
  }
  ASSERT (initdone);

  /* see if client-per-thread entry exists yet */
  cpti = (u_long)cthread_data(cthread_self());
  if ((!cpti) || (clk == 0)) {			/* not yet allocated */
    /* pass off */
    return(CommMrequest(lock,type,replh,replt,clk,tolist,msg,mlen)); 
  }
  cptP = &cpt[cpti];

#if 0				/* FLOW_TIMES -don't know if means much here*/
    NEXT_FT;
    ft_times[fti][FT_TYPE] = type | M_REPLY_BIT; /* reply bit marks "client" */
    ft_times[fti][FT_LOCK] = lock;
    ft_times[fti][FT_TID] = (clk != NULL_CLOCK_VALUE)? clk : LocalClock;
    ft_times[fti][FT_SENDREQ] = FT_READ_TIMER;
#endif


  mutex_lock(&(cptP->m));

  /* see if cpt is for our request */
  if ((cptP->c__tid != clk) || (cptP->c__lock != lock)) {
    return(CommMrequest(lock,type,replh,replt,clk,tolist,msg,mlen));
  }
  ASSERT(cpti == cptP->c__rid);
  ASSERT(cptP->c__pnum == 0);
  /* re-using context; must be an ATM (or at least not sockets) call */
  save_type = cptP->c__optype;
  save_msg = cptP->msg;
  save_mlen = cptP->mlen;
  save_dest = cptP->dest;
  cptP->c__optype = type;
  cptP->msg = msg;
  cptP->mlen = mlen;

  /* might be nice to unlock cpt here (esp if send blocks) but we are
     screwed if anything else looks at the cpt... (we've left it
     inconsistent) (unless ALLPACKETS?) mjz30jan94*/

  /* following is probably awful for 1000's of hosts... tho is minimal
     in information storage size, if have > 200 or so... */
  /* certainly fine for 32 hosts */
  /* for all dests, send message */
  for (i=1,j=0,mask=0x40000000; i<=Comm_hnum; i++) { /**note <= **//* never0*/
    if (mask & tolist[j]) {	/* host is in list */
      cptP->dest = i;
      CommSend(cpti);
      INCR_STAT (mid_comm_stats.multisnds, 1);     /*count of multicast pkts*/
      INCR_STAT (mid_comm_stats.multiBytes, mlen); /*count of multicast bytes*/
    }
    if (!(mask >>= 1)) {	/* shift right one */
      mask = 0x80000000;	/* if fallen off end, restart */
      j++;
    }
  }

  cptP->c__optype = save_type;
  cptP->msg = save_msg;
  cptP->mlen = save_mlen;
  cptP->dest = save_dest;

  mutex_unlock(&(cptP->m));

#if FLOW_TIMES
  ft_times[fti][FT_SENDDONE] = FT_READ_TIMER;
#endif

  /* count number of multicast requests */
  INCR_STAT (mid_comm_stats.multireqs, 1);

  return(0);
} /*CommPMrequest */

/*
 ************************************************************************
 *			C o m m X R e q u e s t				*
 *----------------------------------------------------------------------*
 ************************************************************************
 */

CommXRequest(lock, type,dest,msg,mlen,pnum, last, tid)
     u_long	lock;		/* le object 'd art */
     u_long	type;		/* kind of message */
     int	dest;		/* index into host table */
     char	*msg;		/* type dependent message info */
     u_long	mlen;		/* size of message */
     u_long	pnum;		/* packet number */
     u_long	last;		/* last packet */
     u_long	*tid;		/* transaction id */
{
  char255	smsg;	/* for error messages */
  u_long	cpti; /* client-per-thread index */
  cpt_t	*cptP;
  u_long	seqno;
  u_long	i;
  trantab_t	*tt;
  u_long	stid = 0;
  u_long	state = 0;
  int		gotack = 0;
  
  ASSERT (initdone);

  /* see if client-per-thread entry exists yet */
  cpti = (u_long)cthread_data(cthread_self());
  if (!cpti) { /* not yet allocated */
    cpti = cpt_init();
  }
  cptP = &cpt[cpti];

  if (*tid == 0)
    *tid = LocalClock;		/* set Transaction ID */
  
  mutex_lock(&(cptP->m));	/* too restrictive? */

  /* cpt exists, so fill it in, and then send message & sleep */
  cptP->dest = dest;
  cptP->msg = msg;
  cptP->mlen = mlen;
  cptP->c__optype = type;
  cptP->c__mflags = (last?M_LASTPACK:0);
  cptP->c__rhost = Comm_index;
  cptP->c__rid = cpti;
  cptP->c__tid = *tid;
  cptP->c__lock = lock;
  cptP->c__pnum = pnum;
  ZERO_ADDR_VECTOR (cptP->c__nacks);
  cptP->maxlinear = 0;
  cptP->lastpacket = 0;
  cptP->multicast = 0;
  cptP->retries = 0;
  cptP->pflow = 0;
  midPlistClear(&(cptP->plist)); /* no replies arrived yet */
  ZERO_ADDR_VECTOR (cptP->acks_required);
  ZERO_ADDR_VECTOR (cptP->acks_received);
  if (cptP->oldp == 0) {
    cptP->oldp = (c_pbuf_q_t *) CALLOC(midway_comm_send_ackafter+1,
				       sizeof(c_pbuf_q_t *));
  }
  seqno = (pnum+1) % midway_comm_send_ackafter; /*(1,2,3,...,aca-1,0)*/
  cptP->oldp[seqno].msg = msg; /* save old messages for retrans */
  cptP->oldp[seqno].mlen = mlen;

  CommSend(cpti);

  if (!last) {

#ifdef insane
    if (pnum ==0) {		/* set up trantab for ACK replies */
      /* this is somewhat bogus, but allows us to use same ACK code
	 as the stop after n on replies.  perhaps change later.
	 19oct92mjz*/
      tt = TRANTAB(Comm_index, cpti);
      mutex_lock(trantab_lock);
      /* could sanity check old value.  probably not necessary */
      tt->tranid = *tid;
      tt->u.state = M_R_INPROCESS; /* tra la la */
      mutex_unlock(trantab_lock);
    }/*pnum 0*/
#endif /*insane*/

    /* new ackafter support */
    if (seqno == 0) { /* wait for ack */

      cptP->retries |= M_WAITING;
#ifdef insane
      tt = TRANTAB(Comm_index, cpti);
      mutex_lock(trantab_lock);
      if (tt->tranid != *tid) {	/* insanity */
	state = tt->u.state;
	stid = tt->tranid;
	gotack = -1;
      } else {			/* potentially sane */
	if (tt->u.state == M_R_INPROCESS) {
	  tt->u.state = M_R_WAITING;
	  tt->waiter = cpti;	/* XXX prob dont need except for ACK proc */
	} else if (tt->u.state == M_R_ACKED) {
	  if (tt->waiter == pnum) {	/* packet arrived! */
	    gotack = 1;
	    tt->u.state = M_R_INPROCESS; /* revert back to in process */
	    tt->waiter = 0;	/* XXX */
	  }
	  else {
	    /* can't have received > ack, since haven't sent.  wait */
	    tt->u.state = M_R_WAITING;
	    tt->waiter = cpti;	/* XXX */
	  }
	} else {			/* state not inprocess or acked */
	  state = tt->u.state;
	  stid = tt->tid;
	  gotack = -1;
	  cptP->retries &= M_WAITCLEAR; /* undo */
	}
      } /* potentially sane */
      mutex_unlock(trantab_lock);
#else
      /* no way to represent early ack.  can't get one anyway, so don't
	 sweat it. */
      gotack = 0;
#endif

      if (gotack == 0) {		/* no ack yet, must wait */
	cptP->retries |= M_SCANNABLE;	/* see CommRecv comment */
	signal_scanner();

	while (cptP->retries & M_WAITING)
	  condition_wait(&(cptP->c), &(cptP->m));
	
	cptP->retries &= (M_SCANCLEAR & M_RETRANSCLR); /* reset */
      }

#ifdef insane
      if (gotack < 0) {
	if ((gotack == -1)||(IS_DBG_COMM_WNGS)) {	/* error will robinson */
	  sprintf(smsg, "CommXRequest: going to wait after packet %lu tran %lu, but state was bad 0x%lx tran %lu; continuing.", pnum, *tid, state, stid);
	  StatusMsg(smsg);
	  /* this is a really bad idea here... 'cause we can only send to
	     the other side when it's ready... */
	}
      }
#endif /*insane*/

      /* free old messages */
      for (i=0; i<midway_comm_send_ackafter; i++) {
	mid_pbuf_free(cptP->oldp[i].msg);
	cptP->oldp[i].msg =  0; /* know freed */
      }
    } /* new ackafter support */

    mutex_unlock(&(cptP->m));

    return(0);
  }

  /*
   * Last packet -- must wait for reply a-la commRequest
   */
  cptP->retries = M_SCANNABLE; /* can retransmit... condition? DO LAST */
  
  mutex_unlock(&(cptP->m));
  
  signal_scanner();

#ifdef DEBUG
    if (IS_DBG_COMM_REQ) {
	fprintf(stderr,"CommXRequest: about to wait for condition\n");
    }
#endif /*DEBUG*/
    /* now must wait for reply */
    mutex_lock(&(cptP->m));
    while(!(midAllpackets(cpti) &&
	    midCompareBit(cptP->acksRequired,cptP->acksReceived,
			  Comm_hnum) &&
	    cptP->writers == 0)) {
	condition_wait(&(cptP->c),&(cptP->m)); /* do in while? */
	/* NOTE: if we are signalled, the condition is satisfied because
	   we are now the only ones who can 'unsatisfy' it.  However, we
	   might have gotten real unlucky and been preempted while the
	   condition was satisfied; without the check, we will "miss" the
	   signal (it was actually never sent since no one was queued
	   at the time), and we will hang forever.
	   */
    }
#ifdef DEBUG
    if (IS_DBG_COMM_REQ) {
	fprintf(stderr,"CommXRequest: released from condition\n");
    }
#endif /*DEBUG*/

  /* free any old packets left here */
  if (seqno) {			/* whole range = 1..seqno */
    i = 1;
  } else {			/* adjust so get whole range */
    i = 0;
    seqno = midway_comm_send_ackafter - 1;
  }
  for(; i<=seqno; i++) {
    mid_pbuf_free(cptP->oldp[i].msg);
    cptP->oldp[i].msg =  0; /* know freed */
  }
  cpt->retries &= M_SCANCLEAR;	/* JustInCase */
  mutex_unlock(&(cptP->m));

  return(0);
}   /*CommXRequest */

/*
 ************************************************************************
 *			    C o m m S e n d				*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
CommSend(cpti)
     u_long	cpti;		/* client-per-thread index */
{
    cpt_t	*cptP;
    int rc;			/* return code */
    char255	smsg;		/* for error messages */

  /*
   * Fill in the rest of msg header
   */
  cptP = &cpt[cpti];
  cptP->iov[0].iov_base = (caddr_t)&cptP->msgheader; /* reset each time */
  cptP->iov[0].iov_len = sizeof(cptP->msgheader);
  if(cptP->mlen) {
    cptP->iov[1].iov_base = cptP->msg;
    cptP->iov[1].iov_len = cptP->mlen;
    cptP->msgh.msg_iovlen = 2;		/* 2 items in cptP->iov */
  } else {
    cptP->msgh.msg_iovlen = 1;		/* 1 item in cptP->iov */
  }

  /* convert destination host id to IP address & port */
  cptP->sin.sin_addr.s_addr = htbl[cptP->dest].ip;
  cptP->sin.sin_port = htbl[cptP->dest].uport;

  /* ensure only one sender on a socket at a time.  Really necessary? */
  /* use same lock?  need to ensure different than receiver?? ACK! */
  /* not fair either.... worry about this? */
#ifdef SERIALIZE_SENDS
  mutex_lock(scommlock);
    while (sender_active)
      condition_wait(scomm_signal,scommlock);
    sender_active = 1;
  mutex_unlock(scommlock);
#endif /*SERIALIZE_SENDS*/

  /* at last second, update clock and stick it in message */
  mutex_lock(ClockLock);
  LocalClock = ADJUST_CLOCK (LocalClock,1);
  cptP->msgheader.clock = LocalClock;
  mutex_unlock(ClockLock);

#ifdef DEBUG
  if (IS_DBG_COMM_TRACE) {
    midPmsgTrace(cptP->dest, &cptP->msgheader, cptP->msg, cptP->mlen);
  }
#endif /*DEBUG*/
#ifdef DEBUG_PRT
  if (outcall && IS_DBG_COMM_MSGS)
    (*outcall)(cptP->dest, &cptP->sin, &cptP->msgheader, 0, 0, 
	       cptP->msg, cptP->mlen);
#endif /*DEBUG_PRT*/

#ifdef DEBUG
  if (IS_DBG_COMM_SEND) {
    fprintf(stderr, "commsend: about to send message to %lu\n",
	     cptP->dest);
  }
#endif /*DEBUG*/

#ifdef notdef		/* for historical reasons, may delete19oct92mjz */
/*  I am deleting the stuff from CommForward & CommAck. */
#ifdef FAST_UDP
#ifndef FAST_PRIO		/* define if want better worker prio */
  /** This should probably be outside "sender_active" 18oct92mjz **/
  { /* Slow sender down to prevent overruns. WAS 16 Oct 92 */
#define FAST_UDP_SEND_DELAY 10
    static int send_delay = FAST_UDP_SEND_DELAY;
    if (send_delay)		/* allow dynamic change to 0 */
      timer_pause_ms (send_delay);
  }
#endif /* FAST_PRIO */
#endif /* FAST_UDP */
#endif /* notdef */

  rc = midway_net_fns->sendmsg(midway_udpsock, &cptP->msgh, 0, cptP->dest);
  /* returns -1 on err, or # bytes SENT. */

#ifdef SERIALIZE_SENDS
  mutex_lock(scommlock);
    sender_active = 0;		/* note no longer active */
  mutex_unlock(scommlock);
  condition_signal(scomm_signal); /* let others in */
#endif /*SERIALIZE_SENDS*/

  if (rc == -1) {
    sprintf(smsg,"CommSend: failure sendmsg to dest %d (%lu bytes): %s (%d)",
	    cptP->dest,cptP->mlen,
	    ((errno>sys_nerr)?"unknown":sys_errlist[errno]), errno);
    StatusMsg(smsg);
    sprintf(smsg,"...dest ip=%lx port=%d", cptP->sin.sin_addr.s_addr, cptP->sin.sin_port);
    StatusMsg(smsg);
    return(-1);
  } else if (rc != (cptP->mlen+sizeof(cptP->msgheader))) { /* hrrumph.  *MJZ* */
    sprintf(smsg,"CommSend: sendmsg sent only %d bytes out of %d to %lu; continuing anyway.",
	    rc,cptP->mlen+sizeof(cptP->msgheader),cptP->dest);
    StatusMsg(smsg);
  }

#ifdef STATS

  /* count number of bytes sent */
  INCR_STAT(mid_comm_stats.sndBytes, rc);

  /* keep stats by type of message?  For now, just not HOST_OP &SCHED_OP*/
  if (((cptP->msgheader.type & M_OP_BITS) != M_HOST_OP) &&
      ((cptP->msgheader.type & M_OP_BITS) != M_SCHED_OP)) {

    INCR_STAT (mid_comm_stats.snd,1);

    if (cptP->msgheader.flags & M_RETRANS)
      INCR_STAT (mid_comm_stats.retran, 1);

    if (cptP->msgheader.flags & M_LASTPACK){ 
      if (cptP->msgheader.pacnum > mid_comm_stats.lastpack) {
	INCR_STAT(mid_comm_stats.lastpack, cptP->msgheader.pacnum);
      }
#if MSTATS
      switch(cptP->msgheader.type) {
      case M_ACK:
      case M_NACK:
	break;		/* don't count */
      case M_LOCK_OP:
      case M_INVALIDATE_OP:
	INCR_STAT(mid_comm_stats.reqlout,1);
	break;
      case M_LOCK_OP|M_REPLY_BIT:
      case M_LOCK_INV_OP|M_REPLY_BIT:
      case M_INVALIDATE_OP|M_REPLY_BIT:
	INCR_STAT(mid_comm_stats.replout,1);
	break;
      case M_BARRIER_OP:
	INCR_STAT(mid_comm_stats.reqbout,1);
	break;
      case M_BARRIER_OP|M_REPLY_BIT:
	INCR_STAT(mid_comm_stats.repbout,1);
	break;
      default:
	if (cptP->msgheader.type & M_REPLY_BIT) {
	  INCR_STAT(mid_comm_stats.repout,1);
	} else {
	  INCR_STAT(mid_comm_stats.reqout,1);
	}
	break;
      }
#endif /* MSTATS */
    } /* lastpack*/
  } /* non host or sched */
#endif /* STATS */


  /* if big message, lock&wait on cpt entry.  when ack comes back, send
   * next bit.  Uck!  Fix this to blast and retransmit later. 1/9/91mjz
   */
  return(0);
} /* CommSend */

/*
 ************************************************************************
 *			   C o m m A c k				*
 *----------------------------------------------------------------------*
 * Send an ACK to the designated host/thread pair, except if the pkts	*
 * argument is non-zero, in which case we send a NACK with pkts as the  *
 * list of packets we did not receive.					*
 * c0: no nacks
 ************************************************************************
 */
CommAck(host, thread, tid, lock, pnum, pkts)
     u_long	host;		/* destination host */
     u_long	thread;		/* destination thread */
     u_long	tid;		/* transaction ID we're acking too */
     u_long	lock;		/* le object */
     u_long	pnum;		/* need to know what packets we're about */
     u_long	pkts;		/* if non-zero, packets missed in last blast */
{
  char255	smsg;		/* for error messages */
  struct {
    COMMMSG	msgheader;	/* most of an ACK */
    u_long	badpkts;	/* used only for NACK */
  } ackmsg;
  struct sockaddr_in sin;	/* for destination */
  int rc;			/* return code */

  /* fill in rest; should be able to have most of it filled in once */
  /* at least sin & iov */
  ackmsg.msgheader.version = M_PROTOVER;
  ackmsg.msgheader.type = (pkts)? M_NACK : M_ACK;
  ackmsg.msgheader.flags = M_LASTPACK;
  ackmsg.msgheader.unused = 0;
  ackmsg.msgheader.incarnation = Comm_incarn;
  ackmsg.msgheader.source = Comm_index;	/* index of this host */
  ackmsg.msgheader.id = (u_long) cthread_data(cthread_self()); /* SET YET? --mjz */
  ackmsg.msgheader.rhost = 0;
  ackmsg.msgheader.rid = thread; /* destination */
  ackmsg.msgheader.lock = lock;
/*  ackmsg.msgheader.clock = LocalClock;  done just before sending */
  ackmsg.msgheader.tid = tid;
  ackmsg.msgheader.pacnum = pnum;
  ZERO_ADDR_VECTOR(ackmsg.msgheader.nacks); /* no AcksRequired entries here */
  ackmsg.badpkts = pkts;

  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = htbl[host].ip;
  sin.sin_port = htbl[host].uport;
  bzero(sin.sin_zero,sizeof(sin.sin_zero));

#ifdef SERIALIZE_SENDS
  /* ensure only one sender on a socket at a time.  Really necessary? */
  /* use same lock?  need to ensure different than receiver?? ACK! */
  /* not fair either.... worry about this? */
  mutex_lock(scommlock);
    while (sender_active)
      condition_wait(scomm_signal,scommlock);
    sender_active = 1;
  mutex_unlock(scommlock);
#endif /*SERIALIZE_SENDS*/

  /* at last second, update clock and stick it in message */
  mutex_lock(ClockLock);
  LocalClock = ADJUST_CLOCK (LocalClock, 1);
  ackmsg.msgheader.clock = LocalClock;
  mutex_unlock(ClockLock);

#ifdef DEBUG
  if (IS_DBG_COMM_TRACE) {
    midPmsgTrace(host, &(ackmsg.msgheader), (char *) 0, 0);
  }
#endif /*DEBUG*/
#ifdef DEBUG_PRT
  /* should include badpackets in extra section */
  if (outcall && IS_DBG_COMM_MSGS)
    (*outcall)(host, &sin, &(ackmsg.msgheader), 0, 0, (char *)0, 0);
  /*           dest   sock    our hdr   extra extralen msg messagelen */
#endif /*DEBUG_PRT*/


  rc = midway_net_fns->sendto(midway_udpsock,
			      (char *)&ackmsg, sizeof(ackmsg),
			      0, (struct sockaddr *)&sin, sizeof(sin),
			      host);
			      
  /* returns -1 on err, or # bytes SENT. */



#ifdef SERIALIZE_SENDS
  mutex_lock(scommlock);
    sender_active = 0;		/* note no longer active */
  mutex_unlock(scommlock);
  condition_signal(scomm_signal); /* let others in */
#endif /*SERIALIZE_SENDS */

  if (rc == -1) {
    sprintf(smsg,"CommAck: failure sendto ack to dest %lu: %s (%d)",
	    host,
	    ((errno>sys_nerr)?"unknown":sys_errlist[errno]), errno);
    StatusMsg(smsg);
    return(-1);
  } else if (rc != (sizeof(ackmsg))) { /* hrrumph.  *MJZ* */
    sprintf(smsg,"CommAck: sendto sent only %d bytes out of %d to %lu; continuing anyway.",
	    rc,sizeof(ackmsg),host);
    StatusMsg(smsg);
  }
  INCR_STAT (mid_comm_stats.snd,1);		/* sending a packet */
  INCR_STAT (mid_comm_stats.ack,1);		/* it's an ack */
  INCR_STAT (mid_comm_stats.sndBytes, rc);      /* number of bytes sent */

  return(0);
} /* CommAck */


/*
 ************************************************************************
 *			C o m m F o r w a r d				*
 *----------------------------------------------------------------------*
 * Forward a given message, it wasn't for us...				*
 * MESSAGE CLOCK VALUE							*
 ************************************************************************
 */
CommForward(lock, type, host, rhost, rthread, tid, retrans, msg, msglen)
     u_long	lock;		/* el object */
     u_long	type;		/* type of transaction */
     u_long	rhost;		/* source host */
     u_long	rthread;	/* source thread */
     u_long	tid;		/* transaction ID  */
     int	retrans;	/* non zero if retransmission */
     char	*msg;		/* type specific message */
     u_long	msglen;		/* type specific message length */
{
  char255	smsg;		/* for error messages */
  COMMMSG	msgheader;	/* most of an ACK */
  struct iovec	iov[2];		/* for scatter/gather io */
  struct msghdr msgh;
  struct sockaddr_in sin;	/* for destination */
  int rc;			/* return code */

  /* fill in rest; should be able to have most of it filled in once */
  /* at least sin & iov */
  msgheader.version = M_PROTOVER;
  msgheader.type = type;
  msgheader.flags = (retrans) ? (M_LASTPACK | M_RETRANS) :  M_LASTPACK;
  msgheader.unused = 0;
  msgheader.incarnation = Comm_incarn;
  msgheader.source = Comm_index;	/* index of this host */
  msgheader.id = (u_long) cthread_data(cthread_self()); /* SET YET? --mjz */
  msgheader.rhost = rhost;
  msgheader.rid = rthread;
  msgheader.lock = lock;
  msgheader.tid = tid;
  msgheader.pacnum = 0; /* MULTI PACKET MESSAGES?????????? --mjz */
  /* in response to MPM, never.  We should only be forwarding    */
  /* requests.  the only multi-packet request is a barrier one,  */
  /* and that too will never be forwarded.  (prove, using max    */
  /* length of req msg, given packet list must be in it?)	 */
  ZERO_ADDR_VECTOR(msgheader.nacks);

  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = htbl[host].ip;
  sin.sin_port = htbl[host].uport;
  bzero(sin.sin_zero,sizeof(sin.sin_zero));

  iov[0].iov_base = (caddr_t)&msgheader;
  iov[0].iov_len = sizeof(msgheader);
  if(msglen) {
    iov[1].iov_base = msg;
    iov[1].iov_len = msglen;
    msgh.msg_iovlen = 2;		/* 2 items in iov */
  } else {
    msgh.msg_iovlen = 1;		/* 1 item in iov */
  }

  msgh.msg_name = (caddr_t) &sin;	/* where it goes */
  msgh.msg_namelen = sizeof(sin);
  msgh.msg_iov = iov;
  /* msgh.msg_iovlen set above */
  msgh.msg_accrights = (caddr_t) 0; /* no rights */
  msgh.msg_accrightslen = 0;

#ifdef SERIALIZE_SENDS
  /* ensure only one sender on a socket at a time.  Really necessary? */
  /* use same lock?  need to ensure different than receiver?? ACK! */
  /* not fair either.... worry about this? */
  mutex_lock(scommlock);
    while (sender_active)
      condition_wait(scomm_signal,scommlock);
    sender_active = 1;
  mutex_unlock(scommlock);
#endif /*SERIALIZE_SENDS*/

  /* at last second, update clock and stick it in message */
  mutex_lock(ClockLock);
  LocalClock = ADJUST_CLOCK (LocalClock, 1);
  msgheader.clock = LocalClock;
  mutex_unlock(ClockLock);

#ifdef DEBUG
  if (IS_DBG_COMM_TRACE) {
    midPmsgTrace(host, &msgheader, msg, msglen);
  }
#endif /*DEBUG*/
#ifdef DEBUG_PRT
  if (outcall && IS_DBG_COMM_MSGS)
    (*outcall)(host, &sin, &msgheader, 0, 0, msg, msglen);
#endif /*DEBUG_PRT*/

  rc = midway_net_fns->sendmsg(midway_udpsock, &msgh, 0, host);
  /* returns -1 on err, or # bytes SENT. */

#ifdef SERIALIZE_SENDS
  mutex_lock(scommlock);
    sender_active = 0;		/* note no longer active */
  mutex_unlock(scommlock);
  condition_signal(scomm_signal); /* let others in */
#endif /*SERIALIZE_SENDS*/

  if (rc == -1) {
    sprintf(smsg,"CommForward: failure sendmsg ack to dest %d: %s (%d)",
	    host,
	    ((errno>sys_nerr)?"unknown":sys_errlist[errno]), errno);
    StatusMsg(smsg);
    return(-1);
  } 
  else if (rc != (sizeof(msgheader)+msglen)) { /* hrrumph.  *MJZ* */
    sprintf(smsg,"CommForward: sendmsg sent only %d bytes out of %d to %lu; continuing anyway.",
	    rc,sizeof(msgheader)+msglen,host);
    StatusMsg(smsg);
  }

  /* 9 sep 92 mjz: note forwarded in trantab here? */

  INCR_STAT (mid_comm_stats.snd,1);		/* it's a send */
  INCR_STAT (mid_comm_stats.fwd,1);
  INCR_STAT (mid_comm_stats.sndBytes, rc);      /* number of bytes sent */
  INCR_STAT (mid_comm_stats.fwdBytes, rc);      /* number of bytes forwarded */

  return(0);
} /* CommForward */

/*
 ************************************************************************
 *			    C o m m R e c v				*
 *----------------------------------------------------------------------*
 * receive a new work item.  buffer is provided by network receive prog.*
 * work item.  This routine verifies the version # (looping if a bad	*
 * number is received), and updates LocalClock.				*
 ************************************************************************
 */
CommRecv(char **buf, u_long *buflen)
{
  char255	smsg;		/* for error messages */
  int rc;			/* return code */
  int nogood = 1;		/* no good packet received */
  trantab_t *tt;		/* so only one */
  int havelock;			/* have trantab lock */
  COMMMSG	*msgheader;     /* ptr to message header */
  char          *dataP;         /* ptr to message data */
  struct sockaddr_in sin;	/* to see who packet is from */
#if FLOW_TIMES
  u_int		fti;		/* flow times index */
  u_int		ftclk;
#endif
  u_int		ftstart;

  do {				/* while nogood */

    /*
     * Wait until a message is received
     */
    rc = midway_net_fns->rcv(buf, buflen, &ftstart, (struct sockaddr *) &sin);

    /* get pointer to header & data, returned length (rc) includes both */
    msgheader = (COMMMSG *) *buf;
    dataP     = *buf + sizeof(COMMMSG);

#if FLOW_TIMES
    if (msgheader->type & M_REPLY_BIT) {
      ftclk = FT_READ_TIMER;	/* just remeber time */
    } else {
      /* setup new record */
      NEXT_FT;			/* sets fti */
      ft_times[fti][FT_TYPE] = msgheader->type;
      ft_times[fti][FT_TID] = msgheader->tid;
      ft_times[fti][FT_COMMRECV] = FT_READ_TIMER;
      ft_times[fti][FT_ATMRECV] = ftstart;
    }
#endif


#ifdef DEBUG
  if (IS_DBG_COMM_TRACE) {
    midPmsgTrace(0, msgheader, dataP, rc - sizeof(COMMMSG));
  }
#endif /*DEBUG*/
#ifdef DEBUG_PRT
  if (incall && IS_DBG_COMM_MSGS)
    (*incall)(&sin, rc, msgheader, 0, 0, dataP, rc - sizeof(COMMMSG));
#endif /*DEBUG_PRT*/

    if (rc == -1) {
      sprintf(smsg, "ComRecv: failure - recvfrom(): %s (%d)",
	      ((errno>sys_nerr)?"unknown":sys_errlist[errno]), errno);
      StatusMsg(smsg);
      return(-1);
    }

#ifdef DEBUG9
    /* debug ; warning: inet_ntoa returns value in static area (stomp!)*/
/***/ mutex_lock(commlock);
    sprintf(smsg, "received udp dgram from %s size %d",
	    inet_ntoa(sin.sin_addr),  rc);
/***/ mutex_unlock(commlock);
/*dbg*/ fprintf(stderr,"%s\n", smsg); fflush(stderr);
    StatusMsg(smsg);
#endif /*DEBUG9*/

    if (((msgheader->type & M_OP_BITS) != M_HOST_OP) &&	
	((msgheader->type & M_OP_BITS) != M_SCHED_OP)) {
      INCR_STAT (mid_comm_stats.rcv,1);
      INCR_STAT (mid_comm_stats.rcvBytes,rc);
    }

    /* version check */ /* loopback to receive another? */
    /* remove this if & following to silently discard */
    if (msgheader->version != M_PROTOVER) {
      sprintf(smsg, "invalid protocol version received: %lu", 
	      msgheader->version);
      StatusMsg(smsg);
#if STOP_ON_ERROR
      botch("protocol_version");
#endif
      midway_net_fns->free(*buf); /* toss buffer, 'cause continue */
      continue;			/* no sense checking for duplicates... */
    }
#if FLOW_TIMES

    THIS NO LONGER WORKS. 
    Version 3 msg header does not have space for clock value
    in the version field.  WAS 10Jun94

    if (msgheader->type & M_REPLY_BIT) { /* stash in local var? */
      msgheader->FT_STASH_INMSG = ftclk; /* remember clock value */
    } else {
      /* use version to keep fti in */
      msgheader->FT_STASH_INMSG = fti;
    }
#endif

    /* incarnation check. */
    if (msgheader->incarnation != Comm_incarn) {
      union {
	u_long in_addr;		/* gross hack to print inet addr */
	u_char ca[4];		/* split into bytes */
      } u;
      u_short port = sin.sin_port;
      u_short nport = ntohs(port);

      u.in_addr = sin.sin_addr.s_addr;
      sprintf(smsg, "WARNING: received old %d byte packet from %d.%d.%d.%d port %d(or %d) [incarn 0x%lx != our 0x%lx]",
	      rc, u.ca[0],u.ca[1],u.ca[2],u.ca[3],
	      port, nport,
	      msgheader->incarnation, Comm_incarn);
      StatusMsg(smsg);
      midway_net_fns->free(*buf); /* toss buffer 'cause continue */
      continue;
    } /* incarnation mismatch */
    /* ok, potentially good packet now for... */

    if (!(msgheader->type & M_REPLY_BIT)) {
      /* do only for requests; replies caught by having expected tid */
      tt = TRANTAB(msgheader->rhost,msgheader->rid);
      ASSERT( tt->signature == TRANTAB_SIG );
      mutex_lock(trantab_lock);
        havelock = 1;		/* got it */
        if (msgheader->tid < tt->tranid) { /* really old */
	  INCR_STAT(mid_comm_stats.badclk,1); /* should rename */
	}
        else if (msgheader->tid == tt->tranid) {	/* potential dup */
	  
	  if ((tt->u.state == M_R_INPROCESS)||
	      (tt->u.state == M_R_ACKED)) {
	    mutex_unlock(trantab_lock);	/* lots 'o processing, release lock */
	    havelock = 0;
	    if (((msgheader->type & M_OP_BITS) != M_HOST_OP) &&
	        ((msgheader->type & M_OP_BITS) != M_SCHED_OP)) {
	      INCR_STAT(mid_comm_stats.ipp,1);
	    }
	    if (IS_DBG_COMM_WNGS){
	      sprintf(smsg, "CommRecv: in process packet supressed (clock %lu src %lu thread %lu tid %lu pnum %lu)",
		      msgheader->clock, msgheader->rhost, 
		      msgheader->rid,
		      msgheader->tid,
		      msgheader->pacnum);
	      StatusMsg(smsg);
	    }
	  } /* INPROCESS */

	  else if (tt->u.state == M_R_WAITING) {
	    u_long target = 0;
	    u_long pnum = 0;
	    /* In process, but stuck waiting for ACK */
	    target = tt->waiter; /* for "retransmission" */
	    pnum = cpt[tt->waiter].c__pnum;
	    mutex_unlock(trantab_lock);	/* lots 'o processing, release lock */
	    havelock = 0;

	    INCR_STAT(mid_comm_stats.ipp,1);
	    if (IS_DBG_COMM_WNGS){
	      sprintf(smsg, "CommRecv: in process packet supressed (clock %lu src %lu thread %lu tid %lu pnum %lu)",
		      msgheader->clock, msgheader->rhost, 
		      msgheader->rid,
		      msgheader->tid,
		      msgheader->pacnum);
	      StatusMsg(smsg);
	    }
#ifdef NOT_IF_WE_SCAN_THIS_ENTRY
	    /*
	     * 13 oct 92 (mjz): we have to scan, 'cause if we don't and
	     * we retransmit erroneously, and a packet is dropped, 
	     * then the reply thread might wait forever for an ack.
	     *  As long as we have to scan (although I would rather be driven
	     *  by the other side), we might as well not do this... and
	     *  potentially send even more packets.  Think about this
	     *  some more...
	     */
	    /* since waiting, might have lost mod 8 packet.  resend it */
	    mutex_lock(&(cpt[target].m));
	    /* verify that we're still waiting */
	    mutex_lock(trantab_lock);
	    /* note, we might be able to acquire cpti while holding trantab,
	       but I don't want to have any excessive waiting inside
	       trantab, since it locks the whole thing */
	    if (tt->u.state != M_R_WAITING) {
	      havelock = -1;
	    }
	    mutex_unlock(trantab_lock);
	    if ((havelock == 0) && (cpt[target].pnum == pnum) &&
		(cpt[target].retries & M_WAITING)) {
	      CommSend(target); /* resend *//*XXX turn on RETRANS? */
	      INCR_STAT(mid_comm_stats.retran,1);
	    } else {
	      havelock = 0;	/* ensure back to 0 */
	    }
	    mutex_unlock(&(cpt[target].m));
#endif /* not if we scan this entry */
	  } /* WAITING */

	  else if (tt->u.state == M_R_INPRMULTI) {
	    /* would like to do this elsewhere... */
	    u_long seqno = (msgheader->pacnum+1)%midway_comm_send_ackafter;
	    static u_long plist_all[32] = {  0x00000000, /* index by 1 */
	      0x80000000, 0xC0000000, 0xE0000000, 0xF0000000,
	      0xF8000000, 0xFC000000, 0xFE000000, 0xFF000000,
	      0xFF800000, 0xFFC00000, 0xFFE00000, 0xFFF00000,
	      0xFFF80000, 0xFFFC0000, 0xFFFE0000, 0xFFFF0000,
	      0xFFFF8000, 0xFFFFC000, 0xFFFFE000, 0xFFFFF000,
	      0xFFFFF800, 0xFFFFFC00, 0xFFFFFE00, 0xFFFFFF00,
	      0xFFFFFF80, 0xFFFFFFC0, 0xFFFFFFE0, 0xFFFFFFF0,
	      0xFFFFFFF8, 0xFFFFFFFC, 0xFFFFFFFE/*, 0xFFFFFFFF i by 1*/};
#define ZEROP ((unsigned long) 0x80000000)	    
	    /*
	    sprintf(smsg,"CR: h %d th %d tid %d pn %d seq %d",
		    msgheader->rhost, msgheader->rid, msgheader->tid,
		    msgheader->pacnum, seqno);
	    StatusMsg(smsg);
	    */
	    if ((tt->pstart <= msgheader->pacnum) &&
		(msgheader->pacnum < (tt->pstart+midway_comm_send_ackafter))){
	      /* packet in good range*/
	      if (!((ZEROP >> seqno) & tt->plist)) {
		/*
		sprintf(smsg, "CR: adding seqno %d flg 0x%lx src %d th %d tid %d",
			seqno, (ZEROP >> seqno), msgheader->rhost,
			msgheader->rid, msgheader->tid);
		StatusMsg(smsg);
		*/
		/* 1st time */
		tt->plist |= (ZEROP >> seqno);
		if (msgheader->flags & M_LASTPACK) {
		    tt->plist |= 0x00000001; /* damn.  hate shared state */
		    tt->pl_shift = (seqno?(midway_comm_send_ackafter - seqno - 1):0);
		    /* set the following so we can ensure we're last to hit bar*/
		    tt->inproc_max = msgheader->pacnum;
		}

		/* also implies ackafter < 32 */

		/* see if all packets have arrived */
		/* beware all ye that read this code */
		if ( ((tt->plist ^ 0x00000001) /*so fails if last not present*/
		      |ZEROP)	       /*in case last is not 0 */
		    ==
		    /* following yuck because seqno 1,2,...,n,0 not 0,1,...n */
		     (plist_all[midway_comm_send_ackafter] << tt->pl_shift) ) {
/*		      (seqno?(midway_comm_send_ackafter - seqno - 1):0)) ) {*/
		  msgheader->flags |= M_FINAL;
		  tt->u.state = M_R_INPROCESS; /* got 'em all */
		  /*StatusMsg("CR: FINAL!");*/
		}

		nogood = 0;	/* this one can be passed on */
	      }
	      if (((tt->plist ^ plist_all[midway_comm_send_ackafter]) == 0) && /* all packets arrived */
		  /* don't retrans on retrans of packet other than 8th */
		  !((seqno != 0) & (nogood == 1))
		  ) { /* ack necessary */
		u_long ackid;
		/* do ack */
		/* note will not ack if lastpacket has arrived */

		/* note second condition above redundant 'cause: */
		tt->plist = 0;
		tt->pstart += midway_comm_send_ackafter; /* move "window" */
		ackid = tt->pstart - 1; /* always is last in sequence */

		mutex_unlock(trantab_lock);
		havelock = 0;

		CommAck(msgheader->source, msgheader->id,
			msgheader->tid, msgheader->lock, 
			/* no!  msgheader->pacnum*/ ackid,
			0);
		INCR_STAT (mid_comm_stats.ackRecv1, 1);
	      }
	      else if (((seqno == 0) || (msgheader->flags & M_LASTPACK)) && 
		       (nogood == 1)) {	/* potential nak */
		u_long plist = ~(tt->plist);
		/* do nack */
		/* note WILL nack if lastpack has arrived */
		mutex_unlock(trantab_lock);
		havelock = 0;
		if ((seqno == 0) || 
		    (((plist & plist_all[midway_comm_send_ackafter])
		      | ZEROP) /* acct for no 0 */ != 
		     (plist_all[midway_comm_send_ackafter] << 
		      (midway_comm_send_ackafter-seqno-1))) ){
		  /* complicated test above to not ack if we have gotten
		     all the packets and lastpacket has arrived.
		     We really shouldn't need to test this, because 
		     the worker will change the state to inprocess.
		     But... there is a window of vulnerability here. */
		  CommAck(msgheader->source, msgheader->id,
			  msgheader->tid, msgheader->lock, msgheader->pacnum,
			  plist);
		  INCR_STAT (mid_comm_stats.ackRecv2, 1);
		}
	      } /* potential NAK */

	    } /* end of in good range */
	    else if ((msgheader->pacnum == (tt->pstart - 1)) && 
		     (tt->plist == 0)) { /* =0 => no new packets received */
	      mutex_unlock(trantab_lock);
	      havelock = 0;
	      CommAck(msgheader->source, msgheader->id,
		      msgheader->tid, msgheader->lock, msgheader->pacnum,
		      0);
	      INCR_STAT (mid_comm_stats.ackRecv3, 1);
	    }
	  } /* INPRMULT */
	  
	  else if (tt->u.state == M_R_DROPPED){
	    /* do nothing */
	  }
	  
	  else if (tt->u.state == M_R_DONE){ /* we have already done*/
	    /* Mpack request could be retransmitted (last packet only) if
	       reply was dropped.  Other packets are definitely retrans,
	       and should be dropped. mjz24oct92 */
	    if (msgheader->flags & M_LASTPACK) {
	      msgheader->flags |= (M_DOAGAIN | M_FINAL); /* a-ha! */
	      /* following wasn't there first pass; but without it we never
		 wait if we have to retransmit a multipacket reply (and
		 we would then act on retransmissions even though it was
		 in process, so we could have multiple threads servicing
		 duplicate requests.  We must ensure that the state
		 switches back to done, however, even if this is somehow
		 prematurely terminated (or is in fact a duplicate...
		 if it waits, and there is no reply...) */
	      tt->u.state = M_R_INPROCESS; /*XXX*/
	      nogood = 0;		/* act on it */
	    }
	  }
	  
	  else {		/* probably forwarded.  Make sure */
	    if (tt->u.state & M_R_STATUS) { /* gak.  bad state */
	      sprintf(smsg, "CommRecv: unknown trantab state %lx -  dropping (host %lu thread %lu tid %lu)",
		      tt->u.state, msgheader->rhost, msgheader->rid,
		      msgheader->tid);
	      mutex_unlock(trantab_lock); /* unlock for long operation */
	      havelock = 0;
	      StatusMsg(smsg);
	      /* print it again, in case priting is off... */
	      midPmsgIn(&sin, rc, msgheader, 0, 0, dataP, rc - sizeof(COMMMSG));
            } 
	    else if (tt->u.state == 0) { /* 1st time -- treat as new */
	      /* could also be a do-over, if things were forwarded.
		 ensure that the retransmission flag is set, in case
		 things have gotten reordered... */
	      msgheader->flags |= M_RETRANS;
	      tt->u.state = M_R_INPROCESS;
	      nogood = 0;	/* allow it to be consumed */
	    } /* 1st time */
	    else {		/* really forwarded */
#ifdef COMM_FORWARDING
	      u_long fwd = tt->u.fwd;
	      mutex_unlock(trantab_lock); /* unlock for long operation */
	      havelock = 0;
	      CommForward(msgheader->lock, msgheader->type, fwd,
			  msgheader->rhost, msgheader->rid,
			  msgheader->tid, (msgheader->flags & M_RETRANS),
			  dataP, (rc - sizeof(COMMMSG)));
#else /* !COMM_FORWARDING */
	      sprintf(smsg, "CommRecv: unknown trantab state %lx -  dropping (host %lu thread %lu tid %lu)",
		      tt->u.state, msgheader->rhost, msgheader->rid,
		      msgheader->tid);
	      mutex_unlock(trantab_lock); /* unlock for long operation */
	      havelock = 0;
	      StatusMsg(smsg);
	      /* print it again, in case priting is off... */
	      midPmsgIn(&sin, rc, msgheader, 0, 0, dataP, rc - sizeof(COMMMSG));
#endif /*COMM_FORWARDING */
	    } /* really forwarded */
	  } /* probably forwarded */
	} /* tid is saved tid (potential dup )*/
	else {			/* new packet: > saved tid */
	  tt->tranid = msgheader->tid; /* note we got it */
	  if ((msgheader->pacnum == 0) && (msgheader->flags & M_LASTPACK)) {
	    /* single packet request */
	    tt->u.state = M_R_INPROCESS; /* start with in-process */
	  } else {
	    /* multipacket request.  Must pass to worker */
	    u_long seqno = (msgheader->pacnum+1)%midway_comm_send_ackafter;
	    tt->u.state = M_R_INPRMULTI;
	    tt->pstart = 0;
	    tt->inproc = 0;
	    tt->plist = (ZEROP >> seqno);
	    if (msgheader->flags & M_LASTPACK) {
		tt->plist |= 0x00000001; /* damn.  hate shared state */
		tt->pl_shift = (seqno?(midway_comm_send_ackafter - seqno - 1):0);
		/* set the following so we can ensure we're last to hit bar*/
		tt->inproc_max = msgheader->pacnum;
	    } else {
		tt->pl_shift = 0;
		tt->inproc_max = 0;
	    }
	    
	    /*
	    sprintf(smsg, "CR: init multi req seq %d plist 0x%lx h %d th %d tid %d pn %d",
		    seqno, tt->plist, msgheader->rhost, msgheader->rid,
		    msgheader->tid, msgheader->pacnum);
	    StatusMsg(smsg);
	    */
	  }
 	  nogood = 0;
	}
      if (havelock)
        mutex_unlock(trantab_lock);
    } /* ~M_REPLY_BIT */
    else {
      nogood = 0;		/* all requests are good */
    }

    if (nogood) {
      midway_net_fns->free(*buf);
    }
    
  } while(nogood);

  /* note if message was forwarded... */
  if ((!(msgheader->type && M_REPLY_BIT)) &&        /* a request */
      (msgheader->source != msgheader->rhost)) { /* and not from requestor*/
    msgheader->flags |= M_FORWARDED;
  } /* forwarded request */

  /* **mjz** do this only on request/reply?  only on request?  I dunno... */
  /* also don't need to do LocalClock+1, I think */
  /* also maybe do under protection with receive?  only 1 at a time then?
   * triple-sigh.  (keep RCVLOCK stuff above in, and do clock update before
   * release next thread.) */

  /* Lamport clock update */
  /* XXXX this should not update on receive unless it needs too....
     should update on send!! */
  mutex_lock(ClockLock);	/* protect lock update */

  if (GREATER_THAN_CLOCK (msgheader->clock, LocalClock))
    LocalClock = ADJUST_CLOCK (msgheader->clock, 1);
  else
    LocalClock = ADJUST_CLOCK (LocalClock, 1);

  mutex_unlock(ClockLock);

  /*  *buflen = rc - sizeof(COMMMSG); NO LONGER */ /* amount of actual data in buf */
  /* note: following is automatic for atm, not sockets, tho... */
  *buflen = rc;			/* ensure buflen is updated */
  return(0);
} /* CommRecv */

/*
 ************************************************************************
 *		C o m m P r o c e s s P a c k e t			*
 *----------------------------------------------------------------------*
 * The old guts of CommWorker.  Processes incoming packets.  Here so	*
 * individual threads can process thier own replies.			*
 ************************************************************************
 */
CommProcessPacket(pbuf pbufP,
		  u_long buflen
		  )
{
  u_long	*bufp;
  u_long	thread;		/* src thread id on send, dest on reply*/
  cpt_t		*cptP;
  COMMMSG	*mP;		/* ptr to message header */
  register u_char msg_type;
  int		rcverrs;	/* local receive errors */
  u_long	pnum;		/* packet number << 1 */
  long		pindex;		/* index into bitarray */
  char255	smsg;		/* for errors */
  HREPLYMSG	*hrp;		/* to pick thru HOST_OP replies */
  int		i,j;
  u_long	pi;		/* packet index */
  comm_request_t rc;
  trantab_t *tt;
  int		lasthostop = 0;	/* know all HOST_OP replies processed */
  int		invalSelf = 0;	/* !=0 if told to invalidate ourselves */
#if TSTATS
  tv_t startTime, endTime, diffTime;
#endif TSTATS
#if FLOW_TIMES
  u_int fti;
  u_int *ftp;
  u_int ft_startop = 0, ft_stopop = 0;
  extern int ft_rcv_time_start, ft_rcv_time_copy, ft_rcv_time_rel;
#endif

    /* get ptr to message header & data */
    mP = (COMMMSG *) pbufP;
    buflen -= sizeof(COMMMSG);	/* amount of actual data in buf */
    msg_type = mP->type;

#if FLOW_TIMES
    if (!(msg_type & M_REPLY_BIT)) {
      register i = mP->FT_STASH_INMSG;
      ft_times[i][FT_ATMCOPY] = ft_rcv_time_copy;
      ft_times[i][FT_ATMREL] = ft_rcv_time_rel;
      ft_times[i][FT_WORKER] = FT_READ_TIMER;
      ft_times[i][FT_LOCK] = mP->lock;
    }
#endif

    if (IS_DBG_COMM_WORK) {
      sprintf(smsg,
	     "CommWorker: message - type %x thread %lu tid %lu #%d length %lu",
	      msg_type, mP->rid, mP->tid, mP->pacnum, buflen);
      StatusMsg(smsg);
    }

#define NEXTPACKET goto nextpacket

    /* I think I have a valid message -- decode based on type */
    /* first process ACKs 1-required in message 2-invalidatereplies */
    /* CASE thread == 0??? perchance on InvalidateReplies; must look up
     * using lock id... *mjz* */
    bufp = (u_long *)(((char *)pbufP) + sizeof(COMMMSG));
    
    if (msg_type & M_REPLY_BIT) {
      thread = mP->rid;
      cptP = &cpt[thread];
      
      if (IS_DBG_COMM_WORK) {
	sprintf(smsg,"...in REPLY; tid expected is %lu\n",cptP->c__tid);
	StatusMsg(smsg);
      }

      mutex_lock(&(cptP->m)); /* ensure we don't interfere w scanner */
      if (IS_DBG_COMM_WORK) {
	StatusMsg("worker: after lock of cpt entry");
      }

      /* could have clock modified after we test, which would be bad */
      /* first, ensure is a reply to something we've sent */
      if (cptP->c__tid != mP->tid) {

	if (IS_DBG_COMM_WNGS) {
	  sprintf(smsg,"Warning - bad transaction id received %lx != %lx",
		  mP->tid, cptP->c__tid); /* debug, sorta */
	  StatusMsg(smsg);
	}
	mutex_unlock(&(cptP->m));

	/* gotta check if this is a potential ackafter packet; if so,
	   ack away Just In Case (don't need lock...)*/
	if ((msg_type != M_ACK) && (msg_type != M_NACK)) {
	  /* This one might actually be asking for an ACK.  Test theory */
	  if (((mP->pacnum+1) % midway_comm_reply_ackafter) == 0) {
	      /* Well, there's no ack pending.  Must be a retrans. Ack! */
	    CommAck(mP->source, mP->id, mP->tid, mP->lock, mP->pacnum, 0);
	    INCR_STAT (mid_comm_stats.ackWorker1, 1);
	  }
	} /* don't bother looking at acks or nacks */

	/* Check if this really is just a retransmission from a previous message.
	 * For debugging only.  WAS 24Mar93
	 */
	if (cptP->c__tid == ADJUST_CLOCK(mP->tid,1)) {
	  INCR_STAT (mid_comm_stats.badtidretrans,1);
	}

	INCR_STAT (mid_comm_stats.badtid,1);
	NEXTPACKET;		/* old, drop this packet */
      }

      /* note we got a packet on this connection - don't retrans, please 
       * Handle multicast requests below.
       */
      if (cptP->multicast == 0)
	cptP->retries &= M_RETRANSCLR; /* don't retrans right now */

      /*
       * Acks
       */

      if (msg_type == M_ACK) {	/* ACK REPLY */
	int dup = 0;
	int done = 0;
	if (cptP->c__optype & M_REPLY_BIT) {	/* ack of a waiting reply th */
	  trantab_t *tt;
	  tt = TRANTAB(mP->source, cptP->c__rid);/* get original msg TranTab */
	  ASSERT( tt->signature == TRANTAB_SIG );
	  mutex_lock(trantab_lock);
	  if (tt->tranid == mP->tid) {	/* paranoia.  shouldn't happen. */
	    /* yes, but if != mP->tid, probably operating on retrans of old
	       request which has since been declared bad because of new
	       request.  Release guy if waiting. */
	    /* (we passed the tid check above) */
	    if (tt->u.state == M_R_INPROCESS) {
	      if (mP->pacnum > 0) {	/* real ACK */
		tt->u.state = M_R_ACKED; /* ack arrived with packet number */
		tt->waiter = mP->pacnum;
	      }
	    }
	    else if (tt->u.state == M_R_WAITING) {
	      if (mP->pacnum == cptP->c__pnum) { /* got it! */
		tt->u.state = M_R_INPROCESS;
		tt->waiter = 0;	/* zap *//*XXX Still need? */
		done = 1;
	      }
	    }
	    else if (tt->u.state == M_R_ACKED) {
	      /* in process, but acked */
	      if (mP->pacnum > tt->waiter) { /* newer ack */
		/* therefore previous one must have been a duplicate */
		tt->waiter = mP->pacnum; /* note ack arrived */
	      } else {
		dup = 1;
	      }
	    }
	  } /* paranoia */
	  else {
	    done = 2;		/* failed tid check */
	  }
	  mutex_unlock(trantab_lock);
	} /* ack of reply */
	else {			/* ack of requesting thread */
	  if ((mP->pacnum == cptP->c__pnum) && (mP->pacnum > 0))
	    /* >0, 'cause ==0 => "i'm working on it" */
	    done = 1;
	}
	if (done == 1) {
	  if (cptP->retries & M_WAITING) {
	    cptP->retries &= M_WAITCLEAR;	/* clear waiting flag */
	  } else {
	    sprintf(smsg, "CommWorker: ACK Came in to release thread(%lu pack %lu), but thread is not waiting ret=0x%lx, signalling anyway.",
		   thread, mP->pacnum,
		   cptP->retries);
	    StatusMsg(smsg);
	  }
	} /* done == 1*/
	else if (done == 2) {	/* failed tid.  Release anyway if waiting */
	  /* we passed a tid check above */
	  if ((cptP->retries & M_WAITING) &&
	      (cptP->c__pnum == mP->pacnum)) {
	    cptP->retries &= M_WAITCLEAR;	/* clear waiting */
	  } else {
	    done = 0;		/* don't release if not waiting */
	    dup = 1;		/* must be a dup of some sort */
	  }
	}
	mutex_unlock(&(cptP->m));
	/* note that even if this was an ACK, either this one or the
	   last packet was a duplicate.  Since we don't have any spot
	   to note that except IPP, inc that stat. */
	if (dup) {
	  INCR_STAT(mid_comm_stats.ipp,1);
	}
	if (done){
	  condition_broadcast(&(cptP->c)); /* let waiter go */
	}
	NEXTPACKET;		/* reset retransclear, now done. */
      }				/* ACK REPLY */


      /*
       * nacks
       */
      if (msg_type == M_NACK) {	/* NACK REPLY */
	/* ok.  Want to resend packets in nack list */
	if ((cptP->retries & M_WAITING) &&
	    (cptP->c__pnum == mP->pacnum)) {	/* valid NAK */

	    int ackafter;
	    u_long saveflag = cptP->c__mflags;
	    char * savemsg = cptP->msg; /* in case last set not complete */
	    u_long savemlen = cptP->mlen;

	    if (cptP->c__optype & M_REPLY_BIT) { /* ack of a waiting reply thread */
		ackafter = midway_comm_reply_ackafter;
	    } else {
		ackafter = midway_comm_send_ackafter;
	    }

	  /*
	   * The Deal.  *bufp is the inverse of the "plist" in CommRecv.
	   * the index is by "seqno", which has the relation
	   * seqno = (packet_number+1)%midway_comm_reply_ackafter
	   * concrete example, with ackafter = 8:
	   *       packet# ->15 8 910 11121314 (wraps)
	   *                  7 0 1 2  3 4 5 6
	   *       bits       1 1 1 1  1 1 1 1
	   *       seqno   -> 0 1 2 3  4 5 6 7
	   * Now, we never have to recover packet 7, seqno 0, because it
	   * provoked the NAK in the first place.  Therefore, to get back
	   * the packet # from the seqno, do current packet # (7, 15, ...)
	   * - (ackafter# - seqno).  For example, 7 - (8 - 1) -> resending
	   * packet #'d 0.
	   */

	  cptP->lastpacket = 0;
	  for(pi = 0x40000000, i=1; /* no need to resend 0, it provoked nak */
	      i < ackafter; pi>>=1, i++) {
	    /* here, i plays the role of seqno */
	    if (pi & *((u_long *)bufp)) { /* given packets not sent */
	      cptP->msg = cptP->oldp[i].msg;
	      cptP->mlen = cptP->oldp[i].mlen;
	      cptP->c__pnum = mP->pacnum - (ackafter-i);
	      CommSend(thread);
	    }
	  }
	  cptP->c__mflags = saveflag;
	  cptP->c__pnum = mP->pacnum;
	  cptP->msg = savemsg; /* could probably get away w/0 len now */
	  cptP->mlen = savemlen;
	} /* valid nak */

	mutex_unlock(&(cptP->m));
	NEXTPACKET;
      } /* NACK REPLY*/


      /*
       * other ("real") replies
       */

      /* add others to AcksRequired if necessary */
      MERGE_ADDR_VECTOR (cptP->acks_required, mP->nacks);
      if (midTestBit (mP->nacks.bitArray, Comm_index)) {
	  /* we have just been told (indirectly) that we should invalidate
	     ourselves.  I hope this is a lock reply! */
	  invalSelf = 1;
      }

      /*
       * Handle multicasts.
       * Expect a reply from all recipients of the request 
       */
      if ((cptP->multicast != 0) ||
	  ((msg_type & M_OP_BITS) == M_LOCK_INV_OP) ||
	  ((msg_type & M_OP_BITS) == M_INVALIDATE_OP)) { /* this is an invalidate ack */

	if (IS_DBG_COMM_WORK) {	/* break this one out??? */
	  sprintf(smsg,"MulticastReply: adding %lu to acksReceived", mP->source);
	  StatusMsg(smsg);
	  sprintf(smsg,"allpackets %d areq=arcv %d areq %lx arcv %lx",
		  midAllpackets(thread),
		  midCompareBit(cptP->acksRequired,
				cptP->acksReceived,
				Comm_hnum),
		  cptP->acksRequired[0], cptP->acksReceived[0]);
	  StatusMsg(smsg);
	}

	midSetBit(cptP->acksReceived, mP->source);
	if (midAllpackets(thread) &&
	    midCompareBit(cptP->acksRequired,cptP->acksReceived,
			  Comm_hnum)) {

	  if (IS_DBG_COMM_WORK) {
	    sprintf(smsg,
		    "...all pkts+acks rcvd. writers=%d", cptP->writers);
	    StatusMsg(smsg);
	  }

	  cptP->retries &= M_RETRANSCLR; /* don't retrans right now */
	  cptP->retries &= M_SCANCLEAR; /* no more retrans necessary */

	  if ((((msg_type & M_OP_BITS) == M_LOCK_INV_OP) ||
	       ((msg_type & M_OP_BITS) == M_INVALIDATE_OP)) && /* this is an invalidate ack */
	      (cptP->writers == 0))
	    condition_signal(&(cptP->c));	/* release waiting thread */
	}
      } 

/*    if (mP->type & ~(M_INVALIDATE_BIT | M_ACK | M_NACK | M_REPLY_BIT |
		     M_PAGE_UPDATE | M_PAGE_COPYSET)) {
      if (mP->type & (M_HOST_OP | M_LOCK_OP | M_BARRIER_OP | M_ALLOCATE |
		    M_ALLOC_INFO | M_SCHED_OP | M_PAGE_DATA)) { 
*/
      if (((msg_type & M_OP_BITS) == M_LOCK_OP) ||
	  ((msg_type & M_OP_BITS) == M_LOCK_INV_OP) ||
	  ((msg_type & M_OP_BITS) == M_BARRIER_OP) ||
	  ((msg_type & M_OP_BITS) == M_ALLOCATE) ||
	  ((msg_type & M_OP_BITS) == M_ALLOC_INFO) ||
	  ((msg_type & M_OP_BITS) == M_HOST_OP) ||
	  ((msg_type & M_OP_BITS) == M_SCHED_OP) ||
	  ((msg_type & M_OP_BITS) == M_PAGE_DATA)) {
	/* LOCK, BARRIER and other non-multicasted replies that release
	   the sender. */

	/*process packet number info. */

	if (mP->pacnum > cptP->lastpacket)
	  cptP->lastpacket = mP->pacnum;

	pnum = mP->pacnum << 1;
	if (pnum == cptP->maxlinear) { /* next in sequence */
	  cptP->maxlinear += 2; /* bump */
	} else {
	  pindex = mP->pacnum - (cptP->maxlinear >> 1);
	  /* check for duplicates.  It's a duplicate if (1) pindex is
	   * less than 0, since that means it was one that arrived in
	   * sequence, or (2) it is in the .plist bitvector.
	   */
	  if ((pindex < 0) || midPlistTest(&(cptP->plist),pindex)) {
	    if (IS_DBG_COMM_WNGS) {
	      sprintf(smsg,
		      "comm_worker: Packet %lu tid %lu dropped as duplicate.",
		      mP->pacnum, mP->tid);
	      StatusMsg(smsg);
	    }
	    /* This one might actually be asking for an ACK.  Test theory */
	    if (((mP->pacnum+1) % midway_comm_reply_ackafter) == 0) {
	      /* good shot */
	      if (cptP->ackpnum == 0) {
		/* Well, there's no ack pending.  Must be a retrans. Ack! */
		CommAck(mP->source, mP->id, mP->tid, mP->lock, mP->pacnum, 0);
		INCR_STAT (mid_comm_stats.ackWorker2, 1);
	      }
	    }
	    mutex_unlock(&(cptP->m));
	    INCR_STAT (mid_comm_stats.again,1);
	    NEXTPACKET;		/* drop this packet as duplicate */
	  } else {
	    midPlistSet(&(cptP->plist), pindex);
	  }
#if STATS			/* keep track when we get out of order */
	  if ((cptP->maxlinear & 1) == 0) {
	    cptP->maxlinear |= 1;
	    INCR_STAT (mid_comm_stats.mpseqfail,1);
	  }
#else
	  cptP->maxlinear |= 1; /* ensure will never match again */
#endif
	}

	if (mP->flags & M_LASTPACK){ /*this IS the largest pacnum we will see*/
	  cptP->retries |= M_ALLPACKETS;
	  if (midAllpackets(thread) &&
	      midCompareBit(cptP->acksRequired,cptP->acksReceived,
			    Comm_hnum)){
	    cptP->retries &= M_SCANCLEAR;	/* no need to scan */
	    /*
	     * What happens if a duplicate, or retrans, reply comes
	     * in AFTER we have released the originally waiting thread?
	     * Nothing bad in Comm, perhaps sending another condition_signal
	     * to cpt[].c, but that should fall on the floor.  However, if
	     * it is a lock reply, the lock layer may well decide it is
	     * necessary to scribble on some memory that a running thread
	     * thinks it has exclusively.  This Would Be Bad.  The
	     * following hack should ensure any communication about a
	     * lock that we've already decided to release is dropped.
	     *   Actually, this is probably superfluous; packet index stuff
	     * will drop duplicate packets too...  Hrrumph.
	     */
	    /* ensure retrans of replies die */
	    cptP->c__tid = ADJUST_CLOCK(cptP->c__tid,1);
	    /* (perhaps due to retrans of request, get retransed reply
	       but got first reply too... */
	  }
	}

	if (((mP->pacnum+1)%midway_comm_reply_ackafter) == 0) {
	  /* We should ack after this packet arrives. Note that... */
	  cptP->ackpnum = mP->pacnum;
	  cptP->ackthread = mP->id;
	}


	cptP->writers++;	/* thread will write to memory. */
      } /* type not invalidate */ /* lock & barrier only */

      if (cptP->multicast) {
	/* non-Invalidate multicast requests... these expect a reply
	 * that is handled at a higher level.  Increment writer count
	 * so that the release check after the switch will release
	 * the waiter.  (It should never be true for multicast
	 * that multiple packets can become outstanding... so strictly
	 * speaking, we don't need the writer check if we are multicast.)
	 * (I.e. it could be fixed below :-) --mjz27aug92)
	 */
	cptP->writers++;
      }
	   

      mutex_unlock(&(cptP->m));

#if MSTATS
      if (((msg_type & M_OP_BITS) != M_HOST_OP) &&
	  ((msg_type & M_OP_BITS) != M_SCHED_OP) &&
	  (mP->flags & M_LASTPACK)) {
	/* know acks, nacks already filtered out */
	/* duplicates should be filtered too, no?  Only for multipacket,
	   I guess. */
	switch(msg_type) {
	case M_LOCK_OP|M_REPLY_BIT:
	case M_LOCK_INV_OP|M_REPLY_BIT:
	case M_INVALIDATE_OP|M_REPLY_BIT: /* ? separate for inv? */
	  INCR_STAT(mid_comm_stats.replin,1);
	  break;
	case M_BARRIER_OP|M_REPLY_BIT:
	  INCR_STAT(mid_comm_stats.repbin,1);
	  break;
	case M_PAGE_DATA | M_REPLY_BIT:
	  INCR_STAT(mid_comm_stats.pfrepin,1);
	  break;
	case M_PAGE_UPDATE | M_REPLY_BIT:
	  INCR_STAT(mid_comm_stats.purepin,1);
	  break;
	case M_PAGE_COPYSET | M_REPLY_BIT:
	  INCR_STAT(mid_comm_stats.csrepin,1);
	  break;
	default:
	  INCR_STAT(mid_comm_stats.repin,1);
	  break;
	}
      }
#endif /*MSTATS*/

    } /* is a REPLY, process acks */

    else {
      /* request */
#if MSTATS
      if ((mP->flags & M_LASTPACK) && !(mP->flags & M_DOAGAIN)) {
	switch(msg_type) {
	case M_HOST_OP:
	case M_SCHED_OP:
	  break;		/* do not count */
	case M_LOCK_OP:
	case M_LOCK_INV_OP:
	case M_INVALIDATE_OP:
	  INCR_STAT(mid_comm_stats.reqlin,1);
	  break;
	case M_BARRIER_OP:
	  INCR_STAT(mid_comm_stats.reqbin,1);
	  break;
	case M_PAGE_DATA:
	  INCR_STAT(mid_comm_stats.pfreqin,1);
	  break;
	case M_PAGE_UPDATE:
	  INCR_STAT(mid_comm_stats.pureqin,1);
	  break;
	case M_PAGE_COPYSET:
	  INCR_STAT(mid_comm_stats.csreqin,1);
	  break;
	default:
	  INCR_STAT(mid_comm_stats.reqin,1);
	  break;
	}
      }
#endif /*MSTATS*/
      if (!(mP->pacnum == 0) && (mP->flags & M_LASTPACK)) {
	/* multipacket request */
	if (msg_type != M_BARRIER_OP) {
	  sprintf(smsg, "Worker: ERROR - multipacket request, but type (0x%lx) is not BARRIER.  Your results are probably wrong.", mP->type);
	  StatusMsg(smsg);
	}
      } /* multipacket request */
    } /* request */


/*
 * for request calls other than locks, just update to "done"
 * This should work... (allocinfo?) and allow me to not update the others.
 *			<allocinfo could forward>
 */
#define SIMPLE_TRAN_DONE {\
     mutex_lock(trantab_lock); \
     tt = TRANTAB(mP->rhost,mP->rid); \
     ASSERT( tt->signature == TRANTAB_SIG ); \
     if (tt->tranid != mP->tid) { \
       if (!(mP->flags & M_DOAGAIN)) { \
/*	 sprintf(smsg,"CommWorker: TransTab tranid %lu != our tranid %lu updating state to done (saved state 0x%lx)", \
		 tt->tranid, mP->tid, tt->u.state);*/ \
         mutex_unlock(trantab_lock); \
/*         StatusMsg(smsg);*/ \
         INCR_STAT(mid_comm_stats.useless,1); \
       } \
       else { \
         mutex_unlock(trantab_lock); \
         INCR_STAT(mid_comm_stats.useless,1); \
       } \
     } \
     else { \
       tt->u.state = M_R_DONE; \
       mutex_unlock(trantab_lock); \
     } \
     }



    /*
    **
    ** Handle message based on message type
    **
    */

    switch(msg_type) {
      /* just HOST_OP is special case, only handled by coordinator */

    case M_HOST_OP | M_REPLY_BIT:
      /* perhaps do under lock in case retransmission/duplicate? */
      if (IS_DBG_COMM_WORK)
	fprintf(stderr,"in HostOp reply, writers=%lu, allpack=%d\n",
		cptP->writers,midAllpackets(thread));

      hrp = (HREPLYMSGP) bufp;	/* point to first entry */
      while( buflen >= sizeof(HREPLYMSG) ) { /* still work to do */
	htbl[hrp->index].ip = hrp->ipaddr;
	htbl[hrp->index].uport = hrp->udpport;
	hrp++;			/* point to next hreply entry */
	buflen -= sizeof(HREPLYMSG);
      } /* buflen large enough */

      break;    /* requesting thread is released below */

    case M_SCHED_OP:
      if (IS_DBG_COMM_WORK)
	fprintf(stderr,"worker: about to mid_rsched_rwait\n");

      rc = mid_rsched_rwait(mP->rhost,mP->rid,mP->tid);

      if (rc!=M_R_UNCHANGED) {
	mutex_lock(trantab_lock);
	tt = TRANTAB(mP->rhost,mP->rid);
	ASSERT( tt->signature == TRANTAB_SIG );
	if (tt->tranid != mP->tid) {
	  if (!(mP->flags & M_DOAGAIN)) {
/*	    sprintf(smsg,"CommWorker: TransTab tranid %lu != our tranid %lu updating state to 0x%lx (saved state 0x%lx)",
		    tt->tranid, mP->tid, rc, tt->u.state);*/
	    mutex_unlock(trantab_lock);
/*	    StatusMsg(smsg);*/
	    INCR_STAT(mid_comm_stats.useless,1);
	  }
	  else {
	    mutex_unlock(trantab_lock);
	    INCR_STAT(mid_comm_stats.useless,1);
	  }
	}
	else {
	  tt->u.state = rc;
	  mutex_unlock(trantab_lock);
	} 
      } /* change state */

      if (IS_DBG_COMM_WORK)
	fprintf(stderr,"worker: done mid_rsched_rwait\n");

      break;


    case M_SCHED_OP | M_REPLY_BIT:
      if (IS_DBG_COMM_WORK) {
	fprintf(stderr,"about to midRemoteReleaseRply");
	fprintf(stderr,"writers=%lu allpackets=%d\n",
		cptP->writers, midAllpackets(thread));
      }

      midRemoteReleaseRply(bufp,(u_long)buflen);

      if (IS_DBG_COMM_WORK)
	fprintf(stderr,"done midRemoteReleaseRply");
      break;    /* requesting thread is released below */

    case M_LOCK_OP:
      if (IS_DBG_COMM_WORK)
	StatusMsg("about to midLockRequest");

      rc = midLockRequest(mP->lock,mP->rhost,mP->rid,mP->tid,
		     (mP->flags&(M_RETRANS|M_DOAGAIN|M_FORWARDED)),
		     bufp, (u_long)buflen, mP->FT_STASH_INMSG);

      if (rc!=M_R_UNCHANGED) {
	mutex_lock(trantab_lock);
	tt = TRANTAB(mP->rhost,mP->rid);
	ASSERT( tt->signature == TRANTAB_SIG );
	if (tt->tranid != mP->tid) {
	  if (!(mP->flags & M_DOAGAIN)) {
/*	    sprintf(smsg,"CommWorker: TransTab tranid %lu != our tranid %lu updating state to 0x%lx (saved state 0x%lx)",
		    tt->tranid, mP->tid, rc, tt->u.state);*/
	    mutex_unlock(trantab_lock);
/*	    StatusMsg(smsg);*/
	    INCR_STAT(mid_comm_stats.useless,1);
	  }
	  else {
	    mutex_unlock(trantab_lock);
	    INCR_STAT(mid_comm_stats.useless,1);
	  }
	}
	else {
	  tt->u.state = rc;
	  mutex_unlock(trantab_lock);
	} 
      } /* change state */

      if (IS_DBG_COMM_WORK)
	StatusMsg("done midLockRequest");

#if FLOW_TIMES
      ft_times[mP->FT_STASH_INMSG][FT_WORKERUPDT] = FT_READ_TIMER;
#endif
      break;

    case M_LOCK_OP | M_REPLY_BIT:		/* lock reply */
    case M_LOCK_INV_OP | M_REPLY_BIT: /* no one was reading */
      if (IS_DBG_COMM_WORK) {
	StatusMsg("about to midLockReply");
	sprintf(smsg,"writers=%lu allpackets=%d Required==Received=%d",
		cptP->writers, midAllpackets(thread),
		midCompareBit(cptP->acksRequired,
			      cptP->acksReceived, Comm_hnum));
	StatusMsg(smsg);
      }

#if TSTATS
      my_gettimeofday(&startTime, 0);
#endif /*TSTATS*/
#if FLOW_TIMES
      ft_startop = FT_READ_TIMER;
#endif

      midLockReply(mP->lock,mP->source,mP->id,mP->tid,bufp,(u_long)buflen,
		   0 /* may block */);

#if FLOW_TIMES
      ft_stopop = FT_READ_TIMER;
#endif
#if TSTATS
      my_gettimeofday(&endTime, 0);
      time_diff(&startTime, &endTime, &diffTime);
      INCR_STAT(mid_comm_stats.copytime, TV_TO_DOUBLE(diffTime));
      INCR_STAT(mid_comm_stats.copynum, 1);
#endif /*TSTATS*/


      if (invalSelf) {
	if ( midLockFastInvalidate(mP->lock,Comm_index,mP->id,
				   mP->tid,cptP->acksRequired) ) {
	  /* note: as a side-effect, acksRequired could have been updated,
	     even if we could not invalidate ourselves (because of an active
	     reader) */
	  mutex_lock(&(cptP->m));
	  midSetBit(cptP->acksReceived,Comm_index); /* did it */
	  mutex_unlock(&(cptP->m));
	}
      }
	

      if (IS_DBG_COMM_WORK)
	StatusMsg("done midLockReply");
      break;    /* requesting thread is released below */

    case M_INVALIDATE_OP:
      if (IS_DBG_COMM_WORK) {
	sprintf(smsg,"about to Invalidate lock=%lx host=%lu rid=%lu tid=%lu\n",
		mP->lock,mP->rhost,mP->rid,mP->tid);
	StatusMsg(smsg);
      }

      midLockInvalidate(mP->lock,mP->rhost,mP->rid,mP->tid, 
			0/* not internal*/,
			bufp, (u_long) buflen);

/*      SIMPLE_TRAN_DONE; NOOOOOooo.  If we go to DONE, then a retransmitted
	lock might go to DROPPED, and then we will never retransmit an
	invalidate reply, even if the original reply was lost.  Today,
	just go to know-nothing. 30sep92*/
      mutex_lock(trantab_lock);
      tt = TRANTAB(mP->rhost,mP->rid);
      ASSERT( tt->signature == TRANTAB_SIG );
      /* do not update state if something newer has come in.  I wish we could
	 just squash any processing in progress, but That Would Be Hard. */
      if (tt->tranid == mP->tid) {
	tt->u.state = M_R_DOAGAIN;
	mutex_unlock(trantab_lock);
      } else {
	mutex_unlock(trantab_lock);
	INCR_STAT(mid_comm_stats.useless,1);
      }

      if (IS_DBG_COMM_WORK) {
	sprintf(smsg,"done Invalidate lock=%lx\n",mP->lock);
	StatusMsg(smsg);
      }
      break;

    case M_INVALIDATE_OP | M_REPLY_BIT:
      /*
       * The invalidate represented by this reply has been counted.
       * However, if this indicated we should also invalidate ourself,
       * do it here (since no invalidate request for ourself is going
       * to be issued)
       */

      if (invalSelf) {
	if ( midLockFastInvalidate(mP->lock,Comm_index,mP->id,
				   mP->tid,cptP->acksRequired) ) {
	  /* note: as a side-effect, acksRequired could have been updated,
	     even if we could not invalidate ourselves (because of an active
	     reader) */
	  mutex_lock(&(cptP->m));
	  midSetBit(cptP->acksReceived,Comm_index); /* did it */
	  mutex_unlock(&(cptP->m));
	}
      }

      /*
       * Don't release requesting thread below; does ACK stuff and also
       * decrements writers.  Instead, remove Reply bit, and do the release
       * check inline.  Perhaps we could do this in the end; I didn't
       * do it already, so I am playing it safe by not doing it here.
       * -16feb94mjz
       */
      msg_type = M_INVALIDATE_OP;	/* Don't release requesting thread below */

      if ((cptP->writers == 0) && /* no other writers */
	  (midAllpackets(thread)) &&
	  (midCompareBit(cptP->acksRequired,cptP->acksReceived,
			  Comm_hnum))) { 
	condition_signal(&(cptP->c)); /* all done, let waiter go */
      }

      break;

    case M_LOCK_NEWOWN | M_REPLY_BIT:			/* LOCK NEW OWNER */
      /*
       * 1. stop retrans?  if good, then retransing now doesn't help.
       *    if bad, we will allow a retrans in the future...  how block
       *    out other replies if we are working?  does it matter?
       * 2. ask lock layer if hint is current.
       * 3. if current, fix, resend to new guy and reset retransmission
       * 4. if not current, silently ignore :-)
       */
      mutex_lock(&(cptP->m)); /* re acquire mutex */
        if (cptP->c__tid != mP->tid) {
	  /* sanity -- reply already came in */
	  mutex_unlock(&(cptP->m));
	  NEXTPACKET;
	} else {
	  cptP->retries &= M_SCANCLEAR; /* none of that right now */
	}
      mutex_unlock(&(cptP->m));
      /* need the following if we update transaction id below?  if it
	 came in with the new tid, it must be new, right?  incr'ing tid
	 will throw stuff away earlier, but we might also throw away better
	 information that comes later due to one of our retransmissions.
	 Siiigghhh.  No incr'ing tid for now. 21sep92 mjz */
      rc = midLockNewOwn(mP->lock, mP->source, mP->id, mP->tid,
			 bufp, (u_long) buflen);

      if (rc == M_R_DOAGAIN) {	/* ok, let's do it again */
	mutex_lock(&(cptP->m)); /* re acquire mutex */
          if (cptP->c__tid != mP->tid) {
	    /* sanity -- reply already came in */
	    mutex_unlock(&(cptP->m));
	    NEXTPACKET;
	  } else {
	    cptP->dest = mP->source; /* retarget */
/*	    cptP->c__tid = LocalClock; /X* assign new transaction id XXX*/
	    cptP->c__mflags &= ~M_RETRANS; /* turn off retrans bit now */
	    CommSend(thread);  /* do this (not) under lock??? mjz */
	    cptP->retries = M_SCANNABLE; /* reset */
	  }
	mutex_unlock(&(cptP->m));
	signal_scanner();
      } /* midLockNewOwn says to do it again */
      msg_type &= ~(M_REPLY_BIT);	/* turn off, so no thread released at end */
      break;

    case M_BARRIER_OP:
      if (IS_DBG_COMM_WORK)
	StatusMsg("about to midBarrierRequest");
      /* crufty handling of multiple packet request */
      tt = TRANTAB(mP->rhost, mP->rid); /* get trantab handle */
      ASSERT( tt->signature == TRANTAB_SIG );
      if (((mP->pacnum == 0) && (mP->flags & M_LASTPACK))) {
	/* only */
	mP->flags |= M_FINAL;	/* HACK.  Ensure final too */
      } 
      else if ((mP->flags & M_FINAL)) { /* final, but not only */
	mutex_lock(trantab_lock);
	while ((tt->u.state != M_R_INPROCESS) || /*must be if final? */
	       (tt->inproc_max != tt->inproc)) {
	  condition_wait(trantab_c, trantab_lock); /* BARF only one CV? */
	}
	mutex_unlock(trantab_lock);
      }

      rc = midBarrierRequest(mP->lock,mP->rhost,mP->rid,mP->tid,
			(mP->flags&(M_RETRANS|M_DOAGAIN|M_FORWARDED|M_FINAL)),
			bufp, (u_long) buflen);

      
      /* do SIMPLE_TRAN_DONE, and then some.  Not so simple, anymore */
      /* ensured final was final.  sux to delay here.  Would like barrier
	 to do it, but then it must keep counts from everyone.
	 That might be the right way to go.  do it here for now, move
	 later.  maybe tomorrow if sux.
	 */
      mutex_lock(trantab_lock);
      ASSERT( tt->signature == TRANTAB_SIG );
      /* assuming tranid == tid.  maybe print for now too */
      if (mP->flags & M_FINAL) { /* know all others finished before us */
	if (tt->tranid == mP->tid) /* could have changed if another req arvd */
	  tt->u.state = rc; /* M_R_DONE or M_R_QUEUED */
      } else {			/* not final, perhaps someone waiting */
	tt->inproc++;
	if ((tt->u.state == M_R_INPROCESS) && (tt->inproc_max == tt->inproc)) {
	  /* we were next to last */
	  condition_broadcast(trantab_c); /* may wake up tons 'o folks */
	}
      }
      mutex_unlock(trantab_lock);

      if (IS_DBG_COMM_WORK)
	StatusMsg("done with midBarrierRequest");
      break;

    case M_BARRIER_OP | M_REPLY_BIT:
      if (IS_DBG_COMM_WORK) {
	StatusMsg("about to midBarrierReply");
	sprintf(smsg,"writers=%lu allpackets=%d \n",
	    cptP->writers, midAllpackets(thread));
	StatusMsg(smsg);
      }

      midBarrierReply(mP->lock,mP->source,mP->id,mP->tid,bufp,(u_long)buflen);

      if (IS_DBG_COMM_WORK)
	StatusMsg("done midBarrierReply");
      break;    /* requesting thread is released below */

    case M_ALLOCATE:
      if (IS_DBG_COMM_WORK)
	fprintf(stderr, "about to midAllocateRequest\n");/**/

      midAllocateRequest(mP->lock,mP->rhost,mP->rid,mP->tid,
			 (mP->flags&(M_RETRANS|M_DOAGAIN|M_FORWARDED)));

      SIMPLE_TRAN_DONE;
      if (IS_DBG_COMM_WORK)
	fprintf(stderr, "done with midAllocateRequest\n");/**/
      break;

    case M_ALLOCATE | M_REPLY_BIT:
      if (IS_DBG_COMM_WORK)
	fprintf(stderr, 
		"about to midAllocateReply, writers=%lu, allpackets=%lu\n",
		cptP->writers, midAllpackets(thread));/**/

      midAllocateReply(mP->lock,mP->source,mP->id,mP->tid,bufp,(u_long)buflen);

      if (IS_DBG_COMM_WORK)
	fprintf(stderr,"done midAllocateReply\n");
      break;    /* requesting thread is released below */

    case M_ALLOC_INFO:
      if (IS_DBG_COMM_WORK)
	fprintf(stderr, "about to midAllocInfoRequest\n");/**/

      midAllocInfoRequest(mP->lock,mP->rhost,mP->rid,mP->tid,
			  (mP->flags&(M_RETRANS|M_DOAGAIN|M_FORWARDED)));

      SIMPLE_TRAN_DONE;
      if (IS_DBG_COMM_WORK)
	fprintf(stderr, "done with midAllocInfoRequest\n");/**/
      break;

    case M_ALLOC_INFO | M_REPLY_BIT:
      if (IS_DBG_COMM_WORK)
	fprintf(stderr, 
		"about to midAllocInfoReply, writers=%lu, allpackets=%lu\n",
		cptP->writers, midAllpackets(thread));/**/

      midAllocInfoReply(mP->lock,mP->source,mP->id,mP->tid,bufp,(u_long)buflen);

      if (IS_DBG_COMM_WORK)
	fprintf(stderr,"done midAllocInfoReply\n");
      break;    /* requesting thread is released below */


    case M_PAGE_DATA:
      if (IS_DBG_COMM_WORK)
	fprintf(stderr, "about to mid_page_request\n");/**/

      mid_page_request (mP->lock, mP->rhost, mP->rid, mP->tid,
			bufp, (u_long) buflen);

      SIMPLE_TRAN_DONE;
      if (IS_DBG_COMM_WORK)
	fprintf(stderr, "done with mid_page_request\n");/**/
      break;
    

    case M_PAGE_DATA | M_REPLY_BIT:
      if (IS_DBG_COMM_WORK) {
	StatusMsg("about to mid_page_reply\n");
	sprintf(smsg,"writers=%lu allpackets=%d Required==Received=%d",
		cptP->writers, midAllpackets(thread),
		midCompareBit(cptP->acksRequired,
			      cptP->acksReceived, Comm_hnum));
	StatusMsg(smsg);
      }

      mid_page_reply (mP->lock, mP->rhost, mP->rid, mP->tid,
		      mP->pacnum, bufp, (u_long) buflen);

      if (IS_DBG_COMM_WORK)
	StatusMsg("done mid_page_reply\n");
      break;    /* requesting thread is released below */


    case M_PAGE_UPDATE:
      if (IS_DBG_COMM_WORK)
	fprintf(stderr, "about to mid_updateq_request\n");/**/

      mid_updateq_request ( mP->rhost, mP->rid, mP->tid,
			   bufp, (u_long) buflen);

      SIMPLE_TRAN_DONE;

      if (IS_DBG_COMM_WORK)
	fprintf(stderr, "done with mid_updateq_request\n");/**/
      break;

    case M_PAGE_UPDATE | M_REPLY_BIT:
      if (IS_DBG_COMM_WORK) {
	StatusMsg("about to mid_updateq_reply\n");
	sprintf(smsg,"writers=%lu allpackets=%d Required==Received=%d",
		cptP->writers, midAllpackets(thread),
		midCompareBit(cptP->acksRequired,
			      cptP->acksReceived, Comm_hnum));
	StatusMsg(smsg);
      }

      mid_updateq_reply (mP->rhost, mP->rid, mP->tid,
			 bufp, (u_long) buflen);

      if (IS_DBG_COMM_WORK)
	StatusMsg("done mid_updateq_reply\n");
      break;    /* requesting thread is released below */


    case M_PAGE_COPYSET:
      if (IS_DBG_COMM_WORK)
	fprintf(stderr, "about to mid_copyset_request\n");/**/

      mid_copyset_request (mP->lock, mP->rhost, mP->rid, mP->tid,
			   bufp, (u_long) buflen);

      SIMPLE_TRAN_DONE;

      if (IS_DBG_COMM_WORK)
	fprintf(stderr, "done with mid_copyset_request\n");/**/
      break;


    case M_PAGE_COPYSET | M_REPLY_BIT:
      if (IS_DBG_COMM_WORK) {
	StatusMsg("about to mid_copyset_reply\n");
	sprintf(smsg,"writers=%lu allpackets=%d Required==Received=%d",
		cptP->writers, midAllpackets(thread),
		midCompareBit(cptP->acksRequired,
			      cptP->acksReceived, Comm_hnum));
	StatusMsg(smsg);
      }

      mid_copyset_reply (mP->lock, mP->rhost, mP->rid, mP->tid,
			 bufp, (u_long) buflen);

      if (IS_DBG_COMM_WORK)
	StatusMsg("done mid_copyset_reply\n");
      break;    /* requesting thread is released below */


    case M_SEPPUKU:
      if (IS_DBG_STAT_COMM) {
	mid_comm_print_stats();
      }
      if (IS_DBG_STAT_LOCK) {
	mid_lock_print_stats();
      }
      if (IS_DBG_STAT_PAGEMGR) {
	mid_page_print_stats ();
      }
      if (IS_DBG_STAT_UPDATEQ) {
	mid_updateq_print_stats ();
      }
      if (IS_DBG_STAT_XPAGER) {
	mid_xpager_print_stats ();
      }
      if (IS_DBG_STAT_SHMALLOC) {
	mid_shmalloc_print_stats ();
      }

      StatusMsg("Calling exit(0)");
      midway_exit(0);		/* should terminate the whole thing */
      break;

    default:
      sprintf(smsg,"ERROR - CommWorker: unknown type field in message: %8lx",
	      msg_type);
      StatusMsg(smsg);
      msg_type = 0;    /* reset message type to prevent release below */
      /* better only do these for requests */
      if (!(msg_type & M_REPLY_BIT)) {
	  tt = TRANTAB(mP->rhost, mP->rid);
	  ASSERT( tt->signature == TRANTAB_SIG );
	  mutex_lock(trantab_lock);
	  if (tt->tranid == mP->tid) {
	      tt->u.state = 0; /* funky state */
	  }
	  mutex_unlock(trantab_lock);
      }
      break;
    }/* switch(type) */


    /* 
     * If this was a reply packet, then complete the waiting request 
     */
    if (msg_type & M_REPLY_BIT) {
      mutex_lock(&(cptP->m));

#if FLOW_TIMES
      fti = cptP->pflow;
      if (mP->pacnum <= FT_MAXPAKS) {
	ftp = /*&(ft_times[fti][FT_REPPAKS+(2*mP->pacnum)]);*/
	  FT_MSGINFO(FT_REPPAKS,mP->pacnum);
	/* try to filter out bogus values.. ?*/
	*ftp++ = ft_rcv_time_start; /* to try to get actual rcv time */
	*ftp++ = (u_int) mP->FT_STASH_INMSG; /* sashed receive time */
	*ftp++ = ft_rcv_time_copy; /* XXX Global */
	*ftp++ = ft_rcv_time_rel;
	*ftp++ = ft_startop;	/* time operation itself */
	*ftp++ = ft_stopop;
	/* *ftp = FT_READ_TIMER;*/
      }
      if ((mP->flags & M_LASTPACK) || (mP->pacnum == FT_MAXPAKS))
	ft_times[fti][FT_REPPNUM] = mP->pacnum;
#endif
      /* see if ACK is required... */
      if (cptP->ackpnum &&
	  (cptP->writers <= (midway_comm_reply_ackafter / 2))) {
	/* if we don't wait until ackafter/2 are processed, then if
	   every incoming packet is faulting (as might happen if
	   EC memory writes over non-EC memory (say your default
	   is RC)), then we could get two (or more) windows worth
	   of packets in process, and tie up all the workers.
	 */
	CommAck(mP->source, cptP->ackthread, mP->tid, mP->lock,
		cptP->ackpnum, 0);
	INCR_STAT (mid_comm_stats.ackWorker3, 1);
	cptP->ackpnum = 0;
	cptP->ackthread = 0;
      }
      if (IS_DBG_COMM_WORK) {
	sprintf(smsg,"CommWorker: release waiter? writers=%lu allpackets=%d Required==Received=%d",
		cptP->writers, midAllpackets(thread),
		midCompareBit(cptP->acksRequired,
			      cptP->acksReceived, Comm_hnum));
	StatusMsg(smsg);
      }
      ASSERT(cptP->writers > 0);
      if ((--(cptP->writers) == 0) && /* no other writers */
	  (midAllpackets(thread)) &&
	  (midCompareBit(cptP->acksRequired,cptP->acksReceived,
			  Comm_hnum))) { 
	if (IS_DBG_COMM_WORK)
	  fprintf(stderr,"CommWorker: about to release thread %lu\n",thread);

	/* if we had to invalidate self, it may be that scanner still active */
	cptP->retries &= M_SCANCLEAR;

#if FLOW_TIMES
	ft_times[fti][FT_REPSIGNAL] = FT_READ_TIMER;
#endif
	condition_signal(&(cptP->c)); /* all done, let waiter go */
	if (msg_type == (M_HOST_OP | M_REPLY_BIT)) {
	  lasthostop = 1;	/* note last host op */
	}
      }
      mutex_unlock(&(cptP->m));
    } /* is a reply */
#if (BENCHMARK == 0)
  else {			/* therefore is a request */
      tt = TRANTAB(mP->rhost, mP->rid);
      ASSERT( tt->signature == TRANTAB_SIG );
  }
#endif

 nextpacket:
#if (BENCHMARK == 0)
  tt = TRANTAB(1,2);
  ASSERT( tt->signature == TRANTAB_SIG );
#endif
  return(lasthostop);
} /* CommProcessPacket */
/*
 ************************************************************************
 *			C o m m W o r k e r				*
 *----------------------------------------------------------------------*
 * N concurrent threads of this procedure will loop forever		*
 * receiving messages from the network.  It will perform any ack	*
 * processing that was requested.  It calls an appropriate routine	*
 * based on incoming message type...					*
 *									*
 *    HOST_OP|REPLY_BIT -> midCommHostReply				*
 *    LOCK_OP -> midLockRequest						*
 *    LOCK_OP|REPLY_BIT -> midLockReply (to suck data, hints)		*
 *									*
 * If this is a complete reply, and all InvalidateAcks have been	*
 * received, we will signal the thread to continue.			*
 ************************************************************************
 */
void
CommWorker(workernum)
     u_long workernum;		/* my number, from 0-(n-1) */
{
  pbuf          pbufP;          /* ptr to buffer for this thread */
  u_long	buflen;		/* data in buf, actually */
  int		rcverrs;	/* local receive errors */
  char255	smsg;		/* for errors */
  int		i,j;
#if ATM_IFACE && WIRE_DA_THREADS
  static int	midway_set_good_priority(); /* forward */
#endif
  int		lasthostop;	/* see if the last host op msg received */

  (void) cpt_init();		/* We may use, and lock layer needs :-( */

#if WIRE_DA_THREADS
  /* we are cthread_fork'd; for now wire all cthreads with 3.0 threads*/
  cthread_wire();
#endif

#if ATM_IFACE && WIRE_DA_THREADS
  midway_set_good_priority();
#endif

#if SET_PRIORITY
  /*
   * "new" interface.  Don't use with old ATM_IFACE.  Allow priorities
   * of most everything to be set via global vars...
   */
  if (midway_comm_worker_prio) {
    midway_set_good_priority(midway_comm_worker_prio);
  }
#endif /* SET_PRIORITY */


  /*
   * Loop forever handling incomming messages
   */
  for(rcverrs=0;rcverrs<10;) {	/* *MAGIC* put 10 somewhere else *mjz* */
    /* FUTURE: could listeners++ before, listeners-- after, and if listeners==0,
       crank new thread; someone always listens...  perhaps dealloc'ing
       thread when done.   Hrrumph. *mjz* */
    if (CommRecv(&pbufP, &buflen)) {
      sprintf(smsg,"CommRecv returned with an error");
      StatusMsg(smsg);
      rcverrs++;
      /* sleep? */
      continue;			/* loop de loop */
    }

    lasthostop = CommProcessPacket(pbufP, buflen);

    midway_net_fns->free(pbufP);
    
    if (lasthostop) {
      /* others will be spawned with new network protocol */
      cthread_exit((void *)0);
    }
  } /* (almost) infinite main loop */

  /* if here, then rcverrs exceeded threshold */
  StatusMsg("CommWorker: rcverrs exceeded threshold, dying.");

  cthread_exit((void *)100);
} /* CommWorker */

/*
 ************************************************************************
 *		    C o m m S e t T r a n S t a t e			*
 *----------------------------------------------------------------------*
 * Called by the lock layer to reset the state of an outstanding trans  *
 * when one is dequeued.  So resets queued state to (probably) forwarded*
 ************************************************************************
 */
CommSetTranState(host, thread, tranid, state)
     u_long host;
     u_long thread;
     u_long tranid;
     comm_request_t state;
{
  trantab_t * tt;
  char255 smsg;

  tt = TRANTAB(host,thread);
  ASSERT( tt->signature == TRANTAB_SIG );

  mutex_lock(trantab_lock);

  if (tt->tranid != tranid) {
    comm_request_t st = tt->u.state;
    u_long tid = tt->tranid;
    mutex_unlock(trantab_lock);
    if (IS_DBG_COMM_WNGS) {
      sprintf(smsg, "CommSetTranState h %lu t %lu tid %lu newstate 0x%lx: old tid did not match. tid %lu state 0x%lx",
	      host, thread, tranid, state, tid, st);
      StatusMsg(smsg);
    }
    INCR_STAT(mid_comm_stats.useless,1); /* not sure if this is needed here..*/
    return;
  }
  if (tt->u.state != M_R_QUEUED){
    comm_request_t st = tt->u.state;
    mutex_unlock(trantab_lock);
    if (IS_DBG_COMM_WNGS) {	/* XXX -- have only seen 0->0, tho */
      sprintf(smsg, "CommSetTranState h %lu t %lu tid %lu newstate 0x%lx: old state not queued. state %lx",
	      host, thread, tranid, state, st);
      StatusMsg(smsg);
    }
    return;
  }

  tt->u.state = state;

  mutex_unlock(trantab_lock);
} /* CommSetTranState */
/**** for debugging ****/
trantab_t *
CommTrantab(host, thread)
     u_long host;
     u_long thread;
{
  return(TRANTAB(host,thread));
} /* CommTrantab */

cthread_t
x_cthread_self()
{
  return(cthread_self());
} /* x_cthread_self */

cthread_t
x_cthread_who(stp)
     int stp;			/* a pointer on the stack */
{
  return((ur_cthread_ptr(stp))->incarnation);
} /* x_cthread_who */

void
x_cthread_offsets()
{
#ifdef STACK_GROWTH_UP
  char *sg = "up";
#else
  char *sg = "down";
#endif
  struct ur_cthread uct;

  printf("cthread_stack_mask 0x%lx growth %s [up:and dn:or+1-sizeof(ur_cthread_t *) (%d) offset %d\n", 
	 cthread_stack_mask, sg,
	 sizeof(ur_cthread_t *),
	 (unsigned long)(&(uct.incarnation)) - (unsigned long)(&(uct)));
} /* x_cthread_offsets */


/*
 ************************************************************************
 *			  C o m m R e p l y				*
 *----------------------------------------------------------------------*
 * Called to reply to a request being processed.			*
 ************************************************************************
 */
#if ATM_IFACE
/* return 0: not us, 1: us, and we expect more; -1: us, and we're done */
CommReplyFilter(char *msg, u_long len, any_t opaque)
{
  typedef struct {
    COMMMSG	mh;	/* most of an ACK */
    u_long	pkts;	/* used only for NACK */
  } ackmsg_t;
  ackmsg_t *am;
  u_long cpti;
  cpt_t *cptP;
  char255 smsg;

  INCR_STAT (mid_comm_stats.replfilt, 1);

  if (len != sizeof(ackmsg_t)) {
    /* StatusMsg("failed length test");*/ /* do not do this here! */
    return(0);
  }

  am = (ackmsg_t *) msg;
  cpti = (u_long) opaque;
  cptP = &cpt[cpti];

  if ((am->mh.version == M_PROTOVER) &&
      (am->mh.incarnation == Comm_incarn) &&
      (am->mh.type == M_ACK) &&
      (am->mh.rid == cpti) &&	/* destination id */
      (am->mh.tid == cptP->c__tid) &&
      (am->mh.pacnum == cptP->c__pnum)
      )
    {
      /* Filters need to always update the Lamport clock! */
      if (mutex_try_lock(ClockLock) == 0) {
	mid_comm_stats.repflmiss++;
	return(0);
      }
      if (GREATER_THAN_CLOCK (am->mh.clock, LocalClock))
	LocalClock = ADJUST_CLOCK (am->mh.clock, 1);
      else
	LocalClock = ADJUST_CLOCK (LocalClock, 1);
      mutex_unlock(ClockLock);

      cptP->retries &= M_WAITCLEAR; /* got it! */
      /* could protect update, but just hope for best */
      INCR_STAT (mid_comm_stats.fastreply, 1);
      return(-1);
    }
  else {
    /*
    sprintf(smsg,"crf: packet not for us received, type 0x%x shst %d sthrd %d",
	    am->mh.type, am->mh.source, am->mh.id);
    StatusMsg(smsg);
    */
    return(0);
  }
} /* CommReplyFilter */
#endif
CommReply
    (
     u_long	type,		/* kind of message */
     u_long	host,		/* index into host table */
     u_long	thread,		/* thread@host reply is destined for */
     u_long	tid,		/* tid for this transaction */
     u_long	lock,		/* ??? (u_long) */
     int	last,		/* non zero if last packet */
     u_long	pnum,		/* this packet number (?) */
     addr_vector_t nacks,	/* bitarray of acksRequired */
     char	*msg,		/* type dependent message info */
     u_long	msglen,		/* size of message */
     int        freepack 	/* =C_FREE to free pbuf pointed to by msg */
     )
{
  u_long	cpti;		/* client-per-thread index */
  cpt_t 	*cptP;
  int		mustScan = 0;
  char255	smsg;

  ASSERT (initdone);

  /* see if client-per-thread entry exists yet */
  cpti = (u_long)cthread_data(cthread_self());
  if (!cpti) {			/* not yet allocated */
    cpti = cpt_init();
  }
  cptP = &cpt[cpti];

  /* wait for an ack that should have arrived for the last packet */
  if ((pnum>1) && (((pnum+1) % midway_comm_reply_ackafter) == 1)) {
    trantab_t *tt;
    int gotack = 0;		/* 1=>ack arrived already, -1=>error */
    u_long state = 0;
    u_long stid = 0;

    mutex_lock(&(cptP->m));
    if (cptP->retries & M_WAITING) {
#if ATM_IFACE
      /* try to get reply without context switching */
      midway_atm_wantreply(CommReplyFilter, (any_t) cpti);
      if (cptP->retries & M_WAITING) {
#endif /*ATM_IFACE*/

	/* also note in TRANTAB (we might even have been acked already) */
	tt  = TRANTAB(host, thread);
	ASSERT( tt->signature == TRANTAB_SIG );
	mutex_lock(trantab_lock);
	if (tt->tranid != tid) {	/* insanity */
	  state = tt->u.state;
	  stid = tt->tranid;
	  gotack = -2;	/* well, may be that new request came in on retrans */
	} else {			/* potentially sane */
	  if (tt->u.state == M_R_INPROCESS) {
	    tt->u.state = M_R_WAITING;
	    tt->waiter = cpti;
	  } else if (tt->u.state == M_R_ACKED) {
	    if (tt->waiter == (pnum-1)) {	/* packet arrived! */
	      gotack = 1;
	      tt->u.state = M_R_INPROCESS; /* revert back to in process */
	      tt->waiter = 0;
	    }
	    else {
	      /* can't have received > ack, since haven't sent.  wait */
	      tt->u.state = M_R_WAITING;
	      tt->waiter = cpti;
	    }
	  } else {			/* state not inprocess or acked */
	    state = tt->u.state;
	    stid = tt->tranid;
	    gotack = -1;
	    cptP->retries &= M_WAITCLEAR; /* undo */
	  }
	} /* potentially sane */
	mutex_unlock(trantab_lock);
	
	if (gotack == 0) {		/* no ack yet, must wait */
	  cptP->retries |= M_SCANNABLE;	/* see CommRecv comment */
	  signal_scanner();
	  
	  while (cptP->retries & M_WAITING)
	    condition_wait(&(cptP->c), &(cptP->m));
	  
	  cptP->retries &= (M_SCANCLEAR & M_RETRANSCLR); /* reset */
	}
	
	if (gotack < 0) {
	  if ((gotack == -1)||(IS_DBG_COMM_WNGS)) {	/* error will robinson */
	    sprintf(smsg, "CommReply: going to wait after packet %lu tran %lu, but state was bad 0x%lx tran %lu; continuing.", pnum, tid, state, stid);
	    StatusMsg(smsg);
	  }
	}
#if ATM_IFACE
      } /* waiting after wantreply */
#endif
    } /* still have waiting set */
    else {
#if ATM_IFACE
      midway_atm_clear_filter(CommReplyFilter,cpti);
      mid_comm_stats.repfsetint++;
#endif
    }

    if ((cptP->retries & M_WAITFREE) && (cptP->mlen != 0)) {
      /* gotta free the baby */
      cptP->retries &= M_WAITFREECLR;
      mid_pbuf_free(cptP->msg);
      cptP->msg = 0;
      cptP->mlen = 0;
    }
    mutex_unlock(&(cptP->m));

  } /* new ackafter support */

  /* cpt exists, so fill it in, and then send message & sleep */
  cptP->dest = host;
  cptP->c__tid = tid;	/* used as tid for CommSend */
  cptP->c__optype = type | M_REPLY_BIT;
  cptP->msg = msg;
  cptP->mlen = msglen;
  if (last)
    cptP->c__mflags = M_LASTPACK; /* this is last/only packet for reply */
  else
    cptP->c__mflags = 0;
  cptP->c__pnum = pnum; /* rats.  could all be done here now? (at least with
			    this strategy) */
  cptP->c__rhost = 0;
  cptP->c__rid = thread;
  cptP->c__lock = lock;
  cptP->c__nacks = nacks;
  cptP->retries = 0;
  cptP->multicast = 0;

/* cpt wrong?  Don't want to wait for reply; want to resend until next request,
   right?  or will get request again if no reply?  but what about multipacket
   then... will want to just resend last packet... 
   If ack packets, we're ok; but this is extra traffic; not in critical
   path, though....

*/

  CommSend(cpti);

  /* new ackafter support */
  if ((last == 0) && 
      (((pnum+1) % midway_comm_reply_ackafter) == 0)) { /* wait for ack */
    trantab_t *tt;
    int gotack = 0;		/* 1=>ack arrived already, -1=>error */
    u_long state = 0;
    u_long stid = 0;

    /*
     * we used to wait for an ack here.  However, it would be better if
     * we could just note that an ack might happen, and then process
     * the next packet (a little double buffering).  {Perhaps I could
     * throw away the trantab stuff?}
     */

    mutex_lock(&(cptP->m));
    cptP->retries |= M_WAITING; /* note waiting */
#if ATM_IFACE
    /* try to get reply without context switching */
    midway_atm_set_filter(CommReplyFilter, (any_t) cpti);
    mid_comm_stats.repfset++;
#endif /*ATM_IFACE*/

    if ((freepack == C_FREE) && (cptP->mlen != 0)) {
      cptP->retries |= M_WAITFREE;
    }
    mutex_unlock(&(cptP->m));
  } /* new ackafter support */
  else { /* as before... but now we will not free packet if expecting ack */
    /* free packet */
    /* no, do when ack comes in? (but will get none, and already waited...) */
    if ((freepack == C_FREE) && (cptP->mlen != 0)) {
      mid_pbuf_free(msg);
      cptP->mlen = 0;		/* so never use again */
      cptP->msg = 0;		/* ditto, man */
    }
  }

  return(0);
} /* CommReply */

/*
 ************************************************************************
 *		    C o m m A w a i t i n g A c k			*
 *----------------------------------------------------------------------*
 * Returns 1 if specified thread has received all it's data packets,    *
 * and still is awaiting an invalidate acknowlegement from the specified*
 * host.  This must be done here, because if the lock is being		*
 * transferred to us, only the comm layer knows what invalidate acks    *
 * are outstanding...	Called by the lock layer.			*
 ************************************************************************
 */
int
CommAwaitingAck( cpti, host )
     u_long cpti;		/* cpti gotten from OUR_TID */
     u_long host;		/* processor/host in question */
{
  cpt_t *cptP;
  int ans = 0;
  char255 smsg;

  if ((cpti == 0) || (cpti >= cptnext)) {
    sprintf(smsg,"CommAwaitingAck: given thread id (%lu) is not in the range 1 to %lu",
	    cpti, cptnext);
    StatusMsg(smsg);
    return(0);
  }
  if ((host == 0) || (host > Comm_hnum)) {
    sprintf(smsg,"CommAwaitingAck: given host id (%lu) is not in range 1 to %lu",
	    host, Comm_hnum);
    StatusMsg(smsg);
    return(0);
  }
  cptP = &cpt[cpti];

  mutex_lock(&(cptP->m));

  if (midAllpackets(cpti) &&	/* all data packets have arrived */
      cptP->writers == 0 && /* ..and have been processed */
      midTestBit(cptP->acksRequired, host) && /* and we still */
      !(midTestBit(cptP->acksReceived, host))) /* need it */
    ans = 1;

  mutex_unlock(&(cptP->m));

  return(ans);
} /* CommAwaitingAck */

/* One if all packets received, else 0.					*/
int
midAllpackets(tid)
     u_long tid;		/* thread */
{
  u_long pnum;

  if (!(cpt[tid].retries & M_ALLPACKETS)) /* last packet received */
    return(0);

  pnum = cpt[tid].maxlinear >> 1; /* last packet received in sequence */

/*
fprintf(stderr,"Allpackets: ALLPACKETS set, pnum=%lu lastpacket=%lu\n",
	pnum, cpt[tid].lastpacket);fflush(stderr);
*/
  if (pnum == cpt[tid].lastpacket+1) /* +1 'cause pnum was NEXT to receive */
    return(1);			/* got all in order! */

  if (midPlistCheck(&(cpt[tid].plist), cpt[tid].lastpacket-pnum))
    return(1);			/* all got here anyway */

  return(0);			/* not all packets arrived */
} /* midAllpackets */
/*
 ************************************************************************
 *		p a c k e t   l i s t   f u n c t i o n s		*
 *----------------------------------------------------------------------*
 * Used to operate on packetlists, which are intended to be used to	*
 * tell which packets out of a string have arrived.  Words are alloc-	*
 * cated in blocks (of size _PNUM) to reduce storage overhead due to    *
 * pointers.  These routines allow an (essentially) unlimited number    *
 * of packets to be unacknowledged and in flight.			*
 * These are probably obsolete if we stick with 32 packets max in flight*
 * scheme (5 mar 91 - mjz).						*
 *									*
 * midPlistSet(plist_t *pl, u_long bit) -sets bit in pl, extending the  *
 *				packet list if it is not long enough.   *
 * midPlistCheck(plist_t *pl, u_long lastbit) - ensure that all bits    *
 *				up to lastbit are on.  To be used to    *
 *				see if all packets arrived safely.	*
 * midPlistClear(plist_t *pl) - ensure that all bits are off in a plist *
 ************************************************************************
 */
midPlistSet( pl, bit )
     plist_t	*pl;
     u_long	bit;
{
  u_long	count;		/*  */

  /* ensure in "right" packetlist element.  Create if necessary */
  for (count = bit; count >= 32*_PNUM; count -= 32*_PNUM) {
    /* if here, then need to check next list elment.  Create if
       not available. */
    if (pl->pacnext)
      pl = pl->pacnext;
    else {
      pl->pacnext = (plist_t *) MALLOC(sizeof(plist_t));
      /* if zero, barf *MJZ* */
      bzero((char *)pl->pacnext, sizeof(plist_t));
      pl = pl->pacnext;
    }
  }
  /* now in proper element... set proper bit */
  midSetBit(pl->pacvec, count);
  return;
} /* midPlistSet */
midPlistClearBit( pl, bit )
     plist_t	*pl;
     u_long	bit;
{
  u_long	count;		/*  */

  /* ensure in "right" packetlist element.  Create if necessary */
  for (count = bit; count >= 32*_PNUM; count -= 32*_PNUM) {
    /* if here, then need to check next list elment.  Create if
       not available. */
    if (pl->pacnext)
      pl = pl->pacnext;
    else {
      /* no packetlist element ==> already clear.  quit */
      return(0);
    }
  }
  /* now in proper element... clear proper bit */
  midClearBit(pl->pacvec, count);
  return(0);
} /* midPlistClearBit */
midPlistTest( pl, bit )
     plist_t	*pl;
     u_long	bit;
{
  u_long	count;		/*  */

  /* ensure in "right" packetlist element. */
  for (count = bit; count >= 32*_PNUM; count -= 32*_PNUM) {
    /* if here, then need to check next list elment.  */
    if (pl->pacnext)
      pl = pl->pacnext;
    else {
      /* no next, hence bit cannot be set */
      return(0);
    }
  }
  /* now in proper element... test proper bit */
  return( midTestBit(pl->pacvec, count) );
} /* midPlistTest */
midPlistCheck(pl, lastbit)
     plist_t	*pl;
     u_long	lastbit;
{
  int	i;

  /* ensure all bits up to lastbit are on */
  do {
    /* check this segment of _PNUM */
    for (i=0; i < _PNUM; i++) {
      if (lastbit > 31) {
	if (pl->pacvec[i] != 0xffffffff) {
	  return(0);		/* no match */
	}
	lastbit -= 32;		/* proceed to next chunk */
      } else {			/* last chunk */
	if (pl->pacvec[i] != (0xffffffff<<(31-lastbit))) {
	  return(0);		/* blew it at the end */
	}
	lastbit = 0;		/* note done */
	break; /* for loop */
      }
    } /* end for loop*/
    if (lastbit)		/* more to do */
      pl = pl->pacnext;		/* go to next packlist element */
  } while (lastbit > 0);
  return(1);			/* all OK */
} /* midPlistCheck */
void
midPlistClear(pl)
     plist_t *pl;
{
  int i;

  while(pl) {			/* for each packet list segment */
    /* could unroll following if _PNUM was *really* fixed... */
#if _PNUM == 4
    pl->pacvec[0] = 0;
    pl->pacvec[1] = 0;          /* unroll since _PNUM is defined */
    pl->pacvec[2] = 0;
    pl->pacvec[3] = 0;
#else
    for(i=0; i<_PNUM; i++)
      pl->pacvec[i] = 0;		/* clear the bits */
#endif
    pl = pl->pacnext;		/* and try next segment */
  }
  return;
} /* midPlistClear */

/* -- the way it was
   #ifdef FAST_UDP
   #ifdef FAST_PRIO
*/
#if ATM_IFACE && WIRE_DA_THREADS
/*
 ************************************************************************
 *		m i d w a y _ s e t _ g o o d _ p r i o r i t y		*
 *----------------------------------------------------------------------*
 * I would like this to be in a separate file, but by placing it here   *
 * I don't have to worry about it being a global symbol.  		*
 *									*
 * Reset max priority of this task downward (range 0-31, smaller better)*
 * and this thread in particular to a very nice value.			*
 *									*
 * This task must be running in group kmem or uid 0 to use task_by_pid. *
 *									*
 * We are trying to prevent receive overruns when sucking data directly *
 * out of the packet filter in the Mach kernel.				*
 ************************************************************************
 */
static int			prio_init = 0;	/* initialized? */
static mach_port_t 		host_priv;	/* priv host port */
static processor_set_name_t 	assigned_set;	/* proc set we are asgned to */
static processor_set_t		proc_set;	/* ctl port for our proc set */
static struct mutex prio_mutex = MUTEX_INITIALIZER; /* cheat! */
static mutex_t pmu = &prio_mutex;

/* with ATM: must be lower than interrupt thread... */
int midway_comm_rcv_prio = 10;	/* global for now, leaving 1 as "maximum". */
int midway_comm_rcv_maxp = 1;	/* global for now, leaving 1 as "maximum". */

#define FAILONERR(s)	if (krc != KERN_SUCCESS) { \
				mach_error(s, krc); \
				return(1);}
#define MFAILONERR(s)	if (krc != KERN_SUCCESS) { \
				prio_init = -1; \
				mach_error(s,krc); \
				mutex_unlock(pmu); \
				return(1);}
static int
midway_set_good_priority()
{
  kern_return_t krc;

  if (prio_init == 0) {
    mutex_lock(pmu);
    MIDWAY_MUTEX_SET_NAME(pmu, "comm.c:prio_mutex");
    if (prio_init == 0) {	/* really not initialized */
      host_priv = task_by_pid(-1); /* back door to get priv host port */
      if (host_priv < 0) {	/* Brian did this.  Not sure if proper. */
	perror("bsd_task_by_pid");
	prio_init = -1;		/* show error */
	mutex_unlock(pmu);
	return(1);
      }
      if (host_priv == 0) {	/* see this empirically */
	StatusMsg("Can't get priv host port [not grp=kmem or uid=0?], will not set receive thread priority.");
	prio_init = -1;
	mutex_unlock(pmu);
	return(1);
      }

      krc = task_get_assignment(mach_task_self(), &assigned_set);
      MFAILONERR("task_get_assignment");

      krc = host_processor_set_priv(host_priv, assigned_set, &proc_set);
      MFAILONERR("host_processor_set_priv");

      prio_init = 1;		/* Done, success */
    }
    mutex_unlock(pmu);
  } /* not initialized */

  if (prio_init < 0)
    return(1);			/* failure */

  /* set max priority down  */
  krc = thread_max_priority(mach_thread_self(), proc_set, 
			    midway_comm_rcv_maxp);
  FAILONERR("thread_max_prio");
  
  /* now we can make our priority better */
  krc = thread_priority(mach_thread_self(), midway_comm_rcv_prio, FALSE);
  FAILONERR("thread_priority");

  return(0);			/* success */
} /* midway_set_good_priority */
#undef FAILONERR
#undef MFAILONERR

#endif /*include midway_set_good_priority */

#if FLOW_TIMES
#include "flow_times.c"
#else
print_flow_times() 		/* null function */
{
  return(0);
}
#endif


/*
 * ANCIENT HISTORY
 *  8 mar 91: punted to comm0
 * 15 Mar 91: back to comm; 
 *  8 jul 91 (mjz): added retrans bit to message on retrans; allow it to be
 * 		passed up to lock request (others?) and down on forward.
 * 31 Jul 91 (mjz): C_FREE/C_NOFREE to CommReply
 * 		Add barrier stuff to CommWorker
 *  3 Dec 91 (mjz): add interface for shmalloc
 * 16 Dec 91 (mjz): add conditional printf's [18-dec-91: move to midway-dbg.h]
 * 31 Jan 92 (mjz): fix bug related to CommMRequest -- we hang forever if
 * 		we are told to wait for replies :-(.
 *  3 Feb 92 (mjz): fix race condition in xxxRequest & Reply -- could get
 * 		answer before we wait and we will hang...
 *  5 Feb 92 (mjz): Ack and Forward were locking scommlock, and then waiting
 * 		using commlock (hence never releasing scommlock, and causing
 * 		the whole system to grind to a halt.  Bletch.
 * 20 Mar 92 (mjz): Add cpt_init to Worker calls.  They will use it if
 * 		they ever ACK, and the "TID" is needed for the lock layer.
 * 		[[It uses it when enqueuing, to note stuff queueud.
 * 		This is perhaps wrong.  It could also use the cthread_self()
 * 		identifier instead...  For now fix here.     XXX
 * 26 May 92 (mjz): update so packets with tids "in process" are squashed.
 *  9 Jun 92 (mjz): fix in process to use originator on requests (vs last
 * 		host sent) and also use the originating thread id.
 * 29 Jun 92 (mjz): un-static ss_port for ss_threat_debug
 *  1 Jul 92 (mjz): fixes to propogate timestamps on invalidate messages
 * 		(need corresponding fixes in lock module)
 *  9 Jul 92 (mjz): add CommAwaitingAck for the lock module
 *  3 Aug 92 (mjz): make worker debug go through ss; fix multi-packet replies
 * 11 Aug 92 (was): added clock_t type
 *                  also put stats under macro control & added byte counters
 * 19 Aug 92 (was): added support for M_PAGE_xxx ops.
 * 26 Aug 92 (was): fixed CommMrequest to handle multi-reply requests
 *  1 Sep 92 (mjz): integrate Brian's new scanner (no 2.5 dependencies)
 *  8 Sep 92 (mjz): replace s_tab & workertid with a more standard array
 *		to detect duplicate packets.
 *  4 Oct 92 (mjz): new incarnation # for each execution
 *  9 Oct 92 (was): converted network calls to macros NET_xxxx
 *                  changed recvmsg to recvfrom (no longer use iovec)
 *                  changed CommRecv to no longer take seperate header buffer
 * 12 Oct 92 (mjz): new ackafter support for many multipacket reply messages
 * 15 Oct 92 (was): changed comm worker to use a pbuf, rather than stack space
 *                  Also moved network_init code for FAST_UDP here.
 *                  and fixed off-by-one bug in midPlist code.
 * 18 Oct 92 (mjz): play w/delay + priorities
 * 20 Oct 92 (was): Added midway_task_self for mach_task_self()
 *                  Added calls to mid_page_print_stats when terminating.
 * 27 Oct 92 (mjz): start ATM support
 * 29 Oct 92 (was): Puts statistics under STATS instead of DEBUG
 *  3 Nov 92 (mjz): more stats (TSTATS, XPAGER stats)
 *  5 Nov 92 (was): Added lock stats
 *  6 Nov 92 (mjz): Add MSTATS (message stats)
 * 22 Mar 93 (was): set retransmission interval based on network type
 *  5 May 93 (mjz): add cache to reduce perthread structure use for repeated
 *			create calls.
 * 19 Jul 93 (mjz): sometimes scanner might get inited after first signal.
 */
