/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * lock.c: Lock protocol module for Midway.
 * 
 * This module is closely tied to the communications module (comm.c);
 * comm provides the OS interface and keeps track of hosts and how
 * to get to them.  Lock provides the locking protocol, but they both
 * need access to per-thread data structures and the lock data structure;
 * in order to free a lock, all data must be transferred (comm will ensure);
 * and the AcksReceived set (from InvalidateAck's) must equal the
 * AcksRequired set (from LockReply and InvalidateAck).
 *
 * This module must supply a routine to take care of intermediate
 * LockReply messages.
 *
 * Includes:
 *  midLockReply() - handle data from LockReply messages
 *  midLockRequest() - handle incoming LockRequest messages
 *  midInvalidate() - handle incoming Invalidate messages
 *
 * HISTORY
 *  27 Jun 91 (mjz): fixed up queuing of remote threads; now release all readers
 * 	do better with ->rdrs variable; init queues!
 *   1 Jul 91 (mjz): save tid when xferring lock -- so we can tell if we
 * 		must redo a lock transfer to person we are aquiring from.
 * 		[[check for invalidates too!]]
 *   8 Jul 91 (mjz): Added retrans stuff with comm layer; added stuff in extReq
 * 		so that we forward retrans' to break n-way loop. [10jul]
 *  31 Jul 91 (mjz): converted lock table to global; add barrier stuff in
 * 		release.  Add C_FREE to CommReply.
 *  xx Dec 91 (mjz): added support for transferring static data along with
 * 		lock info at release time.  Split out mlpd routines...
 * 		added if(IS_DBG_LOCK_xxx) around debug info.
 *   5 Feb 92 (mjz): !(lock->eread) & !(lock->ewrite) used in Release, when
 * 		the !'s should not have been used... sigh.
 *  11 Feb 92 (mjz): attempt better diagnostic for updating dirty addr
 *  19 Mar 92 (mjz): more attempts
 *  20 Mar 92 (mjz): changed external queuer handling.  We now zap on transfer,
 * 		and signal zapping by zapping thread id (TID) too.  See
 * 		comments in midLockXfer and midLockRequest.  Removed
 * 		completely bogus zap of queues in midExtLockRequest.
 * 		Transfer barrier cur_time at creation, but this assumes
 * 		the creator is the manager too. XXX
 *   6 Apr 92 (mjz): add more debugging under LOCK_WNG conditional
 * 		Move LRRETURN out of ELSE clause when forwarding under
 * 		L_ACQ (break circ dependency) case... we were forwarding
 * 		AND queuing, sometime later actually transferring (even
 * 		though target had already acquired the lock)
 *   6 May 92 (mjz): if we ever got our own request back, we would act on it.
 * 			THIS IS WRONG.  Drop the packet, it's probably old.
 * 		Except under Own+RO, we do not clear rdrs.  This is Bad.
 *  19 May 92 (mjz): remove races between condition_{signal,bcast}() in 
 * 		Release(), and wait()'s elsewhere.
 *   1 Jul 92 (mjz): use timestamp on lock in invalidate message -- to match
 * 		exactly invalidates with r/o copies. 
 * 		(need corresponding fixes in comm module)
 * 		Change state bits s.t. ACQ is ACQ RO *ONLY*, so I can
 * 		differentiate between acqrw and acqro+stalled acqrw
 *   3 Jul 92 (mjz): Have midLockDecodeData check dirtybit *before* doing
 * 		bcopy; if we are going to stomp on a newer variable
 * 		(possible in ReleaseReply if statics are transferred)
 * 		DON'T.  This is not an error in this case (but it would
 * 		be for midLockReply).
 *  9 Jul 92 (mjz): keep track of who locally is requesting a write lock
 * 		from somewhere else.  We can then see if an incoming r/o req
 * 		corresponds to a pending invalidate ack.  If so, the r/o req
 * 		must be satisfied, otherwise the invalidate ack will never
 *		come...
 * 13 Jul 92 (mjz): incarnation # test reversed on upgrade test
 * 13 Jul 92 (was): added code to support porting of programs
 * 14 Jul 92 (mjz): C precidence bites again! More debugging on Xfer.
 *		Don't zap (or attempt to transfer) Q'd people on a
 *		retransmitted Xfer.  (NOTE: have lost original people
 *		anyway...)
 * 15 Jul 92 (mjz): use L_DATAIN to tell if INV for this incarn should
 *		be delayed.
 * 1 Aug 92 (mjz): don't bump busy when erdr released during acquire r/o.
 *		Could be problem if invalidated before transfer thread runs...
 *		so kill off queuer if invalidated for now.  [Different if
 *		we hold/acquire r/w; then lock would have to TRANSFER (along
 *		with q'ers (eventually), shooting r/o xfer thread.]
 *		Fix midLockXfer to correctly split long dseg's across packets.
 *		{Warning: must be fixed elsewhere once verified correct.}
 *  4 Aug 92 (mjz): more multi-packet reply fixes
 *		..and large # of locks
 * 11 Aug 92 (was): added clock_t type 
 * 13 Aug 92 (was): changed parms to mid_addrmap_xxx
 * 20 Aug 92 (was): removed PORT code.
 *  9 Sep 92 (mjz): modify for new transaction state in comm layer
 * 10 Sep 92 (mjz): finally transfer queuers too
 *  3 Oct 92 (mjz): change names to midway_ for Acquire, et. al.
 * 13 Oct 92 (mjz): fix return->LRRETURN, missed passing return values!
 * 15 Oct 92 (was): changed MAX_DATA_SIZE to mid_max_data_size
 * 19 Oct 92 (was): Init dataclock value to 1 for locks in RemoteReleaseReply
 * 29 Oct 92 (was): Changed mid_bind_synch_object to reset EC memory to
 *                  reset line_info to its initial value, rather than 0 (dirty).
 *  5 Nov 92 (was): Added stats
 *  5 Nov 92 (mjz): move stats to LSTATS, since these may really munch time.
 *  6 Nov 92 (was): split LSTATS and LTIMES to allow cheap counters to work.
 *  8 Nov 92 (was): Added hint segments to lock replies
 *  1 Dec 92 (mjz): convert to mid_addrmap_delta_pack
 * 19 Dec 92 (was): malloc histogram tables instead of defining them in statics
 * 21 Dec 92 (mjz): add check to avoid delta if ext req is upgrade
 * 13 Jan 93 (was): use macro to check if update queue is empty
 * 13 Mar 93 (was): added verification of user parms
 * 22 Mar 93 (was): fixed bug in computing data transferred in midLockXfer
 * 13 May 93 (was): Added code to allocate dense dirty bits in midway_bind_synch_object
 *  6 Jul 93 (was): eliminated rest of USE_DELTA_MAP code.
 *  8 Jan 94 (was): added fastpath for acquire/release locks (FASTLOCKS)
 *                  locktable moved to a fixed/known address in memory
 * 23 Feb 94 (was): support for page-based version of midway
 *  7 Apr 94 (mjz): Fixed fastlocks interaction with the lock code: a RW acq
 *		actually ended up with internal state like a r/o acq.  Fix the
 *		problem, bulletproof a section, and add an assertion check.
 * 20 May 94 (mjz): Found bug: invalidates sent off with dataclock, not incarn.
 * 23 May 94 (mjz): Added Wayne's tracing code
 *		L_XFER: must not set state to 0 at end, local could have req'd.
 * 14 Jun 94 (was): Call MALLOC for malloc, VM_ALLOCATE for vm_allocate
 * 15 Jun 94 (was): minor change to re-enable lock addrmap sorting
 * 19 Jun 94 (was): Change to page-based lock incarnation support.
 *  4 Sep 94 (was): Added midway_rebind_synch_object.
 * 28 Sep 94 (was): Changed page-based to create a diff for every incarnation
 *                  Diffs clean dirty pages, whereas raw data xfers do not.
 */
 /* #include "midwayos.h" -- someday *//* get other pesky (std) include files */
#include <sys/types.h>		/* u_long & friends */
#include <mach.h>
#include <cthreads.h>
#include <errno.h>
#include <stdio.h>
#define KERNEL 1
#include <sys/time.h>
#undef KERNEL

#include "midway_types.h"       /* global midway type definitions */
#include "lock.h"		/* perlock struct, lock request formats */
#include "comm.h"		/* M_xxx_OP */
#include "smexecd_types.h"	/* ensure get char255 */
#include "pbuf.h"		/* packet buffer stuff */
#include "shmalloc.h"		/* for FIRST/LAST_STATIC_SYM */
#include "vm.h"			/* boundaries on shared memory */
#include "updateq.h"		/* prototypes for calling update queue rtns */
#include "midway-dbg.h"		/* filter debug messages */
#ifndef shared			/* allow us to be compiled with cc */
#define shared
#define priv
#endif				/* we don't do shared here */
#include "midway.h"		/* prototypes */
#include <flow_times.h>



#define MIDWAY_FASTLOCKS  1	/* =0 to disable fastlocks in the library */





/* where our thread id is kept.  Gosh... initialization????? *MJZ* */
#define OUR_TID ((u_long) cthread_data(cthread_self()))

#define NOTHREAD	0x80000000 /* really no thread waiting */
/* Hmmm.. could always toss thread, but then have overhead of create thread
   every time.  Also could toss yourself if r/o and r/w waiter already.
   I could have a pool of other kinds of worker threads, but really
   want only one type.  Problem is, I don't want to have to send
   a UDP message to set one up... */


/*
 * Global references
 */
extern int Comm_hnum;		/* number of hosts participating */
extern int Comm_index;		/* number of this host (for init) */
extern clock_t LocalClock;	/* to date lock structures...  */
extern task_t midway_task_self;	/* cache results of mach_task_self() */
extern u_long cpt_init();	/* for init too... BAD STRUCTURE */
extern char *MALLOC();
extern clock_t *midway_dirty_info_address();

/*
 * Forward declarations
 */
void midExtLockRequest();
void midLockXfer();
static void mid_bind_ec_object ();

/*
 * Globals... were Statics. 31jul91 mjz make static again, or change names
 */
pl_t   *locktable;			/* used for debugging only */
mutex_t lockmutex;
static s_addr_map_t *static_amap;	/* addrmap for statics */

/*
 * Define hint update table
 * Contains id of locks we most recently acquired.
 * Table is sent to other nodes requesting a lock from us.
 */
#define HINT_TABLE_SIZE   64
static pl_t    *hintTable[HINT_TABLE_SIZE];
static int      hintIndex;		/* Index to LRU hint in table */
static mutex_t  hintMutex;

#define UPDATE_HINTS 1		/* Enable hint updating */


/*
 * Define vars used for dense dirty bits
 */
static int midway_use_dense_dirtybits = 0;	/* =1 to enable dense dirty bit code */
extern clock_t    *vm_dense_dirty_base;	        /* base of dense dirty bit allocation */
extern clock_t    *dbmalloc (int size, int align);


#if defined(STATS) || defined(LTIMES)

/*
 * Define free running counter
 */
#define FRC_TICKS_PER_USEC 25	/* 1 tick every 40 nanoseconds */
extern u_int *midway_kern_frc;

/*
 * Define table for collecting statistics 
 */
#define STAT_TIME_PER_BUCKET (1000 * FRC_TICKS_PER_USEC)
#define STAT_MAX_TIME        (4 * 1000000 * FRC_TICKS_PER_USEC)
#define STAT_TIME_TABLE_SIZE (STAT_MAX_TIME / STAT_TIME_PER_BUCKET)

struct {
  u_long	allocated;	/* # of locks allocated */
  u_long	acquires;	/* # of lock acquires */
  u_long        acquireRO;      /* # of read-only lock requests */
  u_long        acquireROacq;   /* # of ROs that already acquiring */
  u_long        acquireROownrw;     /* # of ROs that already own */
  u_long        acquireROownrwbusy; /* # of ROs that already own but busy */
  u_long        acquireROownro;     /* # of ROs that already owned read-only */
  u_long        acquireROlockro;    /* # of ROs that already owned & locked read-only */
  u_long        acquireROholdro;    /* # of ROs that already cached read-only */
  u_long        acquireROreqs;  /* # of ROs that require request */
  u_long	acquireRW;	/* # of read-write acquires */
  u_long	acquireRWincon;	/* # of RW that are inconsistent */
  u_long	acquireRWacqrw;	/* # of RW that already acquiring RW */
  u_long	acquireRWacqro;	/* # of RW that already acquiring RO */
  u_long	acquireRWreqs; 	/* # of RW that require request */
  u_long	acquireRWownrw;     /* # of RW that already own */
  u_long	acquireRWownrwbusy; /* # of RW that already own */
  u_long	acquireRWownro; /* # of RW that already own readonly */
  u_long	acquireRWownroinval;/*# of RW that already own readonly */
  u_long	acquireRWownrordrs; /* # of RW that already own readonly */
  u_long	acquireRWownrobusy; /* # of RW that already own readonly */
  u_long	acquireRWholdro; /* # of RW that already hold RO(upgrade req)*/
  u_long	acquireRWxfer;	/* # of RW that already xfering */

  u_long        xferTotalLocks;  /* total locks transferred */
  u_long        xferTotalLockData;  /* total size of lock's transfered */
  u_long        xferTotalData;   /* actual lock data transferred */
  u_long        xferReleases;    /* total number of remote releases */
  u_long        xferReleaseData; /* total release data xfer (not incl locks) */
  u_long        xferEwrtLost;	 /* external writers lost from queue on xfer */
  u_long	xferEwrQLost;	/* Q'd wrtr lost (q'd acquire, most likely) */
  u_long	xferEwrILost;	/* if Inval req on Q when purged */
  u_long        xferErdrLost;	 /* external readers lost from queue on xfer */
  u_long	xferUpgrade;	/* # RW req that had RO before(EXTERNAL) */

  u_long        extReqRequests; 	 /* total remote lock requests */
  u_long        extReqTimeTotal;	 /* total time-of-day time for requests */
  u_long        extReqTimeMax;	 /* max time-of-day time for requests */
  u_long        extReqClockTotal;  /* total free-running-counter times */
  u_long        extReqClockMax;    /* max value of above */
  u_long        extReqReqClockTotal; /*total frc times for CommRequests */
  u_long        extReqReqClockMax;
  u_long        extReqInvalidTime;
  u_long       *extReqROtimeP;
  u_long       *extReqRWtimeP;
  u_long	extReqXferred;	/* # lost, ~tdhldr. should also be in xferE--*/
  u_long	extReqNewOwn;	/* # punted because fwdd retransmissions */

  u_long        acquireTotalTime;
  u_long        invalidTime;

  u_long	invErdrLost;	/* external reads lost from queue on inval */

  u_long	releases;	/* # of releases */
  u_long        releaseNoRdrs;  /* # of releases w/ no one waiting */
  u_long	releaseBusy;	/* # of releases that were still busy*/
  u_long	releaseInval;	/* # of releases that were invalidated */
  u_long	releaseNwriter;	/* # of rels with internal writers waiting*/
  u_long	releaseNreader;	/* # of rels with internal readers waiting*/
  u_long	releaseEwriter;	/* # of rels with external writers waiting*/
  u_long	releaseEreader;	/* # of rels with external readers waiting*/
  u_long        requestsFromSelf; /* # of requests dropped (from self) */

  u_long        denseUnaligned;	/* # of unaligned dirtybits */
  u_long        denseAlloced;	/* # of dense dirtybits allocated to back the unaligned */

#ifdef PAGE_BASED
/*
 * Define stats for page based code
 */
#define STAT_PAGED_INCARN_TABLE_SIZE  128
    u_long	paged_count;		/* # of call to lock xfer */
    u_long	paged_data;		/* actual lock data sent */
    u_long	paged_data_max;		/* max lock data sent */
    u_long      paged_incarn_ok;	/* # of request where diff list held incarn */
    u_long      paged_incarn_missing;	/* # of requests with a missing diff */
    u_long      paged_incarn_max;	/* # of requests where incarn < diff stored */
    u_long      paged_incarn_oversize;	/* # of requests where diff size > lock data size */
    u_long      paged_incarn_discarded;	/* # of requests where diff gen'd & discarded immediately */
    u_long	paged_incarn[STAT_PAGED_INCARN_TABLE_SIZE+1];	/* histo of current incarn - req's */
    u_long      paged_diff;		/* # of diff requests send */
    u_long      paged_diff_nodata;	/* # of zero-length diffs for locks w/ no data */
    u_long      paged_diff_nodiff;;	/* # of zero-length diffs for r/w upgrades */
    u_long      paged_diff_size;	/* total size of diff requests send */
    u_long      paged_diff_resend;	/* # of diff requests resend */
    u_long      paged_diff_resend_size;	/* total size of diff requests resend */
    u_long	paged_diff_buffer_used; /* Sum of sizes of all saved diffs */
    u_long	paged_diff_buffer_max;  /* Max of above */
    u_long      paged_pack_diff;	/* # of calls to pack diffs of data */
    u_long      paged_pack_diff_size;	/* total bytes of diff packed */
    u_long      paged_pack_data;	/* # of calls to pack all data */
    u_long      paged_pack_data_size;	/* total bytes of data packed */
#endif

} midLockStats;
static mutex_t statLock;		/* just in case */
#endif /* STATS or LTIMES */

/* 
 * Define macro to conditionally maintain stat info
 */
#ifdef STATS

# ifdef FAST_STATS
#  define INCR_STATS(x,by) (x) += (by)
# else
#  define INCR_STATS(x,by) \
    { mutex_lock(statLock);  (x) += (by);  mutex_unlock(statLock); }
# endif 

# ifdef PAGE_BASED
#   define INCR_PSTATS(x,by) INCR_STATS(x,by)
# else
#   define INCR_PSTATS(x,by)
#endif


#else /* not STATS */
#define INCR_STATS(x,by)   /* nothing to do  */
#define INCR_PSTATS(x,by)   /* nothing to do  */
#endif

/*
 * Define Stat structure for FastLocks.
 * The structure's layout MUST match the one in midway.h.
 * This should always be defined, since the application code depends on it.
 * WAS 10Jan94
 */
struct {
  unsigned long acqRO_count;
  unsigned long acqRW_count;
  unsigned long acqCA_count;
  unsigned long relRO_count;
  unsigned long relRW_count;
  unsigned long relCA_count;
  unsigned long acqRO_lib;
  unsigned long acqRW_lib;
  unsigned long acqCA_lib;
  unsigned long relRO_lib;
  unsigned long relRW_lib;
  unsigned long relCA_lib;
} midFastLockStats;



#ifdef LTIMES
#define STAT_GET_TIME(x)          \
{ struct timeval tp;              \
  struct timezone tzp;            \
                                  \
  my_gettimeofday(&tp,&tzp);  \
  x = tp.tv_sec & 0xffff;         \
  x *= 1000;                      \
  x += (tp.tv_usec/1000);         \
}
#define GET_CLOCK_VALUE(x) x = *midway_kern_frc;
#else
#define STAT_GET_TIME(x)
#define GET_CLOCK_VALUE(x)
#endif


#ifdef mips

/*
 * Define boundary markers on application's static data
 * Page-based sends all of static data to slave during thread_create
 * since we don't detect writes to that region of memory.
 */
extern vm_offset_t midway_start_data;	/* addr of appl's data segment (initialized extern/static data) */
extern vm_offset_t midway_start_sdata;	/* addr of appl's sdata segment (initialized "small" extern/static data) */
extern vm_offset_t midway_start_bss;	/* addr of appl's bss (uninitialized extern data) */
extern vm_offset_t midway_start_sbss;	/* addr of appl's sbss (uninitialized "small" extern data) */
extern vm_offset_t midway_start_static_bss;	/* addr of appl's bss (uninitialized static data) */
extern vm_offset_t midway_start_static_sbss;	/* addr of appl's sbss (uninitialized "small" static data) */
extern vm_offset_t midway_start_comm;		/* addr of appl's common data (discontiguous part of bss) */
extern vm_offset_t midway_start_scomm;		/* addr of appl's small common data (discontiguous part of sbss) */
extern vm_offset_t midway_start_text;	/* addr of appl's text segment */

extern vm_offset_t midway_end_data;	/* end of appl's data segment (initialized extern/static data) */
extern vm_offset_t midway_end_sdata;	/* end of appl's sdata segment (initialized "small" extern/static data) */
extern vm_offset_t midway_end_bss;	/* end of appl's bss (uninitialized extern data) */
extern vm_offset_t midway_end_sbss;	/* end of appl's sbss (uninitialized "small" extern data) */
extern vm_offset_t midway_end_static_bss;	/* end of appl's bss (uninitialized static data) */
extern vm_offset_t midway_end_static_sbss;	/* end of appl's sbss (uninitialized "small" static data) */
extern vm_offset_t midway_end_comm;		/* end of appl's common data */
extern vm_offset_t midway_end_scomm;		/* end of appl's small common data */
extern vm_offset_t midway_end_text;	/* end of appl's text segment */

#endif /* mips */



#ifdef PAGE_BASED

/*
 * Define header for saved list of diffs for a lock
 * The diffs are stored as a list of pbufs,
 * The header is stored at the top of the pbuf,
 * overwriting the m_lock_req_t originally stored there.
 */
typedef struct {
    clock_t	diff_incarn;		/* incarnation number for diff data */
    u_long	diff_size;		/* size of diff buffer */
} diff_desc_t;

#define DIFF_DESC_PTR(_pbufP) ((diff_desc_t *)(_pbufP))

void midLockIncrDiffs();	/* forward decl */

#endif /* PAGE_BASED */



/*
 ************************************************************************
 *			m i d _ l o c k _ X _ t o _ i n t e r n		*
 *----------------------------------------------------------------------*
 * Convert external FASTLOCK states to internal lock states 		*
 * (and vice-versa)                                             	*
 ************************************************************************
 */

/* 
 * FASTLOCK states correspond to internal lock states as follows:
 *  MIDWAY_LOCKHELD_FREE     = ((state & L_OWN) && (busy == 0) && (rdrs == 1))
 *   -> corresponds to own, r/o with no one else locally or remotely active
 *  MIDWAY_LOCKHELD_RO       = ((state & L_OWN) && (busy == 1) && (rdrs == 1))
 *   -> owning site holds R/O only.  For now, no state for locally R/O and
 *	remotely R/O (although remote cache state exists)
 *  MIDWAY_LOCKHELD_RW       = ((state & L_OWN) && (busy == 1) && (rdrs == 0))
 *   -> since we get here from FREE, rdrs will actually be 1... (--mjz)
 *  MIDWAY_LOCKHELD_CACHE    = (!(state & L_OWN) && (busy >= 0) && (rdrs >= 1))
 *   -> we have a cached copy, but we are not using it.
 *  MIDWAY_LOCKHELD_CACHE_RO = (!(state & L_OWN) && (busy >= 1) && (rdrs >= 1))
 *   -> we have a cached copy, and we are using it.
 *  MIDWAY_LOCKHELD_OTHER    = other internal states
 *
 * When running in application code, with FastLock state anything except OTHER,
 * the internal states MUST correspond to either FREE or CACHE. All other internal
 * states are an internal error.
 *
 * Note that if an INVALIDATE is received for a lock we have a cached copy of, 
 * then rdrs == 0, to prevent new acquires.
 */
#define NO_Q_STATES(lP) ((lP->ewrite == 0) && (lP->eread == 0) && (lP->nreadq == 0) && (lP->nwriteq == 0))
#define VALID_FREESTATES(lP) ((lP->state & L_OWN) && (lP->rdrs == 1) && (lP->busy == 0) && NO_Q_STATES(lP))
#define VALID_CACHESTATES(lP) ((lP->state & L_OWN) == 0)
#define VALID_LOCKSTATES(lP) (VALID_FREESTATES(lP) || VALID_CACHESTATES(lP))

/*
 * When running in the midway library code, with the lock's mutex held, the FastLock
 * state MUST be set to OTHER which disables the FastLock code. Actually, this is
 * only true if the code is checking/changing the internal lock states.
 */
#define VALID_FASTSTATES(lP) (lP->held == MIDWAY_LOCKHELD_OTHER)


#if MIDWAY_FASTLOCKS == 0
#  define mid_lock_X_to_intern(lockP) 
#  define mid_lock_X_to_extern(lockP)
#else

/*
 * Convert FASTLOCKS state to internal lock states.
 * This is done at the all entry points to the real lock code
 * and everytime after the guarding mutex is unlocked or
 * released via a condition_wait.
 * 
 * The FastLocks really need to be better integrated with the internal
 * lock states. Otherwise this will be a constant source of bugs. 
 * WAS 11Jan94
 */
void
mid_lock_X_to_intern (pl_t *lockP)
{
  if (lockP->held == LOCK_RO) {
    ASSERT (VALID_FREESTATES(lockP));
    lockP->rdrs = 1;
    lockP->busy = 1;

  } else if (lockP->held == LOCK_RW) {
    ASSERT (VALID_FREESTATES(lockP));
    lockP->rdrs = 0;
    lockP->busy = 1;
    
  } else if (lockP->held == MIDWAY_LOCKHELD_CACHE_RO) {
    ASSERT (VALID_CACHESTATES(lockP));
    lockP->busy++;		/* may be other rdrs */

  } else if (lockP->held == MIDWAY_LOCKHELD_FREE) {
    ASSERT (VALID_FREESTATES(lockP));
  } else if (lockP->held == MIDWAY_LOCKHELD_CACHE) {
    ASSERT (VALID_CACHESTATES(lockP));
  } else {
    ASSERT (lockP->held == MIDWAY_LOCKHELD_OTHER);
  }
 
  lockP->held = MIDWAY_LOCKHELD_OTHER;
}

/*
 * Convert internal lock state to FASTLOCKS state
 * This is done on exit from the midway_library calls.
 */
void
mid_lock_X_to_extern (pl_t *lockP)
{
  if (VALID_FREESTATES(lockP))
    lockP->held = MIDWAY_LOCKHELD_FREE;
  else if (VALID_CACHESTATES(lockP) && (lockP->rdrs > 0)) {
    lockP->held = MIDWAY_LOCKHELD_CACHE;
  } else 
    lockP->held = MIDWAY_LOCKHELD_OTHER;
}
#endif /* MIDWAY_FASTLOCKS */

/*
 ************************************************************************
 *			m i d _ l o c k _ i n i t			*
 *----------------------------------------------------------------------*
 * Initialize locktab, lockdx, and lockmutex.				*
 ************************************************************************
 */
void
mid_lock_init()
{
  int i;
  vm_offset_t addr;

  /* Verify the locktable has been initialized */
  ASSERT ((vm_locktable_base != 0) && (vm_locktable_end != 0));

  /* allocate & initialize mutex to guard locktable */
  lockmutex = mutex_alloc();
  MIDWAY_MUTEX_SET_NAME(lockmutex, "lockmutex");

  /* set variable to make it easier to display locktable */
  locktable = (pl_t *) vm_locktable_base;


  /* 
   * create addrmaps for static region
   */
#ifdef mips
  { s_addr_map_t *tailP = NULL;
    s_addr_map_t *prevP = NULL;
    vm_size_t static_size;

    /* Note: some of these regions may overlap */
    static_amap = NULL;
    if ((midway_end_data-midway_start_data) > 0)
	mid_addrmap_add (&static_amap, &tailP, &prevP,
			 midway_start_data, midway_end_data-midway_start_data);
    if ((midway_end_sdata-midway_start_sdata) > 0)
	mid_addrmap_add (&static_amap, &tailP, &prevP,
			 midway_start_sdata, midway_end_sdata-midway_start_sdata);
    if ((midway_end_bss-midway_start_bss) > 0)
	mid_addrmap_add (&static_amap, &tailP, &prevP, 
			 midway_start_bss, midway_end_bss-midway_start_bss);
    if ((midway_end_sbss-midway_start_sbss) > 0)
	mid_addrmap_add (&static_amap, &tailP, &prevP, 
			 midway_start_sbss, midway_end_sbss-midway_start_sbss);
    if ((midway_end_static_bss-midway_start_static_bss) > 0)
	mid_addrmap_add (&static_amap, &tailP, &prevP,
			 midway_start_static_bss, midway_end_static_bss-midway_start_static_bss);
    if ((midway_end_static_sbss-midway_start_static_sbss) > 0)
	mid_addrmap_add (&static_amap, &tailP, &prevP,
			 midway_start_static_sbss, midway_end_static_sbss-midway_start_static_sbss);
    if ((midway_end_comm-midway_start_comm) > 0)
	mid_addrmap_add (&static_amap, &tailP, &prevP,
			 midway_start_comm, midway_end_comm-midway_start_comm);
    if ((midway_end_scomm-midway_start_scomm) > 0)
	mid_addrmap_add (&static_amap, &tailP, &prevP,
			 midway_start_scomm, midway_end_scomm-midway_start_scomm);

    /* for debugging */
    vm_static_base = (vm_offset_t) static_amap->addr;
    vm_static_end = (vm_offset_t) tailP->end;
    for (static_size = 0, tailP = static_amap; tailP != NULL; tailP = tailP->next) 
	static_size += tailP->len;

    if (Comm_index == 1) {
	fret("vm_static_total: base = 0x%x end = 0x%x, data = %lu bytes",
	     FIRST_STATIC_SYM, LAST_STATIC_SYM,
	     (u_long) LAST_STATIC_SYM - (u_long)FIRST_STATIC_SYM);
	fret("vm_static_user: base = 0x%x end = 0x%x, data = %lu bytes",
	     vm_static_base, vm_static_end, static_size);
	
	fret("vm_text_total: base = 0x%x end = 0x%x, data = %lu bytes",
	     FIRST_TEXT_SYM, LAST_TEXT_SYM,
	     (u_long) LAST_TEXT_SYM - (u_long)FIRST_TEXT_SYM);
	fret("vm_text_user: base = 0x%x end = 0x%x, data = %lu bytes",
	     midway_start_text, midway_end_text, 
	     (u_long)midway_end_text - (u_long)midway_start_text);
    }
  }
#else /* not mips  */
  /* create dummy addrmap for static region */
  /* uck - should have routine in lockaddr to handle? */
  static_amap = (s_addr_map_t *) MALLOC(sizeof(s_addr_map_t));
  ASSERT(static_amap != 0); 
  static_amap->addr = FIRST_STATIC_SYM;	/* F_S_S from shmalloc.h */
  static_amap->end = LAST_STATIC_SYM;
  static_amap->len = (u_long) LAST_STATIC_SYM - (u_long)FIRST_STATIC_SYM;
  static_amap->next = 0;	/* no next */
#endif /* mips */


#ifdef PAGE_BASED

  /* verify sizeof diff descriptor */
  ASSERT (sizeof(diff_desc_t) <= sizeof(m_lock_req_t));
#endif /* PAGE_BASED */

  /* initialize hint table */
  for (i = 0; i < HINT_TABLE_SIZE; i++)
    hintTable[i] = NULL;
  hintIndex = 0;
  hintMutex = mutex_alloc();
  MIDWAY_MUTEX_SET_NAME(hintMutex,"lock.c:hintMutex");

#if defined(STATS) || defined(LTIMES)
  statLock = mutex_alloc ();
  MIDWAY_MUTEX_SET_NAME(statLock,"lock.c:statLock");
  mid_lock_reset_stats();
  mid_barrier_reset_stats();
#endif

#ifdef LTIMES
  midway_setup_mapped_frc ();

  /* allocate histogram tables, add +1 for overflow */
  midLockStats.extReqROtimeP = (unsigned long *) 
    CALLOC( (STAT_TIME_TABLE_SIZE+1), sizeof(unsigned long) );
  ASSERT (midLockStats.extReqROtimeP != NULL);
  
  midLockStats.extReqRWtimeP = (unsigned long *) 
    CALLOC( (STAT_TIME_TABLE_SIZE+1), sizeof(unsigned long) );
  ASSERT (midLockStats.extReqRWtimeP != NULL);
#endif

  /* 
   * Initialize lock addr code 
   */
  mid_addrmap_init ();

} /* mid_lock_init */


/*
 ************************************************************************
 *		    m i d w a y _ l o c k _ a l l o c			*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
midway_lock_t
midway_lock_alloc()
{
  pl_t *lock;

  /* allocate & initialize a lock structure */
  lock = mid_alloc_synch_object (NULL, L_OWN); /* L_OWN == LOCK */

  INCR_STATS(midLockStats.allocated, 1);

  return ((midway_lock_t) lock);
} /* midway_lock_alloc */


/*
 * Internal routine to allocate either a lock or a barrier.
 */
pl_t *
mid_alloc_synch_object (
			u_long lname,	/* =NULL for a new lock  */
			u_long state	/* Initial state for synch object */
			)
{
  pl_t *lockP, *newP;
  vm_size_t size;
  vm_offset_t addr;
  kern_return_t krc;
  u_long tid;
  char *mname;
  s_addr_map_t *amapP, *next_amapP;


  /*
   * as of 19jun91 this should only happen on main processor... mjz 
   */
  if ((Comm_index != 1)  && (lname == NULL)) {
    StatusMsg("Alloc: used on other than main processor-returning bogus info.");
    return(0);
  }

  /*
   * find next free slot in locktable
   * Note: The following code assumes that locks are never deallocated. WAS 8Jan94
   */
  mutex_lock(lockmutex);

  if (lname == NULL) {
    /* get pointer to next free slot */
    lockP = (pl_t *) vm_locktable_free;
    vm_locktable_free += sizeof(*lockP);
  } else {
    lockP = (pl_t *) lname;
    newP = lockP + 1;
    if (newP > (pl_t *) vm_locktable_free)
      vm_locktable_free = (vm_offset_t) newP;
  }
  
  if (vm_locktable_free > vm_locktable_end) {

    /* 
     * Acquire additional locktable space.
     *
     * Note: This code keeps the locktable contiguous. Since we are using
     * addresses as unique names, the table could be broken up and allocated
     * out of standard private space. Keeping it contiguous make it easier
     * to locate and easier to debug.   WAS 9Jan93
     */
    addr = vm_locktable_end + 1;
    size = (((vm_locktable_free - vm_locktable_end) + (LOCKTABLE_ALLOC_SIZE-1))
	    / LOCKTABLE_ALLOC_SIZE) * LOCKTABLE_ALLOC_SIZE;
    krc = VM_ALLOCATE (midway_task_self, &addr, size, 0); /* 0 => Anywhere is FALSE */
    if ((krc != 0) || (addr != (vm_locktable_end+1))) {
      StatusMsg("midway_lock_alloc: unable to vm_allocate space for locktable.");
      return(0);
    }
    vm_locktable_end += size;
    ASSERT (vm_locktable_free <= vm_locktable_end);

    /* initialize fastlock field for unused locks */
    for (newP = (pl_t *)vm_locktable_free; (vm_offset_t)(newP+1) <= vm_locktable_end; newP++)
      newP->held = MIDWAY_LOCKHELD_OTHER;
  }

  /* release mutex guarding locktable */
  mutex_unlock(lockmutex);


  /* Check if this is a new structure which needs to be allocated */
  if (lockP->lock == 0) {

    /*
     * Initialize lock structure
     */
    mutex_init(&(lockP->m));
#if MIDWAY_MUTEX_NAMES
    mname = MALLOC(12);
    sprintf(mname,"%c%.8x", (state & L_BARRIER) ? 'B' : 'L', lockP);
    mutex_set_name(&(lockP->m), mname);
#endif

    lockP->state = state;
    lockP->bestguess = 1;		/* need to change this too */
    lockP->bgtime = lockP->incarnation;
    lockP->tcreate = LocalClock;	/* get some idea */
    lockP->amap = NULL;
    lockP->amap_tailP = NULL;
    lockP->amap_prevP = NULL;
    lockP->amap_data_size = 0;
    lockP->lock = (u_long)lockP;	/* mark structure as inuse */

#ifdef PAGE_BASED
    { int i;
      for (i = 0; i < MAX_DIFF_INCARN; i++) {
	  lockP->diff[i].diff_size = 0;
	  lockP->diff[i].diff_status = DIFF_STAT_UNKNOWN;
	  lockP->diff[i].diff_incarn = NULL_CLOCK_VALUE;
	  mid_pbuf_qinit(&(lockP->diff[i].diff_queue));
      }
    }
#endif
    
    if (state & L_BARRIER) {
      condition_init(&(lockP->bar_wait));
      mid_extq_init(&(lockP->bar_waiters));
      mid_pbuf_qinit(&(lockP->output));
      mid_pbuf_qinit(&(lockP->old_output));
      lockP->bar_num = 1;		/* assume only need 1 thread to hit barrier locally */
      lockP->bar_curtime = MIN_CLOCK_VALUE;	/* first iteration at 1 not 0 */
      
    } else {  /* type == LOCK */
      lockP->incarnation = 1;	/* 1st owner gets incarnation 1 */
      lockP->rdrs = 1;
      lockP->aerdrs = (BITARRAY *) CALLOC(1, midBITARRAYsize(Comm_hnum));
      ASSERT(lockP->aerdrs != (BITARRAY *) 0);
      condition_init(&(lockP->ireadc));
      condition_init(&(lockP->ereadc));
      condition_init(&(lockP->iwritec));
      condition_init(&(lockP->ewritec));
      condition_init(&(lockP->xferc));
      condition_init(&(lockP->iupc));
      mid_extq_init(&(lockP->erdrs)); /* gotta init the queues!! */
      mid_extq_init(&(lockP->ewrtrs));
    }
  } 
#if 0
  The following fails because the amaps can get split across two packets
    when the RemoteRelease occurs.  We do not want to zap amaps that we
      have already created.  Re-use must be handled differently. 7feb94mjz;
  else {

    /*
     * Otherwise we are reusing an existing lock structure
     * Reinitialize it as necessary.
     * It seems like we should be doing more here.  WAS 11Jan94
     */
    for (amapP = lockP->amap; amapP != NULL; amapP = next_amapP) {
      next_amapP = amapP->next;
      FREE(amapP);
    }
    lockP->amap = NULL;
    lockP->amap_tailP = NULL;
    lockP->amap_prevP = NULL;
    lockP->amap_data_size = 0;

    ** Need to free diff buffers & reinitialize diff_list.
    ** WAS 17May94
  }
#endif /*0*/

  if (OUR_TID == 0) {		/* not yet assigned */
    tid = cpt_init();
  }

  /* initialize FASTLOCKS state vars */
  lockP->held = MIDWAY_LOCKHELD_OTHER;
  mid_lock_X_to_extern (lockP);

  return (lockP);
}

/*
 ************************************************************************
 *		m i d w a y _ p r i n t _ l o c k			*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
void
midway_print_lock(lockn)
     midway_lock_t lockn;	/* lock name */
{
  pl_t *lockP;

  /* convert lockname to a pointer to the lock structure */
  lockP = (pl_t *) lockn;

  fprintf(stderr, "lock %lx%s: held=%lx< ", 
	  lockP, (lockP == (pl_t *)lockP->lock) ? "" : "(INVALID)",
	  lockP->held);
  if (lockP->held == MIDWAY_LOCKHELD_FREE) fprintf(stderr, "free ");
  if (lockP->held & MIDWAY_LOCKHELD_RO) fprintf(stderr, "ro ");
  if (lockP->held & MIDWAY_LOCKHELD_RW) fprintf(stderr, "rw ");
  if (lockP->held & MIDWAY_LOCKHELD_CACHE) fprintf(stderr, "cache ");
  if (lockP->held & MIDWAY_LOCKHELD_OTHER) fprintf(stderr, "other ");
  fprintf(stderr, "> state=%lx< ", lockP->state);
  if (lockP->state & L_OWN) fprintf(stderr,"own ");
  if (lockP->state & L_ACQ) fprintf(stderr,"acq ");
  if (lockP->state & L_ACQRW) fprintf(stderr, "acqrw ");
  if (lockP->state & L_XFER) fprintf(stderr, "xfer ");
  if (lockP->state & L_DATAIN) fprintf(stderr, "datain ");
  fprintf(stderr, ">\n\t busy=%lu rdrs=%lu nwriteq=%lu eread=%lu ewrite=%lu",
	  lockP->busy, lockP->rdrs, lockP->nwriteq, lockP->eread, lockP->ewrite);
  fprintf(stderr, " bguess=%lu xfer=%lu xtid=%lu\n", 
	  lockP->bestguess, lockP->xfersite, lockP->xfertid);
  fflush(stderr);
/*
  mid_addrmap_prt(lockP->lock, lockP->amap);
*/
  return;
} /* midway_print_lock */

/*
 ************************************************************************
 *		m i d w a y _ l o c k _ a c q u i r e			*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
void
midway_lock_acq( lockn, mode )
     midway_lock_t	lockn;		/* lock name */
     midway_acquire_t	mode;		/* shared or exclusive */
{
  int rc;
  pl_t *lock;
  char255 smsg;
#ifdef LTIMES
  u_long startTime, endTime;
#endif


  INCR_STATS(midLockStats.acquires, 1);
  STAT_GET_TIME (startTime);

  /* convert lockname to a pointer to the lock structure */
  lock = (pl_t *) lockn;

  if (IS_DBG_LOCK_OPS) {
    sprintf(smsg, "Acquire(%lu,%s)", lockn, 
	    (mode == LOCK_RO) ? "RO" : 
	    (mode == LOCK_RW) ? "RW" : "INVALID");
    StatusMsg(smsg);
  }
  TRACE_4("ACQ ", lockn, mode, lock->held);

#ifdef DEBUG_USER
  if (((vm_offset_t) lockn < vm_locktable_base) || 
      ((vm_offset_t) lockn >= vm_locktable_end) ||
      ((vm_offset_t) lockn != (vm_offset_t) lock->lock) ||
      (lock->state & L_BARRIER)) {
    sprintf(smsg, "midway_lock_acquire: Invalid lock id %lu\n", lockn);
    StatusMsg (smsg);
#if STOP_ON_ERROR
    botch("midway_lock_acquire");
#else
    return;
#endif
  }
  if ((mode != LOCK_RO) && (mode != LOCK_RW)) {
    sprintf(smsg, "midway_lock_acquire: Invalid lock mode %d\n", mode);
    StatusMsg (smsg);
#if STOP_ON_ERROR
    botch("midway_lock_acquire");
#else
    return;
#endif
  }
#endif 

#if UPDATEQ_FLUSH_AGGRESSIVELY > 0
  /*
   * Flush update queue, before acquiring a global lock 
   * Only flush on locks acquired RW & using PC model.
   */
  if ((mode == LOCK_RW) && 
      (midway_default_model == MODEL_PC) &&
      (UPDATEQ_NOT_EMPTY)) {
    if (IS_DBG_PORT_QUEUE) StatusMsg("Acquire flushing update queue");
    rc = mid_updateq_flush (UPDATEQ_CALLER_ACQUIRE);
  }
#ifdef PAGE_BASED_RC
  /* Flush update queue for RC version of page-based code */
  if (mode == LOCK_RW)
    paged_flush_updates(lockn);
#endif
#endif 
    

  /*
   * Grab mutex protecting lock structure
   */
  mutex_lock(&(lock->m));

  /*
   * Call internal routine to perform actual acquire.
   * This routine also release the lock's mutex.
   */
  midway_lock_acqX (lockn, mode);



#ifdef LTIMES
  STAT_GET_TIME (endTime);
  endTime -= startTime;
  INCR_STATS (midLockStats.acquireTotalTime, endTime);
#endif

  if (IS_DBG_LOCK_OPS) {
    sprintf(smsg, "Acquire of %lu complete.", lockn);
    StatusMsg(smsg);
  }

  return;
} /* midway_lock_acquire */
 

/*
 * Internal routine to perform actual lock acquire.
 * Caller MUST hold mutex guarding lock and 
 * this routine will release it before returning.
 * Note: The release is done here to reduce the size
 * of the macro expansion in the application code.
 */
void
midway_lock_acqX (
		  midway_lock_t		lockn,	/* lock name */
		  midway_acquire_t	mode	/* shared or exclusive */
		  )
{
  int rc;
  u_long i;
  pl_t *lock;
  char255 smsg;


  /* cast lockname to a pointer to its internal structure */
  lock = (pl_t *) lockn;
  TRACE_6("ACQx", lockn, mode, lock->state, lock->busy, lock->rdrs);


#if MIDWAY_FASTLOCKS == 0
  /* disable fastlocks */
#else
  /*
   * Check if the lock is free.
   */
  if (lock->held == MIDWAY_LOCKHELD_FREE) {
    lock->held = mode;		/* FASTLOCKS! */

#ifdef STATS
    if (mode == LOCK_RO) {
      INCR_STATS(midFastLockStats.acqRO_lib, 1);
    } else {
      INCR_STATS(midFastLockStats.acqRW_lib, 1);
    }
#endif /* STATS */

    /* verify lock is in proper state */
    ASSERT (VALID_FREESTATES(lock));
    TRACE_2("Acff", lockn);
    goto done;
  }
  if ((mode == LOCK_RO) &&
      (lock->held == MIDWAY_LOCKHELD_CACHE)) {
    lock->held = MIDWAY_LOCKHELD_CACHE_RO;	/* FASTLOCKS! */

    INCR_STATS(midFastLockStats.acqCA_lib, 1);
    ASSERT (VALID_CACHESTATES(lock));
    TRACE_2("Acfr", lockn);
    goto done;
  }

  /* 
   * Convert FASTLOCKS state to internal lock state
   */
  mid_lock_X_to_intern (lock);
#endif /* MIDWAY_FASTLOCKS */

  
  /*
   * Handle lock request 
   */
  /* first, check for inconsistent transfer state.  Wait for improvement. */
  while ((lock->state & L_XFER) && (lock->xfersite == 0)) {
      if (IS_DBG_LOCK_OPS) {
	  sprintf(smsg,"Acquire(%lu)[RW]: L_XFER set, and xfersite 0.  Waiting.",lockn);
	  StatusMsg(smsg);
      }
      condition_wait(&(lock->xferc), &(lock->m));
      if (IS_DBG_LOCK_OPS) {
	  sprintf(smsg,"Acquire(%lu)[RW]: Back from xferc condition.",lockn);
	  StatusMsg(smsg);
      }
      
      /* Restore internal state from FASTLOCK state after waiting */
      mid_lock_X_to_intern (lock);
      
      INCR_STATS (midLockStats.acquireRWincon, 1);
  }
  /***
   ***
   ***  LOCK_RO
   ***
   **/
  if (mode == LOCK_RO) {	/* read only */
    
    INCR_STATS(midLockStats.acquireRO, 1);
    
    if (lock->state & (L_ACQ|L_ACQRW)) {
      if (IS_DBG_LOCK_OPS) {
	sprintf(smsg,"Acquire(%lu)[RO]: already acquiring so queueing.",lockn);
	StatusMsg(smsg);
      }
      INCR_STATS (midLockStats.acquireROacq, 1);

      /* queue */
      /* don't need to keep tid around, queue is implicit in condition var */
      lock->nreadq++;
      condition_wait(&(lock->ireadc), &(lock->m)); /* wait for improvement */
      /* busy inc'd by signaller */

      /* Restore internal state from FASTLOCK state after waiting */
      mid_lock_X_to_intern (lock);

    } /* case L_ACQ */

    else if ((lock->state & L_OWN) && (lock->rdrs == 0)) { /* own r/w */
      if (IS_DBG_LOCK_OPS) {
	sprintf(smsg,"Acquire(%lu)[RO]: OWN R/W",lockn);
	StatusMsg(smsg);
      }
      /* if busy then queue else demote & grant */
      if (lock->busy) {		/* queue */
	lock->nreadq++;		/* note I exist */
	/* busy will be inc'd by nreadq before release of condition */
	while (lock->rdrs == 0) { /* NOTE: not only condition.  Do all? MJZ */
	  condition_wait(&(lock->ireadc), &(lock->m));
	  mid_lock_X_to_intern (lock);
	}
	INCR_STATS (midLockStats.acquireROownrwbusy, 1);
      } else {			/* demote & grant */
	lock->rdrs++;		/* demote, now r/o */
	/* should be no other q'rs, otherwise Release or the LockRequest
	   (external acquire) would have caught */
	lock->busy++;		/* note another thread */
	INCR_STATS (midLockStats.acquireROownrw, 1);
      }
    } /* case own r/w */

    else if (lock->rdrs > 0) {	/* have r/o already */
      if (IS_DBG_LOCK_OPS) {
	sprintf(smsg,"Acquire(%lu)[RO]: OWN OR HOLD R/O",lockn);
	StatusMsg(smsg);
      }
      /* grant */
      lock->busy++;		/* note another thread... */
#ifdef STATS
      if (lock->state & L_OWN) {
	if (lock->rdrs == 1) {
	  INCR_STATS (midLockStats.acquireROownro, 1);
	} else {
	  INCR_STATS (midLockStats.acquireROlockro, 1);
	}
      } else {
	INCR_STATS (midLockStats.acquireROholdro, 1);
      }
#endif /* STATS */
    }  /* case own or have r/o */

    else {
      /* don't know anything about the lock */
      midExtLockRequest(lock, mode);
      INCR_STATS (midLockStats.acquireROreqs, 1);
    }
  } /* mode == LOCK_RO */

  /***
   ***
   ***  LOCK_RW
   ***
   **/
  else {			/* mode r/w */

    INCR_STATS (midLockStats.acquireRW, 1);


    if (lock->state & L_ACQRW) { /* acquiring exclusively */
      if (IS_DBG_LOCK_OPS) {
	sprintf(smsg,"Acquire(%lu)[RW]: ACQ R/W",lockn);
	StatusMsg(smsg);
      }

      
      /* queue */
      lock->nwriteq++;	/* note I exist */
      /*do { can't do a loop w/o more state (like "RELEASING WRITER") */
      condition_wait(&(lock->iwritec), &(lock->m));
      mid_lock_X_to_intern (lock);
      /*} while ( lock->busy ); */
      lock->nwriteq--;	/* I no longer exist (in q) */
      /* busy now inc'd by signaller 19may92 */
      INCR_STATS (midLockStats.acquireRWacqrw, 1);

    } /* L_ACQRW */
    else if (lock->state & L_ACQ) { /* acquiring r/o copy */
      if (IS_DBG_LOCK_OPS) {
	sprintf(smsg,"Acquire(%lu)[RW]: ACQ R/O",lockn);
	StatusMsg(smsg);
      }
      INCR_STATS (midLockStats.acquireRWacqro, 1);
      INCR_STATS (midLockStats.acquireRWreqs, 1);
      /* request exclusive copy. */
      /* this request will stall until the r/o copy is acquired,
	 simplifying our life and also avoiding a degenerate case.
	 See notes... */
      midExtLockRequest(lock,mode);
    } /* L_ACQ */

    else if (lock->state & L_XFER) { 
      if (IS_DBG_LOCK_OPS) {
	sprintf(smsg,"Acquire(%lu)[RW]: Xferring already",lockn);
	StatusMsg(smsg);
      }
      INCR_STATS (midLockStats.acquireRWxfer, 1);
      INCR_STATS (midLockStats.acquireRWreqs, 1);
      /* forward request back to new owner */ /* ASSERT(bestguess)? */
      midExtLockRequest(lock,mode);
    } /* L_XFER case */

    else if (lock->state & L_OWN) {
      if (lock->rdrs == 0) {	/* own exclusively */
	if (IS_DBG_LOCK_OPS) {
	  sprintf(smsg,"Acquire(%lu)[RW]: OWN R/W",lockn);
	  StatusMsg(smsg);
	}
	INCR_STATS (midLockStats.acquireRWownrw, 1);
	
	if (lock->busy) {
	  /* queue */
	  lock->nwriteq++;	/* note I exist */
	  /*do { can't do -- see above */
	    condition_wait(&(lock->iwritec), &(lock->m));
	    mid_lock_X_to_intern (lock);

	  /*} while( lock->busy ); */
	  lock->nwriteq--;	/* I no longer exist in q */
	  /* busy set by signaller */
	  INCR_STATS (midLockStats.acquireRWownrwbusy, 1);
	} 
	else {
	  /* grant */
	  lock->busy++;		/* note we've got it */
	}
      } 
      else {			/* own r/o, not xferring, not acq */
	if (IS_DBG_LOCK_OPS) {
	  sprintf(smsg,"Acquire(%lu)[RW]: OWN R/O",lockn);
	  StatusMsg(smsg);
	}

	INCR_STATS (midLockStats.acquireRWownro, 1);
	if (lock->rdrs > 1) {	/* there are read only copies */
	  u_long incarnval;	/* save while unlocked. not really necessary */

	  lock->rdrs = 0;	/* block new local read requests */
	  lock->busy++;		/* ensure lock is not taken by others */
	  incarnval = lock->incarnation; /* invalidate needs */

	  mutex_unlock(&(lock->m)); /* allow others in for comm *mjz* */
	  /* request invalidates, must wait for all replies */
	  if (IS_DBG_LOCK_OPS) {
	    sprintf(smsg,"Acquire(%lu)[RW]: going to invalidate.",lockn);
	    StatusMsg(smsg);
	  }
	  INCR_STATS (midLockStats.acquireRWownroinval, 1);
	  
	  CommMrequest(lockn, M_INVALIDATE_OP, 0, 0, 
		       NULL_CLOCK_VALUE, lock->aerdrs,
		       (char *) &incarnval, sizeof(incarnval)); 
	  if (IS_DBG_LOCK_OPS) {
	    sprintf(smsg,"Acquire(%lu)[RW]: invalidate complete.",lockn);
	    StatusMsg(smsg);
	  }
	  INCR_STATS (midLockStats.acquireRWownrordrs, 1);
	  /* if here, all invalidates have come through */
	  /* only state changes should be to add q'ers, either internal
	     or external (since lock was held busy in r/w mode...) */
	  mutex_lock(&(lock->m));

	  /* Restore internal state from FASTLOCK state after waiting */
	  mid_lock_X_to_intern (lock);

	  /* since invalidates have come in, we have no reason to remember */
	  /* who we gave copies to... */
	  bzero((char *) lock->aerdrs, midBITARRAYsize(Comm_hnum));
	  ASSERT (lock->busy > 0);
	  lock->busy--;
	  lock->incarnation++;	/* bump on lock upgrade */
	  TRACE_4("ACQi", lockn, lock->dataclock, lock->incarnation);

#ifdef PAGE_BASED
	  midLockIncrDiffs (lock);
#endif
	} 
	else {			/* no read only copies */
	  lock->rdrs = 0;	/* promote lock time */
	} /* no ro copies */
	/* if here, we now own r/w, and have no r/o copies outstanding */
	/* but there could be other local readers still around... */
	if (lock->busy) {	/* but in use at this site */
	  if (IS_DBG_LOCK_OPS) {
	    sprintf(smsg,"Acquire(%lu)[RW]: lock busy, so queueing.",lockn);
	    StatusMsg(smsg);
	  }
	  /* queue */
	  lock->nwriteq++;	/* note I exist */
	  /*do { can't do -- see above */
	    condition_wait(&(lock->iwritec), &(lock->m));
	    mid_lock_X_to_intern (lock);
	  /*} while( lock->busy ); */
	  lock->nwriteq--;	/* I no longer exist in Q */
	  /* busy set by signaller */
	  if (IS_DBG_LOCK_OPS) {
	    sprintf(smsg,"Acquire(%lu)[RW]: released from queue.",lockn);
	    StatusMsg(smsg);
	  }
	  INCR_STATS (midLockStats.acquireRWownrobusy, 1);
	}  /* busy */
	else {
	  if (IS_DBG_LOCK_OPS) {
	    sprintf(smsg,"Acquire(%lu)[RW]: not busy, claiming lock.",lockn);
	    StatusMsg(smsg);
	  }

	  /* if here, no longer busy */
	  /* grant */
	  lock->busy++;
	}
      } /* end of own r/o */
    } /* end of L_OWN case */

    else {			/* otherwise; none above conditions true */
      /* if we're here, then hold r/o or nothing */
      /* request externally */
      if (IS_DBG_LOCK_OPS) {
	sprintf(smsg,"Acquire(%lu)[RW]: Hold R/O or ignorant",lockn);
	StatusMsg(smsg);
      }
      
      INCR_STATS (midLockStats.acquireRWreqs, 1);
      midExtLockRequest(lock, mode);

      /* caveat: if we are here, then we might still have some readers
	 floating around from a previous existence... wait for busy to
	 go to zero if so... *//* mid...Request inc's busy */
      if ((mode == LOCK_RW) && (lock->busy > 1)) {
	lock->busy--;		/* take ourselves out */
	lock->nwriteq++;	/* note I exist */
	/*do { can't do -- see above */
	  condition_wait(&(lock->iwritec), &(lock->m));
	  mid_lock_X_to_intern (lock);
	/* } while (lock->busy); */
	lock->nwriteq--;	/* i no longer exist in q */
	/* busy set by signaller */
      } /* if have to wait for readers to finish... */
    }
  } /* end mode r/w */

  /* for now, if we acquire this way, we assume fastlocks are disabled
     until release.  Verify this fact. */
  ASSERT(lock->held == MIDWAY_LOCKHELD_OTHER);

done:

  /*
   * Release mutex guarding lock structure before returning.
   */
  TRACE_6("acqx", lockn, lock->held, lock->state, lock->busy, lock->rdrs);
  mutex_unlock(&(lock->m));

  return;
} /* midway_lock_acqX */

/*
 ************************************************************************
 *		m i d w a y _ l o c k _ r e l e a s e			*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
void
midway_lock_rel( lockn )
     midway_lock_t lockn;		/* lock name */
{
  u_long host, thread, tid, junk;
  pl_t *lock;
  char255 smsg;
  int rc;

  /* convert lockname to a pointer to the lock structure */
  lock = (pl_t *) lockn;

  if (IS_DBG_LOCK_OPS) {
    sprintf(smsg, "Release(%lu)", lockn);
    StatusMsg(smsg);
  }
  INCR_STATS (midLockStats.releases, 1);
  TRACE_3("REL ", lockn, lock->held);

#ifdef DEBUG_USER
  if (((vm_offset_t) lockn < vm_locktable_base) || 
      ((vm_offset_t) lockn >= vm_locktable_end) ||
      ((vm_offset_t) lockn != (vm_offset_t) lock->lock) ||
      (lock->state & L_BARRIER)) {
    sprintf(smsg, "midway_lock_release: Invalid lock id %lu\n", lockn);
    StatusMsg (smsg);
#if STOP_ON_ERROR
    botch("midway_lock_release");
#else
    return;
#endif
  }
#endif 

#if UPDATEQ_FLUSH_AGGRESSIVELY > 0
  /*
   * Flush update queue, before acquiring a global lock 
   * Only flush on locks held R/W.
   */
  if ((lock->rdrs == 0) &&	/* not r/o , at least */
      (UPDATEQ_NOT_EMPTY)) {
    if (IS_DBG_PORT_QUEUE) StatusMsg("Release flushing update queue");
    rc = mid_updateq_flush (UPDATEQ_CALLER_RELEASE);
  }
#ifdef PAGE_BASED_RC
  /* Flush update queue for page-based code */
  if (lock->rdrs == 0)
    paged_flush_updates(lockn);
#endif
#endif 


  /*
   * Grab mutex protecting lock structure
   */
  mutex_lock(&(lock->m));

  /*
   * Call internal routine to perform actual release.
   * This routine also release the lock's mutex.
   */
  midway_lock_relX (lockn);

  return;
} /* midway_lock_release */


/*
 * Internal routine to perform actual release
 */
void
midway_lock_relX (
		  midway_lock_t lockn
		  )
{
  u_long host, thread, tid, junk, incarn;
  pl_t *lock;
  char255 smsg;
  void midLockInvalidate();

  /* convert lockname to a pointer to the lock structure */
  lock = (pl_t *) lockn;
  TRACE_5("RELx", lockn, lock->state, lock->busy, lock->rdrs);

#if MIDWAY_FASTLOCKS == 0
  /* disable fastlocks */
#else
  /*
   * Check if the lock is held using a FASTLOCK
   */
  if ((lock->held == MIDWAY_LOCKHELD_RO) ||
      (lock->held == MIDWAY_LOCKHELD_RW)) {

#ifdef STATS
    if (lock->held == LOCK_RO) {
      INCR_STATS(midFastLockStats.relRO_lib, 1);
    } else {
      INCR_STATS(midFastLockStats.relRW_lib, 1);
    }
#endif /* STATS */

    lock->held = MIDWAY_LOCKHELD_FREE;		/* FASTLOCKS! */

    /* verify lock is in proper state */
    ASSERT (VALID_LOCKSTATES(lock));
    TRACE_3("Reff", lockn, lock->held);
    goto done;
  }
  if (lock->held == MIDWAY_LOCKHELD_CACHE_RO) {
    lock->held = MIDWAY_LOCKHELD_CACHE;

    INCR_STATS(midFastLockStats.relCA_lib, 1);
    ASSERT (VALID_CACHESTATES(lock));
    TRACE_3("Refr", lockn, lock->held);
    goto done;
  }

  /* 
   * Convert FASTLOCKS state to internal lock state
   */
  mid_lock_X_to_intern (lock);
#endif /* MIDWAY_FASTLOCKS */


#ifdef DEBUG_USER
  /* Check if lock is held */
  if (lock->busy == 0) {
    sprintf(smsg, "midway_lock_release: lock %lu not held\n", lockn);
    StatusMsg (smsg);
    mutex_unlock(&(lock->m));
    return;
  }
#endif

  /*
   * Perform actual release
   */
  lock->busy--;			/* note one guy down */
  if (!lock->busy) {		/* interesting case, else was reading */
    if (lock->rdrs == 0) {	/* not r/o , at least */
      if (!(lock->state & L_OWN)) { /* no r/o no own -> invalidate arrived */
	if (IS_DBG_LOCK_OPS) {
	  sprintf(smsg,"Release(%lu): !RO !OWN => Invalidate Reply", lockn);
	  StatusMsg(smsg);
	}
	INCR_STATS (midLockStats.releaseInval, 1);

	/* note could be acq r/w (or acq r/o, which would be an error) */
	/* send invalidate reply */
	if (mid_extq_pop( &(lock->ewrtrs), &host, &thread, &tid, &junk, &incarn)) {
	  /* error will robinson */
	  sprintf(smsg,"Release(%lu): rdrs==0, don't own, but no q'd thread to\
invalidate to!", lockn);
	  StatusMsg(smsg);
	  /* don't do anything -- essentially ignores invalidate */
#ifdef STOP_ON_ERROR
	  botch("Release error");
#endif
	} else {		/* got data from ewrtrs extq */
	  /* now really send reply */
	  midLockInvalidate( lock->lock, host, thread, tid, 1 /* internal */,
			    (char *) 0, 0);
	}
      } 
      else {			/* do own lock in exclusive mode */
	if (lock->nwriteq) {	/* internal writer waiting */
	  if (IS_DBG_LOCK_OPS) {
	    sprintf(smsg,"Release(%lu): OWN RW, writer waiting", lockn);
	    StatusMsg(smsg);
	  }
	  lock->busy++;		/* inc to ensure lock not stolen by others */
	  condition_signal(&(lock->iwritec)); /* let it go */
	  INCR_STATS (midLockStats.releaseNwriter, 1);
	} 
	else {		/* no writer, so demote to r/o */
	  if (IS_DBG_LOCK_OPS) {
	    sprintf(smsg,"Release(%lu): OWN RW, !waiting => demoting", lockn);

	    StatusMsg(smsg);
	  }
	  lock->rdrs++;		/* demote */
	  /* release current internal readers */
	  if (lock->nreadq > 0) {
	    lock->busy += lock->nreadq; /* bump, so lock will not be stolen */
	    lock->nreadq = 0;	/* no more queued (after the broadcast) */
	    condition_broadcast(&(lock->ireadc));
	    INCR_STATS (midLockStats.releaseNreader, 1);
	  }
	  if (lock->ewrite) {
#if PAGE_BASED
	    lock->rdrs--;	/* do not demote for writer - can't tolerate ext rdrs */
#endif
	    if (IS_DBG_LOCK_OPS) {
	      sprintf(smsg,"Release(%lu): (now)OWN RO, releasing external writer",lockn);
	      StatusMsg(smsg);
	    }
	    /* release external writer if any (will convert back to r/w) */
	    /* xfer thread must rdrs=0 */
	    lock->state = L_XFER; /* setup so incoming requests will not steal
				   from underneath us */
	    lock->xfersite = 0;	/* Notify others that we're inconsistent */
	    lock->xfertid = 0;
	    condition_signal(&(lock->ewritec));
	    INCR_STATS (midLockStats.releaseEwriter, 1);
	  } 
	  else {		/* no external writer */
	    if (lock->eread) { /* is external reader tho */
	      if (IS_DBG_LOCK_OPS) {
		sprintf(smsg,"Release(%lu): (now)OWN RO, releasing external readers",lockn);
		StatusMsg(smsg);
	      }
	      INCR_STATS (midLockStats.releaseEreader, 1);
	      /* don't bump busy here */
	      condition_signal(&(lock->ereadc)); /* release them */
	    }
	  }/* no ext writer */

	} /* no int writer */
      } /* do not own lock in exclusive mode */

    } /* not r/o case */ else {
      INCR_STATS (midLockStats.releaseNoRdrs, 1);
    }

    /*
     * Convert internal lock state back to external FASTLOCK state
     */
    mid_lock_X_to_extern(lock);

  } /* busy */
  else { /* busy is still > 0 */
    if (IS_DBG_LOCK_OPS) {
      sprintf(smsg, "Release(%lu): busy now %lu", lockn, lock->busy);
      StatusMsg(smsg);
    }
    INCR_STATS (midLockStats.releaseBusy, 1);
  }

done:
  /*
   * Release mutex guarding lock before returning
   */
  TRACE_6("relx", lockn, lock->held, lock->state, lock->busy, lock->rdrs);
  mutex_unlock(&(lock->m));

  return;
} /* midway_lock_relX */

/*
 ************************************************************************
 *	     m i d G e t L o c k I n c a r n T i m e			*
 *----------------------------------------------------------------------*
 * Given the number of a synchronization variable, get it's current time*
 * value (for now, used by CommScanner when resending invalidates)	*
 ************************************************************************
 */
u_long
midGetLockIncarnTime( lockn )
     midway_lock_t lockn;		/* name of lock in question */
{
  char255 smsg;
  u_long i;
  pl_t *lock;

  /* Don't lock for incarnation -- simply a read, so could get old data
     anyway.  In particular, we can get into a deadlock; the lock layer
     sometimes calls into the comm layer, which may lock the comm mutex.
     This is principally called by CommScanner, which holds the comm
     mutex.  Here is the situation I saw:

     T1: ProcessPacket -> midLockRequest [lock] -> CommAwaitingAck(block on c)
     T2: Scanner [c {comm lock}] -> GetLockIncarnTime (block on lock)
     */
  /* convert lockname to a pointer to the lock structure */
  lock = (pl_t *) lockn;

  i = lock->incarnation;

  return(i);
} /* midGetLockTime */


#if 0
/* no longer needed, I think. 31jan94,mjz */
/*
 ************************************************************************
 *		m i d G e t L o c k I n v R e q d			*
 *----------------------------------------------------------------------*
 * If we are required to invalidate due to a reply, get the aerdrs  if  *
 * any exist so that we get a more complete version of who we have to   *
 * wait for.  This should be replaced by a "fastInvalidate" procedure:  *
 * if a reply with an invalidate required for us shows up, then don't   *
 * wait for an additional InvalidateRequest, this should be enough.     *
 * Try to avoid sending a packet to ourselves, then, by just returning  *
 * "done" and an aerdrs struct.  If we can't do it, just return the     *
 * aerdrs struct.  However, this is the third solution I've thought of, *
 * and the first two didn't work, so try this simple (but non-optimal)  *
 * solution first. - 30jan94mjz						*
 ************************************************************************
 */
int *
  midGetLockInvReqd( lockn )
     u_long lockn;		/* name of lock in question */
{
  char255 smsg;
  u_long i;
  pl_t *lock;
  alllock_t *p;

  /* dereference name */
  /* first, find right locktab block */
  for (i=lockn, p=locktab; i>=ALLLCKSIZE; i-=ALLLCKSIZE, p=p->next) ;
  lock = p->addrs[i];		/* bingo */

  if (*(lock->aerdrs) & BAH_USED)
    return(lock->aerdrs);
  else
    return(0);
} /* midGetLockInvReqd */
#endif /*0*/



/*
 ************************************************************************
 *		   m i d E x t L o c k R e q u e s t			*
 *----------------------------------------------------------------------*
 * Request a lock from a remote location...				*
 ************************************************************************
 */
void
midExtLockRequest( lock, mode )
     pl_t		*lock;
     midway_acquire_t	 mode;
{
  m_lock_req_t	req;
  u_long	bestguess;
  u_long        cStart, cReqStart, cReqEnd, cEnd;
  u_long        tStart, tEnd;
  clock_t       save_incarnation;	/* for debugging */
  char255	smsg;

  if (IS_DBG_LOCK_EXTR) {
    sprintf(smsg,"entered midExtLockRequest, lock=%lx[%lx] mode=%lu incarn=%lu\n",
	     lock->lock,lock, mode, lock->incarnation);
    StatusMsg(smsg);
  }
  TRACE_4("EXLR", lock->lock, mode, lock->incarnation);

  /* read clock */
  STAT_GET_TIME (tStart);
  GET_CLOCK_VALUE (cStart);

  /* verify the FASTLOCKS have been disabled */
  ASSERT (lock->held == MIDWAY_LOCKHELD_OTHER);


  if (mode == LOCK_RW) {
    lock->state |= L_ACQRW;	/* note want exclusive copy */
    if (lock->state & L_ACQ) {	/* already acquiring read-only, must stall */
      condition_wait(&(lock->iupc), &(lock->m));
      /* Restore internal state from FASTLOCK state after waiting */
      mid_lock_X_to_intern (lock);
    }
    lock->iwrite = OUR_TID;	/* note we are about to stall in CommReq */
  } else {
    lock->state |= L_ACQ;	/* getting r/o copy */
  }
  req.incarnation = lock->incarnation; /* really only used for R/O requests */
  req.mode = mode;
  req.dataClock = lock->dataclock;
  bestguess = lock->bestguess;	/* steal from lock info */

  ASSERT(bestguess != Comm_index); /* Sending to ourself is Wrong. */

  save_incarnation = lock->incarnation;

  mutex_unlock(&(lock->m));	/* let others give it a whirl */

  /* send request to best guess location */
  if (IS_DBG_LOCK_EXTR) {
    StatusMsg("about to CommRequest to bestguess location");
  }
  GET_CLOCK_VALUE (cReqStart);

  CommRequest(lock->lock, M_LOCK_OP, bestguess, &req, sizeof(req));
  /* should fail-fast here... *mjz* */

  GET_CLOCK_VALUE (cReqEnd);

  /* NOTE: 'req' may be changed by CommRequest (actually by CommScanner),
     if an invalidate needs to be resent! */

  mutex_lock(&(lock->m));	/* lemme in */
  TRACE_6("EXlr", lock->lock, lock->held, lock->state, lock->busy, lock->rdrs);

  /* Restore internal state from FASTLOCK state after waiting */
  mid_lock_X_to_intern (lock);

  /* clean up any leftover other reader indications... */
  /* (since we got the lock, will not get any valid retransmissions
     asking for the lock we had in a previous incarnation) */
  bzero((char *) lock->aerdrs, midBITARRAYsize(Comm_hnum));
#ifdef NOTDEF
  /* 20mar92 yes we do; what if waiting reader while acquiring? */
  /* possibly delete any useful info?  I don't think so 30jun91 mjz */
  mid_extq_null(&(lock->erdrs));
  mid_extq_null(&(lock->ewrtrs));
#endif NOTDEF
  
  if (mode == LOCK_RO) {	/* r/o case, check for stalled r/w */
    lock->rdrs++;		/* will be reading */
    if (lock->nreadq > 0) {
      lock->busy += lock->nreadq;	/* inc by those we will now let go */
      lock->nreadq = 0;
      condition_broadcast(&(lock->ireadc)); /* let internal readers go */
    }
    /* external too ? *mjz* */
    if (lock->eread) {		/* is external reader */
      /* 1aug92mjz: don't do this--we dont' do it in any other case, and
	 we end up with busy bumped one too many times.  see commment at top*/
      /*lock->busy++;	*/	/* inc so not stolen before thread runs */
      condition_signal(&(lock->ereadc));
    }
    if ((lock->state & L_ACQINV) /* invalidate pending */ ||
	(lock->state & L_ACQRW)  /* stalled r/w request */){
      /* turn off invalidate pending and datain indications (if any) */
      lock->state &= (L_XACQINV & L_XDATAIN);
      condition_broadcast(&(lock->iupc)); /* let 'em go */
    }
    if (!(lock->state & L_ACQRW)) { /* no stalled r/w request */
      lock->state = 0;		/* done with acquiring */
    }
    /* else don't disturb state bits */
  }  /* case r/o */
  else {			/* was r/w */
    lock->state = L_OWN;	/* note we own now */
    ASSERT(lock->rdrs == 0);	/* anything else is an error */
    lock->incarnation++;	/* bump-o-matic */
    lock->iwrite = 0;		/* clear, we're back */
    lock->bestguess = Comm_index; /* is a really good guess, too... */
    TRACE_4("EXLi", lock, lock->dataclock, lock->incarnation);

#ifdef PAGE_BASED
    midLockIncrDiffs (lock);
#endif

#ifdef UPDATE_HINTS
    lock->bgtime = lock->incarnation;
    
    /* Add lock to hint table */
    mutex_lock(hintMutex);
    hintTable[hintIndex] = lock;
    hintIndex = (hintIndex + 1) % HINT_TABLE_SIZE;
    mutex_unlock(hintMutex);
#endif
  }
  lock->busy++;			/* now someone else using */

  GET_CLOCK_VALUE (cEnd);
  STAT_GET_TIME (tEnd);
  TRACE_6("exlr", lock->lock, lock->held, lock->state, lock->busy, lock->rdrs);

#ifdef LTIMES 
  if ((tEnd < tStart) ||
      (cEnd < cStart) ||
      (cReqEnd < cReqStart)) {
    INCR_STATS (midLockStats.extReqInvalidTime, 1);
  } else {
    tEnd -= tStart;
    cEnd -= cStart;
    cReqEnd -= cReqStart;
    INCR_STATS (midLockStats.extReqRequests, 1);
    INCR_STATS (midLockStats.extReqClockTotal, cEnd);
    INCR_STATS (midLockStats.extReqReqClockTotal, cReqEnd);
    INCR_STATS (midLockStats.extReqTimeTotal, tEnd);
    if (tEnd > midLockStats.extReqTimeMax)
      midLockStats.extReqTimeMax = tEnd;
    if (cEnd > midLockStats.extReqClockMax) 
      midLockStats.extReqClockMax = cEnd;
    if (cReqEnd > midLockStats.extReqReqClockMax) 
      midLockStats.extReqReqClockMax = cReqEnd;
    
    cReqEnd /= STAT_TIME_PER_BUCKET;
    if (cReqEnd > STAT_TIME_TABLE_SIZE)
      cReqEnd = STAT_TIME_TABLE_SIZE;
    if (mode == LOCK_RO) { 
      INCR_STATS (midLockStats.extReqROtimeP[ cReqEnd ], 1);
    } else {
      INCR_STATS (midLockStats.extReqRWtimeP[ cReqEnd ], 1);
    }
  }
#endif

} /* midExtLockRequest */

/*
 ************************************************************************
 *		     m i d L o c k R e q u e s t			*
 *----------------------------------------------------------------------*
 * take request from external source					*
 ************************************************************************
 */
#if PAGE_BASED
# define PREP_RO_XFER	lock->ROxfers++;
# define FIN_RO_XFER	{lock->ROxfers--; \
	if ((lock->ROxfers == 0) && (lock->state & L_XFER) && \
	    (lock->xfersite != 0) && (lock->ewrite)) { \
	  condition_broadcast(&(lock->ewritec)); \
	}}
# define PREP_RO_XFER_ONLY \
	if (msg->mode == LOCK_RO) { \
	  PREP_RO_XFER; \
	} else { \
	  PREP_RW_XFER; \
	}
# define FIN_RO_XFER_ONLY \
	if (msg->mode == LOCK_RO) { \
	  FIN_RO_XFER; \
	}
# define PREP_RW_XFER	ASSERT(lock->ROxfers == 0)
#else
# define PREP_RO_XFER
# define FIN_RO_XFER
# define PREP_RO_XFER_ONLY
# define FIN_RO_XFER_ONLY
# define PREP_RW_XFER
#endif /*PAGE_BASED*/

comm_request_t
midLockRequest( lockn, host, thread, tid, retrans, msg, msglen, fti )
     u_long lockn;		/* name of lock this is all about */
     u_long host;		/* host requesting transfer */
     u_long thread;		/* thread@host requesting transfer */
     u_long tid;		/* transaction id for this request */
     int retrans;		/* non zero if this is a retransmission */
     m_lock_req_t *msg;		/* request specific info */
     u_long msglen;		/* length of data in msg */
     u_int  fti;		/* for flow times (debugging) */
{
  char255 smsg;
  u_long i;
  pl_t *lock;
  u_long s_host, s_thread, s_tid;
  clock_t s_clock, s_incarn;
  int empty;			/* to check for empty q's */
  comm_request_t rc = M_R_UNCHANGED; /* don't change transaction state */


  /* validate the request */
  if (msglen != sizeof(m_lock_req_t)){ /* Insanity */
    sprintf(smsg,"External lock request had bad length %lu != %d; dropping",
	    msglen, sizeof(m_lock_req_t));
    StatusMsg(smsg);
    return(M_R_DROPPED);	/* for now, drop message *mjz* */
  }

#if FLOW_TIMES
  ft_times[fti][FT_LOCKREQ] = FT_READ_TIMER;
#endif

  /* convert lockname to a pointer to the lock structure */
  lock = (pl_t *) lockn;

  if (IS_DBG_LOCK_EXTR) {
    sprintf(smsg,"External lock req from %lu(%lu)@%lu for %lx[%lx]@%lu mode %lu incarn %lu",
	    host,thread,tid,lockn,lock,msg->dataClock,msg->mode,msg->incarnation);
    StatusMsg(smsg);
  }

  if (host == Comm_index) {	/* is a request from myself */
/***
    StatusMsg("midLockRequest: dropping external request from myself.");
***/
    INCR_STATS (midLockStats.requestsFromSelf, 1);
    return(M_R_DROPPED);
  }

  /*
   * Acquire mutex guarding lock structure
   */
  mutex_lock(&(lock->m));	/* ensure consistent access */

  /* 
   * Convert FASTLOCKS state to internal lock state
   */
  mid_lock_X_to_intern (lock);

  TRACE_6("LKRQ", lock->lock, msg->mode, lock->state, lock->busy, lock->rdrs);

  /* first, check for inconsistent transfer state.  Wait for improvement. */
  while ((lock->state & L_XFER) && (lock->xfersite == 0)) {
    if (IS_DBG_LOCK_EXTR) {
      StatusMsg("midLockReq: XFERRING, and xfersite not consistent.  Waiting.");
    }
    condition_wait(&(lock->xferc), &(lock->m));
    if (IS_DBG_LOCK_EXTR) {
      StatusMsg("midLockReq: back from xferc condition--xfersite consistent");
    }

    /* Restore internal state from FASTLOCK state after waiting */
    mid_lock_X_to_intern (lock);
  }


  /*
   * 8 sep 92: new data from comm -- whether we have ever processed this
   * request before (M_DOAGAIN).  If the state indicates we saw this
   * lock again, then this cannot be a valid retransmission -- drop it.
   * In return, we will tell the comm layer whether we queued, forwarded,
   * or replied to this message.  Only replied ones will come back a second
   * time, and then with the doagain flag... so if we got the lock back,
   * or transferred it somewhere else, then this is very old, and it
   * can be punted.
   * 16 sep 92: forwarded ones come back here too -- 'cause we might
   * get the lock back.
   */

  /* note: due to this check here, the cases below can be made simpler,
     since retransmissions could be handled here...  */
  if (retrans & M_DOAGAIN) {	/* 2nd (or 3rd...) time around */
    /* filter out requests with obvious problems */
    /* ones that should not change state */
    if (
	(lock->state & L_XFER)
	) {
      goto done;		/* return unchanged */
    }
    /* ones that are old news */
    if (
	/* ro request             not transferred to him */    /*OR*/
	((msg->mode == LOCK_RO) && (!(midTestBit(lock->aerdrs,host)))) ||
	/* rw request    and...*/
	((msg->mode == LOCK_RW) &&
	 /* we own (so ownership came back) or we don't own but didn't xfer */
	 ((lock->state & L_OWN) || ((lock->state == 0) && ((lock->rdrs >0)
				    || (lock->xfersite != host)))))
	) {
      rc = M_R_DROPPED;		/* drop the baby */
      goto done;
    }
    /* otherwise attempt to do it again; we should be in a safe state */
  } /* end 1st time through */

  /*
   * CASES - do different things depending on lock's current status
   */


  /* transferring */
  if (lock->state & L_XFER) { 	/* already transferring */
    if (IS_DBG_LOCK_EXTR) {
      sprintf(smsg,"extLockReq(%lu): in L_XFER case",lockn);
      StatusMsg(smsg);
    }
    if (host == lock->xfersite) { /* already xferring to him! */
      /* ReplyPending */
      CommAck(host, thread, tid, lockn, 0, 0); /* ReplyPending */
      /* will resend if he requests again after L_XFER turned off */
    } else {			/* xferring somewhere else */
      /* forward */
      CommForward(lockn, M_LOCK_OP, lock->xfersite,
		  host, thread, tid, retrans&M_RETRANS, msg, msglen);  /* *mjz* */
      /*FWD rc = lock->xfersite;*/	/* where forwarded to */
      rc = M_R_DOAGAIN;
    }
  } /* case L_XFER */

  /* acquiring */
  else if (lock->state & (L_ACQ|L_ACQRW)) {
    if (IS_DBG_LOCK_EXTR) {
      sprintf(smsg,"extLockReq(%lu): in L_ACQ case",lockn);
      StatusMsg(smsg);
    }
    /* hmmm... this evening (30jun91) I got into a circular dependency:
     * one processor asked, the other gave it and immediately asked again
     * for itself.  However, the request and reply got shuffled by the
     * net, and the reply was dropped as being out of date.  The original
     * processor asked again, and got queued.  The second asked again,
     * and also got queued.  After that they just traded ACKs.  (Both
     * were in ACQ_RW mode...  both had each other queued... not pretty.)
     *
     * Today's solution is to recognize this case, and resend the reply.
     * Note that you *must* verify the old transaction ID to distinguish
     * an old request (which we must resend) from a new request (which
     * we must queue).
     *
     * 10jul91 - this happens in 3-way case still... 'cause could
     * have them pointing to each other, but not meet the criteria
     * below.  today's hack is to alow 1st time requests to queue up,
     * but retries that are not in the queue are forwarded.  This will
     * probably end up looping around (unless we hit this n-way case,
     * when the condition below will be satisfied by the retransmission,
     * breaking the loop) -- so we've got to detect a complete loop,
     * and break that by dropping the packet (it's better to resend after
     * a time out than flood the net with these retries going in circles
     * until the right one is passed to the right machine... worse for
     * latency, though).
     */
    if ((lock->xfersite == host) && (lock->bestguess == host) &&
	(lock->xfertid == tid)) {
      /* most likely that we are in the circular dependency case */
      if (IS_DBG_LOCK_WNG) {
	StatusMsg("Circular Dependency case!");
      }
      if ((msg->mode == LOCK_RW) && (lock->incarnation == msg->incarnation)) {
	/* most likely was upgrade, and no need to go through dbits */
	msg->mode |= 0x80000000;
      }
      lock->busy++;		/* for Xfer */
      PREP_RO_XFER_ONLY;
      midLockXfer(lock, msg->dataClock, msg->incarnation, host, thread, tid, msg->mode, fti);
      FIN_RO_XFER_ONLY;
      rc = M_R_DONE;
      goto done;
    }

    if ((msg->mode == LOCK_RW) && ((lock->state & L_ACQRW) == 0)) {
      /* R/W request, and we are acquiring R/O */
      CommForward(lockn, M_LOCK_OP, lock->bestguess,
		  host, thread, tid, retrans&M_RETRANS, msg, msglen);
      /* FWD rc = lock->bestguess; */
      rc = M_R_DOAGAIN;		/* we need to see this packet again */
    } else {
      /* R/O request or R/W req and acquiring R/W -- queue, will sat event. */
      extq_t *q;
      condition_t c;
      u_long *tidholder;	/* holds tid of waiting thread */
      
      if (msg->mode == LOCK_RW) {	/* write request */
	q = &(lock->ewrtrs);
	c = &(lock->ewritec);
	tidholder = &(lock->ewrite);
      } else {			/* read request */
	q = &(lock->erdrs);
	c = &(lock->ereadc);
	tidholder = &(lock->eread);
      }

      
      if(mid_extq_dup(q, host, thread, tid)) {
	CommAck(host, thread, tid, lockn, 0, 0); /* ReplyPending */
	/* since we never time out right now, this is just contributing
	   to net pollution -- mjz27jun91*/
	/* SANITY: since we catch requests in progress by the
	   tid, we should never get here. 17feb94mjz*/
	botch("midLockRequest: duplicate request 1");
	goto done;		/* dup request -- punt */
      }
      
      /* to break n-way circular dependency, forward if retrans rather
       * than queuing, unless we will forward back to the owner.
       */
      if (retrans & M_RETRANS) {
	/* NOTE: this includes Forwarded retransmissions.  We never want to
	   queue those, and we don't.  --17sep92mjz */
	if (((lock->state & (L_ACQRW|L_ACQ)) == L_ACQRW) &&
	    (msg->mode == LOCK_RO) &&	/* r/o request */
	    (lock->iwrite != 0) &&	/* should always be true */
	    /* only awaiting ack; not any data! (so we're consistent) */
	    CommAwaitingAck(lock->iwrite, host) && 
	    (msg->incarnation < lock->incarnation)) {
	  /* this is to break r/o circular dependency.  Problem is if
	     lock is xferred, but r/o requestor didn't get reply.
	     He's hung waiting for reply, and we're hung waiting
	     for an invalidate ack.  Should give him reply.
	     ((future: to cut down on communication, perhaps auto-invalidate
	     and queue request???))  Gotta watch out for case where
	     invalidate REPLY got trashed, and we get a new R/O
	     request almost immediately.  Don't want to respond to
	     NEW request yet!
	   */
	  lock->busy++;		/* for Xfer */
	  PREP_RO_XFER;
	  midLockXfer(lock, msg->dataClock, msg->incarnation, host, thread, tid, msg->mode, fti);
	  FIN_RO_XFER;
	  rc = M_R_DONE;
	  goto done;
	}
	else if (host != lock->bestguess) { /* not all the way around loop */
	  CommForward(lockn, M_LOCK_OP, lock->bestguess, host,
		      thread, tid, retrans&M_RETRANS, msg, msglen);
	  rc = 0;		/* allow to be processed again XXX*/
	  /* what about saying forwarded?  problem is, we may be the ones
	     to satisfy request, and forwarding will go round and round */
	  /* BUT THEN, SATISFY SOMETHING WE SHOULD NOT? */
	} else {		/* host == lock->bestguess */
	  /* don't ack, will retrans faster [danger if max # retrans] */
	  rc = 0; /* XXX */
	}
	goto done;		/* break loop */
      }

      if(mid_extq_add(q, host, thread, tid, msg->dataClock, msg->incarnation)) {
	/* no memory */
	sprintf(smsg,"midLockReq(%lu): extq_add #1 out of memory...",lockn);
	StatusMsg(smsg);
      }
      if(*tidholder) {
	/* already transfer thread waiting */
	if (IS_DBG_LOCK_EXTR) {
	  sprintf(smsg,"mLockReq(%lu): already xfer thread waiting.",lockn);
	  StatusMsg(smsg);
	}
	/* since already thread waiting, and we are queued, leave as
	   INPROCESS */
	goto done;		/* bail out! */
      }
      else {
	*tidholder = OUR_TID;
	do {
	  if (IS_DBG_LOCK_EXTR) {
	    sprintf(smsg, "mLockReq(%lu): about to condition_wait",lockn);
	    StatusMsg(smsg);
	  }
	  condition_wait(c, &(lock->m));
	  if (IS_DBG_LOCK_EXTR) {
	    sprintf(smsg,"mLockReq(%lu): back from L_ACQ condition_wait",lockn);
	    StatusMsg(smsg);
	  }
	  if (*tidholder != OUR_TID) {
	    if (IS_DBG_LOCK_WNG) {
	      sprintf(smsg,"mLockReq(%lu)[ACQ]: tidholder not us -- we were transferred",lockn);
	      StatusMsg(smsg);
	    }
	    INCR_STATS(midLockStats.extReqXferred,1);
	    /* JUST IN CASE: (should have been done by popping us off Q) */
	    rc = M_R_DOAGAIN;	/* mjz17feb94 */
	    goto done;		/* BAIL - we were transferred */
	  }

	  /* Restore internal state from FASTLOCK state after waiting */
	  mid_lock_X_to_intern (lock);

	  if ((msg->mode == LOCK_RO) && (lock->state & L_XFER)) {
	    /* external writer slipped in before we could run */
	    if (IS_DBG_LOCK_WNG) {
	      sprintf(smsg,"mLockReq(%lu)[ACQ]: reader & L_XFER set, bailing out.",lockn);
	      StatusMsg(smsg);
	    }
	    INCR_STATS(midLockStats.extReqXferred, 1);
	    /* if EXT writer slipped in, then we might not have our
	       state changed to do-over... 17feb94mjz*/
	    rc = M_R_DOAGAIN;
	    goto done;
	  }
	} while ((msg->mode == LOCK_RO) && (lock->state == L_OWN) && (lock->rdrs == 0));
	*tidholder = 0;		/* no longer waiting */
	/* cause problems?  Say another request comes in while we are
	   transferring... wouldn't that just set another thread to
	   wait?  Ack!  Think about this some more XXX --mjz 
	   No... we hold the lock; if another were to come in during XFER,
	   XFER would be set, and it wouldn't Q. --mjz a few hours later
	   */
	/* pop ourselves off of q.  Ensure ourselves?  Just take first? */
	if(mid_extq_pop(q, &s_host, &s_thread, &s_tid, &s_clock, &s_incarn)) {
	  /*error -> no more waiters -> we were transferred */
	  /* 30 mar 92, mjz -> no longer true??? XXX */
	  if (IS_DBG_LOCK_WNG) {
	    sprintf(smsg,"mLockReq(%lu): nothing on Q.  we were transferred?",lockn);
	    StatusMsg(smsg);
	  }
	  INCR_STATS(midLockStats.extReqXferred, 1);
	  rc = M_R_DOAGAIN;
	  goto done;		/* nothing to do */
	}
	/* We don't currently continue with the request, so we have to
	 * satisfy everything here.  We also have to pull off all the
	 * requests if this is a read request.  Bletch.
	 */
	if ((s_host != host) || (s_thread != thread) || (s_tid != tid) ||
	    (s_clock != msg->dataClock)) {
	  sprintf(smsg,"extReq: {ACQ} mismatched deq from (mode)%d extq; expect [h,th,tid,ck] %lu.%lu.%lu.%lu got %lu.%lu.%lu.%lu L=%lu",
		  msg->mode, host,thread,tid,msg->dataClock,s_host,s_thread,
		  s_tid,s_clock,lockn);
	  StatusMsg(smsg);
	  /* use the deq'd ones... */
	  /* XXX perhaps also set trantab state to don for one we will not do?
	     - 17feb94 mjz */
	  CommSetTranState(s_host, s_thread, s_tid, M_R_DOAGAIN);
	  if (msg->mode == LOCK_RW) {
	    if (s_clock != 0) {
	      INCR_STATS (midLockStats.xferEwrtLost, 1);
	    } else {
	      INCR_STATS(midLockStats.xferEwrILost, 1);
	    }
	  }
	  while (!mid_extq_pop(q, &s_host, &s_thread, &s_tid, &s_clock, &s_incarn)){
	    CommSetTranState(s_host, s_thread, s_tid, M_R_DOAGAIN);
	    if (msg->mode == LOCK_RW) {
	      if (s_clock != 0) {
		INCR_STATS (midLockStats.xferEwrtLost, 1);
	      } else {
		INCR_STATS(midLockStats.xferEwrILost, 1);
	      }
	    }
	  }
	  rc = M_R_DOAGAIN;
	  goto done;		/* nothing to do */
	}
	/* know rdrs has been inc'd.. */
	/* but, in fact, it could be that another site has come in here and gotten
	   a read-only copy.  Therefore, it could be > 1. Just check for real bogosity.*/
#if PAGE_BASED
	if (lock->rdrs < 0) {	/* pb sets to 0 so no rdrs can sneak in */
#else
	if (lock->rdrs <=  0) {	/* bogus *//*dbg, consistency */
#endif
	  sprintf(smsg,"extLockReq(%lu): rel from ACQ cond, and lock->rdrs <= 0 (%d)",
		  lockn,lock->rdrs);
	  StatusMsg(smsg);
	  lock->rdrs = 1;	/* that should do it */
	}

	/* ..since we're here as the result of a condition_signal */
	PREP_RO_XFER_ONLY;
	do {
	  if (msg->mode == LOCK_RW) {	/* exclusive request */
	    lock->state |= L_XFER; /* now transferring */
	    lock->state &= ~L_OWN; /* just in case (should be done since signalled) */
	    lock->xfersite = s_host; /* where to */
	    lock->xfertid = s_tid;
	    condition_broadcast(&(lock->xferc)); /* xfersite now consistent */
	    /* let other internal readers go? Ha! --mjz XXX */
	    lock->bestguess = s_host; /* where it went */
	    lock->bgtime = lock->incarnation;
	  } else {		/* read request */
	    lock->rdrs++;	/* new external reader */
	    midSetBit(lock->aerdrs, s_host); /* note they have */
	    midSetBit(lock->aerdrs, 0);	/* note aerdrs valid */
	  }
	  lock->busy++;		/* note busy for duration of transfer */
	  midLockXfer(lock, s_clock, s_incarn, s_host, s_thread, s_tid, msg->mode, fti);
	  if (msg->mode == LOCK_RW) {	/* exclusive request */
	    lock->state &= ~L_XFER;	/* and we're gone! */
	    lock->rdrs = 0;	/* no longer own... XXX */
	    rc = M_R_DONE;
	  } else {		/* read request */
	    CommSetTranState(s_host, s_thread, s_tid, M_R_DONE);
	    /* must release *all* external readers */
	    empty = mid_extq_pop(q, &s_host, &s_thread, &s_tid, &s_clock, &s_incarn);
	  }
	} while((msg->mode == LOCK_RO) && (empty == 0));
	/*       read request   AND  still something to do */
	FIN_RO_XFER_ONLY;

      } /* done!  will continue with request.. should have no problems,tho */
    } /* can't forward */
  } /* case L_ACQ */

  /* own */
  else if (lock->state & L_OWN) {

    if (lock->rdrs == 0) {

      /* OWN, RW */

      if (IS_DBG_LOCK_EXTR) {
	sprintf(smsg,"extLockReq(%lu): in L_OWN, r/w case",lockn);
	StatusMsg(smsg);
      }

      /* own r/w */
      if (msg->mode == LOCK_RW) {	/* and r/w request */
	if (lock->busy) {		/* if busy, must queue */
	  /* I never expect not to be busy, since Release currently
	     demotes locks! -4feb91 mjz */
	  /* queue */ /* when return, will own r/w & ~busy */
	  /* must note all here, since may need to REPLY_PENDING */
	  if (mid_extq_dup(&(lock->ewrtrs), host, thread, tid)) {
	    /* already in Q, don't do it again! */
	    CommAck(host, thread, tid, lockn, 0, 0); /* ReplyPending */
	    botch("midLockRequest: duplicate request 2");
	    goto done;
	  }

	  if ((retrans & (M_FORWARDED|M_RETRANS)) == (M_FORWARDED|M_RETRANS)){
	    /* don't queue a forwarded retransmission -- could be old. */
	    /* Have requestor ask directly if he still wants it */
	    CommReply(M_LOCK_NEWOWN, host, thread, tid, lockn, 0 /*lastbpack*/,
		      0 /*packnum*/, NULL_ADDR_VECTOR /*numack*/, (char *)&(lock->incarnation),
		      sizeof(lock->incarnation), C_NOFREE);
	    INCR_STATS(midLockStats.extReqNewOwn,1);
	    rc = M_R_DOAGAIN;
	    goto done;
	  }

	  if(mid_extq_add(&(lock->ewrtrs), host, thread, tid, msg->dataClock, msg->incarnation)) {
	    /* no memory *mjz* */
	    sprintf(smsg,"midLockReq(%lu): extq_add #2 out of memory...",lockn);
	    StatusMsg(smsg);
	  }
	  if (lock->ewrite) {	/* already transfer thread waiting */
	    /* barf *mjz* */
	    if (IS_DBG_LOCK_EXTR) {
	      sprintf(smsg,"mLockReq(%lu): already transfer thread waiting",lockn);
	      StatusMsg(smsg);
	    }
	    goto done;
	  }
	  else {
	    lock->ewrite = OUR_TID;
	    if (IS_DBG_LOCK_EXTR) {
	      sprintf(smsg,"mLockReq(%lu): about to condition_wait",lockn);
	      StatusMsg(smsg);
	    }
	    condition_wait(&(lock->ewritec), &(lock->m));
	    if (IS_DBG_LOCK_EXTR) {
	      sprintf(smsg,"mLockReq(%lu): back from OWN,r/w condition_wait",lockn);
	      StatusMsg(smsg);
	    }
	    if (lock->ewrite != OUR_TID) {
	      sprintf(smsg,"extLockReq(%lu): rel from ewritec, ewrite %lu != OURTID %lu",
		      lockn,lock->ewrite, OUR_TID);
	      StatusMsg(smsg);
	      INCR_STATS(midLockStats.extReqXferred,1);
	      rc = M_R_DOAGAIN;
	      goto done;
	    }

	    /* Restore internal state from FASTLOCK state after waiting */
	    mid_lock_X_to_intern (lock);

	    lock->ewrite = 0;
	    /* pop ourselves off the queue */
	    if (mid_extq_pop(&(lock->ewrtrs), &s_host, &s_thread, &s_tid, &s_clock, &s_incarn)) {
	      /* empty.  Bogus */
	      sprintf(smsg,"extLockReq(%lu): rel from ewritec, and noone on Q!",lockn);
	      StatusMsg(smsg);
	      INCR_STATS(midLockStats.extReqXferred, 1);
	      rc = M_R_DOAGAIN;
	      goto done;
	    }
	    /* could verify that it's us for sanity... *mjz* */
	    if ((s_host != host) || (s_thread != thread) || (s_tid != tid) ||
		(s_clock != msg->dataClock)) {
	      sprintf(smsg,"extReq: {OWN,RW} mismatched deq from (mode)%d extq; expect [h,th,tid,ck] %lu.%lu.%lu.%lu got %lu.%lu.%lu.%lu L=%lu",
		      msg->mode, host,thread,tid,msg->dataClock,
		      s_host,s_thread,s_tid,s_clock, lockn);
	      StatusMsg(smsg);
	      /* use the deq'd ones... */
	      /* ...and set other one to DOAGAIN? -17feb94mjz*/
	    }
	  }
	} /* busy */ 
	else { /* !busy */
	  if ((retrans & (M_FORWARDED|M_RETRANS)) == (M_FORWARDED|M_RETRANS)){
	    /* don't act on a forwarded retransmission -- could be old. */
	    /* Have requestor ask directly if he still wants it */
	    CommReply(M_LOCK_NEWOWN, host, thread, tid, lockn, 0 /*lastbpack*/,
		      0 /*packnum*/, NULL_ADDR_VECTOR /*numack*/, 
		      (char *)&(lock->incarnation), sizeof(lock->incarnation), C_NOFREE);
	    INCR_STATS(midLockStats.extReqNewOwn, 1);
	    rc = M_R_DOAGAIN;
	    goto done;
	  }
	  /* condition_signal'rs demote lock already, don't do it then */
	  if (lock->rdrs == 0) {
	      lock->rdrs = 1;	/* also demote lock *//* should == ++ */
	  }
	  s_host = host;
	  s_thread = thread;
	  s_tid = tid;
	  s_clock = msg->dataClock;
	}
	lock->busy++;		/* note busy for duration of xfer */
	/* If we had to wait because lock was busy, the following is a valid
	   scenerio:  Release (sets L_XFER) Acquire R/O (sets L_ACQ), then
	   we run.  We do not want to stomp on the L_ACQ.  What we need to
	   do is ensure L_OWN is off now, if we did not queue. */
	lock->state |= L_XFER;
	lock->state &= ~L_OWN;
	lock->xfersite = s_host;
	lock->xfertid = s_tid;
	condition_broadcast(&(lock->xferc)); /* xfersite now consistent */

	if (lock->nreadq > 0) {
	  lock->busy += lock->nreadq; /* bump by # of internal readers-to-be */
	  lock->nreadq = 0;
	  condition_broadcast(&(lock->ireadc)); /* let internal rdrs go */
	}
	lock->bestguess = s_host;	/* need here for internal req */
	lock->bgtime = lock->incarnation;
	PREP_RW_XFER;
	midLockXfer(lock, s_clock, s_incarn, s_host, s_thread, s_tid, msg->mode, fti);
	lock->state &= ~L_XFER;	/* don't know anymore  */
	lock->rdrs = 0;		/* XXX */
	rc = M_R_DONE;		/* we satisfied request */
      } /* rw request (own&)*/
      else {			/* and r/o request */
	if (IS_DBG_LOCK_EXTR) {
	  sprintf(smsg,"extLockReq(%lu): own r/w and r/o request",lockn);
	  StatusMsg(smsg);
	}
	if (lock->busy) {	/* mutator at work */
	  if (IS_DBG_LOCK_EXTR) {
	    sprintf(smsg,"extLockReq(%lu): lock busy - checking q's",lockn);
	    StatusMsg(smsg);
	  }
	  /*
	   * test to see if WE have an invalidate outstanding.  
	   * How?  If busy is set,
	   * then we own and someone is inside... EXCEPT if aerdrs has any
	   * valid information (we would have cleared it when the lock
	   * was acquired or upgraded).  SHOULD WALK THRU MATT --3jul92mjz
	   */
	  if (midTestBit(lock->aerdrs,host) &&
	      msg->incarnation < lock->incarnation) {
	    /* OK, if this is set, then we had transferred, and are trying
	       to upgrade.  Unfortunately, we cannot upgrade until we get
	       the invalidate, and we can't get the invalidate without
	       having the other guy get the response he wanted (he can't
	       invalidate something he doesn't have)
	       Test incarnation numbers too to ensure we just didn't
	       lose the ACK and this is a new request.

	       THIS IS A RETRANSMISSION.  Resend the reply.
	     */
	    if (IS_DBG_LOCK_EXTR) {
	      sprintf(smsg,"extLockReq(%lu): busy, but aerdrs for host %lu is set. Assuming retransmission & replying.",
		      lockn, host);
	      StatusMsg(smsg);
	    }
	    lock->busy++;	/* since Xfer will decrement */
	    PREP_RO_XFER;
	    midLockXfer(lock, msg->dataClock, msg->incarnation, host, thread, tid, msg->mode, fti);
	    FIN_RO_XFER;
	    rc = M_R_DONE;
	    goto done;
	  }
	       
	  if (mid_extq_dup(&(lock->erdrs), host, thread, tid)) {
	    /* already on Q, don't do it again! */
	    CommAck(host, thread, tid, lockn, 0, 0); /* ReplyPending */
	    botch("midLockRequest: duplicate request 3");
	    goto done;
	  }
	  if ((retrans & (M_FORWARDED|M_RETRANS)) == (M_FORWARDED|M_RETRANS)){
	    /* don't queue a forwarded retransmission -- could be old. */
	    /* Have requestor ask directly if he still wants it */

	    /*?? a problem for r/o guys? */
	    CommReply(M_LOCK_NEWOWN, host, thread, tid, lockn, 0 /*lastbpack*/,
		      0 /*packnum*/, NULL_ADDR_VECTOR /*numack*/, 
		      (char *)&(lock->incarnation),
		      sizeof(lock->incarnation), C_NOFREE);
	    INCR_STATS(midLockStats.extReqNewOwn, 1);
	    rc = M_R_DOAGAIN;
	    goto done;
	  }
	  if(mid_extq_add(&(lock->erdrs), host, thread, tid, msg->dataClock, msg->incarnation)) {
	    /* no memory *mjz* */
	    sprintf(smsg,"midLockReq(%lu): extq_add #3 out of memory...",lockn);
	    StatusMsg(smsg);
	  }
	  if (lock->eread) {	/* already transfer thread waiting */
	    if (IS_DBG_LOCK_EXTR) {
	      sprintf(smsg,"mLockReq(%lu): already transfer thread waiting",lockn);
	      StatusMsg(smsg);
	    }
	    goto done;
	  }
	  else {
	    lock->eread = OUR_TID;
	    do {		/* loop if intervening local writer grabs */
	      if (IS_DBG_LOCK_EXTR) {
		sprintf(smsg,"mLockReq(%lu): about to condition_wait",lockn);
		StatusMsg(smsg);
	      }
	      condition_wait(&(lock->ereadc), &(lock->m));
	      if (IS_DBG_LOCK_EXTR) {
		sprintf(smsg,"mLockReq(%lu): back from OWN,r/o condition_wait",lockn);
		StatusMsg(smsg);
	      }
	      /* COULD BE XFERRED with write request! *mjz* */
	      if (lock->eread != OUR_TID) {
		if (IS_DBG_LOCK_WNG) {
		  sprintf(smsg,"mLockReq(%lu): tids differ, must have transferred",lockn);
		  StatusMsg(smsg);
		}
		INCR_STATS(midLockStats.extReqXferred, 1);
		rc = M_R_DOAGAIN;
		goto done;		/* assume transferred */
	      }

	      /* Restore internal state from FASTLOCK state after waiting */
	      mid_lock_X_to_intern (lock);

	      if (lock->state & L_XFER) { /* external writer sneaked in */
		/* ... but he hasn't finished xferring, else tid would have
		   been different */
		if (IS_DBG_LOCK_WNG) {
		  sprintf(smsg,"mLockReq(%lu): L_XFER set, we are reader, so exiting",lockn);
		  StatusMsg(smsg);
		}
		  /* tid will be set to zero when queue is drained. */
		INCR_STATS(midLockStats.extReqXferred, 1);
		rc = M_R_DOAGAIN;
		goto done;	/* bye */
	      }
	    } while ((lock->state & L_OWN) && (lock->rdrs == 0)); /*OWN R/W*/
	    lock->eread = 0;
	    if(mid_extq_pop(&(lock->erdrs), &s_host, &s_thread, &s_tid, &s_clock, &s_incarn)) {
	      /* error -> we were transferred */
	      if (IS_DBG_LOCK_WNG) {
		sprintf(smsg,"mLockReq(%lu): nothing on Q",lockn);
		StatusMsg(smsg);
	      }
	      INCR_STATS(midLockStats.extReqXferred, 1);
	      rc = M_R_DOAGAIN;
	      goto done;
	    }
	    if ((s_host != host) || (s_thread != thread) || (s_tid != tid) ||
		(s_clock != msg->dataClock)) {
	      sprintf(smsg,"extReq: {OWN,RO} mismatched deq from (mode)%d extq; expect [h,th,tid,ck] %lu.%lu.%lu.%lu got %lu.%lu.%lu.%lu L=%lu",
		      msg->mode, host,thread,tid,msg->dataClock,
		      s_host,s_thread,s_tid,s_clock,lockn);
	      StatusMsg(smsg);
	      /* use the deq'd ones... */
	    }

	  } /* no transfer thread waiting*/
	} /* busy */
	else {			/* not busy */
	  if (IS_DBG_LOCK_EXTR) {
	    sprintf(smsg,"extLockReq(%lu): not busy.  copying...",lockn);
	    StatusMsg(smsg);
	  }
	  if (lock->rdrs == 0) {
	      lock->rdrs = 1;	/* demote lock.  Already demoted if */
	  }
	  /* ..condition_signal'd */
	  s_host = host;
	  s_thread = thread;
	  s_tid = tid;
	  s_clock = msg->dataClock;
	} /* !busy */
	PREP_RO_XFER;
	do {
	  if (IS_DBG_LOCK_EXTR) {
	    sprintf(smsg,"extLockReq(%lu): top of do loop for r/o xfer",lockn);
	    StatusMsg(smsg);
	  }
	  lock->busy++;		/* note busy for duration of xfer */
	  lock->rdrs++;		/* +1 for new reader */
	  /* no need to let others go, would have demoted on Release */
	  midSetBit(lock->aerdrs, s_host); /* note they have */
	  midSetBit(lock->aerdrs, 0); /* note aerdrs contains good stuff */
	  midLockXfer(lock, s_clock, s_incarn, s_host, s_thread, s_tid, msg->mode, fti);
	  CommSetTranState(s_host, s_thread, s_tid, M_R_DONE); /* note done */
	  if (IS_DBG_LOCK_EXTR) {
	    sprintf(smsg,"extLockReq(%lu): after LockXfer, about to check Q",lockn);
	    StatusMsg(smsg);
	  }
	  /* probably ok to do following, even if we were not signalled...
	     in that case there should be no one waiting */
	  empty = mid_extq_pop(&(lock->erdrs), &s_host, &s_thread,
			       &s_tid, &s_clock, &s_incarn);
	} while(!empty);
	FIN_RO_XFER;
      } /* end r/o request */

    } /* end own r/w */
    else { /* end own r/w, now own r/o */
      if (IS_DBG_LOCK_EXTR) {
	sprintf(smsg,"extLockReq(%lu): in L_OWN r/o case", lockn);
	StatusMsg(smsg);
	mid_addrmap_prt(lock->lock, lock->amap);
      }
      /* own r/o */
      if ((retrans & (M_FORWARDED|M_RETRANS)) == (M_FORWARDED|M_RETRANS)){
	/* don't act on a forwarded retransmission -- could be old. */
	/* Have requestor ask directly if he still wants it */
	CommReply(M_LOCK_NEWOWN, host, thread, tid, lockn, 0 /*lastbpack*/,
		  0 /*packnum*/, NULL_ADDR_VECTOR /*numack*/, 
		  (char *)&(lock->incarnation),
		  sizeof(lock->incarnation), C_NOFREE);
	INCR_STATS(midLockStats.extReqNewOwn, 1);
	rc = M_R_DOAGAIN;
	goto done;
      }

      if (msg->mode == LOCK_RW) {	/* and r/w request */
	int fixup = 0;
	int readers = (lock->rdrs-1); /* remove self */
	u_long oldstate;
#if PAGE_BASED
	/*
	 * Page based r/w xfer must be exclusive with r/o xfers.
	 * 1: set L_XFER, xfersite.
	 * 2: if place on ewrite, queue on ewritec.
	 * 3: When signalled, ensure condition true & continue..
	 * ...this means readers have to bump a variable when they are transferring,
	 * AND IF THERE ARE MULTIPLE, IT CANNOT GO TO ZERO WHILE LOCK IS NOT HELD.
	 */
	if (lock->ROxfers) {	/* active R/O transfer */
	  oldstate = lock->state;
	  lock->state = L_XFER;
	  lock->xfersite = host;
	  lock->xfertid = tid;
	  if (lock->ewrite) {	/* already transfer thread waiting */
	    /* barf *mjz* */
	    if (IS_DBG_LOCK_EXTR) {
	      sprintf(smsg,"mLockReq(%lx): OWN,RO + RWreq already transfer thread waiting",lockn);
	      StatusMsg(smsg);
	      botch("OWN,RO + RWreq & xferthread");
	    }
	    goto done;
	  }
	  else {
	    lock->ewrite = OUR_TID;
	    if (IS_DBG_LOCK_EXTR) {
	      sprintf(smsg,"mLockReq(%lu): about to condition_wait",lockn);
	      StatusMsg(smsg);
	    }
	    condition_wait(&(lock->ewritec), &(lock->m));
	    if (IS_DBG_LOCK_EXTR) {
	      sprintf(smsg,"mLockReq(%lu): back from OWN,r/w condition_wait",lockn);
	      StatusMsg(smsg);
	    }
	    if (lock->ewrite != OUR_TID) {
	      sprintf(smsg,"extLockReq(%lu): rel from ewritec, ewrite %lu != OURTID %lu",
		      lockn,lock->ewrite, OUR_TID);
	      StatusMsg(smsg);
	      INCR_STATS(midLockStats.extReqXferred,1);
	      rc = M_R_DOAGAIN;
	      botch("OWN,RO + RWreq & tids don't match");
	      goto done;
	    }

	    /* Restore internal state from FASTLOCK state after waiting */
	    mid_lock_X_to_intern (lock);

	    lock->ewrite = 0;
	    ASSERT(lock->ROxfers == 0);
	    ASSERT(lock->state & L_XFER);
	  }
	} /* ROxfers */ 

	readers = (lock->rdrs-1); /* remove self */
#endif /*PAGE_BASED*/
	/* don't send invalidate to sender... he will have to figure it
	   out himself anyway (to wait for invalidate) */
	if (midTestBit(lock->aerdrs,host)) {
	  fixup = 1;
	  readers--;
	  midClearBit(lock->aerdrs,host); /* used again, 31Jan94 */
	}
	/* begin invalidate process */
	if (readers>0) {
	  CommMrequest(lockn, M_INVALIDATE_OP, host, thread, tid, lock->aerdrs,
		       (char *) &(lock->incarnation), 
		       sizeof(lock->incarnation)); 
	}
#if BENCHMARK == 0 
	else { /* readers <= 0  -- verify all bits cleared. **32 bits only** */
#if MAX_PROCESSORS != 32
	  barf we must;
#endif
	  int residual = lock->aerdrs[0] & 0x7fffffff;
	  if (residual) {
	    sprintf(smsg,"extLockReq(0x%x):L_OWN,RO acq rw, and rdrs = %d, but residual says should send invalidates: 0x%x (aerdrs[0]: 0x%x)",
		    lockn, lock->rdrs, residual, lock->aerdrs[0]);
	    StatusMsg(smsg);
	    CommMrequest(lockn, M_INVALIDATE_OP, host, thread, tid, 
			 lock->aerdrs, (char *) &(lock->incarnation), 
			 sizeof(lock->incarnation)); 
	  }
	}
#endif /* not benchmark */

	if (fixup) {
	  midSetBit(lock->aerdrs,host);
	  /* if not set, then will never pass as required when transferring
	     ownership, and the requestor will be hosed. */
	}
	/* not scenerio above.  could have issued a r/o L_ACQ, do not disturb rest of bits */
	lock->state |= L_XFER;	/* note transferring ownership |=? 21dec92*/
	lock->state &= ~L_OWN;
	lock->xfersite = host;
	lock->xfertid = tid;
	lock->bestguess = host;	/* he owns now */
	lock->bgtime = lock->incarnation;
	/* we should probably use separate field, but it's 1am 21dec92 mjz */
	if (lock->incarnation == msg->incarnation) {
#if notwork
	  /* the follwing test will not work because of secondary r/o lock
	     acquisition: the owner does not know all the readers.  it only
	     knows who it gave copies too... however those were free to give
	     copies to others as well.
	     */
	  if (!midTestBit(lock->aerdrs,host)) {
	    botch("R/W xfer with incarn = is not in aerdrs list");
	  }
#endif
	  /* note we don't need to even bother looking for data xfer
	     (assuming no user error) -- upgrade should not find any. */
	  msg->mode |= 0x80000000; /* could be = 80000001 */
	}
	lock->busy++;		/* note busy for duration of xfer */
	midLockXfer( lock, msg->dataClock, msg->incarnation, 
		    host, thread, tid, msg->mode, fti); /* here reacq */
	lock->state &= ~L_XFER;	/* now no nothing about... ownership gone */
	lock->rdrs = 0;		/* ??? */
      } 
      else {			/* and r/o request */
	/* bulletproofing: always set bit.  If for some reason rdrs > 1
	   w/o BAH_USED getting set (say, fastlocks set it that way (it
	   did, but doesn't anymore)), then we would never set it, and
	   we would therefore not invalidate cached copies... hence allowing
	   the invalid state of one party having it R/W, and the other having
	   it R/O.  Spend the extra cycles, and set the bit.
	   */
#if 0	/* bulletproofing */
	if (lock->rdrs == 1) {	/* no external readers yet */
#endif
	  midSetBit(lock->aerdrs, 0); /* note aerdrs now with valid data */
#if 0	/* bulletproofing */
	}
#endif
	lock->rdrs++;
	lock->busy++;		/* note busy for duration of xfer */
	midSetBit(lock->aerdrs, host);
	PREP_RO_XFER;
	midLockXfer(lock, msg->dataClock, msg->incarnation, host, thread, tid, msg->mode, fti);
	FIN_RO_XFER;
      } /* end r/o request */
      rc = M_R_DONE;		/* satisfied request */
    } /* end own r/o */

  } /* case L_OWN */

  /* hold r/o */
  else if (lock->rdrs > 0) {
    if (IS_DBG_LOCK_EXTR) {
      sprintf(smsg,"extLockReq(%lu): in hold r/o case",lockn);
      StatusMsg(smsg);
    }
    if ((msg->mode == LOCK_RW) ||
	((retrans & (M_FORWARDED|M_RETRANS)) == (M_FORWARDED|M_RETRANS))
	) {		/* r/w request, or forwarded retransmission */
      /* forward to owner */
      CommForward(lockn, M_LOCK_OP, lock->bestguess,
		  host, thread, tid, retrans&M_RETRANS, msg, msglen); /* *mjz* */
      /*FWD rc = lock->bestguess;*/	/* forwarded */
      rc = M_R_DOAGAIN;
    } 
    else {			/* r/o request */
      lock->busy++;
      lock->rdrs++;
      midSetBit(lock->aerdrs, 0); /* note has valid data */
      midSetBit(lock->aerdrs, host);
      PREP_RO_XFER;
      midLockXfer(lock, msg->dataClock, msg->incarnation, host, thread, tid, msg->mode, fti);
      FIN_RO_XFER;
      rc = M_R_DONE;
    }
  } /* case hold r/o */

  /* don't know anything about this lock */
  else {
    if (retrans & M_DOAGAIN) {
      if (((msg->mode == LOCK_RW) && (host == lock->xfersite)) ||
	  ((msg->mode == LOCK_RO) && midTestBit(lock->aerdrs,host)))
	{
	  /* retransmission, redo request */
	  /**** FILTER BASED ON TID ? ****/
	  /* can check xfertid now -- 1jul91 mjz */
	  lock->busy++;
	  lock->state = L_XFER;
	  if ((msg->mode == LOCK_RW) && (msg->incarnation == lock->incarnation)) {
	    /* should be upgrade retransmission.  no need to give data. */
	    msg->mode |= 0x80000000;
	  }
	  /* state is XFER, no reason to PREP_RO_XFER */
	  midLockXfer(lock, msg->dataClock, msg->incarnation, host, thread, tid, msg->mode, fti);
	  lock->state &= ~L_XFER; /* done */
	  rc = M_R_DONE;
	} else {
	  rc = M_R_DROPPED;	/* forget it */
	}
    } /* M_DOAGAIN */
    else { /* not a do-over */
      /* forward to best guess site */
      CommForward(lockn, M_LOCK_OP, lock->bestguess,
		  host, thread, tid, retrans&M_RETRANS, msg, msglen);
      /*FWD rc = lock->bestguess;*/	/* note forwarded */
      rc = M_R_DOAGAIN;
    }
  } /* case no se nada */

done:

/*  fprintf(stderr,"%s\n-Current State:\n", smsg);fflush(stderr); */
/*  PrintLock(lock->lock); */

  TRACE_6("lkrq", lock->lock, lock->held, lock->state, lock->busy, lock->rdrs);

  mutex_unlock(&(lock->m));
  return(rc);
} /* midLockRequest */

#ifdef UPDATE_HINTS

/*
 ************************************************************************
 *			m i d L o c k X f e r				*
 *----------------------------------------------------------------------*
 * transfer ownership or copy somewhere else...				*
 ************************************************************************
 */
void 
midLockXferHints
    (
     pbuf 	pb,		/* where to assemble messages */
     u_long 	*pblenP		/* in/out: amount of data in packet buffer */
     )
{
    int 	i;
    u_long 	pbleft, pblen;
    hint_t      *hintP;		/* ptr to hint record */
    hseg_t      *hsegP;		/* ptr to hint segment */
    pl_t        *hintLockP;	/* ptr to lock referenced by hint */

    /* add lock location hints, but only if there is room */
    pblen = *pblenP;
    pbleft = mid_max_data_size - pblen;
    if (pbleft >= sizeof(*hsegP)) {

	hsegP = (hseg_t *)(((char *)pb) + pblen);
	pbleft -= (sizeof(*hsegP) - sizeof(hsegP->first));
	hintP = &hsegP->first;
	
	mutex_lock(hintMutex);
	i = hintIndex;
	do {
	    hintLockP = hintTable[i];
	    if (hintLockP == NULL) break;
	    if (hintLockP->state & L_OWN) {    /* only give sure hints */
		ASSERT_PBUF_ADDR (hintP, sizeof(*hintP));
		hintP->lock = hintLockP->lock;
		hintP->incarn = hintLockP->incarnation;
		pbleft -= sizeof(*hintP);
		if (pbleft < sizeof(*hintP)) break;
		hintP++;
	    }
	    i = (i > 0) ? i - 1 : HINT_TABLE_SIZE - 1;
	} while (i != hintIndex);
	mutex_unlock(hintMutex);

	if (hintP != &hsegP->first) {
	    ASSERT_PBUF_ADDR (hsegP, sizeof(*hsegP));
	    hsegP->segtype = HINTSEG;
	    hsegP->num = hintP - &hsegP->first;
	    pblen += ((u_long)hintP) - ((u_long)hsegP);
	    *pblenP = pblen;
	}
    }
}
#endif /* UPDATE_HINTS */  



#ifdef PAGE_BASED
/*
 ************************************************************************
 *		m i d L o c k F r e e D i f f s				*
 *----------------------------------------------------------------------*
 * free queue of diff data for a lock               			*
 ************************************************************************
 */
void 
midLockFreeDiffs 
    (
     diff_list_t *diff_listP    /* ptr to diff list to free */
     )
{
#if STATS
    if (diff_listP->diff_status == DIFF_STAT_VALID) {
        INCR_STATS(midLockStats.paged_diff_buffer_used,
	           -(diff_listP->diff_size));
    }
#endif /*STATS*/
    mid_pbuf_qfree (&diff_listP->diff_queue);
    diff_listP->diff_size = 0;
    diff_listP->diff_status = DIFF_STAT_UNKNOWN;
    diff_listP->diff_incarn = NULL_CLOCK_VALUE;
}

/*
 ************************************************************************
 *		m i d L o c k B u i l d D i f f s			*
 *----------------------------------------------------------------------*
 * build diff of the lock's data                   			*
 ************************************************************************
 */
void 
midLockBuildDiffs 
    (
     pl_t	 *lock,		/* lock in question */
     int 	 diff_type,	/* pack diffs or data */
     int	mark_clean,	/* =1 to mark data as clean */
     diff_list_t *diff_listP    /* ptr to rtnd diff list */
     )
{
    int 	rc;
    pbuf 	pb;		/* where to assemble messages */
    u_long 	pblen;		/* amount of data in packet buffer */
    m_lock_req_t *lr;		/* lock request message */
    s_addr_map_t	*amap;	/* amap for delta_pack */
    char        *amap_addr;	/* parm to delta_pack */
    long        amap_len;
    char        *data_addr;
    long        data_len;
    long	xfer_len;
    long        total_len;
    diff_desc_t *diffP;
    clock_t     diff_incarn;
    char255	smsg;
    
    if (IS_DBG_LOCK_EXTR) { /* use something else? */
	sprintf(smsg,"midLockBuildDiffs: pack%s %s for lock %u incarn %lu}",
		(mark_clean) ? "&clean" : "",
		(diff_type == PAGE_PACK_DATA) ? "DATA" : "DIFF",
		lock->lock, lock->incarnation);
	StatusMsg(smsg);
    }
    ASSERT (diff_listP->diff_status == DIFF_STAT_UNKNOWN);

    /* initialize */
    amap = lock->amap;
    total_len = 0;
    diff_incarn = lock->incarnation;

    /*
     * Loop building diff buffers
     */
    if (amap != NULL) {

	amap_addr = NULL; 
	amap_len = 0; 
	data_addr = NULL; 
	data_len = 0; 
	xfer_len = 0;
	
	do {
	
	    /* allocate a buffer for the reply */
	    pb = mid_pbuf_get();
	    ASSERT (pb != NULL);
	    
	    /* add lock reply data */
	    lr = (m_lock_req_t *) pb;
	    pblen = sizeof(m_lock_req_t);
	    
	    rc = mid_paged_pack_data((u_long)lock, diff_type, mark_clean,
				     &amap, &amap_addr, &amap_len,
				     diff_incarn,
				     (char *)(lr+1), &pblen, &xfer_len);
	    if (xfer_len == 0) {
		mid_pbuf_free(pb);
		break;
	    }
	    
#ifdef STATS
	    if (diff_type == PAGE_PACK_DATA) {
		INCR_PSTATS(midLockStats.paged_pack_data, 1);
		INCR_PSTATS(midLockStats.paged_pack_data_size, xfer_len);
	    } else {
		INCR_PSTATS(midLockStats.paged_pack_diff, 1);
		INCR_PSTATS(midLockStats.paged_pack_diff_size, xfer_len);
	    }
	    INCR_PSTATS(midLockStats.paged_data, xfer_len);
#endif	  
	    
	    diffP = DIFF_DESC_PTR(pb);
	    ASSERT_PBUF_ADDR (diffP, sizeof(*diffP));
	    diffP->diff_incarn = diff_incarn;
	    diffP->diff_size = pblen -  sizeof(m_lock_req_t);
	    mid_pbuf_enqueue (pb, &(diff_listP->diff_queue));
	    
	    total_len += diffP->diff_size;
	    
	    if (IS_DBG_LOCK_MLPD) {
		sprintf(smsg,
			"saving diff_buffer 0x%x incarn %lu size %lu/%lu for lock %lu",
			diffP, diffP->diff_incarn, diffP->diff_size, total_len, lock->lock);
		StatusMsg(smsg);
	    }

	} while (rc != 0);
    }

    /* Set total size of diffs for this incarnation */
    diff_listP->diff_size = total_len;
    diff_listP->diff_status = DIFF_STAT_VALID;
    diff_listP->diff_incarn = diff_incarn;

#if STATS
    INCR_STATS(midLockStats.paged_diff_buffer_used,
	       diff_listP->diff_size);
    if (midLockStats.paged_diff_buffer_used > 
	midLockStats.paged_diff_buffer_max) {
	midLockStats.paged_diff_buffer_max =
	    midLockStats.paged_diff_buffer_used;
    }
#endif /*STATS*/


} /* midLockBuildDiffs */

void
midLockIncrDiffs 
    (
     pl_t	 *lock 		/* lock in question */
     )
{
    int i, delta;

    /* lock incarnation number was just incremented.
     * Adjust incarn queue to account for new incarnation.
     */
    TRACE_6("IncD", lock, lock->incarnation, (lock->incarnation-MAX_DIFF_INCARN)+1,
	    lock->diff[0].diff_incarn, lock->diff[MAX_DIFF_INCARN-1].diff_incarn);
    ASSERT (lock->incarnation > lock->diff[0].diff_incarn);

    if (lock->diff[0].diff_incarn == NULL_CLOCK_VALUE) 
	delta = MAX_DIFF_INCARN;
    else
	delta = lock->incarnation - lock->diff[0].diff_incarn;

    for (i = MAX_DIFF_INCARN-1; i >= 0; i--) {
	
	if ((i + delta) >= MAX_DIFF_INCARN) {
	    /* free this diff queue */
	    midLockFreeDiffs (&lock->diff[i]);
	} else {
	    /* Move this diff queue & reset */
	    lock->diff[i+delta] = lock->diff[i];
	    mid_pbuf_qinit (&lock->diff[i].diff_queue);
	}
	lock->diff[i].diff_size = 0;
	lock->diff[i].diff_status = DIFF_STAT_UNKNOWN;
	lock->diff[i].diff_incarn = NULL_CLOCK_VALUE;
    }

#ifdef DEBUG
    /* Verify the incarnation numbers match up as expected */
    { clock_t incarn = lock->incarnation;
      for (i = 0; i < MAX_DIFF_INCARN; i++, incarn--) {
	  ASSERT((lock->diff[i].diff_incarn == NULL_CLOCK_VALUE) ||
		 (lock->diff[i].diff_incarn == incarn));
	  if (incarn == 0) break;
      }
    }
#endif
}

/*
 ************************************************************************
 *		m i d L o c k X f e r D i f f s				*
 *----------------------------------------------------------------------*
 * Transfer lock data via diffs                        			*
 ************************************************************************
 */
void 
midLockXferDiffs 
    (
     pl_t	*lock,		/* lock in question */
     clock_t    clock,		/* time of requestors newest lock data */
     clock_t    incarn,		/* incarnation of requestor's lock data (page-based) */
     u_long	host,		/* host of requestor */
     u_long	thread,		/* thread@host requesting */
     u_long	tid,		/* transaction id... */
     u_long	mode,		/* mode of request */
     u_int	fti,		/* for FLOW_TIMES */
     addr_vector_t nacks	/* list of acks required */
     )
{
    int 	i, j;
    int         incarn_delta;
    int         c_free, last, packnum;
    int		diff_pack_type;
    int		mark_clean;
    int		save_diff;
    u_long      reply_type;
    u_long	diff_size;
    pbuf        pb, diff_bufP;
    pbufe_t	*e, *enext;
    m_lock_req_t *dlr;
    u_long 	dlr_len;
    diff_desc_t *diffP, diff_saved;
    diff_list_t *diff_listP, diff_list;
    char255	smsg;
#ifdef DEBUG
    int		packcount;
#endif

    /*
     * Page-based code uses the lock incarnation numbers to determine
     * lock data to xfer. The changed data between incarnations is
     * packed into "diff" buffers. When a lock is acquired, the diff
     * buffers are queued off the lock structure (see midLockReply below).
     * When the lock is later transfered, the requestor provides the incarnation
     * number from the last time he saw the lock. If the incarn = lock->incarn
     * then no data need be send. If incarn = lock->incarn-1, then the lock data 
     * must be diff'ed and the differences sent. If incarn < lock->incarn-1, we
     * can scan through the saved diff buffers to see if we have the diff's for
     * all incarnations to make the requestor current. If so, and the diff's are
     * smaller than the actual data, send the saved diffs along with the current
     * diff (calling mid_paged_pack_data). If the we don't have all the diffs to
     * make the requestor current, simply send all of the lock data.
     *
     * For R/W xfers, we can discard the diff buffers after the lock xfer completes.
     * For R/O xfers, we need to hold the diff buffers for next time.
     *
     * WAS 9Mar94
     */
    
    ASSERT (incarn <= lock->incarnation);
    
    INCR_PSTATS(midLockStats.paged_count, 1);
    INCR_PSTATS(midLockStats.paged_data_max, lock->amap_data_size);
#ifdef STATS
    { register u_long i_diff = lock->incarnation - incarn;
      if (i_diff > STAT_PAGED_INCARN_TABLE_SIZE)
	  i_diff = STAT_PAGED_INCARN_TABLE_SIZE;
      INCR_PSTATS(midLockStats.paged_incarn[i_diff], 1);
    }
#endif


     /* 
     ** Determine 
     **  1) what to transfer to requester
     **  2) mark the data as clean or leave it dirty
     **  3) save the diff or discard it after the xfer completes.
     **
     ** initialize default values.
     */
    save_diff = FALSE;
    mark_clean = (mode == LOCK_RW) ? TRUE : FALSE;
    incarn_delta = 0;
    diff_pack_type = PAGE_PACK_DIFFS;

    /* check if there is data bound to this lock */
    if (lock->amap == NULL) goto xfer_data;

    /* check if this is a r/o lock upgrade */
    incarn_delta = lock->incarnation - incarn;
    if (incarn_delta == 0) {
	ASSERT (mode == LOCK_RW);
	ASSERT (lock->diff[0].diff_status != DIFF_STAT_UNKNOWN);
	goto xfer_data;
    }

    /* Check if the requester is sufficiently current for a diff xfer */
    if (incarn_delta > MAX_DIFF_INCARN) {

	/* requester is too far out of date. Send raw data instead */
	diff_pack_type = PAGE_PACK_DATA;
	INCR_PSTATS(midLockStats.paged_incarn_max, 1);
	goto xfer_data;
    }
	
    /*
     * Check if we have all the diffs requested
     * and their size is less than the raw data.
     */
    diff_size = 0;
    for (i = 0; i < incarn_delta; i++) {
	
	/* Check for a missing diff */
	if (lock->diff[i].diff_status == DIFF_STAT_UNKNOWN) {
	    
	    /* its ok if the current incarnation's isn't yet generated */
	    if (i == 0) continue;

	    /* Otherwise we are missing one.
	     * Requester gets a raw data xfer
	     */
	    diff_pack_type = PAGE_PACK_DATA;
	    INCR_PSTATS(midLockStats.paged_incarn_missing, 1);
	    goto xfer_data;
	}
	diff_size += lock->diff[i].diff_size;
    }
		
    /* Don't allow diffs to get too large 
     * This measures the size of the user data not the size of the diff.
     */
    if (diff_size > lock->amap_data_size) {
	diff_pack_type = PAGE_PACK_DATA;
	INCR_PSTATS(midLockStats.paged_incarn_oversize, 1);
	goto xfer_data;
    }

    /*
     * We will generate a diff for this generation
     * Mark the data clean and save the diff for r/o xfers
     */
    mark_clean = TRUE;
    save_diff = (mode == LOCK_RO) ? TRUE : FALSE;

    TRACE_4("XFGd", lock, lock->incarnation, save_diff);

    /* Generate diff for the current incarnation */
    if (lock->diff[0].diff_status == DIFF_STAT_UNKNOWN) {
	midLockBuildDiffs (lock, PAGE_PACK_DIFFS, mark_clean, &lock->diff[0]);
    }
	    
    /* determine final size of diff xfer */
    diff_size = 0;
    for (i = 0; i < incarn_delta; i++) {
	ASSERT  (lock->diff[i].diff_status == DIFF_STAT_VALID);
	diff_size += lock->diff[i].diff_size;
    }
    if (diff_size > lock->amap_data_size) {
	diff_pack_type = PAGE_PACK_DATA;
	INCR_PSTATS(midLockStats.paged_incarn_discarded, 1);

	/* since we already generated a diff for this incarnation
	 * there is no need to clean or save the raw data.
	 */
	mark_clean = FALSE;
	save_diff = FALSE;
	goto xfer_data;
    }

xfer_data:
	
    TRACE_6("XFDF", lock, incarn_delta, 
	    diff_pack_type, mark_clean, save_diff);

    if (diff_pack_type == PAGE_PACK_DIFFS) {

	/* All diffs have been generated for this xfer
	 * go ahead and transfer them
	 */
	diff_listP = lock->diff;
	if (save_diff == FALSE) 
	    c_free = C_FREE;	/* free diffs after lock xfer */
	else
	    c_free = C_NOFREE;	/* save diffs for other R/O acquires */

	/* Must send at least 1 packet */
	if (incarn_delta == 0) {
	    incarn_delta = 1;
#ifdef STATS
	    if (lock->amap) {
		INCR_PSTATS(midLockStats.paged_diff_nodiff, 1);
	    } else {
		INCR_PSTATS(midLockStats.paged_diff_nodata, 1);
	    }
#endif
	}

	/* verify that there is at least 1 packet
	 * in the final incarnation. This is necessary
	 * for the receiver to get the correct incarn number.
	 */
	if (MID_PBUF_QTOP(&(diff_listP[0].diff_queue)) == NULL) {
	    /*
	     * Get a pbuf for the reply
	     * Mark the diff descriptor as zero length
	     */
	    pb = mid_pbuf_get();
	    ASSERT (pb != NULL);
	    diffP = DIFF_DESC_PTR(pb);
	    diffP->diff_incarn = lock->incarnation;
	    diffP->diff_size = 0;

	    diff_listP[0].diff_size = 0;
	    diff_listP[0].diff_status = DIFF_STAT_VALID;
	    diff_listP[0].diff_incarn = lock->incarnation;
	    mid_pbuf_enqueue (pb, &diff_listP[0].diff_queue);
	}
	INCR_PSTATS(midLockStats.paged_incarn_ok, 1);

    } else {
	
	/* 
	 * Generate reply containing the raw data 
	 */
	diff_list.diff_size = 0;
	diff_list.diff_status = DIFF_STAT_UNKNOWN;
	diff_list.diff_incarn = NULL_CLOCK_VALUE;
	mid_pbuf_qinit(&diff_list.diff_queue);

	/* verify we are cleaning a r/w xfer */
	ASSERT ((mode == LOCK_RO) || (mark_clean == TRUE));

	midLockBuildDiffs (lock, PAGE_PACK_DATA, mark_clean, &diff_list);

	diff_list.diff_status = DIFF_STAT_VALID;
	diff_listP = &diff_list;
	incarn_delta = 1;
	c_free = C_FREE;	/* don't keep a copy of raw data */

	/* Verify we are freeing the raw diff data */
	ASSERT (save_diff == FALSE);
    }

    /* 
     * Verify that we are sending the correct incarnation
     */
    ASSERT ((diff_listP->diff_incarn == NULL_CLOCK_VALUE) ||
	    (diff_listP->diff_incarn == lock->incarnation));
    ASSERT (incarn_delta > 0);	/* gotta send at least 1 packet */


#ifdef DEBUG
    packcount = 0;
    for (i = incarn_delta - 1; i >= 0; i--) {
	e = MID_PBUF_QTOP(&(diff_listP[i].diff_queue));
	while (e != 0) {
	    e = MID_PBUF_QNEXT(e);
	    packcount++;
	}
    }
#endif
	    
    
    /*
     * Loop sending reply buffers.
     * Diffs must be sent in incarnation order!
     */
    last = FALSE;
    reply_type = M_LOCK_OP;
    packnum = 0;
    for (i = incarn_delta - 1; i >= 0; i--) {

	e = MID_PBUF_QTOP(&(diff_listP[i].diff_queue));
	if (c_free == C_FREE) {
#if STATS
	  if (diff_listP[i].diff_status == DIFF_STAT_VALID) {
	    INCR_STATS(midLockStats.paged_diff_buffer_used,
		       -(diff_listP[i].diff_size));
	  }
#endif /*STATS*/
	    mid_pbuf_qinit(&(diff_listP[i].diff_queue));
	    diff_listP[i].diff_size = 0;
	    diff_listP[i].diff_status = DIFF_STAT_UNKNOWN;
	    diff_listP[i].diff_incarn = NULL_CLOCK_VALUE;
	}
	while (e != 0) {

	    /* Determine if this is the last buffer */
	    enext = MID_PBUF_QNEXT(e);
	    if (enext == NULL) {

		/* loop to find the next buffer to send */
		for (j = i-1; j >= 0; j--) {
		    if (MID_PBUF_QTOP(&(diff_listP[j].diff_queue)) != NULL) 
			break;
		}
		if (j < 0) last = TRUE;
	    }

	    /* get a pointer to the diff request */
	    diff_bufP = (char *) Q2PBUF(e);
	    diffP = DIFF_DESC_PTR(diff_bufP);
	    diff_saved = *diffP; /* save diff desc  */

	    /*
	     * Generate reply header
	     */
	    dlr = (m_lock_req_t *) diffP;
	    ASSERT_PBUF_ADDR (dlr, sizeof(*dlr));
	    dlr->dataClock = lock->dataclock;
	    dlr->mode = mode;
	    dlr->incarnation = diff_saved.diff_incarn;
	    dlr_len = sizeof(*dlr) + diff_saved.diff_size;

	    INCR_PSTATS(midLockStats.paged_data, mid_paged_live_data(dlr, dlr_len));
	    INCR_PSTATS(midLockStats.paged_diff, 1);
	    INCR_PSTATS(midLockStats.paged_diff_size, dlr_len);
	    INCR_PSTATS(midLockStats.paged_diff_resend, 1);
	    INCR_PSTATS(midLockStats.paged_diff_resend_size, dlr_len);

	    /*
	     * Special case the last packet in the reply
	     */
	    if (last) {
#ifdef UPDATE_HINTS
		/* add lock location hints, but only if there is room */
		midLockXferHints ((pbuf)diffP, &dlr_len);
#endif /* UPDATE_HINTS */  

		/* Update lock state.
		 *
		 * NOTE: This code is copied from midLockXfer.
		 *       WAS 17May94
		 */
		ASSERT (lock->busy > 0);
		lock->busy--;
		if ((mode == LOCK_RO) || (lock->busy)) {
		    /* invalidate will come later */
		    if (mode == LOCK_RW) {	/* busy r/w req */
			/* ensure invalidate goes to right place later, though... */
			if(mid_extq_push(&(lock->ewrtrs), host, thread, tid, 0, 0)) {
			    /* no memory */
			    ASSERT(1 == 0);
			}
			lock->rdrs = 0;	  /* ensure 0 for Release to catch. Superflous? */
		    } /* busy r/w req */
		}
		else {			/* not busy & r/w transfer only */
		    reply_type = M_LOCK_INV_OP; /* do it now */
		}
	    } /* last */
	    else {		/* not last packet */

		/* release mutex on lock desc before sending */
/*********************** Hold lock for duration of xfer.   WAS 4Sep94
		mutex_unlock(&(lock->m));
************************/
	    }
		    
	    if (IS_DBG_LOCK_MLPD) {
		sprintf(smsg,
			"sending diff 0x%x incarn %lu size %lu/%lu to %lu lock %lu",
			diffP, diff_saved.diff_incarn, 
			diff_saved.diff_size, dlr_len,
			host, lock->lock);
		StatusMsg(smsg);
	    }
	    TRACE_6("DFxm", lock, lock->incarnation,
		    packnum, diff_saved.diff_incarn, diff_saved.diff_size);

	    /* verify last packet contains the correct incarnation */
	    ASSERT ((last == 0) || 
		    (dlr->incarnation == lock->incarnation));
		
	    CommReply(reply_type, host, thread, tid, lock->lock, last,
		      packnum, nacks, (char *)dlr, dlr_len, c_free);
	    
	    packnum++;
	    ZERO_ADDR_VECTOR(nacks);	/* only acksRequired in 1st packet */
/*********************** Hold lock for duration of xfer.   WAS 4Sep94
	    if (!last) {	** WARNING: last iteration must be under lock. **
		mutex_lock(&(lock->m));
		mid_lock_X_to_intern (lock);
	    }
***********************/
	    
	    if (c_free == C_NOFREE)
		*diffP = diff_saved; /* restore diff desc */
	    e = enext;
	}
    }
    ASSERT (last != 0);
#ifdef DEBUG
    ASSERT (packcount == packnum);
#endif

    /*
     * If we transferred the lock ownership
     * then be sure all diff buffers have been freed.
     */
    if (mode == LOCK_RW) {
	for (i = 0; i < MAX_DIFF_INCARN; i++) {
	    midLockFreeDiffs (&lock->diff[i]);
	}
    }

    TRACE_5("xfdf", lock, lock->incarnation, mode, host);

} /* midLockXferDiffs */
  

#endif /* PAGE_BASED */


/*
 ************************************************************************
 *			m i d L o c k X f e r				*
 *----------------------------------------------------------------------*
 * transfer ownership or copy somewhere else...				*
 ************************************************************************
 */
void 
midLockXfer(lock, clock, incarn, host, thread, tid, mode, fti)
     pl_t	*lock;		/* lock in question */
     clock_t    clock;		/* time of requestors newest lock data */
     clock_t    incarn;		/* incarnation of requestor's lock data (page-based) */
     u_long	host;		/* host of requestor */
     u_long	thread;		/* thread@host requesting */
     u_long	tid;		/* transaction id... */
     u_long	mode;		/* mode of request */
     u_int	fti;		/* for FLOW_TIMES */
{
  pbuf pb;			/* where to assemble messages */
  u_long pblen;			/* amount of data in packet buffer */
  int		pbleft;		/* free space in packet buffer */
  m_lock_req_t *lr;		/* lock request message */
  qseg_t	*qs;		/* queue segments, if any */
  queuer_t	*qr;		/* queuers within the segments */
  u_long	type;		/* op type... */
  u_long	i,j,mask;
  u_long	packnum;	/* packet number */
  char255	smsg;
  s_addr_map_t	*amap;		/* amap for delta_pack */
  clock_t	maxtime;	/* max time for data to transmit */
  char          *amap_addr;	/* parm to delta_pack */
  long          amap_len;
  char          *data_addr;
  long          data_len;
  clock_t       data_clock;
  long		xferlen;
  u_long	s_host, s_thread, s_tid;
  clock_t	s_clock;
  clock_t	s_incarn;
  u_long	upgrade = 0;	/* upgrade -- no need to look at dbits */
  int           c_free;
  int           last_packet;
  addr_vector_t nacks;
#if FLOW_TIMES
  u_int *ftp;
  extern u_int ft_send_time_start, ft_send_time_end, ft_send_time_alloc;
  extern u_int ft_send_time_prealloc;
#endif

#if FLOW_TIMES
  ft_times[fti][FT_LOCKXFER] = FT_READ_TIMER;
#endif

  if (IS_DBG_LOCK_EXTR) { /* use something else? */
    sprintf(smsg,"midLockXfer: lock %u to %u(%u)[%u]@%u{%lu-%lu}",
	    lock->lock, host, thread, tid, clock, incarn, lock->incarnation);
    StatusMsg(smsg);
    mid_addrmap_prt(lock->lock, lock->amap);
  }
  INCR_STATS (midLockStats.xferTotalLocks, 1);
  INCR_STATS (midLockStats.xferTotalLockData, lock->amap_data_size);
  TRACE_6("XFER", lock->lock, lock->incarnation, clock, incarn, host);

  /* Convert upgrade "mode" back to its original mode */
  if (mode & 0x80000000) {
    mode &= 0x7fffffff;		/* turn off magic bit */
    INCR_STATS(midLockStats.xferUpgrade, 1);
    upgrade = 1;
  }
  ASSERT ((mode == LOCK_RO) || (mode == LOCK_RW));

  /* verify FASTLOCKS state */
  ASSERT (lock->held == MIDWAY_LOCKHELD_OTHER);

  /* should unlock during data transfer */
  /* busy--, so can send invalidate with last packet if necessary */

  /* should go find all the address, length, time triples... */
  /* construct reply packets. --> must fit into mid_max_data_size byte chunks.
     general format of entire blast:
         <lock-mode>{data segments}{qers}{hints}
     This is broken up into packets that have
         <lock-mode>{optional segments = data/qers/hints}
     Data segments (addrseg_t):
         <type><clock>(<length><addr><data>)+   -> all runs of data @clock
	 [ (---)+ means one or more of --- ]
     Q'er segments:
         <type><num>(<host><tid><clock><mode>)^num
     Each packet must contain complete segments, so you can work on them
     as they arrive.
   */
  if (IS_DBG_LOCK_EXTR) {
    sprintf(smsg, "XFER: amapaddr before pbuf acq: %x", lock->amap);
    StatusMsg(smsg);
  }

  /* initialize */
  packnum = 0;			/* first packet is 0 */
  last_packet = 0;
  c_free = C_FREE;


  /* 
   * Generate AcksRequired for the first packet... (do to all?) 
   */
  ZERO_ADDR_VECTOR(nacks);	/* nothing yet */
  if (mode == LOCK_RW) {	/* r/w lock - note invalidates req'd */
      for (i=0,j=0; i<=Comm_hnum; i += 32, j++) {
	  nacks.bitArray[j] = lock->aerdrs[j];
      }
      midClearBit(nacks.bitArray, 0); /* reset BAH_USED bit */

      /* always add ourselves (since this is a r/w request) */
      midSetBit (nacks.bitArray, Comm_index);

      if (lock->state & L_XFER) {	
	  /*
	   * only transfer for writers the first time through -- not on a
	   * retransmission.  On a retransmission, L_XFER is NOT set.
	   * In particular, we could destroy a waiting writer that got
	   * queued because we're in L_ACQRW for the same lock.
	   * If we really transferred the queued info, then perhaps
	   * it would be OK; after all, we are transferring info
	   * that the process needs anyway.   At worst, the info is
	   * thrown away.  Hmmm... perhaps it's a wash.  At any rate,
	   * it's not an error to have queued waiters if we are
	   * retransmitting. --14Jul92 mjz
	   */
      
      /*
       * should transfer waiters here.		ONLY FOR WRITERS--3jul92!
       *
       * we are not, so zap queues & signal waiter.  As of 3/20/92, new policy:
       * if waiter wakes up, & his tid isn't in proper spot, he assumes he
       * was transferred, and goes poof.  I should change the condition_wait's
       * to explicity check state...  SIGH.  Can't just zap queues, 'cause
       * wrtrs queue is overloaded & used for invalidate placeholder (see below).
       */
      /* would transfer writers */
      if (lock->ewrite) {
	/* was waiter.  This is unlikely here, 'cause writers have priority over
	 * readers, and either we are a reader & there was no writer, or we
	 * are the writer that was formerly waiting here.
	 */
#if DEBUG_QDWRITERS
	sprintf(smsg,"Xfer: notifying queued writer to self destruct. l=%lu@%lu->h=%lu t=%lu m=%lu incarn=%lu datac=%lu", 
		lock->lock, tid, host, thread, mode, lock->incarnation,
		lock->dataclock);
	StatusMsg(smsg);
	/*------should be really temporary---------*/
	/* for some reason, the zwrite bombs, (actually it is the second
	   shell spawned; it is probably trying to make up an argument list),
	   my exception server catches the error, and everything looks funny
	   because I get data (address, etc) that is not for my task.
	   I don't know why this is bombing; it doesn't in a test
	   program (but that wasn't multi-threaded...)  Today's solution:
	   don't do it if it hurts! --14jul92 mjz */
#ifdef BOMBS_AWAY
	sprintf(smsg,"zwrite mattz -m 'Xfer: killing queued writer; suspending task %d'",getpid());
	system(smsg);
#endif BOMBS_AWAY
	StatusMsg("Xfer: spinning!  Please debug me.");
	while(1)
	  cthread_yield();
	/*NOTREACHED*/
	/*--------------1jul92 mjz-----------------*/
#else /*debug_qdwriters*/
	lock->ewrite = 0;		/* notify it should self destruct */
	condition_signal(&(lock->ewritec));	/* must be signal? see eread below */
	INCR_STATS(midLockStats.xferEwrQLost, 1);
#endif /*DEBUG_QDWRITERS*/
      }
      /* null after so that we can see who was queued if we stop above. */
      /*mid_extq_null(&(lock->ewrtrs));*/
      while (!mid_extq_pop(&(lock->ewrtrs), &s_host, &s_thread, &s_tid, &s_clock, &s_incarn)){
	CommSetTranState(s_host, s_thread, s_tid, M_R_DOAGAIN);
	if (s_clock != 0) {
	  INCR_STATS (midLockStats.xferEwrtLost, 1);
	} else {
	  INCR_STATS(midLockStats.xferEwrILost, 1);
	}
      }
      
      /*mid_extq_null(&(lock->erdrs));*/
      while (!mid_extq_pop(&(lock->erdrs), &s_host, &s_thread, &s_tid, &s_clock, &s_incarn)){
	/*FWD CommSetTranState(s_host, s_thread, s_tid, host); *//* act as if fwded */
	CommSetTranState(s_host, s_thread, s_tid, M_R_DOAGAIN);
	INCR_STATS (midLockStats.xferErdrLost, 1);
      }
      if (lock->eread) {		/* was waiter */
	lock->eread = 0;		/* signal self destruction */
	condition_signal(&(lock->ereadc)); /* let it go */
	/* if signal doesn't happen immediately, we could signal wrong guy... */
	/* (we don't require execution of the other right now (indeed it can't
	   we hold the mutex) but it must be scheduled now. */
      }
   }  /* L_XFER set */


#if UPDATEQ_FLUSH_ON_DEMAND > 0
    /*
     * Flush update queue for non-EC memory.
     */
    if (UPDATEQ_NOT_EMPTY) {
      int rc;

      if (IS_DBG_PORT_QUEUE) StatusMsg("LockXfer flushing update queue");
      rc = mid_updateq_flush (UPDATEQ_CALLER_LOCKXFER);
    }
#endif 

  } /* rw lock, note invalidates required; and q'ers transferred */

#ifdef PAGE_BASED_RC
  /* Flush update queue for page-based code */
  paged_flush_updates(lock);
#endif

#if FLOW_TIMES
  ft_times[fti][FT_LOCKQRS] = FT_READ_TIMER;
#endif
  

#ifdef PAGE_BASED
  /*
   * Page-based code builds entire reply before sending the first reply packet.
   * This is necessary to determine if diffsize > locksize (and we should
   * send raw data instead). 
   */
  midLockXferDiffs (lock, clock, incarn, host, thread, tid, mode, fti, nacks);
  return;
#endif


  /* allocate a buffer for the reply */
  pb = mid_pbuf_get();
  ASSERT (pb != NULL);

  /* add lock reply data */
  lr = (m_lock_req_t *) pb;
  lr->dataClock = lock->dataclock; /* may change as we run through data vals */
  lr->mode = mode;
  lr->incarnation = lock->incarnation;
  pblen = sizeof(m_lock_req_t);

#if FLOW_TIMES
  /* this doesn't belong here.  figure out what we want to do */
  ft_times[fti][FT_LOCKDLTA] = FT_READ_TIMER;
#endif
  if (upgrade) {
    amap = 0;
  } else {
    amap = lock->amap;
  }

  maxtime = NULL_CLOCK_VALUE; amap_addr = NULL; amap_len = 0; 
  data_addr = NULL; data_len = 0; data_clock = NULL_CLOCK_VALUE;
  if (amap) {
    xferlen = 0;
    while(mid_pack_data(&amap, &amap_addr, &amap_len,
			clock, &maxtime, 
			(char *)(lr+1), &pblen,
			&data_addr, &data_len, &data_clock,
			&xferlen)) {
      /* returned 1, so filled current packet, and needs another */
      ASSERT (xferlen > 0);
      INCR_STATS (midLockStats.xferTotalData, xferlen);
      INCR_PSTATS(midLockStats.paged_diff, 1);
      INCR_PSTATS(midLockStats.paged_diff_size, pblen);
      mutex_unlock(&(lock->m));
      if (lr->dataClock < maxtime)
	lr->dataClock = maxtime;
#if FLOW_TIMES
      if (packnum <= FT_MAXPAKS) {
	ftp = /*&(ft_times[fti][FT_REQPAKS+(2*packnum)]);*/
	  FT_MSGINFO(FT_REQPAKS,packnum);
	*ftp++ = FT_READ_TIMER;
      }
#endif
      CommReply(M_LOCK_OP, host, thread, tid, lock->lock, 0/*lastpack*/,
		packnum, nacks, pb, pblen, C_FREE);
#if FLOW_TIMES
      if (packnum <= FT_MAXPAKS) {
	*ftp++ = ft_send_time_prealloc;
	*ftp++ = ft_send_time_alloc;
	*ftp++ = ft_send_time_start;	/* GAG.  not thread safe. */
	*ftp++ = ft_send_time_end;
	*ftp = FT_READ_TIMER;
      }
#endif

      packnum++;
      ZERO_ADDR_VECTOR(nacks);	/* only acksRequired in 1st packet */
      pb = mid_pbuf_get();	/* get a new one */
      ASSERT (pb != NULL);
      pblen = sizeof(m_lock_req_t);
      lr = (m_lock_req_t *) pb;
      lr->dataClock = lock->dataclock;
      lr->mode = mode;
      mutex_lock(&(lock->m));
      mid_lock_X_to_intern (lock);
      lr->incarnation = lock->incarnation;
    } /* while delta pack */

    /* mid_pack_data returned 0, so this is the last packet */
    INCR_STATS (midLockStats.xferTotalData, xferlen);
  }
	
#ifdef UPDATE_HINTS
  /* add lock location hints, but only if there is room */
  midLockXferHints (pb, &pblen);
#endif /* UPDATE_HINTS */  

  /* update clock now */
  if (lock->dataclock < maxtime) {
    lr->dataClock = maxtime;
    lock->dataclock = maxtime;
  }

  /* this is the last iteration---must be done under lock 
   *
   * NOTE: This code is copied to midLockXferDiffs.
   *       WAS 17May94
   */
  ASSERT (lock->busy > 0);
  lock->busy--;
  if ((mode == LOCK_RO) || (lock->busy)) {
    type = M_LOCK_OP;		/* invalidate will come later */
    if (mode == LOCK_RW) {	/* busy r/w req */
      /* ensure invalidate goes to right place later, though... */
      if(mid_extq_push(&(lock->ewrtrs), host, thread, tid, 0, 0)) {
	/* no memory */
	ASSERT(1 == 0);
      }
      lock->rdrs = 0;	       /* ensure 0 for Release to catch. Superflous? */
    } /* busy r/w req */
  }
  else {			/* not busy & r/w transfer only */
    type = M_LOCK_INV_OP; /* do it now */
  }

  if (IS_DBG_LOCK_MSGS)
    /*								---DEBUG */
    {
      long *f = (long *) pb;
      sprintf(smsg, "%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx",
       *f, *(f+1), *(f+2), *(f+3), *(f+4), *(f+5), *(f+6),
       *(f+7), *(f+8), *(f+9), *(f+10),
       *(f+11), *(f+12));
      StatusMsg(smsg);
      sprintf(smsg,"Total outgoing message length: %u", pblen);
      StatusMsg(smsg);
    }
    /*								---DEBUG */

#if FLOW_TIMES
  if (packnum <= FT_MAXPAKS) {
    ftp = FT_MSGINFO(FT_REQPAKS,packnum);
    *ftp++ = FT_READ_TIMER;
  }
#endif

  CommReply(type, host, thread, tid, lock->lock, 1/* lastpacket*/,
	    packnum, nacks, pb, pblen, C_FREE);

#if FLOW_TIMES
  if (packnum <= FT_MAXPAKS) {
    *ftp++ = ft_send_time_prealloc;
    *ftp++ = ft_send_time_alloc;
    *ftp++ = ft_send_time_start;	/* GAG.  not thread safe. */
    *ftp++ = ft_send_time_end;
    *ftp = FT_READ_TIMER;
  }
  ft_times[fti][FT_REQPNUM] = (packnum<FT_MAXPAKS)?packnum:FT_MAXPAKS;
  ft_times[fti][FT_LOCKXFERDONE] = FT_READ_TIMER;
#endif

} /* midLockXfer */

/*
 ************************************************************************
 *		 m i d L o c k I n v a l i d a t e			*
 *----------------------------------------------------------------------*
 * Handle invalidate requests from others				*
 ************************************************************************
 */
void
midLockInvalidate( lockn, host, thread, tid , internal, msg, mlen)
     u_long lockn;		/* name of lock this is all about */
     u_long host;		/* host requesting transfer */
     u_long thread;		/* thread@host requesting transfer */
     u_long tid;		/* transaction id for this request */
     int internal;		/* !=0 if called from w/in, don't need mutex */
     char * msg;		/* contains dataclock if not internal */
     u_long mlen;		/* length of msg */
{
  int i,j;
  pl_t *lock;
  addr_vector_t nacks;
  char255 smsg;			/* for status message */

  /* if don't (hold r/o or nothing) this is an error...
   * If hold r/o,
   *	if granted to others r/o, send invalidates to them (given list?)
   *	set rdrs=0, xfersite=reply-to-host
   *	if busy, set rdrs=0, so will send invalidate reply on release...
   *		enqueue element on ewrtrs, so have thread & tid.
   *	else
   *		send invalidate reply
   *
   * If nothing, if xfersite (lost tid info...)
   *	if busy, then drop packet (could give keepalive)
   *	else
   *	   send invalidate reply ***includes (saved) list of grantees***
   * else
   *	drop packet & report to status server
   */

  /* convert lockname to a pointer to the lock structure */
  lock = (pl_t *) lockn;

  /*
   * If this is an external request,
   * then acquire mutex guarding lock structure
   * Internal requests have already acquired the mutex.
   */
  if (!internal) {
    mutex_lock(&(lock->m));	/* ensure consistent access */

    /* 
     * Convert FASTLOCKS state to internal lock state
     */
    mid_lock_X_to_intern (lock);

    /* validate incoming request */
    if (mlen != sizeof(lock->incarnation)) {
      sprintf(smsg, "midLockInvalidate: request for lock %lu had bad clock value size: %lu instead of %u.",
	      lockn, mlen, sizeof(lock->incarnation));
      StatusMsg(smsg);
      goto done;
    }

    if (lock->incarnation != *((u_long *)msg)) { /* BYTEORDER */
      /* not for whatever version of the lock we have -- drop */
      /* perhaps should queue if strictly greater than our version? --mjz */
      if (IS_DBG_LOCK_WNG) {
	sprintf(smsg, "midLockInvalidate: lock %lu incarnation %lu != message %lu, dropping.",
		lockn, lock->incarnation, *((u_long *)msg));
	StatusMsg(smsg);
      }
      goto done;
    }
  }/* !internal */

  /* verify FASTLOCKS state */
  ASSERT (lock->held == MIDWAY_LOCKHELD_OTHER);

  TRACE_6("INVA", lock->lock, host, lock->state, lock->busy, lock->rdrs);

  /*
   * must stall if ACQ RO, and this invalidate for this lock.  We already
   * know the invalidate is for the current lock incarnation, because
   * we just passed the dataclock test above.  Therefore, if we are ACQRO,
   * stall this invalidate (so the pending acquire can finish).
   * If we do not, we may invalidate and then the pending acquire will
   * finish, and we will have two threads in the same critical section
   * (Very Bad).
   * 15 jul 92: Use "DATAIN" to mean that data has come in for this lock
   * (so the incarnation # was updated) but the thread has not been
   * freed (so the acquire completed.   If DATAIN is not set, then
   * this is a second acquire for this lock, and the invalidate
   * should be honored.  (We did have it, and thought we invalidated
   * it before.)
   */
  if ((lock->state & (L_ACQ|L_DATAIN)) == (L_ACQ|L_DATAIN)) {
    if (IS_DBG_LOCK_EXTR){
      sprintf(smsg,"midLockInv: lock %lu data %lu stalling for pending RO ACQ to complete.", lockn, lock->dataclock);
      StatusMsg(smsg);
    }
    do {
      lock->state |= L_ACQINV;	/* note we are stalled */
      condition_wait(&(lock->iupc), &(lock->m));

      /* Restore internal state from FASTLOCK state after waiting */
      mid_lock_X_to_intern (lock);

    } while (lock->state & L_DATAIN);

    /* sanity check */
    if (lock->incarnation != *((u_long *)msg)) { /* BYTEORDER */
      /* not for whatever version of the lock we have -- drop */
      if (IS_DBG_LOCK_WNG) {
	sprintf(smsg, "midLockInvalidate: AFTER WAITING lock %lu incarnation %lu != message %lu, dropping.",
		lockn, lock->incarnation, *((u_long *)msg));
	StatusMsg(smsg);
      }
      goto done;
    }
  } /* Were acquiring RO, so stalled */

#ifdef VERBOSITY
  fprintf(stderr,"Invalidating %u, lock current state:\n", lockn); /*dbg*/
  PrintLock(lockn);/*dbg*/
#endif VERBOSITY

  /* case hold r/o */
  if ((lock->state == 0) && (lock->rdrs)) {
    if (*(lock->aerdrs) & BAH_USED) {	/* then other sites have copies */
      int fixup = 0;
      int readers = (lock->rdrs-1); /* remove self */
      /* send out invalidate's, perhaps if not in message? */
      if (midTestBit(lock->aerdrs,host)) {
	fixup = 1;
	readers--;
	midClearBit(lock->aerdrs,host);
      }
      if (readers) {
	CommMrequest(lockn, M_INVALIDATE_OP, host, thread, 
		     (clock_t) tid, lock->aerdrs,
		     (char *) &(lock->incarnation), sizeof(lock->incarnation));
      }
      if (fixup) {
	midSetBit(lock->aerdrs,host);
	/* i suppose no need to do it, 'cause no need to ever know. */
      }
    }
    if (lock->eread) {
      u_long s_host, s_thread, s_tid, s_clock, s_incarn;
      /* 1aug92mjz: there is a external reader transfer thread waiting 
       * that hasn't run yet.  It must have been signalled since we are
       * in hold r/o state.  For correctness, kill it off now.  When
       * it retransmits, the request will be forwarded.
       * (We don't want to let it go after we have invalidated!)
       */
      lock->eread = 0;		/* kill it off */
/*      mid_extq_null(&(lock->erdrs));*/ /* here too */
      while (!mid_extq_pop(&(lock->erdrs), &s_host, &s_thread, &s_tid, &s_clock, &s_incarn)){
	CommSetTranState(s_host, s_thread, s_tid, M_R_DOAGAIN);
	INCR_STATS (midLockStats.invErdrLost, 1);
      }
      condition_signal(&(lock->ereadc));	/* just in case */
    }
    lock->rdrs = 0;		/* allow no other readers in */
    lock->xfersite = host;	/* so we can detect legit retrans */
    if (lock->busy) {		/* cannot invalidate now */
      /* have implicitly queued the invalidate by setting rdrs = 0 */
      /* GAK -- what if not to head of q?  needs to go to head!! push, ~queu */
      if(mid_extq_push(&(lock->ewrtrs), host, thread, tid, 0, 0)) {
	/* no memory *mjz* */
      }
    } else {			/* not busy */
      /* invalidate reply */
      ZERO_ADDR_VECTOR(nacks);	/* nothing yet */
      for (i=0,j=0; i<=Comm_hnum; i += 32, j++) {
	  nacks.bitArray[j] = lock->aerdrs[j];
      }
      midClearBit(nacks.bitArray, 0); /* reset BAH_USED bit */

      CommReply(M_INVALIDATE_OP, host, thread, tid, lockn, 1, 0, nacks,
		(char *) 0, 0, C_NOFREE);
    }
  } /* hold r/o */

  /* case acquiring r/o.  If we get the invalidate before the response
     (musta, since we're here...???what if they're queued on the lock???
     if they came in the interface physically later, the duplicate suppression
     stuff would have quashed the packet) then we lost the response.  We
     send an invalidate, and will retrans the request... which should
     get forwarded to the right place.
   */
  else if (lock->state & L_ACQ) {
    CommReply(M_INVALIDATE_OP, host, thread, tid, lockn, 1, 0, NULL_ADDR_VECTOR,
	      (char *) 0, 0, C_FREE);
  } /* acquiring r/o */

  /* case acquiring r/w (NOTE RO takes precedence over RW 1jul92mjz)*/
  else if (lock->state & L_ACQRW) {
    /*
     * the problem here is we get an invalidate for a lock we previously
     * held a r/o copy of and are trying to promote to r/w.
     * We have to reply to ourselves.  (Yuck).  We have to get the invalidate
     * so we can send any secondary invalidations before the aerdrs entries
     * are zapped (when we actually get the lock).  It's easier to just
     * send the message; otherwise we would have to diddle with the comm
     * data structures and perhaps release the waiting writer thread ourselves
     * (who would then zap the aerdrs struct).  Woof.
     */

    if (lock->rdrs > 0) {	/* first time through, and valid */
      if (*(lock->aerdrs) & BAH_USED) {	/* then other sites have copies */
	int fixup = 0;
	int readers = (lock->rdrs-1); /* remove self */
	/* send out invalidate's, perhaps if not in message? */
	/* Don't invalidate requestor anymore 31jan94mjz */
	if (midTestBit(lock->aerdrs,host)) {
	  fixup = 1;
	  readers--;
	  midClearBit(lock->aerdrs,host);
	}
	if (readers) {
	  CommMrequest(lockn, M_INVALIDATE_OP, host, thread, 
		       (clock_t) tid, lock->aerdrs,
		       (char *) &(lock->incarnation), 
		       sizeof(lock->incarnation)); 
	}
	if (fixup) {
	  midSetBit(lock->aerdrs,host);
	}
      }
      lock->rdrs = 0;		/* allow no other readers in */
      lock->xfersite = host;	/* so we can detect legit retrans */
      if (lock->busy) {		/* cannot invalidate now */
	/* have implicitly queued the invalidate by setting rdrs = 0 */
	if(mid_extq_add(&(lock->ewrtrs), host, thread, tid, 0, 0)) {
	  /* no memory *mjz* */
	  StatusMsg("midLockReq: extq_add #4 out of memory...");
	}
      }
    } /* were reading sites */
    else {			/* not reading sites -- retrans or error */
      if (lock->xfersite != host) {
	/* uh-oh */
	sprintf(smsg, "midLockInvalidate: lock %lu xfersite %lu != host %lu",
		lock->lock, lock->xfersite, host);
	StatusMsg(smsg);
      } else {
	if (lock->busy) {	/* cannot send reply -- ack instead */
	  CommAck(host, thread, tid, lockn, 0, 0); /* ReplyPending */
	}
      }
    } /* lock->rdrs == 0 */

    /* finally, send invalidate if not busy & not error */
    if ((lock->xfersite == host) && (lock->busy == 0)) {
      /* invalidate reply */
      ZERO_ADDR_VECTOR(nacks);	/* nothing yet */
      for (i=0,j=0; i<=Comm_hnum; i += 32, j++) {
	  nacks.bitArray[j] = lock->aerdrs[j];
      }
      midClearBit(nacks.bitArray, 0); /* reset BAH_USED bit */

      CommReply(M_INVALIDATE_OP, host, thread, tid, lockn, 1, 0, nacks,
		(char *) 0, 0, C_NOFREE);
    } /* send invalidate */
  } /* acquiring r/w */

  /* case nothing */
  else if ((lock->state == 0) && (lock->rdrs == 0)) { /* "nothing" */
    if (lock->busy) {
      /* still local reader */
      CommAck(host, thread, tid, lockn, 0, 0); /* ReplyPending */
    } else { /* lock not busy */
      /* retrans invalidateReply *//*--check xfersite?--*/
      ZERO_ADDR_VECTOR(nacks);	/* nothing yet */
      for (i=0,j=0; i<=Comm_hnum; i += 32, j++) {
	  nacks.bitArray[j] = lock->aerdrs[j];
      }
      midClearBit(nacks.bitArray, 0); /* reset BAH_USED bit */

      CommReply(M_INVALIDATE_OP, host, thread, tid, lockn, 1, 0, nacks,
		(char *) 0, 0, C_NOFREE);
    } /* lock not busy */
  } /* "nothing" */

  /* case oops */
  else {
    sprintf(smsg, "Got Invalidate in bad state - state=%lx, rdrs=%lu",
	    lock->state, lock->rdrs);
    StatusMsg(smsg);
#ifdef STOP_ON_ERROR
    botch("Invalid Invalidate");
#endif
  } /* oops */

done:
  /* verify FASTLOCKS state */
  ASSERT (lock->held == MIDWAY_LOCKHELD_OTHER);

  /*
   * Release mutex guarding lock before returning external requests.
   */
  if (!internal) {
    mutex_unlock(&(lock->m));
  }
  
  return;
} /* midLockInvalidate */

/*
 ************************************************************************
 *		 m i d L o c k F a s t I n v a l i d a t e		*
 *----------------------------------------------------------------------*
 * Handle invalidate requests implied in receiving an invalidate        *
 * required from another processor as part of a lock reply message	*
 * Returns the list of other hosts that we passed copies to ORd with	*
 * inv_reqd, and 0/1, 1 if we sucessfully invalidated the local copy	*
 * (have no current readers)						*
 ************************************************************************
 */
int
midLockFastInvalidate( lockn, host, thread, tid ,  inv_reqd)
     u_long lockn;		/* name of lock this is all about */
     u_long host;		/* host requesting transfer */
     u_long thread;		/* thread@host requesting transfer */
     u_long tid;		/* transaction id for this request */
     BITARRAY * inv_reqd;	/* IN/OUT: array of hosts to invalidate */
{
  char255 smsg;			/* for status message */
  pl_t *lock;
  int i;
  int invalidated;

  /* if not acquiring r/w this is an error...
   */

  lock = (pl_t *)lockn;

  invalidated = 0;		/* assume we don't */

  mutex_lock(&(lock->m));

  /* 
   * Convert FASTLOCKS state to internal lock state
   *
   * Just following wsawdon here.
   * ...i imagine not necessary, since we just did a LockRequest
   *  (what if retrans? should just drop, we might bomb. test on enet.
   *  --31jan94mjz
   */
  mid_lock_X_to_intern (lock);
  
#ifdef VERBOSITY
  fprintf(stderr,"Invalidating %u, lock current state:\n", lockn); /*dbg*/
  PrintLock(lockn);/*dbg*/
#endif VERBOSITY

  TRACE_6("FINV", lock->lock, host, lock->state, lock->busy, lock->rdrs);


  /* case acquiring r/w */
  if (lock->state & L_ACQRW) {
    /*
     * the problem here is we get an invalidate for a lock we previously
     * held a r/o copy of and are trying to promote to r/w.
     * We have to reply to ourselves.  (Yuck).  We have to get the invalidate
     * so we can send any secondary invalidations before the aerdrs entries
     * are zapped (when we actually get the lock).  It's easier to just
     * send the message; otherwise we would have to diddle with the comm
     * data structures and perhaps release the waiting writer thread ourselves
     * (who would then zap the aerdrs struct).  Woof.
     */

    if (lock->rdrs > 0) {	/* first time through, and valid */
      if (*(lock->aerdrs) & BAH_USED) {	/* then other sites have copies */
	BITARRAY used_bit = (*inv_reqd & BAH_USED);
	midOrBitArray(inv_reqd,(lock->aerdrs),Comm_hnum); /* pass back */
	if (!used_bit) {
	  *inv_reqd &= ~BAH_USED; /* turn off bit if was off before */
	}
	/* send out invalidate's, perhaps if not in message? */
	CommPMrequest(lockn, M_INVALIDATE_OP, host, thread, 
		     (clock_t) tid, lock->aerdrs,
		     (char *) &(lock->incarnation), 
		     sizeof(lock->incarnation)); 
      }
      lock->rdrs = 0;		/* allow no other readers in */
      lock->xfersite = host;	/* so we can detect legit retrans */
      if (lock->busy) {		/* cannot invalidate now */
	/* have implicitly queued the invalidate by setting rdrs = 0 */
	if(mid_extq_add(&(lock->ewrtrs), host, thread, tid, 0, 0)) {
	  /* no memory *mjz* */
	  StatusMsg("midLockFInval: extq_add out of memory...");
	}
      }
    } /* were reading sites */
    else {			/* not reading sites -- retrans or error */
/*      botch("midLockFastInvalidate bad state");*/
      if (lock->xfersite != host) {
	/* uh-oh */
	sprintf(smsg, "midLockFastInvalidate: lock %lu xfersite %lu != host %lu",
		lock->lock, lock->xfersite, host);
	StatusMsg(smsg);
#if STOP_ON_ERROR
	botch("midLockFastInvalidate bad host");
#endif
      }
      /* ALWAYS ERROR HERE? --MJZ31JAN94 
	 No, could be a retrans inv reply, or because we got an invalidate
	 reply as the result of a release, not an Xfer (which would still
	 contain the bit to say we need to be invalidated) -mjz16feb94 */
    } /* lock->rdrs == 0 */

    /* finally, send invalidate if not busy & not error */
    if ((lock->xfersite == host) && (lock->busy == 0)) {
      /* invalidate reply */
      invalidated = 1;
    } 
  } /* acquiring r/w */

  /* case oops */
  else {
    sprintf(smsg, "Got FastInvalidate in bad state - state=%lx, rdrs=%lu",
	    lock->state, lock->rdrs);
    StatusMsg(smsg);
  } /* oops */

/*done:*/

  /* verify FASTLOCKS state */
  ASSERT (lock->held == MIDWAY_LOCKHELD_OTHER);
  mutex_unlock(&(lock->m));

  return(invalidated);
} /* midLockFastInvalidate */

/*
 ************************************************************************
 *			m i d L o c k R e p l y				*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
comm_request_t
midLockReply(lockn, srchost, destthread, tid, buf, buflen, dontblock)
     u_long lockn;		/* name of lock */
     u_long srchost;
     u_long destthread;
     u_long tid;
     char *buf;
     u_long buflen;
     int dontblock;		/* 1=> don't block; needed for ATM hack */
{
  pl_t *lock;
  m_lock_req_t *lr;		/* lock request message */
  qseg_t	*qs;		/* queue segments, if any */
  queuer_t	*qr;		/* queuers within the segments */
  addrseg_t	*as;		/* data segments, if any */
  addrpair_t    *ap;		/* address/length/data triples, really */
  addrpair_t    *apnext;	/* where next ap will be */
  clock_t	*dirtybit;	/* yaya */
  u_long        dirtybit_len;	/* length of data covered by dirty bit */
  u_long	linesize;	/* cache line size... */
  clock_t	*sdb;		/* save dirtybit TEST */
  int           hintCnt;	/* counter of hints in segment */
  hseg_t        *hsegP;		/* ptr to hint segment */
  hint_t        *hintP;		/* ptr to lock location hint  */
  pl_t          *hintLockP;	/* ptr to lock being updated */
  region_info	ri;
  int		seglength;
  clock_t       *ddbP;		/* ptr to a dense dirty bit */
  char          *ddb_data_addr;	/* address within dense dirty bit line */
  u_long        ddb_array_len;	/* remaining length in dense dirty bit array */
  int           ddb_i;
  char255	smsg;
  u_long	midway_cache_line_size();
#ifdef PAGE_BASED
  int           i, j;
  int		incarn_delta;
  diff_desc_t  *diffP;
  pbuf		diff_bufP;
  u_long 	diff_buflen = buflen;
  addrseg_t    *diff_endP = NULL;
#endif


  /* convert lockname to a pointer to the lock structure */
  lock = (pl_t *) lockn;
  lr = (m_lock_req_t *) buf;


  if (IS_DBG_LOCK_EXTR) {
    sprintf(smsg, "midLockReply: incoming message length %u", buflen);
    StatusMsg(smsg);
  }
  /* update dataclock value under lock (perhaps last...?--mjz) */
  if (dontblock) {
    if (mutex_try_lock(&(lock->m)) == 0) {
      return(M_R_COULDBLOCK);
    }
  } else {
    mutex_lock(&(lock->m));
  }

  /* Convert FASTLOCK state to internal lock state 
   * This shouldn't be necessary unless this is a retransmission.
   * WAS 9Jan94
   */
  mid_lock_X_to_intern(lock);

  TRACE_6("LRPY", lock->lock, srchost, lock->state, lock->busy, lock->rdrs);

#ifdef PAGE_BASED
  /*
   * MUST apply diffs in order!
   * queue out of order requests.
   * WAS 9Mar94
   */
#endif



  if (LESS_THAN_CLOCK (lock->dataclock, lr->dataClock))
    lock->dataclock = lr->dataClock;
  if (lock->incarnation < lr->incarnation) {
    lock->incarnation = lr->incarnation;
    TRACE_4("LRPi", lock->lock, lock->dataclock, lock->incarnation);
#ifdef PAGE_BASED
    midLockIncrDiffs (lock);
#endif
  }
  /* note we have updated incarnation, but are still L_ACQ(RW) */
  lock->state |= L_DATAIN;	
  mutex_unlock(&(lock->m));

  buflen -= sizeof(m_lock_req_t);
  as = (addrseg_t *) (lr+1);	/* point after lock info */
  while (buflen) {
    switch(as->segtype) {

    case DATASEG:
      mid_unpack_data((char **)&as, &buflen, lock->amap,
		      "midLockReply", mid_MUD_lock);
      break; /* case DATASEG */

#ifdef PAGE_BASED
    case DIFFSEG:
      mid_paged_unpack_data((u_long)lock,
			    (char **)&as, &buflen, lock->amap,
			    "midLockReply", mid_MUD_lock);
      break; /* case DIFFSEG */
#endif

    case HINTSEG:
      hsegP = (hseg_t *) as;
      hintP = &hsegP->first;
      hintCnt = hsegP->num;
      buflen -= ((u_long)hintP - (u_long)as); /* account for hintseg header */

      while ((hintCnt-- > 0) && (buflen >= sizeof(*hintP))) {

#ifdef UPDATE_HINTS
	/* locate lock descriptor */
	hintLockP = (pl_t *) hintP->lock;
	ASSERT (hintLockP->lock == hintP->lock);

	mutex_lock(&(hintLockP->m));

	if (IS_DBG_LOCK_EXTR) {
	  sprintf(smsg, 
		  "UpdatingHint lock %lx incarn=%lu ->%lu, was incarn %u bg %lu ->%lu",
		  hintP->lock, hintP->incarn, srchost, 
		  hintLockP->incarnation, hintLockP->bgtime, 
		  hintLockP->bestguess);
	  StatusMsg(smsg);	/* dbg */
	}
	
	/* Check if the hint is better than what we had.
	 * Must be better than last time we saw lock (incarnation)
	 * and better than the last hint update (bgtime).
	 */
	if ((GREATER_THAN_CLOCK (hintP->incarn, hintLockP->incarnation)) &&
	    (GREATER_THAN_CLOCK (hintP->incarn, hintLockP->bgtime))) {
	    
	  /* verify lock state */
	  ASSERT ((hintLockP->state & L_OWN) == 0);

	  /* update hint */
	  hintLockP->bestguess = srchost;
	  hintLockP->bgtime = hintP->incarn;
	}
	mutex_unlock(&(hintLockP->m));

#endif /* UPDATE_HINTS */
	  
	hintP++;
	buflen -= sizeof(*hintP);
      } /* while */

#ifdef PAGE_BASED
      /* save ptr to mark end of diffs */
      if (diff_endP == NULL)  diff_endP = as;
#endif

      as = (addrseg_t *) hintP;	/* next segment would follow */
      break; /* case HINTSEG */

    default:
      sprintf(smsg,"midLockReply: Unknown segment received (%x) decoding lock reply for %lu -- giving up [buflen=%u]",
	      as->segtype, lock->lock,buflen);
      StatusMsg(smsg);
#if STOP_ON_ERROR
      botch("MLR unknown segment");
#endif
      return(M_R_DONE);			/* better not be under lock */
      break;
    } /* switch */
  } /* while(buflen) */


#ifdef PAGE_BASED
  /*
   * Save diff we received to bring us up to the current incarnation.
   *
   * Note: Sender should indicate that diff is entire copy of data
   * If we knew that was the case, we wouldn't need to store it.
   * Alternatively, if we knew this was the last reply buffer
   * we could simply sum the diff buffer sizes, and if greater
   * than the data size, free the buffers.
   */

  if (diff_endP != NULL) 
      diff_buflen = ((vm_offset_t)diff_endP) - ((vm_offset_t)buf);

  if (diff_buflen > sizeof(m_lock_req_t)) {

      /* HACK! Get a pbuf & copy the data into it.
       * Someday we will be able to avoid this copy, but for now...
       * WAS 9Mar94
       */
      diff_bufP = mid_pbuf_get();
      ASSERT (diff_bufP != NULL);
      
      diffP = DIFF_DESC_PTR(diff_bufP);
      diffP->diff_incarn = lock->incarnation;
      diffP->diff_size = diff_buflen -  sizeof(m_lock_req_t);

      /* Save diff data */
      bcopy ((lr+1),					/* source */
	     (char *)diff_bufP + sizeof(m_lock_req_t), 	/* target */
	     diffP->diff_size);				/* length */

      /* Put this diff at the head of the incarnation queue
       * The queues are sorted by incarnation number.
       * If this diff is a different incarn than the 1st queue
       * then push all of the queues down & start a new one.
       */
      if (lock->incarnation != lock->diff[0].diff_incarn) {
	  if (lock->diff[0].diff_incarn != NULL_CLOCK_VALUE) {

	      /* verify diffs are applied in incarn order */
	      ASSERT (lock->incarnation > lock->diff[0].diff_incarn);

	      /* free elements in last queue */
	      midLockFreeDiffs (&lock->diff[MAX_DIFF_INCARN-1]);

	      /* move remaining queues down */
	      for (i = MAX_DIFF_INCARN-1; i > 0; i--) {
		  lock->diff[i] = lock->diff[i-1];
	      }
	      mid_pbuf_qinit (&lock->diff[0].diff_queue);
	  }
	  lock->diff[0].diff_incarn = lock->incarnation;
	  lock->diff[0].diff_size = 0;
      }

      mid_pbuf_enqueue (diff_bufP, &(lock->diff[0].diff_queue));
      lock->diff[0].diff_size += diffP->diff_size;
      
      if (IS_DBG_LOCK_MLPD) {
	  sprintf(smsg,
		  "saving diff_buffer 0x%x incarn %lu size %lu from lock %lu host %lu",
		  diffP, diffP->diff_incarn, diffP->diff_size, lockn, srchost);
	  StatusMsg(smsg);
      }
  }
#endif


  return(M_R_DONE);
} /* midLockReply */


/*
 ************************************************************************
 *			m i d L o c k N e w O w n			*
 *----------------------------------------------------------------------*
 ************************************************************************
 */
comm_request_t		/* not really, but easy for now */
midLockNewOwn(lockn, srchost, destthread, tid, buf, buflen)
     u_long lockn;		/* name of lock */
     u_long srchost;
     u_long destthread;
     u_long tid;
     char *buf;
     u_long buflen;
{
  pl_t *lock;
  comm_request_t rc;
  u_long	incarnation = *((u_long *) buf); /* iteration sent */
  char255	smsg;

  if (buflen != sizeof(u_long)) {
    sprintf(smsg, "midLockNewOwn: bad message size %lu (!= %lu)",
	    buflen, sizeof(u_long));
    StatusMsg(smsg);
    return(M_R_DROPPED);
  }

  /* convert lockname to a pointer to the lock structure */
  lock = (pl_t *) lockn;

  if (IS_DBG_LOCK_EXTR) {
    sprintf(smsg, "midLockNewOwn: lock %lu to %lu at incarnation %lu",
	    lockn, srchost, incarnation);
    StatusMsg(smsg);
  }

  mutex_lock(&(lock->m));
  /* assume lock->notime is only kept here, and is an incarnation number
    (I think that times should work as well, though, if they only come from
     the owners... the owners must have communicated with each other, so
     an increasing time must be meaningful) 21sep92 mjz
   */  
  if (lock->notime >= incarnation) { /* old */
    rc = M_R_DROPPED;		/* bah */
  } else {			/* new! */
    rc = M_R_DOAGAIN;		/* re-issue request */
    lock->bestguess = srchost;
    lock->notime = incarnation;
  }
  mutex_unlock(&(lock->m));
  return(rc);
} /* midLockNewOwn */


/*
 ************************************************************************
 *		m i d R e m o t e R e l e a s e R p l y			*
 *----------------------------------------------------------------------*
 * Accepts data segments from the "main" processor -- updating our idea *
 * of the global lock structure... 					*
 ************************************************************************
 */
midRemoteReleaseRply(buf,buflen)
     char *buf;
     u_long buflen;
{
  m_lock_remote_t *lrp;		/* our message */
  m_lock_remote_seg_t *lseg;	/* lock info segs w/in message */
  u_long number;		/* of segments in the buffer */
  u_long locksegno;
  u_long i;
  u_long s;
  u_long state;
  pl_t *lock;
  u_long tid;
  int default_model;
  m_lock_rem_amap_t *aseg;	/* addrmap segments for lock */
  s_addr_map_t *amap;		/* for running through actual segments */
  u_long anum;			/* # of segments */
  char255 smsg;
  char *mname;
  addrseg_t	*as;		/* for poking through the packet */
  
  /* gotta squirrel away prog & arg... hrrumph. */
  lrp = (m_lock_remote_t *) buf;
  /* let scheduler know what it will do */
  mid_rsched_wait_arg(lrp->func,lrp->arg);
  lseg = (m_lock_remote_seg_t *) (lrp+1); /* follows m_l_r_t entry */

  /* cache global variables before looping */
  default_model = midway_default_model;

  locksegno = 0;
  for (number = 0; number < lrp->nsegs; number++) {
    /* unpack segments */

    /* Allocate/Initialize lock descriptor for each lock */
    state = (lseg->type & L_BARRIER) ?  L_BARRIER | L_BAR_NOREG : 0;
    lock = mid_alloc_synch_object (lseg->name, state);
    ASSERT (lock == (pl_t *)lseg->name);

    /* Set initial values from message */
    lock->tcreate = lseg->tcreate;
    lock->bestguess = lseg->bestguess; /* need to change this too */
    lock->bgtime = 0;
    lock->held = MIDWAY_LOCKHELD_OTHER;
#ifdef PAGE_BASED
    lock->amap_data_size = lseg->anum_data;
#endif

    if (lseg->type == L_BARRIER) {
      lock->bar_curtime = lseg->bar_curtime; /* use manager's time */
    } else {
      lock->rdrs = 0;
      lock->incarnation = 0;
      lock->dataclock = MIN_CLOCK_VALUE;/*hard coded value for creation time */
    }


    /*
     * always attempt to allocate asegs; they might have to be split
     * across packets; they might also be duplicates, but then they will
     * be thrown away (methinks).
     */
    anum = lseg->anum;	/* # of lock addrmap segments following */
    lseg++;			/* point after lseg info */
    /* take care of any addrmap segments that may follow */
    if (anum > 0) {		/* gotta decode */
      aseg = (m_lock_rem_amap_t *) lseg; /* asegs follow lseg */
      while (anum > 0) {
	midway_bind_synch_object ((midway_lock_t)lock->lock, 
				  aseg->addr, aseg->len);

	aseg++;		/* point after this one (next, perhaps) */
	anum--;		/* one less to do */
      }
      if (IS_DBG_LOCK_FORK) {
	mid_addrmap_prt(lock->lock, lock->amap);	/* dbg */
      }
      lseg = (m_lock_remote_seg_t *) aseg; /* adjust lseg pointer */
    }
  } /* unpack segments in message */

  /* decode any data segments that may follow (for changed static data) */
  buflen -= (u_long) lseg - (u_long) buf;
  as = (addrseg_t *) lseg;
  while (buflen) {
    switch(as->segtype) {
    case DATASEG:
      mid_unpack_data((char **)&as, &buflen,  NULL,
		      "midRemoteReleaseReply", mid_MUD_static);
      break;
#ifdef PAGE_BASED
    case DIFFSEG:
      mid_paged_unpack_data(0,
			    (char **)&as, &buflen,  NULL,
			    "midRemoteReleaseReply", mid_MUD_static);
      break;
#endif
    default:
      sprintf(smsg,
	      "midRemoteReleaseReply: Unknown segment received (%x) -- giving up [buflen=%u]",
	      as->segtype, buflen);
      StatusMsg(smsg);
      break;
    }/*switch*/
  }/*while*/

} /* midRemoteReleaseRply */

/*
 ************************************************************************
 *		    m i d R e m o t e R e l e a s e			*
 *----------------------------------------------------------------------*
 * Gonna crank a "remote" thread.  Ensure its version of the global     *
 * lock structure is consistent to start with...			*
 ************************************************************************
 */
midRemoteRelease(host, thread, tid, cval, proc, arg)
     u_long	host;		/* to which we reply */
     u_long	thread;		/* thread on that host */
     u_long	tid;		/* ye olde transaction id */
     clock_t	cval;		/* clock value of last wakeup time */
     void	(*proc)();	/* the procedure to be cthread_fork()ed */
     u_long	arg;		/* the argument to that proc */
{
  pbuf pb;			/* buffer to dump our stuff */
  u_long pbn;			/* buffer # */
  u_long pbl;			/* packet buffer length */
  pl_t         *lockP;  	/* lock we are working on now */
  m_lock_remote_t *lrp;		/* our message */
  m_lock_remote_seg_t *lseg;	/* lock info segs w/in message */
  m_lock_rem_amap_t *aseg;	/* addrmap segments for lock */
  s_addr_map_t *amap;		/* for running through actual segments */
  u_long anum;			/* # of addrmap segs */
  u_long number;		/* of segments in the buffer */
  char255 smsg;
  char	*p;			/* for passing to midLockPackData */
  int		fits;		/* # of addrmap segments that fit in pkt */
  u_long	*anum_p;	/* ptr to right place for anum... */
  clock_t	maxtime;	/* max time for data to transmit */
  char          *amap_addr;	/* parm to delta_pack */
  long          amap_len;
  char          *data_addr;
  long          data_len;
  clock_t       data_clock;
  long		xferlen;

  INCR_STATS (midLockStats.xferReleases, 1);

  mutex_lock(lockmutex);	/* ensure table consistent */

  pbn = 0;			/* 1st buffer packet number */

  if (IS_DBG_LOCK_FORK) {
    StatusMsg("in midRemoteRelease");
  }
  pb = mid_pbuf_get();
  ASSERT (pb != NULL);
  lrp = (m_lock_remote_t *) pb;
  lrp->func = proc;
  lrp->arg = arg;
  pbl = sizeof(m_lock_remote_t);
  lseg = (m_lock_remote_seg_t *) (lrp + 1); /* point to after m_l_r */
  anum = 0;
  
  /* cycle through all locks (if any) */
  lockP = (pl_t *) vm_locktable_base;
  while (lockP < (pl_t *) vm_locktable_free) {
    ASSERT ((pl_t *)lockP->lock == lockP);

    number = 0;			/* no segments yet */
    while ((lockP < (pl_t *) vm_locktable_free) &&
	   (pbl < (mid_max_data_size - sizeof(m_lock_remote_seg_t)))) {

      /* room left in buffer */
      if ((GREATER_THAN_OR_EQUAL_CLOCK (lockP->tcreate, cval)) ||
	  /* must transfer if a barrier to get correct time. */
	  (lockP->state & L_BARRIER)) { 
	/* gotta transfer */
	/* first, see if will fit in buffer.  We assume that a pbuf */
	/* can hold at least the request, one lock, and it's addrmap */
	/* calculate number of addrmap segs */
	if (anum == 0) {	/* none left from last packet */
	  for (amap=lockP->amap; amap != NULL; amap = amap->next) {
	    anum++;
	  }
	}
	/* truly, should be (anum == 0) * sizeof(m_lock_rem_amap_t),
	   but it looks really awful. */
	if ((sizeof(m_lock_rem_amap_t)+sizeof(m_lock_remote_seg_t)+
	     pbl) > mid_max_data_size) { /* can't fit at least one */
	  if (number == 0) {	/* none fit in this segment */
	    /** this will now never happen 18nov92mjz (was anum x size) **/
	    sprintf(smsg,"Lock %u addrmap too big to fit in buffer--dying.",
		    lockP->lock);
	    StatusMsg(smsg);
	    ASSERT(number != 0); /* b a r f */
	  }
	  break;		/* send out partially filled buffer */
	}
	/* ok -- it fits.  Fill in lock_remote_seg, then rem_amap's */
	lseg->name = (u_long)lockP->lock;
	lseg->type = (lockP->state) & L_BARRIER;  /* L_BAR or nothing*/
	if (lseg->type == L_BARRIER)
	  lseg->bar_curtime = lockP->bar_curtime;
	else
	  lseg->bar_curtime = NULL_CLOCK_VALUE; /* ensure 0 if not used. */
	lseg->bestguess = lockP->bestguess;
	lseg->tcreate = lockP->tcreate;
	anum_p = &(lseg->anum);	/* remember this */
/*	lseg->anum = anum;*/	/* # addrmap entries following */
#ifdef PAGE_BASED
  	lseg->anum_data = lockP->amap_data_size;
#endif
	lseg++;
	number++;
	if (anum > 0) {
	  /* compute number that will fit of anum */
	  fits = (mid_max_data_size - pbl - sizeof(m_lock_remote_seg_t)) / 
	    sizeof(m_lock_rem_amap_t); /* # that would fit */
	  if (fits <= 0)
	    botch("fits <= 0");
	  if (fits > anum)
	    fits = anum;
	  anum -= fits;		/* ready for next time (if any) */
	  *anum_p = fits;	/* # that we put in this packet */
	  aseg = (m_lock_rem_amap_t *) lseg;
	  if (amap == 0)
	    amap = lockP->amap;
	  for(; (amap != 0) && (fits > 0); amap = amap->next, fits--) {
	    aseg->addr = amap->addr;
	    aseg->len = amap->len;
	    aseg++;		/* got to next one */
	  } /* for amap */
	  if ((anum > 0) && amap) { /* filled this one up */
	    pbl += (sizeof(m_lock_remote_seg_t) + 
		    (*anum_p)*sizeof(m_lock_rem_amap_t));
	    break;		/* send out partially filled buffer */
	  }
	  if ((anum > 0) || (amap != 0)) {
	    sprintf(smsg, "botched addrmap fill - anum (%d) and amap (0x%lx) not both zero",
		    anum, amap);
	    StatusMsg(smsg);
	    botch("addrmap fill"); /* spin so we can debug */
	  }
	  lseg = (m_lock_remote_seg_t *) aseg; /* ready for next one */
	} else {		/* anum == 0 */
	  *anum_p = 0;
	}
	pbl += (sizeof(m_lock_remote_seg_t) + 
		(*anum_p)*sizeof(m_lock_rem_amap_t));
      } /* gotta transfer */

      lockP++;
    } /* room left in buffer */

    lrp->nsegs = number;
    /*
     * ok: at end of loop... either filled buffer, or have packed all
     * locks into the buffer.  If the former, send off and start again...
     * if latter, hold off so we can send changed static data...
     */
    if (lockP < (pl_t *) vm_locktable_free) {	/* still more to do... send this one off */
      if (IS_DBG_LOCK_MSGS)

      /*							---DEBUG */
      {
	long *f = (long *) pb;
	sprintf(smsg, "%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx",
		*f, *(f+1), *(f+2), *(f+3), *(f+4), *(f+5), *(f+6),
		*(f+7), *(f+8), *(f+9), *(f+10),
		*(f+11), *(f+12));
	StatusMsg(smsg);
      }
      /*							---DEBUG */
      CommReply(M_SCHED_OP, host, thread, tid, 0 /*lock*/, 
		0 /*last*/,
		pbn, NULL_ADDR_VECTOR /* nacks */, pb, pbl, C_FREE);
      pbn++;			/* movin' on (maybe) */
      /* prep new buffer */
      pb = mid_pbuf_get();
      ASSERT (pb != NULL);
      lrp = (m_lock_remote_t *) pb;
      lrp->func = proc;
      lrp->arg = arg;
      pbl = sizeof(m_lock_remote_t);
      lseg = (m_lock_remote_seg_t *) (lrp + 1); /* point to after m_l_r */
    } /* more to do -- send this one off */
  } /* cycle through all locks */

  mutex_unlock(lockmutex);

  if (!IS_DBG_LOCK_NOSTATIC) {	/* add static gunk */
    if (cval == NULL_CLOCK_VALUE) {
      /*
       * memory is always initialized to 1, zero is for currently dirty data
       * (and addrmap_delta will fix that to be the current clock).
       * So, fudge a 0 to 1, otherwise *ALL* static data would be transferred!
       * NOTE NOTE NOTE: we will transfer both shared and unshared static data,
       * and even stuff that probably shouldn't be transferred (like stdio
       * iobuf[] stuff). --18jan92mjz
       */
      cval = MIN_CLOCK_VALUE;
    }
    amap = static_amap;
    p = (char *) lseg;
    maxtime = NULL_CLOCK_VALUE; amap_addr = NULL; amap_len = 0; 
    data_addr = NULL; data_len = 0; data_clock = NULL_CLOCK_VALUE;
    xferlen = 0;
#ifdef PAGE_BASED
    while(mid_paged_pack_data(0,
			      PAGE_PACK_DATA, 
			      FALSE, /* no need to mark static data as clean */
			      &amap, &amap_addr, &amap_len,
			      cval, 
			      p, &pbl,
			      &xferlen)) {
#else
    while(mid_pack_data(&amap, &amap_addr, &amap_len,
			cval, &maxtime, 
			p, &pbl,
			&data_addr, &data_len, &data_clock,
			&xferlen)) {
#endif
      /* returned 1, so filled current packet, and needs another */
      if (IS_DBG_LOCK_FORK)
      /*							---DEBUG */
      {
	long *f = (long *) pb;
	sprintf(smsg, "%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx",
		*f, *(f+1), *(f+2), *(f+3), *(f+4), *(f+5), *(f+6),
		*(f+7), *(f+8), *(f+9), *(f+10),
		*(f+11), *(f+12));
	StatusMsg(smsg);
      }
      /*							---DEBUG */
      CommReply(M_SCHED_OP, host, thread, tid, 0 /*lock*/, 
		0 /*last*/,
		pbn, NULL_ADDR_VECTOR /* nacks */, pb, pbl, C_FREE);
      pbn++;
      /* prep new buffer */
      pb = mid_pbuf_get();
      ASSERT (pb != NULL);
      lrp = (m_lock_remote_t *) pb;
      lrp->func = proc;
      lrp->arg = arg;
      lrp->nsegs = 0;
      pbl = sizeof(m_lock_remote_t);
      p = (char*) (lrp+1);	/* point after m l r */
    } /* while */

#ifdef PAGE_BASED_RC
    /*
     * Page-based version does not use dirtybits like above.
     * Explicitly send the slave all changes to the static data.
     */
    paged_flush_statics(host);

#endif /* PAGE_BASED */
  
  } /* add static gunk */

  /* send off final buffer */
  if (IS_DBG_LOCK_MSGS)
  /*								---DEBUG */
  {
    long *f = (long *) pb;
    sprintf(smsg, "%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx",
	    *f, *(f+1), *(f+2), *(f+3), *(f+4), *(f+5), *(f+6),
	    *(f+7), *(f+8), *(f+9), *(f+10),
	    *(f+11), *(f+12));
    StatusMsg(smsg);
  }
  /*								---DEBUG */
  CommReply(M_SCHED_OP, host, thread, tid, 0 /*lock*/, 
	    1 /*last*/,
	    pbn, NULL_ADDR_VECTOR /* nacks */, pb, pbl, C_FREE);
} /* midRemoteRelease */


/*
 ************************************************************************
 *	    m i d _ b i n d _ d e n s e _ d i r t y b i t               *
 *----------------------------------------------------------------------*
 * Allocate dense dirty bits for an object unaligned w/ cache lines     *
 ************************************************************************
 */
static void
mid_bind_dense_dirtybit 
  (
     region_info riP, 		/* ptr to region info */
     char *addr 		/* start of guarded data */
   )
{
  clock_t *ddbP, *infoP;
  u_long offset;
  u_long linesz_bytes;
  

  if (IS_DBG_LOCK_BIND) { 
    char255 smsg;
    sprintf(smsg,"Adding dense dirty bits for address 0x%x", addr);
    StatusMsg(smsg);
  }

  
  /* check if dense dirty bits were already allocated */
  offset = (int)(addr) - (int)riP->base;
  offset = (offset >> riP->line_size);	/* divide by line size */
  infoP = &riP->line_info[offset];
  if (LINE_UNALIGNED(*infoP))
    return;

  /* malloc space for dense dirty bits */
  linesz_bytes = RI_LINE_SIZE_TO_BYTES(riP->line_size);
  ddbP = (clock_t *) dbmalloc (linesz_bytes*sizeof(*ddbP), FALSE);
  ASSERT (ddbP != NULL);

  /* Set line info to point to dense dirty bits */
  offset = (clock_t *) ddbP - (clock_t *) vm_dense_dirty_base;
  *infoP = LINE_EXCEPTION_FLAG | LINE_UNALIGNED_FLAG | offset;

  INCR_STATS (midLockStats.denseUnaligned, 1);
  INCR_STATS (midLockStats.denseAlloced, linesz_bytes);

  return;

} /* mid_bind_dense_dirtybit */


/*
 ************************************************************************
 *	    m i d _ b i n d _ e c _ o b j e c t		                *
 *----------------------------------------------------------------------*
 * Reset line info from non-EC to EC for objects bound to synch objects *
 ************************************************************************
 */
static void
mid_bind_ec_object
  (
     region_info riP,		/* ptr to region info */
     char *addr,		/* start of guarded data */
     u_long len 		/* length (in bytes) of guarded data */
   )
{
  register clock_t *dbitP;
  register clock_t initialClockForEC;
  u_int line_size_in_bytes;
  u_int data_first_line;
  u_long offset;
  long s_len;
  char *line_addr;
  kern_return_t krc;
  u_long ddb_len, ddb_offset;
  register clock_t *ddbP, *end_ddbP;
  

  if (IS_DBG_PORT_PAGE) {
    char255 smsg;
    sprintf(smsg, "Setting memory at 0x%x for %lu bytes to EC", 
	    addr, len);
    StatusMsg (smsg);
  }

  /* sanity check */
  if (len == 0) return;


  /*
   * Binding data to a synch object implies the memory is EC.
   * Update all cache line tags to indicate the new model.
   *
   * NOTE: This resets the dirty bits for the bound data.
   * Application is NOT allowed to modify data before binding it.
   * WAS 29Oct92
   */
  
  /* locate line_info for starting address */
  offset = (int)(addr) - (int)riP->base;
  offset = (offset >> riP->line_size);	/* divide by line size */
  dbitP = &riP->line_info[offset];

  /* determine how much of object is in first cache line */  
  line_size_in_bytes = RI_LINE_SIZE_TO_BYTES(riP->line_size);
  line_addr = (char *)((int)riP->base + (offset << riP->line_size));
  data_first_line = line_size_in_bytes - (u_int)(addr - line_addr);
  
  /* Get clock value used to initialize new EC memory */
  initialClockForEC = midway_get_local_clock();

  /* loop setting line_info to EC */
  /* don't change if EC already */
  { register clock_t dbit_val;
    register u_long line_exception_and_enqueue_flags 
      = (LINE_EXCEPTION_FLAG|LINE_ENQUEUE_FLAG);
    register u_long line_exception_and_unaligned_flags 
      = (LINE_EXCEPTION_FLAG|LINE_UNALIGNED_FLAG);
    register u_long line_data_value = LINE_DATA_VALUE;
    
    while (s_len > 0) {
      
      dbit_val = *dbitP;
      if ((dbit_val & line_exception_and_enqueue_flags) == 
	  line_exception_and_enqueue_flags) {
	*dbitP = initialClockForEC;
	
      } else if ((dbit_val & line_exception_and_unaligned_flags) == 
		 line_exception_and_unaligned_flags) {
	
	/* handle dense dirty bits */
	if (s_len == len) {
	  ddb_len = (s_len < data_first_line) ? s_len : data_first_line;
	  ddb_offset = (dbit_val & line_data_value) + (line_size_in_bytes - data_first_line);
	} else {
	  ddb_len = (s_len < line_size_in_bytes) ? s_len : line_size_in_bytes;
	  ddb_offset = (dbit_val & line_data_value);
	}
	ddbP = &vm_dense_dirty_base[ ddb_offset ];
	end_ddbP = &vm_dense_dirty_base[ ddb_offset + ddb_len ];
	
	while (ddbP < end_ddbP) {
	  if ((*ddbP & line_exception_and_enqueue_flags) == 
	      line_exception_and_enqueue_flags)
	    *ddbP = initialClockForEC;
	  ddbP++;
	}
      }

      dbitP++;
      if (s_len == len)
	s_len -= data_first_line;
      else
	s_len -= line_size_in_bytes;
    }
  }

  return;
} /* midway_bind_ec_object */



/*
 ************************************************************************
 *	    m i d w a y _ b i n d _ s y n c h _ o b j e c t		*
 *----------------------------------------------------------------------*
 * Associate a sych object (lock) with shared data... (Ensure Shared?)  *
 ************************************************************************
 */
void
midway_bind_synch_object(obj,addr,len)
     midway_lock_t obj;		/* lock name */
     void *addr;		/* start of guarded data */
     u_long len;		/* length (in bytes) of guarded data */
{
  u_long i;
  pl_t *lock;
  clock_t *infoP;
  clock_t initialClockForEC;
  int lineSize;
  region_info riP;
  kern_return_t krc;
  char *line_addr;
  char *end_addr;
  char255 smsg;

  /* compute ending address for object */
  end_addr = ((char *) addr) + len;

  /* convert synch object name to a pointer to the lock structure */
  lock = (pl_t *) obj;

  
#ifdef DEBUG_USER
  if (((vm_offset_t) obj < vm_locktable_base) || 
      ((vm_offset_t) obj >= vm_locktable_end) ||
      ((vm_offset_t) obj != (vm_offset_t) lock->lock)) {
    sprintf(smsg,"midway_bind_synch_object: Invalid synch object %lu", obj);
    StatusMsg (smsg);
    return;
  }
  { char *start_addr = (char *) addr;
    if (((start_addr < (char *) FIRST_STATIC_SYM) ||
	 (end_addr > (char *) LAST_STATIC_SYM)) &&
	((start_addr < (char *) vm_shared_base) ||
	 (end_addr > (char *) vm_shared_end))) {
      sprintf(smsg,
	      "midway_bind_synch_object: Invalid shared address 0x%x for %d bytes",
	      addr, len);
      StatusMsg (smsg);
      return;
    }
  }
  if ((len <= 0) || (len > (vm_shared_end - vm_shared_base))) {
    sprintf(smsg, "midway_bind_synch_object: Invalid length %d", len);
    StatusMsg (smsg);
#if STOP_ON_ERROR
    botch("bind_synch_object -- bad length");
#endif
    return;
  }
#endif 

  /*
  ** Check if object crosses a region boundary.
  ** If so, split the object & make a recursive call to handle each part.
  */

#if !defined(PAGE_BASED) 

  /* locate/allocate the region info & line info for this address */
  riP = region_from_addr (addr);
  if ((riP == NULL) || (riP->line_info == NULL)) {
    if (IS_DBG_PORT_PAGE)
      StatusMsg ("Region not allocated");
      
    krc = get_shmalloc_info (addr);
    ASSERT (krc == KERN_SUCCESS);
    
    riP = region_from_addr (addr);
    if ((riP == NULL) || (riP->line_info == NULL)) {
      fprintf(stderr, "Can't find region info for 0x%x", addr);
#if STOP_ON_ERROR
      task_suspend(mach_task_self()); /* try to catch in act */
#endif
      return;
    }
  }

  if (end_addr > ((char *)(riP->base + REGION_SIZE))) {

    char   *new_addr = ((char *)(riP->base + REGION_SIZE));
    u_long  new_len = end_addr - new_addr; 
    
    /* make recursive call to handle data in next region */
    midway_bind_synch_object ((midway_lock_t)obj, new_addr, new_len);

    /* set length of object within this region */
    len -= new_len;
  }
#endif /* not PAGE_BASED */

  mutex_lock(&(lock->m));	/* is it SAFE? */

  mid_addrmap_add(&lock->amap, &lock->amap_tailP, &lock->amap_prevP, addr, len);
  if (IS_DBG_LOCK_BIND) {
    sprintf(smsg,"Added %lx for %u to object %u", addr, len, obj);
    StatusMsg(smsg);
    mid_addrmap_prt(lock->lock, lock->amap);
  }

  /* keep a sum of addrmap sizes */
  lock->amap_data_size += len;


#if defined(PAGE_BASED) 

  /* Unlock before returning */
  mutex_unlock(&(lock->m));

#else 

  /*
  ** Check if object crosses a cache line boundary.
  ** If so, use dense dirty bits.
  */
  if (midway_use_dense_dirtybits) {

    /* Check if starting address is on cache line boundary */
    line_addr = (char *)(((int)addr >> riP->line_size) << riP->line_size);
    if (line_addr != (char *) addr) {
      mid_bind_dense_dirtybit (riP, line_addr);
    }
    
    /* Check if ending address is on a cache line boundary */
    line_addr = (char *)((((int)addr + len) >> riP->line_size) << riP->line_size);
    if (line_addr != (char *)(addr + len)) {
      mid_bind_dense_dirtybit (riP, line_addr);
    }
  }


  /*
   * Binding data to a synch object implies the memory is EC.
   * Update all cache line tags to indicate the new model.
   */
  if (midway_default_model != MODEL_EC)
    mid_bind_ec_object (riP, addr, len);

  mutex_unlock(&(lock->m));

#if MIDWAY_WIRE_MEMORY
  { kern_return_t krc;
    extern task_t midway_task_self;
    extern mach_port_t midway_host_priv;
    krc = vm_wire(midway_host_priv, midway_task_self, addr,
		  len, VM_PROT_READ|VM_PROT_WRITE);
    if (krc != KERN_SUCCESS) {
      sprintf(smsg,"lock:vm_wire fails, addr 0x%lx len %u",
	      addr, len);
      StatusMsg(smsg);
      mach_error("lock:vm_wire",krc);
    }
  }
#endif
#endif /* not PAGE_BASED */

} /* midway_bind_synch_object */



/*
 ************************************************************************
 *	    m i d w a y _ r e b i n d _ s y n c h _ o b j e c t		*
 *----------------------------------------------------------------------*
 * Clear all assoications for a synchronization object                  *
 ************************************************************************
 */
void
midway_rebind_synch_object(obj)
     midway_lock_t obj;		/* lock name */
{
  u_long i;
  pl_t *lock;
  clock_t *infoP;
  kern_return_t krc;
  char255 smsg;

  /* convert synch object name to a pointer to the lock structure */
  lock = (pl_t *) obj;

#ifdef DEBUG_USER
  if (((vm_offset_t) obj < vm_locktable_base) || 
      ((vm_offset_t) obj >= vm_locktable_end) ||
      ((vm_offset_t) obj != (vm_offset_t) lock->lock)) {
    sprintf(smsg,"midway_rebind_synch_object: Invalid synch object %lu", obj);
    StatusMsg (smsg);
    return;
  }
#endif 

  /* acquire exclusive access to lock */
  mutex_lock(&(lock->m));

  if (IS_DBG_LOCK_BIND) {
    sprintf(smsg,"Removing all bindings object %u", obj);
    StatusMsg(smsg);
  }

  mid_addrmap_purge(&lock->amap, &lock->amap_tailP, &lock->amap_prevP);

  /* keep a sum of addrmap sizes */
  lock->amap_data_size = 0;

  /*
   * Note: once you have unbound, you have no idea what will be bound
   * next.  Therefore, your dataclock value is obsolete -- reset to one.
   *
   * (You also have a problem if you do an acquire without any bindings;
   * the data will come over, but since you don't know what you have,
   * you must ask for it all.  If you got the bindings first, you could
   * give a better answer (by scanning through the dbits).)
   *
   * NOTE: this reset would also have to happen if new data gets
   * bound with an older clock...
   *
   * This will also _not_ revert memory to any non-EC default.
   *
   */
  lock->dataclock = MIN_CLOCK_VALUE;

#ifdef PAGE_BASED
  /*
   * Purge saved diffs since they will no longer apply to new bindings
   * and bump incarnation number, so that xfer doesn't simply diff data.
   *
   *
   * Only increment incarn number if we are currently holding the lock.
   * Its necessary since the caller (qs) rebinds BEFORE acquiring the lock
   * to ensure the dataclock is reset. This is really a big time hack! WAS 8Sep94
   */
  if (lock->state & L_OWN) {
      lock->incarnation += 2;
      midLockIncrDiffs (lock);
      TRACE_3("REBI", lock, lock->incarnation);
  }
#endif

  /* Unlock before returning */
  mutex_unlock(&(lock->m));

} /* midway_rebind_synch_object */


/*
 ************************************************************************
 *	    m i d _ l o c k _ p r i n t _ s t a t s	                *
 *----------------------------------------------------------------------*
 * Print lock manager statistics                                        *
 ************************************************************************
 */
mid_lock_print_stats ()
{
#if !defined(STATS) && !defined(LTIMES)
  StatusMsg ("Not compiled to collect Lock Manager statistics");
#else
  int i, len, slen;
  u_long acqs, reqs;
  char255 smsg, tmp;
#define MAX_STR_LEN 80

  StatusMsg("-=-=- Lock Manager Statistics -=-=-");
  
  acqs = midLockStats.acquires;
  reqs = midLockStats.acquireROreqs + midLockStats.acquireRWreqs;

  sprintf(smsg,
  "Locks alloc=%lu acquired=%lu RO=%lu/RW=%lu  TotalTime=%lu ms Avg=%f ms ReqAvg=%f",
	  midLockStats.allocated, acqs, midLockStats.acquireRO,
	  midLockStats.acquireRW, midLockStats.acquireTotalTime, 
	  (acqs == 0) ? 0.0 : 
	    (double)(((double) midLockStats.acquireTotalTime)
	                / ((double) acqs)),
	  (reqs == 0) ? 0.0 : 
	    (double)(((double) midLockStats.acquireTotalTime)
	                / ((double) reqs)));
  StatusMsg (smsg);

  sprintf(smsg,
  "R/O: acq=%lu ownrw=%lu ownrwbusy=%lu ownro=%lu lockro=%lu cachedro=%lu reqs=%lu", 
	  midLockStats.acquireROacq, midLockStats.acquireROownrw, 
	  midLockStats.acquireROownrwbusy, midLockStats.acquireROownro, 
	  midLockStats.acquireROlockro, midLockStats.acquireROholdro, 
	  midLockStats.acquireROreqs);
  StatusMsg (smsg);

  sprintf(smsg,
  "R/W: acqrw=%lu ownrw=%lu ownrwbusy=%lu xfer=%lu incon=%lu reqs=%lu", 
	  midLockStats.acquireRWacqrw, midLockStats.acquireRWownrw, 
	  midLockStats.acquireRWownrwbusy, midLockStats.acquireRWxfer, 
	  midLockStats.acquireRWincon, midLockStats.acquireRWreqs);
  StatusMsg (smsg);

  sprintf(smsg,
  "R/W: acqro=%lu ownro=%lu ownroinval=%lu ownrordrs=%lu ownrobusy=%lu", 
	  midLockStats.acquireRWacqro, midLockStats.acquireRWownro, 
	  midLockStats.acquireRWownroinval, midLockStats.acquireRWownrordrs,
	  midLockStats.acquireRWownrobusy);
  StatusMsg (smsg);

  sprintf(smsg,
  "Release %lu NoRdr=%lu Busy=%lu Inval=%lu Nwriter=%lu Nrdrs=%lu Ewr=%lu Erd=%lu",
	  midLockStats.releases,  midLockStats.releaseNoRdrs, 
	  midLockStats.releaseBusy, midLockStats.releaseInval, 
	  midLockStats.releaseNwriter, midLockStats.releaseNreader, 
	  midLockStats.releaseEwriter, midLockStats.releaseEreader);
  StatusMsg (smsg);

  sprintf(smsg,
  "FastLocks: AcqRO %lu Rel %lu  AcqRW %lu Rel %lu  AcqCache %lu Rel %lu  ",
	  midFastLockStats.acqRO_count, midFastLockStats.relRO_count, 
	  midFastLockStats.acqRW_count, midFastLockStats.relRW_count,
	  midFastLockStats.acqCA_count, midFastLockStats.relCA_count);
  StatusMsg (smsg);

  sprintf(smsg,
  "FastLib:   AcqRO %lu Rel %lu  AcqRW %lu Rel %lu  AcqCache %lu Rel %lu  ",
	  midFastLockStats.acqRO_lib, midFastLockStats.relRO_lib, 
	  midFastLockStats.acqRW_lib, midFastLockStats.relRW_lib,
	  midFastLockStats.acqCA_lib, midFastLockStats.relCA_lib);
  StatusMsg (smsg);

  sprintf(smsg,
  "LockXfers= %lu data= %lu avg= %f tot= %lu /dirty= %f, RemRels= %lu RelData= %lu Up %lu, LostWrts= %lu Rdrs= %lu QW %lu QI %lu",
	  midLockStats.xferTotalLocks, midLockStats.xferTotalData, 
	  (midLockStats.xferTotalLocks == 0) ? 0.0 :
	    (double)(((double)midLockStats.xferTotalData)
		        / ((double) midLockStats.xferTotalLocks)),
	  midLockStats.xferTotalLockData,
	  (midLockStats.xferTotalLockData == 0) ? 0.0 :
	  (double)(((double)midLockStats.xferTotalData) /
		   ((double)midLockStats.xferTotalLockData)),
	  midLockStats.xferReleases, midLockStats.xferReleaseData,
	  midLockStats.xferUpgrade,
	  midLockStats.xferEwrtLost, midLockStats.xferErdrLost,
	  midLockStats.xferEwrQLost, midLockStats.xferEwrILost);
  StatusMsg (smsg);

  sprintf(smsg,
	  "Requests: fromSelf=%lu lostXferred %lu lostNewOwn %lu invalErdrLost %lu",
	  midLockStats.requestsFromSelf, midLockStats.extReqXferred,
	  midLockStats.extReqNewOwn, midLockStats.invErdrLost);
  StatusMsg(smsg);

  sprintf(smsg,
	  "Dbits: unaligned %lu dense allocated %lu", 
	  midLockStats.denseUnaligned, midLockStats.denseAlloced);
  StatusMsg(smsg);

  mid_lockaddr_print_stats ();
  mid_barrier_print_stats();

#ifdef LTIMES
  sprintf(smsg,
  "AcqTime %lu (ms)  Avg=%f (ms) ReqAvg=%f Inv=%lu",
	  midLockStats.acquireTotalTime, 
	  (acqs == 0) ? 0.0 :
	    (double)(((double) midLockStats.acquireTotalTime) 
	               / ((double) acqs)),
	  (reqs == 0) ? 0.0 :
	    (double)(((double) midLockStats.acquireTotalTime) 
	               / ((double) reqs)),
	  midLockStats.invalidTime);
  StatusMsg (smsg);

  reqs = midLockStats.extReqRequests;
  sprintf(smsg,
  "ExtReqs %lu TOD Time %lu max %lu avg %f (ms),  Invalid Count = %lu",
	  reqs, midLockStats.extReqTimeTotal, midLockStats.extReqTimeMax, 
	  (reqs == 0) ? 0.0 :
	  (double)(((double) midLockStats.extReqTimeTotal) 
		   / ((double) reqs)),
	  midLockStats.extReqInvalidTime);
  StatusMsg (smsg);

  sprintf(smsg,
  "ExtReq FRC Time %lu max %lu avg %f (us)  ReqTime %lu max %lu avg %f (us)",
	  midLockStats.extReqClockTotal / FRC_TICKS_PER_USEC, 
	  midLockStats.extReqClockMax / FRC_TICKS_PER_USEC,
	  (reqs == 0) ? 0.0 :
	    (double)(((double) midLockStats.extReqClockTotal/FRC_TICKS_PER_USEC) 
	               / ((double) reqs)),
	  midLockStats.extReqReqClockTotal / FRC_TICKS_PER_USEC, 
	  midLockStats.extReqReqClockMax / FRC_TICKS_PER_USEC,
	  (reqs == 0) ? 0.0 :
	    (double)(((double)midLockStats.extReqReqClockTotal/FRC_TICKS_PER_USEC) 
	               / ((double) reqs)));
  StatusMsg (smsg);

  sprintf(smsg, "TIME RO:");
  slen = strlen(smsg);
  for (i = 0; i < STAT_TIME_TABLE_SIZE+1; i++) {
    if (midLockStats.extReqROtimeP[i] == 0) continue;
    
    sprintf(tmp, " %d=%lu", i, midLockStats.extReqROtimeP[i]);
    len = strlen(tmp);
    if ((len + slen) > MAX_STR_LEN) {
      StatusMsg(smsg);
      sprintf(smsg, "     RO:");
      slen = strlen(smsg);
    }
    strcat (smsg, tmp);
    slen += len;
  }
  StatusMsg(smsg);

  sprintf(smsg, "TIME RW:");
  slen = strlen(smsg);
  for (i = 0; i < STAT_TIME_TABLE_SIZE+1; i++) {
    if (midLockStats.extReqRWtimeP[i] == 0) continue;
    
    sprintf(tmp, " %d=%lu", i, midLockStats.extReqRWtimeP[i]);
    len = strlen(tmp);
    if ((len + slen) > MAX_STR_LEN) {
      StatusMsg(smsg);
      sprintf(smsg, "     RW:");
      slen = strlen(smsg);
    }
    strcat (smsg, tmp);
    slen += len;
  }
  StatusMsg(smsg);
#endif /* LTIMES */

#endif /* STATS */
}

#ifdef PAGE_BASED
mid_paged_lock_stats ()
{
  int i, len, slen;
  char255 smsg, tmp;
    
  sprintf(smsg,
	  "PgLockXfer: %lu data %lu/%lu (%.3f) incarn ok %lu miss %lu max %lu siz %lu dsc %lu",
	  midLockStats.paged_count, midLockStats.paged_data, midLockStats.paged_data_max,
	  (midLockStats.paged_data_max == 0) ? 0.0 :
	    ((double)((double)midLockStats.paged_data)/((double)midLockStats.paged_data_max)),
	  midLockStats.paged_incarn_ok, midLockStats.paged_incarn_missing,
	  midLockStats.paged_incarn_max, midLockStats.paged_incarn_oversize,
	  midLockStats.paged_incarn_discarded);
  StatusMsg(smsg);

  sprintf(smsg, "Incarn:");
  slen = strlen(smsg);
  for (i = 0; i < STAT_PAGED_INCARN_TABLE_SIZE+1; i++) {
    if (midLockStats.paged_incarn[i] == 0) continue;
    
    sprintf(tmp, " %d=%lu", i, midLockStats.paged_incarn[i]);
    len = strlen(tmp);
    if ((len + slen) > MAX_STR_LEN) {
      StatusMsg(smsg);
      sprintf(smsg, "Incarn:");
      slen = strlen(smsg);
    }
    strcat (smsg, tmp);
    slen += len;
  }
  StatusMsg(smsg);

  sprintf(smsg,
	  "DiffReq: %lu %luK (noData=%lu Upgr=%lu Resnd %lu %luK) Pack (diff %lu %luK data %lu %luK)",
	  midLockStats.paged_diff, midLockStats.paged_diff_size/1024, 
	  midLockStats.paged_diff_nodata, midLockStats.paged_diff_nodiff, 
	  midLockStats.paged_diff_resend, midLockStats.paged_diff_resend_size/1024, 
	  midLockStats.paged_pack_diff, midLockStats.paged_diff_size/1024, 
	  midLockStats.paged_pack_data, midLockStats.paged_pack_data_size/1024);
  StatusMsg(smsg);

  sprintf(smsg,
	  "TotalDiffRequestSize: %lu max %lu MaxDiffIncarn %d",
	  midLockStats.paged_diff_buffer_used, midLockStats.paged_diff_buffer_max,
	  MAX_DIFF_INCARN);
  StatusMsg(smsg);
}
#endif /* PAGE_BASED */

mid_lock_reset_stats ()
{
#if defined(STATS) || defined(LTIMES)
  bzero (&midLockStats, sizeof(midLockStats));
  bzero (&midFastLockStats, sizeof(midFastLockStats));
#endif
  mid_barrier_reset_stats();
} /* mid_lock_reset_stats */


